Package: dnaEPICO
Title: dnaEPICO: Analysis Pipeline for Illumina DNA Methylation Array
        Data
Version: 0.99.34
Date: 2026-01-18
Authors@R: person("Paul", "Ruiz", role = c("aut", "cre"),
            email = "ruizpint@qut.edu.au",
            comment = c(ORCID = "0009-0007-6714-3566"))
Description: 
    A modular and reproducible workflow for preprocessing and analysing
    Illumina DNA methylation array data from the EPICv2, EPIC, and
    450K platforms. The package integrates quality control, probe
    filtering, cell-type deconvolution, phenotype preparation,
    CpG-wise generalised linear models and longitudinal mixed-effects
    models using 'lmerTest'/'lme4' or 'nlme', including optional 'nlme'
    residual correlation structures, and automated report generation.
    It builds on established Bioconductor
    infrastructure and wraps commonly used tools including 'minfi',
    'ENmix', and 'wateRmelon', with support for both local execution
    and high-performance computing workflows.
License: AGPL-3 + file LICENSE
URL: https://github.com/paulYRP/dnaEPICO
BugReports: https://github.com/paulYRP/dnaEPICO/issues
Depends: R (>= 4.4)
Imports: Biobase, RColorBrewer, data.table, ggplot2, methods, minfi,
        openxlsx, limma, wateRmelon, ENmix, ggrepel, glm2, parallel,
        lme4, lmerTest, nlme, quadprog, SummarizedExperiment, utils,
        stats
Suggests: IlluminaHumanMethylation450kmanifest,
        IlluminaHumanMethylation450kanno.ilmn12.hg19,
        IlluminaHumanMethylationEPICv2anno.20a1.hg38,
        IlluminaHumanMethylationEPICv2manifest,
        IlluminaHumanMethylationEPICmanifest,
        IlluminaHumanMethylationEPICanno.ilm10b4.hg19, minfiData,
        GenomicRanges, S4Vectors, remotes, BiocManager, BiocStyle,
        AnnotationHub, RefManageR, sessioninfo, devtools, testthat,
        tiff, magick, knitr, rmarkdown, covr, withr, tinytex,
        FlowSorted.Blood.EPIC, FlowSorted.Blood.450k,
        BeadSorted.Saliva.EPIC
VignetteBuilder: knitr
biocViews: Software, Preprocessing, MethylationArray, QualityControl,
        Epigenetics, Microarray, StatisticalMethod, ChipOnChip
Encoding: UTF-8
Roxygen: list(markdown = TRUE)
RoxygenNote: 7.3.3
Config/testthat/edition: 3
git_url: https://git.bioconductor.org/packages/dnaEPICO
git_branch: devel
git_last_commit: 24662e6
git_last_commit_date: 2026-07-19
Repository: Bioconductor 3.24
Date/Publication: 2026-07-19
NeedsCompilation: no
Packaged: 2026-07-19 21:35:09 UTC; biocbuild
Author: Paul Ruiz [aut, cre] (ORCID: <https://orcid.org/0009-0007-6714-3566>)
Maintainer: Paul Ruiz <ruizpint@qut.edu.au>

# Generated by roxygen2: do not edit by hand

S3method(print,dnaEPICO_dnamReport)
export(analyzeSvaEnmix)
export(annotateMethylationGLMSummaries)
export(annotateMethylationLMESummaries)
export(assessSamplesMinfiEwasWater)
export(buildClockFoundationInputsPreprocessingPheno)
export(buildRawMinfiEwasWater)
export(collectSignificantCpGsMethylationGLM)
export(collectSignificantInteractionsMethylationLME)
export(combineTimepointsPreprocessingPheno)
export(dnamReport)
export(estimateLC)
export(estimateLCMinfiEwasWater)
export(estimateSvaEnmixControls)
export(extractMake)
export(extractMetricsMinfiEwasWater)
export(filterProbesMinfiEwasWater)
export(filterSamplesMinfiEwasWater)
export(fitMethylationGLMModels)
export(fitMethylationLMEModels)
export(loadMetricsPreprocessingPheno)
export(mergeSvaTargetsEnmix)
export(methylationGLM)
export(methylationLME)
export(normalizeMinfiEwasWater)
export(plotAssessmentMinfiEwasWater)
export(plotCtrlMinfiEwasWater)
export(plotMethylationGLMDiagnostics)
export(plotMethylationGLMDistributions)
export(plotMethylationLMEDiagnostics)
export(plotMetricsMinfiEwasWater)
export(plotNormalizationMinfiEwasWater)
export(plotRawDensityMinfiEwasWater)
export(plotSexMinfiEwasWater)
export(plotSvaEnmix)
export(predictSexMinfiEwasWater)
export(prepareDnamReportInputs)
export(prepareMethylationGLMData)
export(prepareMethylationLMEData)
export(preprocessingMinfiEwasWater)
export(preprocessingPheno)
export(readPhenotypeTargets)
export(readRGSetMinfiEwasWater)
export(renderDnamReport)
export(splitTimepointsPreprocessingPheno)
export(summarizeMethylationGLMModels)
export(summarizeMethylationLMEModels)
export(summarizeTimepointsMethylationLME)
export(svaEnmix)
export(writeMethylationGLMOutputs)
export(writeMethylationLMEOutputs)
export(writePhenoLCMinfiEwasWater)
export(writePreprocessingPhenoOutputs)
export(writeSvaEnmixOutputs)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# dnaEPICO <img src="https://raw.githubusercontent.com/paulYRP/dnaEPICO/devel/inst/extdata/dnaEPICORM.svg" align="right" width="140" style="width:140px; height:auto; float:right; margin:0 -24px 16px 20px;" />

<!-- badges: start -->

[![GitHub
issues](https://img.shields.io/github/issues/paulYRP/dnaEPICO)](https://github.com/paulYRP/dnaEPICO/issues)
[![GitHub
pulls](https://img.shields.io/github/issues-pr/paulYRP/dnaEPICO)](https://github.com/paulYRP/dnaEPICO/pulls)
[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
[![codecov](https://codecov.io/gh/paulYRP/dnaEPICO/graph/badge.svg?token=5ZD6K3SMHB)](https://codecov.io/gh/paulYRP/dnaEPICO)

<!-- badges: end -->

<br clear="right" /> <br /> <br />

<p align="center">

<img src="https://raw.githubusercontent.com/paulYRP/dnaEPICO/devel/inst/extdata/dnaEPICO.gif" alt="dnaEPICO workflow preview" width="650" />
</p>

dnaEPICO provides a **modular and reproducible pipeline** for
preprocessing and statistically analysing Illumina DNA methylation array
data from the EPICv2, EPIC, and 450K platforms.

The package supports CpG-wise generalised linear models and longitudinal
mixed-effects models using lmerTest/lme4 or nlme. It also integrates
preprocessing, quality control, phenotype preparation, and automated
report generation for local and High-Performance Computing (HPC)
environments through a **GNU Make-based workflow**.

## Installation instructions

Get the latest stable `R` release from
[CRAN](http://cran.r-project.org/). Then install `dnaEPICO` from
[Bioconductor](http://bioconductor.org/) using the following code:

``` r
if (!requireNamespace("BiocManager", quietly = TRUE)) {
    install.packages("BiocManager")
}

BiocManager::install("dnaEPICO")
```

And the development version from
[GitHub](https://github.com/paulYRP/dnaEPICO) with:

``` r
BiocManager::install("paulYRP/dnaEPICO")
```

Alternatively, install directly from GitHub:

``` r
install.packages('devtools')
devtools::install_github('paulYRP/dnaEPICO')
```

## Articles:

- [**A Pilot Epigenome-Wide Study of Posttraumatic Growth: Identifying
  Novel Candidates for Future
  Research**](https://www.mdpi.com/2075-4655/9/4/39)

## Tutorials:

- [**DNA Methylation
  Tutorial**](https://paulYRP.github.io/2025-cpgpneurogenomics-workshop/tutorial.html)
- [**Getting
  Started**](https://github.com/paulYRP/dnaEPICO/wiki/Getting-Started)
- [**Requirements**](https://github.com/paulYRP/dnaEPICO/wiki/Requirements)

## Citation

Below is the citation output from using `citation('dnaEPICO')` in R.
Please run this yourself to check for any updates on how to cite
**dnaEPICO**.

``` r
print(citation("dnaEPICO"), bibtex = TRUE)
```

    ## dnaEPICO: Analysis Pipeline for Illumina DNA Methylation Array Data,
    ## generated as part of 10.3390/epigenomes9040039
    ## 
    ##   Ruiz P (2026). "dnaEPICO: Analysis Pipeline for Illumina DNA
    ##   Methylation Array Data." _Epigenomes_. doi:10.3390/epigenomes9040039
    ##   <https://doi.org/10.3390/epigenomes9040039>.
    ##   <https://bioconductor.org/packages/dnaEPICO>.
    ## 
    ## A BibTeX entry for LaTeX users is
    ## 
    ##   @Article{,
    ##     title = {dnaEPICO: Analysis Pipeline for Illumina DNA Methylation Array Data},
    ##     doi = {10.3390/epigenomes9040039},
    ##     journal = {Epigenomes},
    ##     author = {Paul Ruiz},
    ##     year = {2026},
    ##     url = {https://bioconductor.org/packages/dnaEPICO},
    ##   }

Please note that the `dnaEPICO` was only made possible thanks to many
other R and bioinformatics software authors, which are cited either in
the vignettes and/or the paper(s) describing this package.

## Code of Conduct

Please note that the `dnaEPICO` project is released with a [Contributor
Code of Conduct](http://bioconductor.org/about/code-of-conduct/). By
contributing to this project, you agree to abide by its terms.

## Development tools

- Continuous code testing is possible thanks to [GitHub
  actions](https://www.tidyverse.org/blog/2020/04/usethis-1-6-0/)
  through *[usethis](https://CRAN.R-project.org/package=usethis)*,
  *[remotes](https://CRAN.R-project.org/package=remotes)*, and
  *[rcmdcheck](https://CRAN.R-project.org/package=rcmdcheck)* customized
  to use [Bioconductor’s docker
  containers](https://www.bioconductor.org/help/docker/) and
  *[BiocCheck](https://bioconductor.org/packages/3.24/BiocCheck)*.
- Code coverage assessment is possible thanks to
  [codecov](https://codecov.io/gh) and
  *[covr](https://CRAN.R-project.org/package=covr)*.
- The [documentation website](http://paulYRP.github.io/dnaEPICO) is
  automatically updated thanks to
  *[pkgdown](https://CRAN.R-project.org/package=pkgdown)*.
- The code is styled automatically thanks to
  *[styler](https://CRAN.R-project.org/package=styler)*.
- The documentation is formatted thanks to
  *[devtools](https://CRAN.R-project.org/package=devtools)* and
  *[roxygen2](https://CRAN.R-project.org/package=roxygen2)*.

For more details, check the `dev` directory.

This package was developed using
*[biocthis](https://bioconductor.org/packages/3.24/biocthis)*.

#' Convenience preprocessing pipeline for Illumina methylation arrays
#'
#' Run the `dnaEPICO` preprocessing workflow as a convenience wrapper around the
#' smaller minfi/ENmix/wateRmelon helper functions in this package. The wrapper
#' returns a structured result object containing the in-memory outputs from each
#' stage. Files are written only when `saveOutputs = TRUE`.
#'
#' @param phenoFile Character. Path to the phenotype CSV file.
#' @param idatFolder Character. Directory containing the IDAT files.
#' @param outputLogs Character. Directory used for log files when `logs = TRUE`.
#' @param nSamples Integer or `NA`. Number of rows to keep from the phenotype
#'   table. Use `NA` to keep all samples.
#' @param SampleID Character. Name of the phenotype column containing sample
#'   identifiers.
#' @param arrayType Character. Illumina array identifier passed to
#'   `Biobase::annotation()`, for example `'IlluminaHumanMethylationEPICv2'`.
#' @param annotationVersion Character. Annotation build passed to
#'   `Biobase::annotation()`, for example `'20a1.hg38'` or `'ilmn12.hg19'`.
#' @param idatForce Logical. Passed to `minfi::read.metharray.exp()` when
#'   reading IDAT files. Use `TRUE` only after confirming that the selected IDAT
#'   files should be read together.
#' @param scriptLabel Character. Label used to name output folders when
#'   `saveOutputs = TRUE`.
#' @param baseDataFolder Character. Base directory used for saved `.RData`
#'   outputs when `saveOutputs = TRUE`.
#' @param figureBaseDir Character. Base directory used for saved figure outputs
#'   when `saveOutputs = TRUE`.
#' @param sepType Character or `NULL`. Field separator used in `phenoFile`. Use
#'   `NULL` for a comma-separated file, `'\\t'` for a tab-delimited file, or
#'   another separator accepted by `utils::read.csv()`.
#' @param tiffWidth Integer. Width of saved TIFF plots in pixels.
#' @param tiffHeight Integer. Height of saved TIFF plots in pixels.
#' @param tiffRes Integer. Resolution in DPI for saved TIFF plots.
#' @param qcCutoff Numeric. QC cutoff passed to `minfi::plotQC()`.
#' @param detPtype Character. Detection P-value mode passed to
#'   `minfi::detectionP()`. Common values in minfi workflows are `'m+u'` and
#'   `'negative'`. The default here is `'m+u'`.
#' @param detPThreshold Numeric. Samples with mean detection P value above this
#'   threshold are removed.
#' @param normMethods Character vector or semicolon-separated string of
#'   normalization methods. Supported values are `'adjustedfunnorm'`,
#'   `'funnorm'`, `'illumina'`, `'quantile'`, and `'swan'`.
#' @param sexColumn Character. Name of the phenotype column containing reported
#'   sex.
#' @param removeSexMismatch Logical. If `TRUE`, remove samples whose reported
#'   sex and methylation-predicted sex are both known and disagree. Missing or
#'   unknown sex values are retained. The default is `FALSE`.
#' @param pvalThreshold Numeric. Probe-level detection P-value threshold used in
#'   the probe filter.
#' @param chrToRemove Character vector or comma-separated string of chromosome
#'   names to remove, for example `'chrX,chrY'`.
#' @param snpsToRemove Character vector or comma-separated string of SNP probe
#'   types to remove, for example `'SBE,CpG'`.
#' @param mafThreshold Numeric. Minor allele frequency threshold passed to
#'   `minfi::dropLociWithSnps()`.
#' @param probeExclusionPath Character vector or semicolon-separated string of
#'   CSV files containing probe IDs to remove.
#' @param probeExclusionIdColumn Character or `NULL`. Column containing probe
#'   IDs. When `NULL` or `''`, each file is auto-detected using `ProbeID`,
#'   `TargetID`, `IlmnID`, or `Name`, then falling back to an unlabeled or
#'   probe-like first column.
#' @param useEpicV2Manifest Logical. If `TRUE`, also remove EPICv2 probes
#'   flagged in the Peters et al. expanded manifest from AnnotationHub resource
#'   `AH116484`.
#' @param epicV2ManifestFlags Named logical vector controlling which EPICv2
#'   manifest flags are removed. Defaults remove `CH_WGBS_evidence`, `CH_BLAT`,
#'   and `MissingPos`, but not `MismatchPos`.
#' @param crossReactivePath Deprecated alias for `probeExclusionPath`.
#' @param crossReactiveIdColumn Deprecated alias for `probeExclusionIdColumn`.
#' @param plotGroupVar Character. Phenotype column used for density and MDS
#'   grouping plots.
#' @param lcRef Character. Reference panel used for cell composition estimation.
#'   `'saliva'` and `'salivaEPIC'` use `estimateLC()`. Other values are passed
#'   to `ENmix::estimateCellProp()`.
#' @param phenoOrder Character vector or semicolon-separated string describing
#'   which phenotype columns should appear first in the merged `phenoLC` table.
#' @param lcPhenoDir Character. Directory used for the saved `phenoLC.csv` file
#'   when `saveOutputs = TRUE`.
#' @param display Logical. If `TRUE`, draw plots on the active graphics device.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#'   The default is `FALSE`.
#' @param logs Logical. If `TRUE`, write log messages to `outputLogs`. The
#'   default is `FALSE`.
#' @param saveOutputs Logical. If `TRUE`, write the `.RData`, figure, and
#'   `phenoLC.csv` outputs to disk. The default is `FALSE`.
#'
#' @return A list with class `'dnaEPICO_preprocessingMinfiEwasWater'`.
#' \describe{
#'   \item{targets}{Filtered phenotype table aligned to the retained samples.}
#'   \item{RGSet}{Filtered `RGChannelSet` used in downstream preprocessing and
#'   available for direct interactive inspection.}
#'   \item{rawData}{Object returned by [buildRawMinfiEwasWater()] containing the
#'   raw `MSet`, `RatioSet`, and genome-mapped object derived from `RGSet`.}
#'   \item{assessment}{Object returned by [assessSamplesMinfiEwasWater()]
#'   containing detection P values, QC summaries, and failed-sample tracking.}
#'   \item{sexData}{Object returned by [predictSexMinfiEwasWater()] containing
#'   predicted sex labels, mismatch summaries, plotting data, and the sample IDs
#'   removed when `removeSexMismatch = TRUE`.}
#'   \item{normData}{Object returned by [normalizeMinfiEwasWater()] containing
#' the requested normalized objects and metadata on the methods that were run.}
#'   \item{filterData}{Object returned by [filterProbesMinfiEwasWater()]
#'   containing the probe-filtered methylation objects at each filtering stage.}
#'   \item{metricsData}{Object returned by [extractMetricsMinfiEwasWater()]
#'   containing the beta-value, M-value, and copy-number matrices used by later
#'   workflow steps.}
#'   \item{lcData}{Object returned by [estimateLCMinfiEwasWater()] containing
#'   the estimated cell-type proportions and the phenotype table augmented with
#'   those proportions.}
#'   \item{logFile}{Resolved path to the optional log file, or `NULL` when
#'   logging was disabled.}
#' }
#' See [dnaEPICO_preprocessingMinfiEwasWater-class] for a class-level overview.
#'
#' @examples
#' if (requireNamespace("minfiData", quietly = TRUE) &&
#'     requireNamespace("IlluminaHumanMethylation450kmanifest", quietly = TRUE) &&
#'     requireNamespace("IlluminaHumanMethylation450kanno.ilmn12.hg19", quietly = TRUE)) {
#'     ex <- dnaEPICO:::exampleMinfiIdatInputsDnaEpico(n = 4)
#'     result <- preprocessingMinfiEwasWater(
#'         phenoFile = ex$phenoFile,
#'         idatFolder = ex$idatFolder,
#'         outputLogs = file.path(ex$tempDir, "logs"),
#'         nSamples = 4,
#'         SampleID = "Sample_Name",
#'         arrayType = ex$arrayType,
#'         annotationVersion = ex$annotationVersion,
#'         scriptLabel = "preprocessingMinfiEwasWater",
#'         baseDataFolder = file.path(ex$tempDir, "rData"),
#'         figureBaseDir = file.path(ex$tempDir, "figures"),
#'         detPThreshold = 1,
#'         normMethods = "quantile",
#'         sexColumn = "Sex",
#'         pvalThreshold = 1,
#'         chrToRemove = "",
#'         snpsToRemove = "SBE",
#'         mafThreshold = 1,
#'         probeExclusionPath = ex$probeExclusionPath,
#'         plotGroupVar = "Sex",
#'         lcRef = "saliva",
#'         phenoOrder = "Sample_Name;Sex;Basename;Sentrix_ID;Sentrix_Position",
#'         lcPhenoDir = ex$tempDir,
#'         saveOutputs = FALSE,
#'         verbose = FALSE,
#'         logs = FALSE
#'     )
#'     inherits(result, "dnaEPICO_preprocessingMinfiEwasWater")
#' }
#'
#' @seealso [dnaEPICO_preprocessingMinfiEwasWater-class]
#'
#' @export
preprocessingMinfiEwasWater <- function(
    phenoFile = "data/preprocessingMinfiEwasWater/pheno.csv",
    idatFolder = "data/preprocessingMinfiEwasWater/idats", outputLogs = "logs",
    nSamples = NA, SampleID = "Sample_Name",
        arrayType = "IlluminaHumanMethylationEPICv2",
    annotationVersion = "20a1.hg38", idatForce = FALSE,
        scriptLabel = "preprocessingMinfiEwasWater",
    baseDataFolder = "rData", figureBaseDir = "figures", sepType = NULL,
    tiffWidth = 2000, tiffHeight = 1000, tiffRes = 150, qcCutoff = 10.5,
    detPtype = "m+u", detPThreshold = 0.05, normMethods = "adjustedfunnorm",
    sexColumn = "Sex", removeSexMismatch = FALSE, pvalThreshold = 0.01,
    chrToRemove = "chrX,chrY", snpsToRemove = "SBE,CpG", mafThreshold = 0.1,
    probeExclusionPath = "data/preprocessingMinfiEwasWater/12864_2024_10027_MOESM8_ESM.csv",
    probeExclusionIdColumn = NULL, useEpicV2Manifest = FALSE,
    epicV2ManifestFlags = c(
        CH_WGBS_evidence = TRUE, CH_BLAT = TRUE,
        MissingPos = TRUE, MismatchPos = FALSE
    ), plotGroupVar = "Sex",
    lcRef = "salivaEPIC",
        phenoOrder = "Sample_Name;Timepoint;Sex;PredSex;Basename;Sentrix_ID;Sentrix_Position",
    lcPhenoDir = "data/preprocessingMinfiEwasWater", display = FALSE,
    verbose = FALSE, logs = FALSE, saveOutputs = FALSE,
        crossReactivePath = NULL,
    crossReactiveIdColumn = NULL
) {
    removeSexMismatch <- validateLogicalScalarDnaEpico(
        removeSexMismatch,
        "removeSexMismatch"
    )

    if (!is.null(crossReactivePath)) {
        probeExclusionPath <- crossReactivePath
    }
    if (!is.null(crossReactiveIdColumn)) {
        probeExclusionIdColumn <- crossReactiveIdColumn
    }

    log_file <- "log_preprocessingMinfiEwasWater.txt"
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = outputLogs,
        log_file = log_file
    )

    norm_method_list <- splitOptionMinfiEwasWater(normMethods,
        sep = ";"
    )

    emitLogMinfiEwasWater(
        c(
            paste(
                "==== Starting", scriptLabel,
                "===="
            ), paste("Start Time:               ", format(Sys.time())),
            paste("Log file path:            ",
                if (is.null(log_path)) "disabled" else log_path),
            paste("Phenotype file:           ", phenoFile), paste(
                "Separator type:           ",
                if (is.null(resolveSeparatorMinfiEwasWater(sepType))) {
                    "default (',')"
                } else {
                    sepType
                }
            ), paste("IDAT folder:              ", idatFolder),
            paste("nSamples limit:           ",
                if (is.na(nSamples)) "all" else nSamples),
            paste("SampleID column:          ", SampleID), paste(
                "Array type:               ",
                arrayType
            ), paste("Annotation version:       ", annotationVersion),
            paste("Force IDAT read:          ", idatForce), paste(
                "Base RData folder:        ",
                baseDataFolder
            ), paste(
                "Base Figure folder:       ",
                figureBaseDir
            ), paste(
                "TIFF size (w x h @ dpi):  ",
                tiffWidth, " x ", tiffHeight, " @ ", tiffRes
            ), paste(
                "QC cutoff (median):       ",
                qcCutoff
            ), paste("Detection P-value type:   ", detPtype),
            paste("Detection p-value threshold:", detPThreshold),
            paste("Normalization methods:    ", paste(norm_method_list,
                collapse = ", "
            )), paste(
                "Sex column:               ",
                sexColumn
            ), paste("Remove sex mismatches:    ", removeSexMismatch),
            paste("Plot grouping variable:   ", plotGroupVar),
                "Probe filtering:",
            paste("  P-value threshold:      ", pvalThreshold), paste(
                "  Chromosomes to remove:  ",
                chrToRemove
            ), paste(
                "  SNP positions filter:   ",
                snpsToRemove
            ), paste(
                "  MAF threshold:          ",
                mafThreshold
            ), paste(
                "  Probe-exclusion file(s):",
                probeExclusionPath
            ), paste(
                "  Probe-exclusion ID col: ",
                if (length(probeExclusionIdColumn) == 0L ||
                    is.na(probeExclusionIdColumn[[1L]]) ||
                    !nzchar(trimws(as.character(probeExclusionIdColumn[[1L]]))) ||
                    identical(
                        toupper(trimws(as.character(probeExclusionIdColumn[[1L]]))),
                        "NULL"
                    )) {
                    "auto"
                } else {
                    probeExclusionIdColumn
                }
            ), paste("  Use EPICv2 manifest:    ", useEpicV2Manifest),
            paste("  EPICv2 manifest flags:  ",
                paste(paste(names(normalizeEpicV2ManifestFlagsMinfiEwasWater(epicV2ManifestFlags)),
                normalizeEpicV2ManifestFlagsMinfiEwasWater(epicV2ManifestFlags),
                sep = "="
            ), collapse = ";")), "Cell composition (estimateLC):",
            paste("  Reference:              ", lcRef), paste(
                "  Leading pheno order:    ",
                phenoOrder
            ), paste(
                "Display plots:            ",
                display
            ), paste("Save outputs:             ", saveOutputs),
            "============================================================"
        ),
        verbose = verbose, log_path = log_path
    )

    withLoggedErrorsMinfiEwasWater(expr = {
        objectDir <- file.path(baseDataFolder, scriptLabel, "objects")
        normDir <- file.path(baseDataFolder, scriptLabel, "normObjects")
        metricsDir <- file.path(
            baseDataFolder, scriptLabel,
            "metrics"
        )
        filterDir <- file.path(baseDataFolder, scriptLabel, "filterObjects")
        qcDir <- file.path(baseDataFolder, scriptLabel, "qc")
        metricsFigDir <- file.path(
            figureBaseDir, scriptLabel,
            "metrics"
        )
        qcFigDir <- file.path(figureBaseDir, scriptLabel, "qc")
        enmixDir <- file.path(figureBaseDir, scriptLabel, "enmix")

        if (isTRUE(saveOutputs)) {
            dirs_to_create <- c(
                objectDir, normDir, metricsDir,
                filterDir, qcDir, metricsFigDir, qcFigDir, enmixDir,
                lcPhenoDir
            )
            for (dir_path in dirs_to_create) {
                dir.create(dir_path, recursive = TRUE, showWarnings = FALSE)
            }
        }

        targets <- readPhenotypeTargets(
            phenoFile = phenoFile,
            sepType = sepType, nSamples = nSamples, SampleID = SampleID,
            verbose = verbose, logs = logs, log_dir = outputLogs,
            log_file = log_file
        )

        RGSet <- readRGSetMinfiEwasWater(
            idatFolder = idatFolder,
            targets = targets, SampleID = SampleID, arrayType = arrayType,
            annotationVersion = annotationVersion, force = idatForce,
            verbose = verbose, logs = logs, log_dir = outputLogs,
            log_file = log_file
        )

        plotCtrlMinfiEwasWater(
            RGSet = RGSet, output_dir = if (isTRUE(saveOutputs)) {
                enmixDir
            } else {
                NULL
            }, verbose = verbose, logs = logs, log_dir = outputLogs,
            log_file = log_file
        )

        rawDataInitial <- buildRawMinfiEwasWater(
            RGSet = RGSet,
            verbose = verbose, logs = logs, log_dir = outputLogs,
            log_file = log_file
        )

        assessment <- assessSamplesMinfiEwasWater(
            rawData = rawDataInitial,
            RGSet = RGSet, qcCutoff = qcCutoff, detPtype = detPtype,
            detPThreshold = detPThreshold, verbose = verbose,
            logs = logs, log_dir = outputLogs, log_file = log_file
        )

        plotAssessmentMinfiEwasWater(
            assessment = assessment,
            plot = "qc", display = display, file = if (isTRUE(saveOutputs)) {
                file.path(qcFigDir, "quality_control(MSet).tiff")
            } else {
                NULL
            }, width = tiffWidth, height = tiffHeight, res = tiffRes,
            verbose = verbose, logs = logs, log_dir = outputLogs,
            log_file = log_file
        )

        plotAssessmentMinfiEwasWater(
            assessment = assessment,
            plot = "detection", display = display,
                file = if (isTRUE(saveOutputs)) {
                file.path(qcFigDir, "detection_pvalues(RGSet).tiff")
            } else {
                NULL
            }, width = tiffWidth, height = tiffHeight, res = tiffRes,
            verbose = verbose, logs = logs, log_dir = outputLogs,
            log_file = log_file
        )

        sampleData <- filterSamplesMinfiEwasWater(
            RGSet = RGSet,
            targets = targets, failedSamples = assessment$failedSamples,
            SampleID = SampleID, verbose = verbose, logs = logs,
            log_dir = outputLogs, log_file = log_file
        )

        rawData <- buildRawMinfiEwasWater(
            RGSet = sampleData$RGSet,
            verbose = verbose, logs = logs, log_dir = outputLogs,
            log_file = log_file
        )

        plotRawDensityMinfiEwasWater(
            rawData = rawData, targets = sampleData$targets,
            plotGroupVar = plotGroupVar, display = display,
                file = if (isTRUE(saveOutputs)) {
                file.path(qcFigDir, "densityBeta(MSet).tiff")
            } else {
                NULL
            }, width = tiffWidth, height = tiffHeight, res = tiffRes,
            verbose = verbose, logs = logs, log_dir = outputLogs,
            log_file = log_file
        )

        sexData <- predictSexMinfiEwasWater(
            rawData = rawData,
            targets = sampleData$targets, SampleID = SampleID,
            sexColumn = sexColumn, verbose = verbose, logs = logs,
            log_dir = outputLogs, log_file = log_file
        )

        mismatch_ids <- as.character(sexData$mismatches[[SampleID]])
        mismatch_ids <-
            mismatch_ids[!is.na(mismatch_ids) & nzchar(mismatch_ids)]
        sexData$removeSexMismatch <- removeSexMismatch
        sexData$removedSampleIDs <- character(0)

        if (isTRUE(removeSexMismatch) && length(mismatch_ids) >
            0L) {
            if (length(mismatch_ids) >= ncol(sampleData$RGSet)) {
                stop("Removing sex mismatches would leave no samples for normalization.",
                    call. = FALSE
                )
            }

            sampleData <- filterSamplesMinfiEwasWater(
                RGSet = sampleData$RGSet,
                targets = sexData$targets, failedSamples = mismatch_ids,
                SampleID = SampleID, verbose = verbose, logs = logs,
                log_dir = outputLogs, log_file = log_file
            )
            rawData <- buildRawMinfiEwasWater(
                RGSet = sampleData$RGSet,
                verbose = verbose, logs = logs, log_dir = outputLogs,
                log_file = log_file
            )
            sexData$removedSampleIDs <- mismatch_ids
        } else {
            sampleData$targets <- sexData$targets
        }

        sexData$retainedTargets <- sampleData$targets
        emitLogMinfiEwasWater(
            c(paste(
                "Remove sex mismatches:    ",
                removeSexMismatch
            ), paste(
                "Sex mismatches removed:   ",
                length(sexData$removedSampleIDs)
            ), if (length(sexData$removedSampleIDs) >
                0L) {
                paste("Removed sample IDs:        ",
                    paste(sexData$removedSampleIDs,
                    collapse = ", "
                ))
            } else {
                "Removed sample IDs:         none"
            }, "============================================================"),
            verbose = verbose, log_path = log_path
        )

        rgset_col_data <- SummarizedExperiment::colData(sampleData$RGSet)
        if (sexColumn %in% colnames(rgset_col_data)) {
            rgset_col_data[[sexColumn]] <- sampleData$targets[[sexColumn]]
        }
        rgset_col_data$PredSex <- sampleData$targets$PredSex
        rgset_col_data$SexMismatch <- sampleData$targets$SexMismatch
        SummarizedExperiment::colData(sampleData$RGSet) <- rgset_col_data

        plotSexMinfiEwasWater(
            sexData = sexData, type = "predicted",
            display = display, file = if (isTRUE(saveOutputs)) {
                file.path(qcFigDir, "sexPrediction(GSet).tiff")
            } else {
                NULL
            }, width = tiffWidth, height = tiffHeight, res = 70L,
            verbose = verbose, logs = logs, log_dir = outputLogs,
            log_file = log_file
        )

        plotSexMinfiEwasWater(
            sexData = sexData, type = "clinical",
            display = display, file = if (isTRUE(saveOutputs)) {
                file.path(qcFigDir, "sexClinical(GSet).tiff")
            } else {
                NULL
            }, width = tiffWidth, height = tiffHeight, res = 70L,
            verbose = verbose, logs = logs, log_dir = outputLogs,
            log_file = log_file
        )

        normData <- normalizeMinfiEwasWater(
            sampleData = sampleData,
            sexColumn = sexColumn, normMethods = normMethods,
            verbose = verbose, logs = logs, log_dir = outputLogs,
            log_file = log_file
        )

        plotNormalizationMinfiEwasWater(
            RGSet = sampleData$RGSet,
            normData = normData, targets = sampleData$targets,
            sexColumn = sexColumn, display = display,
                file = if (isTRUE(saveOutputs)) {
                file.path(qcFigDir, "sexComparison_RawNorm(MSetF).tiff")
            } else {
                NULL
            }, width = tiffWidth, height = tiffHeight, res = tiffRes,
            verbose = verbose, logs = logs, log_dir = outputLogs,
            log_file = log_file
        )

        filterData <- filterProbesMinfiEwasWater(
            normData = normData,
            RGSet = sampleData$RGSet, pvalThreshold = pvalThreshold,
            chrToRemove = chrToRemove, snpsToRemove = snpsToRemove,
            mafThreshold = mafThreshold,
                probeExclusionPath = probeExclusionPath,
            probeExclusionIdColumn = probeExclusionIdColumn,
            useEpicV2Manifest = useEpicV2Manifest,
                epicV2ManifestFlags = epicV2ManifestFlags,
            detPtype = detPtype, verbose = verbose, logs = logs,
            log_dir = outputLogs, log_file = log_file
        )

        metricsData <- extractMetricsMinfiEwasWater(
            filteredData = filterData,
            verbose = verbose, logs = logs, log_dir = outputLogs,
            log_file = log_file
        )

        plotMetricsMinfiEwasWater(
            metricsData = metricsData,
            targets = sampleData$targets, plot = "mds",
                plotGroupVar = plotGroupVar,
            sexColumn = sexColumn, display = display,
                file = if (isTRUE(saveOutputs)) {
                file.path(metricsFigDir,
                    "examineMDS_PostFilteringCrossRect(MSetF_Flt_Rxy_Ds_Rc).tiff")
            } else {
                NULL
            }, width = tiffWidth, height = tiffHeight, res = tiffRes,
            verbose = verbose, logs = logs, log_dir = outputLogs,
            log_file = log_file
        )

        plotMetricsMinfiEwasWater(
            metricsData = metricsData,
            targets = sampleData$targets, plot = "density",
                plotGroupVar = plotGroupVar,
            sexColumn = sexColumn, display = display,
                file = if (isTRUE(saveOutputs)) {
                file.path(metricsFigDir,
                    "densityBeta&M(MSetF_Flt_Rxy_Ds_Rc).tiff")
            } else {
                NULL
            }, width = tiffWidth, height = tiffHeight, res = tiffRes,
            verbose = verbose, logs = logs, log_dir = outputLogs,
            log_file = log_file
        )

        lcData <- estimateLCMinfiEwasWater(
            beta = metricsData$beta,
            targets = sampleData$targets, SampleID = SampleID,
            lcRef = lcRef, phenoOrder = phenoOrder, constrained = FALSE,
            verbose = verbose, logs = logs, log_dir = outputLogs,
            log_file = log_file
        )

        if (isTRUE(saveOutputs)) {
            saveNamedObjectMinfiEwasWater(
                sampleData$RGSet, "RGSet",
                file.path(objectDir, "RGSet.RData")
            )
            saveNamedObjectMinfiEwasWater(
                rawData$MSet, "MSet",
                file.path(objectDir, "MSet.RData")
            )
            saveNamedObjectMinfiEwasWater(
                rawData$RatioSet, "RatioSet",
                file.path(objectDir, "RatioSet.RData")
            )
            saveNamedObjectMinfiEwasWater(
                rawData$GSet, "GSet",
                file.path(objectDir, "GSet.RData")
            )
            saveNamedObjectMinfiEwasWater(
                assessment$detP, "detP",
                file.path(qcDir, "detP_RGSet.RData")
            )

            for (i in seq_along(normData$methods)) {
                saveNamedObjectMinfiEwasWater(
                    normData$normalized[[i]],
                    "normObj", file.path(normDir, paste0(
                        "norm_",
                        normData$methods[[i]], "_RGSet.RData"
                    ))
                )
            }

            saveNamedObjectMinfiEwasWater(
                filterData$detPFiltered,
                "MSetF_Flt", file.path(filterDir, "removProbes_MSetF_Flt.RData")
            )
            saveNamedObjectMinfiEwasWater(
                filterData$chrFiltered,
                "MSetF_Flt_Rxy", file.path(filterDir,
                    "removChrXY_MSetF_Flt_Rxy.RData")
            )
            saveNamedObjectMinfiEwasWater(
                filterData$snpFiltered,
                "MSetF_Flt_Rxy_Ds", file.path(filterDir, paste0(
                    "removSNPs_MAF",
                    mafThreshold, "_MSetF_Flt_Rxy_Ds.RData"
                ))
            )
            saveNamedObjectMinfiEwasWater(
                filterData$filtered,
                "MSetF_Flt_Rxy_Ds_Rc", file.path(filterDir,
                    "removCrossReactive_MSetF_Flt_Rxy_Ds_Rc.RData")
            )
            saveNamedObjectMinfiEwasWater(
                metricsData$m, "m",
                file.path(metricsDir, "m_NomFilt_MSetF_Flt_Rxy_Ds_Rc.RData")
            )
            saveNamedObjectMinfiEwasWater(
                metricsData$beta, "beta",
                file.path(metricsDir, "beta_NomFilt_MSetF_Flt_Rxy_Ds_Rc.RData")
            )
            saveNamedObjectMinfiEwasWater(
                metricsData$cn, "cn",
                file.path(metricsDir, "cn_NomFilt_MSetF_Flt_Rxy_Ds_Rc.RData")
            )

            writePhenoLCMinfiEwasWater(
                lcData = lcData, file = file.path(
                    lcPhenoDir,
                    "phenoLC.csv"
                ), verbose = verbose, logs = logs,
                log_dir = outputLogs, log_file = log_file
            )
        }

        emitLogMinfiEwasWater(
            c(
                paste(
                    "==== Finished", scriptLabel,
                    "===="
                ), paste("End Time:                 ", format(Sys.time())),
                "============================================================"
            ),
            verbose = verbose, log_path = log_path
        )

        structure(list(
            targets = sampleData$targets, RGSet = sampleData$RGSet,
            rawData = rawData, assessment = assessment, sexData = sexData,
            normData = normData, filterData = filterData,
                metricsData = metricsData,
            lcData = lcData, logFile = log_path
        ), class = "dnaEPICO_preprocessingMinfiEwasWater")
    }, log_path = log_path, verbose = verbose,
        context = "preprocessingMinfiEwasWater")
}

test_that("preprocessingMinfiEwasWater runs using staged minfiData IDATs for Bioconductor", {
    testthat::skip_on_ci()
    library(minfi)
    library(minfiData)
    library(IlluminaHumanMethylation450kmanifest)
    library(IlluminaHumanMethylation450kanno.ilmn12.hg19)
    testthat::skip_if_not_installed("tinytex")
    testthat::skip_if_not_installed("tiff")
    testthat::skip_if_not_installed("withr")

    # ------------------------------------------------------------------
    # 1. Locate minfiData example files
    # ------------------------------------------------------------------

    baseDirMinfi <- system.file("extdata", package = "minfiData")

    idatFiles <- list.files(
        baseDirMinfi,
        pattern = "\\.idat$",
        recursive = TRUE,
        full.names = TRUE
    )

    expect_true(length(idatFiles) > 0)

    # ------------------------------------------------------------------
    # 2. Locate dnaEPICO packaged extdata (probe-exclusion reference)
    # ------------------------------------------------------------------

    baseDirDnaEPICO <- system.file("extdata", package = "dnaEPICO")
    expect_true(dir.exists(baseDirDnaEPICO))

    probeExclusionPath <- file.path(
        baseDirDnaEPICO,
        "12864_2024_10027_MOESM8_ESM.csv"
    )

    expect_true(file.exists(probeExclusionPath))

    # ------------------------------------------------------------------
    # 3. Create temporary working directory
    # ------------------------------------------------------------------

    tmpDir <- withr::local_tempdir()

    tmpIdatDir <- file.path(tmpDir, "idats")
    dir.create(tmpIdatDir)

    file.copy(idatFiles, tmpIdatDir, overwrite = TRUE)

    # ------------------------------------------------------------------
    # 4. Create phenotype file (pheno_minfiData.csv)
    # ------------------------------------------------------------------

    targets <- read.csv(
        file.path(baseDirMinfi, "SampleSheet.csv"),
        stringsAsFactors = FALSE,
        skip = 7
    )

    targets$Basename <- paste0(
        targets$Sentrix_ID,
        "_",
        targets$Sentrix_Position
    )

    targets$Timepoint <- 1

    phenoFile <- file.path(tmpDir, "pheno_minfiData.csv")

    write.csv(
        targets,
        phenoFile,
        row.names = FALSE
    )

    expect_true(file.exists(phenoFile))

    # ------------------------------------------------------------------
    # 5. Set working directory safely
    # ------------------------------------------------------------------

    oldWd <- getwd()
    setwd(tmpDir)
    on.exit(setwd(oldWd), add = TRUE)
})

test_that("preprocessingMinfiEwasWater runs using staged minfiData IDATs for Github", {
    library(minfi)
    library(minfiData)
    library(IlluminaHumanMethylation450kmanifest)
    library(IlluminaHumanMethylation450kanno.ilmn12.hg19)
    testthat::skip_if_not_installed("tinytex")
    testthat::skip_if_not_installed("withr")

    # ------------------------------------------------------------------
    # 1. Locate minfiData example files
    # ------------------------------------------------------------------

    baseDirMinfi <- system.file("extdata", package = "minfiData")

    idatFiles <- list.files(
        baseDirMinfi,
        pattern = "\\.idat$",
        recursive = TRUE,
        full.names = TRUE
    )

    expect_true(length(idatFiles) > 0)

    # ------------------------------------------------------------------
    # 2. Locate dnaEPICO packaged extdata (probe-exclusion reference)
    # ------------------------------------------------------------------

    baseDirDnaEPICO <- system.file("extdata", package = "dnaEPICO")
    expect_true(dir.exists(baseDirDnaEPICO))

    probeExclusionPath <- file.path(
        baseDirDnaEPICO,
        "12864_2024_10027_MOESM8_ESM.csv"
    )

    expect_true(file.exists(probeExclusionPath))

    # ------------------------------------------------------------------
    # 3. Create temporary working directory
    # ------------------------------------------------------------------

    tmpDir <- withr::local_tempdir()

    tmpIdatDir <- file.path(tmpDir, "idats")
    dir.create(tmpIdatDir)

    file.copy(idatFiles, tmpIdatDir, overwrite = TRUE)

    # ------------------------------------------------------------------
    # 4. Create phenotype file (pheno_minfiData.csv)
    # ------------------------------------------------------------------

    targets <- read.csv(
        file.path(baseDirMinfi, "SampleSheet.csv"),
        stringsAsFactors = FALSE,
        skip = 7
    )

    targets$Basename <- paste0(
        targets$Sentrix_ID,
        "_",
        targets$Sentrix_Position
    )

    targets$Timepoint <- 1

    phenoFile <- file.path(tmpDir, "pheno_minfiData.csv")

    write.csv(
        targets,
        phenoFile,
        row.names = FALSE
    )

    expect_true(file.exists(phenoFile))

    # ------------------------------------------------------------------
    # 5. Set working directory safely
    # ------------------------------------------------------------------

    oldWd <- getwd()
    setwd(tmpDir)
    on.exit(setwd(oldWd), add = TRUE)

    # ------------------------------------------------------------------
    # 6. Run preprocessing pipeline
    # ------------------------------------------------------------------

    expect_error(
        dnaEPICO::preprocessingMinfiEwasWater(
            phenoFile = phenoFile,
            idatFolder = tmpIdatDir,
            nSamples = 6,
            SampleID = "Sample_Name",
            phenoOrder = "Sample_Name;sex;Basename;Sentrix_ID;Sentrix_Position",
            arrayType = "IlluminaHumanMethylation450k",
            annotationVersion = "ilmn12.hg19",
            sexColumn = "sex",
            plotGroupVar = "sex",
            lcRef = "saliva",
            outputLogs = file.path(tmpDir, "logs"),
            baseDataFolder = file.path(tmpDir, "rData"),
            probeExclusionPath = probeExclusionPath,
            lcPhenoDir = tmpDir,
            saveOutputs = TRUE
        ),
        NA
    )

    # ------------------------------------------------------------------
    # 7. Run sva
    # ------------------------------------------------------------------

    rgsetPath <- file.path(
        tmpDir,
        "rData",
        "preprocessingMinfiEwasWater",
        "objects",
        "RGSet.RData"
    )

    expect_true(file.exists(rgsetPath))

    expect_error(
        dnaEPICO::svaEnmix(
            phenoFile = phenoFile,
            rgsetData = rgsetPath,
            sepType = NULL,
            outputLogs = file.path(tmpDir, "logs"),
            nSamples = NA,
            SampleID = "Sample_Name",
            arrayType = "IlluminaHumanMethylation450k",
            annotationVersion = "ilmn12.hg19",
            SentrixIDColumn = "Sentrix_ID",
            SentrixPositionColumn = "Sentrix_Position",
            ctrlSvaPercVar = 0.90,
            ctrlSvaFlag = 1,
            scriptLabel = "svaEnmix",
            tiffWidth = 2000,
            tiffHeight = 1000,
            tiffRes = 150,
            dataBaseDir = file.path(tmpDir, "data"),
            figureBaseDir = file.path(tmpDir, "figures"),
            rBaseDir = file.path(tmpDir, "rData"),
            saveOutputs = TRUE
        ),
        NA
    )

    # ------------------------------------------------------------------
    # 8. Run preprocessing phenotype files
    # ------------------------------------------------------------------

    betaPath <- file.path(
        tmpDir,
        "rData",
        "preprocessingMinfiEwasWater",
        "metrics",
        "beta_NomFilt_MSetF_Flt_Rxy_Ds_Rc.RData"
    )

    mPath <- file.path(
        tmpDir,
        "rData",
        "preprocessingMinfiEwasWater",
        "metrics",
        "m_NomFilt_MSetF_Flt_Rxy_Ds_Rc.RData"
    )

    cnPath <- file.path(
        tmpDir,
        "rData",
        "preprocessingMinfiEwasWater",
        "metrics",
        "cn_NomFilt_MSetF_Flt_Rxy_Ds_Rc.RData"
    )

    expect_true(file.exists(betaPath))
    expect_true(file.exists(mPath))
    expect_true(file.exists(cnPath))

    expect_error(
        dnaEPICO::preprocessingPheno(
            phenoFile = phenoFile,
            sepType = NULL,
            betaPath = betaPath,
            mPath = mPath,
            cnPath = cnPath,
            SampleID = "Sample_Name",
            timeVar = "Timepoint",
            timepoints = "1",
            combineTimepoints = "1",
            sexColumn = "sex",
            outputPheno = file.path(tmpDir, "data", "preprocessingPheno"),
            outputRData = file.path(tmpDir, "rData", "preprocessingPheno", "metrics"),
            outputRDataMerge = file.path(tmpDir, "rData", "preprocessingPheno", "mergeData"),
            outputLogs = file.path(tmpDir, "logs"),
            outputDir = file.path(tmpDir, "data", "preprocessingPheno"),
            saveOutputs = TRUE
        ),
        NA
    )

    # ------------------------------------------------------------------
    # 9. Generate Report
    # ------------------------------------------------------------------

    report_result <- dnaEPICO::dnamReport(
        outputDir = file.path(tmpDir, "reports"),
        enmixTab = file.path(
            tmpDir, "figures", "preprocessingMinfiEwasWater", "enmix"
        ),
        qcTab = file.path(
            tmpDir, "figures", "preprocessingMinfiEwasWater", "qc"
        ),
        metricTab = file.path(
            tmpDir, "figures", "preprocessingMinfiEwasWater", "metrics"
        ),
        svaTab = file.path(
            tmpDir, "figures", "svaEnmix"
        ),
        verbose = FALSE,
        logs = TRUE,
        logTab = file.path(tmpDir, "logs")
    )
    if (identical(report_result$status, "failed")) {
        testthat::skip(
            paste(
                "Report rendering failed in the current local environment:",
                if (is.null(report_result$errorMessage)) "unknown render error" else report_result$errorMessage
            )
        )
    }
    expect_true(report_result$status %in% c("rendered", "skipped"))
    expect_true(file.exists(file.path(tmpDir, "logs", "log_dnamReport.txt")))
    if (identical(report_result$status, "rendered")) {
        expect_true(file.exists(report_result$outputFile))
    }

    # ------------------------------------------------------------------
    # 10. Extract Makefile
    # ------------------------------------------------------------------
    expect_error(
        makefile_path <- dnaEPICO::extractMake(
            destDir = tmpDir,
            overwrite = TRUE
        ),
        NA
    )
    expect_true(file.exists(makefile_path))

    # ------------------------------------------------------------------
    # 11. Run GLM
    # ------------------------------------------------------------------

    expect_error(
        dnaEPICO::methylationGLM(
            inputPheno = file.path(
                tmpDir, "rData", "preprocessingPheno", "mergeData",
                "phenoBT1.RData"
            ),
            outputLogs = file.path(tmpDir, "logs"),
            outputRData = file.path(tmpDir, "rData", "methylationGLM", "models"),
            outputPlots = file.path(tmpDir, "figures", "methylationGLM"),
            phenotypes = "status",
            covariates = "sex",
            factorVars = "sex,status",
            cpgLimit = 1,
            nCores = 8,
            summaryPval = 1,
            significantCpGDir = file.path(tmpDir, "results", "cpgs", "methylationGLM"),
            summaryTxtDir = file.path(
                tmpDir, "results", "summary", "methylationGLM",
                "glm"
            ),
            annotationCols = "Name,chr,pos,UCSC_RefGene_Group,UCSC_RefGene_Name,Relation_to_Island",
            annotatedGLMOut = file.path(tmpDir, "data", "methylationGLM"),
            annotationPackage = "IlluminaHumanMethylation450kanno.ilmn12.hg19",
            chunkSize = 1,
            significantCpGPval = 1,
            logs = TRUE,
            saveOutputs = TRUE
        ),
        NA
    )

    # ------------------------------------------------------------------
    # 12. Run LME
    # ------------------------------------------------------------------

    expect_error(
        dnaEPICO::methylationLME(
            inputPheno = file.path(tmpDir, "rData", "preprocessingPheno", "mergeData", "phenoBT1.RData"),
            outputLogs = file.path(tmpDir, "logs"),
            outputRData = file.path(tmpDir, "rData", "methylationLME", "models"),
            outputPlots = file.path(tmpDir, "figures", "methylationLME"),
            phenotypes = "status",
            personVar = "person",
            covariates = "sex",
            factorVars = "sex,status",
            cpgLimit = 1,
            nCores = 8,
            summaryPval = 1,
            significantInteractionDir = file.path(tmpDir, "results", "cpgs", "methylationLME"),
            summaryTxtDir = file.path(tmpDir, "results", "summary", "methylationLME"),
            annotatedLMEOut = file.path(tmpDir, "data", "methylationLME"),
            annotationPackage = "IlluminaHumanMethylation450kanno.ilmn12.hg19",
            chunkSize = 1,
            significantInteractionPval = 1,
            saveSignificantInteractions = FALSE,
            annotationCols = "Name,chr,pos,UCSC_RefGene_Group,UCSC_RefGene_Name,Relation_to_Island",
            logs = TRUE,
            saveOutputs = TRUE
        ),
        NA
    )
    expect_false(dir.exists(file.path(
        tmpDir, "data", "methylationGLM", "report-assets"
    )))
    expect_false(dir.exists(file.path(
        tmpDir, "data", "methylationLME", "report-assets"
    )))
})

#' Estimate saliva cell proportions from DNA methylation beta values
#'
#' Estimate cell-type proportions with the saliva reference panels bundled in
#' `dnaEPICO`. This function keeps the original `estimateLC()` interface used by
#' the package while using the internal reference files distributed in
#' `inst/extdata`.
#'
#' @param meth Numeric matrix of beta values with CpGs in rows and samples in
#'   columns. Row names must contain probe identifiers compatible with the
#'   selected reference.
#' @param ref Character. Reference panel name. Supported values are `'saliva'`
#'   and `'salivaEPIC'`.
#' @param constrained Logical. If `TRUE`, estimated cell proportions are
#'   constrained to sum to one.
#'
#' @return A `data.table` with one row per sample and one column per estimated
#'   cell type.
#'
#' @references
#' Murat K, et al. Ewastools: Infinium Human Methylation BeadChip pipeline for
#' population epigenetics integrated into Galaxy. *GigaScience*.
#' 2020;9(5):giaa049.
#' Houseman EA, Accomando WP, Koestler DC, et al. DNA methylation arrays as
#' surrogate measures of cell mixture distribution. *BMC Bioinformatics*.
#' 2012;13:86.
#' Reinius LE, Acevedo N, Joerink M, et al. Differential DNA methylation in
#' purified human blood cells: implications for cell lineage and studies on
#' disease susceptibility. *PLoS One*. 2012;7(7):e41361.
#' Bakulski KM, Feinberg JI, Andrews SV, et al. DNA methylation of cord blood
#' cell types: applications for mixed cell birth studies. *Epigenetics*.
#' 2016;11(5):354-362.
#' de Goede OM, Razzaghian HR, Price EM, et al. Nucleated red blood cells
#' impact DNA methylation and expression analyses of cord blood hematopoietic
#' cells. *Clinical Epigenetics*. 2015;7:95.
#' Gervin K, Salas LA, Bakulski KM, et al. Cell type specific DNA methylation
#' in cord blood: a 450K reference data set and cell count-based validation of
#' estimated cell type composition. *Epigenetics*. 2016;11(9):690-698.
#' Gervin K, Salas LA, Bakulski KM, et al. Systematic evaluation and validation
#' of reference and library selection methods for deconvolution of cord blood
#' DNA methylation data. *bioRxiv*. 2019. doi:10.1101/570457.
#' Salas LA, Koestler DC, Butler RA, et al. An optimized library for
#' reference-based deconvolution of whole-blood biospecimens assayed using the
#' Illumina HumanMethylationEPIC BeadArray. *Genome Biology*. 2018;19:64.
#' Heiss JA, Just AC, Brenner H. Training a model for estimating leukocyte
#' composition using whole-blood DNA methylation and cell counts as reference.
#' *Epigenomics*. 2017;9(1):13-20.
#' Middleton LYM, Dou J, Mill J, et al. Saliva cell type DNA methylation
#' reference panel for epidemiology studies in children. 2020.

#' @examples
#' ref_file <- system.file("extdata", "saliva.txt", package = "dnaEPICO")
#' ref_panel <- as.matrix(utils::read.table(ref_file))
#' meth <- ref_panel[1:20, , drop = FALSE]
#' colnames(meth) <- c("sample1", "sample2")
#' estimateLC(
#'     meth = meth,
#'     ref = "saliva",
#'     constrained = FALSE
#' )
#'
#' @export
estimateLC <- function(meth, ref, constrained = FALSE) {
    constrained <- validateLogicalScalarDnaEpico(
        constrained,
        "constrained"
    )
    validateCellCompositionBetaStructureDnaEpico(
        beta = meth,
        requireSampleNames = FALSE, objectName = "meth"
    )
    validateCellCompositionReferenceDnaEpico(ref)
    prepared <- prepareBetaForCellCompositionDnaEpico(meth)
    if (prepared$boundaries$NaN.Converted > 0L) {
        warning(prepared$boundaries$NaN.Converted,
            " NaN value(s) were converted to NA before cell-composition estimation.",
            call. = FALSE
        )
    }
    if (nrow(prepared$invalidCpGs) > 0L) {
        warning(nrow(prepared$invalidCpGs),
            " invalid CpG(s) were excluded from cell-composition estimation.",
            call. = FALSE
        )
    }
    estimate <- estimateLCFromValidatedBetaDnaEpico(
        meth = prepared$beta,
        ref = ref, constrained = constrained
    )
    attr(estimate, "methylationBoundaries") <- prepared$boundaries
    attr(estimate, "methylationIssues") <- prepared$issues
    attr(estimate, "invalidCpGs") <- prepared$invalidCpGs
    estimate
}

validateCellCompositionBetaStructureDnaEpico <- function(
    beta,
    requireSampleNames = TRUE, objectName = "beta"
) {
    if (!is.matrix(beta) || !is.numeric(beta) || nrow(beta) ==
        0L || ncol(beta) == 0L) {
        stop(objectName, " must be a non-empty numeric matrix.",
            call. = FALSE
        )
    }
    probe_ids <- rownames(beta)
    if (is.null(probe_ids)) {
        stop(objectName, " must have probe identifiers in row names.",
            call. = FALSE
        )
    }
    validateMethylationProbeIdentifiersDnaEpico(probe_ids, paste(
        objectName,
        "row names"
    ))
    if (isTRUE(requireSampleNames) && is.null(colnames(beta))) {
        stop(objectName, " must have sample names in columns.",
            call. = FALSE
        )
    }
    if (!is.null(colnames(beta))) {
        validateSampleIdentifiersDnaEpico(colnames(beta), paste(
            objectName,
            "sample identifiers"
        ))
    }
    invisible(TRUE)
}

validateCellCompositionReferenceDnaEpico <- function(ref) {
    if (!is.character(ref) || length(ref) != 1L || is.na(ref) ||
        !(ref %in% c("saliva", "salivaEPIC"))) {
        stop("Unsupported ref. Supported values are 'saliva' and 'salivaEPIC'.",
            call. = FALSE
        )
    }
    invisible(ref)
}

prepareBetaForCellCompositionDnaEpico <- function(beta) {
    validation <- inspectMethylationMatrixDnaEpico(
        values = beta,
        methylationScale = "beta"
    )
    invalid_ids <- as.character(validation$invalidCpGs$CpG)
    keep <- !(rownames(validation$values) %in% invalid_ids)
    list(
        beta = validation$values[keep, , drop = FALSE],
            boundaries = validation$boundaries,
        issues = validation$issues, invalidCpGs = validation$invalidCpGs
    )
}

estimateLCFromValidatedBetaDnaEpico <- function(meth, ref, constrained) {
    ref_file <- system.file("extdata", paste0(ref, ".txt"),
        package = "dnaEPICO")
    if (!nzchar(ref_file)) {
        stop("The bundled reference file was not found for ref '",
            ref, "'.",
            call. = FALSE
        )
    }

    J <- ncol(meth)

    coefs <- utils::read.table(ref_file)

    coefs <- as.matrix(coefs)
    n_celltypes <- ncol(coefs)


    detect_probe_id_type <- function(x, label) {
        legacy_regex <-
            "^cg[0-9]{8}$|^rs[0-9]+$|^ch\\.[[:alnum:]_]+\\.\\d+[FR]$"
        epicv2_regex <- "^(cg|ch|rs|nv).+_[TB][CO][0-9]+$"

        if (all(grepl(legacy_regex, x))) {
            return("legacy")
        }

        if (all(grepl(epicv2_regex, x))) {
            return("epicv2")
        }

        stop(label, " is (partly) invalid!")
    }

    query_type <- detect_probe_id_type(rownames(coefs), "Query")
    row_type <- detect_probe_id_type(rownames(meth), "Input row names")

    if (identical(query_type, row_type)) {
        markers <- match(rownames(coefs), rownames(meth))
        retained_markers <- !is.na(markers)
        coefs <- coefs[retained_markers, , drop = FALSE]
        marker_meth <- meth[markers[retained_markers], , drop = FALSE]
    } else if (query_type == "legacy" && row_type == "epicv2") {
        row_loci <- sub("_[TB][CO]\\d+$", "", rownames(meth))
        marker_groups <- lapply(rownames(coefs), function(marker) {
            which(row_loci ==
                marker)
        })
        retained_markers <- lengths(marker_groups) > 0L
        coefs <- coefs[retained_markers, , drop = FALSE]
        marker_groups <- marker_groups[retained_markers]
        marker_meth <- t(vapply(marker_groups, function(indices) {
            locus_values <- meth[indices, , drop = FALSE]
            observed <- colSums(!is.na(locus_values))
            locus_means <- rep(NA_real_, J)
            has_values <- observed > 0L
            locus_means[has_values] <- colSums(locus_values[,
                has_values,
                drop = FALSE
            ], na.rm = TRUE) / observed[has_values]
            locus_means
        }, numeric(J)))
    } else if (query_type == "epicv2" && row_type == "legacy") {
        stop("Query contains EPICv2 probe IDs but dataset is of legacy type")
    } else {
        stop("Unsupported probe identifier combination.", call. = FALSE)
    }

    if (nrow(marker_meth) < n_celltypes || qr(coefs)$rank < n_celltypes) {
        stop("Too few independent reference markers overlap meth for cell ",
            "composition estimation.",
            call. = FALSE
        )
    }

    EST <- vapply(seq_len(J), function(j) {
        tmp <- marker_meth[, j]
        i <- !is.na(tmp)

        if (sum(i) < n_celltypes || qr(coefs[i, , drop = FALSE])$rank <
            n_celltypes) {
            stop("Sample ", if (is.null(colnames(meth))) {
                j
            } else {
                colnames(meth)[[j]]
            }, " has too few independent markers for cell composition estimation.",
            call. = FALSE
            )
        }

        if (!isTRUE(constrained)) {
            quadprog::solve.QP(
                t(coefs[i, ]) %*% coefs[i, ],
                t(coefs[i, ]) %*% tmp[i], diag(n_celltypes),
                rep(0, n_celltypes)
            )$solution
        } else {
            quadprog::solve.QP(t(coefs[i, ]) %*% coefs[i, ],
                t(coefs[i, ]) %*% tmp[i], cbind(1, diag(n_celltypes)),
                c(1, rep(0, n_celltypes)),
                meq = 1
            )$solution
        }
    }, FUN.VALUE = numeric(n_celltypes))

    EST <- t(EST)
    colnames(EST) <- colnames(coefs)
    EST <- data.table::data.table(EST)

    EST
}

build_estimate_lc_example_data <- function() {
    ref_file <- system.file("extdata", "saliva.txt", package = "dnaEPICO")
    beta <- as.matrix(utils::read.table(ref_file))[1:20, , drop = FALSE]
    colnames(beta) <- c("sample1", "sample2")

    targets <- data.frame(
        Sample_Name = colnames(beta),
        Timepoint = c("T1", "T2"),
        stringsAsFactors = FALSE
    )

    list(beta = beta, targets = targets)
}

test_that("estimateLC returns the expected cell-type columns", {
    example_data <- build_estimate_lc_example_data()

    result <- estimateLC(
        meth = example_data$beta,
        ref = "saliva",
        constrained = FALSE
    )

    expect_s3_class(result, "data.table")
    expect_identical(
        colnames(result),
        c("Leukocytes", "Epithelial.cells")
    )
    expect_equal(as.numeric(result[1, ]), c(1, 0), tolerance = 1e-8)
    expect_equal(as.numeric(result[2, ]), c(0, 1), tolerance = 1e-8)
})

test_that("estimateLC averages duplicate EPICv2 loci instead of choosing one", {
    example_data <- build_estimate_lc_example_data()
    epic <- example_data$beta
    rownames(epic) <- paste0(rownames(epic), "_TC11")
    duplicate <- epic[1, , drop = FALSE]
    rownames(duplicate) <- sub("_TC11$", "_BC11", rownames(duplicate))
    duplicate[,] <- c(0.2, 0.8)
    epic_with_duplicate <- rbind(epic, duplicate)

    collapsed <- epic
    collapsed[1, ] <- colMeans(epic_with_duplicate[c(1, nrow(epic_with_duplicate)), , drop = FALSE])

    expect_equal(
        as.matrix(estimateLC(epic_with_duplicate, ref = "saliva", constrained = FALSE)),
        as.matrix(estimateLC(collapsed, ref = "saliva", constrained = FALSE)),
        tolerance = 1e-10
    )
})

test_that("estimateLC rejects duplicate exact probe identifiers", {
    example_data <- build_estimate_lc_example_data()
    duplicated_beta <- rbind(example_data$beta, example_data$beta[1, , drop = FALSE])

    expect_error(
        estimateLC(duplicated_beta, ref = "saliva", constrained = FALSE),
        "duplicate CpG identifiers"
    )
})

test_that("estimateLCMinfiEwasWater merges and orders phenoLC columns", {
    example_data <- build_estimate_lc_example_data()

    result <- estimateLCMinfiEwasWater(
        beta = example_data$beta,
        targets = example_data$targets,
        lcRef = "saliva",
        phenoOrder = "Sample_Name;Timepoint"
    )

    expect_s3_class(result, "dnaEPICO_minfiEwasWater_lc")
    expect_identical(
        colnames(result$phenoLC)[1:4],
        c("Sample_Name", "Timepoint", "Leukocytes", "Epithelial.cells")
    )
    expect_equal(result$phenoLC$Leukocytes, c(1, 0), tolerance = 1e-8)
    expect_equal(result$phenoLC$Epithelial.cells, c(0, 1), tolerance = 1e-8)
})

test_that("estimateLCMinfiEwasWater validates reference and beta inputs", {
    example_data <- build_estimate_lc_example_data()
    targets <- data.frame(
        Sample_Name = colnames(example_data$beta),
        stringsAsFactors = FALSE
    )

    expect_error(
        estimateLCMinfiEwasWater(
            beta = example_data$beta,
            targets = targets,
            lcRef = c("saliva", "salivaEPIC")
        ),
        "one non-empty"
    )
    expect_error(
        estimateLCMinfiEwasWater(
            beta = example_data$beta,
            targets = targets,
            lcRef = "saliva",
            constrained = "FALSE"
        ),
        "TRUE or FALSE"
    )

    invalid_beta <- example_data$beta
    invalid_beta[1, 1] <- Inf
    expect_warning(
        result <- estimateLCMinfiEwasWater(
            beta = invalid_beta,
            targets = targets,
            lcRef = "saliva"
        ),
        "invalid CpG"
    )
    expect_equal(nrow(result$invalidCpGs), 1L)
    expect_identical(result$invalidCpGs$CpG, rownames(invalid_beta)[1L])
    expect_equal(result$methylationBoundaries$Positive.Inf, 1L)
    expect_equal(nrow(result$lc), ncol(invalid_beta))
})

test_that("estimateLCMinfiEwasWater converts NaN to NA and reports it", {
    example_data <- build_estimate_lc_example_data()
    nan_beta <- example_data$beta
    nan_beta[1, 1] <- NaN

    expect_warning(
        result <- estimateLCMinfiEwasWater(
            beta = nan_beta,
            targets = example_data$targets,
            lcRef = "saliva"
        ),
        "converted to NA"
    )

    expect_equal(result$methylationBoundaries$NaN.Converted, 1L)
    expect_equal(nrow(result$methylationIssues), 1L)
    expect_identical(
        result$methylationIssues$Status,
        "NaN converted to NA"
    )
    expect_equal(nrow(result$invalidCpGs), 0L)
})

#' Result class returned by preprocessingMinfiEwasWater
#'
#' Objects of class `'dnaEPICO_preprocessingMinfiEwasWater'` are list-based
#' results returned by [preprocessingMinfiEwasWater()]. They are lightweight
#' S3-style containers rather than formal S4 classes.
#'
#' @section Structure:
#' \describe{
#'   \item{targets}{Filtered phenotype table aligned to the retained samples.}
#'   \item{RGSet}{Filtered `RGChannelSet` used in downstream preprocessing.}
#'   \item{rawData}{Object returned by [buildRawMinfiEwasWater()].}
#'   \item{assessment}{Object returned by [assessSamplesMinfiEwasWater()].}
#'   \item{sexData}{Object returned by [predictSexMinfiEwasWater()].}
#'   \item{normData}{Object returned by [normalizeMinfiEwasWater()].}
#'   \item{filterData}{Object returned by [filterProbesMinfiEwasWater()].}
#'   \item{metricsData}{Object returned by [extractMetricsMinfiEwasWater()].}
#'   \item{lcData}{Object returned by [estimateLCMinfiEwasWater()].}
#'   \item{logFile}{Resolved path to the optional log file, or `NULL` when
#'   logging was disabled.}
#' }
#'
#' @seealso [preprocessingMinfiEwasWater()]
#' @name dnaEPICO_preprocessingMinfiEwasWater-class
#' @aliases dnaEPICO_preprocessingMinfiEwasWater
#' @docType class
NULL

#' Result class returned by svaEnmix
#'
#' Objects of class `'dnaEPICO_svaEnmix'` are list-based results returned by
#' [svaEnmix()]. They collect the loaded inputs, surrogate-variable results,
#' association-analysis summaries, and optional file outputs.
#'
#' @section Structure:
#' \describe{
#'   \item{targets}{Phenotype table read from `phenoFile` after any optional row
#'   subsetting.}
#'   \item{RGSet}{Loaded `RGChannelSet` with sample names reset to match
#'   `targets`.}
#'   \item{svaData}{Object returned by [estimateSvaEnmixControls()].}
#'   \item{mergedPheno}{Phenotype table returned by [mergeSvaTargetsEnmix()]
#'   after surrogate variables were appended.}
#'   \item{analysisData}{Object returned by [analyzeSvaEnmix()].}
#'   \item{plotFiles}{Named list of TIFF output paths for the SVA figures when
#'   `saveOutputs = TRUE`, otherwise `NULL` entries for plots not written.}
#'   \item{savedFiles}{Object returned by [writeSvaEnmixOutputs()] when
#'   `saveOutputs = TRUE`, otherwise `NULL`.}
#'   \item{logFile}{Resolved path to the optional log file, or `NULL` when
#'   logging was disabled.}
#' }
#'
#' @seealso [svaEnmix()]
#' @name dnaEPICO_svaEnmix-class
#' @aliases dnaEPICO_svaEnmix
#' @docType class
NULL

#' Result class returned by preprocessingPheno
#'
#' Objects of class `'dnaEPICO_preprocessingPheno'` are list-based results
#' returned by [preprocessingPheno()]. They describe the phenotype data,
#' methylation matrices, timepoint splits, longitudinal merges, and optional
#' exported files.
#'
#' @section Structure:
#' \describe{
#'   \item{pheno}{Phenotype table read from `phenoFile`.}
#'   \item{metricsData}{Object returned by [loadMetricsPreprocessingPheno()].}
#' \item{timepointData}{Object returned by
#' [splitTimepointsPreprocessingPheno()].}
#' \item{combinedData}{Object returned by
#' [combineTimepointsPreprocessingPheno()].}
#'   \item{clockFoundation}{Object returned by
#'   [buildClockFoundationInputsPreprocessingPheno()].}
#'   \item{savedFiles}{Object returned by [writePreprocessingPhenoOutputs()]
#'   when `saveOutputs = TRUE`, otherwise `NULL`.}
#'   \item{logFile}{Resolved path to the optional log file, or `NULL` when
#'   logging was disabled.}
#' }
#'
#' @seealso [preprocessingPheno()]
#' @name dnaEPICO_preprocessingPheno-class
#' @aliases dnaEPICO_preprocessingPheno
#' @docType class
NULL

#' Result class returned by methylationGLM
#'
#' Objects of class `'dnaEPICO_methylationGLM'` are list-based results
#' returned by [methylationGLM()]. They collect the prepared analysis table,
#' fitted models, summaries, diagnostics, annotations, and optional saved files.
#'
#' @section Structure:
#' \describe{
#'   \item{preparedData}{Object returned by [prepareMethylationGLMData()].}
#'   \item{distributionPlots}{Object returned by
#'   [plotMethylationGLMDistributions()].}
#'   \item{modelFits}{Object returned by [fitMethylationGLMModels()].}
#'   \item{modelSummaries}{Object returned by
#'   [summarizeMethylationGLMModels()].}
#'   \item{significantCpGs}{Object returned by
#'   [collectSignificantCpGsMethylationGLM()].}
#'   \item{diagnosticPlots}{Object returned by
#'   [plotMethylationGLMDiagnostics()].}
#'   \item{annotation}{Object returned by
#'   [annotateMethylationGLMSummaries()].}
#'   \item{savedFiles}{Object returned by [writeMethylationGLMOutputs()] when
#'   `saveOutputs = TRUE`, otherwise `NULL`.}
#' }
#'
#' @seealso [methylationGLM()]
#' @name dnaEPICO_methylationGLM-class
#' @aliases dnaEPICO_methylationGLM
#' @docType class
NULL

#' Result class returned by methylationLME
#'
#' Objects of class `'dnaEPICO_methylationLME'` are list-based results
#' returned by [methylationLME()]. They collect the prepared longitudinal
#' analysis table, fitted mixed models, summaries, diagnostics, annotations, and
#' optional saved files.
#'
#' @section Structure:
#' \describe{
#'   \item{preparedData}{Object returned by [prepareMethylationLMEData()].}
#'   \item{modelFits}{Object returned by [fitMethylationLMEModels()].}
#'   \item{modelSummaries}{Object returned by
#'   [summarizeMethylationLMEModels()].}
#'   \item{significantInteractions}{Object returned by
#'   [collectSignificantInteractionsMethylationLME()].}
#'   \item{diagnosticPlots}{Object returned by
#'   [plotMethylationLMEDiagnostics()].}
#'   \item{annotation}{Object returned by
#'   [annotateMethylationLMESummaries()].}
#'   \item{savedFiles}{Object returned by
#'   [writeMethylationLMEOutputs()] when `saveOutputs = TRUE`, otherwise
#'   `NULL`.}
#' }
#'
#' @seealso [methylationLME()]
#' @name dnaEPICO_methylationLME-class
#' @aliases dnaEPICO_methylationLME
#' @docType class
NULL

#' Result class returned by prepareDnamReportInputs
#'
#' Objects of class `'dnaEPICO_dnamReport_prepared'` are list-based results
#' returned by [prepareDnamReportInputs()]. They capture normalized report
#' paths,
#' available figures, and logging metadata before rendering.
#'
#' @section Structure:
#' \describe{
#'   \item{output}{Requested output file name.}
#'   \item{outputDir}{Normalized report output directory.}
#'   \item{outputFile}{Normalized full path to the intended report output file.}
#'   \item{figDir}{Normalized directory used by the report template for copied
#'   figures.}
#'   \item{figureInventory}{Named list describing the available figures for each
#'   report section.}
#'   \item{missingFigureDirectories}{Character vector of expected figure
#'   directories that were not present at preparation time.}
#'   \item{logFile}{Resolved path to the optional log file, or `NULL` when
#'   logging was disabled.}
#' }
#'
#' @seealso [prepareDnamReportInputs()]
#' @name dnaEPICO_dnamReport_prepared-class
#' @aliases dnaEPICO_dnamReport_prepared
#' @docType class
NULL

#' Result class returned by renderDnamReport
#'
#' Objects of class `'dnaEPICO_dnamReport_render'` are list-based results
#' returned by [renderDnamReport()]. They describe whether a prepared report was
#' rendered, skipped, or failed.
#'
#' @section Structure:
#' \describe{
#'   \item{preparedReport}{The input object supplied to [renderDnamReport()].}
#'   \item{status}{Render status string such as `'rendered'`, `'skipped'`, or
#'   `'failed'`.}
#' \item{renderedFile}{Normalized path to the rendered HTML file when rendering
#'   succeeded, otherwise `NULL`.}
#'   \item{errorMessage}{Render error or skip message when available, otherwise
#'   `NULL`.}
#'   \item{logFile}{Resolved path to the optional log file, or `NULL` when
#'   logging was disabled.}
#' }
#'
#' @seealso [renderDnamReport()]
#' @name dnaEPICO_dnamReport_render-class
#' @aliases dnaEPICO_dnamReport_render
#' @docType class
NULL

#' Result class returned by dnamReport
#'
#' Objects of class `'dnaEPICO_dnamReport'` are list-based results returned by
#' [dnamReport()]. They combine the prepared report inputs, render result, and
#' final status metadata into one convenience object.
#'
#' @section Structure:
#' \describe{
#'   \item{preparedReport}{Object returned by [prepareDnamReportInputs()].}
#'   \item{renderResult}{Structured render metadata created by [dnamReport()].}
#'   \item{status}{Final status string such as `'rendered'`, `'skipped'`, or
#'   `'failed'`.}
#'   \item{outputFile}{Path to `docs/index.html`.}
#'   \item{resultTableSources}{Named character vector identifying whether each
#'   model table used a streamed sidecar or the compatible XLSX fallback.}
#'   \item{errorMessage}{Final render error or skip message when available,
#'   otherwise `NULL`.}
#'   \item{logFile}{Resolved path to the optional log file, or `NULL` when
#'   logging was disabled.}
#' }
#'
#' @seealso [dnamReport()]
#' @name dnaEPICO_dnamReport-class
#' @aliases dnaEPICO_dnamReport
#' @docType class
NULL

#' Validate a loaded RGSet for svaEnmix
#'
#' @param RGSet Object loaded from `rgsetData`.
#' @param rgsetData Character. Source file path used in validation messages.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return The validated `RGSet`.
#'
#' @description
#' Internal helper that logs the class and dimensions of the loaded object and
#' stops with a clearer error when the saved file does not contain a usable
#' `RGChannelSet`-like object.
#'
#' @keywords internal
#' @noRd
svaEnmixGetRGSetSampleNames <- function(RGSet) {
    sample_names <- tryCatch(colnames(RGSet), error = function(...) NULL)

    if (length(sample_names) > 0L) {
        return(sample_names)
    }

    if (!isS4(RGSet)) {
        return(character(0))
    }

    slots <- methods::slotNames(RGSet)

    if ("colData" %in% slots) {
        col_data <- methods::slot(RGSet, "colData")
        sample_names <- rownames(as.data.frame(col_data))

        if (length(sample_names) > 0L) {
            return(sample_names)
        }
    }

    if ("phenoData" %in% slots) {
        pheno_data <- Biobase::pData(Biobase::phenoData(RGSet))
        sample_names <- rownames(pheno_data)

        if (length(sample_names) > 0L) {
            return(sample_names)
        }
    }

    character(0)
}

#' @keywords internal
#' @noRd
svaEnmixGetRGSetSampleCount <- function(RGSet) {
    rgset_dim <- tryCatch(dim(RGSet), error = function(...) NULL)

    if (length(rgset_dim) == 2L) {
        return(rgset_dim[[2L]])
    }

    sample_names <- svaEnmixGetRGSetSampleNames(RGSet)

    if (length(sample_names) > 0L) {
        return(length(sample_names))
    }

    if (!isS4(RGSet)) {
        return(NULL)
    }

    slots <- methods::slotNames(RGSet)

    if ("colData" %in% slots) {
        return(nrow(as.data.frame(methods::slot(RGSet, "colData"))))
    }

    if ("phenoData" %in% slots) {
        return(nrow(Biobase::pData(Biobase::phenoData(RGSet))))
    }

    NULL
}

#' @keywords internal
#' @noRd
svaEnmixSetRGSetSampleNames <- function(RGSet, sampleNames) {
    updated_rgset <- tryCatch(
        {
            colnames(RGSet) <- sampleNames
            RGSet
        },
        error = function(...) NULL
    )

    if (!is.null(updated_rgset)) {
        return(updated_rgset)
    }

    if (!isS4(RGSet)) {
        stop("Could not set sample names on the loaded RGSet object.",
            call. = FALSE
        )
    }

    slots <- methods::slotNames(RGSet)

    if ("colData" %in% slots) {
        col_data <- methods::slot(RGSet, "colData")
        rownames(col_data) <- sampleNames
        methods::slot(RGSet, "colData") <- col_data
        return(RGSet)
    }

    if ("phenoData" %in% slots) {
        pheno_data <- Biobase::phenoData(RGSet)
        rownames(Biobase::pData(pheno_data)) <- sampleNames
        Biobase::phenoData(RGSet) <- pheno_data
        return(RGSet)
    }

    stop("Could not set sample names on the loaded RGSet object.",
        call. = FALSE
    )
}

#' @keywords internal
#' @noRd
svaEnmixGetRGSetColData <- function(RGSet) {
    col_data <- tryCatch(SummarizedExperiment::colData(RGSet),
        error = function(...) NULL
    )

    if (!is.null(col_data)) {
        return(col_data)
    }

    if (!isS4(RGSet)) {
        stop("Could not retrieve column metadata from the loaded RGSet.",
            call. = FALSE
        )
    }

    slots <- methods::slotNames(RGSet)

    if ("colData" %in% slots) {
        return(methods::slot(RGSet, "colData"))
    }

    if ("phenoData" %in% slots) {
        return(Biobase::pData(Biobase::phenoData(RGSet)))
    }

    stop("Could not retrieve column metadata from the loaded RGSet.",
        call. = FALSE
    )
}

#' @keywords internal
#' @noRd
svaEnmixValidateRGSet <- function(
    RGSet, rgsetData, verbose = FALSE,
    logs = FALSE, log_dir = NULL, log_file = "log_svaEnmix.txt"
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )
    rgset_class <- paste(class(RGSet), collapse = ", ")
    rgset_dims <- dim(RGSet)
    rgset_ncol <- svaEnmixGetRGSetSampleCount(RGSet)
    slots <- if (isS4(RGSet)) {
        methods::slotNames(RGSet)
    } else {
        character(0)
    }
    log_lines <- c(paste("Loaded RGSet object class: ", rgset_class))

    if (length(rgset_dims) == 2L) {
        log_lines <- c(log_lines, paste(
            "Loaded RGSet dimensions:   ",
            paste(rgset_dims, collapse = " x ")
        ))
    }

    log_lines <- c(log_lines, paste(
        "Loaded RGSet sample count:  ",
        if (length(rgset_ncol) == 1L && !is.na(rgset_ncol)) {
            rgset_ncol
        } else {
            "unavailable"
        }
    ), paste("Loaded RGSet slots:         ", if (length(slots) >
        0L) {
        paste(slots, collapse = ", ")
    } else {
        "unavailable"
    }))

    emitLogMinfiEwasWater(log_lines, verbose = verbose, log_path = log_path)

    if (!(length(rgset_ncol) == 1L && !is.na(rgset_ncol))) {
        stop("The object loaded from ", rgsetData, " has class ",
            rgset_class,
                " and does not expose a usable sample count after loading.",
            call. = FALSE
        )
    }

    list(RGSet = RGSet, sampleCount = rgset_ncol)
}

#' Estimate surrogate variables from ENmix control probes
#'
#' Read the phenotype table and a saved `RGChannelSet`, estimate surrogate
#' variables from ENmix control probes, analyze their association with Sentrix
#' chip and position factors, and return a structured in-memory result. CSV,
#' `.RData`, text-summary, and figure outputs are written only when
#' `saveOutputs = TRUE`.
#'
#' @param phenoFile Character. Path to the phenotype file with cell-composition
#'   data. When `saveOutputs = TRUE`, the validated PC columns are appended to
#'   this file using a rollback-safe replacement.
#' @param rgsetData Character. Path to a saved `RGChannelSet` object. Both
#'   `.RData` and `.rds` files are supported.
#' @param sepType Character or `NULL`. Field separator used in `phenoFile`. Use
#'   `NULL` for a comma-separated file, `'\\t'` for a tab-delimited file, or
#'   another separator accepted by `utils::read.csv()`.
#' @param outputLogs Character. Directory used for log files when `logs = TRUE`.
#' @param nSamples Integer or `NA`. Number of rows to keep from the phenotype
#'   table. Use `NA` to keep all samples.
#' @param SampleID Character. Name of the phenotype column containing sample
#'   identifiers.
#' @param arrayType Character. Illumina array identifier assigned to
#'   `Biobase::annotation(RGSet)`.
#' @param annotationVersion Character. Annotation build assigned to
#'   `Biobase::annotation(RGSet)`.
#' @param SentrixIDColumn Character. Name of the chip identifier column in the
#'   phenotype data.
#' @param SentrixPositionColumn Character. Name of the chip position column in
#'   the phenotype data.
#' @param ctrlSvaPercVar Numeric. Proportion of control-probe variance explained
#'   when running `ENmix::ctrlsva()`.
#' @param ctrlSvaFlag Integer. Control-probe flag passed to `ENmix::ctrlsva()`.
#' @param scriptLabel Character. Label used to name output folders when
#'   `saveOutputs = TRUE`.
#' @param tiffWidth Integer. Width of saved TIFF plots in pixels.
#' @param tiffHeight Integer. Height of saved TIFF plots in pixels.
#' @param tiffRes Integer. Resolution in DPI for saved TIFF plots.
#' @param figureBaseDir Character. Base directory used for saved figure outputs
#'   when `saveOutputs = TRUE`.
#' @param dataBaseDir Character. Base directory used for saved CSV and text
#'   outputs when `saveOutputs = TRUE`.
#' @param rBaseDir Character. Base directory used for saved `.RData` outputs
#'   when `saveOutputs = TRUE`.
#' @param display Logical. If `TRUE`, draw plots on the active graphics device.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#'   The default is `FALSE`.
#' @param logs Logical. If `TRUE`, write the same progress messages to
#'   `outputLogs`. The default is `FALSE`.
#' @param saveOutputs Logical. If `TRUE`, write the CSV, `.RData`, text,
#'   and TIFF outputs to disk. The default is `FALSE`.
#'
#' @return A list with class `'dnaEPICO_svaEnmix'`.
#' \describe{
#'   \item{targets}{Phenotype table read from `phenoFile` after any optional row
#'   subsetting.}
#'   \item{RGSet}{Loaded `RGChannelSet` with sample names realigned to
#'   `targets[[SampleID]]`.}
#'   \item{svaData}{Object returned by [estimateSvaEnmixControls()] containing
#'   the surrogate-variable matrix and the control-probe settings used to
#'   estimate it.}
#'   \item{mergedPheno}{Phenotype table returned by [mergeSvaTargetsEnmix()]
#'   after the surrogate variables were appended as additional columns.}
#'   \item{analysisData}{Object returned by [analyzeSvaEnmix()] containing the
#'   surrogate-variable association models, ANOVA tables, and Sentrix metadata.}
#' \item{plotFiles}{Named list describing the plot file paths requested for the
#'   SVA figures. When `saveOutputs = FALSE`, the entries are typically `NULL`.}
#'   \item{savedFiles}{Object returned by [writeSvaEnmixOutputs()] when
#'   `saveOutputs = TRUE`, otherwise `NULL`.}
#'   \item{logFile}{Resolved path to the optional log file, or `NULL` when
#'   logging was disabled.}
#' }
#' See [dnaEPICO_svaEnmix-class] for a class-level overview.
#'
#' @examples
#' tmp <- tempdir()
#' stopifnot(dir.exists(tmp))
#'
#' if (requireNamespace("minfiData", quietly = TRUE)) {
#'     ex <- dnaEPICO:::exampleMinfiBaseDataDnaEpico()
#'     pheno_file <- file.path(tmp, "pheno.csv")
#'     rgset_path <- file.path(tmp, "RGSet.RData")
#'     RGSet <- ex$RGSet
#'     utils::write.csv(ex$targets, pheno_file, row.names = FALSE)
#'     save(RGSet, file = rgset_path)
#'     sva_result <- svaEnmix(
#'         phenoFile = pheno_file,
#'         rgsetData = rgset_path,
#'         SampleID = "Sample_Name",
#'         arrayType = "IlluminaHumanMethylation450k",
#'         annotationVersion = "ilmn12.hg19",
#'         SentrixIDColumn = "Sentrix_ID",
#'         SentrixPositionColumn = "Sentrix_Position",
#'         outputLogs = file.path(tmp, "logs"),
#'         figureBaseDir = file.path(tmp, "figures"),
#'         dataBaseDir = file.path(tmp, "data"),
#'         rBaseDir = file.path(tmp, "rData"),
#'         saveOutputs = FALSE
#'     )
#'     stopifnot(inherits(sva_result, "dnaEPICO_svaEnmix"))
#' }
#'
#' @seealso [dnaEPICO_svaEnmix-class]
#'
#' @export
svaEnmix <- function(
    phenoFile = "data/preprocessingMinfiEwasWater/phenoLC.csv",
    rgsetData = "rData/preprocessingMinfiEwasWater/objects/RGSet.RData",
    sepType = NULL, outputLogs = "logs", nSamples = NA,
        SampleID = "Sample_Name",
    arrayType = "IlluminaHumanMethylationEPICv2",
        annotationVersion = "20a1.hg38",
    SentrixIDColumn = "Sentrix_ID", SentrixPositionColumn = "Sentrix_Position",
    ctrlSvaPercVar = 0.9, ctrlSvaFlag = 1, scriptLabel = "svaEnmix",
    tiffWidth = 2000, tiffHeight = 1000, tiffRes = 150,
        figureBaseDir = "figures",
    dataBaseDir = "data", rBaseDir = "rData", display = FALSE,
    verbose = FALSE, logs = FALSE, saveOutputs = FALSE
) {
    log_file <- "log_svaEnmix.txt"
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = outputLogs,
        log_file = log_file
    )
    figure_dir <- file.path(figureBaseDir, scriptLabel)

    emitLogMinfiEwasWater(
        c(
            "==== Starting SVA Estimation with Enmix ====",
            paste("Start time:               ", format(Sys.time())),
            paste("Log file path:            ",
                if (is.null(log_path)) "disabled" else log_path),
            paste("Pheno file:               ", phenoFile), paste(
                "RGSet path:               ",
                rgsetData
            ), paste("Separator type:           ",
                if (is.null(resolveSeparatorMinfiEwasWater(sepType))) {
                "default (',')"
            } else {
                sepType
            }), paste("Sample limit:             ",
                if (is.na(nSamples)) "all" else nSamples),
            paste("SampleID column:          ", SampleID), paste(
                "Array type:               ",
                arrayType
            ), paste("Annotation version:       ", annotationVersion),
            paste("Sentrix ID column:        ", SentrixIDColumn),
            paste("Sentrix position column:  ", SentrixPositionColumn),
            paste("ctrlSva percvar:          ", ctrlSvaPercVar),
            paste("ctrlSva flag:             ", ctrlSvaFlag), paste(
                "Script label:             ",
                scriptLabel
            ), paste(
                "TIFF dimensions (WxH):    ",
                tiffWidth, " x ", tiffHeight, " @ ", tiffRes
            ), paste(
                "Display plots:            ",
                display
            ), paste("Save outputs:             ", saveOutputs),
            "============================================================"
        ),
        verbose = verbose, log_path = log_path
    )

    withLoggedErrorsMinfiEwasWater(expr = {
        targets <- readPhenotypeTargets(
            phenoFile = phenoFile,
            sepType = sepType, nSamples = nSamples, SampleID = SampleID,
            verbose = verbose, logs = logs, log_dir = outputLogs,
            log_file = log_file
        )

        RGSet <- loadSavedObjectPreprocessingPheno(
            path = rgsetData,
            preferred_name = "RGSet"
        )
        rgset_validation <- svaEnmixValidateRGSet(
            RGSet = RGSet,
            rgsetData = rgsetData, verbose = verbose, logs = logs,
            log_dir = outputLogs, log_file = log_file
        )
        RGSet <- rgset_validation$RGSet
        rgset_ncol <- rgset_validation$sampleCount

        if (rgset_ncol != nrow(targets)) {
            stop("The saved RGSet contains ", rgset_ncol,
                " samples but the phenotype table contains ",
                nrow(targets), ".",
                call. = FALSE
            )
        }

        target_sample_ids <- validateSampleIdentifiersDnaEpico(
            targets[[SampleID]],
            paste0("Phenotype column '", SampleID, "'")
        )
        rgset_sample_ids <- svaEnmixGetRGSetSampleNames(RGSet)
        rgset_match <- matchSampleIdentifiersDnaEpico(
            query = target_sample_ids,
            reference = rgset_sample_ids, queryLabel = paste0(
                "phenotype column '",
                SampleID, "'"
            ), referenceLabel = "saved RGSet sample identifiers",
            requireSameSet = TRUE
        )
        RGSet <- RGSet[, rgset_match]

        RGSet <- svaEnmixSetRGSetSampleNames(RGSet = RGSet,
            sampleNames = target_sample_ids)
        Biobase::annotation(RGSet) <- c(array = arrayType,
            annotation = annotationVersion)

        emitLogMinfiEwasWater(
            c(
                paste(
                    "RGSet loaded with          ",
                    rgset_ncol, " samples."
                ), paste(
                    "Applied annotation:        ",
                    paste(Biobase::annotation(RGSet), collapse = ", ")
                ),
                "============================================================"
            ),
            verbose = verbose, log_path = log_path
        )

        svaData <- estimateSvaEnmixControls(
            RGSet = RGSet, ctrlSvaPercVar = ctrlSvaPercVar,
            ctrlSvaFlag = ctrlSvaFlag, verbose = verbose, logs = logs,
            log_dir = outputLogs, log_file = log_file
        )
        mergedPheno <- mergeSvaTargetsEnmix(
            targets = targets,
            sva = svaData$sva, SampleID = SampleID, verbose = verbose,
            logs = logs, log_dir = outputLogs, log_file = log_file
        )
        analysisData <- analyzeSvaEnmix(
            sva = svaData$sva, RGSet = RGSet,
            SentrixIDColumn = SentrixIDColumn,
                SentrixPositionColumn = SentrixPositionColumn,
            verbose = verbose, logs = logs, log_dir = outputLogs,
            log_file = log_file
        )

        plot_files <- list(
            sentrixID = plotSvaEnmix(
                analysisData = analysisData,
                plot = "sentrix_id", display = display,
                    file = if (isTRUE(saveOutputs)) {
                    file.path(
                        figure_dir,
                        "sva_SentrixID.tiff"
                    )
                } else {
                    NULL
                }, width = tiffWidth,
                height = tiffHeight, res = tiffRes, verbose = verbose,
                logs = logs, log_dir = outputLogs, log_file = log_file
            ),
            sentrixPosition = plotSvaEnmix(
                analysisData = analysisData,
                plot = "sentrix_position", display = display,
                file = if (isTRUE(saveOutputs)) {
                    file.path(
                        figure_dir,
                        "sva_SentrixPosition.tiff"
                    )
                } else {
                    NULL
                }, width = tiffWidth,
                height = tiffHeight, res = tiffRes, verbose = verbose,
                logs = logs, log_dir = outputLogs, log_file = log_file
            ),
            matrix = plotSvaEnmix(
                analysisData = analysisData,
                plot = "matrix", display = display,
                    file = if (isTRUE(saveOutputs)) {
                    file.path(
                        figure_dir,
                        "sva_SentrixIDPosition.tiff"
                    )
                } else {
                    NULL
                }, width = tiffWidth,
                height = tiffHeight, res = tiffRes, verbose = verbose,
                logs = logs, log_dir = outputLogs, log_file = log_file
            )
        )

        savedFiles <- NULL
        if (isTRUE(saveOutputs)) {
            savedFiles <- writeSvaEnmixOutputs(
                svaData = svaData,
                mergedPheno = mergedPheno, analysisData = analysisData,
                phenoFile = phenoFile, SampleID = SampleID, sepType = sepType,
                dataBaseDir = dataBaseDir, rBaseDir = rBaseDir,
                scriptLabel = scriptLabel, verbose = verbose,
                logs = logs, log_dir = outputLogs, log_file = log_file
            )
        }

        emitLogMinfiEwasWater(
            c(
                "==== Finished SVA Estimation with Enmix ====",
                paste("End time:                 ", format(Sys.time())),
                "============================================================"
            ),
            verbose = verbose, log_path = log_path
        )

        structure(list(
            targets = targets, RGSet = RGSet, svaData = svaData,
            mergedPheno = mergedPheno, analysisData = analysisData,
            plotFiles = plot_files, savedFiles = savedFiles,
            logFile = log_path
        ), class = "dnaEPICO_svaEnmix")
    }, log_path = log_path, verbose = verbose, context = "svaEnmix")
}

#' Prepare phenotype and methylation matrices for downstream modeling
#'
#' Read the phenotype table and the preprocessed beta, M-value, and copy-number
#' matrices; align them by sample identifier; split them by timepoint; prepare
#' combined longitudinal objects for a selected modeling scale; and build Clock
#' Foundation export tables. The
#' function returns a structured in-memory result, while files are
#' written only when `saveOutputs = TRUE`.
#'
#' @param phenoFile Character. Path to the phenotype CSV file.
#' @param sepType Character or `NULL`. Field separator used in `phenoFile`. Use
#'   `NULL` for a comma-separated file, `'\\t'` for a tab-delimited file, or
#'   another separator accepted by `utils::read.csv()`.
#' @param betaPath Character. Path to the saved beta-value object. Both `.RData`
#'   and `.rds` files are supported.
#' @param mPath Character. Path to the saved M-value object. Both `.RData` and
#'   `.rds` files are supported.
#' @param cnPath Character. Path to the saved copy-number object. Both `.RData`
#'   and `.rds` files are supported.
#' @param SampleID Character. Name of the phenotype column containing sample
#'   identifiers used to align phenotype and methylation data.
#' @param timeVar Character. Name of the phenotype column containing timepoint
#'   labels.
#' @param timepoints Character vector or comma-separated string of timepoints to
#'   retain and split into separate in-memory subsets.
#' @param combineTimepoints Character vector or comma-separated string of
#'   timepoints to combine into the longitudinal phenotype-plus-methylation
#'   object.
#' @param methylationScale Character. Methylation metric to use in merged
#'   modeling tables. One of `'Beta'`, `'M'`, or `'CN'`, in any combination
#'   of upper- and lower-case letters. The default is `'beta'`. Beta values
#'   are always used for Clock Foundation exports.
#' @param outputPheno Character. Directory used for saved phenotype CSV files
#'   when `saveOutputs = TRUE`.
#' @param outputRData Character. Directory used for saved metric `.RData` files
#'   when `saveOutputs = TRUE`.
#' @param outputRDataMerge Character. Directory used for saved merged
#'   phenotype-plus-methylation `.RData` files when `saveOutputs = TRUE`.
#' @param sexColumn Character. Name of the phenotype sex column used when
#'   building Clock Foundation exports.
#' @param outputLogs Character. Directory used for log files when `logs = TRUE`.
#' @param outputDir Character. Directory used for Clock Foundation export files
#'   when `saveOutputs = TRUE`.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#'   The default is `FALSE`.
#' @param logs Logical. If `TRUE`, write the same progress messages to
#'   `outputLogs`. The default is `FALSE`.
#' @param saveOutputs Logical. If `TRUE`, write the CSV, ZIP, and `.RData`
#'   outputs to disk. The default is `FALSE`.
#'
#' @return A list with class `'dnaEPICO_preprocessingPheno'`.
#' \describe{
#'   \item{pheno}{Phenotype table read from `phenoFile`.}
#'   \item{metricsData}{Object returned by [loadMetricsPreprocessingPheno()]
#'   containing the beta-value, M-value, and copy-number matrices loaded from
#'   `betaPath`, `mPath`, and `cnPath`.}
#' \item{timepointData}{Object returned by
#' [splitTimepointsPreprocessingPheno()]
#'   containing per-timepoint phenotype tables and methylation matrices.}
#'   \item{combinedData}{Object returned by
#'   [combineTimepointsPreprocessingPheno()] containing the merged longitudinal
#'   phenotype-plus-methylation object and the timepoint combination metadata.}
#'   \item{clockFoundation}{Object returned by
#'   [buildClockFoundationInputsPreprocessingPheno()] containing the beta table
#'   and phenotype table prepared for Clock Foundation export.}
#' \item{savedFiles}{Object returned by [writePreprocessingPhenoOutputs()] when
#'   `saveOutputs = TRUE`, otherwise `NULL`.}
#'   \item{logFile}{Resolved path to the optional log file, or `NULL` when
#'   logging was disabled.}
#' }
#' See [dnaEPICO_preprocessingPheno-class] for a class-level overview.
#'
#' @examples
#' tmp <- tempdir()
#' pheno <- data.frame(
#'     Sample_Name = c("S1", "S2", "S3"),
#'     Timepoint = c("1", "1", "2"),
#'     Sex = c(0, 1, 0),
#'     stringsAsFactors = FALSE
#' )
#' beta <- matrix(
#'     c(0.10, 0.20, 0.30, 0.40, 0.50, 0.60),
#'     nrow = 2,
#'     dimnames = list(c("cg1", "cg2"), pheno$Sample_Name)
#' )
#' m <- beta * 10
#' cn <- beta * 100
#' pheno_file <- file.path(tmp, "pheno.csv")
#' beta_path <- file.path(tmp, "beta.RData")
#' m_path <- file.path(tmp, "m.RData")
#' cn_path <- file.path(tmp, "cn.RData")
#' utils::write.csv(pheno, pheno_file, row.names = FALSE)
#' save(beta, file = beta_path)
#' save(m, file = m_path)
#' save(cn, file = cn_path)
#' result <- preprocessingPheno(
#'     phenoFile = pheno_file,
#'     betaPath = beta_path,
#'     mPath = m_path,
#'     cnPath = cn_path,
#'     SampleID = "Sample_Name",
#'     timeVar = "Timepoint",
#'     timepoints = "1,2",
#'     combineTimepoints = "1,2",
#'     outputPheno = file.path(tmp, "data", "preprocessingPheno"),
#'     outputRData = file.path(tmp, "rData", "preprocessingPheno", "metrics"),
#'     outputRDataMerge = file.path(tmp, "rData", "preprocessingPheno", "mergeData"),
#'     sexColumn = "Sex",
#'     outputLogs = file.path(tmp, "logs"),
#'     outputDir = file.path(tmp, "clockFoundation"),
#'     saveOutputs = FALSE
#' )
#' stopifnot(inherits(result, "dnaEPICO_preprocessingPheno"))
#'
#' @seealso [dnaEPICO_preprocessingPheno-class]
#'
#' @export
preprocessingPheno <- function(
    phenoFile = "data/preprocessingMinfiEwasWater/phenoLC.csv",
    sepType = NULL,
        betaPath = "rData/preprocessingMinfiEwasWater/metrics/beta_NomFilt_MSetF_Flt_Rxy_Ds_Rc.RData",
    mPath = "rData/preprocessingMinfiEwasWater/metrics/m_NomFilt_MSetF_Flt_Rxy_Ds_Rc.RData",
    cnPath = "rData/preprocessingMinfiEwasWater/metrics/cn_NomFilt_MSetF_Flt_Rxy_Ds_Rc.RData",
    SampleID = "Sample_Name", timeVar = "Timepoint", timepoints = "1,2",
    combineTimepoints = "1,2", methylationScale = "beta",
        outputPheno = "data/preprocessingPheno",
    outputRData = "rData/preprocessingPheno/metrics",
        outputRDataMerge = "rData/preprocessingPheno/mergeData",
    sexColumn = "Sex", outputLogs = "logs",
        outputDir = "data/preprocessingPheno",
    verbose = FALSE, logs = FALSE, saveOutputs = FALSE
) {
    methylationScale <- normalizeMethylationScaleDnaEpico(methylationScale)
    methylationLabel <- methylationScaleResponseLabelDnaEpico(methylationScale)
    methylationObjectPrefix <-
        methylationScaleObjectPrefixDnaEpico(methylationScale)
    log_file <- "log_preprocessingPheno.txt"
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = outputLogs,
        log_file = log_file
    )

    emitLogMinfiEwasWater(
        c(
            "==== Starting preprocessingPheno ====",
            paste("Start Time:               ", format(Sys.time())),
            paste("Log file path:            ",
                if (is.null(log_path)) "disabled" else log_path),
            paste("Phenotype file:           ", phenoFile), paste(
                "Separator type:           ",
                if (is.null(resolveSeparatorMinfiEwasWater(sepType))) {
                    "default (',')"
                } else {
                    sepType
                }
            ), paste("Beta path:                ", betaPath),
            paste("M-values path:            ", mPath), paste(
                "CN path:                  ",
                cnPath
            ), paste("Identifier column:        ", SampleID),
            paste("Timepoint column:         ", timeVar), paste(
                "Timepoints:               ",
                timepoints
            ), paste(
                "Combine timepoints:       ",
                combineTimepoints
            ), paste(
                "Merged modeling object:   ",
                methylationObjectPrefix, "*"
            ), "Clock Foundation scale:     Beta values",
            paste("Sex column:               ", sexColumn), paste(
                "Output phenotype dir:     ",
                outputPheno
            ), paste(
                "RData metrics dir:        ",
                outputRData
            ), paste(
                "RData merge dir:          ",
                outputRDataMerge
            ), paste(
                "Clock Foundation dir:     ",
                outputDir
            ), paste("Save outputs:             ", saveOutputs),
            "============================================================"
        ),
        verbose = verbose, log_path = log_path
    )

    withLoggedErrorsMinfiEwasWater(expr = {
        pheno <- readPhenotypeTargets(
            phenoFile = phenoFile,
            sepType = sepType, SampleID = SampleID, verbose = verbose,
            logs = logs, log_dir = outputLogs, log_file = log_file
        )
        metricsData <- loadMetricsPreprocessingPheno(
            betaPath = betaPath,
            mPath = mPath, cnPath = cnPath, verbose = verbose,
            logs = logs, log_dir = outputLogs, log_file = log_file
        )
        timepointData <- splitTimepointsPreprocessingPheno(
            pheno = pheno,
            metricsData = metricsData, SampleID = SampleID, timeVar = timeVar,
            timepoints = timepoints, methylationScale = methylationScale,
            verbose = verbose, logs = logs, log_dir = outputLogs,
            log_file = log_file
        )
        combinedData <- combineTimepointsPreprocessingPheno(
            timepointData = timepointData,
            combineTimepoints = combineTimepoints,
                methylationScale = methylationScale,
            verbose = verbose, logs = logs, log_dir = outputLogs,
            log_file = log_file
        )
        clockFoundation <- buildClockFoundationInputsPreprocessingPheno(
            beta = metricsData$beta,
            pheno = pheno, SampleID = SampleID, sexColumn = sexColumn,
            verbose = verbose, logs = logs, log_dir = outputLogs,
            log_file = log_file
        )

        result <- list(
            pheno = pheno, metricsData = metricsData,
            timepointData = timepointData, combinedData = combinedData,
            methylationScale = methylationScale,
                methylationLabel = methylationLabel,
            methylationObjectPrefix = methylationObjectPrefix,
            clockFoundation = clockFoundation, savedFiles = NULL,
            logFile = log_path
        )

        if (isTRUE(saveOutputs)) {
            result$savedFiles <- writePreprocessingPhenoOutputs(
                preprocessingData = result,
                outputPheno = outputPheno, outputRData = outputRData,
                outputRDataMerge = outputRDataMerge, outputDir = outputDir,
                verbose = verbose, logs = logs, log_dir = outputLogs,
                log_file = log_file
            )
        }

        emitLogMinfiEwasWater(
            c(
                "==== Finished preprocessingPheno ====",
                paste("End Time:                 ", format(Sys.time())),
                "============================================================"
            ),
            verbose = verbose, log_path = log_path
        )

        structure(result, class = "dnaEPICO_preprocessingPheno")
    }, log_path = log_path, verbose = verbose, context = "preprocessingPheno")
}

# Data Description

# Purpose:
#   Document how external and internal data files used by the dnaEPICO
# package were obtained, prepared, and stored for reproducibility.
#
# This script is for documentation and reproducibility purposes only.
# It is NOT executed during package installation.

# -----------------------------------------------------------
# Data sources
# -----------------------------------------------------------

# 1. Probe-exclusion list:
#   12864_2024_10027_MOESM8_ESM.csv
#
# This file was obtained from the supplementary material of the
# corresponding publication and is included for probe filtering
# demonstration purposes. The file size was reduced (<100 MB) to
# comply with package distribution requirements.
#
# Source URL:
#   https://link.springer.com/article/10.1186/s12864-024-10027-5
#
#
# 2. Cell composition reference coefficients (Houseman method):
#
#   saliva.txt
#   salivaEPIC.txt
#
# These reference coefficient matrices are used by the function
# `estimateLC()` to estimate cell type proportions from DNA
# methylation beta values using the Houseman algorithm.
#
# The reference data originate from the ewastools pipeline:
#
#   Murat K, et al. (2020)
# "Ewastools: Infinium Human Methylation BeadChip pipeline for
#    population epigenetics integrated into Galaxy."
# GigaScience, 9(5):giaa049.
# https://academic.oup.com/gigascience/article/9/5/giaa049/5836679
#
# The saliva reference panel corresponds to:
#
#   Middleton LYM, et al. (2020)
# "Saliva cell type DNA methylation reference panel for
#    epidemiology studies in children."
# https://pubmed.ncbi.nlm.nih.gov/33588693/

# -----------------------------------------------------------
# Make/ directory
# -----------------------------------------------------------

# The `make/` directory contains workflow orchestration files used to
# run the complete dnaEPICO analysis pipeline in a reproducible and
# automated manner. These files are not executed during package
# installation and are not required for normal package usage.
#
# Files in the `make/` directory include:
#
#   - Makefile rules that define parameters between preprocessing,
# modelling, and reporting steps
#   - Makefile model pipeline configuration variables and models shared
# across scripts
#
# These files allow users and developers to reproduce the full analysis
# performed by dnaEPICO outside of the package runtime environment,
# ensuring transparency and methodological reproducibility.

# -----------------------------------------------------------
# dnaEPICO logo/report graphics
# -----------------------------------------------------------

# dnaEPICO.svg is the packaged logo used by `dnamReport()`.
# dnaEPICORM.svg is the logo used in Readme and vignette files.
# dnaEPICO.gif is a rasterized version of the logo used in the package manual.
# cpg-viewer.js renders paged CpG result tables in the report.

# -----------------------------------------------------------
# Final location in the package
# -----------------------------------------------------------

# The prepared files were placed in:
#
#   inst/extdata/
#
#   - 12864_2024_10027_MOESM8_ESM.csv
#   - saliva.txt
#   - salivaEPIC.txt
#   - dnaEPICO.svg
#   - dnaEPICORM.svg
#   - dnaEPICO.gif
#   - cpg-viewer.js

# inst/extdata/make/
#
#   - Workflow reproducible pipeline execution files

#' dnaEPICO: DNA methylation preprocessing and modeling workflows
#'
#' The `dnaEPICO` package provides a structured workflow for preprocessing and
#' analyzing Illumina DNA methylation array data, including quality control,
#' normalization, cell-type estimation, surrogate-variable analysis, phenotype
#' preparation, CpG-wise generalised linear models, longitudinal mixed-effects
#' models using `lmerTest`/`lme4` or `nlme` with optional residual correlation
#' structures, and interactive web reporting.
#'
#' The package supports two complementary usage styles:
#' \itemize{
#'   \item interactive use, where functions return structured in-memory result
#'   objects for inspection and composition; and
#'   \item file-based pipeline use, where the same functions can write logs,
#'   plots, tables, and serialized objects when `saveOutputs = TRUE`.
#' }
#'
#' The main high-level entry points are:
#' \itemize{
#'   \item [preprocessingMinfiEwasWater()]
#'   \item [svaEnmix()]
#'   \item [preprocessingPheno()]
#'   \item [methylationGLM()]
#'   \item [methylationLME()]
#'   \item [dnamReport()]
#' }
#'
#' @name dnaEPICO
#' @docType package
#' @keywords package
"_PACKAGE"

#' Ensure a person identifier column exists for longitudinal LME analyses
#'
#' @param data Data frame containing the longitudinal phenotype-plus-beta data.
#' @param personVar Character. Name of the subject identifier column.
#' @param sidVar Character. Name of the fallback sample identifier column used
#' to
#'   derive `personVar` when it is missing.
#'
#' @return A list containing the updated data frame, whether `personVar` was
#'   created, an optional SID-to-person preview, and counts per person.
#'
#' @description
#' Internal helper that derives a subject identifier from `SID` when the
#' requested `personVar` column is not present.
#'
#' @keywords internal
#' @noRd
ensurePersonColumnMethylationLME <- function(
    data, personVar = "person",
    sidVar = "SID"
) {
    person_created <- FALSE
    mapping_preview <- NULL

    if (!(personVar %in% colnames(data))) {
        if (!(sidVar %in% colnames(data))) {
            stop("Column '", personVar,
                "' was not found and cannot be created because '",
                sidVar, "' is missing.",
                call. = FALSE
            )
        }

        sid_values <- validateSampleIdentifiersDnaEpico(
            data[[sidVar]],
            paste0("inputPheno$", sidVar)
        )
        if (any(!grepl("[AB]$", sid_values))) {
            stop("Cannot safely derive '", personVar, "' from '",
                sidVar, "': every SID must end in A or B. Supply an explicit subject identifier column.",
                call. = FALSE
            )
        }
        person_values <- sub("[AB]$", "", sid_values)
        visit_values <- sub("^.*([AB])$", "\\1", sid_values)
        if (any(!nzchar(person_values))) {
            stop("Derived subject identifiers cannot be empty.",
                call. = FALSE
            )
        }
        if (anyDuplicated(paste(person_values, visit_values,
            sep = "\r"
        ))) {
            stop("SID contains duplicate subject/visit identifiers.",
                call. = FALSE
            )
        }
        data[[personVar]] <- person_values
        person_created <- TRUE
        mapping_preview <- utils::head(
            data[order(
                data[[personVar]],
                data[[sidVar]]
            ), c(sidVar, personVar), drop = FALSE],
            20L
        )
    }

    person_values <- as.character(data[[personVar]])
    if (anyNA(person_values) || any(!nzchar(trimws(person_values)))) {
        stop("personVar contains missing or blank subject identifiers.",
            call. = FALSE
        )
    }
    person_counts <- table(data[[personVar]], useNA = "ifany")
    if (length(person_counts) < 2L) {
        stop("At least two subjects are required for mixed-effects modeling.",
            call. = FALSE
        )
    }
    if (max(person_counts) < 2L) {
        stop("At least one subject must have repeated observations for mixed-effects modeling.",
            call. = FALSE
        )
    }

    list(
        data = data, personCreated = person_created,
            mappingPreview = mapping_preview,
        personCounts = person_counts
    )
}

#' Build a mixed-effects formula for methylationLME helpers
#'
#' @param phenotype Character. Phenotype variable of interest.
#' @param personVar Character. Subject identifier variable used as a random
#'   intercept.
#' @param covariates Character vector of additional fixed-effect covariates.
#' @param interactionTerm Character or `NULL`. Optional interaction term.
#'
#' @return Character scalar containing a model formula.
#'
#' @description
#' Internal helper that builds the per-CpG linear mixed-effects formula used by
#' `methylationLME()` and its modular helpers.
#'
#' @keywords internal
#' @noRd
buildFormulaMethylationLME <- function(
    phenotype, personVar,
    covariates = character(0), interactionTerm = NULL, includeRandomTerm = TRUE,
    responseVar = "beta"
) {
    if (!is.null(interactionTerm) && nzchar(interactionTerm) &&
        identical(interactionTerm, phenotype)) {
        stop("interactionTerm must differ from the phenotype being modelled.",
            call. = FALSE
        )
    }
    quoted_phenotype <- quoteNamesMethylationGLM(phenotype)
    quoted_person <- quoteNamesMethylationGLM(personVar)
    fixed_terms <- unique(c(covariates))

    if (!is.null(interactionTerm) && nzchar(interactionTerm)) {
        quoted_interaction <- quoteNamesMethylationGLM(interactionTerm)
        interaction_part <- paste(quoted_phenotype, quoted_interaction,
            sep = " * "
        )
        fixed_terms <- setdiff(fixed_terms, c(
            interactionTerm,
            phenotype
        ))
        fixed_formula_terms <- c(interaction_part,
            quoteNamesMethylationGLM(fixed_terms))
    } else {
        fixed_terms <- setdiff(fixed_terms, phenotype)
        fixed_formula_terms <- c(quoted_phenotype,
            quoteNamesMethylationGLM(fixed_terms))
    }

    fixed_formula_terms <-
        unique(fixed_formula_terms[nzchar(fixed_formula_terms)])
    if (length(fixed_formula_terms) == 0L) {
        stop("At least one fixed-effect term is required.", call. = FALSE)
    }

    formula_text <- paste(responseVar, "~", paste(fixed_formula_terms,
        collapse = " + "
    ))

    if (isTRUE(includeRandomTerm)) {
        formula_text <- paste(
            formula_text, "+ (1 |", quoted_person,
            ")"
        )
    }

    formula_text
}

normalizeCorrelationStructureMethylationLME <-
    function(correlationStructure = "none") {
    if (length(correlationStructure) == 0L ||
        all(is.na(correlationStructure))) {
        return("none")
    }
    if (length(correlationStructure) != 1L) {
        stop("correlationStructure must contain one value.",
            call. = FALSE
        )
    }

    structure_value <- tolower(trimws(as.character(correlationStructure[[1L]])))
    if (structure_value %in% c("", "none", "null", "na")) {
        return("none")
    }
    if (identical(structure_value, "ar1")) {
        return("AR1")
    }
    if (identical(structure_value, "car1")) {
        return("CAR1")
    }

    stop("correlationStructure must be one of: none, AR1, CAR1.",
        call. = FALSE
    )
}

normalizeCorrelationVariableMethylationLME <- function(correlationVar = NULL) {
    if (is.null(correlationVar) || length(correlationVar) ==
        0L || all(is.na(correlationVar))) {
        return(NULL)
    }

    parsed <- splitOptionMinfiEwasWater(correlationVar, sep = ",")
    if (length(parsed) == 0L) {
        return(NULL)
    }
    if (length(parsed) > 1L) {
        stop("correlationVar must contain a single variable name.",
            call. = FALSE
        )
    }

    parsed <- trimws(as.character(parsed[[1L]]))
    if (!nzchar(parsed) || tolower(parsed) %in% c("null", "na")) {
        return(NULL)
    }

    parsed
}

resolveLmeLibrariesMethylationLME <- function(lmeLibs = "lme4,lmerTest") {
    requested <- splitOptionMinfiEwasWater(lmeLibs, sep = ",")
    if (length(requested) == 0L) {
        requested <- c("lme4", "lmerTest")
    }

    requested_lower <- tolower(requested)
    has_nlme <- "nlme" %in% requested_lower
    has_lme4 <- any(requested_lower %in% c("lme4", "lmertest"))

    if (isTRUE(has_nlme) && isTRUE(has_lme4)) {
        stop("lmeLibs must choose either 'lme4,lmerTest' or 'nlme', not both.",
            call. = FALSE
        )
    }

    if (isTRUE(has_nlme)) {
        return(list(
            engine = "nlme", requestedPackages = requested,
            requiredPackages = "nlme"
        ))
    }

    if (isTRUE(has_lme4)) {
        return(list(
            engine = "lme4", requestedPackages = requested,
            requiredPackages = unique(c(requested, "lme4", "lmerTest"))
        ))
    }

    stop("lmeLibs must contain either 'lme4'/'lmerTest' or 'nlme'.",
        call. = FALSE
    )
}

coerceCorrelationTimeMethylationLME <- function(x) {
    if (is.numeric(x)) {
        return(as.numeric(x))
    }

    character_x <- trimws(as.character(x))
    non_missing <- !is.na(x)
    numeric_pattern <- "^[+-]?(?:\\d+(?:\\.\\d*)?|\\.\\d+)(?:[eE][+-]?\\d+)?$"
    is_numeric_text <- grepl(numeric_pattern, character_x)

    if (all(is_numeric_text[non_missing])) {
        numeric_x <- rep(NA_real_, length(x))
        numeric_x[non_missing] <- as.numeric(character_x[non_missing])
        return(numeric_x)
    }

    stop("correlationVar must be numeric or contain numeric text; categorical values cannot define AR1/CAR1 spacing.",
        call. = FALSE
    )
}

addCorrelationTimeVariableMethylationLME <- function(
    modelData,
    correlationVar
) {
    correlation_time_var <- "dnaEPICO_lme_correlation_time"
    while (correlation_time_var %in% colnames(modelData)) {
        correlation_time_var <- paste0(
            correlation_time_var,
            "_"
        )
    }

    modelData[[correlation_time_var]] <-
        coerceCorrelationTimeMethylationLME(modelData[[correlationVar]])

    list(data = modelData, correlationTimeVar = correlation_time_var)
}

validateCorrelationTimeMethylationLME <- function(
    modelData,
    correlationTimeVar, personVar, correlationStructure
) {
    time_values <- modelData[[correlationTimeVar]]
    person_values <- modelData[[personVar]]

    if (anyNA(time_values) || any(!is.finite(time_values))) {
        stop("correlationVar contains missing or non-finite values after conversion.",
            call. = FALSE
        )
    }
    person_time <- interaction(person_values, time_values, drop = TRUE)
    if (anyDuplicated(person_time)) {
        stop("correlationVar must be unique within each subject for AR1/CAR1 models.",
            call. = FALSE
        )
    }
    if (identical(correlationStructure, "AR1") && any(abs(time_values -
        round(time_values)) > sqrt(.Machine$double.eps))) {
        stop("AR1 requires integer-valued correlationVar observations.",
            call. = FALSE
        )
    }

    invisible(TRUE)
}

buildCorrelationMethylationLME <- function(
    correlationStructure = "none",
    correlationTimeVar = NULL, personVar
) {
    if (identical(correlationStructure, "none")) {
        return(NULL)
    }

    if (is.null(correlationTimeVar) || !nzchar(correlationTimeVar)) {
        stop("An internal correlation time variable is required for AR1/CAR1 models.",
            call. = FALSE
        )
    }

    correlation_formula <- stats::as.formula(paste(
        "~", quoteNamesMethylationGLM(correlationTimeVar),
        "|", quoteNamesMethylationGLM(personVar)
    ))

    switch(correlationStructure,
        AR1 = nlme::corAR1(form = correlation_formula),
        CAR1 = nlme::corCAR1(form = correlation_formula),
        stop("Unsupported correlation structure.",
            call. = FALSE
        )
    )
}

coerceCoefficientTableMethylationLME <- function(coefTable, engine = "lme4") {
    coef_table <- as.data.frame(coefTable, check.names = FALSE)

    if (identical(engine, "nlme")) {
        rename_map <- c(
            Value = "Estimate", Std.Error = "Std. Error",
            DF = "df", `t-value` = "t value", `p-value` = "Pr(>|t|)"
        )
        for (old_name in names(rename_map)) {
            if (old_name %in% colnames(coef_table)) {
                colnames(coef_table)[colnames(coef_table) ==
                    old_name] <- rename_map[[old_name]]
            }
        }
    }

    as.matrix(coef_table)
}

#' Find coefficient rows for a longitudinal phenotype effect or interaction
#'
#' @param coefNames Character vector of coefficient names.
#' @param phenotype Character. Phenotype variable of interest.
#' @param interactionTerm Character or `NULL`. Optional interaction term.
#'
#' @return Character vector of matching coefficient names.
#'
#' @description
#' Internal helper that identifies the longitudinal fixed-effect terms that
#' should be extracted from each fitted mixed-effects model.
#'
#' @keywords internal
#' @noRd
findCoefficientRowsMethylationLME <- function(
    coefNames, phenotype,
    interactionTerm = NULL, coefficientTerms = NULL
) {
    findCoefficientRowsMethylationGLM(
        coefNames = coefNames,
        variable = phenotype, interactionTerm = interactionTerm,
        coefficientTerms = coefficientTerms
    )
}

fitStatusValuesMethylationLME <- function(modelObj) {
    if (inherits(modelObj, "dnaEPICO_methylationLME_fit_error")) {
        return(list(
            status = if (is.null(modelObj$status)) "failed" else as.character(modelObj$status),
            singular = NA,
                converged = if (is.null(modelObj$converged)) NA else as.logical(modelObj$converged)[[1L]],
            convergenceMessage = if (is.null(modelObj$convergenceMessage)) {
                NA_character_
            } else {
                as.character(modelObj$convergenceMessage)[[1L]]
            }, warning = NA_character_
        ))
    }

    fit_status <- modelObj$fitStatus
    singular_fit <- if (is.null(fit_status$singular)) {
        NA
    } else {
        as.logical(fit_status$singular)[[1L]]
    }
    converged <- if (is.null(fit_status$converged)) {
        length(fit_status$convergenceMessages) == 0L
    } else {
        isTRUE(fit_status$converged)
    }
    convergence_messages <- unique(as.character(fit_status$convergenceMessages))
    convergence_messages <- convergence_messages[!is.na(convergence_messages) &
        nzchar(convergence_messages)]
    fit_warnings <- unique(as.character(fit_status$warnings))
    fit_warnings <- fit_warnings[!is.na(fit_warnings) & nzchar(fit_warnings)]

    list(status = if (!isTRUE(converged)) {
        "fitted_not_converged"
    } else if (isTRUE(singular_fit)) {
        "fitted_singular"
    } else if (length(fit_warnings) > 0L) {
        "fitted_with_warning"
    } else {
        "fitted"
    }, singular = singular_fit, converged = converged,
        convergenceMessage = if (length(convergence_messages) >
        0L) {
        paste(convergence_messages, collapse = " | ")
    } else {
        NA_character_
    }, warning = if (length(fit_warnings) > 0L) {
        paste(fit_warnings, collapse = " | ")
    } else {
        NA_character_
    })
}

collectFitDiagnosticsMethylationLME <- function(fits) {
    rows <- list()
    row_index <- 1L
    for (phenotype in names(fits)) {
        fit_group <- fits[[phenotype]]
        for (cpg in names(fit_group)) {
            fit_object <- fit_group[[cpg]]
            values <- fitStatusValuesMethylationLME(fit_object)
            inference_included <- startsWith(values$status, "fitted")
            rows[[row_index]] <- data.frame(
                Phenotype = phenotype,
                CpG = cpg, Fit.Status = values$status,
                    Singular.Fit = values$singular,
                Converged = values$converged,
                    Convergence.Message = values$convergenceMessage,
                Fit.Warning = values$warning,
                    Inference.Included = inference_included,
                Exclusion.Reason = if (!inference_included &&
                    !is.null(fit_object$error)) {
                    as.character(fit_object$error)
                } else {
                    NA_character_
                }, stringsAsFactors = FALSE, check.names = FALSE
            )
            row_index <- row_index + 1L
        }
    }

    if (length(rows) == 0L) {
        return(data.frame(
            Phenotype = character(0), CpG = character(0),
            Fit.Status = character(0), Singular.Fit = logical(0),
            Converged = logical(0), Convergence.Message = character(0),
            Fit.Warning = character(0), Inference.Included = logical(0),
            Exclusion.Reason = character(0), stringsAsFactors = FALSE,
            check.names = FALSE
        ))
    }

    diagnostics <- do.call(rbind, rows)
    rownames(diagnostics) <- NULL
    diagnostics
}

#' Summarize a single CpG-level mixed-effects model fit
#'
#' @param cpg Character. CpG identifier.
#' @param modelObj List returned by `fitCpGModelMethylationLME()`.
#' @param phenotype Character. Phenotype variable of interest.
#' @param interactionTerm Character or `NULL`. Optional interaction term.
#'
#' @return A data frame for the requested CpG, or `NULL`.
#'
#' @description
#' Internal helper that extracts phenotype-specific fixed-effect rows from a
#' single CpG-level mixed-effects model fit.
#'
#' @keywords internal
#' @noRd
summarizeCpGFitMethylationLME <- function(
    cpg, modelObj, phenotype,
    interactionTerm = NULL
) {
    if (is.null(modelObj) || inherits(modelObj,
        "dnaEPICO_methylationLME_fit_error")) {
        return(NULL)
    }

    coef_table <- modelObj$coef
    if (is.null(coef_table)) {
        return(NULL)
    }

    matched_terms <- findCoefficientRowsMethylationLME(
        coefNames = rownames(coef_table),
        phenotype = phenotype, interactionTerm = interactionTerm,
        coefficientTerms = modelObj$coefficientTerms
    )
    if (length(matched_terms) == 0L) {
        return(NULL)
    }

    fit_status <- fitStatusValuesMethylationLME(modelObj)

    do.call(rbind, lapply(matched_terms, function(term) {
        coef_row <- coef_table[term, ]
        data.frame(
            CpG = cpg, Interaction.Term = term,
                Estimate = unname(coef_row["Estimate"]),
            Std.Error = unname(coef_row["Std. Error"]),
                t.value = unname(coef_row["t value"]),
            P.value = unname(coef_row["Pr(>|t|)"]),
                Fit.Status = fit_status$status,
            Singular.Fit = fit_status$singular,
                Converged = fit_status$converged,
            Convergence.Message = fit_status$convergenceMessage,
            Fit.Warning = fit_status$warning, Inference.Included = TRUE,
            Exclusion.Reason = NA_character_, stringsAsFactors = FALSE,
            row.names = NULL
        )
    }))
}

applyFitQualityExclusionsMethylationLME <- function(
    summaryDf,
    excludeSingular = FALSE, excludeNonConverged = FALSE
) {
    summary_df <- summaryDf
    if (is.null(summary_df) || nrow(summary_df) == 0L) {
        return(summary_df)
    }
    reasons <- rep(NA_character_, nrow(summary_df))
    singular <- rep(FALSE, nrow(summary_df))
    if (isTRUE(excludeSingular) && "Singular.Fit" %in% names(summary_df)) {
        singular_values <- as.logical(summary_df[["Singular.Fit"]])
        singular <- !is.na(singular_values) & singular_values
    }
    not_converged <- rep(FALSE, nrow(summary_df))
    if (isTRUE(excludeNonConverged) && "Converged" %in% names(summary_df)) {
        converged <- as.logical(summary_df[["Converged"]])
        not_converged <- !is.na(converged) & !converged
    }
    reasons[singular] <- "singular random-effects fit"
    reasons[not_converged] <- ifelse(is.na(reasons[not_converged]),
        "model did not converge", paste(reasons[not_converged],
            "model did not converge",
            sep = "; "
        )
    )
    excluded <- singular | not_converged
    summary_df$Inference.Included <- !excluded
    summary_df$Exclusion.Reason <- reasons
    if (any(excluded) && "P.value" %in% names(summary_df)) {
        summary_df$P.value[excluded] <- NA_real_
    }
    summary_df
}

filterSummaryByPvalueMethylationLME <- function(summaryDf, pValueFilter) {
    summary_df <- summaryDf
    if (is.null(summary_df) || nrow(summary_df) == 0L) {
        return(data.frame())
    }

    if (nrow(summary_df) > 0L && !is.na(pValueFilter)) {
        keep <- is.finite(summary_df$P.value) & summary_df$P.value <
            pValueFilter
        summary_df <- summary_df[keep, , drop = FALSE]
    }
    rownames(summary_df) <- NULL

    summary_df
}

fitCpGModelMethylationLME <- function(
    cpg, cpgValues, modelData,
    formulaText, personVar, lmeEngine = "lme4", correlationStructure = "none",
    correlationTimeVar = NULL, responseVar = "beta"
) {
    tryCatch(
        {
            model_data <- modelData
            model_data[[responseVar]] <- as.numeric(cpgValues)
            observed_response <-
                model_data[[responseVar]][is.finite(model_data[[responseVar]])]
            if (length(unique(observed_response)) < 2L) {
                stop("The CpG response has no observed variation.",
                    call. = FALSE
                )
            }

            warning_state <- new.env(parent = emptyenv())
            warning_state$messages <- character(0)

            if (identical(lmeEngine, "nlme")) {
                fit <- withCallingHandlers(
                    nlme::lme(
                        fixed = stats::as.formula(formulaText),
                        random = stats::as.formula(paste("~ 1 |",
                            quoteNamesMethylationGLM(personVar))),
                        correlation = buildCorrelationMethylationLME(
                            correlationStructure = correlationStructure,
                            correlationTimeVar = correlationTimeVar,
                                personVar = personVar
                        ),
                        data = model_data, na.action = stats::na.exclude,
                        method = "REML",
                            control = nlme::lmeControl(returnObject = TRUE)
                    ),
                    warning = function(condition) {
                        warning_state$messages <- c(
                            warning_state$messages,
                            conditionMessage(condition)
                        )
                        invokeRestart("muffleWarning")
                    }
                )
                coef_table <-
                    coerceCoefficientTableMethylationLME(summary(fit)$tTable,
                    engine = lmeEngine
                )
                ranef_values <- nlme::ranef(fit)
                fixef_values <- nlme::fixef(fit)
                singular_fit <- NA
                fit_warnings <- warning_state$messages
                convergence_messages <- fit_warnings
            } else {
                fit <- withCallingHandlers(lmerTest::lmer(
                    formula = stats::as.formula(formulaText),
                    data = model_data, na.action = stats::na.exclude,
                    REML = TRUE
                ), warning = function(condition) {
                    warning_state$messages <- c(
                        warning_state$messages,
                        conditionMessage(condition)
                    )
                    invokeRestart("muffleWarning")
                })
                fit_warnings <- warning_state$messages
                coef_table <-
                    coerceCoefficientTableMethylationLME(summary(fit)$coefficients,
                    engine = lmeEngine
                )
                ranef_values <- lme4::ranef(fit)
                fixef_values <- lme4::fixef(fit)
                singular_fit <- lme4::isSingular(fit, tol = 1e-04)
                raw_optimizer_codes <- unlist(fit@optinfo$conv$opt,
                    use.names = FALSE
                )
                optimizer_codes <- vapply(raw_optimizer_codes, function(value) {
                    converted <- utils::type.convert(as.character(value),
                        as.is = TRUE, na.strings = character()
                    )
                    if (is.numeric(converted)) {
                        as.numeric(converted)
                    } else {
                        NA_real_
                    }
                }, numeric(1))
                optimizer_codes <- optimizer_codes[is.finite(optimizer_codes) &
                    optimizer_codes != 0]
                optimizer_messages <- if (length(optimizer_codes) >
                    0L) {
                    paste0("lme4 optimizer convergence code: ",
                        paste(unique(optimizer_codes),
                        collapse = ", "
                    ))
                } else {
                    character(0)
                }
                convergence_messages <- c(
                    fit_warnings, fit@optinfo$conv$lme4$messages,
                    optimizer_messages
                )
            }

            fit_warnings <- unique(as.character(convergence_messages))
            fit_warnings <-
                fit_warnings[!is.na(fit_warnings) & nzchar(fit_warnings)]
            convergence_messages <-
                fit_warnings[grepl("converg|unable to evaluate|degenerate|failed to|gradient|iteration limit",
                fit_warnings,
                ignore.case = TRUE
            )]

            required_statistics <- c(
                "Estimate", "Std. Error", "t value",
                "Pr(>|t|)"
            )
            if (!all(required_statistics %in% colnames(coef_table)) ||
                any(!is.finite(as.matrix(coef_table[, required_statistics,
                    drop = FALSE
                ])))) {
                stop("The mixed model returned missing or non-finite coefficient statistics.",
                    call. = FALSE
                )
            }

            coefficient_terms <- buildCoefficientTermMapMethylationModels(
                formulaText = formulaText,
                data = model_data, removeRandomEffects = !identical(
                    lmeEngine,
                    "nlme"
                )
            )

            list(
                coef = coef_table, residuals = stats::residuals(fit),
                fitted = stats::fitted(fit), ranef = ranef_values,
                fixef = fixef_values, coefficientTerms = coefficient_terms,
                fitStatus = list(singular = singular_fit,
                    converged = length(convergence_messages) ==
                    0L, warnings = fit_warnings,
                        convergenceMessages = convergence_messages)
            )
        },
        error = function(error) {
            reason <- conditionMessage(error)
            convergence_failure <-
                grepl("converg|unable to evaluate|degenerate|failed to|gradient|iteration limit",
                reason,
                ignore.case = TRUE
            )
            newMethylationFitErrorDnaEpico(
                reason = reason,
                    errorClass = "dnaEPICO_methylationLME_fit_error",
                status = if (convergence_failure) {
                    "not_converged"
                } else {
                    "failed"
                }, converged = if (convergence_failure) {
                    FALSE
                } else {
                    NA
                }, convergenceMessage = if (convergence_failure) {
                    reason
                } else {
                    NA_character_
                }
            )
        }
    )
}

fitMethylationLMEBatch <- function(
    cpgBatch, data, modelData,
    formulaText, personVar, lmeEngine = "lme4", correlationStructure = "none",
    correlationTimeVar = NULL, phenotype, interactionTerm = NULL,
    responseVar = "beta", invalidCpgReasons = character(0)
) {
    fits <- vector("list", length(cpgBatch))
    names(fits) <- cpgBatch
    summaries <- vector("list", length(cpgBatch))
    names(summaries) <- cpgBatch

    for (cpg in cpgBatch) {
        invalid_reason <- if (cpg %in% names(invalidCpgReasons)) {
            invalidCpgReasons[[cpg]]
        } else {
            NULL
        }
        model_obj <- if (!is.null(invalid_reason)) {
            newMethylationFitErrorDnaEpico(
                reason = invalid_reason,
                errorClass = "dnaEPICO_methylationLME_fit_error",
                status = "invalid"
            )
        } else {
            fitCpGModelMethylationLME(
                cpg = cpg, cpgValues = data[[cpg]],
                modelData = modelData, formulaText = formulaText,
                personVar = personVar, lmeEngine = lmeEngine,
                correlationStructure = correlationStructure,
                correlationTimeVar = correlationTimeVar,
                    responseVar = responseVar
            )
        }
        fits[[cpg]] <- model_obj
        summaries[[cpg]] <- summarizeCpGFitMethylationLME(
            cpg = cpg,
            modelObj = model_obj, phenotype = phenotype,
                interactionTerm = interactionTerm
        )
    }

    summaries <- Filter(Negate(is.null), summaries)
    summary_df <- if (length(summaries) == 0L) {
        data.frame()
    } else {
        out <- do.call(rbind, summaries)
        rownames(out) <- NULL
        out
    }

    list(fits = fits, summaries = summary_df)
}

#' Summarize phenotype values by timepoint for longitudinal methylation analyses
#'
#' @param data Data frame containing the longitudinal phenotype-plus-beta data.
#' @param timeVar Character. Name of the time variable.
#' @param phenotypes Character vector or comma-separated string of phenotype
#'   variables to summarize.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return A data frame with one row per timepoint and summary columns for each
#'   requested phenotype.
#'
#' @description
#' Summarize the requested phenotype variables by timepoint. Numeric phenotypes
#' are reported with mean, standard deviation, and non-missing counts;
#' non-numeric
#' phenotypes are reported with non-missing counts and the observed levels.
#'
#' @examples
#' ex <- dnaEPICO:::exampleMethylationLMEStateDnaEpico()
#' timepoint_summary <- summarizeTimepointsMethylationLME(
#'     data = ex$preparedData$data,
#'     timeVar = "Timepoint",
#'     phenotypes = "score",
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#' nrow(timepoint_summary)
#'
#' @export
summarizeTimepointsMethylationLME <- function(
    data, timeVar = "Timepoint",
    phenotypes, verbose = FALSE, logs = FALSE, log_dir = NULL,
    log_file = "log_methylationLME.txt"
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )
    phenotype_list <- unique(splitOptionMinfiEwasWater(phenotypes,
        sep = ","
    ))

    if (!(timeVar %in% colnames(data))) {
        stop("timeVar column not found: ", timeVar, call. = FALSE)
    }

    missing_phenotypes <- setdiff(phenotype_list, colnames(data))
    if (length(missing_phenotypes) > 0L) {
        stop("Phenotype columns not found for timepoint summary: ",
            paste(missing_phenotypes, collapse = ", "),
            call. = FALSE
        )
    }

    split_data <- split(data, as.character(data[[timeVar]]))
    summaries <- lapply(split_data, function(df) {
        out <- list()
        out[[timeVar]] <- as.character(df[[timeVar]][[1L]])

        for (phenotype in phenotype_list) {
            values <- df[[phenotype]]
            if (is.numeric(values)) {
                finite_values <- values[is.finite(values)]
                out[[paste0(phenotype, "_mean")]] <-
                    if (length(finite_values)) {
                    mean(finite_values)
                } else {
                    NA_real_
                }
                out[[paste0(phenotype, "_sd")]] <- if (length(finite_values) >
                    1L) {
                    stats::sd(finite_values)
                } else {
                    NA_real_
                }
                out[[paste0(phenotype, "_n")]] <- length(finite_values)
            } else {
                observed_levels <- unique(as.character(values[!is.na(values)]))
                out[[paste0(phenotype, "_n")]] <- sum(!is.na(values))
                out[[paste0(phenotype, "_levels")]] <- paste(observed_levels,
                    collapse = ","
                )
            }
        }

        as.data.frame(out, stringsAsFactors = FALSE)
    })
    summary_df <- do.call(rbind, summaries)
    rownames(summary_df) <- NULL

    emitLogMinfiEwasWater(
        c(
            "============================================================",
            "Summary statistics for phenotype scores by timepoint:",
            previewLinesMinfiEwasWater(summary_df),
                "============================================================"
        ),
        verbose = verbose, log_path = log_path
    )

    summary_df
}

#' Prepare longitudinal phenotype-plus-methylation data for mixed-effects
#' analyses
#'
#' @param inputPheno Character. Path to the merged longitudinal phenotype-plus-
#'   methylation object created by `preprocessingPheno()`.
#' @param personVar Character. Name of the subject identifier column.
#' @param timeVar Character. Name of the time variable.
#' @param phenotypes Character vector or comma-separated string of phenotype
#'   variables to model.
#' @param covariates Character vector or comma-separated string of covariate
#'   variables to adjust for.
#' @param factorVars Character vector or comma-separated string of variables
#' that
#'   should be converted to factors before modeling.
#' @param scaleVars Character vector, comma-separated variable names, or `NULL`.
#'   Numeric fixed-effect variables to standardize before model fitting.
#' @param prsMap Character vector or comma-separated string of phenotype-to-PRS
#'   mappings in the form `'Phenotype:PRS'`.
#' @param cpgPrefix Character. Prefix used to identify methylation columns.
#' @param cpgLimit Integer or `NA`. Maximum number of CpGs to retain. `NA`
#'   keeps all matching CpGs.
#' @param methylationScale Character. Methylation metric represented by the CpG
#'   columns. One of `'Beta'`, `'M'`, or `'CN'`, case-insensitively.
#' @param interactionTerm Character or `NULL`. Optional interaction term.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return A list with class `'dnaEPICO_methylationLME_data'` containing
#'   the prepared analysis data, parsed variable selections, CpG columns,
#'   timepoint summaries, and subject-ID diagnostics.
#'
#' @description
#' Load the merged longitudinal phenotype-plus-methylation object, ensure that
#' a subject
#' identifier column is available, validate the requested modeling variables,
#' convert selected variables to factors, and return a single in-memory object
#' for downstream mixed-effects modeling helpers.
#'
#' @examples
#' ex <- dnaEPICO:::exampleMethylationLMEStateDnaEpico()
#' prepared_data <- prepareMethylationLMEData(
#'     inputPheno = ex$inputPath,
#'     personVar = "person",
#'     timeVar = "Timepoint",
#'     phenotypes = "score",
#'     covariates = "sex",
#'     factorVars = "sex",
#'     cpgLimit = 2,
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#' names(prepared_data)
#'
#' @export
prepareMethylationLMEData <- function(
    inputPheno, personVar = "person",
    timeVar = "Timepoint", phenotypes, covariates, factorVars,
    scaleVars = NULL, prsMap = NULL, cpgPrefix = "cg", cpgLimit = NA,
    methylationScale = "beta", interactionTerm = NULL, verbose = FALSE,
    logs = FALSE, log_dir = NULL, log_file = "log_methylationLME.txt"
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )
    phenotype_list <- unique(splitOptionMinfiEwasWater(phenotypes,
        sep = ","
    ))
    covariate_list <- splitOptionMinfiEwasWater(covariates, sep = ",")
    factor_list <- splitOptionMinfiEwasWater(factorVars, sep = ",")
    scale_list <- normalizeScaleVariablesDnaEpico(scaleVars)
    prs_map <- parsePrsMapMethylationGLM(prsMap)
    cpg_limit <-
        validateCpgLimitMethylationModels(normalizeOptionalNumericMethylationGLM(cpgLimit))
    cpgPrefix <- validateCpgPrefixDnaEpico(cpgPrefix)
    methylation_scale <- normalizeMethylationScaleDnaEpico(methylationScale)
    methylation_label <-
        methylationScaleResponseLabelDnaEpico(methylation_scale)
    methylation_prefix <-
        methylationScaleObjectPrefixDnaEpico(methylation_scale)
    response_column <- methylationScaleResponseColumnDnaEpico(methylation_scale)
    input_object_name <- sub("[.][^.]+$", "", basename(inputPheno))
    scale_named_input <- if (startsWith(input_object_name,
        methylation_prefix)) {
        input_object_name
    } else {
        character(0)
    }
    legacy_input_name <- if (identical(methylation_scale, "beta")) {
        "phenoBT1T2"
    } else {
        character(0)
    }
    analysis_data <- loadSavedObjectPreprocessingPheno(inputPheno,
        preferred_name = c(scale_named_input, paste0(
            methylation_prefix,
            "T1T2"
        ), legacy_input_name, input_object_name)
    )

    if (!is.data.frame(analysis_data)) {
        analysis_data <- as.data.frame(analysis_data, stringsAsFactors = FALSE)
    }

    if (length(phenotype_list) == 0L) {
        stop("At least one phenotype must be supplied.", call. = FALSE)
    }

    person_data <- ensurePersonColumnMethylationLME(
        data = analysis_data,
        personVar = personVar
    )
    analysis_data <- person_data$data

    if (!(timeVar %in% colnames(analysis_data))) {
        stop("timeVar column not found in inputPheno: ", timeVar,
            call. = FALSE
        )
    }
    time_values <- analysis_data[[timeVar]]
    invalid_time <-
        is.na(time_values) | !nzchar(trimws(as.character(time_values)))
    if (is.numeric(time_values)) {
        invalid_time <- invalid_time | !is.finite(time_values)
    }
    if (any(invalid_time)) {
        stop("timeVar contains missing, blank, or non-finite values.",
            call. = FALSE
        )
    }

    missing_phenotypes <- setdiff(phenotype_list, colnames(analysis_data))
    if (length(missing_phenotypes) > 0L) {
        stop("Phenotype columns not found in inputPheno: ",
            paste(missing_phenotypes,
            collapse = ", "
        ), call. = FALSE)
    }

    missing_covariates <- setdiff(covariate_list, colnames(analysis_data))
    if (length(missing_covariates) > 0L) {
        stop("Covariate columns not found in inputPheno: ",
            paste(missing_covariates,
            collapse = ", "
        ), call. = FALSE)
    }

    mapped_prs <- unname(prs_map[names(prs_map) %in% phenotype_list])
    missing_prs <- setdiff(mapped_prs, colnames(analysis_data))
    if (length(missing_prs) > 0L) {
        stop("PRS columns not found in inputPheno: ", paste(unique(missing_prs),
            collapse = ", "
        ), call. = FALSE)
    }

    resolved_interaction <- normalizeOptionalColumnMethylationModels(
        interactionTerm,
        "interactionTerm"
    )
    if (!is.null(resolved_interaction) && nzchar(resolved_interaction)) {
        if (!(resolved_interaction %in% colnames(analysis_data))) {
            stop("interactionTerm column not found in inputPheno: ",
                resolved_interaction,
                call. = FALSE
            )
        }
    }

    missing_factor_vars <- setdiff(factor_list, colnames(analysis_data))
    if (length(missing_factor_vars) > 0L) {
        stop("Factor columns not found in inputPheno: ",
            paste(missing_factor_vars,
            collapse = ", "
        ), call. = FALSE)
    }

    for (var in intersect(c(personVar, factor_list), colnames(analysis_data))) {
        if (identical(var, personVar)) {
            next
        }
        analysis_data[[var]] <- as.factor(analysis_data[[var]])
    }

    cpg_columns <- grep(paste0("^", escapeRegexMethylationGLM(cpgPrefix)),
        colnames(analysis_data),
        value = TRUE
    )
    if (!is.na(cpg_limit)) {
        cpg_columns <- utils::head(cpg_columns, as.integer(cpg_limit))
    }
    if (length(cpg_columns) == 0L) {
        stop("No CpG columns were found with prefix '", cpgPrefix,
            "'.",
            call. = FALSE
        )
    }
    validateMethylationProbeIdentifiersDnaEpico(
        cpg_columns,
        "CpG columns in the LME input"
    )
    methylation_validation <- inspectMethylationColumnsDnaEpico(
        data = analysis_data,
        cpgColumns = cpg_columns, methylationScale = methylation_scale
    )
    analysis_data <- methylation_validation$data
    if (methylation_validation$boundaries$NaN.Converted > 0L) {
        warning(methylation_validation$boundaries$NaN.Converted,
            " NaN methylation values were converted to NA. See methylationIssues.",
            call. = FALSE
        )
    }
    if (nrow(methylation_validation$invalidCpGs) > 0L) {
        warning(nrow(methylation_validation$invalidCpGs),
            " invalid CpG columns will be reported as failed fits instead of stopping the analysis.",
            call. = FALSE
        )
    }

    fixed_effect_variables <- unique(c(
        phenotype_list, covariate_list,
        mapped_prs, resolved_interaction
    ))
    fixed_effect_variables <-
        fixed_effect_variables[!is.na(fixed_effect_variables) &
        nzchar(fixed_effect_variables)]
    scaled_cpgs <- intersect(scale_list, cpg_columns)
    if (length(scaled_cpgs) > 0L) {
        stop("CpG methylation response columns cannot be listed in scaleVars: ",
            paste(scaled_cpgs, collapse = ", "),
            call. = FALSE
        )
    }
    model_columns <- setdiff(colnames(analysis_data), cpg_columns)
    scaling <- scaleModelVariablesDnaEpico(
        data = analysis_data[,
            model_columns,
            drop = FALSE
        ], scaleVars = scale_list,
        factorVars = factor_list, eligibleVars = fixed_effect_variables,
        protectedVars = c(personVar, cpg_columns)
    )

    timepoint_summary <- summarizeTimepointsMethylationLME(
        data = analysis_data,
        timeVar = timeVar, phenotypes = phenotype_list, verbose = FALSE,
        logs = FALSE
    )

    log_lines <- c(
        "============================================================",
        paste(
            "Loaded longitudinal phenotype + methylation data from:",
            inputPheno
        ), paste(
            "Merged modeling object:            ",
            methylation_prefix, "*"
        ), paste(
            "Data dimensions:                  ",
            paste(dim(analysis_data), collapse = " x ")
        ), paste(
            "Person variable:                  ",
            personVar
        ), paste(
            "Time variable:                    ",
            timeVar
        ), paste(
            "Phenotypes:                       ",
            paste(phenotype_list, collapse = ", ")
        ), paste(
            "Covariates:                       ",
            paste(covariate_list, collapse = ", ")
        ), paste(
            "Factor variables:                 ",
            paste(factor_list, collapse = ", ")
        ), formatScalingMetadataLogDnaEpico(scaling$metadata),
        paste("CpG columns retained:             ", length(cpg_columns)),
        formatMethylationBoundariesLogDnaEpico(methylation_validation$boundaries),
        if (nrow(methylation_validation$issues) > 0L) {
            c("Methylation value issues:",
                previewLinesMinfiEwasWater(methylation_validation$issues))
        } else {
            "Methylation value issues:            none"
        }, if (isTRUE(person_data$personCreated)) {
            paste("Created person variable from SID: ", personVar)
        } else {
            paste("Person variable already present:  ", personVar)
        }, "Count of records per person ID:",
            previewLinesMinfiEwasWater(person_data$personCounts),
        paste("Values observed in", timeVar, ":"),
            previewLinesMinfiEwasWater(table(analysis_data[[timeVar]],
            useNA = "ifany"
        ))
    )

    if (!is.null(person_data$mappingPreview)) {
        log_lines <- c(
            log_lines, "Example mapping of SID to person ID:",
            previewLinesMinfiEwasWater(person_data$mappingPreview)
        )
    }

    log_lines <- c(
        log_lines, "Summary statistics for phenotype scores by timepoint:",
        previewLinesMinfiEwasWater(timepoint_summary),
            "============================================================"
    )
    emitLogMinfiEwasWater(log_lines, verbose = verbose, log_path = log_path)

    structure(
        list(
            data = analysis_data, modelData = scaling$data,
            personVar = personVar, timeVar = timeVar,
                phenotypes = phenotype_list,
            covariates = covariate_list, factorVars = factor_list,
            scaleVars = scaling$scaleVars, scalingMetadata = scaling$metadata,
            prsMap = prs_map, cpgColumns = cpg_columns, cpgPrefix = cpgPrefix,
            cpgLimit = cpg_limit, methylationScale = methylation_scale,
            responseLabel = methylation_label,
                methylationObjectPrefix = methylation_prefix,
            internalResponseColumn = response_column,
                interactionTerm = resolved_interaction,
            requestedInteractionTerm = interactionTerm,
                methylationBoundaries = methylation_validation$boundaries,
            methylationIssues = methylation_validation$issues,
                invalidCpGs = methylation_validation$invalidCpGs,
            personCreated = person_data$personCreated,
                personCounts = person_data$personCounts,
            personMappingPreview = person_data$mappingPreview,
                timepointSummary = timepoint_summary
        ),
        class = "dnaEPICO_methylationLME_data"
    )
}
#' Fit CpG-wise mixed-effects models for longitudinal methylation analyses
#'
#' @param preparedData Object returned by `prepareMethylationLMEData()`.
#' @param nCores Integer. Maximum number of worker processes to use. Automatic
#'   fitting uses the empirical lme4 or nlme crossover and caps workers by
#'   available CpGs and detected physical cores.
#' @param libPath Character vector or `NULL`. Optional library paths forwarded
#'   to worker processes.
#' @param lmeLibs Character vector or comma-separated string of package names to
#'   check on worker processes. The default is `'lme4,lmerTest'`.
#' @param correlationStructure Character. Residual correlation structure used
#'   when `lmeLibs` selects `'nlme'`. One of `'none'`, `'AR1'`, or `'CAR1'`.
#' @param correlationVar Character or `NULL`. Variable used to order repeated
#'   observations within `personVar` for `AR1` or `CAR1` residual correlation
#'   structures. Must be supplied explicitly for `AR1` or `CAR1`.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return A list with class `'dnaEPICO_methylationLME_models'`
#'   containing fitted model lists, model formulas, and counts of failed CpG
#'   fits.
#'
#' @description
#' Fit one linear mixed-effects model per CpG for each phenotype requested in
#' the
#' object returned by `prepareMethylationLMEData()`.
#'
#' @examples
#' ex <- dnaEPICO:::exampleMethylationLMEStateDnaEpico()
#' model_results <- fitMethylationLMEModels(
#'     preparedData = ex$preparedData,
#'     nCores = 1,
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#' names(model_results$fits)
#'
#' @export
fitMethylationLMEModels <- function(
    preparedData, nCores = 1L,
    libPath = NULL, lmeLibs = "lme4,lmerTest", correlationStructure = "none",
    correlationVar = NULL, verbose = FALSE, logs = FALSE, log_dir = NULL,
    log_file = "log_methylationLME.txt"
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )

    if (is.null(libPath)) {
        libPath <- .libPaths()
    }

    lme_config <- resolveLmeLibrariesMethylationLME(lmeLibs)
    lme_lib_list <- lme_config$requestedPackages
    required_lme_lib_list <- lme_config$requiredPackages
    lme_engine <- lme_config$engine
    correlation_structure <-
        normalizeCorrelationStructureMethylationLME(correlationStructure)
    correlation_var <-
        normalizeCorrelationVariableMethylationLME(correlationVar = correlationVar)
    if (!identical(lme_engine, "nlme") && !identical(
        correlation_structure,
        "none"
    )) {
        stop("correlationStructure can only be AR1 or CAR1 when lmeLibs selects 'nlme'.",
            call. = FALSE
        )
    }
    if (!identical(correlation_structure, "none")) {
        if (is.null(correlation_var)) {
            stop("correlationVar must be supplied when correlationStructure is AR1 or CAR1.",
                call. = FALSE
            )
        }
        if (correlation_var %in% preparedData$scaleVars) {
            stop("correlationVar cannot also be listed in scaleVars; provide a separate scaled fixed-effect column if required.",
                call. = FALSE
            )
        }
    }

    analysis_data <- preparedData$data
    model_data <- if (!is.null(preparedData$modelData)) {
        preparedData$modelData
    } else {
        analysis_data
    }
    cpg_columns <- preparedData$cpgColumns
    invalid_cpg_reasons <- invalidMethylationReasonsDnaEpico(preparedData)
    n_cores <- validatePositiveIntegerMethylationModels(
        nCores,
        "nCores"
    )
    fits <- list()
    summary_cache <- list()
    formulas <- stats::setNames(
        character(length(preparedData$phenotypes)),
        preparedData$phenotypes
    )
    failure_counts <- stats::setNames(
        integer(length(preparedData$phenotypes)),
        preparedData$phenotypes
    )
    failure_reasons <- list()
    singular_counts <- stats::setNames(
        integer(length(preparedData$phenotypes)),
        preparedData$phenotypes
    )
    convergence_issue_counts <- stats::setNames(
        integer(length(preparedData$phenotypes)),
        preparedData$phenotypes
    )
    parallel_plan <- resolveParallelPlanMethylationModels(
        engine = lme_engine,
        nCores = n_cores, nCpGs = length(cpg_columns)
    )
    backend <- parallel_plan$backend
    worker_count <- parallel_plan$workerCount
    cpg_batches <- chunkCpGColumnsMethylationModels(
        cpgColumns = cpg_columns,
        nCores = worker_count, batchesPerCore = 8L
    )
    psock_cluster <- NULL
    if (identical(backend, "psock") && length(cpg_batches) >
        1L) {
        psock_cluster <- makePsockClusterMethylationModels(min(
            worker_count,
            length(cpg_batches)
        ))
        on.exit(
            {
                if (!is.null(psock_cluster)) {
                    try(parallel::stopCluster(psock_cluster), silent = TRUE)
                }
            },
            add = TRUE
        )
        parallel::clusterExport(psock_cluster,
            varlist = c(
                "analysis_data",
                "invalid_cpg_reasons", "libPath", "required_lme_lib_list",
                "validateWorkerPackagesMethylationModels",
                    "newMethylationFitErrorDnaEpico",
                "fitMethylationLMEBatch", "fitCpGModelMethylationLME",
                "buildCorrelationMethylationLME",
                    "coerceCoefficientTableMethylationLME",
                "buildCoefficientTermMapMethylationModels",
                    "removeRandomInterceptMethylationModels",
                "summarizeCpGFitMethylationLME",
                    "fitStatusValuesMethylationLME",
                "findCoefficientRowsMethylationLME",
                    "findCoefficientRowsMethylationGLM",
                "quoteNamesMethylationGLM", "escapeRegexMethylationGLM"
            ),
            envir = environment()
        )
        parallel::clusterEvalQ(psock_cluster,
            validateWorkerPackagesMethylationModels(
            libPath = libPath,
            packages = required_lme_lib_list
        ))
    }

    for (phenotype in preparedData$phenotypes) {
        prs_var <- character(0)
        if (phenotype %in% names(preparedData$prsMap)) {
            prs_var <- unname(preparedData$prsMap[[phenotype]])
        }
        covariates <- unique(c(preparedData$covariates, prs_var))
        model_vars <- unique(c(
            preparedData$personVar, phenotype,
            covariates, preparedData$interactionTerm, if (!identical(
                correlation_structure,
                "none"
            )) {
                correlation_var
            } else {
                character(0)
            }
        ))
        model_vars <- model_vars[!is.na(model_vars) & nzchar(model_vars)]

        missing_vars <- setdiff(model_vars, colnames(model_data))
        if (length(missing_vars) > 0L) {
            stop("Model variables not found for phenotype ",
                phenotype, ": ", paste(missing_vars, collapse = ", "),
                call. = FALSE
            )
        }

        display_formula_text <- buildFormulaMethylationLME(
            phenotype = phenotype,
            personVar = preparedData$personVar, covariates = covariates,
            interactionTerm = preparedData$interactionTerm,
                includeRandomTerm = TRUE,
            responseVar = preparedData$internalResponseColumn
        )
        formula_text <- buildFormulaMethylationLME(
            phenotype = phenotype,
            personVar = preparedData$personVar, covariates = covariates,
            interactionTerm = preparedData$interactionTerm,
                includeRandomTerm = !identical(
                lme_engine,
                "nlme"
            ), responseVar = preparedData$internalResponseColumn
        )

        base_model_data <- model_data[, model_vars, drop = FALSE]
        factor_vars <- preparedData$factorVars
        person_var <- preparedData$personVar
        base_model_data[[person_var]] <-
            as.factor(base_model_data[[person_var]])
        for (var in intersect(factor_vars, colnames(base_model_data))) {
            base_model_data[[var]] <- as.factor(base_model_data[[var]])
        }
        validateFixedEffectDesignMethylationModels(
            formulaText = formula_text,
            data = base_model_data, removeRandomEffects = !identical(
                lme_engine,
                "nlme"
            )
        )
        correlation_time_var <- NULL
        if (!identical(correlation_structure, "none")) {
            correlation_data <- addCorrelationTimeVariableMethylationLME(
                modelData = base_model_data,
                correlationVar = correlation_var
            )
            base_model_data <- correlation_data$data
            correlation_time_var <- correlation_data$correlationTimeVar
            validateCorrelationTimeMethylationLME(
                modelData = base_model_data,
                correlationTimeVar = correlation_time_var,
                    personVar = person_var,
                correlationStructure = correlation_structure
            )
        }
        batch_worker <- fitMethylationLMEBatch
        resolved_interaction <- preparedData$interactionTerm
        response_var <- preparedData$internalResponseColumn

        if (!identical(backend, "serial") && length(cpg_batches) >
            1L) {
            cluster_size <- min(worker_count, length(cpg_batches))

            if (identical(backend, "fork")) {
                batch_results <- parallel::mclapply(cpg_batches,
                    function(batch) {
                        validateWorkerPackagesMethylationModels(
                            libPath = libPath,
                            packages = required_lme_lib_list
                        )
                        batch_worker(
                            cpgBatch = batch, data = analysis_data,
                            modelData = base_model_data,
                                formulaText = formula_text,
                            personVar = person_var, lmeEngine = lme_engine,
                            correlationStructure = correlation_structure,
                            correlationTimeVar = correlation_time_var,
                            phenotype = phenotype,
                                interactionTerm = resolved_interaction,
                            responseVar = response_var,
                                invalidCpgReasons = invalid_cpg_reasons
                        )
                    },
                    mc.cores = cluster_size, mc.preschedule = FALSE
                )
            } else {
                parallel::clusterExport(psock_cluster,
                    varlist = c(
                        "base_model_data",
                        "formula_text", "phenotype", "resolved_interaction",
                        "person_var", "lme_engine", "correlation_structure",
                        "correlation_time_var", "response_var", "batch_worker"
                    ),
                    envir = environment()
                )
                batch_results <- parallel::parLapplyLB(
                    psock_cluster,
                    cpg_batches, function(batch) {
                        batch_worker(
                            cpgBatch = batch, data = analysis_data,
                            modelData = base_model_data,
                                formulaText = formula_text,
                            personVar = person_var, lmeEngine = lme_engine,
                            correlationStructure = correlation_structure,
                            correlationTimeVar = correlation_time_var,
                            phenotype = phenotype,
                                interactionTerm = resolved_interaction,
                            responseVar = response_var,
                                invalidCpgReasons = invalid_cpg_reasons
                        )
                    }
                )
            }
        } else {
            validateWorkerPackagesMethylationModels(
                libPath = libPath,
                packages = required_lme_lib_list
            )
            batch_results <- lapply(cpg_batches, function(batch) {
                batch_worker(
                    cpgBatch = batch, data = analysis_data,
                    modelData = base_model_data, formulaText = formula_text,
                    personVar = person_var, lmeEngine = lme_engine,
                    correlationStructure = correlation_structure,
                    correlationTimeVar = correlation_time_var,
                    phenotype = phenotype,
                        interactionTerm = resolved_interaction,
                    responseVar = response_var,
                        invalidCpgReasons = invalid_cpg_reasons
                )
            })
        }

        combined_results <- combineFitBatchResultsMethylationModels(
            batchResults = batch_results,
            cpgColumns = cpg_columns
        )
        fit_list <- combined_results$fits
        phenotype_summary_cache <- filterSummaryByPvalueMethylationLME(
            summaryDf = combined_results$summaries,
            pValueFilter = NA_real_
        )
        failures <- vapply(fit_list, function(x) {
            inherits(
                x,
                "dnaEPICO_methylationLME_fit_error"
            )
        }, logical(1))
        error_counts <- summarizeFitErrorsMethylationModels(
            fitList = fit_list,
            errorClass = "dnaEPICO_methylationLME_fit_error"
        )
        valid_fits <- fit_list[!failures]
        singular_fits <- vapply(
            valid_fits, function(x) isTRUE(x$fitStatus$singular),
            logical(1)
        )
        convergence_issues <- vapply(valid_fits, function(x) {
            identical(
                x$fitStatus$converged,
                FALSE
            )
        }, logical(1))

        fits[[phenotype]] <- fit_list
        summary_cache[[phenotype]] <- phenotype_summary_cache
        formulas[[phenotype]] <- display_formula_text
        failure_counts[[phenotype]] <- sum(failures)
        failure_reasons[[phenotype]] <- error_counts
        singular_counts[[phenotype]] <- sum(singular_fits)
        convergence_issue_counts[[phenotype]] <- sum(convergence_issues)

        emitLogMinfiEwasWater(
            c(
                "============================================================",
                paste("Fitted phenotype:            ", phenotype),
                paste("Formula:                     ", display_formula_text),
                paste("Correlation structure:       ", correlation_structure),
                paste("Correlation variable:        ", if (identical(
                    correlation_structure,
                    "none"
                )) {
                    "None"
                } else {
                    correlation_var
                }), paste(
                    "CpGs attempted:              ",
                    length(cpg_columns)
                ), paste(
                    "Failed CpG fits:             ",
                    failure_counts[[phenotype]]
                ), paste(
                    "Singular CpG fits:           ",
                    singular_counts[[phenotype]]
                ), paste(
                    "Non-converged CpG fits:      ",
                    convergence_issue_counts[[phenotype]]
                ), paste(
                    "Top fit errors:              ",
                    formatFitErrorsMethylationModels(error_counts)
                ),
                paste("Parallel backend:            ", backend),
                paste("Effective workers:           ", worker_count),
                paste("Parallel crossover CpGs:     ",
                    parallel_plan$crossoverCpGs),
                paste("Parallel selection:          ", parallel_plan$reason),
                paste("Fit batches:                 ", length(cpg_batches)),
                paste("Fit batch size:              ",
                    if (length(cpg_batches) ==
                    0L) {
                    0L
                } else {
                    max(vapply(cpg_batches, length, integer(1)))
                }),
                paste("Fit-time summary rows cached:",
                    nrow(phenotype_summary_cache)),
                "============================================================"
            ),
            verbose = verbose, log_path = log_path
        )

        if (length(fit_list) > 0L && all(failures)) {
            warning("All CpG ", toupper(lme_engine),
                " fits failed for phenotype '",
                phenotype, "'. Top failure reasons: ",
                    formatFitErrorsMethylationModels(error_counts),
                ". The failure inventory was retained and the analysis continued.",
                call. = FALSE
            )
        }
    }

    fit_failures <- collectFitFailuresMethylationModels(
        fits = fits,
        errorClass = "dnaEPICO_methylationLME_fit_error"
    )
    fit_diagnostics <- collectFitDiagnosticsMethylationLME(fits)

    if (!is.null(psock_cluster)) {
        parallel::stopCluster(psock_cluster)
        psock_cluster <- NULL
    }

    structure(
        list(
            fits = fits, summaryCache = summary_cache,
            formulas = formulas, phenotypes = names(fits),
                failureCounts = failure_counts,
            failureReasons = failure_reasons, fitFailures = fit_failures,
            fitDiagnostics = fit_diagnostics,
                methylationBoundaries = preparedData$methylationBoundaries,
            methylationIssues = preparedData$methylationIssues,
                invalidCpGs = preparedData$invalidCpGs,
            singularCounts = singular_counts,
                convergenceIssueCounts = convergence_issue_counts,
            settings = list(
                nCores = n_cores, parallelBackend = backend,
                workerCount = worker_count,
                    resourceWorkerCap = parallel_plan$resourceWorkerCap,
                parallelCrossoverCpGs = parallel_plan$crossoverCpGs,
                parallelSelectionReason = parallel_plan$reason,
                    clusterReusedAcrossPhenotypes = identical(
                    backend,
                    "psock"
                ) && length(preparedData$phenotypes) >
                    1L, fitBatchCount = length(cpg_batches), libPath = libPath,
                lmeLibs = lme_lib_list, lmeEngine = lme_engine,
                    correlationStructure = correlation_structure,
                correlationVar = if (identical(
                    correlation_structure,
                    "none"
                )) {
                    NULL
                } else {
                    correlation_var
                }, methylationScale = preparedData$methylationScale,
                methylationObjectPrefix = preparedData$methylationObjectPrefix,
                responseLabel = preparedData$responseLabel,
                    internalResponseColumn = preparedData$internalResponseColumn,
                interactionTerm = preparedData$interactionTerm,
                    phenotypes = preparedData$phenotypes,
                covariates = preparedData$covariates,
                    factorVars = preparedData$factorVars,
                factorLevels = lapply(stats::setNames(
                    preparedData$factorVars,
                    preparedData$factorVars
                ), function(variable) levels(preparedData$data[[variable]])),
                scaleVars = preparedData$scaleVars,
                    scalingMetadata = preparedData$scalingMetadata,
                sampleCount = nrow(preparedData$data),
                    personVar = preparedData$personVar,
                timeVar = preparedData$timeVar
            ), responseLabel = preparedData$responseLabel
        ),
        class = "dnaEPICO_methylationLME_models"
    )
}

#' Summarize CpG-wise mixed-effects model fits for longitudinal analyses
#'
#' @param modelResults Object returned by `fitMethylationLMEModels()`.
#' @param preparedData Object returned by `prepareMethylationLMEData()`.
#' @param summaryPval Numeric or `NA`. Optional p-value filter applied to the
#'   returned summary tables. `NA` keeps all rows.
#' @param excludeSingular Logical. If `TRUE`, retain singular CpG rows and
#'   reasons but set their inferential p-values to `NA`.
#' @param excludeNonConverged Logical. If `TRUE`, retain non-converged CpG rows
#'   and reasons but set their inferential p-values to `NA`.
#' @param nCores Integer. Number of worker processes to use while extracting
#'   summary rows.
#' @param chunkSize Integer or `NULL`. Number of CpGs processed per parallel
#'   chunk. `NULL` chooses a value automatically.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return A list with class `'dnaEPICO_methylationLME_summaries'`
#'   containing the optionally filtered summary tables in `summaries` and the
#'   complete CpG-level tables in `diagnosticSummaries`. Diagnostics,
#'   annotation, and report output use the complete tables so `summaryPval`
#'   does not remove CpGs from those outputs. `fitDiagnostics` retains the
#'   phenotype-specific fit status, lme4 singular-fit flag, and warning text for
#'   every attempted CpG.
#'
#' @description
#' Extract phenotype-specific fixed-effect tables from the fitted mixed-effects
#' model object returned by `fitMethylationLMEModels()`.
#'
#' @examples
#' ex <- dnaEPICO:::exampleMethylationLMEStateDnaEpico()
#' summary_results <- summarizeMethylationLMEModels(
#'     modelResults = ex$modelResults,
#'     preparedData = ex$preparedData,
#'     summaryPval = NA,
#'     nCores = 1,
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#' names(summary_results$summaries)
#'
#' @export
summarizeMethylationLMEModels <- function(
    modelResults, preparedData,
    summaryPval = NA, excludeSingular = FALSE, excludeNonConverged = FALSE,
    nCores = 1L, chunkSize = NULL, verbose = FALSE, logs = FALSE,
    log_dir = NULL, log_file = "log_methylationLME.txt"
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )
    p_value_filter <- normalizeOptionalNumericMethylationGLM(summaryPval)
    p_value_filter <- validateProbabilityDnaEpico(p_value_filter,
        "summaryPval",
        allowNA = TRUE
    )
    chunk_size <- normalizeChunkSizeMethylationGLM(chunkSize)
    n_cores <- validatePositiveIntegerMethylationModels(
        nCores,
        "nCores"
    )
    summaries <- list()
    diagnostic_summaries <- list()
    lme_engine <- modelResults$settings$lmeEngine
    if (is.null(lme_engine) || !nzchar(lme_engine)) {
        lme_engine <- "lme4"
    }
    if (identical(lme_engine, "nlme") && isTRUE(excludeSingular)) {
        warning("excludeSingular is not applicable to nlme fits and was ignored.",
            call. = FALSE
        )
    }
    remove_inapplicable_singular <- function(data) {
        if (identical(lme_engine, "nlme") && "Singular.Fit" %in%
            colnames(data)) {
            data$Singular.Fit <- NULL
        }
        data
    }

    for (phenotype in names(modelResults$fits)) {
        if (!is.null(modelResults$summaryCache) &&
            !is.null(modelResults$summaryCache[[phenotype]])) {
            diagnostic_df <- filterSummaryByPvalueMethylationLME(
                summaryDf = modelResults$summaryCache[[phenotype]],
                pValueFilter = NA_real_
            )
            diagnostic_df <- applyFitQualityExclusionsMethylationLME(
                summaryDf = diagnostic_df,
                excludeSingular = excludeSingular && identical(
                    lme_engine,
                    "lme4"
                ), excludeNonConverged = excludeNonConverged
            )
            diagnostic_df <- remove_inapplicable_singular(diagnostic_df)
            summary_df <- filterSummaryByPvalueMethylationLME(
                summaryDf = diagnostic_df,
                pValueFilter = p_value_filter
            )
            diagnostic_summaries[[phenotype]] <- diagnostic_df
            summaries[[phenotype]] <- summary_df

            emitLogMinfiEwasWater(
                c(
                    "============================================================",
                    paste("Summarized phenotype:        ", phenotype),
                    paste("LME summary rows returned:   ", nrow(summary_df)),
                    "Summary source:              fit-time cache",
                    if (is.na(p_value_filter)) {
                        "P-value filter:              none"
                    } else {
                        paste("P-value filter:              ", p_value_filter)
                    }, "============================================================"
                ),
                verbose = verbose, log_path = log_path
            )
            next
        }

        fit_list <- modelResults$fits[[phenotype]]
        cpg_names <- names(fit_list)

        local_chunk_size <- chunk_size
        if (is.null(local_chunk_size)) {
            local_chunk_size <- max(10L, floor(length(cpg_names) / max(n_cores *
                4L, 1L)))
        }
        local_chunk_size <- max(1L, as.integer(local_chunk_size))
        cpg_chunks <- split(cpg_names,
            ceiling(seq_along(cpg_names) / local_chunk_size))

        summary_worker <- summarizeCpGFitMethylationLME
        resolved_interaction <- preparedData$interactionTerm

        if (n_cores > 1L && length(cpg_chunks) > 1L) {
            cluster_size <- min(n_cores, length(cpg_chunks))
            cl <- parallel::makeCluster(cluster_size)
            on.exit(parallel::stopCluster(cl), add = TRUE)

            parallel::clusterExport(cl, varlist = c(
                "fit_list",
                "phenotype", "resolved_interaction", "summary_worker",
                "fitStatusValuesMethylationLME"
            ), envir = environment())

            result_chunks <- parallel::parLapplyLB(
                cl, cpg_chunks,
                function(chunk) {
                    rows <- lapply(chunk, function(cpg) {
                        summary_worker(
                            cpg = cpg, modelObj = fit_list[[cpg]],
                            phenotype = phenotype,
                                interactionTerm = resolved_interaction
                        )
                    })
                    rows <- Filter(Negate(is.null), rows)
                    if (length(rows) == 0L) {
                        return(NULL)
                    }
                    do.call(rbind, rows)
                }
            )
            parallel::stopCluster(cl)
            on.exit(NULL, add = FALSE)
        } else {
            result_chunks <- lapply(cpg_chunks, function(chunk) {
                rows <- lapply(chunk, function(cpg) {
                    summary_worker(
                        cpg = cpg, modelObj = fit_list[[cpg]],
                        phenotype = phenotype,
                            interactionTerm = resolved_interaction
                    )
                })
                rows <- Filter(Negate(is.null), rows)
                if (length(rows) == 0L) {
                    return(NULL)
                }
                do.call(rbind, rows)
            })
        }

        result_chunks <- Filter(Negate(is.null), result_chunks)
        if (length(result_chunks) == 0L) {
            summary_df <- data.frame()
        } else {
            summary_df <- do.call(rbind, result_chunks)
            rownames(summary_df) <- NULL
        }

        summary_df <- applyFitQualityExclusionsMethylationLME(
            summaryDf = summary_df,
            excludeSingular = excludeSingular && identical(
                lme_engine,
                "lme4"
            ), excludeNonConverged = excludeNonConverged
        )
        summary_df <- remove_inapplicable_singular(summary_df)
        diagnostic_summaries[[phenotype]] <- summary_df
        if (nrow(summary_df) > 0L && !is.na(p_value_filter)) {
            keep <- is.finite(summary_df$P.value) & summary_df$P.value <
                p_value_filter
            summary_df <- summary_df[keep, , drop = FALSE]
            rownames(summary_df) <- NULL
        }

        summaries[[phenotype]] <- summary_df

        emitLogMinfiEwasWater(
            c(
                "============================================================",
                paste("Summarized phenotype:        ", phenotype),
                paste("LME summary rows returned:   ", nrow(summary_df)),
                paste("Summary chunk size:          ", local_chunk_size),
                if (is.na(p_value_filter)) {
                    "P-value filter:              none"
                } else {
                    paste("P-value filter:              ", p_value_filter)
                }, "============================================================"
            ),
            verbose = verbose, log_path = log_path
        )
    }

    structure(list(
        summaries = summaries, diagnosticSummaries = diagnostic_summaries,
        phenotypes = names(summaries), fitFailures = modelResults$fitFailures,
        fitDiagnostics = resolveFitDiagnosticsMethylationLME(modelSummaries = list(fitDiagnostics = if (!is.null(modelResults$fitDiagnostics)) {
            modelResults$fitDiagnostics
        } else {
            collectFitDiagnosticsMethylationLME(modelResults$fits)
        }, fitFailures = modelResults$fitFailures),
            summaryList = diagnostic_summaries),
        settings = list(
            summaryPval = p_value_filter,
                excludeSingular = isTRUE(excludeSingular),
            excludeNonConverged = isTRUE(excludeNonConverged),
            chunkSize = chunk_size,
                interactionTerm = preparedData$interactionTerm,
            lmeEngine = modelResults$settings$lmeEngine
        )
    ), class = "dnaEPICO_methylationLME_summaries")
}

#' Collect significant longitudinal model terms from fitted mixed-effects models
#'
#' @param modelResults Object returned by `fitMethylationLMEModels()`.
#' @param pvalThreshold Numeric. Threshold applied to the extracted phenotype or
#'   interaction p-values.
#' @param interactionTerm Character or `NULL`. Optional interaction term. When
#'   `NULL`, phenotype main effects are used.
#' @param excludeSingular Logical. If `TRUE`, do not collect significant CpGs
#'   with singular lme4 fits.
#' @param excludeNonConverged Logical. If `TRUE`, do not collect significant
#'   CpGs whose mixed model did not converge. Excluded CpG rows and reasons
#'   remain available in diagnostic and annotated output.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return A list with class `'dnaEPICO_methylationLME_significant'`
#'   containing the retained coefficient tables for each phenotype.
#'
#' @description
#' Collect raw coefficient tables for CpGs whose phenotype main effect or
#' requested interaction p-value passes the chosen threshold.
#'
#' @examples
#' ex <- dnaEPICO:::exampleMethylationLMEStateDnaEpico()
#' significant_hits <- collectSignificantInteractionsMethylationLME(
#'     modelResults = ex$modelResults,
#'     pvalThreshold = 1,
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#' names(significant_hits)
#'
#' @export
collectSignificantInteractionsMethylationLME <- function(
    modelResults,
    pvalThreshold = 0.05, interactionTerm = NULL, excludeSingular = FALSE,
    excludeNonConverged = FALSE, verbose = FALSE, logs = FALSE,
    log_dir = NULL, log_file = "log_methylationLME.txt"
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )
    threshold <- validateProbabilityDnaEpico(pvalThreshold, "pvalThreshold")
    retained <- list()

    for (phenotype in names(modelResults$fits)) {
        fit_list <- modelResults$fits[[phenotype]]
        phenotype_hits <- list()
        if (!is.null(modelResults$summaryCache) &&
            !is.null(modelResults$summaryCache[[phenotype]]) &&
            optionalTermMatchesMethylationModels(
                requested = interactionTerm,
                cached = modelResults$settings$interactionTerm
            )) {
            cached_summary <- modelResults$summaryCache[[phenotype]]
            if (nrow(cached_summary) > 0L && !is.na(threshold)) {
                hit_cpgs <- unique(cached_summary$CpG[cached_summary$P.value <
                    threshold])
                hit_cpgs <- hit_cpgs[!is.na(hit_cpgs) & hit_cpgs %in%
                    names(fit_list)]
                for (cpg in hit_cpgs) {
                    model_obj <- fit_list[[cpg]]
                    if (is.null(model_obj) || inherits(
                        model_obj,
                        "dnaEPICO_methylationLME_fit_error"
                    )) {
                        next
                    }
                    fit_quality <- fitStatusValuesMethylationLME(model_obj)
                    if ((isTRUE(excludeSingular) &&
                        isTRUE(fit_quality$singular)) ||
                        (isTRUE(excludeNonConverged) && identical(
                            fit_quality$converged,
                            FALSE
                        ))) {
                        next
                    }
                    if (!is.null(model_obj$coef)) {
                        phenotype_hits[[cpg]] <- as.data.frame(model_obj$coef)
                    }
                }
            }

            retained[[phenotype]] <- phenotype_hits
            next
        }

        for (cpg in names(fit_list)) {
            model_obj <- fit_list[[cpg]]
            if (is.null(model_obj) || inherits(model_obj,
                "dnaEPICO_methylationLME_fit_error")) {
                next
            }
            fit_quality <- fitStatusValuesMethylationLME(model_obj)
            if ((isTRUE(excludeSingular) && isTRUE(fit_quality$singular)) ||
                (isTRUE(excludeNonConverged) && identical(
                    fit_quality$converged,
                    FALSE
                ))) {
                next
            }

            coef_table <- model_obj$coef
            if (is.null(coef_table)) {
                next
            }

            matched_rows <- findCoefficientRowsMethylationLME(
                coefNames = rownames(coef_table),
                phenotype = phenotype, interactionTerm = interactionTerm,
                coefficientTerms = model_obj$coefficientTerms
            )
            if (length(matched_rows) == 0L) {
                next
            }

            matched_pvals <- coef_table[matched_rows, "Pr(>|t|)",
                drop = TRUE
            ]
            if (any(matched_pvals < threshold, na.rm = TRUE)) {
                phenotype_hits[[cpg]] <- as.data.frame(coef_table)
            }
        }

        retained[[phenotype]] <- phenotype_hits
    }

    hit_counts <- vapply(retained, length, integer(1))
    emitLogMinfiEwasWater(
        c(
            "============================================================",
            paste(
                "Significant longitudinal terms retained at p <",
                threshold, ":"
            ), paste(names(hit_counts), hit_counts,
                sep = ": ", collapse = "; "
            ), "============================================================"
        ),
        verbose = verbose, log_path = log_path
    )

    structure(retained, class = "dnaEPICO_methylationLME_significant")
}
#' Plot longitudinal mixed-effects model diagnostics
#'
#' @param modelSummaries Object returned by `summarizeMethylationLMEModels()`.
#' @param preparedData Object returned by `prepareMethylationLMEData()`.
#' @param fdrThreshold Numeric. False-discovery-rate threshold used to highlight
#'   CpGs in the diagnostic plots.
#' @param padjmethod Character. P-value adjustment method passed to
#'   `stats::p.adjust()`.
#' @param outputDir Character or `NULL`. Directory used for TIFF files. When
#'   `NULL`, plots are returned in memory only.
#' @param plotWidth Integer. TIFF width in pixels when plots are written to
#' disk.
#' @param plotHeight Integer. TIFF height in pixels when plots are written to
#'   disk.
#' @param plotDPI Integer. TIFF resolution in DPI when plots are written to
#'   disk.
#' @param display Logical. If `TRUE`, draw plots on the active graphics device.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return A list with class `'dnaEPICO_methylationLME_diagnostic_plots'`
#'   containing the generated `ggplot2` objects, genomic inflation factors, and
#'   any saved TIFF file paths.
#'
#' @description
#' Create Q-Q and standard-error diagnostic plots from the mixed-effects summary
#' tables returned by `summarizeMethylationLMEModels()`.
#'
#' @examples
#' ex <- dnaEPICO:::exampleMethylationLMEStateDnaEpico()
#' diagnostic_plots <- plotMethylationLMEDiagnostics(
#'     modelSummaries = ex$modelSummaries,
#'     preparedData = ex$preparedData,
#'     display = FALSE,
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#' names(diagnostic_plots$plots)
#'
#' @export
plotMethylationLMEDiagnostics <- function(
    modelSummaries, preparedData,
    fdrThreshold = 0.05, padjmethod = "fdr", outputDir = NULL,
    plotWidth = 2000L, plotHeight = 1000L, plotDPI = 150L, display = FALSE,
    verbose = FALSE, logs = FALSE, log_dir = NULL,
        log_file = "log_methylationLME.txt"
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )
    fdrThreshold <- validateProbabilityDnaEpico(
        fdrThreshold,
        "fdrThreshold"
    )
    padjmethod <- validatePAdjustmentMethodMethylationModels(padjmethod)
    summary_list <- if (!is.null(modelSummaries$diagnosticSummaries)) {
        modelSummaries$diagnosticSummaries
    } else {
        modelSummaries$summaries
    }
    diagnostic_mean <- diagnosticMeanMethylationModels(preparedData)
    plot_list <- list()
    inflation_factors <- list()
    saved_files <- list()

    for (phenotype in names(summary_list)) {
        summary_df <- summary_list[[phenotype]]
        if (is.null(summary_df) || nrow(summary_df) == 0L) {
            next
        }

        summary_df$FDR <- adjustPvaluesByTermMethylationModels(
            pValues = summary_df$P.value,
            terms = summary_df$Interaction.Term, method = padjmethod
        )
        interaction_terms <- unique(as.character(summary_df$Interaction.Term))
        interaction_terms <- interaction_terms[!is.na(interaction_terms)]
        multiple_terms <- length(interaction_terms) > 1L
        phenotype_plots <- phenotype_files <- list()
        phenotype_inflation <- numeric(0)

        for (term_index in seq_along(interaction_terms)) {
            term <- interaction_terms[[term_index]]
            term_diagnostics <- buildMethylationTermDiagnosticsDnaEpico(
                summaryData = summary_df,
                phenotype = phenotype, term = term,
                    termColumn = "Interaction.Term",
                pValueColumn = "P.value", yColumn = "Std.Error",
                yLabel = "Standard Error", diagnosticMean = diagnostic_mean,
                fdrThreshold = fdrThreshold
            )
            if (is.null(term_diagnostics)) {
                next
            }

            file_key <- phenotype
            if (multiple_terms) {
                file_key <- paste0(phenotype, "_", sprintf(
                    "%02d",
                    term_index
                ), "_", sanitizeDiagnosticTermDnaEpico(term))
            }
            term_files <- list(
                qqplot = NULL, residualSD = NULL,
                residualSignificance = NULL
            )
            if (!is.null(outputDir)) {
                term_files <- list(
                    qqplot = file.path(
                        outputDir,
                        paste0("qqplot_", file_key, ".tiff")
                    ), residualSD = file.path(
                        outputDir,
                        paste0("residualSD_", file_key, ".tiff")
                    ),
                    residualSignificance = file.path(
                        outputDir,
                        paste0(
                            "residualSignificance_", file_key,
                            ".tiff"
                        )
                    )
                )
            }

            for (plot_name in names(term_diagnostics$plots)) {
                plot_object <- term_diagnostics$plots[[plot_name]]
                if (is.null(plot_object)) {
                    next
                }
                runPlotMinfiEwasWater(
                    draw_fun = function() drawPlotObjectMinfiEwasWater(plot_object),
                    display = display, file = term_files[[plot_name]],
                    width = plotWidth, height = plotHeight, res = plotDPI
                )
            }

            phenotype_inflation[[term]] <- term_diagnostics$lambda
            if (multiple_terms) {
                phenotype_plots[[term]] <- term_diagnostics$plots
                phenotype_files[[term]] <- term_files
            } else {
                phenotype_plots <- term_diagnostics$plots
                phenotype_files <- term_files
            }
        }

        if (length(phenotype_inflation) > 0L) {
            inflation_factors[[phenotype]] <- if (multiple_terms) {
                phenotype_inflation
            } else {
                unname(phenotype_inflation[[1L]])
            }
            plot_list[[phenotype]] <- phenotype_plots
            saved_files[[phenotype]] <- phenotype_files
        }
    }

    emitLogMinfiEwasWater(
        c(
            "============================================================",
            paste("Diagnostic plots generated for phenotypes:",
                length(plot_list)),
            if (is.null(outputDir)) {
                "Diagnostic plots were returned in memory only."
            } else {
                paste("Diagnostic plots saved to:    ", outputDir)
            }, "============================================================"
        ),
        verbose = verbose, log_path = log_path
    )

    structure(list(
        plots = plot_list, inflationFactors = inflation_factors,
        files = saved_files
    ), class = "dnaEPICO_methylationLME_diagnostic_plots")
}

resolveFitDiagnosticsMethylationLME <- function(
    modelSummaries,
    summaryList
) {
    required_columns <- c(
        "Phenotype", "CpG", "Fit.Status", "Singular.Fit",
        "Converged", "Convergence.Message", "Fit.Warning", "Inference.Included",
        "Exclusion.Reason"
    )
    rows <- list()
    row_index <- 1L
    for (phenotype in names(summaryList)) {
        summary_df <- summaryList[[phenotype]]
        if (is.null(summary_df) || nrow(summary_df) == 0L ||
            !("CpG" %in% names(summary_df))) {
            next
        }
        value_or_default <- function(column, default) {
            if (column %in% names(summary_df)) {
                summary_df[[column]]
            } else {
                rep(default, nrow(summary_df))
            }
        }
        phenotype_rows <- unique(data.frame(
            Phenotype = phenotype,
            CpG = as.character(summary_df$CpG),
                Fit.Status = as.character(value_or_default(
                "Fit.Status",
                "fitted"
            )), Singular.Fit = as.logical(value_or_default(
                "Singular.Fit",
                NA
            )), Converged = as.logical(value_or_default(
                "Converged",
                TRUE
            )), Convergence.Message = as.character(value_or_default(
                "Convergence.Message",
                NA_character_
            )), Fit.Warning = as.character(value_or_default(
                "Fit.Warning",
                NA_character_
            )), Inference.Included = as.logical(value_or_default(
                "Inference.Included",
                TRUE
            )), Exclusion.Reason = as.character(value_or_default(
                "Exclusion.Reason",
                NA_character_
            )), stringsAsFactors = FALSE, check.names = FALSE
        ))
        rows[[row_index]] <- phenotype_rows
        row_index <- row_index + 1L
    }

    diagnostics <- modelSummaries$fitDiagnostics
    if (is.data.frame(diagnostics) && all(c(
        "Phenotype", "CpG",
        "Fit.Status"
    ) %in% names(diagnostics))) {
        defaults <- list(
            Singular.Fit = NA, Converged = NA,
                Convergence.Message = NA_character_,
            Fit.Warning = NA_character_, Inference.Included = startsWith(
                as.character(diagnostics$Fit.Status),
                "fitted"
            ), Exclusion.Reason = NA_character_
        )
        for (column in names(defaults)) {
            if (!(column %in% names(diagnostics))) {
                value <- defaults[[column]]
                diagnostics[[column]] <- if (length(value) ==
                    1L) {
                    rep(value, nrow(diagnostics))
                } else {
                    value
                }
            }
        }
        rows[[row_index]] <- diagnostics[, required_columns,
            drop = FALSE
        ]
        row_index <- row_index + 1L
    }

    fit_failures <- modelSummaries$fitFailures
    if (is.data.frame(fit_failures) && nrow(fit_failures) > 0L) {
        rows[[row_index]] <- data.frame(
            Phenotype = as.character(fit_failures$Phenotype),
            CpG = as.character(fit_failures$CpG),
                Fit.Status = as.character(fit_failures$Status),
            Singular.Fit = NA, Converged = ifelse(fit_failures$Status ==
                "not_converged", FALSE, NA),
                    Convergence.Message = ifelse(fit_failures$Status ==
                "not_converged", as.character(fit_failures$Error),
            NA_character_
            ), Fit.Warning = NA_character_,
            Inference.Included = FALSE,
                Exclusion.Reason = as.character(fit_failures$Error),
            stringsAsFactors = FALSE, check.names = FALSE
        )
    }

    if (length(rows) == 0L) {
        return(data.frame(
            Phenotype = character(0), CpG = character(0),
            Fit.Status = character(0), Singular.Fit = logical(0),
            Converged = logical(0), Convergence.Message = character(0),
            Fit.Warning = character(0), Inference.Included = logical(0),
            Exclusion.Reason = character(0), stringsAsFactors = FALSE,
            check.names = FALSE
        ))
    }

    diagnostics <- do.call(rbind, rows)
    key <- paste(diagnostics$Phenotype, diagnostics$CpG, sep = "\r")
    diagnostics <- diagnostics[!duplicated(key), , drop = FALSE]
    rownames(diagnostics) <- NULL
    diagnostics
}

buildAnnotationFitDiagnosticsMethylationLME <- function(fitDiagnostics) {
    if (!is.data.frame(fitDiagnostics) || nrow(fitDiagnostics) ==
        0L) {
        return(list(overall = data.frame(), byPhenotype = data.frame()))
    }

    aggregate_logical <- function(values, falseDominates = TRUE) {
        values <- as.logical(values)
        values <- values[!is.na(values)]
        if (length(values) == 0L) {
            return(NA)
        }
        if (isTRUE(falseDominates) && any(!values)) {
            return(FALSE)
        }
        if (!isTRUE(falseDominates) && any(values)) {
            return(TRUE)
        }
        if (isTRUE(falseDominates)) {
            TRUE
        } else {
            FALSE
        }
    }
    diagnostics_by_cpg <- split(fitDiagnostics, fitDiagnostics$CpG,
        drop = TRUE
    )
    overall_rows <- lapply(names(diagnostics_by_cpg), function(cpg) {
        diagnostic <- diagnostics_by_cpg[[cpg]]
        statuses <- as.character(diagnostic$Fit.Status)
        failed <- statuses %in% c("failed", "invalid", "not_converged")
        status <- if (any(failed)) {
            if (!all(failed)) {
                "partial"
            } else if (all(statuses == "invalid")) {
                "invalid"
            } else if (all(statuses == "not_converged")) {
                "not_converged"
            } else {
                "failed"
            }
        } else if (any(statuses == "fitted_not_converged")) {
            "fitted_not_converged"
        } else if (any(statuses == "fitted_singular")) {
            "fitted_singular"
        } else if (any(statuses == "fitted_with_warning")) {
            "fitted_with_warning"
        } else {
            "fitted"
        }
        reason_index <- !is.na(diagnostic$Exclusion.Reason) &
            nzchar(diagnostic$Exclusion.Reason)
        reasons <- unique(paste0(
            diagnostic$Phenotype[reason_index],
            ": ", diagnostic$Exclusion.Reason[reason_index]
        ))
        data.frame(
            CpG = cpg, Diagnostic.Fit.Status = status,
            Singular.Fit = aggregate_logical(diagnostic$Singular.Fit,
                falseDominates = FALSE
            ), Converged = aggregate_logical(diagnostic$Converged),
            Inference.Included = aggregate_logical(diagnostic$Inference.Included),
            Fit.Quality.Reason = if (length(reasons)) {
                paste(reasons, collapse = " | ")
            } else {
                NA_character_
            }, stringsAsFactors = FALSE, check.names = FALSE
        )
    })
    overall <- do.call(rbind, overall_rows)
    rownames(overall) <- NULL

    diagnostic_columns <- setdiff(names(fitDiagnostics), c(
        "Phenotype",
        "CpG"
    ))
    phenotype_tables <- lapply(
        unique(as.character(fitDiagnostics$Phenotype)),
        function(phenotype) {
            diagnostic <- fitDiagnostics[fitDiagnostics$Phenotype ==
                phenotype, c("CpG", diagnostic_columns), drop = FALSE]
            diagnostic <- unique(diagnostic)
            colnames(diagnostic)[-1L] <- paste0(
                phenotype, "_",
                diagnostic_columns
            )
            diagnostic
        }
    )
    phenotype_tables <- Filter(Negate(is.null), phenotype_tables)
    by_phenotype <- if (length(phenotype_tables) == 0L) {
        data.frame()
    } else if (length(phenotype_tables) == 1L) {
        phenotype_tables[[1L]]
    } else {
        Reduce(
            function(x, y) merge(x, y, by = "CpG", all = TRUE),
            phenotype_tables
        )
    }

    list(overall = overall, byPhenotype = by_phenotype)
}

#' Annotate longitudinal mixed-effects summary tables with array annotation
#' metadata
#'
#' @param modelSummaries Object returned by
#'   `summarizeMethylationLMEModels()` or a named list of summary data
#'   frames.
#' @param annotationObject Character package/object name, annotation data frame,
#'   or annotation object understood by `minfi::getAnnotation()`.
#' @param annotationCols Character vector or comma-separated string of
#' annotation
#'   columns to append.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return A list with class `'dnaEPICO_methylationLME_annotation'`
#'   containing the annotated summary table and any requested annotation columns
#'   that were unavailable in the chosen annotation object.
#'
#' @description
#' Merge phenotype-specific longitudinal summary tables with probe annotation
#' metadata and return a single annotated result table.
#'
#' @examples
#' ex <- dnaEPICO:::exampleMethylationLMEStateDnaEpico()
#' annotation_data <- annotateMethylationLMESummaries(
#'     modelSummaries = ex$modelSummaries,
#'     annotationObject = ex$annotationData,
#'     annotationCols = "Name,chr,pos",
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#' names(annotation_data)
#'
#' @export
annotateMethylationLMESummaries <- function(
    modelSummaries, annotationObject,
    annotationCols = c(
        "Name", "chr", "pos", "UCSC_RefGene_Group",
        "UCSC_RefGene_Name", "Relation_to_Island", "GencodeV41_Group"
    ),
    verbose = FALSE, logs = FALSE, log_dir = NULL,
        log_file = "log_methylationLME.txt"
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )
    fit_failures <- modelSummaries$fitFailures
    summary_phenotypes <- modelSummaries$phenotypes
    summary_list <- modelSummaries
    if (!is.null(modelSummaries$diagnosticSummaries)) {
        summary_list <- modelSummaries$diagnosticSummaries
    } else if (!is.null(modelSummaries$summaries)) {
        summary_list <- modelSummaries$summaries
    }
    if (is.null(summary_phenotypes)) {
        summary_phenotypes <- names(summary_list)
    }
    fit_diagnostics <- resolveFitDiagnosticsMethylationLME(
        modelSummaries = modelSummaries,
        summaryList = summary_list
    )
    annotation_diagnostics <-
        buildAnnotationFitDiagnosticsMethylationLME(fitDiagnostics = fit_diagnostics)

    annotation_cols <- splitOptionMinfiEwasWater(annotationCols,
        sep = ","
    )
    annotation_df <- coerceAnnotationDataMethylationGLM(annotationObject)

    cleaned_summaries <- lapply(names(summary_list), function(model_name) {
        df <- summary_list[[model_name]]
        if (is.null(df) || nrow(df) == 0L) {
            return(NULL)
        }
        if (!all(c("CpG", "Interaction.Term", "P.value") %in%
            colnames(df))) {
            return(NULL)
        }

        df_split <- split(df, df$Interaction.Term)
        model_tables <- lapply(names(df_split), function(term) {
            sub_df <- df_split[[term]][, c("CpG", "P.value"),
                drop = FALSE
            ]
            clean_term <- gsub("`", "", term, fixed = TRUE)
            interaction_suffix <- clean_term
            if (startsWith(clean_term, paste0(model_name, "."))) {
                interaction_suffix <- sub(paste0(
                    "^", escapeRegexMethylationGLM(model_name),
                    "\\."
                ), "", clean_term)
            }
            p_col <- paste0(
                model_name, "_", interaction_suffix,
                "_P.Value"
            )
            colnames(sub_df)[2] <- p_col
            sub_df
        })

        Reduce(
            function(x, y) merge(x, y, by = "CpG", all = TRUE),
            model_tables
        )
    })
    cleaned_summaries <- Filter(Negate(is.null), cleaned_summaries)

    if (length(cleaned_summaries) == 0L) {
        merged_summary <- data.frame(CpG = character(0))
    } else if (length(cleaned_summaries) == 1L) {
        merged_summary <- cleaned_summaries[[1L]]
    } else {
        merged_summary <- Reduce(function(x, y) {
            merge(x, y,
                by = "CpG",
                all = TRUE
            )
        }, cleaned_summaries)
    }
    lme_engine <- modelSummaries$settings$lmeEngine
    if (is.null(lme_engine) || !nzchar(lme_engine)) {
        lme_engine <- "lme4"
    }
    if (nrow(annotation_diagnostics$byPhenotype) > 0L) {
        lme_diagnostics <- annotation_diagnostics$byPhenotype
        if (!identical(lme_engine, "lme4")) {
            lme_diagnostics <- lme_diagnostics[, !grepl(
                "_Singular\\.Fit$",
                colnames(lme_diagnostics)
            ), drop = FALSE]
        }
        merged_summary <- merge(merged_summary, lme_diagnostics,
            by = "CpG", all = TRUE
        )
    }

    available_annotation_cols <- intersect(annotation_cols,
        colnames(annotation_df))
    missing_annotation_cols <- setdiff(annotation_cols, colnames(annotation_df))
    annotated_results <- merge(merged_summary, annotation_df[,
        c("CpG", available_annotation_cols),
        drop = FALSE
    ],
    by = "CpG",
    all.x = TRUE
    )
    if ("CpG" %in% colnames(annotated_results)) {
        colnames(annotated_results)[colnames(annotated_results) ==
            "CpG"] <- "IlmnID"
    }
    annotated_results <- orderAnnotatedModelColumnsDnaEpico(
        data = annotated_results,
        annotationCols = available_annotation_cols, includeSingular = identical(
            lme_engine,
            "lme4"
        )
    )

    emitLogMinfiEwasWater(
        c(
            "============================================================",
            paste("Annotated CpG rows:          ", nrow(annotated_results)),
            paste("Annotation columns used:      ",
                paste(available_annotation_cols,
                collapse = ", "
            )), if (length(missing_annotation_cols) ==
                0L) {
                "Missing annotation columns:   none"
            } else {
                paste("Missing annotation columns:   ",
                    paste(missing_annotation_cols,
                    collapse = ", "
                ))
            }, "============================================================"
        ),
        verbose = verbose, log_path = log_path
    )

    structure(list(
        data = annotated_results, fitFailures = fit_failures,
        fitDiagnostics = fit_diagnostics,
            annotationColumnsUsed = available_annotation_cols,
        missingAnnotationCols = missing_annotation_cols
    ), class = "dnaEPICO_methylationLME_annotation")
}

#' Write optional disk outputs for longitudinal mixed-effects analyses
#'
#' @param modelResults Object returned by `fitMethylationLMEModels()`.
#' @param modelSummaries Object returned by `summarizeMethylationLMEModels()`.
#' @param annotatedResults Object returned by
#'   `annotateMethylationLMESummaries()` or a compatible data frame.
#' @param significantInteractions Object returned by
#'   `collectSignificantInteractionsMethylationLME()` or `NULL`.
#' @param outputRData Character. Directory used for serialized model and summary
#'   outputs.
#' @param summaryTxtDir Character. Directory used for tab-delimited summary
#'   tables.
#' @param significantInteractionDir Character. Directory used for significant
#'   interaction coefficient tables.
#' @param annotatedLMEOut Character. Directory used for the annotated summary
#'   XLSX workbook.
#' @param reportAssetsDir Character or `NULL`. Report results directory used for
#'   the compressed TSV table and compact metadata sidecars.
#' @param saveTxtSummaries Logical. If `TRUE`, write tab-delimited summary
#' tables.
#' @param saveSignificantInteractions Logical. If `TRUE`, write significant
#'   interaction coefficient tables.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return A list with class `'dnaEPICO_methylationLME_paths'`
#'   containing the paths of the files written to disk, including the annotated
#'   workbook and any requested report-table sidecars.
#'
#' @description
#' Write optional serialized outputs, summary tables, significant interaction
#' tables, and annotated results from the longitudinal mixed-effects workflow.
#'
#' @examples
#' ex <- dnaEPICO:::exampleMethylationLMEStateDnaEpico()
#' annotation_data <- annotateMethylationLMESummaries(
#'     modelSummaries = ex$modelSummaries,
#'     annotationObject = ex$annotationData,
#'     annotationCols = "Name,chr,pos",
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#' significant_hits <- collectSignificantInteractionsMethylationLME(
#'     modelResults = ex$modelResults,
#'     pvalThreshold = 1,
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#' output_paths <- writeMethylationLMEOutputs(
#'     modelResults = ex$modelResults,
#'     modelSummaries = ex$modelSummaries,
#'     annotatedResults = annotation_data,
#'     significantInteractions = significant_hits,
#'     outputRData = file.path(ex$tempDir, "models"),
#'     summaryTxtDir = file.path(ex$tempDir, "summary"),
#'     significantInteractionDir = file.path(ex$tempDir, "significant"),
#'     annotatedLMEOut = file.path(ex$tempDir, "annotated"),
#'     saveTxtSummaries = TRUE,
#'     saveSignificantInteractions = TRUE,
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#' names(output_paths)
#'
#' @export
writeMethylationLMEOutputs <- function(
    modelResults, modelSummaries,
    annotatedResults, significantInteractions = NULL, outputRData,
    summaryTxtDir, significantInteractionDir, annotatedLMEOut,
    reportAssetsDir = NULL, saveTxtSummaries = TRUE,
    saveSignificantInteractions = FALSE,
    verbose = FALSE, logs = FALSE, log_dir = NULL,
        log_file = "log_methylationLME.txt"
) {
    if (!is.null(reportAssetsDir) &&
        (!is.character(reportAssetsDir) || length(reportAssetsDir) != 1L ||
            is.na(reportAssetsDir) || !nzchar(reportAssetsDir))) {
        stop("reportAssetsDir must be NULL or one non-empty directory path.",
            call. = FALSE
        )
    }
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )
    dir.create(outputRData, recursive = TRUE, showWarnings = FALSE)
    dir.create(annotatedLMEOut, recursive = TRUE, showWarnings = FALSE)

    model_files <- stats::setNames(
        character(length(modelResults$fits)),
        names(modelResults$fits)
    )
    summary_files <- stats::setNames(
        character(length(modelSummaries$summaries)),
        names(modelSummaries$summaries)
    )
    summary_txt_files <- list()
    significant_files <- list()
    fit_failure_file <- character(0)

    for (phenotype in names(modelResults$fits)) {
        model_file <- file.path(outputRData, paste0(
            phenotype,
            "LME.rds"
        ))
        summary_file <- file.path(outputRData, paste0(
            phenotype,
            "SummaryLME.rds"
        ))
        saveRDS(modelResults$fits[[phenotype]], file = model_file)
        saveRDS(modelSummaries$summaries[[phenotype]], file = summary_file)
        model_files[[phenotype]] <- model_file
        summary_files[[phenotype]] <- summary_file
    }

    if (isTRUE(saveTxtSummaries)) {
        dir.create(summaryTxtDir, recursive = TRUE, showWarnings = FALSE)
        for (phenotype in names(modelSummaries$summaries)) {
            summary_df <- modelSummaries$summaries[[phenotype]]
            if (is.null(summary_df) || nrow(summary_df) == 0L) {
                next
            }

            if ("P.value" %in% colnames(summary_df)) {
                summary_df <- summary_df[order(summary_df$P.value), ,
                    drop = FALSE
                ]
            }

            output_file <- file.path(summaryTxtDir, paste0(
                phenotype,
                "SummaryLME.txt"
            ))
            utils::write.table(summary_df,
                file = output_file,
                sep = "\t", row.names = FALSE, quote = FALSE
            )
            summary_txt_files[[phenotype]] <- output_file
        }
    }

    if (isTRUE(saveSignificantInteractions) &&
        !is.null(significantInteractions)) {
        dir.create(significantInteractionDir,
            recursive = TRUE,
            showWarnings = FALSE
        )
        for (phenotype in names(significantInteractions)) {
            phenotype_hits <- significantInteractions[[phenotype]]
            if (length(phenotype_hits) == 0L) {
                next
            }

            significant_files[[phenotype]] <- character(0)
            phenotype_dir <- file.path(
                significantInteractionDir,
                phenotype
            )
            dir.create(phenotype_dir, recursive = TRUE, showWarnings = FALSE)

            for (cpg in names(phenotype_hits)) {
                cpg_dir <- file.path(phenotype_dir, cpg)
                dir.create(cpg_dir, recursive = TRUE, showWarnings = FALSE)
                output_file <- file.path(cpg_dir, paste0(
                    cpg,
                    ".txt"
                ))
                utils::write.table(phenotype_hits[[cpg]],
                    file = output_file,
                    sep = "\t", quote = FALSE
                )
                significant_files[[phenotype]] <- c(
                    significant_files[[phenotype]],
                    output_file
                )
            }
        }
    }

    annotated_df <- annotatedResults
    if (!is.null(annotatedResults$data)) {
        annotated_df <- annotatedResults$data
    }
    report_table <- sortReportTableDnaEpico(annotated_df)
    annotated_df <- report_table$data
    annotated_file <- file.path(annotatedLMEOut, "annotatedLME.xlsx")
    fit_failures <- modelSummaries$fitFailures
    if (is.null(fit_failures)) {
        fit_failures <- modelResults$fitFailures
    }
    if (is.data.frame(fit_failures) && nrow(fit_failures) > 0L) {
        fit_failure_file <- file.path(annotatedLMEOut, "CpGFitFailuresLME.txt")
        utils::write.table(fit_failures,
            file = fit_failure_file,
            sep = "\t", row.names = FALSE, quote = FALSE
        )
    }
    dictionary <- buildAnnotatedWorkbookDictionaryMethylationGLM(
        columns = colnames(annotated_df),
        modelDescription = "Pvalue from LME model",
            formulaText = modelResults$formulas,
        modelLabel = "LME",
            responseLabel = inferMethylationValueLabelMethylationGLM(modelResults)
    )
    metadata <- buildModelWorkbookMetadataDnaEpico(
        modelResults = modelResults,
        modelSummaries = modelSummaries, annotatedResults = annotatedResults,
        analysis = "lme"
    )
    writeAnnotatedWorkbookMethylationGLM(
        annotated_df = annotated_df,
        file = annotated_file, resultSheet = "annotatedLME",
        dictionary = dictionary, metadata = metadata
    )
    report_sidecar <- list(
        table = NULL, metadata = NULL, dictionary = NULL,
        workbookMetadata = NULL
    )
    if (!is.null(reportAssetsDir)) {
        report_sidecar <- writeReportTableSidecarDnaEpico(
            tableData = annotated_df,
            workbookFile = annotated_file, sidecarDir = reportAssetsDir,
            sheet = "annotatedLME", idColumn = report_table$idColumn,
            dictionary = dictionary, workbookMetadata = metadata
        )
    }

    emitLogMinfiEwasWater(
        c(
            "============================================================",
            paste("Serialized model outputs:    ", length(model_files)),
            paste("Serialized summary outputs:  ", length(summary_files)),
            paste("Annotated results file:      ", annotated_file),
            if (is.null(report_sidecar$table)) {
                "Report table sidecar:          not requested"
            } else {
                paste("Report table sidecar:         ", report_sidecar$table)
            },
            paste("CpG fit failures reported:   ", nrow(fit_failures)),
            if (isTRUE(saveTxtSummaries)) {
                paste("Summary text files written:   ",
                    length(summary_txt_files))
            } else {
                "Summary text files written:   0"
            }, if (isTRUE(saveSignificantInteractions)) {
                paste("Significant interaction files:", sum(vapply(
                    significant_files,
                    length, integer(1)
                )))
            } else {
                "Significant interaction files: 0"
            }, "============================================================"
        ),
        verbose = verbose, log_path = log_path
    )

    structure(
        list(
            modelFiles = model_files, summaryFiles = summary_files,
            summaryTxtFiles = summary_txt_files,
                significantInteractionFiles = significant_files,
            fitFailureFile = fit_failure_file, annotatedLME = annotated_file,
            annotatedLMEText = report_sidecar$table,
                annotatedLMEReportMetadata = report_sidecar$metadata,
            annotatedLMEDictionary = report_sidecar$dictionary,
                annotatedLMEMetadata = report_sidecar$workbookMetadata
        ),
        class = "dnaEPICO_methylationLME_paths"
    )
}

create_methylation_lme_example <- function(path, include_person = TRUE) {
    phenoBT1T2 <- data.frame(
        SID = c("P1A", "P1B", "P2A", "P2B", "P3A", "P3B", "P4A", "P4B"),
        Timepoint = factor(c("1", "2", "1", "2", "1", "2", "1", "2")),
        score = c(10, 12, 9, 11, 13, 14, 8, 9),
        sex = factor(c("F", "F", "M", "M", "F", "F", "M", "M")),
        cg00000029 = c(0.25, 0.27, 0.20, 0.22, 0.30, 0.31, 0.18, 0.20),
        cg00000108 = c(0.50, 0.53, 0.55, 0.57, 0.48, 0.49, 0.60, 0.61),
        check.names = FALSE
    )

    if (isTRUE(include_person)) {
        phenoBT1T2$person <- c(1, 1, 2, 2, 3, 3, 4, 4)
    }

    input_path <- file.path(path, "phenoBT1T2.RData")
    save(phenoBT1T2, file = input_path)

    list(
        inputPheno = input_path,
        phenoBT1T2 = phenoBT1T2
    )
}

create_methylation_lme_ar1_example <- function(path) {
    set.seed(1)
    pheno <- data.frame(
        SID = paste0("S", rep(seq_len(8), each = 3), "_", rep(seq_len(3), times = 8)),
        person = rep(seq_len(8), each = 3),
        Timepoint = factor(rep(c("T1", "T2", "T3"), times = 8), levels = c("T1", "T2", "T3")),
        VisitOrder = rep(seq_len(3), times = 8),
        status = factor(rep(rep(c("Control", "Case"), each = 4), each = 3)),
        sex = factor(rep(c("F", "M"), length.out = 24)),
        cg00000029 =
            0.2 +
                rep(rep(c(0, 0.03), each = 4), each = 3) +
                rep(seq(0, 0.02, length.out = 3), times = 8) +
                stats::rnorm(24, 0, 0.005),
        cg00000108 =
            0.5 +
                rep(rep(c(0, -0.02), each = 4), each = 3) +
                rep(seq(0, 0.01, length.out = 3), times = 8) +
                stats::rnorm(24, 0, 0.005),
        check.names = FALSE
    )

    input_path <- file.path(path, "phenoMT1T2T3.RData")
    save(pheno, file = input_path)

    list(
        inputPheno = input_path,
        annotation = data.frame(
            CpG = c("cg00000029", "cg00000108"),
            chr = c("chr1", "chr1"),
            pos = c(1, 2),
            stringsAsFactors = FALSE
        )
    )
}

test_that("lme4 singular fits remain available and are reported per CpG", {
    model_data <- data.frame(
        person = factor(rep(seq_len(8), each = 3)),
        visit = rep(c(-1, 0, 1), times = 8)
    )
    cpg_values <-
        0.4 +
        (0.03 * model_data$visit) +
        rep(c(0.001, -0.002, 0.001), times = 8)

    fit <- suppressMessages(dnaEPICO:::fitCpGModelMethylationLME(
        cpg = "cg_singular",
        cpgValues = cpg_values,
        modelData = model_data,
        formulaText = "beta ~ visit + (1 | person)",
        personVar = "person",
        lmeEngine = "lme4",
        responseVar = "beta"
    ))
    expect_false(inherits(fit, "dnaEPICO_methylationLME_fit_error"))
    expect_true(isTRUE(fit$fitStatus$singular))

    summary_table <- dnaEPICO:::summarizeCpGFitMethylationLME(
        cpg = "cg_singular",
        modelObj = fit,
        phenotype = "visit"
    )
    expect_equal(summary_table$Fit.Status, "fitted_singular")
    expect_true(summary_table$Singular.Fit)
    expect_match(summary_table$Fit.Warning, "singular", ignore.case = TRUE)

    fits <- list(visit = list(cg_singular = fit))
    fit_diagnostics <- dnaEPICO:::collectFitDiagnosticsMethylationLME(fits)
    fit_failures <- dnaEPICO:::collectFitFailuresMethylationModels(
        fits,
        "dnaEPICO_methylationLME_fit_error"
    )
    annotation <- annotateMethylationLMESummaries(
        modelSummaries = list(
            diagnosticSummaries = list(visit = summary_table),
            summaries = list(visit = summary_table),
            phenotypes = "visit",
            fitFailures = fit_failures,
            fitDiagnostics = fit_diagnostics
        ),
        annotationObject = data.frame(
            CpG = "cg_singular",
            chr = "chr1",
            pos = 1,
            stringsAsFactors = FALSE
        ),
        annotationCols = c("chr", "pos"),
        verbose = FALSE,
        logs = FALSE
    )
    expect_true(annotation$data$visit_Singular.Fit)
    expect_equal(annotation$data$visit_Fit.Status, "fitted_singular")
    expect_match(annotation$data$visit_Fit.Warning, "singular", ignore.case = TRUE)
    expect_true(annotation$data$visit_Converged)
    expect_false(any(c("Fit.Status", "Singular.Fit", "Converged") %in% names(annotation$data)))
    expect_equal(names(annotation$data)[[1L]], "IlmnID")
    expect_lt(match("pos", names(annotation$data)), match("visit_Fit.Status", names(annotation$data)))

    excluded <- summarizeMethylationLMEModels(
        modelResults = list(
            fits = fits,
            summaryCache = list(visit = summary_table),
            fitFailures = fit_failures,
            fitDiagnostics = fit_diagnostics,
            settings = list(interactionTerm = NULL)
        ),
        preparedData = list(interactionTerm = NULL),
        summaryPval = NA,
        excludeSingular = TRUE,
        nCores = 1,
        logs = FALSE
    )
    excluded_row <- excluded$summaries$visit
    expect_equal(nrow(excluded_row), 1L)
    expect_true(is.finite(excluded_row$Estimate))
    expect_true(is.na(excluded_row$P.value))
    expect_false(excluded_row$Inference.Included)
    expect_match(excluded_row$Exclusion.Reason, "singular")

    excluded_annotation <- annotateMethylationLMESummaries(
        modelSummaries = excluded,
        annotationObject = data.frame(
            CpG = "cg_singular",
            chr = "chr1",
            stringsAsFactors = FALSE
        ),
        annotationCols = "chr",
        logs = FALSE
    )
    expect_equal(excluded_annotation$data$IlmnID, "cg_singular")
    expect_true(excluded_annotation$data$visit_Singular.Fit)
    expect_false(excluded_annotation$data$visit_Inference.Included)
    expect_match(excluded_annotation$data$visit_Exclusion.Reason, "singular")
})

test_that("non-converged mixed fits retain coefficients and can be excluded", {
    coefficient_table <- matrix(
        c(0.4, 0.01, 40, 1e-08, 0.03, 0.01, 3, 0.01),
        nrow = 2,
        byrow = TRUE,
        dimnames = list(
            c("(Intercept)", "visit"),
            c("Estimate", "Std. Error", "t value", "Pr(>|t|)")
        )
    )
    fit <- list(
        coef = coefficient_table,
        coefficientTerms = c("(Intercept)" = "(Intercept)", visit = "visit"),
        fitStatus = list(
            singular = FALSE,
            converged = FALSE,
            warnings = "failed to converge with max|grad| = 0.01",
            convergenceMessages = "failed to converge with max|grad| = 0.01"
        )
    )
    summary_table <- dnaEPICO:::summarizeCpGFitMethylationLME(
        cpg = "cg_nonconverged",
        modelObj = fit,
        phenotype = "visit"
    )
    expect_equal(summary_table$Fit.Status, "fitted_not_converged")
    expect_false(summary_table$Converged)
    expect_match(summary_table$Convergence.Message, "failed to converge")

    excluded <- dnaEPICO:::applyFitQualityExclusionsMethylationLME(
        summary_table,
        excludeNonConverged = TRUE
    )
    expect_equal(nrow(excluded), 1L)
    expect_true(is.na(excluded$P.value))
    expect_false(excluded$Inference.Included)
    expect_match(excluded$Exclusion.Reason, "did not converge")
})

test_that("methylationLME returns in-memory results quietly by default", {
    testthat::skip_if_not_installed("IlluminaHumanMethylation450kanno.ilmn12.hg19")
    testthat::skip_if_not_installed("lmerTest")

    tmp <- withr::local_tempdir()
    example_data <- create_methylation_lme_example(tmp, include_person = TRUE)

    expect_message(
        result <- methylationLME(
            inputPheno = example_data$inputPheno,
            outputLogs = file.path(tmp, "logs"),
            outputRData = file.path(tmp, "rData", "methylationLME", "models"),
            outputPlots = file.path(tmp, "figures", "methylationLME"),
            summaryTxtDir = file.path(tmp, "results", "summary"),
            significantInteractionDir = file.path(tmp, "results", "significant"),
            annotatedLMEOut = file.path(tmp, "data", "methylationLME"),
            personVar = "person",
            timeVar = "Timepoint",
            phenotypes = "score",
            covariates = "sex",
            factorVars = "sex",
            cpgLimit = 2,
            nCores = 1,
            summaryPval = 1,
            saveSignificantInteractions = TRUE,
            significantInteractionPval = 1,
            annotationPackage = "IlluminaHumanMethylation450kanno.ilmn12.hg19",
            annotationCols = "Name,chr,pos",
            display = FALSE,
            verbose = FALSE,
            logs = FALSE,
            saveOutputs = FALSE
        ),
        NA
    )

    expect_s3_class(result, "dnaEPICO_methylationLME")
    expect_null(result$savedFiles)
    expect_true("score" %in% names(result$modelFits$fits))
    expect_true("score" %in% names(result$modelSummaries$summaries))
    expect_equal(result$modelFits$settings$parallelBackend, "serial")
    expect_equal(nrow(result$modelFits$summaryCache$score), 2)
    expect_s3_class(result$diagnosticPlots$plots$score$qqplot, "ggplot")
    expect_true("IlmnID" %in% colnames(result$annotation$data))
    expect_false(isTRUE(result$preparedData$personCreated))
    expect_equal(result$runSettings$analysisLabel, "methylationLME")
    expect_equal(result$runSettings$internalResponseColumn, "beta")
    expect_false(dir.exists(file.path(tmp, "figures", "methylationLME")))
})

test_that("methylationLME updates the internal response column for copy-number runs", {
    testthat::skip_if_not_installed("lmerTest")

    tmp <- withr::local_tempdir()
    example_data <- create_methylation_lme_example(tmp, include_person = TRUE)

    prepared <- prepareMethylationLMEData(
        inputPheno = example_data$inputPheno,
        personVar = "person",
        timeVar = "Timepoint",
        phenotypes = "score",
        covariates = "sex",
        factorVars = "sex",
        cpgLimit = 1,
        methylationScale = "cn",
        verbose = FALSE,
        logs = FALSE
    )
    fits <- fitMethylationLMEModels(
        preparedData = prepared,
        lmeLibs = "lme4,lmerTest",
        nCores = 1,
        verbose = FALSE,
        logs = FALSE
    )

    expect_equal(prepared$internalResponseColumn, "cn")
    expect_equal(prepared$responseLabel, "Copy number values")
    expect_equal(fits$settings$internalResponseColumn, "cn")
    expect_equal(unname(fits$formulas["score"]), "cn ~ `score` + `sex` + (1 | `person` )")
})

test_that("methylationLME can write outputs and derive person IDs on request", {
    testthat::skip_if_not_installed("IlluminaHumanMethylation450kanno.ilmn12.hg19")
    testthat::skip_if_not_installed("lmerTest")

    tmp <- withr::local_tempdir()
    example_data <- create_methylation_lme_example(tmp, include_person = FALSE)
    report_assets <- file.path(
        tmp, "reports", "model1", "assets", "results", "lme_results"
    )

    expect_message(
        result <- methylationLME(
            inputPheno = example_data$inputPheno,
            outputLogs = file.path(tmp, "logs"),
            outputRData = file.path(tmp, "rData", "methylationLME", "models"),
            outputPlots = file.path(tmp, "figures", "methylationLME"),
            personVar = "person",
            timeVar = "Timepoint",
            phenotypes = "score",
            covariates = "sex",
            factorVars = "sex",
            cpgLimit = 2,
            nCores = 1,
            summaryPval = 1,
            saveSignificantInteractions = TRUE,
            significantInteractionDir = file.path(tmp, "results", "cpgs", "methylationLME"),
            significantInteractionPval = 1,
            saveTxtSummaries = TRUE,
            summaryTxtDir = file.path(tmp, "results", "summary", "methylationLME"),
            annotationPackage = "IlluminaHumanMethylation450kanno.ilmn12.hg19",
            annotationCols = "Name,chr,pos",
            annotatedLMEOut = file.path(tmp, "data", "methylationLME"),
            reportAssetsDir = report_assets,
            display = FALSE,
            verbose = TRUE,
            logs = TRUE,
            saveOutputs = TRUE
        ),
        "Starting DNAm LME Analysis"
    )

    expect_s3_class(result$savedFiles, "dnaEPICO_methylationLME_paths")
    log_file <- file.path(tmp, "logs", "log_methylationLME.txt")
    expect_true(file.exists(log_file))
    log_text <- paste(readLines(log_file, warn = FALSE), collapse = "\n")
    expect_match(log_text, "Fit-time summary rows cached")
    expect_match(log_text, "Summary source:\\s+fit-time cache")
    expect_equal(result$runSettings$methylationObjectPrefix, "phenoB")
    expect_equal(result$modelFits$settings$internalResponseColumn, "beta")
    expect_true(file.exists(result$savedFiles$modelFiles[["score"]]))
    expect_true(file.exists(result$savedFiles$summaryFiles[["score"]]))
    expect_true(file.exists(result$savedFiles$summaryTxtFiles[["score"]]))
    expect_true(file.exists(result$savedFiles$annotatedLME))
    expect_true(file.exists(result$savedFiles$annotatedLMEText))
    expect_true(file.exists(result$savedFiles$annotatedLMEReportMetadata))
    expect_true(file.exists(result$savedFiles$annotatedLMEDictionary))
    expect_true(file.exists(result$savedFiles$annotatedLMEMetadata))
    expect_match(result$savedFiles$annotatedLME, "annotatedLME\\.xlsx$")
    expect_equal(
        normalizePath(
            dirname(result$savedFiles$annotatedLMEText),
            winslash = "/", mustWork = FALSE
        ),
        normalizePath(report_assets, winslash = "/", mustWork = FALSE)
    )
    expect_false(dir.exists(file.path(
        dirname(result$savedFiles$annotatedLME), "report-assets"
    )))
    expect_equal(
        sort(basename(unlist(result$savedFiles[c(
            "annotatedLMEText",
            "annotatedLMEReportMetadata",
            "annotatedLMEDictionary",
            "annotatedLMEMetadata"
        )]))),
        sort(c(
            "annotatedLME.tsv.gz",
            "annotatedLME.report.tsv",
            "annotatedLME.dictionary.tsv",
            "annotatedLME.metadata.tsv"
        ))
    )
    expect_equal(
        openxlsx::getSheetNames(result$savedFiles$annotatedLME),
        c("annotatedLME", "metadata", "dictionary")
    )
    dictionary <- openxlsx::read.xlsx(
        result$savedFiles$annotatedLME,
        sheet = "dictionary",
        check.names = FALSE
    )
    expect_equal(colnames(dictionary), c("Column", "Description", "Formula"))
    expect_true(any(dictionary$Description == "Pvalue from LME model"))
    expect_true(any(dictionary$Formula == "LME: Beta values ~ `score` + `sex` + (1 | `person` )"))
    metadata <- openxlsx::read.xlsx(
        result$savedFiles$annotatedLME,
        sheet = "metadata",
        check.names = FALSE
    )
    expect_true(all(c("Key", "Value") %in% colnames(metadata)))
    expect_equal(metadata$Value[metadata$Key == "backend"], "lme4")
    expect_equal(metadata$Value[metadata$Key == "fitting_function"], "lmerTest::lmer")
    expect_true(file.exists(file.path(tmp, "figures", "methylationLME", "qqplot_score.tiff")))
    expect_true(isTRUE(result$preparedData$personCreated))
    expect_true(length(result$savedFiles$significantInteractionFiles$score) >= 1)
    expect_true(all(file.exists(result$savedFiles$significantInteractionFiles$score)))
})

test_that("methylationLME supports nlme AR1 models through lmeLibs", {
    testthat::skip_if_not_installed("nlme")

    tmp <- withr::local_tempdir()
    example_data <- create_methylation_lme_ar1_example(tmp)
    report_assets <- file.path(
        tmp, "reports", "nlme", "assets", "results", "lme_results"
    )

    expect_message(
        result <- methylationLME(
            inputPheno = example_data$inputPheno,
            outputLogs = file.path(tmp, "logs"),
            outputRData = file.path(tmp, "rData", "methylationLME", "models"),
            outputPlots = file.path(tmp, "figures", "methylationLME"),
            personVar = "person",
            timeVar = "Timepoint",
            phenotypes = "status",
            covariates = "sex",
            factorVars = "status,sex",
            summaryTxtDir = file.path(tmp, "results", "summary", "nlme"),
            significantInteractionDir = file.path(tmp, "results", "significant", "nlme"),
            annotatedLMEOut = file.path(tmp, "data", "nlme"),
            reportAssetsDir = report_assets,
            lmeLibs = "nlme",
            correlationStructure = "AR1",
            correlationVar = "VisitOrder",
            methylationScale = "m",
            cpgLimit = 2,
            nCores = 1,
            summaryPval = 1,
            saveSignificantInteractions = TRUE,
            significantInteractionPval = 1,
            annotationPackage = example_data$annotation,
            annotationCols = "chr,pos",
            display = FALSE,
            verbose = FALSE,
            logs = TRUE,
            saveOutputs = TRUE
        ),
        NA
    )

    expect_s3_class(result, "dnaEPICO_methylationLME")
    expect_equal(result$modelFits$settings$lmeLibs, "nlme")
    expect_equal(result$modelFits$settings$correlationStructure, "AR1")
    expect_equal(result$modelFits$settings$correlationVar, "VisitOrder")
    expect_equal(result$runSettings$correlationStructure, "AR1")
    expect_equal(result$runSettings$correlationVar, "VisitOrder")
    expect_equal(result$runSettings$methylationScale, "m")
    expect_equal(nrow(result$modelFits$summaryCache$status), 2)
    expect_true(all(c(
        "CpG", "Interaction.Term", "Estimate", "Std.Error", "t.value", "P.value",
        "Fit.Status", "Converged", "Convergence.Message",
        "Fit.Warning", "Inference.Included", "Exclusion.Reason"
    ) %in%
        colnames(result$modelSummaries$summaries$status)))
    expect_true(all(is.na(result$modelFits$fitDiagnostics$Singular.Fit)))
    expect_true(all(result$modelFits$fitDiagnostics$Converged))
    expect_equal(
        openxlsx::getSheetNames(result$savedFiles$annotatedLME),
        c("annotatedLME", "metadata", "dictionary")
    )
    metadata <- openxlsx::read.xlsx(
        result$savedFiles$annotatedLME,
        sheet = "metadata",
        check.names = FALSE
    )
    expect_equal(metadata$Value[metadata$Key == "backend"], "nlme")
    expect_equal(metadata$Value[metadata$Key == "fitting_function"], "nlme::lme")
    annotated <- openxlsx::read.xlsx(
        result$savedFiles$annotatedLME,
        sheet = "annotatedLME",
        check.names = FALSE
    )
    expect_false(any(grepl("Singular\\.Fit$", colnames(annotated))))
    expect_true(file.exists(file.path(report_assets, "annotatedLME.tsv.gz")))
    expect_false(dir.exists(file.path(
        dirname(result$savedFiles$annotatedLME), "report-assets"
    )))

    log_text <- paste(readLines(file.path(tmp, "logs", "log_methylationLME.txt"), warn = FALSE), collapse = "\n")
    expect_match(log_text, "LME libraries:\\s+nlme")
    expect_match(log_text, "Correlation structure:\\s+AR1")
    expect_match(log_text, "Correlation variable:\\s+VisitOrder")
    expect_false(grepl("Resolved LME engine", log_text, fixed = TRUE))
})

test_that("scaleVars standardizes selected longitudinal fixed effects", {
    tmp <- withr::local_tempdir()
    example_data <- create_methylation_lme_example(tmp)
    prepared <- prepareMethylationLMEData(
        inputPheno = example_data$inputPheno,
        personVar = "person",
        timeVar = "Timepoint",
        phenotypes = "score",
        covariates = "sex",
        factorVars = "sex",
        scaleVars = "score",
        cpgLimit = 1,
        logs = FALSE
    )

    expect_equal(prepared$data$score, example_data$phenoBT1T2$score)
    expect_equal(mean(prepared$modelData$score), 0, tolerance = 1e-12)
    expect_equal(stats::sd(prepared$modelData$score), 1, tolerance = 1e-12)
    expect_equal(prepared$scalingMetadata$Variable, "score")

    prepared_with_correlation_scaled <- prepared
    prepared_with_correlation_scaled$scaleVars <- c(
        prepared_with_correlation_scaled$scaleVars,
        "Timepoint"
    )
    expect_error(
        fitMethylationLMEModels(
            preparedData = prepared_with_correlation_scaled,
            nCores = 1,
            lmeLibs = "nlme",
            correlationStructure = "AR1",
            correlationVar = "Timepoint",
            logs = FALSE
        ),
        "correlationVar cannot also"
    )
})

test_that("methylationLME requires a correlation variable for AR1 models", {
    testthat::skip_if_not_installed("nlme")

    tmp <- withr::local_tempdir()
    example_data <- create_methylation_lme_ar1_example(tmp)

    expect_error(
        methylationLME(
            inputPheno = example_data$inputPheno,
            personVar = "person",
            timeVar = "Timepoint",
            phenotypes = "status",
            covariates = "sex",
            factorVars = "status,sex",
            lmeLibs = "nlme",
            correlationStructure = "AR1",
            methylationScale = "m",
            annotationPackage = example_data$annotation,
            display = FALSE,
            verbose = FALSE,
            logs = FALSE,
            saveOutputs = FALSE
        ),
        "correlationVar must be supplied"
    )

})

test_that("methylationLME formulas support direct interaction variables", {
    formula_text <- dnaEPICO:::buildFormulaMethylationLME(
        phenotype = "caseStatus",
        personVar = "person",
        covariates = "Sex",
        interactionTerm = "Age",
        includeRandomTerm = TRUE
    )

    expect_equal(
        formula_text,
        "beta ~ `caseStatus` * `Age` + `Sex` + (1 | `person` )"
    )
})

test_that("annotateMethylationLMESummaries accepts annotation package names", {
    testthat::skip_if_not_installed("IlluminaHumanMethylation450kanno.ilmn12.hg19")
    testthat::skip_if_not_installed("lmerTest")

    ex <- dnaEPICO:::exampleMethylationLMEStateDnaEpico()

    annotation_data <- annotateMethylationLMESummaries(
        modelSummaries = ex$modelSummaries,
        annotationObject = "IlluminaHumanMethylation450kanno.ilmn12.hg19",
        annotationCols = "Name,chr,pos",
        verbose = FALSE,
        logs = FALSE
    )

    expect_s3_class(annotation_data, "dnaEPICO_methylationLME_annotation")
    expect_true("IlmnID" %in% colnames(annotation_data$data))
})

## ----include = FALSE----------------------------------------------------------
startTime <- Sys.time()
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>",
    crop = NULL
)

library(dnaEPICO)

## ----overview-helpers, echo = FALSE, results = "asis"-------------------------
cat("
<style>
.overview-figure {
    margin: 1.5rem 0 2rem;
}

.overview-controls {
    display: flex;
    flex-wrap: wrap;
    gap: 8px;
    margin: 0 0 0.5rem;
}

.overview-zoom-button {
    background: #ffffff;
    border: 1px solid #2c7fb8;
    border-radius: 6px;
    color: #2c7fb8;
    cursor: pointer;
    font-size: 0.875rem;
    line-height: 1.2;
    padding: 5px 9px;
}

.overview-zoom-button:hover,
.overview-zoom-button:focus {
    background: #edf6fb;
}

.overview-zoom-value {
    align-items: center;
    color: #4b5563;
    display: inline-flex;
    font-size: 0.875rem;
    min-width: 44px;
}

.overview-svg-frame {
    background: #ffffff;
    border: 1px solid #d9e2ec;
    border-radius: 6px;
    max-height: 680px;
    overflow: auto;
    padding: 10px;
}

.overview-svg-frame img {
    display: block;
    height: auto;
    max-width: none;
    width: calc(1600px * var(--overview-scale, 1));
}

.overview-figure:fullscreen {
    background: #ffffff;
    margin: 0;
    overflow: hidden;
    padding: 16px;
}

.overview-figure:fullscreen .overview-svg-frame {
    height: calc(100vh - 88px);
    max-height: none;
}

@media (max-width: 768px) {
    .overview-svg-frame {
    max-height: 520px;
    }

    .overview-svg-frame img {
    width: calc(1200px * var(--overview-scale, 1));
    }
}
</style>
<script>
function setOverviewZoom(figureId, scale) {
    var figure = document.getElementById(figureId);
    if (!figure) {
    return;
    }

    var next = Math.max(0.5, Math.min(3, scale));
    figure.dataset.scale = next.toFixed(2);
    figure.style.setProperty('--overview-scale', next.toString());

    var label = figure.querySelector('.overview-zoom-value');
    if (label) {
    label.textContent = Math.round(next * 100) + '%';
    }
}

function changeOverviewZoom(figureId, delta) {
    var figure = document.getElementById(figureId);
    var current = figure ? parseFloat(figure.dataset.scale || '1') : 1;
    setOverviewZoom(figureId, current + delta);
}

function openOverviewFullscreen(figureId) {
    var figure = document.getElementById(figureId);
    if (!figure || !figure.requestFullscreen) {
    return;
    }

    figure.requestFullscreen();
}
</script>
")

overview_svgs <- c(
    preprocessingMinfiEwasWater =
        "preprocessingMinfiEwasWater_pipeline_overview.svg",
    svaEnmix =
        "svaEnmix_pipeline_overview.svg",
    preprocessingPheno =
        "preprocessingPheno_pipeline_overview.svg",
    methylationGLM =
        "methylationGLM_pipeline_overview.svg",
    methylationLME =
        "methylationLME_pipeline_overview.svg",
    dnamReport =
        "dnamReport_pipeline_overview.svg"
)

html_escape <- function(x) {
    x <- gsub("&", "&amp;", x, fixed = TRUE)
    x <- gsub("<", "&lt;", x, fixed = TRUE)
    x <- gsub(">", "&gt;", x, fixed = TRUE)
    x <- gsub('"', "&quot;", x, fixed = TRUE)
    x
}

svg_to_html <- function(svg_path, title = NULL) {
    if (!file.exists(svg_path)) {
        stop("Missing overview SVG: ", svg_path)
    }

    if (is.null(title)) {
        title <- sub(
            "_pipeline_overview$",
            "() overview",
            tools::file_path_sans_ext(basename(svg_path))
        )
    }

    figure_id <- gsub("[^A-Za-z0-9_-]", "-", tools::file_path_sans_ext(
        basename(svg_path)
    ))

    cat(sprintf(
        '<div class="overview-figure" id="%s" data-scale="1">\n',
        figure_id
    ))
    cat('<div class="overview-controls">\n')
    cat(sprintf(
        '<button class="overview-zoom-button" type="button" onclick="changeOverviewZoom(\'%s\', 0.25)">Zoom +</button>\n',
        figure_id
    ))
    cat(sprintf(
        '<button class="overview-zoom-button" type="button" onclick="changeOverviewZoom(\'%s\', -0.25)">Zoom -</button>\n',
        figure_id
    ))
    cat(sprintf(
        '<button class="overview-zoom-button" type="button" onclick="setOverviewZoom(\'%s\', 1)">Reset</button>\n',
        figure_id
    ))
    cat(sprintf(
        '<button class="overview-zoom-button" type="button" onclick="openOverviewFullscreen(\'%s\')">Full screen</button>\n',
        figure_id
    ))
    cat('<span class="overview-zoom-value">100%</span>\n')
    cat('</div>\n')
    cat('<div class="overview-svg-frame">\n')
    cat(sprintf(
        '<img src="%s" alt="%s">\n',
        html_escape(svg_path),
        html_escape(title)
    ))
    cat('\n</div>\n</div>\n')
}

## ----preprocessingMinfiEwasWater-overview, echo = FALSE, results = "asis"-----
svg_to_html(
    overview_svgs[["preprocessingMinfiEwasWater"]]
)

## ----svaEnmix-overview, echo = FALSE, results = "asis"------------------------
svg_to_html(
    overview_svgs[["svaEnmix"]])

## ----preprocessingPheno-overview, echo = FALSE, results = "asis"--------------
svg_to_html(
    overview_svgs[["preprocessingPheno"]])

## ----methylationGLM-overview, echo = FALSE, results = "asis"------------------
svg_to_html(
    overview_svgs[["methylationGLM"]])

## ----methylationLME-overview, echo = FALSE, results = "asis"------------------
svg_to_html(
    overview_svgs[["methylationLME"]])

## ----dnamReport-overview, echo = FALSE, results = "asis"----------------------
svg_to_html(
    overview_svgs[["dnamReport"]])

## ----echo = FALSE-------------------------------------------------------------
Sys.time()

## ----echo = FALSE-------------------------------------------------------------
totalTime <- diff(c(startTime, Sys.time()))
round(totalTime, digits = 3)

## ----echo = FALSE-------------------------------------------------------------
sessionInfo()

#' Internal example helpers for documentation
#'
#' These helpers create small, reproducible objects used by runnable examples in
#' the package documentation.
#'
#' @keywords internal
#' @noRd

requireExamplePackageDnaEpico <- function(package_name) {
    if (!requireNamespace(package_name, quietly = TRUE)) {
        stop("Package '", package_name, "' is required for this example.",
            call. = FALSE
        )
    }
}

exampleCacheDnaEpico <- new.env(parent = emptyenv())

getCachedExampleDnaEpico <- function(key, builder) {
    if (exists(key, envir = exampleCacheDnaEpico, inherits = FALSE)) {
        return(get(key, envir = exampleCacheDnaEpico, inherits = FALSE))
    }

    value <- builder()
    assign(key, value, envir = exampleCacheDnaEpico)
    value
}

exampleTempDirDnaEpico <- function(prefix) {
    temp_dir <- tempfile(pattern = prefix)
    dir.create(temp_dir, recursive = TRUE, showWarnings = FALSE)
    temp_dir
}

exampleMinfiBaseDataDnaEpico <- function() {
    getCachedExampleDnaEpico(key = "minfi_base", builder = function() {
        requireExamplePackageDnaEpico("minfiData")

        example_env <- new.env(parent = emptyenv())
        utils::data("RGsetEx", package = "minfiData", envir = example_env)

        rgset <- get("RGsetEx", envir = example_env)
        keep_samples <- seq_len(min(4L, ncol(rgset)))
        rgset <- rgset[, keep_samples]
        targets <- as.data.frame(SummarizedExperiment::colData(rgset))

        targets$Sample_Name <- colnames(rgset)

        if (!("Sex" %in% colnames(targets))) {
            if ("sex" %in% colnames(targets)) {
                sex_values <- as.character(targets$sex)
                targets$Sex <- ifelse(sex_values %in% c(
                    "0",
                    "F", "Female"
                ), "F", "M")
            } else {
                targets$Sex <- rep(c("F", "M"), length.out = nrow(targets))
            }
        }

        if (!("Sentrix_ID" %in% colnames(targets))) {
            if ("Slide" %in% colnames(targets)) {
                targets$Sentrix_ID <- as.character(targets$Slide)
            } else {
                targets$Sentrix_ID <- paste0("Slide", seq_len(nrow(targets)))
            }
        }

        if (!("Sentrix_Position" %in% colnames(targets))) {
            if ("Array" %in% colnames(targets)) {
                targets$Sentrix_Position <- as.character(targets$Array)
            } else {
                targets$Sentrix_Position <- paste0(
                    "Position",
                    seq_len(nrow(targets))
                )
            }
        }

        if (!("Timepoint" %in% colnames(targets))) {
            targets$Timepoint <- rep(c("1", "2"), length.out = nrow(targets))
        }

        rgset_col_data <- SummarizedExperiment::colData(rgset)
        rgset_col_data$Sex <- targets$Sex
        rgset_col_data$Sentrix_ID <- targets$Sentrix_ID
        rgset_col_data$Sentrix_Position <- targets$Sentrix_Position
        rgset_col_data$Timepoint <- targets$Timepoint
        SummarizedExperiment::colData(rgset) <- rgset_col_data

        list(RGSet = rgset, targets = targets,
            probeExclusionPath = system.file("extdata",
            "12864_2024_10027_MOESM8_ESM.csv",
            package = "dnaEPICO"
        ))
    })
}

exampleMinfiIdatInputsDnaEpico <- function(n = 6L) {
    getCachedExampleDnaEpico(
        key = paste0("minfi_idat_", as.integer(n)),
        builder = function() {
            requireExamplePackageDnaEpico("minfiData")
            requireExamplePackageDnaEpico("IlluminaHumanMethylation450kmanifest")
            requireExamplePackageDnaEpico("IlluminaHumanMethylation450kanno.ilmn12.hg19")

            base_dir <- system.file("extdata", package = "minfiData")
            targets <- utils::read.csv(file.path(base_dir, "SampleSheet.csv"),
                stringsAsFactors = FALSE, skip = 7
            )
            targets <- utils::head(targets, n = as.integer(n))
            targets$Basename <- paste0(
                targets$Sentrix_ID, "_",
                targets$Sentrix_Position
            )

            if (!("Sex" %in% colnames(targets)) && "sex" %in%
                colnames(targets)) {
                targets$Sex <- targets$sex
            }

            temp_dir <- exampleTempDirDnaEpico("dnaEPICO-idat-example-")
            idat_dir <- file.path(temp_dir, "idats")
            dir.create(idat_dir, recursive = TRUE, showWarnings = FALSE)

            available_idats <- list.files(base_dir,
                pattern = "[.]idat$",
                recursive = TRUE, full.names = TRUE
            )

            for (basename_value in targets$Basename) {
                matched_files <- grep(paste0(
                    basename_value,
                    ".*[.]idat$"
                ), available_idats, value = TRUE)
                file.copy(matched_files, idat_dir, overwrite = TRUE)
            }

            pheno_file <- file.path(temp_dir, "pheno.csv")
            utils::write.csv(targets, pheno_file, row.names = FALSE)

            list(
                tempDir = temp_dir, idatFolder = idat_dir,
                    phenoFile = pheno_file,
                targets = targets, arrayType = "IlluminaHumanMethylation450k",
                annotationVersion = "ilmn12.hg19",
                    probeExclusionPath = system.file("extdata",
                    "12864_2024_10027_MOESM8_ESM.csv",
                    package = "dnaEPICO"
                )
            )
        }
    )
}
exampleMinfiMetricsStateDnaEpico <- function() {
    getCachedExampleDnaEpico(key = "minfi_metrics", builder = function() {
        sample_names <- paste0("S", seq_len(4L))
        probe_names <- sprintf("cg%08d", seq_len(24L))
        beta <- matrix(
            seq(0.1, 0.9, length.out = length(probe_names) *
                length(sample_names)),
            nrow = length(probe_names),
            dimnames = list(probe_names, sample_names)
        )
        m <- log2(beta / (1 - beta))
        cn <- matrix(1,
            nrow = nrow(beta), ncol = ncol(beta),
            dimnames = dimnames(beta)
        )
        ratio_set <- minfi::RatioSet(
            Beta = beta, M = m, CN = cn,
            annotation = c(
                array = "IlluminaHumanMethylation450k",
                annotation = "ilmn12.hg19"
            )
        )
        methyl_set <- minfi::MethylSet(Meth = beta * 1000, Unmeth = (1 -
            beta) * 1000, annotation = c(
            array = "IlluminaHumanMethylation450k",
            annotation = "ilmn12.hg19"
        ))
        targets <- data.frame(Sample_Name = sample_names, Sex = rep(c(
            "F",
            "M"
        ), length.out = length(sample_names)), Timepoint = rep(c(
            "T1",
            "T2"
        ), length.out = length(sample_names)), stringsAsFactors = FALSE)
        filtered_data <- structure(list(filtered = ratio_set),
            class = "dnaEPICO_minfiEwasWater_filter"
        )
        metrics_data <- extractMetricsMinfiEwasWater(
            filteredData = filtered_data,
            verbose = FALSE, logs = FALSE
        )

        list(
            filteredData = filtered_data, metricsData = metrics_data,
            rawData = list(MSet = methyl_set),
                normData = list(primary = ratio_set),
            beta = beta, targets = targets
        )
    })
}
exampleSexPlotStateDnaEpico <- function() {
    getCachedExampleDnaEpico(key = "sex_plot", builder = function() {
        sample_names <- paste0("S", seq_len(4L))
        p_sex <- data.frame(
            predictedSex = c("F", "M", "F", "M"),
            xMed = c(12.2, 12.5, 12.1, 12.6), yMed = c(
                8.4, 10.1,
                8.5, 10.2
            ), row.names = sample_names, stringsAsFactors = FALSE
        )
        targets <- data.frame(Sample_Name = sample_names, Sex = c(
            0L,
            1L, 0L, 1L
        ), PredSex = c(0L, 1L, 0L, 1L), stringsAsFactors = FALSE)
        sex_plot_data <- p_sex
        sex_plot_data$SampleID <- sample_names
        sex_plot_data$Sex <- targets$Sex

        structure(list(
            pSex = p_sex, targets = targets, sexPlotData = sex_plot_data,
            mismatches = targets[0L, , drop = FALSE], SampleID = "Sample_Name",
            sexColumn = "Sex"
        ), class = "dnaEPICO_minfiEwasWater_sex")
    })
}

exampleMinfiWorkflowStateDnaEpico <- function() {
    getCachedExampleDnaEpico(key = "minfi_workflow", builder = function() {
        example_data <- exampleMinfiBaseDataDnaEpico()

        rgset <- example_data$RGSet
        targets <- example_data$targets

        raw_data <- buildRawMinfiEwasWater(
            RGSet = rgset, verbose = FALSE,
            logs = FALSE
        )
        assessment <- assessSamplesMinfiEwasWater(
            rawData = raw_data,
            RGSet = rgset, detPThreshold = 1, verbose = FALSE,
            logs = FALSE
        )
        sample_data <- filterSamplesMinfiEwasWater(
            RGSet = rgset,
            targets = targets, failedSamples = character(0),
            SampleID = "Sample_Name", verbose = FALSE, logs = FALSE
        )
        raw_filtered <- buildRawMinfiEwasWater(
            RGSet = sample_data$RGSet,
            verbose = FALSE, logs = FALSE
        )
        sex_data <- predictSexMinfiEwasWater(
            rawData = raw_filtered,
            targets = sample_data$targets, SampleID = "Sample_Name",
            sexColumn = "Sex", verbose = FALSE, logs = FALSE
        )

        rgset_col_data <- SummarizedExperiment::colData(sample_data$RGSet)
        rgset_col_data$Sex <- sample_data$targets$Sex
        rgset_col_data$PredSex <- sample_data$targets$PredSex
        SummarizedExperiment::colData(sample_data$RGSet) <- rgset_col_data

        norm_data <- normalizeMinfiEwasWater(
            sampleData = sample_data,
            sexColumn = "Sex", normMethods = "quantile", verbose = FALSE,
            logs = FALSE
        )
        filtered_data <- filterProbesMinfiEwasWater(
            normData = norm_data,
            RGSet = sample_data$RGSet, pvalThreshold = 1, chrToRemove = "chrY",
            snpsToRemove = "SBE", mafThreshold = 1,
                probeExclusionPath = example_data$probeExclusionPath,
            detPtype = "m+u", verbose = FALSE, logs = FALSE
        )
        metrics_data <- extractMetricsMinfiEwasWater(
            filteredData = filtered_data,
            verbose = FALSE, logs = FALSE
        )

        list(
            RGSet = rgset, targets = targets, rawData = raw_data,
            assessment = assessment, sampleData = sample_data,
            rawFiltered = raw_filtered, sexData = sex_data,
                normData = norm_data,
            filteredData = filtered_data, metricsData = metrics_data,
            probeExclusionPath = example_data$probeExclusionPath
        )
    })
}

examplePreprocessingPhenoStateDnaEpico <- function() {
    getCachedExampleDnaEpico(key = "preprocessing_pheno", builder = function() {
        temp_dir <-
            exampleTempDirDnaEpico("dnaEPICO-preprocessingPheno-example-")

        pheno <- data.frame(
            Sample_Name = c(
                "S1", "S2", "S3",
                "S4"
            ), Timepoint = c("1", "1", "2", "2"), Sex = c(
                "Female",
                "Male", "Female", "Male"
            ), Age = c(20, 22, 21, 23),
            stringsAsFactors = FALSE
        )

        beta <- matrix(
            c(
                0.2, 0.25, 0.22, 0.27, 0.6, 0.55, 0.52,
                0.58, 0.1, 0.15, 0.12, 0.16
            ),
            nrow = 3, byrow = TRUE,
            dimnames = list(
                c("cg00000029", "cg00000108", "cg00000109"),
                pheno$Sample_Name
            )
        )
        m <- log2(beta / (1 - beta))
        cn <- matrix(1,
            nrow = nrow(beta), ncol = ncol(beta),
            dimnames = dimnames(beta)
        )

        pheno_path <- file.path(temp_dir, "phenoLC.csv")
        beta_path <- file.path(temp_dir, "beta.RData")
        m_path <- file.path(temp_dir, "m.RData")
        cn_path <- file.path(temp_dir, "cn.RData")

        utils::write.csv(pheno, pheno_path, row.names = FALSE)
        saveNamedObjectMinfiEwasWater(beta, "beta", beta_path)
        saveNamedObjectMinfiEwasWater(m, "m", m_path)
        saveNamedObjectMinfiEwasWater(cn, "cn", cn_path)

        metrics_data <- loadMetricsPreprocessingPheno(
            betaPath = beta_path,
            mPath = m_path, cnPath = cn_path, verbose = FALSE,
            logs = FALSE
        )
        timepoint_data <- splitTimepointsPreprocessingPheno(
            pheno = pheno,
            metricsData = metrics_data, SampleID = "Sample_Name",
            timeVar = "Timepoint", timepoints = "1,2", verbose = FALSE,
            logs = FALSE
        )
        combined_data <- combineTimepointsPreprocessingPheno(
            timepointData = timepoint_data,
            combineTimepoints = "1,2", verbose = FALSE, logs = FALSE
        )
        clock_foundation <- buildClockFoundationInputsPreprocessingPheno(
            beta = timepoint_data$data[["1"]]$beta,
            pheno = timepoint_data$data[["1"]]$pheno, SampleID = "Sample_Name",
            sexColumn = "Sex", verbose = FALSE, logs = FALSE
        )

        preprocessing_data <- list(
            pheno = pheno, metrics = metrics_data,
            timepointData = timepoint_data, combinedData = combined_data,
            clockFoundation = clock_foundation
        )

        list(
            tempDir = temp_dir, pheno = pheno, phenoPath = pheno_path,
            betaPath = beta_path, mPath = m_path, cnPath = cn_path,
            metricsData = metrics_data, timepointData = timepoint_data,
            combinedData = combined_data, clockFoundation = clock_foundation,
            preprocessingData = preprocessing_data
        )
    })
}

exampleSvaAnalysisStateDnaEpico <- function() {
    getCachedExampleDnaEpico(key = "sva_analysis", builder = function() {
        sample_names <- paste0("S", seq_len(8L))
        targets <- data.frame(Sample_Name = sample_names, Sentrix_ID = rep(c(
            "Slide1",
            "Slide2"
        ), each = 4L), Sentrix_Position = rep(c(
            "R01C01",
            "R02C01", "R03C01", "R04C01"
        ), 2L), stringsAsFactors = FALSE)
        RGSet <- SummarizedExperiment::SummarizedExperiment(assays = list(signal = matrix(0,
            nrow = 2L, ncol = length(sample_names)
        )), colData = S4Vectors::DataFrame(
            Sentrix_ID = targets$Sentrix_ID,
            Sentrix_Position = targets$Sentrix_Position,
                row.names = sample_names
        ))
        colnames(RGSet) <- sample_names

        sva <- cbind(PC1 = c(
            0.1, 0.25, 0.15, 0.45, 0.22, 0.3,
            0.5, 0.48
        ), PC2 = c(
            0.5, 0.4, 0.55, 0.35, 0.42, 0.6,
            0.38, 0.52
        ))
        rownames(sva) <- sample_names

        merged_pheno <- mergeSvaTargetsEnmix(
            targets = targets,
            sva = sva, SampleID = "Sample_Name", verbose = FALSE,
            logs = FALSE
        )
        analysis_data <- analyzeSvaEnmix(
            sva = sva, RGSet = RGSet,
            SentrixIDColumn = "Sentrix_ID",
                SentrixPositionColumn = "Sentrix_Position",
            verbose = FALSE, logs = FALSE
        )

        list(
            RGSet = RGSet, targets = targets, sva = sva,
                mergedPheno = merged_pheno,
            analysisData = analysis_data
        )
    })
}

exampleMethylationGLMStateDnaEpico <- function() {
    getCachedExampleDnaEpico(key = "methylation_glm", builder = function() {
        temp_dir <- exampleTempDirDnaEpico("dnaEPICO-glm-example-")
        input_path <- file.path(temp_dir, "phenoBT1.RData")

        phenoBT1 <- data.frame(
            Sample_Name = paste0("S", seq_len(8L)),
            status = factor(rep(c("Case", "Control"), each = 4L)),
            sex = factor(rep(c("F", "M"), times = 4L)), age = c(
                20,
                22, 24, 26, 21, 23, 25, 27
            ), cg00000029 = c(
                0.2,
                0.25, 0.22, 0.27, 0.21, 0.24, 0.23, 0.26
            ), cg00000108 = c(
                0.6,
                0.55, 0.52, 0.58, 0.57, 0.54, 0.53, 0.56
            ), check.names = FALSE
        )
        save(phenoBT1, file = input_path)

        prepared_data <- prepareMethylationGLMData(
            inputPheno = input_path,
            phenotypes = "status", covariates = "sex,age",
                factorVars = "status,sex",
            cpgLimit = 2, verbose = FALSE, logs = FALSE
        )
        model_results <- fitMethylationGLMModels(
            preparedData = prepared_data,
            nCores = 1, verbose = FALSE, logs = FALSE
        )
        model_summaries <- summarizeMethylationGLMModels(
            modelResults = model_results,
            preparedData = prepared_data, summaryResidualSD = TRUE,
            summaryPval = NA, nCores = 1, verbose = FALSE, logs = FALSE
        )

        list(
            tempDir = temp_dir, inputPath = input_path,
                preparedData = prepared_data,
            modelResults = model_results, modelSummaries = model_summaries,
            annotationData = data.frame(
                CpG = c(
                    "cg00000029",
                    "cg00000108"
                ), Name = c("cg00000029", "cg00000108"),
                chr = c("chr1", "chr2"), pos = c(100L, 200L),
                stringsAsFactors = FALSE
            )
        )
    })
}
exampleMethylationLMEStateDnaEpico <- function() {
    getCachedExampleDnaEpico(key = "methylation_lme", builder = function() {
        temp_dir <- exampleTempDirDnaEpico("dnaEPICO-lme-example-")
        input_path <- file.path(temp_dir, "phenoBT1T2.RData")

        phenoBT1T2 <- data.frame(
            Sample_Name = paste0("S", seq_len(8)),
            person = rep(seq_len(4), each = 2), Timepoint = rep(c(
                "T1",
                "T2"
            ), 4), score = c(
                10, 11, 8, 9, 7, 8, 12,
                13
            ), sex = factor(rep(c("F", "M"), each = 4)),
            cg00000029 = c(
                0.2, 0.22, 0.25, 0.24, 0.18, 0.19,
                0.28, 0.29
            ), cg00000108 = c(
                0.6, 0.59, 0.55,
                0.56, 0.52, 0.53, 0.58, 0.6
            ), check.names = FALSE
        )
        save(phenoBT1T2, file = input_path)

        prepared_data <- prepareMethylationLMEData(
            inputPheno = input_path,
            personVar = "person", timeVar = "Timepoint", phenotypes = "score",
            covariates = "sex", factorVars = "sex", cpgLimit = 2,
            verbose = FALSE, logs = FALSE
        )
        model_results <- fitMethylationLMEModels(
            preparedData = prepared_data,
            nCores = 1, verbose = FALSE, logs = FALSE
        )
        model_summaries <- summarizeMethylationLMEModels(
            modelResults = model_results,
            preparedData = prepared_data, summaryPval = NA, nCores = 1,
            verbose = FALSE, logs = FALSE
        )

        list(
            tempDir = temp_dir, inputPath = input_path,
                preparedData = prepared_data,
            modelResults = model_results, modelSummaries = model_summaries,
            annotationData = data.frame(
                CpG = c(
                    "cg00000029",
                    "cg00000108"
                ), Name = c("cg00000029", "cg00000108"),
                chr = c("chr1", "chr2"), pos = c(100L, 200L),
                stringsAsFactors = FALSE
            )
        )
    })
}

## ----include = FALSE----------------------------------------------------------
startTime <- Sys.time()
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>",
    crop = NULL
)

library(dnaEPICO)

can_run_preprocessing <- requireNamespace("minfiData", quietly = TRUE) &&
    requireNamespace("IlluminaHumanMethylation450kmanifest", quietly = TRUE) &&
    requireNamespace("IlluminaHumanMethylation450kanno.ilmn12.hg19", quietly = TRUE)

can_run_sva <- requireNamespace("minfiData", quietly = TRUE)

## ----eval = FALSE-------------------------------------------------------------
# if (!requireNamespace("BiocManager", quietly = TRUE)) {
#     install.packages("BiocManager")
# }
# 
# BiocManager::install("dnaEPICO")
# 
# BiocManager::valid()

## ----eval = FALSE, message = FALSE--------------------------------------------
# library("dnaEPICO")

## ----preprocessing-local-inputs, eval = can_run_preprocessing, message = FALSE, warning = FALSE----
preprocessing_inputs <- dnaEPICO:::exampleMinfiIdatInputsDnaEpico(n = 6L)

names(preprocessing_inputs)
preprocessing_inputs$tempDir
basename(preprocessing_inputs$phenoFile)
head(basename(list.files(preprocessing_inputs$idatFolder, full.names = TRUE)), 4)
basename(preprocessing_inputs$probeExclusionPath)

## ----preprocessing-local, eval = can_run_preprocessing, message = FALSE, warning = FALSE----
preprocessResult <- preprocessingMinfiEwasWater(
    phenoFile = preprocessing_inputs$phenoFile,
    idatFolder = preprocessing_inputs$idatFolder,
    outputLogs = file.path(preprocessing_inputs$tempDir, "logs"),
    nSamples = 6,
    SampleID = "Sample_Name",
    arrayType = preprocessing_inputs$arrayType,
    annotationVersion = preprocessing_inputs$annotationVersion,
    scriptLabel = "preprocessingMinfiEwasWater",
    baseDataFolder = file.path(preprocessing_inputs$tempDir, "rData"),
    figureBaseDir = file.path(preprocessing_inputs$tempDir, "figures"),
    detPThreshold = 1,
    normMethods = "quantile",
    sexColumn = "Sex",
    removeSexMismatch = FALSE,
    pvalThreshold = 1,
    chrToRemove = "",
    snpsToRemove = "SBE",
    mafThreshold = 1,
    probeExclusionPath = preprocessing_inputs$probeExclusionPath,
    plotGroupVar = "Sex",
    lcRef = "saliva",
    phenoOrder = "Sample_Name;Sex;Basename;Sentrix_ID;Sentrix_Position",
    lcPhenoDir = preprocessing_inputs$tempDir,
    display = FALSE,
    verbose = FALSE,
    logs = FALSE,
    saveOutputs = FALSE
)

class(preprocessResult)
names(preprocessResult)
head(preprocessResult$targets[, c(
    "Sample_Name",
    "Sex",
    "Sentrix_ID",
    "Sentrix_Position"
)])

## ----preprocessing-local-inspect, eval = can_run_preprocessing, message = FALSE, warning = FALSE----
class(preprocessResult$RGSet)
names(preprocessResult$metricsData)
head(preprocessResult$lcData$phenoLC[, 1:6])

## ----sva-local-inputs, eval = can_run_sva, message = FALSE, warning = FALSE----
sva_inputs <- dnaEPICO:::exampleMinfiBaseDataDnaEpico()
sva_temp_dir <- tempdir()
sva_targets_file <- file.path(sva_temp_dir, "sva_targets.csv")
sva_rgset_file <- file.path(sva_temp_dir, "sva_RGSet.RData")

names(sva_inputs)
class(sva_inputs$RGSet)
utils::write.csv(sva_inputs$targets, sva_targets_file, row.names = FALSE)
RGSet <- sva_inputs$RGSet
save(RGSet, file = sva_rgset_file)

basename(c(sva_targets_file, sva_rgset_file))
file.exists(c(sva_targets_file, sva_rgset_file))

## ----sva-local, eval = can_run_sva, message = FALSE, warning = FALSE----------
svaResult <- svaEnmix(
    phenoFile = sva_targets_file,
    rgsetData = sva_rgset_file,
    outputLogs = file.path(tempdir(), "logs"),
    SampleID = "Sample_Name",
    arrayType = "IlluminaHumanMethylation450k",
    annotationVersion = "ilmn12.hg19",
    SentrixIDColumn = "Sentrix_ID",
    SentrixPositionColumn = "Sentrix_Position",
    ctrlSvaPercVar = 0.90,
    ctrlSvaFlag = 1,
    scriptLabel = "svaEnmix",
    display = FALSE,
    verbose = FALSE,
    logs = FALSE,
    saveOutputs = FALSE
)

class(svaResult)
names(svaResult)
dim(svaResult$svaData$sva)
head(svaResult$mergedPheno[, 1:min(6, ncol(svaResult$mergedPheno))])

## ----sva-local-inspect, eval = can_run_sva, message = FALSE, warning = FALSE----
class(svaResult$analysisData)
names(svaResult$analysisData)

## ----preprocessingPheno-local-inputs, message = FALSE, warning = FALSE--------
pheno_inputs <- dnaEPICO:::examplePreprocessingPhenoStateDnaEpico()

names(pheno_inputs)
pheno_inputs$tempDir
basename(c(
    pheno_inputs$phenoPath,
    pheno_inputs$betaPath,
    pheno_inputs$mPath,
    pheno_inputs$cnPath
))

## ----preprocessingPheno-local, message = FALSE, warning = FALSE---------------
phenoResult <- preprocessingPheno(
    phenoFile = pheno_inputs$phenoPath,
    betaPath = pheno_inputs$betaPath,
    mPath = pheno_inputs$mPath,
    cnPath = pheno_inputs$cnPath,
    SampleID = "Sample_Name",
    timeVar = "Timepoint",
    timepoints = "1,2",
    combineTimepoints = "1,2",
    outputPheno = file.path(pheno_inputs$tempDir, "data", "preprocessingPheno"),
    outputRData = file.path(
        pheno_inputs$tempDir,
        "rData",
        "preprocessingPheno",
        "metrics"
    ),
    outputRDataMerge = file.path(
        pheno_inputs$tempDir,
        "rData",
        "preprocessingPheno",
        "mergeData"
    ),
    sexColumn = "Sex",
    outputLogs = file.path(pheno_inputs$tempDir, "logs"),
    outputDir = file.path(pheno_inputs$tempDir, "clockFoundation"),
    verbose = FALSE,
    logs = FALSE,
    saveOutputs = FALSE
)

class(phenoResult)
names(phenoResult)
names(phenoResult$combinedData)
head(phenoResult$combinedData$phenoMethylation[, 1:6])

## ----preprocessingPheno-local-inspect, message = FALSE, warning = FALSE-------
names(phenoResult$clockFoundation)
head(phenoResult$clockFoundation$phenoCF)

## ----echo = FALSE-------------------------------------------------------------
Sys.time()

## ----echo = FALSE-------------------------------------------------------------
totalTime <- diff(c(startTime, Sys.time()))
round(totalTime, digits = 3)

## ----echo = FALSE-------------------------------------------------------------
sessionInfo()

#' Read phenotype targets for shared dnaEPICO workflows
#'
#' Read the phenotype table used by shared `dnaEPICO` workflows, validate the
#' sample identifier column, optionally subset the first `nSamples`, and return
#' the targets as a base `data.frame`.
#'
#' @param phenoFile Character. Path to the phenotype table on disk.
#' @param sepType Character or `NULL`. Field separator used in `phenoFile`. Use
#'   `NULL` for a standard comma-separated file, `'\\t'` for a tab-delimited
#' file, or another single-character separator accepted by `utils::read.csv()`.
#' @param nSamples Integer or `NA`. Number of rows to keep from the start of the
#'   phenotype table. The default `NA` reads and returns all rows.
#' @param SampleID Character. Name of the column containing sample identifiers
#'   that will later be used to name methylation-array samples.
#' @param verbose Logical. If `TRUE`, emit progress and preview messages with
#'   `message()`. The default is `FALSE`, so the function is quiet unless the
#'   user explicitly requests messages.
#' @param logs Logical. If `TRUE`, write the same progress messages to a log
#'   file. The default is `FALSE`.
#' @param log_dir Character or `NULL`. Directory where the log file should be
#'   written when `logs = TRUE`. If `NULL`, the current working directory is
#'   used.
#' @param log_file Character. File name used when `logs = TRUE`. The default is
#'   `'log_readPhenotypeTargets.txt'`.
#'
#' @return A `data.frame` containing the phenotype targets.
#'
#' @examples
#' tmp <- tempdir()
#' pheno <- data.frame(
#'     Sample_Name = c("S1", "S2"),
#'     Sex = c("F", "M"),
#'     stringsAsFactors = FALSE
#' )
#' pheno_file <- file.path(tmp, "pheno.csv")
#' utils::write.csv(pheno, pheno_file, row.names = FALSE)
#' targets <- readPhenotypeTargets(
#'     phenoFile = pheno_file,
#'     SampleID = "Sample_Name"
#' )
#' stopifnot(is.data.frame(targets))
#' stopifnot(nrow(targets) == 2L)
#'
#' @export
readPhenotypeTargets <- function(
    phenoFile, sepType = NULL, nSamples = NA,
    SampleID = "Sample_Name", verbose = FALSE, logs = FALSE,
    log_dir = NULL, log_file = "log_readPhenotypeTargets.txt"
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )

    if (!file.exists(phenoFile)) {
        stop("phenoFile does not exist: ", phenoFile, call. = FALSE)
    }

    if (!is.character(SampleID) || length(SampleID) != 1L ||
        is.na(SampleID) || !nzchar(trimws(SampleID))) {
        stop("SampleID must be a single, non-empty character string.",
            call. = FALSE
        )
    }

    if (length(nSamples) != 1L) {
        stop("nSamples must be a single positive integer or NA.",
            call. = FALSE
        )
    }
    if (!is.na(nSamples)) {
        if (!is.numeric(nSamples) || !is.finite(nSamples) ||
            nSamples < 1L || nSamples != floor(nSamples)) {
            stop("nSamples must be a single positive integer or NA.",
                call. = FALSE
            )
        }

        nSamples <- as.integer(nSamples)
    }

    emitLogMinfiEwasWater(lines = c(
        "============================================================",
        paste("Phenotype file:           ", phenoFile), paste(
            "Separator type:           ",
            if (is.null(resolveSeparatorMinfiEwasWater(sepType))) {
                "default (',')"
            } else {
                sepType
            }
        ), paste("SampleID column:          ", SampleID),
        paste("nSamples limit:           ", if (is.na(nSamples)) {
            "all"
        } else {
            nSamples
        })
    ), verbose = verbose, log_path = log_path)

    sep_char <- resolveSeparatorMinfiEwasWater(sepType)

    if (is.null(sep_char)) {
        targets <- utils::read.csv(phenoFile, stringsAsFactors = FALSE)
    } else {
        targets <- utils::read.csv(phenoFile,
            sep = sep_char,
            stringsAsFactors = FALSE
        )
    }

    if (!(SampleID %in% colnames(targets))) {
        stop("SampleID column not found in phenotype data: ",
            SampleID,
            call. = FALSE
        )
    }
    if (nrow(targets) == 0L) {
        stop("The phenotype file contains no sample rows.", call. = FALSE)
    }

    targets[[SampleID]] <- validateSampleIdentifiersDnaEpico(
        targets[[SampleID]],
        paste0("Phenotype column '", SampleID, "'")
    )

    if (!is.na(nSamples) && nSamples < nrow(targets)) {
        targets <- targets[seq_len(nSamples), , drop = FALSE]
        subset_line <- paste("Subsetting to", nSamples, "samples for testing.")
    } else {
        subset_line <- paste("Using all", nrow(targets), "samples.")
    }

    preview_cols <- seq_len(min(ncol(targets), 5L))
    preview_lines <- utils::capture.output(utils::head(targets[,
        preview_cols,
        drop = FALSE
    ]))

    emitLogMinfiEwasWater(
        lines = c(
            subset_line, paste(
                "Phenotype file loaded with",
                nrow(targets), "samples and", ncol(targets), "columns."
            ),
            "Preview of targets:", preview_lines,
                "============================================================"
        ),
        verbose = verbose, log_path = log_path
    )

    targets
}

test_that("readPhenotypeTargets is quiet by default and subsets rows", {
    pheno <- data.frame(
        Sample_Name = c("S1", "S2", "S3"),
        Sex = c("F", "M", "F"),
        stringsAsFactors = FALSE
    )

    pheno_file <- tempfile(fileext = ".csv")
    utils::write.csv(pheno, pheno_file, row.names = FALSE)

    targets <- NULL

    expect_silent({
        targets <- dnaEPICO::readPhenotypeTargets(
            phenoFile = pheno_file,
            SampleID = "Sample_Name",
            nSamples = 2
        )
    })

    expect_s3_class(targets, "data.frame")
    expect_equal(nrow(targets), 2L)
    expect_identical(targets$Sample_Name, c("S1", "S2"))
})

test_that("readPhenotypeTargets treats absent separator values as comma defaults", {
    pheno <- data.frame(
        Sample_Name = c("S1", "S2"),
        Sex = c("F", "M"),
        stringsAsFactors = FALSE
    )

    pheno_file <- tempfile(fileext = ".csv")
    utils::write.csv(pheno, pheno_file, row.names = FALSE)

    for (sep_type in list(NULL, "", "NULL")) {
        targets <- dnaEPICO::readPhenotypeTargets(
            phenoFile = pheno_file,
            sepType = sep_type,
            SampleID = "Sample_Name"
        )

        expect_identical(targets$Sample_Name, pheno$Sample_Name)
        expect_identical(targets$Sex, pheno$Sex)
    }
})

test_that("readPhenotypeTargets preserves explicit tab separators", {
    pheno <- data.frame(
        Sample_Name = c("S1", "S2"),
        Sex = c("F", "M"),
        stringsAsFactors = FALSE
    )

    pheno_file <- tempfile(fileext = ".tsv")
    utils::write.table(
        pheno,
        pheno_file,
        sep = "\t",
        row.names = FALSE,
        quote = FALSE
    )

    for (sep_type in list("\t", "\\t")) {
        targets <- dnaEPICO::readPhenotypeTargets(
            phenoFile = pheno_file,
            sepType = sep_type,
            SampleID = "Sample_Name"
        )

        expect_identical(targets$Sample_Name, pheno$Sample_Name)
        expect_identical(targets$Sex, pheno$Sex)
    }
})

test_that("readPhenotypeTargets can emit verbose messages and write logs", {
    pheno <- data.frame(
        Sample_Name = c("S1", "S2"),
        Sex = c("F", "M"),
        stringsAsFactors = FALSE
    )

    pheno_file <- tempfile(fileext = ".csv")
    log_dir <- tempfile("readTargets-log-")
    utils::write.csv(pheno, pheno_file, row.names = FALSE)

    expect_message(
        dnaEPICO::readPhenotypeTargets(
            phenoFile = pheno_file,
            SampleID = "Sample_Name",
            verbose = TRUE,
            logs = TRUE,
            log_dir = log_dir
        ),
        "Phenotype file loaded with 2 samples and 2 columns."
    )

    log_file <- file.path(log_dir, "log_readPhenotypeTargets.txt")
    expect_true(file.exists(log_file))

    log_lines <- readLines(log_file, warn = FALSE)
    expect_true(any(grepl("Preview of targets:", log_lines, fixed = TRUE)))
})

test_that("readPhenotypeTargets validates the SampleID column", {
    pheno <- data.frame(
        sample_name = c("S1", "S2"),
        Sex = c("F", "M"),
        stringsAsFactors = FALSE
    )

    pheno_file <- tempfile(fileext = ".csv")
    utils::write.csv(pheno, pheno_file, row.names = FALSE)

    expect_error(
        dnaEPICO::readPhenotypeTargets(
            phenoFile = pheno_file,
            SampleID = "Sample_Name"
        ),
        "SampleID column not found in phenotype data"
    )
    expect_error(
        dnaEPICO::readPhenotypeTargets(
            phenoFile = pheno_file,
            SampleID = NA_character_
        ),
        "SampleID must be"
    )
})

#' Estimate cell composition for preprocessingMinfiEwasWater
#'
#' Estimate cell proportions from beta values using `estimateLC()` for saliva
#' reference panels or `ENmix::estimateCellProp()` for other supported
#' references, then merge the estimates into the phenotype table.
#'
#' @param beta Numeric matrix of beta values with probes in rows and samples in
#'   columns.
#' @param targets Phenotype data frame aligned with the columns of `beta`.
#' @param SampleID Character. Phenotype column containing sample identifiers.
#' @param lcRef Character. Cell-composition reference. Internal saliva-based
#'   references supported through `estimateLC()` are `'saliva'` and
#'   `'salivaEPIC'`. Other references are passed to
#'   `ENmix::estimateCellProp()`.
#' @param phenoOrder Character vector or semicolon-separated string describing
#'   the phenotype columns that should appear first in the merged `phenoLC`
#'   output.
#' @param constrained Logical. Passed to `estimateLC()` when an internal saliva
#'   reference is used. If `TRUE`, estimated proportions are constrained to sum
#'   to one.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return A list with class `'dnaEPICO_minfiEwasWater_lc'` containing the cell
#'   proportion matrix, merged phenotype table, reference name, and method used.
#'
#' @examples
#' ref_file <- system.file("extdata", "saliva.txt", package = "dnaEPICO")
#' beta <- as.matrix(utils::read.table(ref_file))[1:20, , drop = FALSE]
#' colnames(beta) <- c("sample1", "sample2")
#' targets <- data.frame(
#'     Sample_Name = colnames(beta),
#'     Timepoint = c("T1", "T2"),
#'     stringsAsFactors = FALSE
#' )
#' lc_data <- estimateLCMinfiEwasWater(
#'     beta = beta,
#'     targets = targets,
#'     lcRef = "saliva",
#'     phenoOrder = "Sample_Name;Timepoint"
#' )
#' stopifnot(is.data.frame(lc_data$phenoLC))
#'
#' @export
estimateLCMinfiEwasWater <- function(
    beta, targets, SampleID = "Sample_Name",
    lcRef = "salivaEPIC",
        phenoOrder = "Sample_Name;Timepoint;Sex;PredSex;Basename;Sentrix_ID;Sentrix_Position",
    constrained = FALSE, verbose = FALSE, logs = FALSE, log_dir = NULL,
    log_file = "log_estimateLCMinfiEwasWater.txt"
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )

    constrained <- validateLogicalScalarDnaEpico(
        constrained,
        "constrained"
    )
    if (!is.character(lcRef) || length(lcRef) != 1L || is.na(lcRef) ||
        !nzchar(trimws(lcRef))) {
        stop("lcRef must be one non-empty reference name.", call. = FALSE)
    }
    ewas_refs <- c("saliva", "salivaEPIC")
    use_internal <- lcRef %in% ewas_refs

    validateCellCompositionBetaStructureDnaEpico(
        beta = beta,
        requireSampleNames = TRUE, objectName = "beta"
    )
    beta_sample_ids <- as.character(colnames(beta))
    beta_validation <- prepareBetaForCellCompositionDnaEpico(beta)
    beta <- beta_validation$beta
    if (beta_validation$boundaries$NaN.Converted > 0L) {
        warning(beta_validation$boundaries$NaN.Converted,
            " NaN value(s) were converted to NA before cell-composition estimation.",
            call. = FALSE
        )
    }
    if (nrow(beta_validation$invalidCpGs) > 0L) {
        warning(nrow(beta_validation$invalidCpGs),
            " invalid CpG(s) were excluded from cell-composition estimation.",
            call. = FALSE
        )
    }
    if (!(SampleID %in% colnames(targets))) {
        stop("SampleID column not found in targets: ", SampleID,
            call. = FALSE
        )
    }
    target_sample_ids <- validateSampleIdentifiersDnaEpico(
        targets[[SampleID]],
        paste0("Phenotype column '", SampleID, "'")
    )
    target_match <- matchSampleIdentifiersDnaEpico(
        query = beta_sample_ids,
        reference = target_sample_ids,
            queryLabel = "Beta-matrix sample identifiers",
        referenceLabel = paste0(
            "phenotype column '", SampleID,
            "'"
        ), requireSameSet = TRUE
    )
    targets <- targets[target_match, , drop = FALSE]
    rownames(targets) <- NULL

    emitLogMinfiEwasWater(c(
        paste(
            "Cell composition reference:",
            lcRef
        ), if (use_internal) {
            "Using internal Houseman implementation (estimateLC)."
        } else {
            "Using ENmix Houseman-based cell composition."
        }, formatMethylationBoundariesLogDnaEpico(beta_validation$boundaries),
        if (nrow(beta_validation$issues) > 0L) {
            c("Methylation value issues:",
                previewLinesMinfiEwasWater(beta_validation$issues))
        } else {
            "Methylation value issues:     none"
        }
    ), verbose = verbose, log_path = log_path)

    if (use_internal) {
        lc <- estimateLCFromValidatedBetaDnaEpico(
            meth = beta,
            ref = lcRef, constrained = constrained
        )
        method <- "estimateLC"
    } else {
        lc <- ENmix::estimateCellProp(
            userdata = beta, refdata = lcRef,
            nonnegative = TRUE, normalize = FALSE, nProbes = 50,
            refplot = FALSE
        )
        method <- "ENmix::estimateCellProp"
    }

    if (nrow(lc) != ncol(beta)) {
        stop("Cell-composition output has ", nrow(lc), " rows for ",
            ncol(beta), " beta-matrix samples.",
            call. = FALSE
        )
    }

    phenoLC <- cbind(targets, as.data.frame(lc, stringsAsFactors = FALSE))
    phenoLC <- phenoLC[, !duplicated(colnames(phenoLC)), drop = FALSE]

    lead_cols <- splitOptionMinfiEwasWater(phenoOrder, sep = ";")
    lead_cols <- lead_cols[lead_cols %in% colnames(phenoLC)]
    remaining_cols <- setdiff(colnames(phenoLC), lead_cols)
    phenoLC <- phenoLC[, c(lead_cols, remaining_cols), drop = FALSE]

    emitLogMinfiEwasWater(
        c(
            paste(
                "Cell composition method:   ",
                method
            ), paste("phenoLC columns:           ", ncol(phenoLC)),
            paste("phenoLC rows:              ", nrow(phenoLC)),
            "============================================================"
        ),
        verbose = verbose, log_path = log_path
    )

    structure(
        list(
            lc = lc, phenoLC = phenoLC, sampleIDs = beta_sample_ids,
            SampleID = SampleID, ref = lcRef, method = method,
                methylationBoundaries = beta_validation$boundaries,
            methylationIssues = beta_validation$issues,
                invalidCpGs = beta_validation$invalidCpGs
        ),
        class = "dnaEPICO_minfiEwasWater_lc"
    )
}

#' Write the merged phenotype plus cell-composition table
#'
#' @param lcData Object returned by `estimateLCMinfiEwasWater()`.
#' @param file Character. Path to the CSV file to write.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return Invisibly returns `file`.
#'
#' @examples
#' ref_file <- system.file("extdata", "saliva.txt", package = "dnaEPICO")
#' beta <- as.matrix(utils::read.table(ref_file))[1:20, , drop = FALSE]
#' colnames(beta) <- c("sample1", "sample2")
#' targets <- data.frame(
#'     Sample_Name = colnames(beta),
#'     Timepoint = c("T1", "T2"),
#'     stringsAsFactors = FALSE
#' )
#' lc_data <- estimateLCMinfiEwasWater(
#'     beta = beta,
#'     targets = targets,
#'     lcRef = "saliva",
#'     phenoOrder = "Sample_Name;Timepoint"
#' )
#' output_file <- file.path(tempdir(), "phenoLC.csv")
#' writePhenoLCMinfiEwasWater(lcData = lc_data, file = output_file)
#' file.exists(output_file)
#'
#' @export
writePhenoLCMinfiEwasWater <- function(
    lcData, file, verbose = FALSE,
    logs = FALSE, log_dir = NULL,
        log_file = "log_writePhenoLCMinfiEwasWater.txt"
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )

    dir.create(dirname(file), recursive = TRUE, showWarnings = FALSE)
    utils::write.csv(lcData$phenoLC, file = file, row.names = FALSE)

    emitLogMinfiEwasWater(
        c(paste(
            "Saved phenoLC:             ",
            file
        ), "============================================================"),
        verbose = verbose, log_path = log_path
    )

    invisible(file)
}

# Internal methylation-scale definitions and value
# validation

validateProbabilityDnaEpico <- function(value, name, allowNA = FALSE) {
    if (length(value) != 1L) {
        stop(name, " must be a single number between 0 and 1.",
            call. = FALSE
        )
    }
    numeric_value <- utils::type.convert(as.character(value),
        as.is = TRUE,
        na.strings = character()
    )
    if (isTRUE(allowNA) && is.na(numeric_value)) {
        return(NA_real_)
    }
    if (!is.numeric(numeric_value) || is.na(numeric_value) ||
        !is.finite(numeric_value) || numeric_value < 0 || numeric_value >
        1) {
        stop(name, " must be a single number between 0 and 1.",
            call. = FALSE
        )
    }
    numeric_value
}

validateCpgPrefixDnaEpico <- function(cpgPrefix) {
    if (!is.character(cpgPrefix) || length(cpgPrefix) != 1L ||
        is.na(cpgPrefix) || !nzchar(cpgPrefix)) {
        stop("cpgPrefix must be one non-empty character string.",
            call. = FALSE
        )
    }
    cpgPrefix
}

normalizeScaleVariablesDnaEpico <- function(scaleVars = NULL) {
    variables <- unique(splitOptionMinfiEwasWater(scaleVars,
        sep = ","
    ))
    variables <- variables[!is.na(variables) & nzchar(variables) &
        !(tolower(variables) %in% c("null", "none", "na"))]
    variables
}

scaleModelVariablesDnaEpico <- function(
    data, scaleVars = NULL,
    factorVars = NULL, eligibleVars, protectedVars = NULL
) {
    scale_vars <- normalizeScaleVariablesDnaEpico(scaleVars)
    factor_vars <- unique(splitOptionMinfiEwasWater(factorVars,
        sep = ","
    ))
    eligible_vars <- unique(as.character(eligibleVars))
    protected_vars <- unique(as.character(protectedVars))

    empty_metadata <- data.frame(
        Variable = character(0), Center = numeric(0),
        Scale = numeric(0), Finite.Values = integer(0),
            Missing.Values = integer(0),
        stringsAsFactors = FALSE, check.names = FALSE
    )
    if (length(scale_vars) == 0L) {
        return(list(data = data, scaleVars = scale_vars,
            metadata = empty_metadata))
    }

    missing_vars <- setdiff(scale_vars, colnames(data))
    if (length(missing_vars) > 0L) {
        stop("Scale columns not found in inputPheno: ", paste(missing_vars,
            collapse = ", "
        ), call. = FALSE)
    }
    factor_overlap <- intersect(scale_vars, factor_vars)
    if (length(factor_overlap) > 0L) {
        stop("Variables cannot be listed in both factorVars and scaleVars: ",
            paste(factor_overlap, collapse = ", "),
            call. = FALSE
        )
    }
    protected_overlap <- intersect(scale_vars, protected_vars)
    if (length(protected_overlap) > 0L) {
        stop("Identifier or protected variables cannot be scaled: ",
            paste(protected_overlap, collapse = ", "),
            call. = FALSE
        )
    }
    unused_vars <- setdiff(scale_vars, eligible_vars)
    if (length(unused_vars) > 0L) {
        stop("scaleVars must contain fixed-effect variables used by at least one model: ",
            paste(unused_vars, collapse = ", "),
            call. = FALSE
        )
    }

    model_data <- data
    metadata_rows <- vector("list", length(scale_vars))
    for (index in seq_along(scale_vars)) {
        variable <- scale_vars[[index]]
        values <- model_data[[variable]]
        if (!is.numeric(values)) {
            stop("scaleVars must be numeric: ", variable, call. = FALSE)
        }
        if (any(is.infinite(values))) {
            stop("scaleVars contains infinite values: ", variable,
                call. = FALSE
            )
        }
        finite_values <- values[is.finite(values)]
        if (length(finite_values) < 2L) {
            stop("scaleVars requires at least two finite observations: ",
                variable,
                call. = FALSE
            )
        }
        center <- mean(finite_values)
        scale_value <- stats::sd(finite_values)
        if (!is.finite(scale_value) || scale_value <= 0) {
            stop("scaleVars has zero or undefined variance: ",
                variable,
                call. = FALSE
            )
        }
        transformed <- values
        finite_index <- is.finite(values)
        transformed[finite_index] <- (values[finite_index] -
            center) / scale_value
        model_data[[variable]] <- transformed
        metadata_rows[[index]] <- data.frame(
            Variable = variable,
            Center = center, Scale = scale_value,
                Finite.Values = length(finite_values),
            Missing.Values = sum(is.na(values)), stringsAsFactors = FALSE,
            check.names = FALSE
        )
    }

    list(data = model_data, scaleVars = scale_vars, metadata = do.call(
        rbind,
        metadata_rows
    ))
}

formatScalingMetadataLogDnaEpico <- function(metadata) {
    if (!is.data.frame(metadata) || nrow(metadata) == 0L) {
        return("Scaled variables:             none")
    }
    c(
        paste("Scaled variables:            ", paste(metadata$Variable,
            collapse = ", "
        )), "Scaling parameters (sample SD):",
        previewLinesMinfiEwasWater(metadata)
    )
}

validateMethylationProbeIdentifiersDnaEpico <- function(
    probeIDs,
    label = "Methylation probe identifiers"
) {
    probe_ids <- as.character(probeIDs)
    if (length(probe_ids) == 0L) {
        stop(label, " must contain at least one CpG identifier.",
            call. = FALSE
        )
    }
    if (anyNA(probe_ids) || any(!nzchar(trimws(probe_ids)))) {
        stop(label, " contains missing or blank CpG identifiers.",
            call. = FALSE
        )
    }
    duplicated_ids <- unique(probe_ids[duplicated(probe_ids)])
    if (length(duplicated_ids) > 0L) {
        stop(label, " contains duplicate CpG identifiers: ",
            paste(duplicated_ids, collapse = ", "),
            call. = FALSE
        )
    }
    probe_ids
}

normalizeMethylationScaleDnaEpico <- function(methylationScale = "beta") {
    choices <- c("beta", "m", "cn")
    if (!is.character(methylationScale) || length(methylationScale) !=
        1L || is.na(methylationScale)) {
        stop("methylationScale must be one of: Beta, M, CN (case-insensitive).",
            call. = FALSE
        )
    }

    methylation_scale <- tolower(trimws(methylationScale))
    if (!(methylation_scale %in% choices)) {
        stop("methylationScale must be one of: Beta, M, CN (case-insensitive).",
            call. = FALSE
        )
    }

    methylation_scale
}

methylationScaleDefinitionDnaEpico <- function(methylationScale = "beta") {
    definitions <- list(
        beta = list(
            objectPrefix = "phenoB",
            responseLabel = "Beta values", responseColumn = "beta",
            domain = "[0, 1]"
        ), m = list(
            objectPrefix = "phenoM",
            responseLabel = "M-values", responseColumn = "m",
                domain = "Unbounded; finite values required for modeling"
        ),
        cn = list(
            objectPrefix = "phenoCN", responseLabel = "Copy number values",
            responseColumn = "cn",
                domain = "Unbounded; finite values required for modeling"
        )
    )
    definitions[[normalizeMethylationScaleDnaEpico(methylationScale)]]
}

methylationScaleObjectPrefixDnaEpico <- function(methylationScale = "beta") {
    methylationScaleDefinitionDnaEpico(methylationScale)$objectPrefix
}

methylationScaleResponseLabelDnaEpico <- function(methylationScale = "beta") {
    methylationScaleDefinitionDnaEpico(methylationScale)$responseLabel
}

methylationScaleResponseColumnDnaEpico <- function(methylationScale = "beta") {
    methylationScaleDefinitionDnaEpico(methylationScale)$responseColumn
}

methylationScaleDomainDnaEpico <- function(methylationScale = "beta") {
    methylationScaleDefinitionDnaEpico(methylationScale)$domain
}

buildMethylationValidationDnaEpico <- function(
    cpgIDs, methylationScale,
    numericCpGs, finiteCounts, naCounts, nanCounts, positiveInfCounts,
    negativeInfCounts, outOfRangeCounts, lowerBoundaryCounts,
    upperBoundaryCounts, observedMinimum, observedMaximum,
        overallMinimum = NA_real_,
    overallMaximum = NA_real_
) {
    methylation_scale <- normalizeMethylationScaleDnaEpico(methylationScale)
    n_cpgs <- length(cpgIDs)
    invalid_cpgs <- !numericCpGs | positiveInfCounts > 0L | negativeInfCounts >
        0L | outOfRangeCounts > 0L | (numericCpGs & finiteCounts ==
        0L)
    issue_cpgs <- invalid_cpgs | nanCounts > 0L
    issue_indices <- which(issue_cpgs)

    reason_rules <- list(
        `CpG column is not numeric` = !numericCpGs,
        `NaN converted to NA` = nanCounts > 0L,
            `infinite methylation value` = positiveInfCounts >
            0L | negativeInfCounts > 0L,
                `beta value outside [0, 1]` = outOfRangeCounts >
            0L, `no finite methylation values` = numericCpGs &
            finiteCounts == 0L
    )
    reasons <- vapply(issue_indices, function(index) {
        active <- vapply(reason_rules, `[[`, logical(1), index)
        paste(names(reason_rules)[active], collapse = "; ")
    }, character(1))

    issues <- data.frame(
        CpG = cpgIDs[issue_indices], Scale = rep(
            methylation_scale,
            length(issue_indices)
        ), Status = ifelse(invalid_cpgs[issue_indices],
            "invalid", "NaN converted to NA"
        ), Observed.Minimum = observedMinimum[issue_indices],
        Observed.Maximum = observedMaximum[issue_indices],
            Finite.Values = finiteCounts[issue_indices],
        NA.Values = naCounts[issue_indices],
            NaN.Converted = nanCounts[issue_indices],
        Positive.Inf = positiveInfCounts[issue_indices],
            Negative.Inf = negativeInfCounts[issue_indices],
        Out.Of.Range = outOfRangeCounts[issue_indices], Reason = reasons,
        stringsAsFactors = FALSE, check.names = FALSE
    )

    beta_scale <- identical(methylation_scale, "beta")
    boundaries <- data.frame(
        Scale = methylation_scale,
            Defined.Domain = methylationScaleDomainDnaEpico(methylation_scale),
        Observed.Minimum = overallMinimum, Observed.Maximum = overallMaximum,
        Finite.Values = sum(finiteCounts), NA.Values = sum(naCounts),
        NaN.Converted = sum(nanCounts), Positive.Inf = sum(positiveInfCounts),
        Negative.Inf = sum(negativeInfCounts),
            Out.Of.Range = sum(outOfRangeCounts),
        At.Lower.Boundary = if (beta_scale) {
            sum(lowerBoundaryCounts)
        } else {
            NA_real_
        }, At.Upper.Boundary = if (beta_scale) {
            sum(upperBoundaryCounts)
        } else {
            NA_real_
        }, CpGs = n_cpgs, Numeric.CpGs = sum(numericCpGs),
        Invalid.CpGs = sum(invalid_cpgs), stringsAsFactors = FALSE,
        check.names = FALSE
    )

    list(boundaries = boundaries, issues = issues,
        invalidCpGs = issues[issues$Status ==
        "invalid", , drop = FALSE])
}

inspectMethylationMatrixDnaEpico <- function(values,
    methylationScale = "beta") {
    methylation_scale <- normalizeMethylationScaleDnaEpico(methylationScale)
    if (!is.matrix(values) || !is.numeric(values)) {
        stop("Methylation values must be supplied as a numeric matrix.",
            call. = FALSE
        )
    }

    cpg_ids <- rownames(values)
    if (is.null(cpg_ids)) {
        cpg_ids <- as.character(seq_len(nrow(values)))
    }

    nan_by_cpg <- rowSums(is.nan(values))
    values[is.nan(values)] <- NA_real_
    finite_index <- is.finite(values)
    finite_by_cpg <- rowSums(finite_index)
    na_by_cpg <- rowSums(is.na(values))
    positive_inf_by_cpg <- rowSums(is.infinite(values) & values >
        0)
    negative_inf_by_cpg <- rowSums(is.infinite(values) & values <
        0)
    beta_scale <- identical(methylation_scale, "beta")
    out_of_range_by_cpg <- if (beta_scale) {
        rowSums(finite_index & (values < 0 | values > 1))
    } else {
        integer(nrow(values))
    }
    lower_boundary_by_cpg <- if (beta_scale) {
        rowSums(values == 0, na.rm = TRUE)
    } else {
        integer(nrow(values))
    }
    upper_boundary_by_cpg <- if (beta_scale) {
        rowSums(values == 1, na.rm = TRUE)
    } else {
        integer(nrow(values))
    }

    finite_values <- values[finite_index]
    overall_minimum <- if (length(finite_values)) {
        min(finite_values)
    } else {
        NA_real_
    }
    overall_maximum <- if (length(finite_values)) {
        max(finite_values)
    } else {
        NA_real_
    }
    observed_minimum <- rep(NA_real_, nrow(values))
    observed_maximum <- rep(NA_real_, nrow(values))
    issue_indices <- which(nan_by_cpg > 0L | positive_inf_by_cpg >
        0L | negative_inf_by_cpg > 0L | out_of_range_by_cpg >
        0L | finite_by_cpg == 0L)
    for (index in issue_indices) {
        finite_row <- values[index, is.finite(values[index, ]),
            drop = TRUE
        ]
        if (length(finite_row)) {
            observed_minimum[[index]] <- min(finite_row)
            observed_maximum[[index]] <- max(finite_row)
        }
    }

    validation <- buildMethylationValidationDnaEpico(
        cpgIDs = cpg_ids,
        methylationScale = methylation_scale, numericCpGs = rep(
            TRUE,
            nrow(values)
        ), finiteCounts = finite_by_cpg, naCounts = na_by_cpg,
        nanCounts = nan_by_cpg, positiveInfCounts = positive_inf_by_cpg,
        negativeInfCounts = negative_inf_by_cpg,
            outOfRangeCounts = out_of_range_by_cpg,
        lowerBoundaryCounts = lower_boundary_by_cpg,
            upperBoundaryCounts = upper_boundary_by_cpg,
        observedMinimum = observed_minimum, observedMaximum = observed_maximum,
        overallMinimum = overall_minimum, overallMaximum = overall_maximum
    )
    c(list(values = values), validation)
}

inspectMethylationColumnsDnaEpico <- function(
    data, cpgColumns,
    methylationScale = "beta"
) {
    methylation_scale <- normalizeMethylationScaleDnaEpico(methylationScale)
    n_cpgs <- length(cpgColumns)
    numeric_cpgs <- vapply(data[cpgColumns], is.numeric, logical(1))
    finite_counts <- na_counts <- nan_counts <- positive_inf_counts <-
        negative_inf_counts <- out_of_range_counts <- lower_boundary_counts <-
        upper_boundary_counts <- integer(n_cpgs)
    observed_minimum <- observed_maximum <- rep(NA_real_, n_cpgs)
    beta_scale <- identical(methylation_scale, "beta")

    for (index in which(numeric_cpgs)) {
        cpg <- cpgColumns[[index]]
        values <- data[[cpg]]
        nan_index <- is.nan(values)
        nan_counts[[index]] <- sum(nan_index)
        values[nan_index] <- NA_real_
        data[[cpg]] <- values

        finite_values <- values[is.finite(values)]
        finite_counts[[index]] <- length(finite_values)
        na_counts[[index]] <- sum(is.na(values))
        positive_inf_counts[[index]] <- sum(is.infinite(values) &
            values > 0)
        negative_inf_counts[[index]] <- sum(is.infinite(values) &
            values < 0)
        if (length(finite_values)) {
            observed_minimum[[index]] <- min(finite_values)
            observed_maximum[[index]] <- max(finite_values)
        }
        if (beta_scale) {
            out_of_range_counts[[index]] <- sum(finite_values <
                0 | finite_values > 1)
            lower_boundary_counts[[index]] <- sum(finite_values ==
                0)
            upper_boundary_counts[[index]] <- sum(finite_values ==
                1)
        }
    }
    na_counts[!numeric_cpgs] <- vapply(
        data[cpgColumns[!numeric_cpgs]],
        function(values) sum(is.na(values)), integer(1)
    )

    finite_minimum <- observed_minimum[is.finite(observed_minimum)]
    finite_maximum <- observed_maximum[is.finite(observed_maximum)]
    validation <- buildMethylationValidationDnaEpico(
        cpgIDs = cpgColumns,
        methylationScale = methylation_scale, numericCpGs = numeric_cpgs,
        finiteCounts = finite_counts, naCounts = na_counts,
            nanCounts = nan_counts,
        positiveInfCounts = positive_inf_counts,
            negativeInfCounts = negative_inf_counts,
        outOfRangeCounts = out_of_range_counts,
            lowerBoundaryCounts = lower_boundary_counts,
        upperBoundaryCounts = upper_boundary_counts,
            observedMinimum = observed_minimum,
        observedMaximum = observed_maximum,
            overallMinimum = if (length(finite_minimum)) {
            min(finite_minimum)
        } else {
            NA_real_
        }, overallMaximum = if (length(finite_maximum)) {
            max(finite_maximum)
        } else {
            NA_real_
        }
    )
    c(list(data = data), validation)
}

formatMethylationBoundariesLogDnaEpico <- function(boundaries) {
    if (!is.data.frame(boundaries) || nrow(boundaries) == 0L) {
        return("Methylation boundaries: unavailable")
    }
    row <- boundaries[1L, , drop = FALSE]
    c(
        paste("Methylation scale:           ", row$Scale), paste(
            "Defined domain:              ",
            row$Defined.Domain
        ), paste(
            "Observed finite minimum:     ",
            row$Observed.Minimum
        ), paste(
            "Observed finite maximum:     ",
            row$Observed.Maximum
        ), paste(
            "Values at lower boundary:    ",
            row$At.Lower.Boundary
        ), paste(
            "Values at upper boundary:    ",
            row$At.Upper.Boundary
        ), paste(
            "NaN values converted to NA:  ",
            row$NaN.Converted
        ), paste(
            "Positive/negative infinity:  ",
            paste(row$Positive.Inf, row$Negative.Inf, sep = "/")
        ),
        paste("Out-of-range beta values:    ", row$Out.Of.Range),
        paste("Invalid CpGs:                ", row$Invalid.CpGs)
    )
}

invalidMethylationReasonsDnaEpico <- function(validation) {
    invalid <- validation$invalidCpGs
    if (!is.data.frame(invalid) || nrow(invalid) == 0L) {
        return(stats::setNames(character(0), character(0)))
    }
    stats::setNames(as.character(invalid$Reason), as.character(invalid$CpG))
}

newMethylationFitErrorDnaEpico <- function(
    reason, errorClass,
    status = "failed", converged = NA, convergenceMessage = NA_character_
) {
    structure(
        list(
            error = as.character(reason), status = as.character(status),
            converged = as.logical(converged),
                convergenceMessage = as.character(convergenceMessage)
        ),
        class = errorClass
    )
}

#' Normalize optional numeric inputs for methylationGLM helpers
#'
#' @param value Numeric, character, or `NA` input.
#'
#' @return A numeric scalar or `NA_real_`.
#'
#' @description
#' Internal helper that standardizes optional numeric arguments such as
#' `cpgLimit` and `summaryPval`.
#'
#' @keywords internal
#' @noRd
normalizeOptionalNumericMethylationGLM <- function(value) {
    if (is.null(value) || length(value) == 0L) {
        return(NA_real_)
    }
    if (length(value) != 1L) {
        stop("Optional numeric inputs must contain one value.",
            call. = FALSE
        )
    }

    if (is.character(value)) {
        trimmed <- trimws(value)
        if (is.na(trimmed)) {
            return(NA_real_)
        }
        if (!nzchar(trimmed) || tolower(trimmed) %in% c(
            "na",
            "null"
        )) {
            return(NA_real_)
        }
        numeric_value <- utils::type.convert(trimmed,
            as.is = TRUE,
            na.strings = character()
        )
        if (!is.numeric(numeric_value) || is.na(numeric_value)) {
            stop("Optional numeric inputs must be numeric, NA, or NULL.",
                call. = FALSE
            )
        }
        return(numeric_value)
    }

    if (is.na(value)) {
        return(NA_real_)
    }
    if (!is.numeric(value)) {
        stop("Optional numeric inputs must be numeric, NA, or NULL.",
            call. = FALSE
        )
    }

    as.numeric(value)
}

validateCpgLimitMethylationModels <- function(value) {
    if (is.na(value)) {
        return(NA_integer_)
    }
    if (!is.finite(value) || value < 1 || value != floor(value)) {
        stop("cpgLimit must be NA or a positive whole number.",
            call. = FALSE
        )
    }
    as.integer(value)
}

validatePositiveIntegerMethylationModels <- function(value, name) {
    if (length(value) != 1L) {
        stop(name, " must be a positive whole number.", call. = FALSE)
    }
    numeric_value <- utils::type.convert(as.character(value),
        as.is = TRUE,
        na.strings = character()
    )
    if (!is.numeric(numeric_value) || is.na(numeric_value) ||
        !is.finite(numeric_value) || numeric_value < 1 || numeric_value !=
        floor(numeric_value)) {
        stop(name, " must be a positive whole number.", call. = FALSE)
    }
    as.integer(numeric_value)
}

validatePAdjustmentMethodMethylationModels <- function(method) {
    if (length(method) != 1L || is.na(method) || !(method %in%
        stats::p.adjust.methods)) {
        stop("padjmethod must be one of: ", paste(stats::p.adjust.methods,
            collapse = ", "
        ), call. = FALSE)
    }
    method
}

normalizeOptionalColumnMethylationModels <- function(value, name) {
    if (is.null(value) || length(value) == 0L || all(is.na(value))) {
        return(NULL)
    }
    if (!is.character(value) || length(value) != 1L || is.na(value)) {
        stop(name, " must be NULL or one column name.", call. = FALSE)
    }
    value <- trimws(value)
    if (!nzchar(value) || tolower(value) %in% c("null", "na")) {
        return(NULL)
    }
    value
}

adjustPvaluesByTermMethylationModels <- function(
    pValues, terms,
    method
) {
    adjusted <- rep(NA_real_, length(pValues))
    groups <- split(seq_along(pValues), as.character(terms),
        drop = TRUE
    )
    for (indices in groups) {
        adjusted[indices] <- stats::p.adjust(pValues[indices],
            method = method
        )
    }
    adjusted
}

collectFitFailuresMethylationModels <- function(fits, errorClass) {
    rows <- list()
    row_index <- 1L
    for (phenotype in names(fits)) {
        fit_group <- fits[[phenotype]]
        for (cpg in names(fit_group)) {
            fit_object <- fit_group[[cpg]]
            if (!inherits(fit_object, errorClass)) {
                next
            }
            rows[[row_index]] <- data.frame(
                Phenotype = phenotype,
                CpG = cpg, Status = if (is.null(fit_object$status)) {
                    "failed"
                } else {
                    as.character(fit_object$status)
                }, Error = as.character(fit_object$error),
                stringsAsFactors = FALSE, check.names = FALSE
            )
            row_index <- row_index + 1L
        }
    }
    if (length(rows) == 0L) {
        return(data.frame(
            Phenotype = character(0), CpG = character(0),
            Status = character(0), Error = character(0),
                stringsAsFactors = FALSE,
            check.names = FALSE
        ))
    }
    out <- do.call(rbind, rows)
    rownames(out) <- NULL
    out
}

diagnosticMeanMethylationModels <- function(preparedData) {
    methylation_matrix <- preparedData$data[, preparedData$cpgColumns,
        drop = FALSE
    ]
    means <- vapply(
        methylation_matrix, meanFiniteOrNADnaEpico,
        numeric(1)
    )
    invalid_cpgs <- preparedData$invalidCpGs$CpG
    if (length(invalid_cpgs) > 0L) {
        means[names(means) %in% invalid_cpgs] <- NA_real_
    }
    scale <- preparedData$methylationScale
    if (is.null(scale) || identical(scale, "beta")) {
        return(list(values = means, label = "Average Beta"))
    }
    response_label <- preparedData$responseLabel
    if (is.null(response_label) || !nzchar(response_label)) {
        response_label <- toupper(scale)
    }
    list(values = means, label = paste("Average", response_label))
}

sanitizeDiagnosticTermDnaEpico <- function(term) {
    component <- gsub("[^A-Za-z0-9.-]+", "_", as.character(term))
    component <- gsub("^_+|_+$", "", component)
    if (nzchar(component)) {
        component
    } else {
        "term"
    }
}

buildMethylationTermDiagnosticsDnaEpico <- function(
    summaryData,
    phenotype, term, termColumn, pValueColumn, yColumn, yLabel,
    diagnosticMean, fdrThreshold
) {
    term_values <- as.character(summaryData[[termColumn]])
    term_data <- summaryData[!is.na(term_values) & term_values ==
        term, , drop = FALSE]
    valid_p <-
        is.finite(term_data[[pValueColumn]]) & term_data[[pValueColumn]] >=
        0 & term_data[[pValueColumn]] <= 1
    term_data <- term_data[valid_p, , drop = FALSE]
    if (nrow(term_data) == 0L) {
        return(NULL)
    }

    p_values <- pmax(term_data[[pValueColumn]], .Machine$double.xmin)
    chi_square <- stats::qchisq(p_values, df = 1, lower.tail = FALSE)
    lambda <- round(stats::median(chi_square) / stats::qchisq(0.5,
        df = 1
    ), 3)
    diagnostic_label <- paste0(phenotype, " [", term, "]")
    qq_data <- data.frame(
        expected = -log10(stats::ppoints(length(p_values))),
        observed = -log10(sort(p_values))
    )
    qq_plot <- ggplot2::ggplot(qq_data, ggplot2::aes(
        x = expected,
        y = observed
    )) +
        ggplot2::geom_point(color = "black") +
        ggplot2::geom_abline(intercept = 0, slope = 1, color = "red") +
        ggplot2::labs(title = paste0(
            "Q-Q Plot of p-values for ",
            diagnostic_label, "\nGenomic Inflation Factor = ",
            lambda
        ), x = "Expected -log10(p)", y = "Observed -log10(p)") +
        ggplot2::theme_minimal()

    term_data$meanMethylation <- diagnosticMean$values[term_data$CpG]
    term_data$diagnosticP <- p_values
    residual_plot <- NULL
    residual_significance_plot <- NULL
    if (yColumn %in% colnames(term_data)) {
        term_data$diagnosticY <- term_data[[yColumn]]
        residual_plot <- ggplot2::ggplot(term_data, ggplot2::aes(
            x = meanMethylation,
            y = diagnosticY
        )) +
            ggplot2::geom_point(
                alpha = 0.6,
                color = "black"
            ) +
            ggplot2::labs(
                title = paste(
                    yLabel,
                    "vs", diagnosticMean$label, "for", diagnostic_label
                ),
                x = diagnosticMean$label, y = yLabel
            ) +
            ggplot2::theme_minimal()

        complete_diagnostic <- is.finite(term_data$meanMethylation) &
            is.finite(term_data$diagnosticY)
        if (sum(complete_diagnostic) >= 3L) {
            residual_plot <- residual_plot + ggplot2::geom_smooth(
                method = "loess",
                formula = y ~ x, se = FALSE, color = "red"
            )
        }

        significant_data <- term_data[!is.na(term_data$FDR) &
            term_data$FDR < fdrThreshold, , drop = FALSE]
        residual_significance_plot <- ggplot2::ggplot(
            term_data,
            ggplot2::aes(
                x = -log10(diagnosticP), y = diagnosticY,
                color = FDR < fdrThreshold
            )
        ) +
            ggplot2::geom_point(alpha = 0.6) +
            ggrepel::geom_text_repel(
                data = significant_data,
                ggplot2::aes(label = CpG)
            ) +
            ggplot2::scale_color_manual(values = c(
                `FALSE` = "grey50",
                `TRUE` = "firebrick"
            )) +
            ggplot2::labs(
                title = paste(
                    yLabel,
                    "vs Significance for", diagnostic_label
                ), x = "-log10(p-value)",
                y = yLabel, color = paste("FDR <", fdrThreshold)
            ) +
            ggplot2::theme_minimal()
    }

    list(lambda = lambda, plots = list(
        qqplot = qq_plot, residualSD = residual_plot,
        residualSignificance = residual_significance_plot
    ))
}

#' Normalize optional chunk-size inputs for methylationGLM helpers
#'
#' @param chunkSize Integer, character, or `NULL` input.
#'
#' @return An integer scalar or `NULL`.
#'
#' @description
#' Internal helper that standardizes optional chunk-size arguments used while
#' extracting CpG-level summaries.
#'
#' @keywords internal
#' @noRd
normalizeChunkSizeMethylationGLM <- function(chunkSize) {
    if (is.null(chunkSize) || length(chunkSize) == 0L) {
        return(NULL)
    }

    if (length(chunkSize) == 1L && is.character(chunkSize)) {
        trimmed <- trimws(chunkSize)
        if (!nzchar(trimmed) || tolower(trimmed) %in% c(
            "na",
            "null"
        )) {
            return(NULL)
        }
        return(validatePositiveIntegerMethylationModels(
            trimmed,
            "chunkSize"
        ))
    }

    validatePositiveIntegerMethylationModels(chunkSize, "chunkSize")
}

#' Escape regular-expression metacharacters for methylationGLM helpers
#'
#' @param x Character scalar to escape.
#'
#' @return Character scalar safe to use in `grep()` or `grepl()`.
#'
#' @description
#' Internal helper that escapes user-supplied model terms before matching
#' coefficient names.
#'
#' @keywords internal
#' @noRd
escapeRegexMethylationGLM <- function(x) {
    gsub("([][{}()+*^$|\\?.])", "\\\\\\1", x)
}

#' Backtick variable names for methylationGLM formulas
#'
#' @param x Character vector of variable names.
#'
#' @return Character vector with names wrapped in backticks.
#'
#' @description
#' Internal helper that keeps formula construction robust when variable names
#' contain punctuation.
#'
#' @keywords internal
#' @noRd
quoteNamesMethylationGLM <- function(x) {
    if (length(x) == 0L) {
        return(character(0))
    }
    paste0("`", gsub("`", "", x, fixed = TRUE), "`")
}

#' Parse phenotype-to-PRS mappings for methylationGLM helpers
#'
#' @param prsMap Character scalar or vector describing mappings in the form
#'   `'Phenotype:PRS'`.
#'
#' @return A named character vector.
#'
#' @description
#' Internal helper that parses the optional `prsMap` argument.
#'
#' @keywords internal
#' @noRd
parsePrsMapMethylationGLM <- function(prsMap = NULL) {
    if (is.null(prsMap) || length(prsMap) == 0L) {
        return(stats::setNames(character(0), character(0)))
    }

    map_entries <- splitOptionMinfiEwasWater(prsMap, sep = ",")
    if (length(map_entries) == 0L) {
        return(stats::setNames(character(0), character(0)))
    }

    pieces <- strsplit(map_entries, split = ":", fixed = TRUE)
    invalid <- !vapply(pieces, function(x) length(x) == 2L, logical(1))
    if (any(invalid)) {
        stop("Each prsMap entry must follow the format 'Phenotype:PRS'. Invalid entries: ",
            paste(map_entries[invalid], collapse = ", "),
            call. = FALSE
        )
    }

    keys <- vapply(pieces, function(x) trimws(x[[1L]]), character(1))
    values <- vapply(pieces, function(x) trimws(x[[2L]]), character(1))
    if (any(!nzchar(keys)) || any(!nzchar(values))) {
        stop("prsMap phenotype and PRS column names cannot be blank.",
            call. = FALSE
        )
    }
    duplicated_keys <- unique(keys[duplicated(keys)])
    if (length(duplicated_keys) > 0L) {
        stop("prsMap contains duplicate phenotype mappings: ",
            paste(duplicated_keys, collapse = ", "),
            call. = FALSE
        )
    }
    stats::setNames(values, keys)
}

#' Find coefficient rows for a phenotype or interaction term
#'
#' @param coefNames Character vector of model coefficient names.
#' @param variable Character. Phenotype variable of interest.
#' @param interactionTerm Character or `NULL`. Optional interaction term.
#'
#' @return Character vector of matching coefficient names.
#'
#' @description
#' Internal helper that locates the phenotype main-effect or interaction rows in
#' a CpG-level coefficient table.
#'
#' @keywords internal
#' @noRd
findCoefficientRowsMethylationGLM <- function(
    coefNames, variable,
    interactionTerm = NULL, coefficientTerms = NULL
) {
    if (!is.null(coefficientTerms)) {
        coefficient_names <- intersect(coefNames, names(coefficientTerms))
        term_values <- coefficientTerms[coefficient_names]
        normalize_term <- function(x) {
            gsub("`", "", gsub("[[:space:]]+", "", x), fixed = TRUE)
        }
        normalized_variable <- normalize_term(variable)
        normalized_terms <- normalize_term(term_values)

        if (!is.null(interactionTerm) && nzchar(interactionTerm)) {
            normalized_interaction <- normalize_term(interactionTerm)
            matches <- vapply(strsplit(normalized_terms, ":",
                fixed = TRUE
            ), function(parts) {
                normalized_variable %in% parts && normalized_interaction %in%
                    parts
            }, logical(1))
        } else {
            matches <- normalized_terms == normalized_variable
        }

        return(coefficient_names[matches])
    }

    normalized_names <- gsub("`", "", coefNames, fixed = TRUE)
    variable_pattern <- escapeRegexMethylationGLM(variable)

    if (!is.null(interactionTerm) && nzchar(interactionTerm)) {
        interaction_pattern <- escapeRegexMethylationGLM(interactionTerm)
        matches <- grepl(paste0(
            "^", variable_pattern, ".*:",
            interaction_pattern
        ), normalized_names)
    } else {
        matches <- grepl(paste0("^", variable_pattern), normalized_names)
    }

    coefNames[matches]
}

#' Map model-matrix columns to their source formula terms
#'
#' @keywords internal
#' @noRd
removeRandomInterceptMethylationModels <- function(formulaText) {
    fixed_text <- sub("\\s*\\+\\s*\\(\\s*1\\s*\\|\\s*[^)]+\\)\\s*$",
        "", formulaText,
        perl = TRUE
    )
    if (identical(fixed_text, formulaText)) {
        stop("Could not isolate the fixed-effect formula from the mixed model.",
            call. = FALSE
        )
    }
    fixed_text
}

buildCoefficientTermMapMethylationModels <- function(
    formulaText,
    data, removeRandomEffects = FALSE
) {
    if (isTRUE(removeRandomEffects)) {
        formulaText <- removeRandomInterceptMethylationModels(formulaText)
    }
    formula_object <- stats::as.formula(formulaText)
    terms_object <- stats::terms(formula_object, data = data)
    design_terms <- stats::delete.response(terms_object)
    design <- stats::model.matrix(design_terms, data = data)
    assignments <- attr(design, "assign")
    term_labels <- attr(design_terms, "term.labels")
    mapped_terms <- rep("(Intercept)", length(assignments))
    non_intercept <- assignments > 0L
    mapped_terms[non_intercept] <- term_labels[assignments[non_intercept]]
    stats::setNames(mapped_terms, colnames(design))
}

#' Validate a fixed-effect model matrix before fitting CpGs
#'
#' @keywords internal
#' @noRd
validateFixedEffectDesignMethylationModels <- function(
    formulaText,
    data, removeRandomEffects = FALSE
) {
    if (isTRUE(removeRandomEffects)) {
        formulaText <- removeRandomInterceptMethylationModels(formulaText)
    }
    formula_object <- stats::as.formula(formulaText)
    design_terms <- stats::delete.response(stats::terms(formula_object,
        data = data
    ))
    design <- stats::model.matrix(design_terms, data = data)

    if (nrow(design) == 0L) {
        stop("No complete observations remain for the requested model.",
            call. = FALSE
        )
    }
    if (any(!is.finite(design))) {
        stop("The fixed-effect design contains non-finite numeric values. Check phenotypes, covariates, and interactions.",
            call. = FALSE
        )
    }
    design_rank <- qr(design)$rank
    if (design_rank < ncol(design)) {
        stop("The fixed-effect design matrix is rank deficient (rank ",
            design_rank, " of ", ncol(design),
                "). Check duplicated covariates, factor levels, and interactions.",
            call. = FALSE
        )
    }

    invisible(list(
        rows = nrow(design), columns = ncol(design),
        rank = design_rank
    ))
}

summarizeFitErrorsMethylationModels <- function(fitList, errorClass) {
    failed <- vapply(fitList, inherits, logical(1), what = errorClass)
    if (!any(failed)) {
        return(integer(0))
    }
    messages <- vapply(fitList[failed], function(x) x$error,
        character(1),
        USE.NAMES = FALSE
    )
    sort(table(messages), decreasing = TRUE)
}

formatFitErrorsMethylationModels <- function(errorCounts, maximum = 3L) {
    if (length(errorCounts) == 0L) {
        return("none")
    }
    keep <- seq_len(min(length(errorCounts), as.integer(maximum)))
    paste0(names(errorCounts)[keep], " (n=", as.integer(errorCounts[keep]),
        ")",
        collapse = "; "
    )
}

#' Build a GLM formula string for methylationGLM helpers
#'
#' @param phenotype Character. Phenotype variable of interest.
#' @param covariates Character vector of covariate variables.
#' @param interactionTerm Character or `NULL`. Optional interaction term.
#'
#' @return Character scalar containing a formula string.
#'
#' @description
#' Internal helper that builds the one-CpG Gaussian GLM formula.
#'
#' @keywords internal
#' @noRd
buildFormulaMethylationGLM <- function(
    phenotype, covariates = character(0),
    interactionTerm = NULL, responseVar = "beta"
) {
    if (!is.null(interactionTerm) && nzchar(interactionTerm) &&
        identical(interactionTerm, phenotype)) {
        stop("interactionTerm must differ from the phenotype being modelled.",
            call. = FALSE
        )
    }
    quoted_phenotype <- quoteNamesMethylationGLM(phenotype)
    quoted_covariates <- quoteNamesMethylationGLM(covariates)

    if (!is.null(interactionTerm) && nzchar(interactionTerm)) {
        quoted_interaction <- quoteNamesMethylationGLM(interactionTerm)
        interaction_part <- paste(quoted_phenotype, quoted_interaction,
            sep = " * "
        )
        fixed_terms <- setdiff(covariates, interactionTerm)
        quoted_fixed_terms <- quoteNamesMethylationGLM(fixed_terms)
        terms <- c(interaction_part, quoted_fixed_terms)
    } else {
        terms <- c(quoted_phenotype, quoted_covariates)
    }

    terms <- unique(terms[nzchar(terms)])
    if (length(terms) == 0L) {
        stop("At least one phenotype or covariate term is required.",
            call. = FALSE
        )
    }

    paste(responseVar, "~", paste(terms, collapse = " + "))
}

fitStatusValuesMethylationGLM <- function(modelObj) {
    if (inherits(modelObj, "dnaEPICO_methylationGLM_fit_error")) {
        converged <- if (is.null(modelObj$converged)) {
            NA
        } else {
            as.logical(modelObj$converged)[[1L]]
        }
        convergence_message <- if (is.null(modelObj$convergenceMessage)) {
            NA_character_
        } else {
            as.character(modelObj$convergenceMessage)[[1L]]
        }
        return(list(
            status = if (is.null(modelObj$status)) "failed" else as.character(modelObj$status),
            singular = NA, converged = converged,
                convergenceMessage = convergence_message,
            warning = NA_character_
        ))
    }

    fit_status <- modelObj$fitStatus
    converged <- if (is.null(fit_status$converged)) {
        TRUE
    } else {
        isTRUE(fit_status$converged)
    }
    convergence_messages <- unique(as.character(fit_status$convergenceMessages))
    convergence_messages <- convergence_messages[!is.na(convergence_messages) &
        nzchar(convergence_messages)]
    fit_warnings <- unique(as.character(fit_status$warnings))
    fit_warnings <- fit_warnings[!is.na(fit_warnings) & nzchar(fit_warnings)]

    list(status = if (!isTRUE(converged)) {
        "fitted_not_converged"
    } else if (length(fit_warnings) > 0L) {
        "fitted_with_warning"
    } else {
        "fitted"
    }, singular = NA, converged = converged,
        convergenceMessage = if (length(convergence_messages) >
        0L) {
        paste(convergence_messages, collapse = " | ")
    } else {
        NA_character_
    }, warning = if (length(fit_warnings) > 0L) {
        paste(fit_warnings, collapse = " | ")
    } else {
        NA_character_
    })
}

collectFitDiagnosticsMethylationGLM <- function(fits) {
    rows <- list()
    row_index <- 1L
    for (phenotype in names(fits)) {
        fit_group <- fits[[phenotype]]
        for (cpg in names(fit_group)) {
            fit_object <- fit_group[[cpg]]
            values <- fitStatusValuesMethylationGLM(fit_object)
            inference_included <- startsWith(values$status, "fitted")
            rows[[row_index]] <- data.frame(
                Phenotype = phenotype,
                CpG = cpg, Fit.Status = values$status,
                    Singular.Fit = values$singular,
                Converged = values$converged,
                    Convergence.Message = values$convergenceMessage,
                Fit.Warning = values$warning,
                    Inference.Included = inference_included,
                Exclusion.Reason = if (!inference_included &&
                    !is.null(fit_object$error)) {
                    as.character(fit_object$error)
                } else if (identical(values$converged, FALSE)) {
                    "model did not converge"
                } else {
                    NA_character_
                }, stringsAsFactors = FALSE, check.names = FALSE
            )
            row_index <- row_index + 1L
        }
    }

    if (length(rows) == 0L) {
        return(data.frame(
            Phenotype = character(0), CpG = character(0),
            Fit.Status = character(0), Singular.Fit = logical(0),
            Converged = logical(0), Convergence.Message = character(0),
            Fit.Warning = character(0), Inference.Included = logical(0),
            Exclusion.Reason = character(0), stringsAsFactors = FALSE,
            check.names = FALSE
        ))
    }

    diagnostics <- do.call(rbind, rows)
    rownames(diagnostics) <- NULL
    diagnostics
}

applyFitQualityExclusionsMethylationGLM <- function(
    summaryDf,
    excludeNonConverged = FALSE
) {
    summary_df <- summaryDf
    if (is.null(summary_df) || nrow(summary_df) == 0L) {
        return(summary_df)
    }
    if (!("Inference.Included" %in% names(summary_df))) {
        summary_df$Inference.Included <- TRUE
    }
    if (!("Exclusion.Reason" %in% names(summary_df))) {
        summary_df$Exclusion.Reason <- NA_character_
    }

    excluded <- rep(FALSE, nrow(summary_df))
    if (isTRUE(excludeNonConverged) && "Converged" %in% names(summary_df)) {
        converged <- as.logical(summary_df[["Converged"]])
        excluded <- !is.na(converged) & !converged
    }
    summary_df$Inference.Included[excluded] <- FALSE
    summary_df$Exclusion.Reason[excluded] <- "model did not converge"
    if (any(excluded) && "Pr(>|t|)" %in% names(summary_df)) {
        summary_df[["Pr(>|t|)"]][excluded] <- NA_real_
    }
    summary_df
}

#' Summarize a single CpG-level GLM fit
#'
#' @param cpg Character. CpG identifier.
#' @param modelObj List returned by `fitCpGModelMethylationGLM()`.
#' @param variable Character. Phenotype variable of interest.
#' @param interactionTerm Character or `NULL`. Optional interaction term.
#' @param includeResidualSD Logical. If `TRUE`, append residual standard
#'   deviations to the output.
#'
#' @return A data frame for the requested CpG, or `NULL`.
#'
#' @description
#' Internal helper that extracts phenotype-specific coefficient rows from one
#' CpG-level GLM fit.
#'
#' @keywords internal
#' @noRd
summarizeCpGFitMethylationGLM <- function(
    cpg, modelObj, variable,
    interactionTerm = NULL, includeResidualSD = TRUE
) {
    if (is.null(modelObj) || inherits(modelObj,
        "dnaEPICO_methylationGLM_fit_error")) {
        return(NULL)
    }

    coef_table <- modelObj$coef
    if (is.null(coef_table)) {
        return(NULL)
    }

    matched_rows <- findCoefficientRowsMethylationGLM(
        coefNames = rownames(coef_table),
        variable = variable, interactionTerm = interactionTerm,
        coefficientTerms = modelObj$coefficientTerms
    )
    if (length(matched_rows) == 0L) {
        return(NULL)
    }

    summary_df <- as.data.frame(coef_table[matched_rows, , drop = FALSE])
    summary_df$CpG <- cpg
    summary_df$Coefficient <- rownames(summary_df)
    fit_status <- fitStatusValuesMethylationGLM(modelObj)
    summary_df$Fit.Status <- fit_status$status
    summary_df$Converged <- fit_status$converged
    summary_df$Convergence.Message <- fit_status$convergenceMessage
    summary_df$Fit.Warning <- fit_status$warning
    summary_df$Inference.Included <- !identical(
        fit_status$converged,
        FALSE
    )
    summary_df$Exclusion.Reason <- if (identical(
        fit_status$converged,
        FALSE
    )) {
        "model did not converge"
    } else {
        NA_character_
    }

    if (isTRUE(includeResidualSD) && !is.null(modelObj$residuals)) {
        summary_df$ResidualSD <- if (!is.null(modelObj$residualSD)) {
            modelObj$residualSD
        } else {
            finite_residuals <-
                modelObj$residuals[is.finite(modelObj$residuals)]
            if (length(finite_residuals) > 1L) {
                stats::sd(finite_residuals)
            } else {
                NA_real_
            }
        }
    }

    summary_df
}

resolveParallelBackendMethylationModels <- function(nCores) {
    n_cores <- validatePositiveIntegerMethylationModels(
        nCores,
        "nCores"
    )
    requested_backend <- tolower(Sys.getenv(
        "DNAEPICO_PARALLEL_BACKEND",
        "auto"
    ))
    if (!(requested_backend %in% c("auto", "fork", "psock", "serial"))) {
        requested_backend <- "auto"
    }

    if (n_cores <= 1L || identical(requested_backend, "serial")) {
        return("serial")
    }

    fork_supported <- !identical(.Platform$OS.type, "windows")
    if (identical(requested_backend, "fork")) {
        return(if (fork_supported) "fork" else "psock")
    }
    if (identical(requested_backend, "psock")) {
        return("psock")
    }

    if (fork_supported) {
        return("fork")
    }

    "psock"
}

parallelCrossoverMethylationModels <- function(engine) {
    engine <- tolower(as.character(engine[[1L]]))
    defaults <- c(glm2 = 25000L, lme4 = 1500L, nlme = 5000L)
    if (!(engine %in% names(defaults))) {
        stop("Unsupported methylation model engine: ", engine,
            call. = FALSE
        )
    }

    configured <- getOption("dnaEPICO.parallelCrossover")
    if (!is.null(configured) && !is.null(names(configured)) &&
        engine %in% names(configured)) {
        return(validatePositiveIntegerMethylationModels(
            configured[[engine]],
            paste0("dnaEPICO.parallelCrossover['", engine, "']")
        ))
    }

    env_name <- paste0("DNAEPICO_PARALLEL_CROSSOVER_", toupper(engine))
    env_value <- Sys.getenv(env_name, unset = "")
    if (nzchar(env_value)) {
        return(validatePositiveIntegerMethylationModels(
            env_value,
            env_name
        ))
    }

    unname(defaults[[engine]])
}

availableWorkersMethylationModels <- function() {
    detect_cores <- function(logical) {
        tryCatch(parallel::detectCores(logical = logical),
            warning = function(condition) NA_integer_,
            error = function(condition) NA_integer_
        )
    }
    physical_cores <- detect_cores(FALSE)
    logical_cores <- detect_cores(TRUE)
    available <- if (length(physical_cores) == 1L &&
        is.finite(physical_cores)) {
        physical_cores
    } else if (length(logical_cores) == 1L && is.finite(logical_cores)) {
        logical_cores
    } else {
        1L
    }

    configured_cap <- Sys.getenv("DNAEPICO_MAX_WORKERS", unset = "")
    if (nzchar(configured_cap)) {
        available <- min(available, validatePositiveIntegerMethylationModels(
            configured_cap,
            "DNAEPICO_MAX_WORKERS"
        ))
    }

    max(1L, as.integer(available))
}

resolveParallelPlanMethylationModels <- function(
    engine, nCores,
    nCpGs
) {
    requested_cores <- validatePositiveIntegerMethylationModels(
        nCores,
        "nCores"
    )
    n_cpgs <- max(0L, as.integer(nCpGs))
    requested_backend <- tolower(Sys.getenv(
        "DNAEPICO_PARALLEL_BACKEND",
        "auto"
    ))
    if (!(requested_backend %in% c("auto", "fork", "psock", "serial"))) {
        requested_backend <- "auto"
    }
    crossover <- parallelCrossoverMethylationModels(engine)
    resource_cap <- availableWorkersMethylationModels()
    worker_count <- min(requested_cores, resource_cap, max(
        1L,
        n_cpgs
    ))

    if (requested_cores <= 1L || worker_count <= 1L) {
        backend <- "serial"
        reason <- "one effective worker"
    } else if (identical(requested_backend, "serial")) {
        backend <- "serial"
        worker_count <- 1L
        reason <- "serial backend requested"
    } else if (identical(requested_backend, "auto") && n_cpgs <
        crossover) {
        backend <- "serial"
        worker_count <- 1L
        reason <- paste0(
            "workload below the ", engine, " crossover of ",
            crossover, " CpGs"
        )
    } else {
        backend <- resolveParallelBackendMethylationModels(worker_count)
        reason <- if (identical(requested_backend, "auto")) {
            paste0("workload reached the ", engine, " crossover")
        } else {
            paste0(requested_backend, " backend requested")
        }
    }

    list(
        engine = tolower(as.character(engine[[1L]])), backend = backend,
        requestedCores = requested_cores, workerCount = worker_count,
        resourceWorkerCap = resource_cap, workloadCpGs = n_cpgs,
        crossoverCpGs = crossover, reason = reason, forced = !identical(
            requested_backend,
            "auto"
        )
    )
}

chunkCpGColumnsMethylationModels <- function(
    cpgColumns, nCores = 1L,
    batchesPerCore = 8L
) {
    if (length(cpgColumns) == 0L) {
        return(list())
    }

    n_cores <- validatePositiveIntegerMethylationModels(
        nCores,
        "nCores"
    )
    batches_per_core <- max(1L, as.integer(batchesPerCore))
    target_batches <- min(length(cpgColumns), max(1L, n_cores *
        batches_per_core))
    chunk_size <- max(1L, ceiling(length(cpgColumns) / target_batches))

    split(cpgColumns, ceiling(seq_along(cpgColumns) / chunk_size))
}

validateWorkerPackagesMethylationModels <- function(
    libPath = NULL,
    packages = character(0)
) {
    if (!is.null(libPath)) {
        .libPaths(unique(c(libPath, .libPaths())))
    }

    packages <- packages[!is.na(packages) & nzchar(packages)]
    for (pkg in packages) {
        if (!requireNamespace(pkg, quietly = TRUE)) {
            stop("Failed to load package: ", pkg, call. = FALSE)
        }
    }

    invisible(TRUE)
}

makePsockClusterMethylationModels <- function(clusterSize) {
    tryCatch(parallel::makeCluster(clusterSize,
        type = "PSOCK",
        useXDR = FALSE, methods = FALSE
    ), error = function(error) {
        parallel::makeCluster(clusterSize, type = "PSOCK")
    })
}

combineFitBatchResultsMethylationModels <- function(
    batchResults,
    cpgColumns
) {
    fit_chunks <- lapply(batchResults, function(result) result$fits)
    fit_chunks <- Filter(function(x) {
        !is.null(x) && length(x) >
            0L
    }, fit_chunks)
    fit_list <- if (length(fit_chunks) == 0L) {
        list()
    } else {
        out <- list()
        for (chunk in fit_chunks) {
            for (cpg in names(chunk)) {
                out[[cpg]] <- chunk[[cpg]]
            }
        }
        out
    }
    fit_list <- fit_list[cpgColumns[cpgColumns %in% names(fit_list)]]

    summary_chunks <- lapply(batchResults, function(result) result$summaries)
    summary_chunks <- Filter(function(x) {
        !is.null(x) && nrow(x) >
            0L
    }, summary_chunks)
    summary_df <- if (length(summary_chunks) == 0L) {
        data.frame()
    } else {
        out <- do.call(rbind, summary_chunks)
        rownames(out) <- NULL
        out
    }

    list(fits = fit_list, summaries = summary_df)
}

optionalTermMatchesMethylationModels <- function(requested, cached) {
    normalize <- function(value) {
        if (is.null(value) || length(value) == 0L || is.na(value[[1L]])) {
            return("")
        }
        value <- as.character(value[[1L]])
        if (!nzchar(value)) {
            return("")
        }
        value
    }

    identical(normalize(requested), normalize(cached))
}

filterSummaryByPvalueMethylationGLM <- function(
    summaryDf, pValueFilter,
    includeResidualSD = TRUE
) {
    summary_df <- summaryDf
    if (is.null(summary_df) || nrow(summary_df) == 0L) {
        return(data.frame())
    }

    ordered_columns <- c(
        "CpG", "Coefficient", "Estimate", "Std. Error",
        "t value", "Pr(>|t|)", if (isTRUE(includeResidualSD)) "ResidualSD",
        "Fit.Status", "Converged", "Convergence.Message", "Fit.Warning",
        "Inference.Included", "Exclusion.Reason"
    )
    ordered_columns <- intersect(ordered_columns, colnames(summary_df))
    summary_df <- summary_df[, ordered_columns, drop = FALSE]

    if (nrow(summary_df) > 0L && !is.na(pValueFilter)) {
        keep <- is.finite(summary_df[["Pr(>|t|)"]]) & summary_df[["Pr(>|t|)"]] <
            pValueFilter
        summary_df <- summary_df[keep, , drop = FALSE]
    }
    rownames(summary_df) <- NULL

    summary_df
}

fitCpGModelMethylationGLM <- function(
    cpg, cpgValues, modelData,
    formulaText, responseVar = "beta"
) {
    tryCatch(
        {
            model_data <- modelData
            model_data[[responseVar]] <- as.numeric(cpgValues)
            observed_response <-
                model_data[[responseVar]][!is.na(model_data[[responseVar]])]
            if (length(unique(observed_response)) < 2L) {
                stop("The CpG response has no observed variation.",
                    call. = FALSE
                )
            }

            warning_state <- new.env(parent = emptyenv())
            warning_state$messages <- character(0)
            fit <- withCallingHandlers(
                glm2::glm2(
                    formula = stats::as.formula(formulaText),
                    data = model_data, family = stats::gaussian(),
                        na.action = stats::na.exclude
                ),
                warning = function(condition) {
                    warning_state$messages <- c(
                        warning_state$messages,
                        conditionMessage(condition)
                    )
                    invokeRestart("muffleWarning")
                }
            )
            fit_warnings <- warning_state$messages
            if (!isTRUE(fit$converged)) {
                stop("The GLM did not converge.", call. = FALSE)
            }
            if (fit$rank < length(stats::coef(fit)) ||
                anyNA(stats::coef(fit))) {
                stop("The CpG-specific design became rank deficient after excluding missing values.",
                    call. = FALSE
                )
            }
            if (fit$df.residual <= 0L) {
                stop("The GLM has no residual degrees of freedom.",
                    call. = FALSE
                )
            }
            coef_table <- summary(fit)$coefficients
            required_statistics <- c(
                "Estimate", "Std. Error", "t value",
                "Pr(>|t|)"
            )
            if (!all(required_statistics %in% colnames(coef_table)) ||
                any(!is.finite(as.matrix(coef_table[, required_statistics,
                    drop = FALSE
                ])))) {
                stop("The GLM returned missing or non-finite coefficient statistics.",
                    call. = FALSE
                )
            }

            coefficient_terms <- buildCoefficientTermMapMethylationModels(
                formulaText = formulaText,
                data = model_data
            )
            residuals <- stats::residuals(fit)
            residual_sd <- if (fit$df.residual > 0L) {
                sqrt(sum(residuals^2, na.rm = TRUE) / fit$df.residual)
            } else {
                NA_real_
            }

            list(
                coef = coef_table, residuals = residuals,
                    fitted = stats::fitted(fit),
                residualSD = residual_sd, coefficientTerms = coefficient_terms,
                fitStatus = list(
                    converged = TRUE,
                        warnings = unique(as.character(fit_warnings)),
                    convergenceMessages = character(0)
                )
            )
        },
        error = function(error) {
            reason <- conditionMessage(error)
            convergence_failure <- grepl("did not converge|convergence",
                reason,
                ignore.case = TRUE
            )
            newMethylationFitErrorDnaEpico(
                reason = reason,
                    errorClass = "dnaEPICO_methylationGLM_fit_error",
                status = if (convergence_failure) {
                    "not_converged"
                } else {
                    "failed"
                }, converged = if (convergence_failure) {
                    FALSE
                } else {
                    NA
                }, convergenceMessage = if (convergence_failure) {
                    reason
                } else {
                    NA_character_
                }
            )
        }
    )
}

fitMethylationGLMBatch <- function(
    cpgBatch, data, modelData,
    formulaText, phenotype, interactionTerm = NULL, responseVar = "beta",
    invalidCpgReasons = character(0)
) {
    fits <- vector("list", length(cpgBatch))
    names(fits) <- cpgBatch
    summaries <- vector("list", length(cpgBatch))
    names(summaries) <- cpgBatch

    for (cpg in cpgBatch) {
        invalid_reason <- if (cpg %in% names(invalidCpgReasons)) {
            invalidCpgReasons[[cpg]]
        } else {
            NULL
        }
        model_obj <- if (!is.null(invalid_reason)) {
            newMethylationFitErrorDnaEpico(
                reason = invalid_reason,
                errorClass = "dnaEPICO_methylationGLM_fit_error",
                status = "invalid"
            )
        } else {
            fitCpGModelMethylationGLM(
                cpg = cpg, cpgValues = data[[cpg]],
                modelData = modelData, formulaText = formulaText,
                responseVar = responseVar
            )
        }
        fits[[cpg]] <- model_obj
        summaries[[cpg]] <- summarizeCpGFitMethylationGLM(
            cpg = cpg,
            modelObj = model_obj, variable = phenotype,
                interactionTerm = interactionTerm,
            includeResidualSD = TRUE
        )
    }

    summaries <- Filter(Negate(is.null), summaries)
    summary_df <- if (length(summaries) == 0L) {
        data.frame()
    } else {
        out <- do.call(rbind, summaries)
        rownames(out) <- NULL
        out
    }

    list(fits = fits, summaries = summary_df)
}

#' Create a distribution plot used by methylationGLM helpers
#'
#' @param values Vector of phenotype or covariate values to plot.
#' @param variable Character. Variable label.
#' @param type Character. Either `'hist'` or `'bar'`.
#' @param fill Character. Fill colour used for the bars.
#'
#' @return A `ggplot2` object.
#'
#' @description
#' Internal helper that builds exploratory distribution plots.
#'
#' @keywords internal
#' @noRd
createDistributionPlotMethylationGLM <- function(
    values, variable,
    type = c("hist", "bar"), fill = "steelblue"
) {
    type <- match.arg(type)

    if (identical(type, "hist")) {
        plot_data <- data.frame(value = as.numeric(values))
        return(ggplot2::ggplot(plot_data, ggplot2::aes(x = value)) +
            ggplot2::geom_histogram(bins = 30, fill = fill, color = "white") +
            ggplot2::labs(
                title = paste("Distribution of", variable),
                x = variable, y = "Frequency"
            ) +
            ggplot2::theme_minimal())
    }

    plot_data <- data.frame(value = as.factor(values))
    ggplot2::ggplot(plot_data, ggplot2::aes(x = value)) +
        ggplot2::geom_bar(fill = fill) +
        ggplot2::labs(
            title = paste("Distribution of", variable),
            x = variable, y = "Count"
        ) +
        ggplot2::theme_minimal()
}

#' Resolve an annotation object for methylationGLM helpers
#'
#' @param annotationObject Character package/object name, annotation data frame,
#'   or annotation object understood by `minfi::getAnnotation()`.
#'
#' @return An object accepted by `minfi::getAnnotation()`.
#'
#' @description
#' Internal helper that loads or validates the annotation resource used to
#' annotate CpG-level summary tables.
#'
#' @keywords internal
#' @noRd
resolveAnnotationObjectMethylationGLM <- function(annotationObject) {
    if (!is.character(annotationObject) || length(annotationObject) !=
        1L) {
        return(annotationObject)
    }

    if (requireNamespace(annotationObject, quietly = TRUE)) {
        annotation_lookup <-
            suppressPackageStartupMessages(tryCatch(minfi::getAnnotation(annotationObject),
            error = function(e) NULL
        ))
        if (!is.null(annotation_lookup)) {
            return(annotationObject)
        }

        annotation_namespace <- asNamespace(annotationObject)
        if (exists(annotationObject,
            envir = annotation_namespace,
            inherits = FALSE
        )) {
            return(get(annotationObject,
                envir = annotation_namespace,
                inherits = FALSE
            ))
        }
    }

    loaded_namespaces <- loadedNamespaces()
    for (namespace_name in loaded_namespaces) {
        namespace_env <- asNamespace(namespace_name)
        if (exists(annotationObject, envir = namespace_env, inherits = FALSE)) {
            return(get(annotationObject,
                envir = namespace_env,
                inherits = FALSE
            ))
        }
    }

    if (exists(annotationObject, inherits = TRUE)) {
        return(get(annotationObject, inherits = TRUE))
    }

    stop("Annotation package or object was not found: ", annotationObject,
        call. = FALSE
    )
}

coerceAnnotationDataMethylationGLM <- function(annotationObject) {
    if (is.data.frame(annotationObject)) {
        annotation_df <- annotationObject
        if (!("CpG" %in% colnames(annotation_df))) {
            if ("IlmnID" %in% colnames(annotation_df)) {
                annotation_df$CpG <- annotation_df$IlmnID
            } else if (.row_names_info(annotation_df, type = 1L) <
                0L) {
                stop("annotationObject data frames must include a CpG or IlmnID column, or explicit probe row names.",
                    call. = FALSE
                )
            } else {
                annotation_df$CpG <- rownames(annotation_df)
            }
        }
    } else {
        annotation_source <-
            resolveAnnotationObjectMethylationGLM(annotationObject)
        annotation_df <- as.data.frame(minfi::getAnnotation(annotation_source))
        annotation_df$CpG <- rownames(annotation_df)
    }

    annotation_df$CpG <- validateMethylationProbeIdentifiersDnaEpico(
        annotation_df$CpG,
        "Annotation probe identifiers"
    )
    annotation_df
}

buildAnnotatedWorkbookDictionaryMethylationGLM <- function(
    columns,
    modelDescription, formulaText, modelLabel, responseLabel
) {
    pvalue_columns <- grepl("P\\.Value$|P\\.value$", columns)
    default_formula <- paste(responseLabel, "~ formula unavailable")
    formula_values <- stats::setNames(
        as.character(formulaText),
        names(formulaText)
    )
    formula_values <- formula_values[!is.na(formula_values) &
        nzchar(formula_values)]

    select_formula <- function(column) {
        if (length(formula_values) == 0L) {
            return(default_formula)
        }

        formula_names <- names(formula_values)
        if (!is.null(formula_names) && any(nzchar(formula_names))) {
            for (formula_name in formula_names[nzchar(formula_names)]) {
                if (grepl(
                    paste0("^", escapeRegexMethylationGLM(formula_name)),
                    column
                )) {
                    return(unname(formula_values[[formula_name]]))
                }
            }
        }

        if (length(formula_values) == 1L) {
            return(unname(formula_values[[1L]]))
        }

        paste(unique(unname(formula_values)), collapse = "; ")
    }

    display_formula <- function(column) {
        formula_text <- select_formula(column)
        formula_parts <- trimws(unlist(strsplit(formula_text,
            "\\s*;\\s*",
            perl = TRUE
        ), use.names = FALSE))
        formula_parts <- vapply(formula_parts, function(formula_part) {
            sub("^\\s*[^~]+\\s*~", paste(responseLabel, "~"),
                formula_part,
                ignore.case = TRUE
            )
        }, character(1))
        paste0(modelLabel, ": ", paste(formula_parts, collapse = "; "))
    }

    descriptions <- vapply(seq_along(columns), function(index) {
        column <- columns[[index]]
        if (pvalue_columns[[index]]) {
            return(modelDescription)
        }
        if (column %in% c("IlmnID", "CpG", "Name")) {
            return("CpG probe identifier")
        }
        if (identical(column, "Fit.Status")) {
            return("Overall CpG model-fitting status across the requested phenotypes")
        }
        if (grepl("_Fit\\.Status$", column)) {
            return("Phenotype-specific CpG model-fitting status")
        }
        if (grepl("_Singular\\.Fit$", column)) {
            return("Whether lme4 detected a singular random-effects fit; NA when not assessed")
        }
        if (identical(column, "Singular.Fit")) {
            return("Whether any requested lme4 model had a singular random-effects fit; NA when not assessed")
        }
        if (identical(column, "Converged")) {
            return("Whether all assessed CpG models converged")
        }
        if (grepl("_Converged$", column)) {
            return("Whether the phenotype-specific CpG model converged")
        }
        if (grepl("_Convergence\\.Message$", column)) {
            return("Phenotype-specific convergence message retained for interpretation")
        }
        if (grepl("_Fit\\.Warning$", column)) {
            return("Phenotype-specific model-fitting warning retained for interpretation")
        }
        if (identical(column, "Inference.Included")) {
            return("Whether all requested phenotype-specific fits were included in inference")
        }
        if (grepl("_Inference\\.Included$", column)) {
            return("Whether the phenotype-specific CpG fit was included in inference")
        }
        if (identical(column, "Fit.Quality.Reason")) {
            return("Reason a CpG fit was excluded from inference while retaining its row")
        }
        if (grepl("_Exclusion\\.Reason$", column)) {
            return("Phenotype-specific reason the CpG fit was excluded from inference")
        }
        if (identical(column, "Failure.Phenotypes")) {
            return("Phenotypes for which the CpG model was invalid or failed")
        }
        if (identical(column, "Failure.Reason")) {
            return("Reported reason for an invalid or failed CpG model")
        }
        "Genomic annotation or supporting result column"
    }, character(1))
    formulas <- rep("", length(columns))
    formulas[pvalue_columns] <- vapply(
        columns[pvalue_columns],
        display_formula, character(1)
    )

    data.frame(
        Column = columns, Description = descriptions,
        Formula = formulas, stringsAsFactors = FALSE, check.names = FALSE
    )
}

orderAnnotatedModelColumnsDnaEpico <- function(
    data, annotationCols = character(0),
    includeSingular = FALSE
) {
    columns <- colnames(data)
    id_columns <- intersect(c("IlmnID", "CpG"), columns)
    pvalue_columns <- columns[grepl(
        "P\\.Value$|P\\.value$",
        columns
    )]
    annotation_columns <- intersect(annotationCols, columns)
    diagnostic_suffixes <- c(
        "Fit.Status", if (isTRUE(includeSingular)) "Singular.Fit",
        "Converged", "Convergence.Message", "Fit.Warning", "Inference.Included",
        "Exclusion.Reason"
    )
    diagnostic_pattern <- paste0("_(?:", paste(gsub(
        "\\.", "\\\\.",
        diagnostic_suffixes
    ), collapse = "|"), ")$")
    diagnostic_columns <- columns[grepl(diagnostic_pattern, columns,
        perl = TRUE
    )]
    if (!isTRUE(includeSingular)) {
        diagnostic_columns <- diagnostic_columns[!grepl(
            "_Singular\\.Fit$",
            diagnostic_columns
        )]
    }
    supporting_columns <- setdiff(columns, c(
        id_columns, pvalue_columns,
        annotation_columns, diagnostic_columns
    ))
    ordered_columns <- unique(c(
        id_columns, pvalue_columns, annotation_columns,
        supporting_columns, diagnostic_columns
    ))
    data[, ordered_columns, drop = FALSE]
}

modelSettingTextDnaEpico <- function(value, empty = "None") {
    if (is.null(value) || length(value) == 0L || all(is.na(value))) {
        return(empty)
    }
    value <- as.character(value)
    value <- value[!is.na(value) & nzchar(value)]
    if (length(value) == 0L) {
        empty
    } else {
        paste(value, collapse = ",")
    }
}

buildModelWorkbookMetadataDnaEpico <- function(
    modelResults,
    modelSummaries, annotatedResults, analysis = c("glm", "lme")
) {
    analysis <- match.arg(analysis)
    settings <- modelResults$settings
    diagnostics <- modelResults$fitDiagnostics
    if (!is.data.frame(diagnostics)) {
        diagnostics <- data.frame()
    }
    fit_status <- if ("Fit.Status" %in% colnames(diagnostics)) {
        as.character(diagnostics$Fit.Status)
    } else {
        character(0)
    }
    fitted <- startsWith(fit_status, "fitted")
    converged <- if ("Converged" %in% colnames(diagnostics)) {
        as.logical(diagnostics$Converged)
    } else {
        logical(0)
    }
    included <- if ("Inference.Included" %in% colnames(diagnostics)) {
        as.logical(diagnostics$Inference.Included)
    } else {
        logical(0)
    }
    singular <- if ("Singular.Fit" %in% colnames(diagnostics)) {
        as.logical(diagnostics$Singular.Fit)
    } else {
        logical(0)
    }
    lme_engine <- if (identical(analysis, "lme")) {
        modelSettingTextDnaEpico(settings$lmeEngine, empty = "lme4")
    } else {
        "glm2"
    }
    engine_package <- if (identical(analysis, "glm")) {
        "glm2"
    } else if (identical(lme_engine, "nlme")) {
        "nlme"
    } else {
        "lmerTest"
    }
    engine_version <-
        tryCatch(as.character(utils::packageVersion(engine_package)),
        error = function(error) "unavailable"
    )
    annotation_used <- if (!is.null(annotatedResults$annotationColumnsUsed)) {
        annotatedResults$annotationColumnsUsed
    } else {
        character(0)
    }
    annotation_missing <-
        if (!is.null(annotatedResults$missingAnnotationCols)) {
        annotatedResults$missingAnnotationCols
    } else {
        character(0)
    }

    keys <- c(
        "analysis", "backend", "fitting_function", "engine_version",
        "formulas", "response_label", "methylation_scale", "sample_count",
        "phenotypes", "covariates", "factor_variables", "scaled_variables",
        "interaction_term", "cpg_count", "fit_count", "fitted_count",
        "failed_or_invalid_count", "non_converged_count", "singular_count",
        "inference_excluded_count", "parallel_backend", "worker_count",
        "annotation_columns", "missing_annotation_columns", "created"
    )
    fitting_function <- if (identical(analysis, "glm")) {
        "glm2::glm2"
    } else if (identical(lme_engine, "nlme")) {
        "nlme::lme"
    } else {
        "lmerTest::lmer"
    }
    values <- c(
        if (identical(analysis, "glm")) "methylationGLM" else "methylationLME",
        lme_engine, fitting_function, engine_version,
            paste(paste(names(modelResults$formulas),
            modelResults$formulas,
            sep = ": "
        ), collapse = " | "),
        inferMethylationValueLabelMethylationGLM(modelResults),
        modelSettingTextDnaEpico(settings$methylationScale),
        modelSettingTextDnaEpico(settings$sampleCount),
            modelSettingTextDnaEpico(modelResults$phenotypes),
        modelSettingTextDnaEpico(settings$covariates),
            modelSettingTextDnaEpico(settings$factorVars),
        modelSettingTextDnaEpico(settings$scaleVars),
            modelSettingTextDnaEpico(settings$interactionTerm),
        as.character(if (!is.null(annotatedResults$data)) nrow(annotatedResults$data) else NA_integer_),
        as.character(nrow(diagnostics)), as.character(sum(fitted,
            na.rm = TRUE
        )), as.character(sum(!fitted, na.rm = TRUE)),
        as.character(sum(!is.na(converged) & !converged)),
            as.character(sum(!is.na(singular) &
            singular)), as.character(sum(!is.na(included) & !included)),
        modelSettingTextDnaEpico(settings$parallelBackend),
            modelSettingTextDnaEpico(settings$workerCount),
        modelSettingTextDnaEpico(annotation_used),
            modelSettingTextDnaEpico(annotation_missing),
        format(Sys.time(), tz = "UTC", usetz = TRUE)
    )
    metadata <- data.frame(
        Key = keys, Value = values, stringsAsFactors = FALSE,
        check.names = FALSE
    )

    factor_levels <- settings$factorLevels
    if (is.list(factor_levels) && length(factor_levels) > 0L) {
        factor_rows <- data.frame(Key = paste0(
            "factor.", names(factor_levels),
            ".levels"
        ), Value = vapply(
            factor_levels, function(levels) modelSettingTextDnaEpico(levels),
            character(1)
        ), stringsAsFactors = FALSE, check.names = FALSE)
        metadata <- rbind(metadata, factor_rows)
    }
    if (identical(analysis, "lme")) {
        lme_rows <- data.frame(
            Key = c(
                "libraries", "person_variable",
                "time_variable", "correlation_structure",
                    "correlation_variable",
                "exclude_singular", "exclude_non_converged"
            ), Value = c(
                modelSettingTextDnaEpico(settings$lmeLibs),
                modelSettingTextDnaEpico(settings$personVar),
                    modelSettingTextDnaEpico(settings$timeVar),
                modelSettingTextDnaEpico(settings$correlationStructure,
                    empty = "none"
                ), modelSettingTextDnaEpico(settings$correlationVar),
                modelSettingTextDnaEpico(modelSummaries$settings$excludeSingular),
                modelSettingTextDnaEpico(modelSummaries$settings$excludeNonConverged)
            ),
            stringsAsFactors = FALSE, check.names = FALSE
        )
        metadata <- rbind(metadata, lme_rows)
    } else {
        glm_rows <- data.frame(
            Key = c("libraries", "exclude_non_converged"),
            Value = c(
                modelSettingTextDnaEpico(settings$glmLibs),
                modelSettingTextDnaEpico(modelSummaries$settings$excludeNonConverged)
            ),
            stringsAsFactors = FALSE, check.names = FALSE
        )
        metadata <- rbind(metadata, glm_rows)
    }

    scaling <- settings$scalingMetadata
    if (is.data.frame(scaling) && nrow(scaling) > 0L) {
        scaling_rows <- do.call(rbind, lapply(
            seq_len(nrow(scaling)),
            function(index) {
                variable <- scaling$Variable[[index]]
                data.frame(
                    Key = c(
                        paste0(
                            "scale.", variable,
                            ".center"
                        ), paste0("scale.", variable, ".sd"),
                        paste0("scale.", variable, ".finite_values"),
                        paste0("scale.", variable, ".missing_values")
                    ),
                    Value = as.character(c(
                        scaling$Center[[index]],
                        scaling$Scale[[index]], scaling$Finite.Values[[index]],
                        scaling$Missing.Values[[index]]
                    )), stringsAsFactors = FALSE,
                    check.names = FALSE
                )
            }
        ))
        metadata <- rbind(metadata, scaling_rows)
    }
    rownames(metadata) <- NULL
    metadata
}

writeAnnotatedWorkbookMethylationGLM <- function(
    annotated_df,
    file, resultSheet, dictionary, metadata = NULL
) {
    workbook <- openxlsx::createWorkbook()
    openxlsx::addWorksheet(workbook, resultSheet)
    openxlsx::writeData(workbook, sheet = resultSheet, x = annotated_df)
    if (is.data.frame(metadata) && nrow(metadata) > 0L) {
        openxlsx::addWorksheet(workbook, "metadata")
        openxlsx::writeData(workbook, sheet = "metadata", x = metadata)
    }
    openxlsx::addWorksheet(workbook, "dictionary")
    openxlsx::writeData(workbook, sheet = "dictionary", x = dictionary)
    openxlsx::saveWorkbook(workbook, file = file, overwrite = TRUE)

    invisible(file)
}

sortReportTableDnaEpico <- function(tableData) {
    table_data <- as.data.frame(tableData,
        stringsAsFactors = FALSE,
        check.names = FALSE
    )
    id_column <- if ("IlmnID" %in% names(table_data)) {
        "IlmnID"
    } else if ("CpG" %in% names(table_data)) {
        "CpG"
    } else if (ncol(table_data) > 0L) {
        names(table_data)[[1L]]
    } else {
        ""
    }

    if (nzchar(id_column) && nrow(table_data) > 1L) {
        row_order <- order(as.character(table_data[[id_column]]),
            na.last = TRUE
        )
        if (!identical(row_order, seq_len(nrow(table_data)))) {
            table_data <- table_data[row_order, , drop = FALSE]
            rownames(table_data) <- NULL
        }
    }

    list(data = table_data, idColumn = id_column)
}

reportTableSidecarPathsDnaEpico <- function(sidecarDir, sheet) {
    list(
        table = file.path(sidecarDir, paste0(sheet, ".tsv.gz")),
        metadata = file.path(sidecarDir, paste0(sheet, ".report.tsv")),
        dictionary = file.path(sidecarDir, paste0(sheet, ".dictionary.tsv")),
        workbookMetadata = file.path(sidecarDir, paste0(
            sheet,
            ".metadata.tsv"
        ))
    )
}

writeReportTableSidecarDnaEpico <- function(
    tableData, workbookFile, sidecarDir,
    sheet, idColumn, dictionary, workbookMetadata = NULL
) {
    paths <- reportTableSidecarPathsDnaEpico(sidecarDir, sheet)
    dir.create(dirname(paths$table), recursive = TRUE, showWarnings = FALSE)
    if (!(idColumn %in% names(tableData))) {
        stop("The report table identifier column was not found.",
            call. = FALSE
        )
    }
    row_order <- order(as.character(tableData[[idColumn]]), na.last = TRUE)
    if (!identical(row_order, seq_len(nrow(tableData)))) {
        stop("The report table must be sorted by its identifier column.",
            call. = FALSE
        )
    }
    data.table::fwrite(tableData,
        file = paths$table, sep = "\t",
        quote = TRUE, na = "", compress = "gzip", showProgress = FALSE
    )

    sidecar_metadata <- data.frame(Key = c(
        "format_version",
        "sheet", "workbook_md5", "rows", "columns", "id_column",
        "sorted_by_id"
    ), Value = c(
        "1", sheet, unname(tools::md5sum(workbookFile)),
        as.character(nrow(tableData)), as.character(ncol(tableData)),
        idColumn, "TRUE"
    ), stringsAsFactors = FALSE, check.names = FALSE)
    data.table::fwrite(dictionary,
        file = paths$dictionary, sep = "\t",
        quote = TRUE, na = ""
    )
    if (is.data.frame(workbookMetadata) && nrow(workbookMetadata) >
        0L) {
        data.table::fwrite(workbookMetadata,
            file = paths$workbookMetadata,
            sep = "\t", quote = TRUE, na = ""
        )
    } else if (file.exists(paths$workbookMetadata)) {
        unlink(paths$workbookMetadata, force = TRUE)
    }
    data.table::fwrite(sidecar_metadata,
        file = paths$metadata,
        sep = "\t", quote = TRUE, na = ""
    )

    paths
}

resolveReportTableSidecarDnaEpico <- function(workbookFile, sidecarDir, sheet) {
    paths <- reportTableSidecarPathsDnaEpico(sidecarDir, sheet)
    unavailable <- function(reason) {
        list(
            ok = FALSE, reason = reason, table = paths$table,
            metadata = paths$metadata
        )
    }
    if (!file.exists(paths$table) || !file.exists(paths$metadata) ||
        !file.exists(paths$dictionary)) {
        return(unavailable("A report table sidecar was not found."))
    }

    metadata <- tryCatch(
        utils::read.delim(paths$metadata,
            check.names = FALSE,
            stringsAsFactors = FALSE, quote = "\"", comment.char = ""
        ),
        error = function(error) error
    )
    if (inherits(metadata, "error") || !all(c("Key", "Value") %in%
        names(metadata))) {
        return(unavailable("The report table sidecar metadata is invalid."))
    }
    values <- as.character(metadata$Value)
    names(values) <- as.character(metadata$Key)
    required <- c(
        "format_version", "sheet", "workbook_md5",
        "rows", "columns", "id_column", "sorted_by_id"
    )
    if (!all(required %in% names(values))) {
        return(unavailable("The report table sidecar metadata is incomplete."))
    }
    if (!identical(values[["format_version"]], "1") || !identical(
        values[["sheet"]],
        sheet
    ) || !identical(values[["sorted_by_id"]], "TRUE")) {
        return(unavailable("The report table sidecar metadata is incompatible."))
    }
    workbook_md5 <- unname(tools::md5sum(workbookFile))
    if (!identical(values[["workbook_md5"]], workbook_md5)) {
        return(unavailable("The report table sidecar does not match the workbook."))
    }

    list(
        ok = TRUE, reason = NULL, table = paths$table,
            metadata = paths$metadata,
        dictionary = paths$dictionary,
            workbookMetadata = paths$workbookMetadata,
        rows = as.integer(values[["rows"]]),
            columns = as.integer(values[["columns"]]),
        idColumn = values[["id_column"]]
    )
}

streamReportTableDnaEpico <- function(
    tableFile, chunkSize = 5000L,
    expectedRows = NULL, expectedColumns = NULL, chunkHandler = NULL
) {
    chunk_size <- max(1L, as.integer(chunkSize))
    connection <- if (grepl("\\.gz$", tableFile, ignore.case = TRUE)) {
        gzfile(tableFile, open = "rt")
    } else {
        file(tableFile, open = "rt")
    }
    on.exit(close(connection), add = TRUE)

    header_line <- readLines(connection, n = 1L, warn = FALSE)
    if (length(header_line) == 0L) {
        stop("The report table sidecar is empty.", call. = FALSE)
    }
    columns <- names(utils::read.delim(
        text = header_line, nrows = 0L,
        check.names = FALSE, quote = "\"", comment.char = ""
    ))
    if (!is.null(expectedColumns) && !identical(
        length(columns),
        as.integer(expectedColumns)
    )) {
        stop("The report table sidecar columns do not match its metadata.",
            call. = FALSE
        )
    }

    n_rows <- 0L
    chunk_number <- 1L
    maximum_chunk_rows <- 0L
    repeat {
        chunk <- utils::read.delim(connection,
            header = FALSE,
            col.names = columns, nrows = chunk_size, sep = "\t",
            quote = "\"", comment.char = "", colClasses = "character",
            check.names = FALSE, na.strings = ""
        )
        if (nrow(chunk) == 0L) {
            break
        }
        if (is.function(chunkHandler)) {
            chunkHandler(chunk, chunk_number)
        }
        n_rows <- n_rows + nrow(chunk)
        maximum_chunk_rows <- max(maximum_chunk_rows, nrow(chunk))
        chunk_number <- chunk_number + 1L
    }
    close(connection)
    on.exit(NULL, add = FALSE)

    if (!is.null(expectedRows) && !identical(
        as.integer(n_rows),
        as.integer(expectedRows)
    )) {
        stop("The report table sidecar row count does not match its metadata.",
            call. = FALSE
        )
    }

    list(columns = columns, nRows = n_rows, nChunks = chunk_number -
        1L, maximumChunkRows = maximum_chunk_rows)
}

inferMethylationValueLabelMethylationGLM <- function(modelResults) {
    if (!is.null(modelResults$responseLabel) &&
        nzchar(modelResults$responseLabel)) {
        return(modelResults$responseLabel)
    }

    if (!is.null(modelResults$settings$methylationScale)) {
        return(methylationScaleResponseLabelDnaEpico(modelResults$settings$methylationScale))
    }

    if (is.null(modelResults$fits) || length(modelResults$fits) ==
        0L) {
        return("Beta values")
    }

    found_response <- FALSE
    for (fit_group in modelResults$fits) {
        if (!is.list(fit_group)) {
            next
        }

        for (fit_object in fit_group) {
            if (!is.list(fit_object) || inherits(
                fit_object,
                "dnaEPICO_methylationGLM_fit_error"
            ) || inherits(
                fit_object,
                "dnaEPICO_methylationLME_fit_error"
            ) || is.null(fit_object$fitted) ||
                is.null(fit_object$residuals)) {
                next
            }

            fitted_values <- fit_object$fitted
            residual_values <- fit_object$residuals
            if (!is.numeric(fitted_values) || !is.numeric(residual_values) ||
                length(fitted_values) != length(residual_values)) {
                next
            }

            response_values <- fitted_values + residual_values
            response_values <- response_values[is.finite(response_values)]
            if (length(response_values) == 0L) {
                next
            }

            found_response <- TRUE
            if (any(response_values < 0 | response_values > 1,
                na.rm = TRUE
            )) {
                return("M-values")
            }
        }
    }

    if (isTRUE(found_response)) {
        return("Beta values")
    }

    "Beta values"
}

#' Prepare phenotype-plus-methylation data for one-timepoint GLM analyses
#'
#' @param inputPheno Character. Path to the merged phenotype-plus-methylation
#' object
#'   created by `preprocessingPheno()`.
#' @param phenotypes Character vector or comma-separated string of phenotype
#'   variables to model.
#' @param covariates Character vector or comma-separated string of covariate
#'   variables to adjust for.
#' @param factorVars Character vector or comma-separated string of variables
#' that
#'   should be converted to factors before modeling.
#' @param scaleVars Character vector, comma-separated variable names, or `NULL`.
#'   Numeric fixed-effect variables to standardize before model fitting.
#' @param cpgPrefix Character. Prefix used to identify methylation columns.
#' @param cpgLimit Integer or `NA`. Maximum number of CpGs to retain. `NA`
#'   keeps all matching CpGs.
#' @param methylationScale Character. Methylation metric represented by the CpG
#'   columns. One of `'Beta'`, `'M'`, or `'CN'`, case-insensitively.
#' @param interactionTerm Character or `NULL`. Optional interaction term.
#' @param prsMap Character vector or comma-separated string of phenotype-to-PRS
#'   mappings in the form `'Phenotype:PRS'`.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return A list with class `'dnaEPICO_methylationGLM_data'` containing the
#'   prepared analysis data, parsed variable selections, CpG columns, and
#'   exploratory summaries.
#'
#' @description
#' Load the merged phenotype-plus-methylation input object, validate the
#' requested
#' modeling variables, convert selected variables to factors, and return a
#' single in-memory object for downstream helpers.
#'
#' @examples
#' ex <- dnaEPICO:::exampleMethylationGLMStateDnaEpico()
#' prepared_data <- prepareMethylationGLMData(
#'     inputPheno = ex$inputPath,
#'     phenotypes = "status",
#'     covariates = "sex,age",
#'     factorVars = "status,sex",
#'     cpgLimit = 2,
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#' names(prepared_data)
#'
#' @export
prepareMethylationGLMData <- function(
    inputPheno, phenotypes,
    covariates, factorVars, scaleVars = NULL, cpgPrefix = "cg",
    cpgLimit = NA, methylationScale = "beta", interactionTerm = NULL,
    prsMap = NULL, verbose = FALSE, logs = FALSE, log_dir = NULL,
    log_file = "log_methylationGLM.txt"
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )
    phenotype_list <- unique(splitOptionMinfiEwasWater(phenotypes,
        sep = ","
    ))
    covariate_list <- splitOptionMinfiEwasWater(covariates, sep = ",")
    factor_list <- splitOptionMinfiEwasWater(factorVars, sep = ",")
    scale_list <- normalizeScaleVariablesDnaEpico(scaleVars)
    prs_map <- parsePrsMapMethylationGLM(prsMap)
    cpg_limit <-
        validateCpgLimitMethylationModels(normalizeOptionalNumericMethylationGLM(cpgLimit))
    cpgPrefix <- validateCpgPrefixDnaEpico(cpgPrefix)
    methylation_scale <- normalizeMethylationScaleDnaEpico(methylationScale)
    methylation_label <-
        methylationScaleResponseLabelDnaEpico(methylation_scale)
    methylation_prefix <-
        methylationScaleObjectPrefixDnaEpico(methylation_scale)
    response_column <- methylationScaleResponseColumnDnaEpico(methylation_scale)
    input_object_name <- sub("[.][^.]+$", "", basename(inputPheno))
    scale_named_input <- if (startsWith(input_object_name,
        methylation_prefix)) {
        input_object_name
    } else {
        character(0)
    }
    legacy_input_name <- if (identical(methylation_scale, "beta")) {
        "phenoBT1"
    } else {
        character(0)
    }
    analysis_data <- loadSavedObjectPreprocessingPheno(inputPheno,
        preferred_name = c(scale_named_input, paste0(
            methylation_prefix,
            "T1"
        ), legacy_input_name, input_object_name)
    )

    if (!is.data.frame(analysis_data)) {
        analysis_data <- as.data.frame(analysis_data, stringsAsFactors = FALSE)
    }

    if (length(phenotype_list) == 0L) {
        stop("At least one phenotype must be supplied.", call. = FALSE)
    }

    missing_phenotypes <- setdiff(phenotype_list, colnames(analysis_data))
    if (length(missing_phenotypes) > 0L) {
        stop("Phenotype columns not found in inputPheno: ",
            paste(missing_phenotypes,
            collapse = ", "
        ), call. = FALSE)
    }

    missing_covariates <- setdiff(covariate_list, colnames(analysis_data))
    if (length(missing_covariates) > 0L) {
        stop("Covariate columns not found in inputPheno: ",
            paste(missing_covariates,
            collapse = ", "
        ), call. = FALSE)
    }

    mapped_prs <- unname(prs_map[names(prs_map) %in% phenotype_list])
    missing_prs <- setdiff(mapped_prs, colnames(analysis_data))
    if (length(missing_prs) > 0L) {
        stop("PRS columns not found in inputPheno: ", paste(unique(missing_prs),
            collapse = ", "
        ), call. = FALSE)
    }

    resolved_interaction <- normalizeOptionalColumnMethylationModels(
        interactionTerm,
        "interactionTerm"
    )
    if (!is.null(resolved_interaction) && nzchar(resolved_interaction)) {
        if (!(resolved_interaction %in% colnames(analysis_data))) {
            stop("interactionTerm column not found in inputPheno: ",
                resolved_interaction,
                call. = FALSE
            )
        }
    }

    missing_factor_vars <- setdiff(factor_list, colnames(analysis_data))
    if (length(missing_factor_vars) > 0L) {
        stop("Factor columns not found in inputPheno: ",
            paste(missing_factor_vars,
            collapse = ", "
        ), call. = FALSE)
    }

    for (var in intersect(factor_list, colnames(analysis_data))) {
        analysis_data[[var]] <- as.factor(analysis_data[[var]])
    }

    cpg_columns <- grep(paste0("^", escapeRegexMethylationGLM(cpgPrefix)),
        colnames(analysis_data),
        value = TRUE
    )
    if (!is.na(cpg_limit)) {
        cpg_columns <- utils::head(cpg_columns, as.integer(cpg_limit))
    }
    if (length(cpg_columns) == 0L) {
        stop("No CpG columns were found with prefix '", cpgPrefix,
            "'.",
            call. = FALSE
        )
    }
    validateMethylationProbeIdentifiersDnaEpico(
        cpg_columns,
        "CpG columns in the GLM input"
    )
    methylation_validation <- inspectMethylationColumnsDnaEpico(
        data = analysis_data,
        cpgColumns = cpg_columns, methylationScale = methylation_scale
    )
    analysis_data <- methylation_validation$data
    if (methylation_validation$boundaries$NaN.Converted > 0L) {
        warning(methylation_validation$boundaries$NaN.Converted,
            " NaN methylation values were converted to NA. See methylationIssues.",
            call. = FALSE
        )
    }
    if (nrow(methylation_validation$invalidCpGs) > 0L) {
        warning(nrow(methylation_validation$invalidCpGs),
            " invalid CpG columns will be reported as failed fits instead of stopping the analysis.",
            call. = FALSE
        )
    }

    fixed_effect_variables <- unique(c(
        phenotype_list, covariate_list,
        mapped_prs, resolved_interaction
    ))
    fixed_effect_variables <-
        fixed_effect_variables[!is.na(fixed_effect_variables) &
        nzchar(fixed_effect_variables)]
    scaled_cpgs <- intersect(scale_list, cpg_columns)
    if (length(scaled_cpgs) > 0L) {
        stop("CpG methylation response columns cannot be listed in scaleVars: ",
            paste(scaled_cpgs, collapse = ", "),
            call. = FALSE
        )
    }
    model_columns <- setdiff(colnames(analysis_data), cpg_columns)
    scaling <- scaleModelVariablesDnaEpico(
        data = analysis_data[,
            model_columns,
            drop = FALSE
        ], scaleVars = scale_list,
        factorVars = factor_list, eligibleVars = fixed_effect_variables,
        protectedVars = cpg_columns
    )

    requested_columns <- unique(c(phenotype_list, covariate_list))
    missing_counts <- vapply(
        requested_columns,
            function(column_name) sum(is.na(analysis_data[[column_name]])),
        integer(1)
    )
    variable_summary <- summary(analysis_data[, requested_columns,
        drop = FALSE
    ])
    interaction_table <- NULL
    if (!is.null(resolved_interaction) && nzchar(resolved_interaction)) {
        interaction_table <- table(analysis_data[[resolved_interaction]],
            useNA = "ifany"
        )
    }

    log_lines <- c(
        "============================================================",
        paste("Loaded phenotype + methylation data from:", inputPheno),
        paste(
            "Merged modeling object:     ", methylation_prefix,
            "*"
        ), paste("Data dimensions:             ", paste(dim(analysis_data),
            collapse = " x "
        )), paste(
            "Phenotypes:                  ",
            paste(phenotype_list, collapse = ", ")
        ), paste(
            "Covariates:                  ",
            paste(covariate_list, collapse = ", ")
        ), paste(
            "Factor variables:            ",
            paste(factor_list, collapse = ", ")
        ), formatScalingMetadataLogDnaEpico(scaling$metadata),
        paste("CpG columns retained:        ", length(cpg_columns)),
        formatMethylationBoundariesLogDnaEpico(methylation_validation$boundaries),
        if (nrow(methylation_validation$issues) > 0L) {
            c("Methylation value issues:",
                previewLinesMinfiEwasWater(methylation_validation$issues))
        } else {
            "Methylation value issues:     none"
        }, "Missing summary:", paste(names(missing_counts), missing_counts,
            sep = ": ", collapse = "; "
        ), "Summary statistics:",
        previewLinesMinfiEwasWater(variable_summary)
    )
    if (!is.null(interaction_table)) {
        log_lines <- c(log_lines, paste(
            "Interaction table for",
            resolved_interaction, ":"
        ), previewLinesMinfiEwasWater(interaction_table))
    }
    log_lines <- c(log_lines,
        "============================================================")
    emitLogMinfiEwasWater(log_lines, verbose = verbose, log_path = log_path)

    structure(list(
        data = analysis_data, modelData = scaling$data,
        phenotypes = phenotype_list, covariates = covariate_list,
        factorVars = factor_list, scaleVars = scaling$scaleVars,
        scalingMetadata = scaling$metadata, cpgColumns = cpg_columns,
        cpgPrefix = cpgPrefix, cpgLimit = cpg_limit,
            methylationScale = methylation_scale,
        responseLabel = methylation_label,
            methylationObjectPrefix = methylation_prefix,
        internalResponseColumn = response_column, prsMap = prs_map,
        interactionTerm = resolved_interaction,
            requestedInteractionTerm = interactionTerm,
        methylationBoundaries = methylation_validation$boundaries,
        methylationIssues = methylation_validation$issues,
            invalidCpGs = methylation_validation$invalidCpGs,
        missingCounts = missing_counts, variableSummary = variable_summary,
        interactionTable = interaction_table
    ), class = "dnaEPICO_methylationGLM_data")
}
#' Plot phenotype and covariate distributions for one-timepoint GLM analyses
#'
#' @param preparedData Object returned by `prepareMethylationGLMData()`.
#' @param plotWidth Integer. TIFF width in pixels when plots are written to
#' disk.
#' @param plotHeight Integer. TIFF height in pixels when plots are written to
#'   disk.
#' @param plotDPI Integer. TIFF resolution in DPI when plots are written to
#'   disk.
#' @param outputDir Character or `NULL`. Directory used for TIFF files. When
#'   `NULL`, plots are returned in memory only.
#' @param display Logical. If `TRUE`, draw plots on the active graphics device.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return A list with class `'dnaEPICO_methylationGLM_distribution_plots'`
#'   containing the generated `ggplot2` objects and any saved TIFF file paths.
#'
#' @description
#' Create phenotype, factor-variable, and numeric-covariate distribution plots
#' from the object returned by `prepareMethylationGLMData()`.
#'
#' @examples
#' ex <- dnaEPICO:::exampleMethylationGLMStateDnaEpico()
#' distribution_plots <- plotMethylationGLMDistributions(
#'     preparedData = ex$preparedData,
#'     display = FALSE,
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#' names(distribution_plots)
#'
#' @export
plotMethylationGLMDistributions <- function(
    preparedData, plotWidth = 2000L,
    plotHeight = 1000L, plotDPI = 150L, outputDir = NULL, display = FALSE,
    verbose = FALSE, logs = FALSE, log_dir = NULL,
        log_file = "log_methylationGLM.txt"
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )
    analysis_data <- preparedData$data
    phenotype_plots <- list()
    factor_plots <- list()
    covariate_plots <- list()
    saved_files <- list(
        phenotypes = list(), factors = list(),
        covariates = list()
    )

    for (var in preparedData$phenotypes) {
        if (!var %in% colnames(analysis_data)) {
            next
        }

        plot_type <- if (is.numeric(analysis_data[[var]])) {
            "hist"
        } else {
            "bar"
        }
        plot_object <- createDistributionPlotMethylationGLM(
            values = analysis_data[[var]],
            variable = var, type = plot_type, fill = "steelblue"
        )
        phenotype_plots[[var]] <- plot_object

        file_path <- NULL
        if (!is.null(outputDir)) {
            file_path <- file.path(outputDir, paste0(if (identical(
                plot_type,
                "hist"
            )) {
                "hist_"
            } else {
                "bar_"
            }, var, ".tiff"))
        }

        runPlotMinfiEwasWater(
            draw_fun = function() drawPlotObjectMinfiEwasWater(plot_object),
            display = display, file = file_path, width = plotWidth,
            height = plotHeight, res = plotDPI
        )

        if (!is.null(file_path)) {
            saved_files$phenotypes[[var]] <- file_path
        }
    }

    factor_vars <- intersect(preparedData$factorVars, colnames(analysis_data))
    for (var in factor_vars) {
        plot_object <- createDistributionPlotMethylationGLM(
            values = analysis_data[[var]],
            variable = var, type = "bar", fill = "darkorange"
        )
        factor_plots[[var]] <- plot_object

        file_path <- NULL
        if (!is.null(outputDir)) {
            file_path <- file.path(outputDir, paste0(
                "bar_",
                var, ".tiff"
            ))
        }

        runPlotMinfiEwasWater(
            draw_fun = function() drawPlotObjectMinfiEwasWater(plot_object),
            display = display, file = file_path, width = plotWidth,
            height = plotHeight, res = plotDPI
        )

        if (!is.null(file_path)) {
            saved_files$factors[[var]] <- file_path
        }
    }

    numeric_covariates <- setdiff(preparedData$covariates,
        preparedData$factorVars)
    numeric_covariates <- intersect(numeric_covariates, colnames(analysis_data))
    for (var in numeric_covariates) {
        plot_object <- createDistributionPlotMethylationGLM(
            values = analysis_data[[var]],
            variable = var, type = "hist", fill = "darkgreen"
        )
        covariate_plots[[var]] <- plot_object

        file_path <- NULL
        if (!is.null(outputDir)) {
            file_path <- file.path(outputDir, paste0(
                "hist_",
                var, ".tiff"
            ))
        }

        runPlotMinfiEwasWater(
            draw_fun = function() drawPlotObjectMinfiEwasWater(plot_object),
            display = display, file = file_path, width = plotWidth,
            height = plotHeight, res = plotDPI
        )

        if (!is.null(file_path)) {
            saved_files$covariates[[var]] <- file_path
        }
    }

    emitLogMinfiEwasWater(
        c(
            "============================================================",
            paste("Phenotype distribution plots: ", length(phenotype_plots)),
            paste("Factor distribution plots:    ", length(factor_plots)),
            paste("Numeric covariate plots:      ", length(covariate_plots)),
            if (is.null(outputDir)) {
                "Distribution plots were returned in memory only."
            } else {
                paste("Distribution plots saved to:  ", outputDir)
            }, "============================================================"
        ),
        verbose = verbose, log_path = log_path
    )

    structure(list(
        phenotypes = phenotype_plots, factors = factor_plots,
        covariates = covariate_plots, files = saved_files
    ), class = "dnaEPICO_methylationGLM_distribution_plots")
}

#' Fit CpG-wise Gaussian GLMs for one-timepoint methylation analyses
#'
#' @param preparedData Object returned by `prepareMethylationGLMData()`.
#' @param nCores Integer. Maximum number of worker processes to use. Automatic
#'   fitting remains serial below the empirical glm2 crossover and caps workers
#'   by available CpGs and detected physical cores.
#' @param libPath Character vector or `NULL`. Optional library paths forwarded
#'   to worker processes.
#' @param glmLibs Character vector or comma-separated string of package names to
#'   check on worker processes. The default is `'glm2'`.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return A list with class `'dnaEPICO_methylationGLM_models'` containing
#'   fitted model lists, model formulas, and counts of failed CpG fits.
#'
#' @description
#' Fit one Gaussian GLM per CpG for each phenotype requested in the object
#' returned by `prepareMethylationGLMData()`.
#'
#' @examples
#' ex <- dnaEPICO:::exampleMethylationGLMStateDnaEpico()
#' model_results <- fitMethylationGLMModels(
#'     preparedData = ex$preparedData,
#'     nCores = 1,
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#' names(model_results$fits)
#'
#' @export
fitMethylationGLMModels <- function(
    preparedData, nCores = 1L,
    libPath = NULL, glmLibs = "glm2", verbose = FALSE, logs = FALSE,
    log_dir = NULL, log_file = "log_methylationGLM.txt"
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )

    if (is.null(libPath)) {
        libPath <- .libPaths()
    }

    glm_lib_list <- splitOptionMinfiEwasWater(glmLibs, sep = ",")
    if (length(glm_lib_list) == 0L) {
        glm_lib_list <- "glm2"
    }

    analysis_data <- preparedData$data
    model_data <- if (!is.null(preparedData$modelData)) {
        preparedData$modelData
    } else {
        analysis_data
    }
    cpg_columns <- preparedData$cpgColumns
    invalid_cpg_reasons <- invalidMethylationReasonsDnaEpico(preparedData)
    n_cores <- validatePositiveIntegerMethylationModels(
        nCores,
        "nCores"
    )
    fits <- list()
    summary_cache <- list()
    formulas <- stats::setNames(
        character(length(preparedData$phenotypes)),
        preparedData$phenotypes
    )
    failure_counts <- stats::setNames(
        integer(length(preparedData$phenotypes)),
        preparedData$phenotypes
    )
    failure_reasons <- list()
    parallel_plan <- resolveParallelPlanMethylationModels(
        engine = "glm2",
        nCores = n_cores, nCpGs = length(cpg_columns)
    )
    backend <- parallel_plan$backend
    worker_count <- parallel_plan$workerCount
    cpg_batches <- chunkCpGColumnsMethylationModels(
        cpgColumns = cpg_columns,
        nCores = worker_count, batchesPerCore = 8L
    )
    psock_cluster <- NULL
    if (identical(backend, "psock") && length(cpg_batches) >
        1L) {
        psock_cluster <- makePsockClusterMethylationModels(min(
            worker_count,
            length(cpg_batches)
        ))
        on.exit(
            {
                if (!is.null(psock_cluster)) {
                    try(parallel::stopCluster(psock_cluster), silent = TRUE)
                }
            },
            add = TRUE
        )
        parallel::clusterExport(psock_cluster,
            varlist = c(
                "analysis_data",
                "invalid_cpg_reasons", "libPath", "glm_lib_list",
                "validateWorkerPackagesMethylationModels",
                    "newMethylationFitErrorDnaEpico",
                "fitMethylationGLMBatch", "fitCpGModelMethylationGLM",
                "buildCoefficientTermMapMethylationModels",
                    "removeRandomInterceptMethylationModels",
                "summarizeCpGFitMethylationGLM",
                    "fitStatusValuesMethylationGLM",
                "findCoefficientRowsMethylationGLM", "escapeRegexMethylationGLM"
            ),
            envir = environment()
        )
        parallel::clusterEvalQ(psock_cluster,
            validateWorkerPackagesMethylationModels(
            libPath = libPath,
            packages = glm_lib_list
        ))
    }

    for (phenotype in preparedData$phenotypes) {
        prs_var <- character(0)
        if (phenotype %in% names(preparedData$prsMap)) {
            prs_var <- unname(preparedData$prsMap[[phenotype]])
        }
        covariates <- unique(c(preparedData$covariates, prs_var))
        model_vars <- unique(c(phenotype, covariates,
            preparedData$interactionTerm))
        model_vars <- model_vars[!is.na(model_vars) & nzchar(model_vars)]

        missing_vars <- setdiff(model_vars, colnames(model_data))
        if (length(missing_vars) > 0L) {
            stop("Model variables not found for phenotype ",
                phenotype, ": ", paste(missing_vars, collapse = ", "),
                call. = FALSE
            )
        }

        formula_text <- buildFormulaMethylationGLM(
            phenotype = phenotype,
            covariates = covariates,
                interactionTerm = preparedData$interactionTerm,
            responseVar = preparedData$internalResponseColumn
        )

        base_model_data <- model_data[, model_vars, drop = FALSE]
        factor_vars <- preparedData$factorVars
        for (var in intersect(factor_vars, colnames(base_model_data))) {
            base_model_data[[var]] <- as.factor(base_model_data[[var]])
        }
        validateFixedEffectDesignMethylationModels(
            formulaText = formula_text,
            data = base_model_data
        )
        batch_worker <- fitMethylationGLMBatch
        resolved_interaction <- preparedData$interactionTerm
        response_var <- preparedData$internalResponseColumn

        if (!identical(backend, "serial") && length(cpg_batches) >
            1L) {
            cluster_size <- min(worker_count, length(cpg_batches))

            if (identical(backend, "fork")) {
                batch_results <- parallel::mclapply(cpg_batches,
                    function(batch) {
                        validateWorkerPackagesMethylationModels(
                            libPath = libPath,
                            packages = glm_lib_list
                        )
                        batch_worker(
                            cpgBatch = batch, data = analysis_data,
                            modelData = base_model_data,
                                formulaText = formula_text,
                            phenotype = phenotype,
                                interactionTerm = resolved_interaction,
                            responseVar = response_var,
                                invalidCpgReasons = invalid_cpg_reasons
                        )
                    },
                    mc.cores = cluster_size, mc.preschedule = FALSE
                )
            } else {
                parallel::clusterExport(psock_cluster, varlist = c(
                    "base_model_data",
                    "formula_text", "phenotype", "resolved_interaction",
                    "response_var", "batch_worker"
                ), envir = environment())
                batch_results <- parallel::parLapplyLB(
                    psock_cluster,
                    cpg_batches, function(batch) {
                        batch_worker(
                            cpgBatch = batch, data = analysis_data,
                            modelData = base_model_data,
                                formulaText = formula_text,
                            phenotype = phenotype,
                                interactionTerm = resolved_interaction,
                            responseVar = response_var,
                                invalidCpgReasons = invalid_cpg_reasons
                        )
                    }
                )
            }
        } else {
            validateWorkerPackagesMethylationModels(
                libPath = libPath,
                packages = glm_lib_list
            )
            batch_results <- lapply(cpg_batches, function(batch) {
                batch_worker(
                    cpgBatch = batch, data = analysis_data,
                    modelData = base_model_data, formulaText = formula_text,
                    phenotype = phenotype,
                        interactionTerm = resolved_interaction,
                    responseVar = response_var,
                        invalidCpgReasons = invalid_cpg_reasons
                )
            })
        }

        combined_results <- combineFitBatchResultsMethylationModels(
            batchResults = batch_results,
            cpgColumns = cpg_columns
        )
        fit_list <- combined_results$fits
        phenotype_summary_cache <- filterSummaryByPvalueMethylationGLM(
            summaryDf = combined_results$summaries,
            pValueFilter = NA_real_, includeResidualSD = TRUE
        )
        failures <- vapply(fit_list, function(x) {
            inherits(
                x,
                "dnaEPICO_methylationGLM_fit_error"
            )
        }, logical(1))
        error_counts <- summarizeFitErrorsMethylationModels(
            fitList = fit_list,
            errorClass = "dnaEPICO_methylationGLM_fit_error"
        )
        fits[[phenotype]] <- fit_list
        summary_cache[[phenotype]] <- phenotype_summary_cache
        formulas[[phenotype]] <- formula_text
        failure_counts[[phenotype]] <- sum(failures)
        failure_reasons[[phenotype]] <- error_counts

        emitLogMinfiEwasWater(
            c(
                "============================================================",
                paste("Fitted phenotype:            ", phenotype),
                paste("Formula:                     ", formula_text),
                paste("CpGs attempted:              ", length(cpg_columns)),
                paste("Failed CpG fits:             ",
                    failure_counts[[phenotype]]),
                paste("Top fit errors:               ",
                    formatFitErrorsMethylationModels(error_counts)),
                paste("Parallel backend:            ", backend),
                paste("Effective workers:           ", worker_count),
                paste("Parallel crossover CpGs:     ",
                    parallel_plan$crossoverCpGs),
                paste("Parallel selection:          ", parallel_plan$reason),
                paste("Fit batches:                 ", length(cpg_batches)),
                paste("Fit batch size:              ",
                    if (length(cpg_batches) ==
                    0L) {
                    0L
                } else {
                    max(vapply(cpg_batches, length, integer(1)))
                }),
                paste("Fit-time summary rows cached:",
                    nrow(phenotype_summary_cache)),
                "============================================================"
            ),
            verbose = verbose, log_path = log_path
        )

        if (length(fit_list) > 0L && all(failures)) {
            warning("All CpG GLM fits failed for phenotype '",
                phenotype, "'. Top failure reasons: ",
                    formatFitErrorsMethylationModels(error_counts),
                ". The failure inventory was retained and the analysis continued.",
                call. = FALSE
            )
        }
    }

    fit_failures <- collectFitFailuresMethylationModels(
        fits = fits,
        errorClass = "dnaEPICO_methylationGLM_fit_error"
    )
    fit_diagnostics <- collectFitDiagnosticsMethylationGLM(fits)

    if (!is.null(psock_cluster)) {
        parallel::stopCluster(psock_cluster)
        psock_cluster <- NULL
    }

    structure(
        list(
            fits = fits, summaryCache = summary_cache,
            formulas = formulas, phenotypes = names(fits),
                failureCounts = failure_counts,
            failureReasons = failure_reasons, fitFailures = fit_failures,
            fitDiagnostics = fit_diagnostics,
                methylationBoundaries = preparedData$methylationBoundaries,
            methylationIssues = preparedData$methylationIssues,
                invalidCpGs = preparedData$invalidCpGs,
            settings = list(
                nCores = n_cores, parallelBackend = backend,
                workerCount = worker_count,
                    resourceWorkerCap = parallel_plan$resourceWorkerCap,
                parallelCrossoverCpGs = parallel_plan$crossoverCpGs,
                parallelSelectionReason = parallel_plan$reason,
                    clusterReusedAcrossPhenotypes = identical(
                    backend,
                    "psock"
                ) && length(preparedData$phenotypes) >
                    1L, fitBatchCount = length(cpg_batches), libPath = libPath,
                glmLibs = glm_lib_list,
                    methylationScale = preparedData$methylationScale,
                methylationObjectPrefix = preparedData$methylationObjectPrefix,
                responseLabel = preparedData$responseLabel,
                    internalResponseColumn = preparedData$internalResponseColumn,
                interactionTerm = preparedData$interactionTerm,
                    phenotypes = preparedData$phenotypes,
                covariates = preparedData$covariates,
                    factorVars = preparedData$factorVars,
                factorLevels = lapply(stats::setNames(
                    preparedData$factorVars,
                    preparedData$factorVars
                ), function(variable) levels(preparedData$data[[variable]])),
                scaleVars = preparedData$scaleVars,
                    scalingMetadata = preparedData$scalingMetadata,
                sampleCount = nrow(preparedData$data)
            ), responseLabel = preparedData$responseLabel
        ),
        class = "dnaEPICO_methylationGLM_models"
    )
}

#' Summarize CpG-wise Gaussian GLM fits for one-timepoint analyses
#'
#' @param modelResults Object returned by `fitMethylationGLMModels()`.
#' @param preparedData Object returned by `prepareMethylationGLMData()`.
#' @param summaryResidualSD Logical. If `TRUE`, add residual standard deviations
#'   to each CpG summary row.
#' @param summaryPval Numeric or `NA`. Optional p-value filter applied to the
#'   returned summary tables. `NA` keeps all rows.
#' @param excludeNonConverged Logical. If `TRUE`, retain non-converged CpG rows
#'   and their reasons but set their inferential p-values to `NA`.
#' @param nCores Integer. Number of worker processes to use while extracting
#'   summary rows.
#' @param libPath Character vector or `NULL`. Optional library paths forwarded
#'   to worker processes.
#' @param glmLibs Character vector or comma-separated string of package names to
#'   check on worker processes. The default is `'glm2'`.
#' @param chunkSize Integer or `NULL`. Number of CpGs to process per parallel
#'   chunk. `NULL` chooses a value automatically.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return A list with class `'dnaEPICO_methylationGLM_summaries'`
#'   containing the optionally filtered summary tables in `summaries` and the
#'   complete CpG-level tables in `diagnosticSummaries`. Diagnostics,
#'   annotation, and report output use the complete tables so `summaryPval`
#'   does not remove CpGs from those outputs.
#'
#' @description
#' Extract phenotype-specific CpG coefficient tables from the fitted model
#' object returned by `fitMethylationGLMModels()`.
#'
#' @examples
#' ex <- dnaEPICO:::exampleMethylationGLMStateDnaEpico()
#' summary_results <- summarizeMethylationGLMModels(
#'     modelResults = ex$modelResults,
#'     preparedData = ex$preparedData,
#'     summaryResidualSD = TRUE,
#'     summaryPval = NA,
#'     nCores = 1,
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#' names(summary_results$summaries)
#'
#' @export
summarizeMethylationGLMModels <- function(
    modelResults, preparedData,
    summaryResidualSD = TRUE, summaryPval = NA, excludeNonConverged = FALSE,
    nCores = 1L, libPath = NULL, glmLibs = "glm2", chunkSize = NULL,
    verbose = FALSE, logs = FALSE, log_dir = NULL,
        log_file = "log_methylationGLM.txt"
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )

    if (is.null(libPath)) {
        libPath <- .libPaths()
    }

    glm_lib_list <- splitOptionMinfiEwasWater(glmLibs, sep = ",")
    if (length(glm_lib_list) == 0L) {
        glm_lib_list <- "glm2"
    }

    p_value_filter <- normalizeOptionalNumericMethylationGLM(summaryPval)
    p_value_filter <- validateProbabilityDnaEpico(p_value_filter,
        "summaryPval",
        allowNA = TRUE
    )
    chunk_size <- normalizeChunkSizeMethylationGLM(chunkSize)
    n_cores <- validatePositiveIntegerMethylationModels(
        nCores,
        "nCores"
    )
    summaries <- list()
    diagnostic_summaries <- list()

    for (phenotype in names(modelResults$fits)) {
        if (!is.null(modelResults$summaryCache) &&
            !is.null(modelResults$summaryCache[[phenotype]])) {
            diagnostic_df <- filterSummaryByPvalueMethylationGLM(
                summaryDf = modelResults$summaryCache[[phenotype]],
                pValueFilter = NA_real_,
                    includeResidualSD = isTRUE(summaryResidualSD)
            )
            diagnostic_df <- applyFitQualityExclusionsMethylationGLM(
                summaryDf = diagnostic_df,
                excludeNonConverged = excludeNonConverged
            )
            summary_df <- filterSummaryByPvalueMethylationGLM(
                summaryDf = diagnostic_df,
                pValueFilter = p_value_filter,
                    includeResidualSD = isTRUE(summaryResidualSD)
            )
            diagnostic_summaries[[phenotype]] <- diagnostic_df
            summaries[[phenotype]] <- summary_df

            emitLogMinfiEwasWater(
                c(
                    "============================================================",
                    paste("Summarized phenotype:        ", phenotype),
                    paste("CpG summary rows returned:   ", nrow(summary_df)),
                    "Summary source:              fit-time cache",
                    if (is.na(p_value_filter)) {
                        "P-value filter:              none"
                    } else {
                        paste("P-value filter:              ", p_value_filter)
                    }, "============================================================"
                ),
                verbose = verbose, log_path = log_path
            )
            next
        }

        fit_list <- modelResults$fits[[phenotype]]
        cpg_names <- names(fit_list)

        local_chunk_size <- chunk_size
        if (is.null(local_chunk_size)) {
            local_chunk_size <- max(10L, floor(length(cpg_names) / max(n_cores *
                4L, 1L)))
        }
        local_chunk_size <- max(1L, as.integer(local_chunk_size))
        cpg_chunks <- split(cpg_names,
            ceiling(seq_along(cpg_names) / local_chunk_size))

        summary_worker <- summarizeCpGFitMethylationGLM
        resolved_interaction <- preparedData$interactionTerm
        include_residual_sd <- isTRUE(summaryResidualSD)

        if (n_cores > 1L && length(cpg_chunks) > 1L) {
            cluster_size <- min(n_cores, length(cpg_chunks))
            cl <- parallel::makeCluster(cluster_size)
            on.exit(parallel::stopCluster(cl), add = TRUE)

            parallel::clusterExport(cl,
                varlist = c(
                    "fit_list",
                    "phenotype", "resolved_interaction", "include_residual_sd",
                    "summary_worker", "libPath", "glm_lib_list"
                ),
                envir = environment()
            )

            parallel::clusterEvalQ(cl, {
                if (!is.null(libPath)) {
                    .libPaths(unique(c(libPath, .libPaths())))
                }

                for (pkg in glm_lib_list) {
                    if (!requireNamespace(pkg, quietly = TRUE)) {
                        stop("Failed to load package: ", pkg, call. = FALSE)
                    }
                }

                NULL
            })

            result_chunks <- parallel::parLapplyLB(
                cl, cpg_chunks,
                function(chunk) {
                    rows <- lapply(chunk, function(cpg) {
                        summary_worker(
                            cpg = cpg, modelObj = fit_list[[cpg]],
                            variable = phenotype,
                                interactionTerm = resolved_interaction,
                            includeResidualSD = include_residual_sd
                        )
                    })
                    rows <- Filter(Negate(is.null), rows)
                    if (length(rows) == 0L) {
                        return(NULL)
                    }
                    do.call(rbind, rows)
                }
            )
            parallel::stopCluster(cl)
            on.exit(NULL, add = FALSE)
        } else {
            result_chunks <- lapply(cpg_chunks, function(chunk) {
                rows <- lapply(chunk, function(cpg) {
                    summary_worker(
                        cpg = cpg, modelObj = fit_list[[cpg]],
                        variable = phenotype,
                            interactionTerm = resolved_interaction,
                        includeResidualSD = include_residual_sd
                    )
                })
                rows <- Filter(Negate(is.null), rows)
                if (length(rows) == 0L) {
                    return(NULL)
                }
                do.call(rbind, rows)
            })
        }

        result_chunks <- Filter(Negate(is.null), result_chunks)
        if (length(result_chunks) == 0L) {
            summary_df <- data.frame()
        } else {
            summary_df <- do.call(rbind, result_chunks)
            summary_df <- summary_df[, c(
                "CpG", "Coefficient",
                "Estimate", "Std. Error", "t value", "Pr(>|t|)",
                if (isTRUE(summaryResidualSD)) "ResidualSD",
                "Fit.Status", "Singular.Fit", "Converged",
                    "Convergence.Message",
                "Fit.Warning", "Inference.Included", "Exclusion.Reason"
            ),
            drop = FALSE
            ]
            rownames(summary_df) <- NULL
        }

        summary_df <- applyFitQualityExclusionsMethylationGLM(
            summaryDf = summary_df,
            excludeNonConverged = excludeNonConverged
        )
        diagnostic_summaries[[phenotype]] <- summary_df
        if (nrow(summary_df) > 0L && !is.na(p_value_filter)) {
            keep <- is.finite(summary_df[["Pr(>|t|)"]]) & summary_df[["Pr(>|t|)"]] <
                p_value_filter
            summary_df <- summary_df[keep, , drop = FALSE]
            rownames(summary_df) <- NULL
        }

        summaries[[phenotype]] <- summary_df

        emitLogMinfiEwasWater(
            c(
                "============================================================",
                paste("Summarized phenotype:        ", phenotype),
                paste("CpG summary rows returned:   ", nrow(summary_df)),
                paste("Summary chunk size:          ", local_chunk_size),
                if (is.na(p_value_filter)) {
                    "P-value filter:              none"
                } else {
                    paste("P-value filter:              ", p_value_filter)
                }, "============================================================"
            ),
            verbose = verbose, log_path = log_path
        )
    }

    structure(list(
        summaries = summaries, diagnosticSummaries = diagnostic_summaries,
        phenotypes = names(summaries), fitFailures = modelResults$fitFailures,
        fitDiagnostics = resolveFitDiagnosticsMethylationGLM(modelSummaries = list(fitDiagnostics = if (!is.null(modelResults$fitDiagnostics)) {
            modelResults$fitDiagnostics
        } else {
            collectFitDiagnosticsMethylationGLM(modelResults$fits)
        }, fitFailures = modelResults$fitFailures),
            summaryList = diagnostic_summaries),
        settings = list(
            summaryResidualSD = isTRUE(summaryResidualSD),
            summaryPval = p_value_filter,
                excludeNonConverged = isTRUE(excludeNonConverged),
            chunkSize = chunk_size
        )
    ), class = "dnaEPICO_methylationGLM_summaries")
}

#' Collect significant CpG coefficient tables from fitted one-timepoint GLMs
#'
#' @param modelResults Object returned by `fitMethylationGLMModels()`.
#' @param pvalThreshold Numeric. Threshold applied to phenotype main-effect or
#'   interaction p-values.
#' @param interactionTerm Character or `NULL`. Optional interaction term.
#' @param excludeNonConverged Logical. If `TRUE`, do not collect CpGs whose GLM
#'   did not converge; their rows and reasons remain in diagnostic and annotated
#'   output.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return A list with class `'dnaEPICO_methylationGLM_significant_cpgs'`.
#'
#' @description
#' Collect the raw coefficient tables for CpGs whose phenotype main effect or
#' interaction p-value passes the requested threshold.
#'
#' @examples
#' ex <- dnaEPICO:::exampleMethylationGLMStateDnaEpico()
#' significant_cpgs <- collectSignificantCpGsMethylationGLM(
#'     modelResults = ex$modelResults,
#'     pvalThreshold = 1,
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#' names(significant_cpgs)
#'
#' @export
collectSignificantCpGsMethylationGLM <- function(
    modelResults,
    pvalThreshold = 0.05, interactionTerm = NULL, excludeNonConverged = FALSE,
    verbose = FALSE, logs = FALSE, log_dir = NULL,
        log_file = "log_methylationGLM.txt"
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )
    threshold <- validateProbabilityDnaEpico(pvalThreshold, "pvalThreshold")
    retained <- list()

    for (phenotype in names(modelResults$fits)) {
        fit_list <- modelResults$fits[[phenotype]]
        phenotype_hits <- list()
        if (!is.null(modelResults$summaryCache) &&
            !is.null(modelResults$summaryCache[[phenotype]]) &&
            optionalTermMatchesMethylationModels(
                requested = interactionTerm,
                cached = modelResults$settings$interactionTerm
            )) {
            cached_summary <- modelResults$summaryCache[[phenotype]]
            if (nrow(cached_summary) > 0L && !is.na(threshold)) {
                hit_cpgs <-
                    unique(cached_summary$CpG[cached_summary[["Pr(>|t|)"]] <
                    threshold])
                hit_cpgs <- hit_cpgs[!is.na(hit_cpgs) & hit_cpgs %in%
                    names(fit_list)]
                for (cpg in hit_cpgs) {
                    model_obj <- fit_list[[cpg]]
                    if (is.null(model_obj) || inherits(
                        model_obj,
                        "dnaEPICO_methylationGLM_fit_error"
                    )) {
                        next
                    }
                    if (isTRUE(excludeNonConverged) && identical(
                        fitStatusValuesMethylationGLM(model_obj)$converged,
                        FALSE
                    )) {
                        next
                    }
                    if (!is.null(model_obj$coef)) {
                        phenotype_hits[[cpg]] <- as.data.frame(model_obj$coef)
                    }
                }
            }

            retained[[phenotype]] <- phenotype_hits
            next
        }

        for (cpg in names(fit_list)) {
            model_obj <- fit_list[[cpg]]
            if (is.null(model_obj) || inherits(model_obj,
                "dnaEPICO_methylationGLM_fit_error")) {
                next
            }
            if (isTRUE(excludeNonConverged) && identical(
                fitStatusValuesMethylationGLM(model_obj)$converged,
                FALSE
            )) {
                next
            }

            coef_table <- model_obj$coef
            if (is.null(coef_table)) {
                next
            }

            matched_rows <- findCoefficientRowsMethylationGLM(
                coefNames = rownames(coef_table),
                variable = phenotype, interactionTerm = interactionTerm,
                coefficientTerms = model_obj$coefficientTerms
            )
            if (length(matched_rows) == 0L) {
                next
            }

            matched_pvals <- coef_table[matched_rows, "Pr(>|t|)",
                drop = TRUE
            ]
            if (any(matched_pvals < threshold, na.rm = TRUE)) {
                phenotype_hits[[cpg]] <- as.data.frame(coef_table)
            }
        }

        retained[[phenotype]] <- phenotype_hits
    }

    hit_counts <- vapply(retained, length, integer(1))
    emitLogMinfiEwasWater(
        c(
            "============================================================",
            paste(
                "Significant CpGs retained at p <", threshold,
                ":"
            ), paste(names(hit_counts), hit_counts,
                sep = ": ",
                collapse = "; "
            ), "============================================================"
        ),
        verbose = verbose, log_path = log_path
    )

    structure(retained, class = "dnaEPICO_methylationGLM_significant_cpgs")
}
#' Plot diagnostic summaries for one-timepoint methylation GLMs
#'
#' @param modelSummaries Object returned by `summarizeMethylationGLMModels()`.
#' @param preparedData Object returned by `prepareMethylationGLMData()`.
#' @param fdrThreshold Numeric. False-discovery-rate threshold used to highlight
#'   CpGs in the diagnostic plots.
#' @param padjmethod Character. P-value adjustment method passed to
#'   `stats::p.adjust()`.
#' @param outputDir Character or `NULL`. Directory used for TIFF files. When
#'   `NULL`, plots are returned in memory only.
#' @param plotWidth Integer. TIFF width in pixels when plots are written to
#' disk.
#' @param plotHeight Integer. TIFF height in pixels when plots are written to
#'   disk.
#' @param plotDPI Integer. TIFF resolution in DPI when plots are written to
#'   disk.
#' @param display Logical. If `TRUE`, draw plots on the active graphics device.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return A list with class `'dnaEPICO_methylationGLM_diagnostic_plots'`
#'   containing the generated `ggplot2` objects, genomic inflation factors, and
#'   any saved TIFF file paths.
#'
#' @description
#' Create Q-Q and residual-diagnostic plots from the CpG summary tables returned
#' by `summarizeMethylationGLMModels()`.
#'
#' @examples
#' ex <- dnaEPICO:::exampleMethylationGLMStateDnaEpico()
#' diagnostic_plots <- plotMethylationGLMDiagnostics(
#'     modelSummaries = ex$modelSummaries,
#'     preparedData = ex$preparedData,
#'     display = FALSE,
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#' names(diagnostic_plots$plots)
#'
#' @export
plotMethylationGLMDiagnostics <- function(
    modelSummaries, preparedData,
    fdrThreshold = 0.05, padjmethod = "fdr", outputDir = NULL,
    plotWidth = 2000L, plotHeight = 1000L, plotDPI = 150L, display = FALSE,
    verbose = FALSE, logs = FALSE, log_dir = NULL,
        log_file = "log_methylationGLM.txt"
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )
    fdrThreshold <- validateProbabilityDnaEpico(
        fdrThreshold,
        "fdrThreshold"
    )
    padjmethod <- validatePAdjustmentMethodMethylationModels(padjmethod)
    summary_list <- if (!is.null(modelSummaries$diagnosticSummaries)) {
        modelSummaries$diagnosticSummaries
    } else {
        modelSummaries$summaries
    }
    diagnostic_mean <- diagnosticMeanMethylationModels(preparedData)
    plot_list <- list()
    inflation_factors <- list()
    saved_files <- list()

    for (phenotype in names(summary_list)) {
        summary_df <- summary_list[[phenotype]]
        if (is.null(summary_df) || nrow(summary_df) == 0L) {
            next
        }

        summary_df$FDR <- adjustPvaluesByTermMethylationModels(
            pValues = summary_df[["Pr(>|t|)"]],
            terms = summary_df$Coefficient, method = padjmethod
        )
        coefficient_terms <- unique(as.character(summary_df$Coefficient))
        coefficient_terms <- coefficient_terms[!is.na(coefficient_terms)]
        multiple_terms <- length(coefficient_terms) > 1L
        phenotype_plots <- phenotype_files <- list()
        phenotype_inflation <- numeric(0)

        for (term_index in seq_along(coefficient_terms)) {
            term <- coefficient_terms[[term_index]]
            term_diagnostics <- buildMethylationTermDiagnosticsDnaEpico(
                summaryData = summary_df,
                phenotype = phenotype, term = term, termColumn = "Coefficient",
                pValueColumn = "Pr(>|t|)", yColumn = "ResidualSD",
                yLabel = "Residual SD", diagnosticMean = diagnostic_mean,
                fdrThreshold = fdrThreshold
            )
            if (is.null(term_diagnostics)) {
                next
            }

            file_key <- phenotype
            if (multiple_terms) {
                file_key <- paste0(phenotype, "_", sprintf(
                    "%02d",
                    term_index
                ), "_", sanitizeDiagnosticTermDnaEpico(term))
            }
            term_files <- list(
                qqplot = NULL, residualSD = NULL,
                residualSignificance = NULL
            )
            if (!is.null(outputDir)) {
                term_files <- list(qqplot = file.path(
                    outputDir,
                    paste0("qqplot_", file_key, ".tiff")
                ), residualSD = if (!is.null(term_diagnostics$plots$residualSD)) {
                    file.path(outputDir, paste0(
                        "residualSD_",
                        file_key, ".tiff"
                    ))
                } else {
                    NULL
                }, residualSignificance = if (!is.null(term_diagnostics$plots$residualSignificance)) {
                    file.path(outputDir, paste0(
                        "residualSignificance_",
                        file_key, ".tiff"
                    ))
                } else {
                    NULL
                })
            }

            for (plot_name in names(term_diagnostics$plots)) {
                plot_object <- term_diagnostics$plots[[plot_name]]
                if (is.null(plot_object)) {
                    next
                }
                runPlotMinfiEwasWater(
                    draw_fun = function() drawPlotObjectMinfiEwasWater(plot_object),
                    display = display, file = term_files[[plot_name]],
                    width = plotWidth, height = plotHeight, res = plotDPI
                )
            }

            phenotype_inflation[[term]] <- term_diagnostics$lambda
            if (multiple_terms) {
                phenotype_plots[[term]] <- term_diagnostics$plots
                phenotype_files[[term]] <- term_files
            } else {
                phenotype_plots <- term_diagnostics$plots
                phenotype_files <- term_files
            }
        }

        if (length(phenotype_inflation) > 0L) {
            inflation_factors[[phenotype]] <- if (multiple_terms) {
                phenotype_inflation
            } else {
                unname(phenotype_inflation[[1L]])
            }
            plot_list[[phenotype]] <- phenotype_plots
            saved_files[[phenotype]] <- phenotype_files
        }
    }

    emitLogMinfiEwasWater(
        c(
            "============================================================",
            paste("Diagnostic plots generated for phenotypes:",
                length(plot_list)),
            if (is.null(outputDir)) {
                "Diagnostic plots were returned in memory only."
            } else {
                paste("Diagnostic plots saved to:    ", outputDir)
            }, "============================================================"
        ),
        verbose = verbose, log_path = log_path
    )

    structure(list(
        plots = plot_list, inflationFactors = inflation_factors,
        files = saved_files
    ), class = "dnaEPICO_methylationGLM_diagnostic_plots")
}

resolveFitDiagnosticsMethylationGLM <- function(
    modelSummaries,
    summaryList
) {
    required_columns <- c(
        "Phenotype", "CpG", "Fit.Status", "Singular.Fit",
        "Converged", "Convergence.Message", "Fit.Warning", "Inference.Included",
        "Exclusion.Reason"
    )
    rows <- list()
    row_index <- 1L
    for (phenotype in names(summaryList)) {
        summary_df <- summaryList[[phenotype]]
        if (is.null(summary_df) || nrow(summary_df) == 0L ||
            !("CpG" %in% names(summary_df))) {
            next
        }
        value_or_default <- function(column, default) {
            if (column %in% names(summary_df)) {
                summary_df[[column]]
            } else {
                rep(default, nrow(summary_df))
            }
        }
        rows[[row_index]] <- unique(data.frame(
            Phenotype = phenotype,
            CpG = as.character(summary_df$CpG),
                Fit.Status = as.character(value_or_default(
                "Fit.Status",
                "fitted"
            )), Singular.Fit = as.logical(value_or_default(
                "Singular.Fit",
                NA
            )), Converged = as.logical(value_or_default(
                "Converged",
                TRUE
            )), Convergence.Message = as.character(value_or_default(
                "Convergence.Message",
                NA_character_
            )), Fit.Warning = as.character(value_or_default(
                "Fit.Warning",
                NA_character_
            )), Inference.Included = as.logical(value_or_default(
                "Inference.Included",
                TRUE
            )), Exclusion.Reason = as.character(value_or_default(
                "Exclusion.Reason",
                NA_character_
            )), stringsAsFactors = FALSE, check.names = FALSE
        ))
        row_index <- row_index + 1L
    }

    diagnostics <- modelSummaries$fitDiagnostics
    if (is.data.frame(diagnostics) && all(required_columns %in%
        names(diagnostics))) {
        rows[[row_index]] <- diagnostics[, required_columns,
            drop = FALSE
        ]
    }
    if (length(rows) == 0L) {
        empty <- lapply(required_columns, function(x) character(0))
        names(empty) <- required_columns
        empty$Singular.Fit <- logical(0)
        empty$Converged <- logical(0)
        empty$Inference.Included <- logical(0)
        return(as.data.frame(empty,
            stringsAsFactors = FALSE,
            check.names = FALSE
        ))
    }

    out <- do.call(rbind, rows)
    key <- paste(out$Phenotype, out$CpG, sep = "\r")
    out <- out[!duplicated(key), , drop = FALSE]
    rownames(out) <- NULL
    out
}

buildAnnotationFitDiagnosticsMethylationGLM <- function(fitDiagnostics) {
    if (!is.data.frame(fitDiagnostics) || nrow(fitDiagnostics) ==
        0L) {
        return(list(overall = data.frame(), byPhenotype = data.frame()))
    }

    aggregate_logical <- function(values, falseDominates = TRUE) {
        values <- as.logical(values)
        values <- values[!is.na(values)]
        if (length(values) == 0L) {
            return(NA)
        }
        if (isTRUE(falseDominates) && any(!values)) {
            return(FALSE)
        }
        if (!isTRUE(falseDominates) && any(values)) {
            return(TRUE)
        }
        if (isTRUE(falseDominates)) {
            TRUE
        } else {
            FALSE
        }
    }
    diagnostics_by_cpg <- split(fitDiagnostics, fitDiagnostics$CpG,
        drop = TRUE
    )
    overall_rows <- lapply(names(diagnostics_by_cpg), function(cpg) {
        diagnostic <- diagnostics_by_cpg[[cpg]]
        statuses <- as.character(diagnostic$Fit.Status)
        failed <- statuses %in% c("failed", "invalid", "not_converged")
        status <- if (any(failed)) {
            if (!all(failed)) {
                "partial"
            } else if (all(statuses == "invalid")) {
                "invalid"
            } else if (all(statuses == "not_converged")) {
                "not_converged"
            } else {
                "failed"
            }
        } else if (any(statuses == "fitted_not_converged")) {
            "fitted_not_converged"
        } else if (any(statuses == "fitted_with_warning")) {
            "fitted_with_warning"
        } else {
            "fitted"
        }
        reasons <-
            unique(paste0(diagnostic$Phenotype[!is.na(diagnostic$Exclusion.Reason) &
            nzchar(diagnostic$Exclusion.Reason)], ": ",
                diagnostic$Exclusion.Reason[!is.na(diagnostic$Exclusion.Reason) &
            nzchar(diagnostic$Exclusion.Reason)]))
        data.frame(
            CpG = cpg, Diagnostic.Fit.Status = status,
            Singular.Fit = aggregate_logical(diagnostic$Singular.Fit,
                falseDominates = FALSE
            ), Converged = aggregate_logical(diagnostic$Converged),
            Inference.Included = aggregate_logical(diagnostic$Inference.Included),
            Fit.Quality.Reason = if (length(reasons)) {
                paste(reasons, collapse = " | ")
            } else {
                NA_character_
            }, stringsAsFactors = FALSE, check.names = FALSE
        )
    })
    overall <- do.call(rbind, overall_rows)
    rownames(overall) <- NULL

    diagnostic_columns <- setdiff(names(fitDiagnostics), c(
        "Phenotype",
        "CpG"
    ))
    phenotype_tables <- lapply(
        unique(as.character(fitDiagnostics$Phenotype)),
        function(phenotype) {
            diagnostic <- fitDiagnostics[fitDiagnostics$Phenotype ==
                phenotype, c("CpG", diagnostic_columns), drop = FALSE]
            diagnostic <- unique(diagnostic)
            colnames(diagnostic)[-1L] <- paste0(
                phenotype, "_",
                diagnostic_columns
            )
            diagnostic
        }
    )
    by_phenotype <- if (length(phenotype_tables) == 1L) {
        phenotype_tables[[1L]]
    } else {
        Reduce(
            function(x, y) merge(x, y, by = "CpG", all = TRUE),
            phenotype_tables
        )
    }
    list(overall = overall, byPhenotype = by_phenotype)
}

#' Annotate one-timepoint GLM summary tables with array annotation metadata
#'
#' @param modelSummaries Object returned by `summarizeMethylationGLMModels()`
#'   or a named list of CpG summary data frames.
#' @param annotationObject Character package/object name, annotation data frame,
#'   or annotation object understood by `minfi::getAnnotation()`.
#' @param annotationCols Character vector or comma-separated string of
#' annotation
#'   columns to append.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return A list with class `'dnaEPICO_methylationGLM_annotation'`
#'   containing the annotated summary table and any requested annotation columns
#'   that were unavailable in the chosen annotation object.
#'
#' @description
#' Merge phenotype-specific CpG summary tables with probe annotation metadata
#' and
#' return a single annotated result table.
#'
#' @examples
#' ex <- dnaEPICO:::exampleMethylationGLMStateDnaEpico()
#' annotation_data <- annotateMethylationGLMSummaries(
#'     modelSummaries = ex$modelSummaries,
#'     annotationObject = ex$annotationData,
#'     annotationCols = "Name,chr,pos",
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#' names(annotation_data)
#'
#' @export
annotateMethylationGLMSummaries <- function(
    modelSummaries, annotationObject,
    annotationCols = c(
        "Name", "chr", "pos", "UCSC_RefGene_Group",
        "UCSC_RefGene_Name", "Relation_to_Island", "GencodeV41_Group"
    ),
    verbose = FALSE, logs = FALSE, log_dir = NULL,
        log_file = "log_methylationGLM.txt"
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )
    fit_failures <- modelSummaries$fitFailures
    summary_phenotypes <- modelSummaries$phenotypes
    summary_list <- modelSummaries
    if (!is.null(modelSummaries$diagnosticSummaries)) {
        summary_list <- modelSummaries$diagnosticSummaries
    } else if (!is.null(modelSummaries$summaries)) {
        summary_list <- modelSummaries$summaries
    }
    if (is.null(summary_phenotypes)) {
        summary_phenotypes <- names(summary_list)
    }
    fit_diagnostics <- resolveFitDiagnosticsMethylationGLM(
        modelSummaries = modelSummaries,
        summaryList = summary_list
    )
    annotation_diagnostics <-
        buildAnnotationFitDiagnosticsMethylationGLM(fitDiagnostics = fit_diagnostics)

    annotation_cols <- splitOptionMinfiEwasWater(annotationCols,
        sep = ","
    )
    annotation_df <- coerceAnnotationDataMethylationGLM(annotationObject)
    coefficient_occurrences <- unlist(lapply(summary_list, function(x) {
        if (is.null(x) || nrow(x) == 0L || !("Coefficient" %in%
            colnames(x))) {
            return(character(0))
        }
        unique(as.character(x$Coefficient))
    }), use.names = FALSE)
    duplicated_coefficients <-
        unique(coefficient_occurrences[duplicated(coefficient_occurrences)])

    merged_summary_list <- lapply(names(summary_list), function(phenotype) {
        summary_df <- summary_list[[phenotype]]
        if (is.null(summary_df) || nrow(summary_df) == 0L) {
            return(NULL)
        }

        coefficient_names <- unique(summary_df$Coefficient)
        coefficient_tables <- lapply(coefficient_names,
            function(coefficient_name) {
            sub_df <- summary_df[summary_df$Coefficient == coefficient_name,
                c("CpG", "Pr(>|t|)"),
                drop = FALSE
            ]
            clean_name <- gsub("`", "", coefficient_name, fixed = TRUE)
            colnames(sub_df)[2] <- if (coefficient_name %in%
                duplicated_coefficients) {
                paste0(phenotype, "_", clean_name, "_P.Value")
            } else {
                paste0(clean_name, "P.Value")
            }
            sub_df
        })

        if (length(coefficient_tables) == 0L) {
            return(NULL)
        }

        Reduce(
            function(x, y) merge(x, y, by = "CpG", all = TRUE),
            coefficient_tables
        )
    })
    merged_summary_list <- Filter(Negate(is.null), merged_summary_list)

    if (length(merged_summary_list) == 0L) {
        merged_summary <- data.frame(CpG = character(0))
    } else if (length(merged_summary_list) == 1L) {
        merged_summary <- merged_summary_list[[1L]]
    } else {
        merged_summary <- Reduce(function(x, y) {
            merge(x, y,
                by = "CpG",
                all = TRUE
            )
        }, merged_summary_list)
    }
    if (nrow(annotation_diagnostics$byPhenotype) > 0L) {
        glm_diagnostics <- annotation_diagnostics$byPhenotype
        glm_diagnostics <- glm_diagnostics[, !grepl(
            "_Singular\\.Fit$",
            colnames(glm_diagnostics)
        ), drop = FALSE]
        merged_summary <- merge(merged_summary, glm_diagnostics,
            by = "CpG", all = TRUE
        )
    }

    available_annotation_cols <- intersect(annotation_cols,
        colnames(annotation_df))
    missing_annotation_cols <- setdiff(annotation_cols, colnames(annotation_df))
    annotated_results <- merge(merged_summary, annotation_df[,
        c("CpG", available_annotation_cols),
        drop = FALSE
    ],
    by = "CpG",
    all.x = TRUE
    )
    if ("CpG" %in% colnames(annotated_results)) {
        colnames(annotated_results)[colnames(annotated_results) ==
            "CpG"] <- "IlmnID"
    }
    annotated_results <- orderAnnotatedModelColumnsDnaEpico(
        data = annotated_results,
        annotationCols = available_annotation_cols, includeSingular = FALSE
    )

    emitLogMinfiEwasWater(
        c(
            "============================================================",
            paste("Annotated CpG rows:          ", nrow(annotated_results)),
            paste("Annotation columns used:      ",
                paste(available_annotation_cols,
                collapse = ", "
            )), if (length(missing_annotation_cols) ==
                0L) {
                "Missing annotation columns:   none"
            } else {
                paste("Missing annotation columns:   ",
                    paste(missing_annotation_cols,
                    collapse = ", "
                ))
            }, "============================================================"
        ),
        verbose = verbose, log_path = log_path
    )

    structure(list(
        data = annotated_results, fitFailures = fit_failures,
        fitDiagnostics = fit_diagnostics,
            annotationColumnsUsed = available_annotation_cols,
        missingAnnotationCols = missing_annotation_cols
    ), class = "dnaEPICO_methylationGLM_annotation")
}

#' Write optional disk outputs for one-timepoint GLM analyses
#'
#' @param modelResults Object returned by `fitMethylationGLMModels()`.
#' @param modelSummaries Object returned by `summarizeMethylationGLMModels()`.
#' @param annotatedResults Object returned by
#'   `annotateMethylationGLMSummaries()` or a compatible data frame.
#' @param significantCpGs Object returned by
#'   `collectSignificantCpGsMethylationGLM()` or `NULL`.
#' @param outputRData Character. Directory used for serialized model and summary
#'   outputs.
#' @param summaryTxtDir Character. Directory used for tab-delimited summary
#'   tables.
#' @param significantCpGDir Character. Directory used for significant-CpG
#'   coefficient tables.
#' @param annotatedGLMOut Character. Directory used for the annotated summary
#'   XLSX workbook.
#' @param reportAssetsDir Character or `NULL`. Report results directory used for
#'   the compressed TSV table and compact metadata sidecars.
#' @param saveTxtSummaries Logical. If `TRUE`, write tab-delimited summary
#'   tables.
#' @param saveSignificantCpGs Logical. If `TRUE`, write significant-CpG
#'   coefficient tables.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return A list with class `'dnaEPICO_methylationGLM_paths'` containing
#'   the paths of the files written to disk, including the annotated workbook
#'   and any requested report-table sidecars.
#'
#' @description
#' Write optional serialized outputs, summary tables, significant-CpG tables,
#' and annotated results from the one-timepoint GLM workflow.
#'
#' @examples
#' ex <- dnaEPICO:::exampleMethylationGLMStateDnaEpico()
#' annotation_data <- annotateMethylationGLMSummaries(
#'     modelSummaries = ex$modelSummaries,
#'     annotationObject = ex$annotationData,
#'     annotationCols = "Name,chr,pos",
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#' significant_cpgs <- collectSignificantCpGsMethylationGLM(
#'     modelResults = ex$modelResults,
#'     pvalThreshold = 1,
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#' output_paths <- writeMethylationGLMOutputs(
#'     modelResults = ex$modelResults,
#'     modelSummaries = ex$modelSummaries,
#'     annotatedResults = annotation_data,
#'     significantCpGs = significant_cpgs,
#'     outputRData = file.path(ex$tempDir, "models"),
#'     summaryTxtDir = file.path(ex$tempDir, "summary"),
#'     significantCpGDir = file.path(ex$tempDir, "significant"),
#'     annotatedGLMOut = file.path(ex$tempDir, "annotated"),
#'     saveTxtSummaries = TRUE,
#'     saveSignificantCpGs = TRUE,
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#' names(output_paths)
#'
#' @export
writeMethylationGLMOutputs <- function(
    modelResults, modelSummaries,
    annotatedResults, significantCpGs = NULL, outputRData, summaryTxtDir,
    significantCpGDir, annotatedGLMOut, reportAssetsDir = NULL,
    saveTxtSummaries = TRUE,
    saveSignificantCpGs = FALSE, verbose = FALSE, logs = FALSE,
    log_dir = NULL, log_file = "log_methylationGLM.txt"
) {
    if (!is.null(reportAssetsDir) &&
        (!is.character(reportAssetsDir) || length(reportAssetsDir) != 1L ||
            is.na(reportAssetsDir) || !nzchar(reportAssetsDir))) {
        stop("reportAssetsDir must be NULL or one non-empty directory path.",
            call. = FALSE
        )
    }
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )
    dir.create(outputRData, recursive = TRUE, showWarnings = FALSE)
    dir.create(annotatedGLMOut, recursive = TRUE, showWarnings = FALSE)

    model_files <- stats::setNames(
        character(length(modelResults$fits)),
        names(modelResults$fits)
    )
    summary_files <- stats::setNames(
        character(length(modelSummaries$summaries)),
        names(modelSummaries$summaries)
    )
    summary_txt_files <- list()
    significant_files <- list()
    fit_failure_file <- character(0)

    for (phenotype in names(modelResults$fits)) {
        model_file <- file.path(outputRData, paste0(
            phenotype,
            "GLM.rds"
        ))
        summary_file <- file.path(outputRData, paste0(
            phenotype,
            "SummaryGLM.rds"
        ))
        saveRDS(modelResults$fits[[phenotype]], file = model_file)
        saveRDS(modelSummaries$summaries[[phenotype]], file = summary_file)
        model_files[[phenotype]] <- model_file
        summary_files[[phenotype]] <- summary_file
    }

    if (isTRUE(saveTxtSummaries)) {
        dir.create(summaryTxtDir, recursive = TRUE, showWarnings = FALSE)
        for (phenotype in names(modelSummaries$summaries)) {
            summary_df <- modelSummaries$summaries[[phenotype]]
            if (is.null(summary_df) || nrow(summary_df) == 0L) {
                next
            }

            if ("Pr(>|t|)" %in% colnames(summary_df)) {
                summary_df <- summary_df[order(summary_df[["Pr(>|t|)"]]), ,
                    drop = FALSE
                ]
            }

            output_file <- file.path(summaryTxtDir, paste0(
                phenotype,
                "SummaryGLM.txt"
            ))
            utils::write.table(summary_df,
                file = output_file,
                sep = "\t", row.names = FALSE, quote = FALSE
            )
            summary_txt_files[[phenotype]] <- output_file
        }
    }

    if (isTRUE(saveSignificantCpGs) && !is.null(significantCpGs)) {
        dir.create(significantCpGDir, recursive = TRUE, showWarnings = FALSE)
        for (phenotype in names(significantCpGs)) {
            phenotype_hits <- significantCpGs[[phenotype]]
            if (length(phenotype_hits) == 0L) {
                next
            }

            significant_files[[phenotype]] <- character(0)
            phenotype_dir <- file.path(significantCpGDir, phenotype)
            dir.create(phenotype_dir, recursive = TRUE, showWarnings = FALSE)

            for (cpg in names(phenotype_hits)) {
                cpg_dir <- file.path(phenotype_dir, cpg)
                dir.create(cpg_dir, recursive = TRUE, showWarnings = FALSE)
                output_file <- file.path(cpg_dir, paste0(
                    cpg,
                    ".txt"
                ))
                utils::write.table(phenotype_hits[[cpg]],
                    file = output_file,
                    sep = "\t", quote = FALSE
                )
                significant_files[[phenotype]] <- c(
                    significant_files[[phenotype]],
                    output_file
                )
            }
        }
    }

    annotated_df <- annotatedResults
    if (!is.null(annotatedResults$data)) {
        annotated_df <- annotatedResults$data
    }
    report_table <- sortReportTableDnaEpico(annotated_df)
    annotated_df <- report_table$data
    annotated_file <- file.path(annotatedGLMOut, "annotatedGLM.xlsx")
    fit_failures <- modelSummaries$fitFailures
    if (is.null(fit_failures)) {
        fit_failures <- modelResults$fitFailures
    }
    if (is.data.frame(fit_failures) && nrow(fit_failures) > 0L) {
        fit_failure_file <- file.path(annotatedGLMOut, "CpGFitFailuresGLM.txt")
        utils::write.table(fit_failures,
            file = fit_failure_file,
            sep = "\t", row.names = FALSE, quote = FALSE
        )
    }
    dictionary <- buildAnnotatedWorkbookDictionaryMethylationGLM(
        columns = colnames(annotated_df),
        modelDescription = "Pvalue from GLM model",
            formulaText = modelResults$formulas,
        modelLabel = "GLM",
            responseLabel = inferMethylationValueLabelMethylationGLM(modelResults)
    )
    metadata <- buildModelWorkbookMetadataDnaEpico(
        modelResults = modelResults,
        modelSummaries = modelSummaries, annotatedResults = annotatedResults,
        analysis = "glm"
    )
    writeAnnotatedWorkbookMethylationGLM(
        annotated_df = annotated_df,
        file = annotated_file, resultSheet = "annotatedGLM",
        dictionary = dictionary, metadata = metadata
    )
    report_sidecar <- list(
        table = NULL, metadata = NULL, dictionary = NULL,
        workbookMetadata = NULL
    )
    if (!is.null(reportAssetsDir)) {
        report_sidecar <- writeReportTableSidecarDnaEpico(
            tableData = annotated_df,
            workbookFile = annotated_file, sidecarDir = reportAssetsDir,
            sheet = "annotatedGLM", idColumn = report_table$idColumn,
            dictionary = dictionary, workbookMetadata = metadata
        )
    }

    emitLogMinfiEwasWater(
        c(
            "============================================================",
            paste("Serialized model outputs:    ", length(model_files)),
            paste("Serialized summary outputs:  ", length(summary_files)),
            paste("Annotated results file:      ", annotated_file),
            if (is.null(report_sidecar$table)) {
                "Report table sidecar:          not requested"
            } else {
                paste("Report table sidecar:         ", report_sidecar$table)
            },
            paste("CpG fit failures reported:   ", nrow(fit_failures)),
            if (isTRUE(saveTxtSummaries)) {
                paste("Summary text files written:   ",
                    length(summary_txt_files))
            } else {
                "Summary text files written:   0"
            }, if (isTRUE(saveSignificantCpGs)) {
                paste("Significant CpG text files:  ", sum(vapply(
                    significant_files,
                    length, integer(1)
                )))
            } else {
                "Significant CpG text files:  0"
            }, "============================================================"
        ),
        verbose = verbose, log_path = log_path
    )

    structure(
        list(
            modelFiles = model_files, summaryFiles = summary_files,
            summaryTxtFiles = summary_txt_files,
                significantCpGFiles = significant_files,
            fitFailureFile = fit_failure_file, annotatedGLM = annotated_file,
            annotatedGLMText = report_sidecar$table,
                annotatedGLMReportMetadata = report_sidecar$metadata,
            annotatedGLMDictionary = report_sidecar$dictionary,
                annotatedGLMMetadata = report_sidecar$workbookMetadata
        ),
        class = "dnaEPICO_methylationGLM_paths"
    )
}

create_methylation_glm_t1_example <- function(path) {
    phenoBT1 <- data.frame(
        Sample_Name = paste0("S", seq_len(8)),
        status = factor(rep(c("Case", "Case", "Control", "Control"), 2)),
        sex = factor(rep(c("F", "M", "F", "M"), 2)),
        cg00000029 = c(0.20, 0.25, 0.22, 0.27, 0.23, 0.29, 0.26, 0.30),
        cg00000108 = c(0.60, 0.55, 0.52, 0.58, 0.62, 0.57, 0.50, 0.56),
        check.names = FALSE
    )

    input_path <- file.path(path, "phenoBT1.RData")
    save(phenoBT1, file = input_path)

    list(
        inputPheno = input_path,
        phenoBT1 = phenoBT1
    )
}

test_that("methylationGLM returns in-memory results quietly by default", {
    testthat::skip_if_not_installed("IlluminaHumanMethylation450kanno.ilmn12.hg19")

    tmp <- withr::local_tempdir()
    example_data <- create_methylation_glm_t1_example(tmp)

    expect_message(
        result <- methylationGLM(
            inputPheno = example_data$inputPheno,
            outputLogs = file.path(tmp, "logs"),
            outputRData = file.path(tmp, "rData", "methylationGLM", "models"),
            outputPlots = file.path(tmp, "figures", "methylationGLM"),
            phenotypes = "status",
            covariates = "sex",
            factorVars = "status,sex",
            cpgLimit = 2,
            nCores = 1,
            summaryPval = 1,
            saveSignificantCpGs = TRUE,
            significantCpGPval = 1,
            annotationPackage = "IlluminaHumanMethylation450kanno.ilmn12.hg19",
            annotationCols = "Name,chr,pos",
            saveTxtSummaries = TRUE,
            display = FALSE,
            verbose = FALSE,
            logs = FALSE,
            saveOutputs = FALSE
        ),
        NA
    )

    expect_s3_class(result, "dnaEPICO_methylationGLM")
    expect_null(result$savedFiles)
    expect_true("status" %in% names(result$modelFits$fits))
    expect_true("status" %in% names(result$modelSummaries$summaries))
    expect_equal(result$modelFits$settings$parallelBackend, "serial")
    expect_equal(nrow(result$modelFits$summaryCache$status), 2)
    expect_s3_class(result$distributionPlots$phenotypes$status, "ggplot")
    expect_s3_class(result$diagnosticPlots$plots$status$qqplot, "ggplot")
    expect_true("IlmnID" %in% colnames(result$annotation$data))
    expect_equal(result$runSettings$analysisLabel, "methylationGLM")
    expect_equal(result$runSettings$internalResponseColumn, "beta")
    expect_false(dir.exists(file.path(tmp, "figures", "methylationGLM")))
})

test_that("methylationGLM updates the internal response column for M-value runs", {
    tmp <- withr::local_tempdir()
    example_data <- create_methylation_glm_t1_example(tmp)

    prepared <- prepareMethylationGLMData(
        inputPheno = example_data$inputPheno,
        phenotypes = "status",
        covariates = "sex",
        factorVars = "status,sex",
        cpgLimit = 1,
        methylationScale = "m",
        verbose = FALSE,
        logs = FALSE
    )
    fits <- fitMethylationGLMModels(
        preparedData = prepared,
        nCores = 1,
        verbose = FALSE,
        logs = FALSE
    )

    expect_equal(prepared$internalResponseColumn, "m")
    expect_equal(prepared$responseLabel, "M-values")
    expect_equal(fits$settings$internalResponseColumn, "m")
    expect_equal(unname(fits$formulas["status"]), "m ~ `status` + `sex`")
})

test_that("methylationGLM can write outputs and logs on request", {
    testthat::skip_if_not_installed("IlluminaHumanMethylation450kanno.ilmn12.hg19")

    tmp <- withr::local_tempdir()
    example_data <- create_methylation_glm_t1_example(tmp)
    report_assets <- file.path(
        tmp, "reports", "model1", "assets", "results", "glm_results"
    )

    expect_message(
        result <- methylationGLM(
            inputPheno = example_data$inputPheno,
            outputLogs = file.path(tmp, "logs"),
            outputRData = file.path(tmp, "rData", "methylationGLM", "models"),
            outputPlots = file.path(tmp, "figures", "methylationGLM"),
            phenotypes = "status",
            covariates = "sex",
            factorVars = "status,sex",
            cpgLimit = 2,
            nCores = 1,
            summaryPval = 1,
            saveSignificantCpGs = TRUE,
            significantCpGDir = file.path(tmp, "results", "cpgs", "methylationGLM"),
            significantCpGPval = 1,
            saveTxtSummaries = TRUE,
            summaryTxtDir = file.path(tmp, "results", "summary", "methylationGLM"),
            annotationPackage = "IlluminaHumanMethylation450kanno.ilmn12.hg19",
            annotationCols = "Name,chr,pos",
            annotatedGLMOut = file.path(tmp, "data", "methylationGLM"),
            reportAssetsDir = report_assets,
            display = FALSE,
            verbose = TRUE,
            logs = TRUE,
            saveOutputs = TRUE
        ),
        "Starting DNAm GLM Analysis"
    )

    expect_s3_class(result$savedFiles, "dnaEPICO_methylationGLM_paths")
    log_file <- file.path(tmp, "logs", "log_methylationGLM.txt")
    expect_true(file.exists(log_file))
    log_text <- paste(readLines(log_file, warn = FALSE), collapse = "\n")
    expect_match(log_text, "Fit-time summary rows cached")
    expect_match(log_text, "Summary source:\\s+fit-time cache")
    expect_equal(result$runSettings$methylationObjectPrefix, "phenoB")
    expect_equal(result$modelFits$settings$internalResponseColumn, "beta")
    expect_true(file.exists(result$savedFiles$modelFiles[["status"]]))
    expect_true(file.exists(result$savedFiles$summaryFiles[["status"]]))
    expect_true(file.exists(result$savedFiles$summaryTxtFiles[["status"]]))
    expect_true(file.exists(result$savedFiles$annotatedGLM))
    expect_true(file.exists(result$savedFiles$annotatedGLMText))
    expect_true(file.exists(result$savedFiles$annotatedGLMReportMetadata))
    expect_true(file.exists(result$savedFiles$annotatedGLMDictionary))
    expect_true(file.exists(result$savedFiles$annotatedGLMMetadata))
    expect_match(result$savedFiles$annotatedGLM, "annotatedGLM\\.xlsx$")
    expect_equal(
        normalizePath(
            dirname(result$savedFiles$annotatedGLMText),
            winslash = "/", mustWork = FALSE
        ),
        normalizePath(report_assets, winslash = "/", mustWork = FALSE)
    )
    expect_false(dir.exists(file.path(
        dirname(result$savedFiles$annotatedGLM), "report-assets"
    )))
    expect_equal(
        sort(basename(unlist(result$savedFiles[c(
            "annotatedGLMText",
            "annotatedGLMReportMetadata",
            "annotatedGLMDictionary",
            "annotatedGLMMetadata"
        )]))),
        sort(c(
            "annotatedGLM.tsv.gz",
            "annotatedGLM.report.tsv",
            "annotatedGLM.dictionary.tsv",
            "annotatedGLM.metadata.tsv"
        ))
    )
    expect_equal(
        openxlsx::getSheetNames(result$savedFiles$annotatedGLM),
        c("annotatedGLM", "metadata", "dictionary")
    )
    dictionary <- openxlsx::read.xlsx(
        result$savedFiles$annotatedGLM,
        sheet = "dictionary",
        check.names = FALSE
    )
    expect_equal(colnames(dictionary), c("Column", "Description", "Formula"))
    expect_true(any(dictionary$Description == "Pvalue from GLM model"))
    expect_true(any(dictionary$Formula == "GLM: Beta values ~ `status` + `sex`"))
    metadata <- openxlsx::read.xlsx(
        result$savedFiles$annotatedGLM,
        sheet = "metadata",
        check.names = FALSE
    )
    expect_equal(metadata$Value[metadata$Key == "backend"], "glm2")
    expect_equal(metadata$Value[metadata$Key == "fitting_function"], "glm2::glm2")
    expect_true(file.exists(file.path(tmp, "figures", "methylationGLM", "bar_status.tiff")))
    expect_true(file.exists(file.path(tmp, "figures", "methylationGLM", "qqplot_status.tiff")))
    expect_true(length(result$savedFiles$significantCpGFiles$status) >= 1)
    expect_true(all(file.exists(result$savedFiles$significantCpGFiles$status)))
})

test_that("annotated workbook dictionary reports M-values when the fitted response uses M scale", {
    model_results <- list(
        settings = list(
            methylationScale = "m",
            internalResponseColumn = "m"
        ),
        fits = list(
            status = list(
                cg00000029 = list(
                    fitted = c(-0.4, 0.2, 1.3),
                    residuals = c(0, 0, 0)
                )
            )
        ),
        formulas = c(status = "m ~ `status` + `sex`")
    )

    dictionary <- dnaEPICO:::buildAnnotatedWorkbookDictionaryMethylationGLM(
        columns = c("CpG", "statusCaseP.Value"),
        modelDescription = "Pvalue from GLM model",
        formulaText = model_results$formulas,
        modelLabel = "GLM",
        responseLabel = dnaEPICO:::inferMethylationValueLabelMethylationGLM(model_results)
    )

    expect_equal(
        dictionary$Formula[dictionary$Column == "statusCaseP.Value"],
        "GLM: M-values ~ `status` + `sex`"
    )
})

test_that("annotated workbook dictionary reports copy number when requested", {
    model_results <- list(
        settings = list(
            methylationScale = "cn",
            internalResponseColumn = "cn"
        ),
        formulas = c(status = "cn ~ `status` + `sex`")
    )

    dictionary <- dnaEPICO:::buildAnnotatedWorkbookDictionaryMethylationGLM(
        columns = c("CpG", "statusCaseP.Value"),
        modelDescription = "Pvalue from GLM model",
        formulaText = model_results$formulas,
        modelLabel = "GLM",
        responseLabel = dnaEPICO:::inferMethylationValueLabelMethylationGLM(model_results)
    )

    expect_equal(
        dictionary$Formula[dictionary$Column == "statusCaseP.Value"],
        "GLM: Copy number values ~ `status` + `sex`"
    )
})

test_that("multiple GLM formulas use one consistent response label", {
    dictionary <- dnaEPICO:::buildAnnotatedWorkbookDictionaryMethylationGLM(
        columns = c("CpG", "unmappedP.Value"),
        modelDescription = "Pvalue from GLM model",
        formulaText = c(
            status = "beta ~ `status`",
            age = "beta ~ `age`"
        ),
        modelLabel = "GLM",
        responseLabel = "Beta values"
    )

    expect_equal(
        dictionary$Formula[dictionary$Column == "unmappedP.Value"],
        "GLM: Beta values ~ `status`; Beta values ~ `age`"
    )

    mapped_dictionary <- dnaEPICO:::buildAnnotatedWorkbookDictionaryMethylationGLM(
        columns = c("CpG", "statusCaseP.Value", "ageP.Value"),
        modelDescription = "Pvalue from GLM model",
        formulaText = c(
            status = "beta ~ `status`",
            age = "beta ~ `age`"
        ),
        modelLabel = "GLM",
        responseLabel = "Beta values"
    )

    expect_equal(
        mapped_dictionary$Formula[
            mapped_dictionary$Column == "statusCaseP.Value"
        ],
        "GLM: Beta values ~ `status`"
    )
    expect_equal(
        mapped_dictionary$Formula[
            mapped_dictionary$Column == "ageP.Value"
        ],
        "GLM: Beta values ~ `age`"
    )
})

test_that("GLM diagnostics retain convergence quality in summaries and annotation", {
    tmp <- withr::local_tempdir()
    example_data <- create_methylation_glm_t1_example(tmp)
    prepared <- prepareMethylationGLMData(
        inputPheno = example_data$inputPheno,
        phenotypes = "status",
        covariates = "sex",
        factorVars = "status,sex",
        cpgLimit = 1,
        logs = FALSE
    )
    fits <- fitMethylationGLMModels(prepared, nCores = 1, logs = FALSE)
    summaries <- summarizeMethylationGLMModels(
        fits,
        prepared,
        excludeNonConverged = TRUE,
        logs = FALSE
    )
    summary_table <- summaries$summaries$status
    quality_columns <- c(
        "Fit.Status", "Converged", "Convergence.Message",
        "Fit.Warning", "Inference.Included", "Exclusion.Reason"
    )
    expect_true(all(quality_columns %in% names(summary_table)))
    expect_true(summary_table$Converged)
    expect_true(summary_table$Inference.Included)
    expect_false("Singular.Fit" %in% names(summary_table))

    annotation <- annotateMethylationGLMSummaries(
        modelSummaries = summaries,
        annotationObject = data.frame(
            CpG = "cg00000029",
            chr = "chr1",
            stringsAsFactors = FALSE
        ),
        annotationCols = "chr",
        logs = FALSE
    )
    expect_true(all(c(
        "status_Converged", "status_Inference.Included",
        "status_Convergence.Message"
    ) %in% names(annotation$data)))
    expect_false(any(c("Fit.Status", "Converged", "Inference.Included") %in% names(annotation$data)))
    expect_true(annotation$data$status_Converged)
    expect_equal(names(annotation$data)[[1L]], "IlmnID")
    expect_lt(match("chr", names(annotation$data)), match("status_Fit.Status", names(annotation$data)))
})

test_that("non-converged GLM failures retain their CpG and convergence reason", {
    failed_fit <- dnaEPICO:::newMethylationFitErrorDnaEpico(
        reason = "The GLM did not converge.",
        errorClass = "dnaEPICO_methylationGLM_fit_error",
        status = "not_converged",
        converged = FALSE,
        convergenceMessage = "The GLM did not converge."
    )
    fits <- list(status = list(cg_failed = failed_fit))
    diagnostics <- dnaEPICO:::collectFitDiagnosticsMethylationGLM(fits)
    failures <- dnaEPICO:::collectFitFailuresMethylationModels(
        fits,
        "dnaEPICO_methylationGLM_fit_error"
    )
    annotation <- annotateMethylationGLMSummaries(
        modelSummaries = list(
            diagnosticSummaries = list(status = data.frame()),
            summaries = list(status = data.frame()),
            phenotypes = "status",
            fitFailures = failures,
            fitDiagnostics = diagnostics
        ),
        annotationObject = data.frame(
            CpG = "cg_failed",
            chr = "chr1",
            stringsAsFactors = FALSE
        ),
        annotationCols = "chr",
        logs = FALSE
    )

    expect_equal(annotation$data$IlmnID, "cg_failed")
    expect_equal(annotation$data$status_Fit.Status, "not_converged")
    expect_false(annotation$data$status_Converged)
    expect_match(annotation$data$status_Convergence.Message, "did not converge")
    expect_match(annotation$data$status_Exclusion.Reason, "did not converge")
})

test_that("scaleVars standardizes only selected GLM fixed effects", {
    tmp <- withr::local_tempdir()
    example_data <- create_methylation_glm_t1_example(tmp)
    phenoBT1 <- example_data$phenoBT1
    phenoBT1$age <- seq(20, 55, length.out = nrow(phenoBT1))
    save(phenoBT1, file = example_data$inputPheno)

    prepared <- prepareMethylationGLMData(
        inputPheno = example_data$inputPheno,
        phenotypes = "status",
        covariates = "sex,age",
        factorVars = "status,sex",
        scaleVars = "age",
        cpgLimit = 1,
        logs = FALSE
    )

    expect_equal(prepared$data$age, phenoBT1$age)
    expect_equal(mean(prepared$modelData$age), 0, tolerance = 1e-12)
    expect_equal(stats::sd(prepared$modelData$age), 1, tolerance = 1e-12)
    expect_equal(prepared$scaleVars, "age")
    expect_equal(prepared$scalingMetadata$Variable, "age")
    expect_error(
        prepareMethylationGLMData(
            inputPheno = example_data$inputPheno,
            phenotypes = "status",
            covariates = "sex,age",
            factorVars = "status,sex",
            scaleVars = "sex",
            cpgLimit = 1,
            logs = FALSE
        ),
        "both factorVars and scaleVars"
    )
    expect_error(
        prepareMethylationGLMData(
            inputPheno = example_data$inputPheno,
            phenotypes = "status",
            covariates = "sex",
            factorVars = "status,sex",
            scaleVars = "age",
            cpgLimit = 1,
            logs = FALSE
        ),
        "fixed-effect variables"
    )
})

test_that("parallel plans use engine crossover and resource caps", {
    withr::local_envvar(c(
        DNAEPICO_PARALLEL_BACKEND = "auto",
        DNAEPICO_MAX_WORKERS = "2"
    ))
    expect_equal(dnaEPICO:::parallelCrossoverMethylationModels("glm2"), 25000L)
    expect_equal(dnaEPICO:::parallelCrossoverMethylationModels("lme4"), 1500L)
    expect_equal(dnaEPICO:::parallelCrossoverMethylationModels("nlme"), 5000L)

    small <- dnaEPICO:::resolveParallelPlanMethylationModels(
        engine = "glm2",
        nCores = 8,
        nCpGs = 100
    )
    expect_equal(small$backend, "serial")
    expect_equal(small$workerCount, 1L)
    expect_match(small$reason, "below")

    withr::local_envvar(DNAEPICO_PARALLEL_BACKEND = "psock")
    forced <- dnaEPICO:::resolveParallelPlanMethylationModels(
        engine = "glm2",
        nCores = 8,
        nCpGs = 3
    )
    expect_lte(forced$workerCount, 2L)
    expect_lte(forced$workerCount, 3L)
})

test_that("annotateMethylationGLMSummaries accepts annotation package names", {
    testthat::skip_if_not_installed("IlluminaHumanMethylation450kanno.ilmn12.hg19")

    ex <- dnaEPICO:::exampleMethylationGLMStateDnaEpico()

    annotation_data <- annotateMethylationGLMSummaries(
        modelSummaries = ex$modelSummaries,
        annotationObject = "IlluminaHumanMethylation450kanno.ilmn12.hg19",
        annotationCols = "Name,chr,pos",
        verbose = FALSE,
        logs = FALSE
    )

    expect_s3_class(annotation_data, "dnaEPICO_methylationGLM_annotation")
    expect_true("IlmnID" %in% colnames(annotation_data$data))
})

test_that("methylation-scale validation follows Beta, M, and CN definitions", {
    expect_identical(
        vapply(
            c("beta", "Beta", "BETA", "m", "M", "cn", "CN", "cN"),
            dnaEPICO:::normalizeMethylationScaleDnaEpico,
            character(1)
        ),
        c(
            beta = "beta", Beta = "beta", BETA = "beta",
            m = "m", M = "m", cn = "cn", CN = "cn", cN = "cn"
        )
    )
    expect_error(
        dnaEPICO:::normalizeMethylationScaleDnaEpico("unsupported"),
        "Beta, M, CN"
    )

    beta <- rbind(
        cg_boundary = c(0, 1),
        cg_nan = c(NaN, 0.5),
        cg_low = c(-0.01, 0.2),
        cg_inf = c(Inf, 0.3),
        cg_missing = c(NA_real_, NA_real_)
    )
    beta_result <- dnaEPICO:::inspectMethylationMatrixDnaEpico(beta, "beta")

    expect_true(is.na(beta_result$values["cg_nan", 1]))
    expect_equal(beta_result$boundaries$Observed.Minimum, -0.01)
    expect_equal(beta_result$boundaries$Observed.Maximum, 1)
    expect_equal(beta_result$boundaries$At.Lower.Boundary, 1L)
    expect_equal(beta_result$boundaries$At.Upper.Boundary, 1L)
    expect_equal(beta_result$boundaries$NaN.Converted, 1L)
    expect_equal(beta_result$boundaries$Invalid.CpGs, 3L)
    expect_setequal(
        beta_result$invalidCpGs$CpG,
        c("cg_low", "cg_inf", "cg_missing")
    )
    expect_identical(
        beta_result$issues$Status[beta_result$issues$CpG == "cg_nan"],
        "NaN converted to NA"
    )

    m_values <- rbind(cg_finite = c(-100, 100), cg_inf = c(-Inf, 0))
    m_result <- dnaEPICO:::inspectMethylationMatrixDnaEpico(m_values, "m")
    expect_equal(m_result$boundaries$Out.Of.Range, 0L)
    expect_identical(m_result$invalidCpGs$CpG, "cg_inf")

    cn_values <- rbind(cg_finite = c(-20, 30), cg_inf = c(Inf, 0))
    cn_result <- dnaEPICO:::inspectMethylationMatrixDnaEpico(cn_values, "cn")
    expect_equal(cn_result$boundaries$Out.Of.Range, 0L)
    expect_identical(cn_result$invalidCpGs$CpG, "cg_inf")
})

test_that("methylation probe identifiers must be present and unique", {
    expect_identical(
        dnaEPICO:::validateMethylationProbeIdentifiersDnaEpico(
            c("cg00000001", "cg00000002")
        ),
        c("cg00000001", "cg00000002")
    )
    expect_error(
        dnaEPICO:::validateMethylationProbeIdentifiersDnaEpico(
            c("cg00000001", "cg00000001")
        ),
        "duplicate CpG identifiers"
    )
    expect_error(
        dnaEPICO:::validateMethylationProbeIdentifiersDnaEpico(c("cg00000001", "")),
        "missing or blank CpG identifiers"
    )
})

test_that("GLM skips and reports invalid CpGs without stopping valid fits", {
    tmp <- withr::local_tempdir()
    phenoBT1 <- data.frame(
        Sample_Name = paste0("S", seq_len(12)),
        phenotype = rep(c(0, 1), 6),
        cg_valid = seq(0.1, 0.65, length.out = 12),
        cg_nan = c(NaN, seq(0.2, 0.7, length.out = 11)),
        cg_invalid = c(1.1, seq(0.2, 0.7, length.out = 11)),
        check.names = FALSE
    )
    input_file <- file.path(tmp, "phenoBT1.RData")
    save(phenoBT1, file = input_file)

    prepared <- suppressWarnings(prepareMethylationGLMData(
        inputPheno = input_file,
        phenotypes = "phenotype",
        covariates = character(0),
        factorVars = character(0),
        methylationScale = "beta",
        verbose = FALSE,
        logs = FALSE
    ))
    fits <- fitMethylationGLMModels(
        preparedData = prepared,
        nCores = 1,
        verbose = FALSE,
        logs = FALSE
    )
    summaries <- summarizeMethylationGLMModels(
        modelResults = fits,
        preparedData = prepared,
        verbose = FALSE,
        logs = FALSE
    )
    annotation <- annotateMethylationGLMSummaries(
        modelSummaries = summaries,
        annotationObject = data.frame(
            CpG = c("cg_valid", "cg_nan", "cg_invalid"),
            chr = "chr1",
            stringsAsFactors = FALSE
        ),
        annotationCols = "chr",
        verbose = FALSE,
        logs = FALSE
    )

    expect_identical(fits$fitFailures$CpG, "cg_invalid")
    expect_identical(fits$fitFailures$Status, "invalid")
    expect_match(fits$fitFailures$Error, "outside \\[0, 1\\]")
    expect_setequal(summaries$summaries$phenotype$CpG, c("cg_valid", "cg_nan"))
    expect_setequal(annotation$data$IlmnID, c("cg_valid", "cg_nan", "cg_invalid"))
    expect_identical(
        annotation$data$phenotype_Fit.Status[
            annotation$data$IlmnID == "cg_invalid"
        ],
        "invalid"
    )
    expect_match(
        annotation$data$phenotype_Exclusion.Reason[
            annotation$data$IlmnID == "cg_invalid"
        ],
        "outside \\[0, 1\\]"
    )
})

test_that("GLM retains an inventory when every CpG is invalid", {
    tmp <- withr::local_tempdir()
    phenoBT1 <- data.frame(
        Sample_Name = paste0("S", seq_len(8)),
        phenotype = rep(c(0, 1), 4),
        cg_invalid = rep(1.2, 8),
        check.names = FALSE
    )
    input_file <- file.path(tmp, "phenoBT1.RData")
    save(phenoBT1, file = input_file)
    prepared <- suppressWarnings(prepareMethylationGLMData(
        inputPheno = input_file,
        phenotypes = "phenotype",
        covariates = character(0),
        factorVars = character(0),
        verbose = FALSE,
        logs = FALSE
    ))

    expect_warning(
        fits <- fitMethylationGLMModels(prepared, nCores = 1, logs = FALSE),
        "failure inventory was retained"
    )
    expect_equal(nrow(fits$fitFailures), 1L)
    expect_identical(fits$fitFailures$Status, "invalid")
})

test_that("lme4 and nlme skip the same invalid longitudinal CpG", {
    testthat::skip_if_not_installed("lmerTest")
    testthat::skip_if_not_installed("nlme")

    tmp <- withr::local_tempdir()
    set.seed(11)
    phenoBT1T2 <- data.frame(
        SID = paste0("P", rep(seq_len(10), each = 2), c("A", "B")),
        person = rep(seq_len(10), each = 2),
        Timepoint = factor(rep(c("T1", "T2"), 10)),
        phenotype = rep(c(0, 1), each = 10),
        cg_valid = 0.3 + rep(c(0, 0.04), 10) + stats::rnorm(20, 0, 0.01),
        cg_invalid = c(-0.1, rep(0.4, 19)),
        check.names = FALSE
    )
    input_file <- file.path(tmp, "phenoBT1T2.RData")
    save(phenoBT1T2, file = input_file)
    prepared <- suppressWarnings(prepareMethylationLMEData(
        inputPheno = input_file,
        personVar = "person",
        timeVar = "Timepoint",
        phenotypes = "phenotype",
        covariates = character(0),
        factorVars = "phenotype",
        verbose = FALSE,
        logs = FALSE
    ))

    lme4_fits <- fitMethylationLMEModels(
        preparedData = prepared,
        lmeLibs = "lme4,lmerTest",
        nCores = 1,
        logs = FALSE
    )
    nlme_fits <- fitMethylationLMEModels(
        preparedData = prepared,
        lmeLibs = "nlme",
        nCores = 1,
        logs = FALSE
    )

    expect_identical(lme4_fits$fitFailures$CpG, "cg_invalid")
    expect_identical(nlme_fits$fitFailures$CpG, "cg_invalid")
    expect_identical(lme4_fits$fitFailures$Status, "invalid")
    expect_identical(nlme_fits$fitFailures$Status, "invalid")
    expect_false(inherits(lme4_fits$fits$phenotype$cg_valid, "dnaEPICO_methylationLME_fit_error"))
    expect_false(inherits(nlme_fits$fits$phenotype$cg_valid, "dnaEPICO_methylationLME_fit_error"))
})

test_that("model preparation selects scale-specific data from multi-object external files", {
    tmp <- withr::local_tempdir()
    phenoMT1 <- data.frame(
        Sample_Name = paste0("S", seq_len(8)),
        phenotype = rep(c(0, 1), 4),
        cg_external = seq(-4, 3),
        check.names = FALSE
    )
    distractor <- data.frame(not_the_model_input = 1)
    glm_file <- file.path(tmp, "phenoMT1.RData")
    save(phenoMT1, distractor, file = glm_file)

    glm_data <- prepareMethylationGLMData(
        inputPheno = glm_file,
        phenotypes = "phenotype",
        covariates = character(0),
        factorVars = character(0),
        methylationScale = "m",
        verbose = FALSE,
        logs = FALSE
    )
    expect_equal(glm_data$data$cg_external, phenoMT1$cg_external)
    expect_identical(glm_data$methylationScale, "m")

    phenoCNT1T2 <- data.frame(
        SID = paste0("P", rep(seq_len(4), each = 2), rep(c("A", "B"), 4)),
        person = rep(seq_len(4), each = 2),
        Timepoint = rep(c(1, 2), 4),
        phenotype = rep(c(0, 1), 4),
        cg_external = seq(-20, 15, length.out = 8),
        check.names = FALSE
    )
    lme_file <- file.path(tmp, "phenoCNT1T2.RData")
    save(phenoCNT1T2, distractor, file = lme_file)

    lme_data <- prepareMethylationLMEData(
        inputPheno = lme_file,
        personVar = "person",
        timeVar = "Timepoint",
        phenotypes = "phenotype",
        covariates = character(0),
        factorVars = character(0),
        methylationScale = "cn",
        verbose = FALSE,
        logs = FALSE
    )
    expect_identical(lme_data$data$cg_external, phenoCNT1T2$cg_external)
    expect_identical(lme_data$methylationScale, "cn")
})

create_dnam_report_example <- function(path) {
    dirs <- list(
        qcDir = file.path(path, "figures", "preprocessingMinfiEwasWater", "enmix"),
        preprocessingDir = file.path(path, "figures", "preprocessingMinfiEwasWater", "qc"),
        postprocessingDir = file.path(path, "figures", "preprocessingMinfiEwasWater", "metrics"),
        svaDir = file.path(path, "figures", "svaEnmix"),
        glmDir = file.path(path, "figures", "methylationGLM"),
        lmeDir = file.path(path, "figures", "methylationLME"),
        glmTableDir = file.path(path, "data", "qpasst1", "methylationGLM"),
        lmeTableDir = file.path(path, "data", "qpasst1", "methylationLME"),
        phenoTabDir = file.path(path, "data", "qpasst1", "preprocessingMinfiEwasWater"),
        figDir = file.path(path, "reports", "figures"),
        outputDir = file.path(path, "reports"),
        glmReportDir = file.path(path, "reports", "assets", "results", "glm_results"),
        lmeReportDir = file.path(path, "reports", "assets", "results", "lme_results"),
        logDir = file.path(path, "logs")
    )

    invisible(lapply(dirs, dir.create, recursive = TRUE, showWarnings = FALSE))

    glm_subdir <- file.path(dirs$glmDir, "status")
    lme_subdir <- file.path(dirs$lmeDir, "score")
    dir.create(glm_subdir, recursive = TRUE, showWarnings = FALSE)
    dir.create(lme_subdir, recursive = TRUE, showWarnings = FALSE)

    file.create(file.path(dirs$qcDir, "qc_1.jpg"))
    file.create(file.path(dirs$preprocessingDir, "pre_1.tiff"))
    file.create(file.path(dirs$postprocessingDir, "post_1.tiff"))
    file.create(file.path(dirs$svaDir, "sva_1.tiff"))
    file.create(file.path(glm_subdir, "qqplot_status.tiff"))
    file.create(file.path(lme_subdir, "qqplot_score.tiff"))

    dirs$glmTablePath <- file.path(dirs$glmTableDir, "annotatedGLM.xlsx")
    dirs$lmeTablePath <- file.path(dirs$lmeTableDir, "annotatedLME.xlsx")
    dirs$phenoTab <- file.path(dirs$phenoTabDir, "phenoLC.csv")
    utils::write.csv(
        data.frame(
            person = c("P1", "P2"),
            Sentrix_ID = c("Chip1", "Chip1"),
            Timepoint = c(1, 2),
            check.names = FALSE
        ),
        dirs$phenoTab,
        row.names = FALSE
    )
    glm_dictionary <- data.frame(
        Column = c("P.Value", "IlmnID", "Name", "chr", "pos"),
        Description = c(
            "Pvalue from GLM model",
            "CpG probe identifier",
            "CpG probe identifier",
            "Genomic annotation or supporting result column",
            "Genomic annotation or supporting result column"
        ),
        Formula = c(
            "GLM: Beta values ~ `status` + `sex`",
            "",
            "",
            "",
            ""
        ),
        check.names = FALSE
    )
    glm_data <- data.frame(
            IlmnID = "cg00000029",
            Name = "cg00000029",
            P.Value = 0.001,
            chr = "chr16",
            pos = 53434200,
            UCSC_RefGene_Group = "TSS1500",
            UCSC_RefGene_Name = "RBL2",
            Relation_to_Island = "OpenSea",
            GencodeV41_Group = "TSS",
            check.names = FALSE
    )
    openxlsx::write.xlsx(
        list(annotatedGLM = glm_data, dictionary = glm_dictionary),
        dirs$glmTablePath,
        overwrite = TRUE
    )
    dnaEPICO:::writeReportTableSidecarDnaEpico(
        tableData = glm_data,
        workbookFile = dirs$glmTablePath,
        sidecarDir = dirs$glmReportDir,
        sheet = "annotatedGLM",
        idColumn = "IlmnID",
        dictionary = glm_dictionary
    )
    lme_dictionary <- data.frame(
        Column = c("phenotype:Timepoint P.Value", "IlmnID", "Name", "chr", "pos"),
        Description = c(
            "Pvalue from LME model",
            "CpG probe identifier",
            "CpG probe identifier",
            "Genomic annotation or supporting result column",
            "Genomic annotation or supporting result column"
        ),
        Formula = c(
            "LME: Beta values ~ `score` + `Timepoint` + `sex` + (1 | `person` )",
            "",
            "",
            "",
            ""
        ),
        check.names = FALSE
    )
    lme_data <- data.frame(
            IlmnID = "cg00000108",
            Name = "cg00000108",
            "phenotype:Timepoint P.Value" = 0.002,
            chr = "chr3",
            pos = 37417715,
            UCSC_RefGene_Group = "Body",
            UCSC_RefGene_Name = "C3orf35",
            Relation_to_Island = "Island",
            GencodeV41_Group = "gene body",
            check.names = FALSE
    )
    openxlsx::write.xlsx(
        list(annotatedLME = lme_data, dictionary = lme_dictionary),
        dirs$lmeTablePath,
        overwrite = TRUE
    )
    dnaEPICO:::writeReportTableSidecarDnaEpico(
        tableData = lme_data,
        workbookFile = dirs$lmeTablePath,
        sidecarDir = dirs$lmeReportDir,
        sheet = "annotatedLME",
        idColumn = "IlmnID",
        dictionary = lme_dictionary
    )

    dirs
}

write_multiple_formula_workbooks <- function(example_dirs, backend = "lme4") {
    phenotypes <- c("T1D", "HbA1c", "C-peptide")
    glm_formulas <- c(
        "GLM: M-values ~ `T1D` + `Age` + `Sex` + `PC1` + `PC2` + `PC3`",
        "GLM: M-values ~ `HbA1c` + `Age` + `Sex` + `PC1` + `PC2` + `PC3`",
        "GLM: M-values ~ `C-peptide` + `Age` + `Sex` + `PC1` + `PC2` + `PC3`"
    )
    lme_formulas <- c(
        "LME: M-values ~ `T1D` + `Age` + `Sex` + `PC1` + `PC2` + `PC3` + (1 | `SID` )",
        "LME: M-values ~ `HbA1c` + `Age` + `Sex` + `PC1` + `PC2` + `PC3` + (1 | `SID` )",
        "LME: M-values ~ `C-peptide` + `Age` + `Sex` + `PC1` + `PC2` + `PC3` + (1 | `SID` )"
    )
    pvalue_columns <- paste0(phenotypes, "_P.Value")
    annotation_columns <- c("IlmnID", "Name", "chr", "pos")

    glm_data <- data.frame(IlmnID = "cg00000029", check.names = FALSE)
    glm_data[pvalue_columns] <- list(0.001, 0.02, 0.03)
    glm_data$Name <- "cg00000029"
    glm_data$chr <- "chr16"
    glm_data$pos <- 53434200
    glm_dictionary <- data.frame(
        Column = c(pvalue_columns, annotation_columns),
        Description = c(
            rep("Pvalue from GLM model", length(pvalue_columns)),
            "CpG probe identifier",
            "CpG probe identifier",
            "Genomic annotation or supporting result column",
            "Genomic annotation or supporting result column"
        ),
        Formula = c(glm_formulas, rep("", length(annotation_columns))),
        check.names = FALSE
    )
    openxlsx::write.xlsx(
        list(annotatedGLM = glm_data, dictionary = glm_dictionary),
        example_dirs$glmTablePath,
        overwrite = TRUE
    )

    lme_data <- data.frame(IlmnID = "cg00000108", check.names = FALSE)
    lme_data[pvalue_columns] <- list(0.002, 0.01, 0.04)
    lme_data$Name <- "cg00000108"
    lme_data$chr <- "chr3"
    lme_data$pos <- 37417715
    lme_dictionary <- data.frame(
        Column = c(pvalue_columns, annotation_columns),
        Description = c(
            rep("Pvalue from LME model", length(pvalue_columns)),
            "CpG probe identifier",
            "CpG probe identifier",
            "Genomic annotation or supporting result column",
            "Genomic annotation or supporting result column"
        ),
        Formula = c(lme_formulas, rep("", length(annotation_columns))),
        check.names = FALSE
    )
    lme_metadata <- data.frame(
        Key = c(
            "backend",
            "fitting_function",
            "libraries",
            "correlation_structure",
            "correlation_variable",
            "interaction_term",
            "phenotypes"
        ),
        Value = c(
            backend,
            if (identical(backend, "nlme")) "nlme::lme" else "lmerTest::lmer",
            if (identical(backend, "nlme")) "nlme" else "lme4,lmerTest",
            if (identical(backend, "nlme")) "AR1" else "none",
            if (identical(backend, "nlme")) "Timepoint" else "None",
            "None",
            paste(phenotypes, collapse = ",")
        ),
        stringsAsFactors = FALSE
    )
    openxlsx::write.xlsx(
        list(annotatedLME = lme_data, dictionary = lme_dictionary, metadata = lme_metadata),
        example_dirs$lmeTablePath,
        overwrite = TRUE
    )
    dnaEPICO:::writeReportTableSidecarDnaEpico(
        tableData = glm_data,
        workbookFile = example_dirs$glmTablePath,
        sidecarDir = example_dirs$glmReportDir,
        sheet = "annotatedGLM",
        idColumn = "IlmnID",
        dictionary = glm_dictionary
    )
    dnaEPICO:::writeReportTableSidecarDnaEpico(
        tableData = lme_data,
        workbookFile = example_dirs$lmeTablePath,
        sidecarDir = example_dirs$lmeReportDir,
        sheet = "annotatedLME",
        idColumn = "IlmnID",
        dictionary = lme_dictionary,
        workbookMetadata = lme_metadata
    )

    invisible(example_dirs)
}

test_that("report table sidecars are read in bounded chunks without losing rows", {
    tmp <- withr::local_tempdir()
    table_path <- file.path(tmp, "annotatedLME.tsv.gz")
    table_data <- data.frame(
        IlmnID = sprintf("cg%08d", seq_len(10025)),
        P.Value = seq_len(10025) / 10026,
        Fit.Status = "fitted",
        stringsAsFactors = FALSE,
        check.names = FALSE
    )
    data.table::fwrite(
        table_data,
        file = table_path,
        sep = "\t",
        quote = TRUE,
        compress = "gzip"
    )
    identifiers <- character(0)
    stream_result <- dnaEPICO:::streamReportTableDnaEpico(
        tableFile = table_path,
        chunkSize = 5000,
        expectedRows = nrow(table_data),
        expectedColumns = ncol(table_data),
        chunkHandler = function(chunk, chunk_number) {
            identifiers <<- c(identifiers, chunk$IlmnID)
        }
    )

    expect_equal(stream_result$nRows, 10025)
    expect_equal(stream_result$nChunks, 3)
    expect_equal(stream_result$maximumChunkRows, 5000)
    expect_identical(identifiers, table_data$IlmnID)
})

test_that("prepareDnamReportInputs returns a structured inventory quietly", {
    tmp <- withr::local_tempdir()
    example_dirs <- create_dnam_report_example(tmp)

    expect_message(
        result <- prepareDnamReportInputs(
            outputDir = example_dirs$outputDir,
            qcDir = example_dirs$qcDir,
            preprocessingDir = example_dirs$preprocessingDir,
            postprocessingDir = example_dirs$postprocessingDir,
            svaDir = example_dirs$svaDir,
            glmDir = example_dirs$glmDir,
            lmeDir = example_dirs$lmeDir,
            figDir = example_dirs$figDir,
            verbose = FALSE,
            logs = FALSE
        ),
        NA
    )

    expect_s3_class(result, "dnaEPICO_dnamReport_prepared")
    expect_equal(result$figureInventory$qc$count, 1)
    expect_equal(result$figureInventory$glm$count, 1)
    expect_equal(result$figureInventory$lme$count, 1)
    expect_length(result$missingFigureDirectories, 0)
    expect_match(result$outputFile, "docs/index\\.html$")
})

test_that("dnamReport renders the dashboard and writes logs on request", {
    tmp <- withr::local_tempdir()
    example_dirs <- create_dnam_report_example(tmp)

    expect_message(
        result <- dnamReport(
            outputDir = example_dirs$outputDir,
            phenoTab = example_dirs$phenoTab,
            enmixTab = example_dirs$qcDir,
            qcTab = example_dirs$preprocessingDir,
            svaTab = example_dirs$svaDir,
            metricTab = example_dirs$postprocessingDir,
            glmTab = example_dirs$glmTablePath,
            lmeTab = example_dirs$lmeTablePath,
            verbose = TRUE,
            logs = TRUE,
            logTab = example_dirs$logDir
        ),
        "Report path:"
    )

    expect_s3_class(result, "dnaEPICO_dnamReport")
    expect_true(result$status %in% c("rendered", "skipped", "failed"))
    expect_s3_class(result$preparedReport, "dnaEPICO_dnamReport_prepared")
    expect_s3_class(result$renderResult, "dnaEPICO_dnamReport_render")
    expect_true(file.exists(result$logFile))
    report_log <- paste(readLines(result$logFile, warn = FALSE), collapse = "\n")
    expect_match(report_log, "GLM report table source: streamed_sidecar", fixed = TRUE)
    expect_match(report_log, "LME report table source: streamed_sidecar", fixed = TRUE)
    if (identical(result$status, "rendered")) {
        expect_true(file.exists(result$outputFile))
    }
    expect_true(any(grepl("glm\\.qmd$", result$sourceFiles)))
    expect_true(any(grepl("lme\\.qmd$", result$sourceFiles)))
    expect_equal(
        unname(result$resultTableSources[c("GLM", "LME")]),
        c("streamed_sidecar", "streamed_sidecar")
    )

    glm_qmd <- readLines(file.path(result$projectDir, "glm.qmd"), warn = FALSE)
    lme_qmd <- readLines(file.path(result$projectDir, "lme.qmd"), warn = FALSE)
    data_qmd <- readLines(file.path(result$projectDir, "index.qmd"), warn = FALSE)
    expect_true(any(grepl('data-result-key="phenotype_data"', data_qmd, fixed = TRUE)))
    expect_true(any(grepl('data-role="filter-column"', data_qmd, fixed = TRUE)))
    expect_equal(sum(grepl("dnaepico-viewer-pagination", data_qmd, fixed = TRUE)), 2)
    expect_false(any(grepl("DT::datatable", data_qmd, fixed = TRUE)))
    expect_false(any(grepl("htmlwidget", data_qmd, fixed = TRUE)))
    expect_true(any(grepl(
        "Table 1. Generalised Linear Model Results and Genomic Annotation of CpG Sites by Phenotype(s)",
        glm_qmd,
        fixed = TRUE
    )))
    expect_false(any(grepl("paged viewer", glm_qmd, fixed = TRUE)))
    expect_true(any(grepl("fitted with `glm2`", glm_qmd, fixed = TRUE)))
    expect_true(any(grepl("The recorded model formula is", glm_qmd, fixed = TRUE)))
    expect_true(any(grepl("GLM: Beta values ~ status + sex", glm_qmd, fixed = TRUE)))
    expect_true(any(grepl("data-result-key=\"glm_results\"", glm_qmd, fixed = TRUE)))
    expect_false(any(grepl("DT::datatable", glm_qmd, fixed = TRUE)))
    expect_false(any(grepl("openxlsx::read.xlsx", glm_qmd, fixed = TRUE)))
    expect_true(any(grepl(
        "Table 1. Linear Mixed-Effects Model Results and Genomic Annotation of CpG Sites by Phenotype(s)",
        lme_qmd,
        fixed = TRUE
    )))
    expect_false(any(grepl("paged viewer", lme_qmd, fixed = TRUE)))
    expect_true(any(grepl("`lmerTest::lmer()`", lme_qmd, fixed = TRUE)))
    expect_true(any(grepl("No phenotype-by-time interaction was fitted", lme_qmd, fixed = TRUE)))
    expect_true(any(grepl("The recorded model formula is", lme_qmd, fixed = TRUE)))
    expect_true(any(grepl("data-result-key=\"lme_results\"", lme_qmd, fixed = TRUE)))
    expect_false(any(grepl("DT::datatable", lme_qmd, fixed = TRUE)))
    expect_false(any(grepl("openxlsx::read.xlsx", lme_qmd, fixed = TRUE)))

    result_assets <- file.path(result$projectDir, "assets", "results")
    expect_true(file.exists(file.path(result_assets, "phenotype_data", "manifest.js")))
    expect_true(file.exists(file.path(result_assets, "phenotype_data", "chunk-0001.js")))
    expect_true(file.exists(file.path(result_assets, "phenotype_data", basename(example_dirs$phenoTab))))
    expect_true(file.exists(file.path(result_assets, "glm_results", "manifest.js")))
    expect_true(file.exists(file.path(result_assets, "glm_results", "chunk-0001.js")))
    expect_false(file.exists(file.path(result_assets, "glm_results", "annotatedGLM.xlsx")))
    expect_true(file.exists(file.path(result_assets, "glm_results", "annotatedGLM.tsv.gz")))
    expect_true(file.exists(file.path(result_assets, "lme_results", "manifest.js")))
    expect_true(file.exists(file.path(result_assets, "lme_results", "chunk-0001.js")))
    expect_false(file.exists(file.path(result_assets, "lme_results", "annotatedLME.xlsx")))
    expect_true(file.exists(file.path(result_assets, "lme_results", "annotatedLME.tsv.gz")))
    expect_true(file.exists(file.path(result$projectDir, "assets", "cpg-viewer.js")))
    viewer_script <- readLines(
        file.path(result$projectDir, "assets", "cpg-viewer.js"),
        warn = FALSE
    )
    expect_true(any(grepl("maxCachedChunks", viewer_script, fixed = TRUE)))
    expect_false(any(grepl("Download all results (XLSX)", glm_qmd, fixed = TRUE)))
    expect_false(any(grepl("Download all results (XLSX)", lme_qmd, fixed = TRUE)))

    enmix_qmd <- readLines(file.path(result$projectDir, "enmix-qc.qmd"), warn = FALSE)
    metrics_qmd <- readLines(file.path(result$projectDir, "metrics.qmd"), warn = FALSE)
    logs_qmd <- readLines(file.path(result$projectDir, "logs.qmd"), warn = FALSE)
    expect_true(any(grepl("`ENmix` produces control plots", enmix_qmd, fixed = TRUE)))
    expect_true(any(grepl("`Figure 1`", metrics_qmd, fixed = TRUE)))
    expect_true(any(grepl("Displays the methylation preprocessing log", logs_qmd, fixed = TRUE)))
    expect_true(any(grepl("Displays the phenotype preparation log", logs_qmd, fixed = TRUE)))
    expect_true(any(grepl("Displays the hidden-effect and surrogate-variable analysis log", logs_qmd, fixed = TRUE)))
    expect_true(any(grepl('::: {.card title="GLM Analysis"}', logs_qmd, fixed = TRUE)))
    expect_true(any(grepl("methylationGLM.txt", logs_qmd, fixed = TRUE)))
    expect_true(any(grepl('::: {.card title="LME Analysis"}', logs_qmd, fixed = TRUE)))
    expect_true(any(grepl("methylationLME.txt", logs_qmd, fixed = TRUE)))

    report_qmd <- readLines(file.path(result$projectDir, "report.qmd"), warn = FALSE)
    expect_true(any(grepl("The table has 2 rows and 3 columns.", report_qmd, fixed = TRUE)))
    expect_true(any(grepl(
        "Using person as the participant identifier",
        report_qmd,
        fixed = TRUE
    )))
    expect_false(any(grepl(
        "Using Sentrix_ID as the participant identifier",
        report_qmd,
        fixed = TRUE
    )))
    expect_false(any(grepl("The Data tab summarises", report_qmd, fixed = TRUE)))
    expect_false(any(grepl("The ENmix QC tab summarises", report_qmd, fixed = TRUE)))
    expect_false(any(grepl("The GLM Analysis tab summarises", report_qmd, fixed = TRUE)))
    expect_false(any(grepl("The LME Analysis tab summarises", report_qmd, fixed = TRUE)))
    expect_false(any(grepl("The Logs tab summarises", report_qmd, fixed = TRUE)))
    expect_false(any(grepl("Generated Outputs", report_qmd, fixed = TRUE)))

    if (identical(result$status, "rendered")) {
        expect_lt(file.info(file.path(result$projectDir, "docs", "index.html"))$size, 100000)
        expect_lt(file.info(file.path(result$projectDir, "docs", "lme.html"))$size, 500000)
        expect_lt(file.info(file.path(result$projectDir, "docs", "glm.html"))$size, 500000)
        data_html <- readLines(file.path(result$projectDir, "docs", "index.html"), warn = FALSE)
        glm_html <- readLines(file.path(result$projectDir, "docs", "glm.html"), warn = FALSE)
        lme_html <- readLines(file.path(result$projectDir, "docs", "lme.html"), warn = FALSE)
        expect_true(any(grepl('data-role="body"', data_html, fixed = TRUE)))
        expect_false(any(grepl("html-widget", data_html, fixed = TRUE)))
        expect_false(any(grepl("jquery.dataTables", data_html, fixed = TRUE)))
        expect_true(any(grepl('data-role="body"', glm_html, fixed = TRUE)))
        expect_true(any(grepl('data-role="body"', lme_html, fixed = TRUE)))
        expect_false(any(grepl("&lt;table class=", glm_html, fixed = TRUE)))
        expect_false(any(grepl("&lt;table class=", lme_html, fixed = TRUE)))
        expect_false(any(grepl("&lt;label&gt;Rows per page", glm_html, fixed = TRUE)))
        expect_false(any(grepl("&lt;label&gt;Rows per page", lme_html, fixed = TRUE)))
        expect_true(any(grepl(
            'Table 1. Generalised Linear Model Results',
            glm_html,
            fixed = TRUE
        )))
        expect_true(any(grepl(
            'Table 1. Linear Mixed-Effects Model Results',
            lme_html,
            fixed = TRUE
        )))
        expect_true(any(grepl(
            'Data Preview',
            data_html,
            fixed = TRUE
        )))
        expect_true(any(grepl("card-header", glm_html, fixed = TRUE)))
        expect_true(any(grepl("card-header", lme_html, fixed = TRUE)))
        expect_true(any(grepl("card-header", data_html, fixed = TRUE)))
    }

    printed <- utils::capture.output(print(result))
    expect_equal(
        printed,
        c(
            "Class type: dnaEPICO_dnamReport",
            paste("Log output path:", normalizePath(result$logFile, winslash = "/", mustWork = FALSE)),
            paste("Report output path:", normalizePath(result$outputFile, winslash = "/", mustWork = FALSE))
        )
    )
    expect_false(any(grepl("\\$preparedReport|\\$figureInventory|\\$renderResult", printed)))
})

test_that("dnamReport labels and describes nlme reports from workbook metadata", {
    tmp <- withr::local_tempdir()
    example_dirs <- create_dnam_report_example(tmp)
    workbook <- openxlsx::loadWorkbook(example_dirs$lmeTablePath)
    openxlsx::addWorksheet(workbook, "metadata")
    openxlsx::writeData(
        workbook,
        "metadata",
        data.frame(
            Key = c(
                "backend",
                "fitting_function",
                "libraries",
                "correlation_structure",
                "correlation_variable",
                "interaction_term"
            ),
            Value = c("nlme", "nlme::lme", "nlme", "AR1", "Timepoint", "None"),
            stringsAsFactors = FALSE
        )
    )
    openxlsx::saveWorkbook(workbook, example_dirs$lmeTablePath, overwrite = TRUE)
    dnaEPICO:::writeReportTableSidecarDnaEpico(
        tableData = openxlsx::read.xlsx(
            example_dirs$lmeTablePath,
            sheet = "annotatedLME",
            check.names = FALSE
        ),
        workbookFile = example_dirs$lmeTablePath,
        sidecarDir = file.path(tmp, "quarto", "assets", "results", "lme_results"),
        sheet = "annotatedLME",
        idColumn = "IlmnID",
        dictionary = openxlsx::read.xlsx(
            example_dirs$lmeTablePath,
            sheet = "dictionary",
            check.names = FALSE
        ),
        workbookMetadata = openxlsx::read.xlsx(
            example_dirs$lmeTablePath,
            sheet = "metadata",
            check.names = FALSE
        )
    )

    result <- dnamReport(
        outputDir = file.path(tmp, "quarto"),
        phenoTab = example_dirs$phenoTab,
        enmixTab = example_dirs$qcDir,
        qcTab = example_dirs$preprocessingDir,
        svaTab = example_dirs$svaDir,
        metricTab = example_dirs$postprocessingDir,
        glmTab = example_dirs$glmTablePath,
        lmeTab = example_dirs$lmeTablePath,
        verbose = FALSE,
        logs = FALSE,
        logTab = example_dirs$logDir
    )

    quarto_yml <- readLines(file.path(result$projectDir, "_quarto.yml"), warn = FALSE)
    lme_qmd <- readLines(file.path(result$projectDir, "lme.qmd"), warn = FALSE)
    logs_qmd <- readLines(file.path(result$projectDir, "logs.qmd"), warn = FALSE)
    expect_true(any(grepl('text: "nlme Analysis"', quarto_yml, fixed = TRUE)))
    expect_true(any(grepl('title: "nlme Analysis"', lme_qmd, fixed = TRUE)))
    expect_true(any(grepl("analysed using `nlme::lme()`", lme_qmd, fixed = TRUE)))
    expect_true(any(grepl("`AR1` residual correlation structure", lme_qmd, fixed = TRUE)))
    expect_true(any(grepl('class="dnaepico-model-formula"', lme_qmd, fixed = TRUE)))
    expect_false(any(grepl("~ `", lme_qmd, fixed = TRUE)))
    expect_true(any(grepl('::: {.card title="nlme Analysis"}', logs_qmd, fixed = TRUE)))
})

test_that("dnamReport presents multiple GLM, LME, and nlme formulas by phenotype", {
    tmp <- withr::local_tempdir()
    example_dirs <- create_dnam_report_example(tmp)
    write_multiple_formula_workbooks(example_dirs, backend = "lme4")

    lme_result <- dnamReport(
        outputDir = file.path(tmp, "quarto-lme"),
        phenoTab = example_dirs$phenoTab,
        enmixTab = example_dirs$qcDir,
        qcTab = example_dirs$preprocessingDir,
        svaTab = example_dirs$svaDir,
        metricTab = example_dirs$postprocessingDir,
        glmTab = example_dirs$glmTablePath,
        lmeTab = example_dirs$lmeTablePath,
        verbose = FALSE,
        logs = FALSE,
        logTab = example_dirs$logDir
    )
    glm_qmd <- paste(readLines(file.path(lme_result$projectDir, "glm.qmd"), warn = FALSE), collapse = "\n")
    lme_qmd <- paste(readLines(file.path(lme_result$projectDir, "lme.qmd"), warn = FALSE), collapse = "\n")
    report_qmd <- paste(readLines(file.path(lme_result$projectDir, "report.qmd"), warn = FALSE), collapse = "\n")

    for (page in list(glm_qmd, lme_qmd)) {
        expect_match(page, "3 phenotype-specific models were fitted\\.")
        expect_match(page, "View recorded model formulas \\(3\\)")
        expect_match(page, "<strong>T1D</strong>", fixed = TRUE)
        expect_match(page, "<strong>HbA1c</strong>", fixed = TRUE)
        expect_match(page, "<strong>C-peptide</strong>", fixed = TRUE)
        expect_false(grepl("The recorded model formula is", page, fixed = TRUE))
    }
    expect_match(glm_qmd, "GLM: M-values ~ T1D + Age + Sex + PC1 + PC2 + PC3", fixed = TRUE)
    expect_match(lme_qmd, "LME: M-values ~ T1D + Age + Sex + PC1 + PC2 + PC3 + (1 | SID)", fixed = TRUE)
    expect_false(grepl("View recorded model formulas", report_qmd, fixed = TRUE))
    expect_false(grepl("dnaepico-model-formula-item", report_qmd, fixed = TRUE))

    write_multiple_formula_workbooks(example_dirs, backend = "nlme")
    nlme_result <- dnamReport(
        outputDir = file.path(tmp, "quarto-nlme"),
        phenoTab = example_dirs$phenoTab,
        enmixTab = example_dirs$qcDir,
        qcTab = example_dirs$preprocessingDir,
        svaTab = example_dirs$svaDir,
        metricTab = example_dirs$postprocessingDir,
        glmTab = example_dirs$glmTablePath,
        lmeTab = example_dirs$lmeTablePath,
        verbose = FALSE,
        logs = FALSE,
        logTab = example_dirs$logDir
    )
    nlme_qmd <- paste(readLines(file.path(nlme_result$projectDir, "lme.qmd"), warn = FALSE), collapse = "\n")
    expect_match(nlme_qmd, 'title: "nlme Analysis"', fixed = TRUE)
    expect_match(nlme_qmd, "3 phenotype-specific models were fitted.", fixed = TRUE)
    expect_match(nlme_qmd, "nlme: M-values ~ T1D + Age + Sex + PC1 + PC2 + PC3 + (1 | SID)", fixed = TRUE)
})

test_that("dnamReport validates the detection P-value threshold before writing", {
    output_dir <- file.path(withr::local_tempdir(), "quarto")

    expect_error(
        dnamReport(
            outputDir = output_dir,
            detPThreshold = 2,
            verbose = FALSE,
            logs = FALSE
        ),
        "detPThreshold must be a single number between 0 and 1",
        fixed = TRUE
    )
    expect_false(dir.exists(output_dir))
})

test_that("sample identifiers are reordered explicitly and duplicates fail", {
    expect_identical(
        dnaEPICO:::matchSampleIdentifiersDnaEpico(
            query = c("S2", "S1"),
            reference = c("S1", "S2"),
            requireSameSet = TRUE
        ),
        c(2L, 1L)
    )

    expect_error(
        dnaEPICO:::matchSampleIdentifiersDnaEpico(
            query = c("S1", "S1"),
            reference = c("S1", "S2")
        ),
        "duplicate sample identifiers"
    )
    expect_error(
        dnaEPICO:::matchSampleIdentifiersDnaEpico(
            query = c("S1", "S3"),
            reference = c("S1", "S2")
        ),
        "S3"
    )
})

test_that("sex labels are canonicalized without changing reported values", {
    original <- c("Female", "M", "0", "1", "unknown", NA)
    result <- dnaEPICO:::canonicalizeSexDnaEpico(original)

    expect_identical(result$original, original)
    expect_identical(result$code, c(0L, 1L, 0L, 1L, NA_integer_, NA_integer_))
    expect_identical(result$label, c("F", "M", "F", "M", NA, NA))
    expect_identical(result$unknown, "unknown")
    expect_identical(formals(preprocessingMinfiEwasWater)$removeSexMismatch, FALSE)

    expect_identical(
        dnaEPICO:::resolveNormalizationSexDnaEpico(
            data.frame(
                Sex = c("F", "unknown", NA),
                PredSex = c(0L, 1L, 0L)
            ),
            "Sex"
        ),
        c("F", "M", "F")
    )
})

test_that("sex mismatch removal keeps RGSet and phenotype rows aligned", {
    rgset <- matrix(seq_len(12), nrow = 3)
    colnames(rgset) <- c("S3", "S1", "S2", "S4")
    targets <- data.frame(
        Sample_Name = c("S1", "S2", "S3", "S4"),
        Sex = c("F", "M", "F", "M"),
        stringsAsFactors = FALSE
    )

    result <- filterSamplesMinfiEwasWater(
        RGSet = rgset,
        targets = targets,
        failedSamples = c("S2", "S4"),
        verbose = FALSE,
        logs = FALSE
    )

    expect_identical(colnames(result$RGSet), c("S3", "S1"))
    expect_identical(result$targets$Sample_Name, c("S3", "S1"))
})

test_that("coefficient extraction follows formula terms exactly", {
    model_data <- data.frame(
        status = factor(rep(c("Control", "Case"), each = 4)),
        sex = factor(rep(c("F", "M"), times = 4))
    )
    formula_text <- "beta ~ status * sex"
    term_map <- dnaEPICO:::buildCoefficientTermMapMethylationModels(
        formulaText = formula_text,
        data = model_data
    )

    interaction_rows <- dnaEPICO:::findCoefficientRowsMethylationGLM(
        coefNames = names(term_map),
        variable = "status",
        interactionTerm = "sex",
        coefficientTerms = term_map
    )
    main_rows <- dnaEPICO:::findCoefficientRowsMethylationGLM(
        coefNames = names(term_map),
        variable = "status",
        coefficientTerms = term_map
    )

    expect_true(length(interaction_rows) > 0L)
    expect_true(all(term_map[interaction_rows] == "status:sex"))
    expect_true(length(main_rows) > 0L)
    expect_true(all(term_map[main_rows] == "status"))
    expect_false(any(interaction_rows %in% main_rows))
})

test_that("rank-deficient fixed-effect designs fail before CpG fitting", {
    model_data <- data.frame(x = 1:8, duplicate_x = 1:8)
    expect_error(
        dnaEPICO:::validateFixedEffectDesignMethylationModels(
            formulaText = "beta ~ x + duplicate_x",
            data = model_data
        ),
        "rank deficient"
    )
    expect_error(
        dnaEPICO:::validateFixedEffectDesignMethylationModels(
            formulaText = "beta ~ x",
            data = data.frame(x = c(1, Inf, 3))
        ),
        "non-finite numeric values"
    )
})

test_that("a phenotype cannot be used as its own interaction term", {
    expect_error(
        dnaEPICO:::buildFormulaMethylationGLM(
            phenotype = "status",
            interactionTerm = "status"
        ),
        "must differ"
    )
    expect_error(
        dnaEPICO:::buildFormulaMethylationLME(
            phenotype = "score",
            personVar = "person",
            interactionTerm = "score"
        ),
        "must differ"
    )
})

test_that("phenotype-only GLM and LME formulas contain no empty term", {
    expect_identical(
        dnaEPICO:::buildFormulaMethylationGLM(
            phenotype = "status",
            covariates = character(0)
        ),
        "beta ~ `status`"
    )
    expect_identical(
        dnaEPICO:::buildFormulaMethylationLME(
            phenotype = "score",
            personVar = "person",
            covariates = character(0)
        ),
        "beta ~ `score` + (1 | `person` )"
    )
})

test_that("longitudinal subject derivation accepts only explicit A/B visits", {
    valid <- data.frame(SID = c("P1A", "P1B", "P2A", "P2B"))
    derived <- dnaEPICO:::ensurePersonColumnMethylationLME(valid)
    expect_identical(derived$data$person, c("P1", "P1", "P2", "P2"))

    invalid <- data.frame(SID = c("P1_visit1", "P1_visit2", "P2_visit1"))
    expect_error(
        dnaEPICO:::ensurePersonColumnMethylationLME(invalid),
        "Supply an explicit subject identifier"
    )
})

test_that("probability and CpG-limit options reject invalid values", {
    expect_equal(
        dnaEPICO:::validateProbabilityDnaEpico(0.05, "threshold"),
        0.05
    )
    expect_error(
        dnaEPICO:::validateProbabilityDnaEpico(1.2, "threshold"),
        "between 0 and 1"
    )
    expect_error(
        dnaEPICO:::validateCpgLimitMethylationModels(2.5),
        "positive whole number"
    )
    expect_error(
        dnaEPICO:::validateCpgPrefixDnaEpico(""),
        "cpgPrefix"
    )
    expect_error(
        dnaEPICO:::normalizeOptionalNumericMethylationGLM(c(1, 2)),
        "one value"
    )
    expect_error(
        dnaEPICO:::normalizeOptionalNumericMethylationGLM("not-a-number"),
        "must be numeric"
    )
})

test_that("phenotype-to-PRS mappings are unambiguous", {
    expect_identical(
        dnaEPICO:::parsePrsMapMethylationGLM("T1D:PRS_T1D,HbA1c:PRS_HbA1c"),
        c(T1D = "PRS_T1D", HbA1c = "PRS_HbA1c")
    )
    expect_error(
        dnaEPICO:::parsePrsMapMethylationGLM("T1D:PRS1,T1D:PRS2"),
        "duplicate phenotype mappings"
    )
    expect_error(
        dnaEPICO:::parsePrsMapMethylationGLM("T1D:"),
        "format 'Phenotype:PRS'"
    )
    expect_error(
        dnaEPICO:::parsePrsMapMethylationGLM("T1D:PRS:extra"),
        "format 'Phenotype:PRS'"
    )
})

test_that("constant CpG responses are reported as failed GLM fits", {
    result <- dnaEPICO:::fitCpGModelMethylationGLM(
        cpg = "cg00000001",
        cpgValues = rep(0.5, 8),
        modelData = data.frame(status = rep(c(0, 1), each = 4)),
        formulaText = "beta ~ status"
    )

    expect_s3_class(result, "dnaEPICO_methylationGLM_fit_error")
    expect_match(result$error, "no observed variation")
})

test_that("constant CpG responses are reported before mixed-model fitting", {
    result <- dnaEPICO:::fitCpGModelMethylationLME(
        cpg = "cg00000001",
        cpgValues = rep(0.5, 8),
        modelData = data.frame(
            person = factor(rep(seq_len(4), each = 2)),
            score = rep(c(0, 1), 4)
        ),
        formulaText = "beta ~ score + (1 | person)",
        personVar = "person",
        lmeEngine = "lme4"
    )

    expect_s3_class(result, "dnaEPICO_methylationLME_fit_error")
    expect_match(result$error, "no observed variation")
})

test_that("external annotation identifiers cannot be implicit or duplicated", {
    expect_error(
        dnaEPICO:::coerceAnnotationDataMethylationGLM(data.frame(chr = "chr1")),
        "must include a CpG or IlmnID column"
    )
    expect_error(
        dnaEPICO:::coerceAnnotationDataMethylationGLM(data.frame(
            CpG = c("cg00000001", "cg00000001"),
            chr = c("chr1", "chr1")
        )),
        "duplicate CpG identifiers"
    )
})

test_that("finite summaries return NA instead of creating NaN", {
    expect_equal(dnaEPICO:::meanFiniteOrNADnaEpico(c(NA_real_, NaN, Inf)), NA_real_)
    expect_equal(dnaEPICO:::meanFiniteOrNADnaEpico(c(NA_real_, 0.2, 0.4)), 0.3)

    timepoint_summary <- summarizeTimepointsMethylationLME(
        data = data.frame(
            Timepoint = c("T1", "T1", "T2", "T2"),
            score = c(NA_real_, NaN, 1, Inf)
        ),
        timeVar = "Timepoint",
        phenotypes = "score",
        logs = FALSE
    )
    expect_true(is.na(timepoint_summary$score_mean[timepoint_summary$Timepoint == "T1"]))
    expect_false(any(is.nan(timepoint_summary$score_mean)))
    expect_identical(timepoint_summary$score_n, c(0L, 1L))
})

test_that("diagnostic means stay on the declared methylation scale", {
    prepared <- list(
        data = data.frame(
            cg_boundary = c(0, 0),
            cg_valid = c(0.2, 0.4),
            cg_invalid = c(Inf, NA_real_)
        ),
        cpgColumns = c("cg_boundary", "cg_valid", "cg_invalid"),
        methylationScale = "beta",
        responseLabel = "Beta values",
        invalidCpGs = data.frame(CpG = "cg_invalid")
    )
    result <- dnaEPICO:::diagnosticMeanMethylationModels(prepared)

    expect_identical(result$label, "Average Beta")
    expect_equal(unname(result$values[c("cg_boundary", "cg_valid")]), c(0, 0.3))
    expect_true(is.na(result$values[["cg_invalid"]]))
    expect_false(any(is.nan(result$values)))
})

test_that("saved plot devices close before an optional display draw", {
    tmp <- withr::local_tempdir()
    plot_file <- file.path(tmp, "device-test.tiff")
    draws <- 0L
    grDevices::pdf(file.path(tmp, "display-device.pdf"))
    on.exit(grDevices::dev.off(), add = TRUE)
    device_before <- grDevices::dev.cur()

    dnaEPICO:::runPlotMinfiEwasWater(
        draw_fun = function() {
            draws <<- draws + 1L
            graphics::plot.new()
            graphics::text(0.5, 0.5, "device test")
        },
        display = TRUE,
        file = plot_file,
        width = 400,
        height = 400,
        res = 72
    )

    expect_equal(draws, 2L)
    expect_true(file.exists(plot_file))
    expect_identical(grDevices::dev.cur(), device_before)
})

test_that("GLM filtering does not truncate diagnostics or annotated CpGs", {
    summary_cache <- data.frame(
        CpG = rep(c("cg1", "cg2"), each = 2L),
        Coefficient = rep(c("statusB", "statusC"), times = 2L),
        Estimate = c(0.2, -0.1, 0.05, 0.08),
        `Std. Error` = rep(0.05, 4L),
        `t value` = c(4, -2, 1, 1.6),
        `Pr(>|t|)` = c(0.001, 0.04, 0.5, 0.2),
        ResidualSD = rep(0.1, 4L),
        check.names = FALSE
    )
    prepared <- list(
        data = data.frame(cg1 = c(0.2, 0.3), cg2 = c(0.4, 0.5)),
        cpgColumns = c("cg1", "cg2"),
        methylationScale = "beta",
        responseLabel = "Beta values",
        interactionTerm = NULL,
        invalidCpGs = data.frame(CpG = character(0))
    )
    summaries <- summarizeMethylationGLMModels(
        modelResults = list(
            fits = list(status = list()),
            summaryCache = list(status = summary_cache),
            fitFailures = data.frame()
        ),
        preparedData = prepared,
        summaryPval = 0.05,
        logs = FALSE
    )

    expect_equal(nrow(summaries$summaries$status), 2L)
    expect_equal(nrow(summaries$diagnosticSummaries$status), 4L)

    diagnostics <- plotMethylationGLMDiagnostics(
        modelSummaries = summaries,
        preparedData = prepared,
        display = FALSE,
        logs = FALSE
    )
    expect_setequal(names(diagnostics$plots$status), c("statusB", "statusC"))
    expect_setequal(names(diagnostics$inflationFactors$status), c("statusB", "statusC"))

    annotation <- annotateMethylationGLMSummaries(
        modelSummaries = summaries,
        annotationObject = data.frame(CpG = c("cg1", "cg2"), chr = c("chr1", "chr2")),
        annotationCols = "chr",
        logs = FALSE
    )
    expect_setequal(annotation$data$IlmnID, c("cg1", "cg2"))
})

test_that("LME filtering does not truncate diagnostics or annotated CpGs", {
    summary_cache <- data.frame(
        CpG = rep(c("cg1", "cg2"), each = 2L),
        Interaction.Term = rep(c("score", "score:visit"), times = 2L),
        Estimate = c(0.2, -0.1, 0.05, 0.08),
        Std.Error = rep(0.05, 4L),
        t.value = c(4, -2, 1, 1.6),
        P.value = c(0.001, 0.04, 0.5, 0.2),
        stringsAsFactors = FALSE
    )
    prepared <- list(
        data = data.frame(cg1 = c(0.2, 0.3), cg2 = c(0.4, 0.5)),
        cpgColumns = c("cg1", "cg2"),
        methylationScale = "beta",
        responseLabel = "Beta values",
        interactionTerm = "visit",
        invalidCpGs = data.frame(CpG = character(0))
    )
    summaries <- summarizeMethylationLMEModels(
        modelResults = list(
            fits = list(score = list()),
            summaryCache = list(score = summary_cache),
            fitFailures = data.frame()
        ),
        preparedData = prepared,
        summaryPval = 0.05,
        logs = FALSE
    )

    expect_equal(nrow(summaries$summaries$score), 2L)
    expect_equal(nrow(summaries$diagnosticSummaries$score), 4L)

    diagnostics <- plotMethylationLMEDiagnostics(
        modelSummaries = summaries,
        preparedData = prepared,
        display = FALSE,
        logs = FALSE
    )
    expect_setequal(names(diagnostics$plots$score), c("score", "score:visit"))
    expect_setequal(names(diagnostics$inflationFactors$score), c("score", "score:visit"))

    annotation <- annotateMethylationLMESummaries(
        modelSummaries = summaries,
        annotationObject = data.frame(CpG = c("cg1", "cg2"), chr = c("chr1", "chr2")),
        annotationCols = "chr",
        logs = FALSE
    )
    expect_setequal(annotation$data$IlmnID, c("cg1", "cg2"))
})

## ----include = FALSE----------------------------------------------------------
startTime <- Sys.time()
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>",
    crop = NULL
)

library(dnaEPICO)

can_run_preprocessing <- requireNamespace("minfiData", quietly = TRUE) &&
    requireNamespace("IlluminaHumanMethylation450kmanifest", quietly = TRUE) &&
    requireNamespace("IlluminaHumanMethylation450kanno.ilmn12.hg19", quietly = TRUE)

can_run_sva <- can_run_preprocessing
can_run_glm <- requireNamespace(
    "IlluminaHumanMethylation450kanno.ilmn12.hg19",
    quietly = TRUE
)
can_run_lme <- can_run_glm && requireNamespace("lmerTest", quietly = TRUE)

## ----eval = FALSE-------------------------------------------------------------
# if (!requireNamespace("BiocManager", quietly = TRUE)) {
#     install.packages("BiocManager")
# }
# 
# BiocManager::install("dnaEPICO")
# 
# BiocManager::valid()

## ----eval = FALSE, message = FALSE--------------------------------------------
# library("dnaEPICO")

## ----makefile-example---------------------------------------------------------
makefilePath <- extractMake(destDir = tempdir(), overwrite = TRUE)
basename(makefilePath)

## ----preprocessing-pipeline-inputs, eval = can_run_preprocessing, message = FALSE, warning = FALSE----
preprocessing_inputs <- dnaEPICO:::exampleMinfiIdatInputsDnaEpico(n = 6L)

names(preprocessing_inputs)
preprocessing_inputs$tempDir
basename(preprocessing_inputs$phenoFile)
head(basename(list.files(preprocessing_inputs$idatFolder, full.names = TRUE)), 4)
basename(preprocessing_inputs$probeExclusionPath)

## ----preprocessing-pipeline, eval = can_run_preprocessing, message = FALSE, warning = FALSE----
preprocessPipelineResult <- preprocessingMinfiEwasWater(
    phenoFile = preprocessing_inputs$phenoFile,
    idatFolder = preprocessing_inputs$idatFolder,
    outputLogs = file.path(preprocessing_inputs$tempDir, "logs"),
    nSamples = 6,
    SampleID = "Sample_Name",
    arrayType = preprocessing_inputs$arrayType,
    annotationVersion = preprocessing_inputs$annotationVersion,
    scriptLabel = "preprocessingMinfiEwasWater",
    baseDataFolder = file.path(preprocessing_inputs$tempDir, "rData"),
    figureBaseDir = file.path(preprocessing_inputs$tempDir, "figures"),
    detPThreshold = 1,
    normMethods = "quantile",
    sexColumn = "Sex",
    removeSexMismatch = FALSE,
    pvalThreshold = 1,
    chrToRemove = "",
    snpsToRemove = "SBE",
    mafThreshold = 1,
    probeExclusionPath = preprocessing_inputs$probeExclusionPath,
    plotGroupVar = "Sex",
    lcRef = "saliva",
    phenoOrder = "Sample_Name;Sex;Basename;Sentrix_ID;Sentrix_Position",
    lcPhenoDir = file.path(
        preprocessing_inputs$tempDir,
        "data",
        "preprocessingMinfiEwasWater"
    ),
    display = FALSE,
    verbose = FALSE,
    logs = TRUE,
    saveOutputs = TRUE
)

preprocess_paths <- c(
    phenoLC = file.path(
        preprocessing_inputs$tempDir,
        "data",
        "preprocessingMinfiEwasWater",
        "phenoLC.csv"
    ),
    rgset = file.path(
        preprocessing_inputs$tempDir,
        "rData",
        "preprocessingMinfiEwasWater",
        "objects",
        "RGSet.RData"
    ),
    beta = file.path(
        preprocessing_inputs$tempDir,
        "rData",
        "preprocessingMinfiEwasWater",
        "metrics",
        "beta_NomFilt_MSetF_Flt_Rxy_Ds_Rc.RData"
    ),
    qc = file.path(
        preprocessing_inputs$tempDir,
        "figures",
        "preprocessingMinfiEwasWater",
        "qc",
        "quality_control(MSet).tiff"
    )
)

class(preprocessPipelineResult)
basename(preprocess_paths)
file.exists(preprocess_paths)

## ----sva-pipeline-inputs, eval = can_run_sva, message = FALSE, warning = FALSE----
sva_targets_file <- file.path(
    preprocessing_inputs$tempDir,
    "data",
    "preprocessingMinfiEwasWater",
    "phenoLC.csv"
)
sva_rgset_file <- file.path(
    preprocessing_inputs$tempDir,
    "rData",
    "preprocessingMinfiEwasWater",
    "objects",
    "RGSet.RData"
)

basename(c(sva_targets_file, sva_rgset_file))
file.exists(c(sva_targets_file, sva_rgset_file))

## ----sva-pipeline, eval = can_run_sva, message = FALSE, warning = FALSE-------
svaPipelineResult <- svaEnmix(
    phenoFile = sva_targets_file,
    rgsetData = sva_rgset_file,
    outputLogs = file.path(preprocessing_inputs$tempDir, "logs"),
    SampleID = "Sample_Name",
    arrayType = "IlluminaHumanMethylation450k",
    annotationVersion = "ilmn12.hg19",
    SentrixIDColumn = "Sentrix_ID",
    SentrixPositionColumn = "Sentrix_Position",
    ctrlSvaPercVar = 0.90,
    ctrlSvaFlag = 1,
    scriptLabel = "svaEnmix",
    dataBaseDir = file.path(preprocessing_inputs$tempDir, "data"),
    rBaseDir = file.path(preprocessing_inputs$tempDir, "rData"),
    figureBaseDir = file.path(preprocessing_inputs$tempDir, "figures"),
    display = FALSE,
    verbose = FALSE,
    logs = TRUE,
    saveOutputs = TRUE
)

class(svaPipelineResult$savedFiles)
names(svaPipelineResult$savedFiles)
basename(unlist(svaPipelineResult$savedFiles, use.names = FALSE))

## ----preprocessingPheno-pipeline-inputs, message = FALSE, warning = FALSE-----
pheno_inputs <- dnaEPICO:::examplePreprocessingPhenoStateDnaEpico()

names(pheno_inputs)
pheno_inputs$tempDir
basename(c(
    pheno_inputs$phenoPath,
    pheno_inputs$betaPath,
    pheno_inputs$mPath,
    pheno_inputs$cnPath
))

## ----preprocessingPheno-pipeline, message = FALSE, warning = FALSE------------
phenoPipelineResult <- preprocessingPheno(
    phenoFile = pheno_inputs$phenoPath,
    betaPath = pheno_inputs$betaPath,
    mPath = pheno_inputs$mPath,
    cnPath = pheno_inputs$cnPath,
    SampleID = "Sample_Name",
    timeVar = "Timepoint",
    timepoints = "1,2",
    combineTimepoints = "1,2",
    outputPheno = file.path(pheno_inputs$tempDir, "data", "preprocessingPheno"),
    outputRData = file.path(
        pheno_inputs$tempDir,
        "rData",
        "preprocessingPheno",
        "metrics"
    ),
    outputRDataMerge = file.path(
        pheno_inputs$tempDir,
        "rData",
        "preprocessingPheno",
        "mergeData"
    ),
    sexColumn = "Sex",
    outputLogs = file.path(pheno_inputs$tempDir, "logs"),
    outputDir = file.path(pheno_inputs$tempDir, "clockFoundation"),
    verbose = FALSE,
    logs = TRUE,
    saveOutputs = TRUE
)

names(phenoPipelineResult$savedFiles)
names(phenoPipelineResult$savedFiles$timepoints)
basename(unlist(phenoPipelineResult$savedFiles$timepoints[["1"]]))
basename(unlist(phenoPipelineResult$savedFiles[c(
    "combinedPheno",
    "combinedPhenoMethylation",
    "betaCSV",
    "phenoCF"
)]))

## ----glm-pipeline-inputs, eval = can_run_glm, message = FALSE, warning = FALSE----
glm_inputs <- dnaEPICO:::exampleMethylationGLMStateDnaEpico()

names(glm_inputs)
glm_inputs$tempDir
basename(glm_inputs$inputPath)
file.exists(glm_inputs$inputPath)

## ----glm-pipeline, eval = can_run_glm, message = FALSE, warning = FALSE-------
glmPipelineResult <- methylationGLM(
    inputPheno = glm_inputs$inputPath,
    phenotypes = "status",
    covariates = "sex,age",
    factorVars = "status,sex",
    scaleVars = "age",
    cpgLimit = 2,
    nCores = 1,
    outputLogs = file.path(glm_inputs$tempDir, "logs"),
    outputRData = file.path(glm_inputs$tempDir, "rData", "models"),
    outputPlots = file.path(glm_inputs$tempDir, "figures"),
    significantCpGDir = file.path(glm_inputs$tempDir, "significant"),
    summaryTxtDir = file.path(glm_inputs$tempDir, "summaries"),
    summaryPval = 1,
    significantCpGPval = 1,
    annotationPackage = "IlluminaHumanMethylation450kanno.ilmn12.hg19",
    annotationCols = "Name,chr,pos",
    annotatedGLMOut = file.path(glm_inputs$tempDir, "annotated"),
    reportAssetsDir = file.path(
        preprocessing_inputs$tempDir,
        "reports", "example", "assets", "results", "glm_results"
    ),
    display = FALSE,
    verbose = FALSE,
    logs = TRUE,
    saveOutputs = TRUE
)

class(glmPipelineResult$savedFiles)
names(glmPipelineResult$savedFiles)

## ----lme-pipeline-inputs, eval = can_run_lme, message = FALSE, warning = FALSE----
lme_inputs <- dnaEPICO:::exampleMethylationLMEStateDnaEpico()

names(lme_inputs)
lme_inputs$tempDir
basename(lme_inputs$inputPath)
file.exists(lme_inputs$inputPath)

## ----lme-pipeline, eval = can_run_lme, message = FALSE, warning = FALSE-------
lmePipelineResult <- methylationLME(
    inputPheno = lme_inputs$inputPath,
    outputLogs = file.path(lme_inputs$tempDir, "logs"),
    outputRData = file.path(lme_inputs$tempDir, "rData", "models"),
    outputPlots = file.path(lme_inputs$tempDir, "figures"),
    personVar = "person",
    timeVar = "Timepoint",
    phenotypes = "score",
    covariates = "sex",
    factorVars = "sex",
    scaleVars = "score",
    cpgLimit = 2,
    nCores = 1,
    summaryPval = 1,
    saveSignificantInteractions = TRUE,
    significantInteractionDir = file.path(
        lme_inputs$tempDir,
        "results",
        "cpgs",
        "methylationLME"
    ),
    significantInteractionPval = 1,
    saveTxtSummaries = TRUE,
    summaryTxtDir = file.path(
        lme_inputs$tempDir,
        "results",
        "summary",
        "methylationLME"
    ),
    annotationPackage = "IlluminaHumanMethylation450kanno.ilmn12.hg19",
    annotationCols = "Name,chr,pos",
    annotatedLMEOut = file.path(lme_inputs$tempDir, "annotated"),
    reportAssetsDir = file.path(
        preprocessing_inputs$tempDir,
        "reports", "example", "assets", "results", "lme_results"
    ),
    display = FALSE,
    verbose = FALSE,
    logs = TRUE,
    saveOutputs = TRUE
)

class(lmePipelineResult$savedFiles)
names(lmePipelineResult$savedFiles)

## ----quarto-check-r, eval = FALSE---------------------------------------------
# Sys.which("quarto")
# system2("quarto", "--version")

## ----quarto-path-r, eval = FALSE----------------------------------------------
# Sys.setenv(QUARTO_BIN = "/full/path/to/quarto")

## ----dnamReport-pipeline-example, eval = FALSE--------------------------------
# report_log_dir <- file.path(preprocessing_inputs$tempDir, "logs", "report")
# dir.create(report_log_dir, recursive = TRUE, showWarnings = FALSE)
# 
# report_logs <- c(
#     file.path(
#         preprocessing_inputs$tempDir,
#         "logs",
#         "log_preprocessingMinfiEwasWater.txt"
#     ),
#     file.path(pheno_inputs$tempDir, "logs", "log_preprocessingPheno.txt"),
#     file.path(preprocessing_inputs$tempDir, "logs", "log_svaEnmix.txt"),
#     file.path(glm_inputs$tempDir, "logs", "log_methylationGLM.txt"),
#     file.path(lme_inputs$tempDir, "logs", "log_methylationLME.txt")
# )
# 
# existing_logs <- report_logs[file.exists(report_logs)]
# if (length(existing_logs)) {
#     file.copy(existing_logs, report_log_dir, overwrite = TRUE)
# }
# 
# reportResult <- dnamReport(
#     outputDir = file.path(preprocessing_inputs$tempDir, "reports", "example"),
#     phenoTab = file.path(
#         preprocessing_inputs$tempDir,
#         "data",
#         "preprocessingMinfiEwasWater",
#         "phenoLC.csv"
#     ),
#     enmixTab = file.path(
#         preprocessing_inputs$tempDir,
#         "figures",
#         "preprocessingMinfiEwasWater",
#         "enmix"
#     ),
#     qcTab = file.path(
#         preprocessing_inputs$tempDir,
#         "figures",
#         "preprocessingMinfiEwasWater",
#         "qc"
#     ),
#     svaTab = file.path(preprocessing_inputs$tempDir, "figures", "svaEnmix"),
#     metricTab = file.path(
#         preprocessing_inputs$tempDir,
#         "figures",
#         "preprocessingMinfiEwasWater",
#         "metrics"
#     ),
#     glmTab = glmPipelineResult$savedFiles$annotatedGLM,
#     lmeTab = lmePipelineResult$savedFiles$annotatedLME,
#     logTab = report_log_dir,
#     detPPath = file.path(
#         preprocessing_inputs$tempDir,
#         "rData",
#         "preprocessingMinfiEwasWater",
#         "qc",
#         "detP_RGSet.RData"
#     ),
#     detPThreshold = 0.01,
#     verbose = FALSE,
#     logs = TRUE
# )
# 
# reportResult$outputFile

## ----makefile-template-full, echo = FALSE, results = "asis"-------------------
makefile_lines <- readLines(makefilePath, warn = FALSE)
cat("```make\n", paste(makefile_lines, collapse = "\n"), "\n```\n", sep = "")

## ----echo = FALSE-------------------------------------------------------------
Sys.time()

## ----echo = FALSE-------------------------------------------------------------
totalTime <- diff(c(startTime, Sys.time()))
round(totalTime, digits = 3)

## ----echo = FALSE-------------------------------------------------------------
sessionInfo()

#' Fit CpG-wise linear mixed-effects models for longitudinal methylation
#' analyses
#'
#' @param inputPheno Character. Path to the merged longitudinal
#'   phenotype-plus-methylation `.RData` or `.rds` object created by
#'   `preprocessingPheno()`. The default points to the combined timepoint object
#'   produced by the package workflow.
#' @param outputLogs Character. Directory used for optional log files.
#' @param outputRData Character. Directory used for optional serialized
#' mixed-model
#'   and summary outputs.
#' @param outputPlots Character. Directory used for optional TIFF diagnostic
#' plots.
#' @param personVar Character. Subject identifier variable used for the random
#'   intercept. When this column is missing, it is derived from `SID` using the
#'   package's existing sample naming convention.
#' @param timeVar Character. Name of the longitudinal time variable used for
#'   timepoint summaries and preprocessing checks.
#' @param phenotypes Character vector or comma-separated phenotype variables to
#'   model.
#' @param covariates Character. Comma-separated fixed-effect covariates included
#'   in every mixed model.
#' @param factorVars Character. Comma-separated variables that should be coerced
#'   to factors before modeling. This usually includes categorical phenotypes,
#'   covariates, or interaction variables.
#' @param scaleVars Character vector, comma-separated variable names, or `NULL`.
#' Numeric fixed-effect variables to centre and divide by their sample standard
#'   deviations before fitting.
#' @param lmeLibs Character. Comma-separated package names to validate on worker
#'   processes and select the LME backend. Use `'lme4,lmerTest'` or `'lme4'`
#'   for the `lmerTest`/`lme4` path, or `'nlme'` for the `nlme::lme()` path.
#' @param correlationStructure Character. Residual correlation structure used
#'   when `lmeLibs = 'nlme'`. One of `'none'`, `'AR1'`, or `'CAR1'`. The
#'   default is `'none'`.
#' @param correlationVar Character or `NULL`. Variable used to order repeated
#'   observations within `personVar` for `AR1` or `CAR1` residual correlation
#'   structures. Must be supplied explicitly for `AR1` or `CAR1`.
#' @param prsMap Character or `NULL`. Optional phenotype-to-PRS mapping in the
#'   form `'Phenotype1:PRS_1,Phenotype2:PRS_2'`.
#' @param libPath Character vector or `NULL`. Optional library paths forwarded
#' to
#'   worker processes. By default, the current `.libPaths()` are used.
#' @param cpgPrefix Character. Prefix used to identify methylation columns in
#' the
#'   merged phenotype-plus-methylation input object. The default is `'cg'`.
#' @param cpgLimit Integer or `NA`. Maximum number of CpGs to analyse. Use `NA`
#'   to keep all CpGs matching `cpgPrefix`.
#' @param methylationScale Character. Methylation metric represented by the CpG
#'   columns. One of `'Beta'`, `'M'`, or `'CN'`, in any combination of
#'   upper- and lower-case letters. The default is `'beta'`.
#' @param nCores Integer. Number of worker processes to use while fitting models
#'   and extracting summaries. Automatic fitting uses engine-specific empirical
#' crossovers and caps workers by the CpG workload and detected physical cores.
#' @param summaryPval Numeric or `NA`. Optional p-value threshold applied to the
#' returned longitudinal CpG summary tables. Use `NA` to keep all summary rows.
#' @param excludeSingular Logical. If `TRUE`, retain singular lme4 CpG rows and
#'   reasons but exclude their p-values from inference.
#' @param excludeNonConverged Logical. If `TRUE`, retain non-converged CpG rows
#'   and reasons but exclude their p-values from inference.
#' @param plotWidth Integer. TIFF width in pixels when plots are written to
#' disk.
#' @param plotHeight Integer. TIFF height in pixels when plots are written to
#' disk.
#' @param plotDPI Integer. TIFF resolution in DPI when plots are written to
#' disk.
#' @param interactionTerm Character or `NULL`. Optional interaction term. When
#'   supplied and present in the input data, the phenotype is modeled together
#'   with its interaction against this variable.
#' @param saveSignificantInteractions Logical. If `TRUE`, collect coefficient
#'   tables for CpGs passing `significantInteractionPval` in the returned object
#'   and optionally write them to disk when `saveOutputs = TRUE`.
#' @param significantInteractionDir Character. Directory used for optional
#'   significant-interaction coefficient tables.
#' @param significantInteractionPval Numeric. P-value threshold used to collect
#' or
#'   write significant interaction coefficient tables.
#' @param saveTxtSummaries Logical. If `TRUE` and `saveOutputs = TRUE`, write
#'   tab-delimited summary tables to `summaryTxtDir`.
#' @param chunkSize Integer or `NULL`. Number of CpGs processed per summary
#'   extraction chunk. `NULL` chooses a value automatically.
#' @param summaryTxtDir Character. Directory used for optional tab-delimited LME
#'   summary tables.
#' @param fdrThreshold Numeric. False-discovery-rate threshold used to highlight
#'   CpGs in the residual-significance diagnostic plots.
#' @param padjmethod Character. P-value adjustment method passed to
#'   `stats::p.adjust()`. The default is `'fdr'`.
#' @param annotationPackage Character. Annotation package or object name passed
#' to
#'   `minfi::getAnnotation()`, for example
#'   `'IlluminaHumanMethylationEPICv2anno.20a1.hg38'`.
#' @param annotationCols Character vector or comma-separated annotation columns
#'   to append to the combined LME summary table. Available columns depend on
#'   the selected annotation package.
#' @param annotatedLMEOut Character. Directory used for the optional annotated
#' LME
#'   summary XLSX workbook.
#' @param reportAssetsDir Character or `NULL`. Report results directory used for
#'   the compressed TSV table and its compact metadata sidecars. `NULL` writes
#'   only the model outputs and annotated workbook.
#' @param display Logical. If `TRUE`, draw diagnostic plots on the active
#'   graphics device.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' The
#'   default is `FALSE`, so the function is quiet unless requested.
#' @param logs Logical. If `TRUE`, write the same progress messages to
#'   `file.path(outputLogs, 'log_methylationLME.txt')`.
#' @param saveOutputs Logical. If `TRUE`, write optional serialized model files,
#'   summary tables, significant-interaction tables, annotated results, and TIFF
#'   plots to the requested output directories. The default is `FALSE`, so the
#'   function returns in-memory results without writing files.
#'
#' @return A list with class `'dnaEPICO_methylationLME'`.
#' \describe{
#'   \item{preparedData}{Object returned by [prepareMethylationLMEData()]
#' containing the merged longitudinal phenotype-plus-methylation analysis table
#' and
#'   modeling metadata.}
#'   \item{modelFits}{Object returned by [fitMethylationLMEModels()]
#'   containing the per-phenotype CpG mixed-effects model fits.}
#'   \item{modelSummaries}{Object returned by
#'   [summarizeMethylationLMEModels()] containing the combined CpG summary
#'   tables used for reporting and annotation.}
#'   \item{significantInteractions}{Object returned by
#'   [collectSignificantInteractionsMethylationLME()] containing optional
#'   phenotype-specific significant-interaction tables.}
#'   \item{diagnosticPlots}{Object returned by
#'   [plotMethylationLMEDiagnostics()] describing the diagnostic plot
#'   objects and any written TIFF files.}
#'   \item{annotation}{Object returned by
#'   [annotateMethylationLMESummaries()] containing the annotated combined
#'   summary table.}
#'   \item{savedFiles}{Object returned by
#'   [writeMethylationLMEOutputs()] when `saveOutputs = TRUE`, otherwise
#'   `NULL`.}
#'   \item{runSettings}{High-level run metadata including the generic analysis
#'   label, methylation scale, display label, selected merged-object prefix,
#'   internal response-column name, and longitudinal time variable.}
#' }
#' See [dnaEPICO_methylationLME-class] for a class-level overview.
#'
#' @description
#' `methylationLME()` is the high-level coordinator for the longitudinal
#' linear mixed-effects stage of the `dnaEPICO` workflow. It prepares the merged
#' phenotype-plus-methylation input, fits one mixed-effects model per CpG for
#' each
#' requested phenotype, extracts phenotype-specific coefficient summaries,
#' optionally collects significant interaction tables, generates diagnostic
#' plots,
#' annotates the combined summary table, and optionally writes model outputs to
#' disk. By default, the function runs quietly and returns its results in
#' memory.
#'
#' @examples
#' if (
#'     requireNamespace("IlluminaHumanMethylation450kanno.ilmn12.hg19", quietly = TRUE) &&
#'         requireNamespace("lmerTest", quietly = TRUE)
#' ) {
#'     tmp <- tempdir()
#'     toy_path <- file.path(tmp, "phenoBT1T2.RData")
#'     phenoBT1T2 <- data.frame(
#'         SID = c("P1A", "P1B", "P2A", "P2B", "P3A", "P3B", "P4A", "P4B"),
#'         person = c(1, 1, 2, 2, 3, 3, 4, 4),
#'         Timepoint = factor(c("1", "2", "1", "2", "1", "2", "1", "2")),
#'         score = c(10, 12, 9, 11, 13, 14, 8, 9),
#'         sex = factor(c("F", "F", "M", "M", "F", "F", "M", "M")),
#'         cg00000029 = c(0.25, 0.27, 0.20, 0.22, 0.30, 0.31, 0.18, 0.20),
#'         cg00000108 = c(0.50, 0.53, 0.55, 0.57, 0.48, 0.49, 0.60, 0.61),
#'         check.names = FALSE
#'     )
#'     save(phenoBT1T2, file = toy_path)
#'
#'     result <- methylationLME(
#'         inputPheno = toy_path,
#'         phenotypes = "score",
#'         covariates = "sex",
#'         factorVars = "sex",
#'         cpgLimit = 2,
#'         nCores = 1,
#'         summaryPval = 1,
#'         annotationPackage = "IlluminaHumanMethylation450kanno.ilmn12.hg19",
#'         annotationCols = "Name,chr,pos",
#'         display = FALSE,
#'         verbose = FALSE,
#'         logs = FALSE,
#'         saveOutputs = FALSE
#'     )
#'
#'     class(result)
#' }
#'
#' @seealso [dnaEPICO_methylationLME-class]
#'
#' @export
methylationLME <- function(
    inputPheno = "rData/preprocessingPheno/mergeData/phenoBT1T2.RData",
    outputLogs = "logs", outputRData = "rData/methylationLME/models",
    outputPlots = "figures/methylationLME", personVar = "person",
    timeVar = "Timepoint", phenotypes = c(
        "DASS_Depression",
        "DASS_Anxiety", "DASS_Stress", "PCL5_TotalScore", "MHCSF_TotalScore",
        "BRS_TotalScore"
    ), covariates = "Sex,Age,Ethnicity,TraumaDefinition,Leukocytes,Epithelial.cells",
    factorVars = "Sex,Ethnicity,TraumaDefinition", scaleVars = NULL,
    lmeLibs = "lme4,lmerTest", correlationStructure = "none",
    correlationVar = NULL, prsMap = NULL, libPath = NULL, cpgPrefix = "cg",
    cpgLimit = NA, methylationScale = "beta", nCores = 32, summaryPval = NA,
    excludeSingular = FALSE, excludeNonConverged = FALSE, plotWidth = 2000,
    plotHeight = 1000, plotDPI = 150, interactionTerm = NULL,
    saveSignificantInteractions = TRUE,
        significantInteractionDir = "preliminaryResults/cpgs/methylationLME",
    significantInteractionPval = 0.05, saveTxtSummaries = TRUE,
    chunkSize = NULL,
        summaryTxtDir = "preliminaryResults/summary/methylationLME",
    fdrThreshold = 0.05, padjmethod = "fdr",
        annotationPackage = "IlluminaHumanMethylationEPICv2anno.20a1.hg38",
    annotationCols = c(
        "Name", "chr", "pos", "UCSC_RefGene_Group",
        "UCSC_RefGene_Name", "Relation_to_Island", "GencodeV41_Group"
    ),
    annotatedLMEOut = "data/methylationLME", reportAssetsDir = NULL,
    display = FALSE,
    verbose = FALSE, logs = FALSE, saveOutputs = FALSE
) {
    cpgLimit <- normalizeOptionalNumericMethylationGLM(cpgLimit)
    summaryPval <- normalizeOptionalNumericMethylationGLM(summaryPval)
    chunkSize <- normalizeChunkSizeMethylationGLM(chunkSize)
    correlationStructure <-
        normalizeCorrelationStructureMethylationLME(correlationStructure)
    correlationVar <-
        normalizeCorrelationVariableMethylationLME(correlationVar = correlationVar)
    if (!identical(correlationStructure, "none")) {
        if (is.null(correlationVar)) {
            stop("correlationVar must be supplied when correlationStructure is AR1 or CAR1.",
                call. = FALSE
            )
        }
    }
    methylationScale <- normalizeMethylationScaleDnaEpico(methylationScale)
    methylationLabel <- methylationScaleResponseLabelDnaEpico(methylationScale)
    methylationObjectPrefix <-
        methylationScaleObjectPrefixDnaEpico(methylationScale)
    responseColumn <- methylationScaleResponseColumnDnaEpico(methylationScale)
    log_file <- "log_methylationLME.txt"

    if (is.null(libPath)) {
        libPath <- .libPaths()
    }

    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = outputLogs,
        log_file = log_file
    )

    emitLogMinfiEwasWater(
        c(
            "==== Starting DNAm LME Analysis ====",
            paste("Start time:                     ", format(Sys.time())),
            paste("Input phenotype + methylation:  ", inputPheno),
            paste(
                "Merged modeling object:         ", methylationObjectPrefix,
                "*"
            ), paste("Output RData folder:            ", outputRData),
            paste("Output logs folder:             ", outputLogs),
            paste("Output plots folder:            ", outputPlots),
            paste("Report assets folder:           ",
                if (is.null(reportAssetsDir)) "None" else reportAssetsDir),
            paste("Person ID variable:             ", personVar),
            paste("Time variable:                  ", timeVar), paste(
                "Phenotypes:                     ",
                phenotypes
            ), paste(
                "Covariates:                     ",
                covariates
            ), paste(
                "Factor variables:               ",
                factorVars
            ), paste(
                "Scale variables:                ",
                if (length(normalizeScaleVariablesDnaEpico(scaleVars))) {
                    paste(normalizeScaleVariablesDnaEpico(scaleVars),
                        collapse = ","
                    )
                } else {
                    "None"
                }
            ), paste("LME libraries:                  ", lmeLibs),
            paste("Correlation structure:          ", correlationStructure),
            paste("Correlation variable:           ", if (identical(
                correlationStructure,
                "none"
            )) {
                "None"
            } else {
                correlationVar
            }), paste(
                "PRS mapping:                    ",
                if (is.null(prsMap)) "None" else prsMap
            ), paste(
                "CpG column prefix:              ",
                cpgPrefix
            ), paste(
                "CpG limit:                      ",
                if (is.na(cpgLimit)) "All" else cpgLimit
            ), paste(
                "Number of cores:                ",
                as.integer(nCores)
            ), paste(
                "Summary p-value filter:         ",
                if (is.na(summaryPval)) "None" else summaryPval
            ),
            paste("Exclude singular fits:          ", isTRUE(excludeSingular)),
            paste("Exclude non-converged fits:     ",
                isTRUE(excludeNonConverged)),
            paste("Interaction term:               ",
                if (is.null(interactionTerm)) "None" else interactionTerm),
            paste("Save significant interactions:  ",
                isTRUE(saveSignificantInteractions)),
            paste("Significant interaction p-value:",
                significantInteractionPval),
            paste("Save text summaries:            ", isTRUE(saveTxtSummaries)),
            paste("Chunk size:                     ",
                if (is.null(chunkSize)) "Auto" else chunkSize),
            paste("FDR threshold:                  ", fdrThreshold),
            paste("P-value adjustment method:      ", padjmethod),
            paste("Display plots:                  ", isTRUE(display)),
            paste("Save outputs:                   ", isTRUE(saveOutputs)),
            "============================================================"
        ),
        verbose = verbose, log_path = log_path
    )

    withLoggedErrorsMinfiEwasWater(expr = {
        preparedData <- prepareMethylationLMEData(
            inputPheno = inputPheno,
            personVar = personVar, timeVar = timeVar, phenotypes = phenotypes,
            covariates = covariates, factorVars = factorVars,
            scaleVars = scaleVars, prsMap = prsMap, cpgPrefix = cpgPrefix,
            cpgLimit = cpgLimit, methylationScale = methylationScale,
            interactionTerm = interactionTerm, verbose = verbose,
            logs = logs, log_dir = outputLogs, log_file = log_file
        )

        modelFits <- fitMethylationLMEModels(
            preparedData = preparedData,
            nCores = nCores, libPath = libPath, lmeLibs = lmeLibs,
            correlationStructure = correlationStructure,
                correlationVar = correlationVar,
            verbose = verbose, logs = logs, log_dir = outputLogs,
            log_file = log_file
        )

        modelSummaries <- summarizeMethylationLMEModels(
            modelResults = modelFits,
            preparedData = preparedData, summaryPval = summaryPval,
            excludeSingular = excludeSingular,
                excludeNonConverged = excludeNonConverged,
            nCores = nCores, chunkSize = chunkSize, verbose = verbose,
            logs = logs, log_dir = outputLogs, log_file = log_file
        )

        significantInteractions <- NULL
        if (isTRUE(saveSignificantInteractions)) {
            significantInteractions <-
                collectSignificantInteractionsMethylationLME(
                modelResults = modelFits,
                pvalThreshold = significantInteractionPval,
                    interactionTerm = preparedData$interactionTerm,
                excludeSingular = excludeSingular,
                    excludeNonConverged = excludeNonConverged,
                verbose = verbose, logs = logs, log_dir = outputLogs,
                log_file = log_file
            )
        }

        diagnosticPlots <- plotMethylationLMEDiagnostics(
            modelSummaries = modelSummaries,
            preparedData = preparedData, fdrThreshold = fdrThreshold,
            padjmethod = padjmethod, outputDir = if (isTRUE(saveOutputs)) {
                outputPlots
            } else {
                NULL
            }, plotWidth = plotWidth, plotHeight = plotHeight,
            plotDPI = plotDPI, display = display, verbose = verbose,
            logs = logs, log_dir = outputLogs, log_file = log_file
        )

        annotation <- annotateMethylationLMESummaries(
            modelSummaries = modelSummaries,
            annotationObject = annotationPackage,
                annotationCols = annotationCols,
            verbose = verbose, logs = logs, log_dir = outputLogs,
            log_file = log_file
        )

        savedFiles <- NULL
        if (isTRUE(saveOutputs)) {
            savedFiles <- writeMethylationLMEOutputs(
                modelResults = modelFits,
                modelSummaries = modelSummaries, annotatedResults = annotation,
                significantInteractions = significantInteractions,
                outputRData = outputRData, summaryTxtDir = summaryTxtDir,
                significantInteractionDir = significantInteractionDir,
                annotatedLMEOut = annotatedLMEOut,
                    reportAssetsDir = reportAssetsDir,
                    saveTxtSummaries = saveTxtSummaries,
                saveSignificantInteractions = saveSignificantInteractions,
                verbose = verbose, logs = logs, log_dir = outputLogs,
                log_file = log_file
            )
        }

        emitLogMinfiEwasWater(
            c(
                "============================================================",
                paste("Finished DNAm LME Analysis:", format(Sys.time())),
                if (isTRUE(saveOutputs)) {
                    paste("Annotated LME output:          ",
                        savedFiles$annotatedLME)
                } else {
                    "Outputs were returned in memory only."
                }, "============================================================"
            ),
            verbose = verbose, log_path = log_path
        )

        structure(list(
            preparedData = preparedData, modelFits = modelFits,
            modelSummaries = modelSummaries,
                significantInteractions = significantInteractions,
            diagnosticPlots = diagnosticPlots, annotation = annotation,
            savedFiles = savedFiles, runSettings = list(
                analysisLabel = "methylationLME",
                methylationScale = methylationScale,
                    methylationLabel = methylationLabel,
                methylationObjectPrefix = methylationObjectPrefix,
                internalResponseColumn = responseColumn,
                    scaleVars = preparedData$scaleVars,
                scalingMetadata = preparedData$scalingMetadata,
                timeVar = timeVar, excludeSingular = isTRUE(excludeSingular),
                excludeNonConverged = isTRUE(excludeNonConverged),
                reportAssetsDir = reportAssetsDir,
                correlationStructure = correlationStructure,
                correlationVar = if (identical(
                    correlationStructure,
                    "none"
                )) {
                    NULL
                } else {
                    correlationVar
                }
            )
        ), class = "dnaEPICO_methylationLME")
    }, log_path = log_path, verbose = verbose, context = "methylationLME")
}

# ================================================================
# dnaEPICO - Global variable declarations for NSE compatibility
# ================================================================
#
# Purpose:
# Declare variables used via Non-Standard Evaluation (NSE) so that
# R CMD check does not produce "no visible binding for global variable"
#
# This file does not modify runtime behaviour. It only informs the
# static code analyser about variables used inside:
#   - data.table operations
#   - modelling outputs
#   - annotation tables
#
# ================================================================

utils::globalVariables(c(
    # ---------- Illumina annotation / manifest ----------
    "IlmnID", "Name", "AddressA_ID", "AddressB_ID", "Next_Base",
    "Color_Channel", "Probe_Type", "Infinium_Design_Type",
    "CHR", "MAPINFO", "Rep_Num",
    "CHR_37", "MAPINFO_37",
    "CHR_hg38", "Start_hg38", "Strand_hg38",

    # ---------- Chromosome / probe mapping ----------
    "chr37", "chr38", "probe_id",
    "probe_design", "channel", "probe_type", "next_base",

    # ---------- DNAm metrics ----------
    "m", "cn",

    # ---------- Cell composition / estimateLC ----------
    "CpG", "FDR", "P.value", "Std.Error",

    # ---------- GLM / LME outputs ----------
    "Pr(>|t|)", "ResidualSD",
    "value", "expected", "observed", "meanMethylation", "diagnosticP",
        "diagnosticY",

    # ---------- Pipeline variables ----------
    "idatFolder", "rBaseDir"
))

#' Copy dnaEPICO Makefile to a user directory
#'
#' Copies the example Makefile pipeline shipped with dnaEPICO to a
#' user-specified directory for local execution or modification.
#'
#' @param destDir Character. Destination directory where the Makefile
#'   will be copied.
#' @param overwrite Logical. Whether to overwrite an existing `Makefile` in
#'   `destDir`. The default is `FALSE`.
#'
#' @return Character scalar containing the path to the copied `Makefile`.
#'
#' @examples
#' tmp <- file.path(tempdir(), "dnaEPICO-make-example")
#' dir.create(tmp, recursive = TRUE, showWarnings = FALSE)
#' makefile_path <- extractMake(
#'     destDir = tmp,
#'     overwrite = TRUE
#' )
#' stopifnot(file.exists(makefile_path))
#'
#' @export
extractMake <- function(destDir, overwrite = FALSE) {
    if (!is.character(destDir) || length(destDir) != 1L || !nzchar(destDir)) {
        stop("destDir must be a single, non-empty character path.",
            call. = FALSE
        )
    }

    if (!is.logical(overwrite) || length(overwrite) != 1L ||
        is.na(overwrite)) {
        stop("overwrite must be a single TRUE or FALSE value.",
            call. = FALSE
        )
    }

    if (!dir.exists(destDir)) {
        stop("Destination directory does not exist: ", destDir,
            call. = FALSE
        )
    }

    makefileSrc <- system.file("extdata", "make", "Makefile.model.pipeline",
        package = "dnaEPICO", mustWork = TRUE
    )

    makefileDest <- file.path(destDir, "Makefile")

    copied <- file.copy(makefileSrc, makefileDest, overwrite = overwrite)

    if (!isTRUE(copied)) {
        stop("Failed to copy the packaged Makefile to: ", makefileDest,
            call. = FALSE
        )
    }

    normalizePath(makefileDest, winslash = "/", mustWork = FALSE)
}

test_that("extractMake copies the packaged Makefile and returns its path", {
    tmp <- withr::local_tempdir()

    expect_error(
        makefile_path <- extractMake(
            destDir = tmp,
            overwrite = TRUE
        ),
        NA
    )

    expect_true(is.character(makefile_path))
    expect_length(makefile_path, 1)
    expect_true(file.exists(makefile_path))
    expect_match(basename(makefile_path), "^Makefile$")
    makefile <- readLines(makefile_path, warn = FALSE)
    expect_true(any(grepl("SCALE_VARS = NULL", makefile, fixed = TRUE)))
    expect_true(any(grepl(
        "REMOVE_SEX_MISMATCH = FALSE",
        makefile,
        fixed = TRUE
    )))
    expect_true(any(grepl("Makefile.rules.pipeline", makefile, fixed = TRUE)))
    expect_true(any(grepl(
        "DNAPIPE_MK := $(subst $(DNAPIPE_SPACE),\\ ,$(DNAPIPE_MK_RAW))",
        makefile,
        fixed = TRUE
    )))

    rules_file <- system.file(
        "extdata", "make", "Makefile.rules.pipeline",
        package = "dnaEPICO",
        mustWork = TRUE
    )
    rules <- readLines(rules_file, warn = FALSE)
    expect_true(any(grepl(
        "STEP1_PRIMARY = $(RDATA_DIR)/$(MODEL)/$(STEP1)/$(OBJ_DIR)/RGSet.RData",
        rules,
        fixed = TRUE
    )))
    expect_true(any(grepl("$(STEP1_DERIVED): $(STEP1_PRIMARY)", rules, fixed = TRUE)))
    expect_true(any(grepl(
        "STEP3_PRIMARY = $(COMBINED_PHENO_METHYLATION)",
        rules,
        fixed = TRUE
    )))
    expect_true(any(grepl("$(STEP3_DERIVED): $(STEP3_PRIMARY)", rules, fixed = TRUE)))
    expect_true(any(grepl(
        "mkdir -p \"$(STEP6)/$(MODEL)\" \"$(MODEL_LOG_DIR)\"",
        rules,
        fixed = TRUE
    )))
    expect_true(any(grepl(
        "identical(result[['status']], 'rendered')",
        rules,
        fixed = TRUE
    )))
    expect_true(any(grepl(
        "file.exists(result[['outputFile']])",
        rules,
        fixed = TRUE
    )))
    expect_false(any(grepl("result$$status", rules, fixed = TRUE)))
    expect_true(any(grepl(
        "PHENO_LC = $(DATA_DIR)/$(MODEL)/$(STEP1)/phenoLC.csv",
        rules,
        fixed = TRUE
    )))
    expect_true(any(grepl(
        "SVA_MARKER = $(DATA_DIR)/$(MODEL)/$(STEP2)/.sva_complete",
        rules,
        fixed = TRUE
    )))
    expect_true(any(grepl(
        "$(SVA_MARKER): $(PHENO_LC)",
        rules,
        fixed = TRUE
    )))
    expect_true(any(grepl(
        "if (file.exists('$(SVA_MARKER)')) unlink('$(SVA_MARKER)')",
        rules,
        fixed = TRUE
    )))
    expect_true(any(grepl(
        "$(STEP3_PRIMARY): $(SVA_MARKER) $(PHENO_LC)",
        rules,
        fixed = TRUE
    )))
    expect_true(any(grepl(
        "COVARIATES_ARG = $(call optional_text_arg,$(COVARIATES))",
        rules,
        fixed = TRUE
    )))
    expect_true(any(grepl(
        "covariates = $(COVARIATES_ARG)",
        rules,
        fixed = TRUE
    )))
    expect_true(any(grepl(
        "factorVars = $(FACTOR_VARS_ARG)",
        rules,
        fixed = TRUE
    )))
    expect_true(any(grepl(
        "scaleVars = $(SCALE_VARS_GLM_ARG)",
        rules,
        fixed = TRUE
    )))
    expect_true(any(grepl(
        "scaleVars = $(SCALE_VARS_LME_ARG)",
        rules,
        fixed = TRUE
    )))
    expect_true(any(grepl(
        "$(GLM_OUTPUTS) &: $(GLM_PHENO_METHYLATION)",
        rules,
        fixed = TRUE
    )))
    expect_true(any(grepl(
        "$(LME_OUTPUTS) &: $(COMBINED_PHENO_METHYLATION)",
        rules,
        fixed = TRUE
    )))
    expect_true(any(grepl(
        "reportAssetsDir = '$(GLM_REPORT_ASSET_DIR)'",
        rules,
        fixed = TRUE
    )))
    expect_true(any(grepl(
        "reportAssetsDir = '$(LME_REPORT_ASSET_DIR)'",
        rules,
        fixed = TRUE
    )))
    expect_false(any(grepl("phenoLC_withSVA", rules, fixed = TRUE)))
    expect_false(any(grepl("outputPhenoFile", rules, fixed = TRUE)))
    expect_false(any(grepl("PRE_SVA_PHENO", rules, fixed = TRUE)))
    expect_false(any(grepl("FINAL_PHENO", rules, fixed = TRUE)))
})

#' Fit CpG-wise GLMs for one-timepoint methylation analyses
#'
#' @param inputPheno Character. Path to the merged phenotype-plus-methylation
#' `.RData`
#'   or `.rds` object created by `preprocessingPheno()`. The default points to
#'   the timepoint-1 object produced by the package workflow.
#' @param outputLogs Character. Directory used for optional log files.
#' @param outputRData Character. Directory used for optional serialized model
#' and
#'   summary outputs.
#' @param outputPlots Character. Directory used for optional TIFF plots.
#' @param phenotypes Character vector or comma-separated phenotype variables to
#'   model.
#' @param covariates Character. Comma-separated covariate variables included in
#'   each GLM.
#' @param factorVars Character. Comma-separated variables that should be treated
#'   as factors before modeling.
#' @param scaleVars Character vector, comma-separated variable names, or `NULL`.
#' Numeric fixed-effect variables to centre and divide by their sample standard
#'   deviations before model fitting.
#' @param cpgPrefix Character. Prefix used to identify methylation columns in
#' the
#'   merged phenotype-plus-methylation input object. The default is `'cg'`.
#' @param cpgLimit Integer or `NA`. Maximum number of CpGs to analyse. Use `NA`
#'   to keep all CpGs matching `cpgPrefix`.
#' @param methylationScale Character. Methylation metric represented by the CpG
#'   columns. One of `'Beta'`, `'M'`, or `'CN'`, in any combination of
#'   upper- and lower-case letters. The default is `'beta'`.
#' @param nCores Integer. Number of worker processes to use while fitting models
#'   and extracting summaries. Automatic fitting remains serial below the
#'   empirical glm2 crossover and caps workers by the CpG workload and detected
#'   physical cores.
#' @param plotWidth Integer. TIFF width in pixels when plots are written to
#' disk.
#' @param plotHeight Integer. TIFF height in pixels when plots are written to
#'   disk.
#' @param plotDPI Integer. TIFF resolution in DPI when plots are written to
#'   disk.
#' @param interactionTerm Character or `NULL`. Optional interaction term. When
#'   supplied and present in the input data, the phenotype is modeled together
#'   with its interaction against this variable.
#' @param libPath Character vector or `NULL`. Optional library paths forwarded
#' to
#'   worker processes. By default, the current `.libPaths()` are used.
#' @param glmLibs Character. Comma-separated package names to validate on worker
#'   processes. The default is `'glm2'`.
#' @param prsMap Character or `NULL`. Optional phenotype-to-PRS mapping in the
#'   form `'Phenotype1:PRS_1,Phenotype2:PRS_2'`.
#' @param summaryPval Numeric or `NA`. Optional p-value threshold applied to the
#'   returned CpG summary tables. Use `NA` to keep all summary rows.
#' @param summaryResidualSD Logical. If `TRUE`, append residual standard
#'   deviations to the CpG summary tables and residual diagnostic plots.
#' @param excludeNonConverged Logical. If `TRUE`, retain non-converged CpG rows
#'   and reasons but exclude their p-values from inference.
#' @param saveSignificantCpGs Logical. If `TRUE`, collect significant CpG
#'   coefficient tables in the returned object and optionally write them to disk
#'   when `saveOutputs = TRUE`.
#' @param significantCpGDir Character. Directory used for optional significant
#'   CpG coefficient tables.
#' @param significantCpGPval Numeric. P-value threshold used to collect or write
#'   significant CpG coefficient tables.
#' @param saveTxtSummaries Logical. If `TRUE` and `saveOutputs = TRUE`, write
#'   tab-delimited summary tables to `summaryTxtDir`.
#' @param chunkSize Integer or `NULL`. Number of CpGs processed per summary
#'   extraction chunk. `NULL` chooses a value automatically.
#' @param summaryTxtDir Character. Directory used for optional tab-delimited GLM
#'   summary tables.
#' @param fdrThreshold Numeric. False-discovery-rate threshold used to highlight
#'   CpGs in the residual-significance diagnostic plots.
#' @param padjmethod Character. P-value adjustment method passed to
#'   `stats::p.adjust()`. The default is `'fdr'`.
#' @param annotationPackage Character. Annotation package or object name passed
#'   to `minfi::getAnnotation()`, for example
#'   `'IlluminaHumanMethylationEPICv2anno.20a1.hg38'`.
#' @param annotationCols Character vector or comma-separated annotation columns
#'   to append to the combined GLM summary table. Available columns depend on
#'   the selected annotation package.
#' @param annotatedGLMOut Character. Directory used for the optional annotated
#'   GLM summary XLSX workbook.
#' @param reportAssetsDir Character or `NULL`. Report results directory used for
#'   the compressed TSV table and its compact metadata sidecars. `NULL` writes
#'   only the model outputs and annotated workbook.
#' @param display Logical. If `TRUE`, draw exploratory and diagnostic plots on
#'   the active graphics device.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#'   The default is `FALSE`, so the function is quiet unless requested.
#' @param logs Logical. If `TRUE`, write the same progress messages to
#'   `file.path(outputLogs, 'log_methylationGLM.txt')`.
#' @param saveOutputs Logical. If `TRUE`, write optional serialized model files,
#' summary tables, significant-CpG tables, annotated results, and TIFF plots to
#'   the requested output directories. The default is `FALSE`, so the function
#'   returns in-memory results without writing files.
#'
#' @return A list with class `'dnaEPICO_methylationGLM'`.
#' \describe{
#'   \item{preparedData}{Object returned by [prepareMethylationGLMData()]
#' containing the merged phenotype-plus-methylation analysis table and modeling
#'   metadata.}
#'   \item{distributionPlots}{Object returned by
#'   [plotMethylationGLMDistributions()] describing any exploratory plots that
#'   were generated or written.}
#'   \item{modelFits}{Object returned by [fitMethylationGLMModels()]
#'   containing the per-phenotype CpG model fits.}
#'   \item{modelSummaries}{Object returned by
#'   [summarizeMethylationGLMModels()] containing the combined CpG summary
#'   tables used for reporting and annotation.}
#'   \item{significantCpGs}{Object returned by
#'   [collectSignificantCpGsMethylationGLM()] containing optional
#'   phenotype-specific significant-CpG tables.}
#'   \item{diagnosticPlots}{Object returned by
#'   [plotMethylationGLMDiagnostics()] describing the diagnostic plot objects
#'   and any written TIFF files.}
#'   \item{annotation}{Object returned by
#'   [annotateMethylationGLMSummaries()] containing the annotated combined
#'   summary table.}
#'   \item{savedFiles}{Object returned by [writeMethylationGLMOutputs()] when
#'   `saveOutputs = TRUE`, otherwise `NULL`.}
#'   \item{runSettings}{High-level run metadata including the generic analysis
#'   label, methylation scale, display label, selected merged-object prefix, and
#'   internal response-column name.}
#' }
#' See [dnaEPICO_methylationGLM-class] for a class-level overview.
#'
#' @description
#' `methylationGLM()` is the high-level coordinator for the methylation GLM
#' stage of the `dnaEPICO` workflow. It prepares the merged
#' phenotype-plus-methylation
#' input, optionally creates exploratory plots, fits one Gaussian GLM per CpG
#' for
#' each requested phenotype, extracts CpG-level summaries, optionally collects
#' significant CpG coefficient tables, generates diagnostic plots, annotates the
#' combined summary table, and optionally writes model outputs to disk. By
#' default, the function runs quietly and returns its results in memory.
#'
#' @examples
#' if (requireNamespace("IlluminaHumanMethylation450kanno.ilmn12.hg19", quietly = TRUE)) {
#'     tmp <- tempdir()
#'     toy_path <- file.path(tmp, "phenoBT1.RData")
#'     phenoBT1 <- data.frame(
#'         Sample_Name = c("S1", "S2", "S3", "S4"),
#'         status = factor(c("Case", "Case", "Control", "Control")),
#'         sex = factor(c("F", "M", "F", "M")),
#'         cg00000029 = c(0.20, 0.25, 0.22, 0.27),
#'         cg00000108 = c(0.60, 0.55, 0.52, 0.58),
#'         check.names = FALSE
#'     )
#'     save(phenoBT1, file = toy_path)
#'
#'     result <- methylationGLM(
#'         inputPheno = toy_path,
#'         phenotypes = "status",
#'         covariates = "sex",
#'         factorVars = "status,sex",
#'         cpgLimit = 2,
#'         nCores = 1,
#'         summaryPval = 1,
#'         annotationPackage = "IlluminaHumanMethylation450kanno.ilmn12.hg19",
#'         annotationCols = "Name,chr,pos",
#'         display = FALSE,
#'         verbose = FALSE,
#'         logs = FALSE,
#'         saveOutputs = FALSE
#'     )
#'
#'     class(result)
#' }
#'
#' @seealso [dnaEPICO_methylationGLM-class]
#'
#' @export
methylationGLM <- function(
    inputPheno = "rData/preprocessingPheno/mergeData/phenoBT1.RData",
    outputLogs = "logs", outputRData = "rData/methylationGLM/models",
    outputPlots = "figures/methylationGLM", phenotypes = c(
        "DASS_Depression",
        "DASS_Anxiety", "DASS_Stress", "PCL5_TotalScore", "MHCSF_TotalScore",
        "BRS_TotalScore"
    ), covariates = "Sex,Age,Ethnicity,TraumaDefinition,Leukocytes,Epithelial.cells",
    factorVars = "Sex,Ethnicity,TraumaDefinition", scaleVars = NULL,
    cpgPrefix = "cg", cpgLimit = NA, methylationScale = "beta",
    nCores = 32, plotWidth = 2000, plotHeight = 1000, plotDPI = 150,
    interactionTerm = NULL, libPath = NULL, glmLibs = "glm2",
    prsMap = NULL, summaryPval = NA, summaryResidualSD = TRUE,
    excludeNonConverged = FALSE, saveSignificantCpGs = FALSE,
    significantCpGDir = "preliminaryResults/cpgs/methylationGLM",
    significantCpGPval = 0.05, saveTxtSummaries = TRUE, chunkSize = NULL,
    summaryTxtDir = "preliminaryResults/summary/methylationGLM",
    fdrThreshold = 0.05, padjmethod = "fdr",
        annotationPackage = "IlluminaHumanMethylationEPICv2anno.20a1.hg38",
    annotationCols = c(
        "Name", "chr", "pos", "UCSC_RefGene_Group",
        "UCSC_RefGene_Name", "Relation_to_Island", "GencodeV41_Group"
    ),
    annotatedGLMOut = "data/methylationGLM", reportAssetsDir = NULL,
    display = FALSE,
    verbose = FALSE, logs = FALSE, saveOutputs = FALSE
) {
    cpgLimit <- normalizeOptionalNumericMethylationGLM(cpgLimit)
    summaryPval <- normalizeOptionalNumericMethylationGLM(summaryPval)
    chunkSize <- normalizeChunkSizeMethylationGLM(chunkSize)
    methylationScale <- normalizeMethylationScaleDnaEpico(methylationScale)
    methylationLabel <- methylationScaleResponseLabelDnaEpico(methylationScale)
    methylationObjectPrefix <-
        methylationScaleObjectPrefixDnaEpico(methylationScale)
    responseColumn <- methylationScaleResponseColumnDnaEpico(methylationScale)
    log_file <- "log_methylationGLM.txt"

    if (is.null(libPath)) {
        libPath <- .libPaths()
    }

    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = outputLogs,
        log_file = log_file
    )

    emitLogMinfiEwasWater(
        c(
            "==== Starting DNAm GLM Analysis ====",
            paste("Start time:                ", format(Sys.time())),
            paste("Input phenotype + methylation:", inputPheno),
            paste(
                "Merged modeling object:    ", methylationObjectPrefix,
                "*"
            ), paste("Output RData folder:       ", outputRData),
            paste("Output logs folder:        ", outputLogs), paste(
                "Output plots folder:       ",
                outputPlots
            ), paste(
                "Report assets folder:      ",
                if (is.null(reportAssetsDir)) "None" else reportAssetsDir
            ), paste(
                "Phenotypes:                ",
                phenotypes
            ), paste(
                "Covariates:                ",
                covariates
            ), paste(
                "Factor variables:          ",
                factorVars
            ), paste(
                "Scale variables:           ",
                if (length(normalizeScaleVariablesDnaEpico(scaleVars))) {
                    paste(normalizeScaleVariablesDnaEpico(scaleVars),
                        collapse = ","
                    )
                } else {
                    "None"
                }
            ), paste("CpG column prefix:         ", cpgPrefix),
            paste("CpG limit:                 ",
                if (is.na(cpgLimit)) "All" else cpgLimit),
            paste("Number of cores:           ", as.integer(nCores)),
            paste("Interaction term:          ",
                if (is.null(interactionTerm)) "None" else interactionTerm),
            paste("GLM libraries:             ", glmLibs), paste(
                "PRS mapping:               ",
                if (is.null(prsMap)) "None" else prsMap
            ), paste(
                "Summary p-value filter:    ",
                if (is.na(summaryPval)) "None" else summaryPval
            ),
            paste("Include Residual SD:       ", isTRUE(summaryResidualSD)),
            paste("Exclude non-converged:     ", isTRUE(excludeNonConverged)),
            paste("Chunk size:                ",
                if (is.null(chunkSize)) "Auto" else chunkSize),
            paste("FDR threshold:             ", fdrThreshold), paste(
                "P-value adjustment method: ",
                padjmethod
            ), paste(
                "Display plots:             ",
                isTRUE(display)
            ), paste(
                "Save outputs:              ",
                isTRUE(saveOutputs)
            ), "============================================================"
        ),
        verbose = verbose, log_path = log_path
    )

    withLoggedErrorsMinfiEwasWater(expr = {
        preparedData <- prepareMethylationGLMData(
            inputPheno = inputPheno,
            phenotypes = phenotypes, covariates = covariates,
            factorVars = factorVars, scaleVars = scaleVars,
                cpgPrefix = cpgPrefix,
            cpgLimit = cpgLimit, methylationScale = methylationScale,
            interactionTerm = interactionTerm, prsMap = prsMap,
            verbose = verbose, logs = logs, log_dir = outputLogs,
            log_file = log_file
        )

        distributionPlots <- plotMethylationGLMDistributions(
            preparedData = preparedData,
            plotWidth = plotWidth, plotHeight = plotHeight, plotDPI = plotDPI,
            outputDir = if (isTRUE(saveOutputs)) {
                outputPlots
            } else {
                NULL
            }, display = display, verbose = verbose,
            logs = logs, log_dir = outputLogs, log_file = log_file
        )

        modelFits <- fitMethylationGLMModels(
            preparedData = preparedData,
            nCores = nCores, libPath = libPath, glmLibs = glmLibs,
            verbose = verbose, logs = logs, log_dir = outputLogs,
            log_file = log_file
        )

        modelSummaries <- summarizeMethylationGLMModels(
            modelResults = modelFits,
            preparedData = preparedData, summaryResidualSD = summaryResidualSD,
            summaryPval = summaryPval,
                excludeNonConverged = excludeNonConverged,
            nCores = nCores, libPath = libPath, glmLibs = glmLibs,
            chunkSize = chunkSize, verbose = verbose, logs = logs,
            log_dir = outputLogs, log_file = log_file
        )

        significantCpGs <- NULL
        if (isTRUE(saveSignificantCpGs)) {
            significantCpGs <- collectSignificantCpGsMethylationGLM(
                modelResults = modelFits,
                pvalThreshold = significantCpGPval,
                    interactionTerm = preparedData$interactionTerm,
                excludeNonConverged = excludeNonConverged, verbose = verbose,
                logs = logs, log_dir = outputLogs, log_file = log_file
            )
        }

        diagnosticPlots <- plotMethylationGLMDiagnostics(
            modelSummaries = modelSummaries,
            preparedData = preparedData, fdrThreshold = fdrThreshold,
            padjmethod = padjmethod, outputDir = if (isTRUE(saveOutputs)) {
                outputPlots
            } else {
                NULL
            }, plotWidth = plotWidth, plotHeight = plotHeight,
            plotDPI = plotDPI, display = display, verbose = verbose,
            logs = logs, log_dir = outputLogs, log_file = log_file
        )

        annotation <- annotateMethylationGLMSummaries(
            modelSummaries = modelSummaries,
            annotationObject = annotationPackage,
                annotationCols = annotationCols,
            verbose = verbose, logs = logs, log_dir = outputLogs,
            log_file = log_file
        )

        savedFiles <- NULL
        if (isTRUE(saveOutputs)) {
            savedFiles <- writeMethylationGLMOutputs(
                modelResults = modelFits,
                modelSummaries = modelSummaries, annotatedResults = annotation,
                significantCpGs = significantCpGs, outputRData = outputRData,
                summaryTxtDir = summaryTxtDir,
                    significantCpGDir = significantCpGDir,
                annotatedGLMOut = annotatedGLMOut,
                    reportAssetsDir = reportAssetsDir,
                    saveTxtSummaries = saveTxtSummaries,
                saveSignificantCpGs = saveSignificantCpGs, verbose = verbose,
                logs = logs, log_dir = outputLogs, log_file = log_file
            )
        }

        emitLogMinfiEwasWater(
            c(
                "============================================================",
                paste("Finished DNAm GLM Analysis:", format(Sys.time())),
                if (isTRUE(saveOutputs)) {
                    paste("Annotated GLM output:         ",
                        savedFiles$annotatedGLM)
                } else {
                    "Outputs were returned in memory only."
                }, "============================================================"
            ),
            verbose = verbose, log_path = log_path
        )

        structure(
            list(
                preparedData = preparedData,
                    distributionPlots = distributionPlots,
                modelFits = modelFits, modelSummaries = modelSummaries,
                significantCpGs = significantCpGs,
                    diagnosticPlots = diagnosticPlots,
                annotation = annotation, savedFiles = savedFiles,
                runSettings = list(
                    analysisLabel = "methylationGLM",
                    methylationScale = methylationScale,
                        methylationLabel = methylationLabel,
                    methylationObjectPrefix = methylationObjectPrefix,
                    internalResponseColumn = responseColumn,
                        scaleVars = preparedData$scaleVars,
                    scalingMetadata = preparedData$scalingMetadata,
                    excludeNonConverged = isTRUE(excludeNonConverged),
                    reportAssetsDir = reportAssetsDir
                )
            ),
            class = "dnaEPICO_methylationGLM"
        )
    }, log_path = log_path, verbose = verbose, context = "methylationGLM")
}

#' Parse timepoint selections for preprocessingPheno helpers
#'
#' @param values Character vector or comma-separated string of requested
#'   timepoints.
#' @param label Character. Label used in validation messages.
#'
#' @return Character vector of parsed timepoints.
#'
#' @description
#' Internal helper that standardizes timepoint parsing for
#' `preprocessingPheno()` and related helper functions.
#'
#' @keywords internal
#' @noRd
parseTimepointsPreprocessingPheno <- function(values, label = "timepoints") {
    parsed <- splitOptionMinfiEwasWater(values, sep = ",")

    if (length(parsed) == 0L) {
        stop(label, " must contain at least one timepoint.",
            call. = FALSE
        )
    }

    unique(parsed)
}

#' Extract a preferred named element from a loaded object
#'
#' @param object Loaded object.
#' @param preferred_name Character vector or `NULL`. Preferred element names,
#'   in selection order, when `object` is a named list.
#'
#' @return The original object or its preferred named element.
#'
#' @description
#' Internal helper that unwraps list-style saved objects used by
#' `preprocessingPheno()` and downstream wrappers.
#'
#' @keywords internal
#' @noRd
extractPreferredObjectPreprocessingPheno <- function(
    object,
    preferred_name = NULL
) {
    if (is.null(preferred_name)) {
        return(object)
    }

    preferred_names <- unique(as.character(preferred_name))
    preferred_names <- preferred_names[!is.na(preferred_names) &
        nzchar(preferred_names)]
    matching_names <- preferred_names[preferred_names %in% names(object)]
    if (is.list(object) && !is.data.frame(object) && length(matching_names) >
        0L) {
        return(object[[matching_names[[1L]]]])
    }

    object
}

#' Load a saved object used by preprocessingPheno helpers
#'
#' @param path Character. Path to an `.RData` or `.rds` file.
#' @param preferred_name Character vector or `NULL`. Preferred object names,
#'   in selection order, when `path` points to an `.RData` file containing
#'   multiple objects.
#'
#' @return The loaded object.
#'
#' @description
#' Internal helper that loads a saved object from disk while keeping
#' `preprocessingPheno()` compatible with the package's existing `.RData`
#' outputs.
#'
#' @keywords internal
#' @noRd
loadSavedObjectPreprocessingPheno <- function(path, preferred_name = NULL) {
    if (!file.exists(path)) {
        stop("Input file does not exist: ", path, call. = FALSE)
    }

    if (grepl("\\.rds$", path, ignore.case = TRUE)) {
        return(extractPreferredObjectPreprocessingPheno(
            object = readRDS(path),
            preferred_name = preferred_name
        ))
    }

    load_env <- new.env(parent = emptyenv())
    loaded_names <- load(path, envir = load_env)

    preferred_names <- unique(as.character(preferred_name))
    preferred_names <- preferred_names[!is.na(preferred_names) &
        nzchar(preferred_names)]
    matching_names <- preferred_names[preferred_names %in% loaded_names]
    if (length(matching_names) > 0L) {
        selected_name <- matching_names[[1L]]
        return(extractPreferredObjectPreprocessingPheno(
            object = load_env[[selected_name]],
            preferred_name = preferred_names
        ))
    }

    if (length(loaded_names) == 1L) {
        return(extractPreferredObjectPreprocessingPheno(
            object = load_env[[loaded_names[[1L]]]],
            preferred_name = preferred_name
        ))
    }

    stop("Could not determine which object to load from ", path,
        ". Available objects: ", paste(loaded_names, collapse = ", "),
        call. = FALSE
    )
}

#' Merge phenotype rows with methylation values for preprocessingPheno helpers
#'
#' @param phenoFrame Data frame containing phenotype rows.
#' @param methylationMatrix Numeric matrix with probes in rows and samples in
#'   columns.
#' @param id Character. Name of the sample identifier column in `phenoFrame`.
#' @param methylationScale Character. Selected methylation scale: `'Beta'`,
#'   `'M'`, or `'CN'`, case-insensitively.
#'
#' @return A data frame with phenotype columns followed by transposed
#' methylation
#'   values.
#'
#' @description
#' Internal helper that aligns phenotype rows with methylation-matrix columns
#' before
#' building merged objects used by downstream modeling functions.
#'
#' @keywords internal
#' @noRd
mergePhenoMethylationPreprocessingPheno <- function(
    phenoFrame,
    methylationMatrix, id = "Sample_Name", methylationScale = "beta"
) {
    methylation_scale <- normalizeMethylationScaleDnaEpico(methylationScale)
    methylation_label <-
        methylationScaleResponseLabelDnaEpico(methylation_scale)

    if (!(id %in% colnames(phenoFrame))) {
        stop("Sample identifier column not found in phenotype data: ",
            id,
            call. = FALSE
        )
    }

    if (!is.matrix(methylationMatrix) || !is.numeric(methylationMatrix) ||
        is.null(rownames(methylationMatrix)) ||
            is.null(colnames(methylationMatrix))) {
        stop(methylation_label,
            " must be a numeric matrix with sample identifiers in column names.",
            call. = FALSE
        )
    }
    validateMethylationProbeIdentifiersDnaEpico(
        rownames(methylationMatrix),
        paste0(methylation_label, " row names")
    )

    sample_ids <- validateSampleIdentifiersDnaEpico(
        phenoFrame[[id]],
        paste0("phenotype column '", id, "'")
    )
    matrix_ids <- validateSampleIdentifiersDnaEpico(
        colnames(methylationMatrix),
        paste0(methylation_label, " column names")
    )
    missing_samples <- setdiff(sample_ids, matrix_ids)
    if (length(missing_samples) > 0L) {
        stop("Phenotype samples are missing from ", methylation_label,
            ": ", paste(utils::head(missing_samples, 10L), collapse = ", "),
            call. = FALSE
        )
    }
    matched_ids <- sample_ids

    if (length(matched_ids) == 0L) {
        stop("No matching sample identifiers were found between phenotype data and ",
            methylation_label, ".",
            call. = FALSE
        )
    }

    if (anyDuplicated(matched_ids) > 0L) {
        stop("Duplicate sample identifiers were found while merging phenotype and methylation data: ",
            paste(unique(matched_ids[duplicated(matched_ids)]),
                collapse = ", "
            ),
            call. = FALSE
        )
    }

    phenoFrame <- phenoFrame[match(matched_ids, sample_ids), ,
        drop = FALSE
    ]
    methylationMatrix <- methylationMatrix[, matched_ids, drop = FALSE]

    cbind(phenoFrame, as.data.frame(t(methylationMatrix),
        stringsAsFactors = FALSE))
}

#' Load methylation metric matrices for preprocessingPheno
#'
#' @param betaPath Character. Path to the saved beta-value object.
#' @param mPath Character. Path to the saved M-value object.
#' @param cnPath Character. Path to the saved copy-number object.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return A list with class `'dnaEPICO_preprocessingPheno_metrics'`
#'   containing `beta`, `m`, and `cn`.
#'
#' @examples
#' ex <- dnaEPICO:::examplePreprocessingPhenoStateDnaEpico()
#' metrics_data <- loadMetricsPreprocessingPheno(
#'     betaPath = ex$betaPath,
#'     mPath = ex$mPath,
#'     cnPath = ex$cnPath,
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#' names(metrics_data)
#'
#' @description
#' Load the metric matrices generated by `preprocessingMinfiEwasWater()` and
#' return them as a single in-memory object for downstream phenotype alignment.
#'
#' @export
loadMetricsPreprocessingPheno <- function(
    betaPath, mPath, cnPath,
    verbose = FALSE, logs = FALSE, log_dir = NULL,
        log_file = "log_loadMetricsPreprocessingPheno.txt"
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )

    beta <- loadSavedObjectPreprocessingPheno(betaPath, preferred_name = "beta")
    m <- loadSavedObjectPreprocessingPheno(mPath, preferred_name = "m")
    cn <- loadSavedObjectPreprocessingPheno(cnPath, preferred_name = "cn")

    metric_objects <- list(beta = beta, m = m, cn = cn)
    invalid_metrics <- names(metric_objects)[!vapply(
        metric_objects,
        function(x) is.matrix(x) && is.numeric(x), logical(1)
    )]
    if (length(invalid_metrics) > 0L) {
        stop("Metric objects must be numeric matrices: ", paste(invalid_metrics,
            collapse = ", "
        ), call. = FALSE)
    }

    if (is.null(rownames(beta)) || is.null(colnames(beta))) {
        stop("The beta matrix must have probe and sample names.",
            call. = FALSE
        )
    }
    validateMethylationProbeIdentifiersDnaEpico(
        rownames(beta),
        "Beta-matrix row names"
    )
    validateSampleIdentifiersDnaEpico(colnames(beta),
        "Beta-matrix sample identifiers")
    for (metric_name in c("m", "cn")) {
        metric <- metric_objects[[metric_name]]
        if (!identical(dim(metric), dim(beta)) || !identical(
            rownames(metric),
            rownames(beta)
        ) || !identical(colnames(metric), colnames(beta))) {
            stop(metric_name,
                " must have the same dimensions, probe order, and sample order as beta.",
                call. = FALSE
            )
        }
    }

    validations <- lapply(names(metric_objects), function(metric_name) {
        inspectMethylationMatrixDnaEpico(
            values = metric_objects[[metric_name]],
            methylationScale = metric_name
        )
    })
    names(validations) <- names(metric_objects)
    metric_objects <- lapply(validations, `[[`, "values")
    beta <- metric_objects$beta
    m <- metric_objects$m
    cn <- metric_objects$cn

    converted_nan <- sum(vapply(
        validations, function(validation) validation$boundaries$NaN.Converted,
        numeric(1)
    ))
    invalid_cpgs <- sum(vapply(
        validations, function(validation) validation$boundaries$Invalid.CpGs,
        numeric(1)
    ))
    if (converted_nan > 0L) {
        warning(converted_nan,
            " NaN methylation values were converted to NA while loading metrics.",
            call. = FALSE
        )
    }
    if (invalid_cpgs > 0L) {
        warning(invalid_cpgs,
            " invalid scale-specific CpG records were retained for downstream reporting.",
            call. = FALSE
        )
    }

    preview_rows <- seq_len(min(nrow(beta), 5L))
    preview_cols <- seq_len(min(ncol(beta), 5L))

    emitLogMinfiEwasWater(
        c(
            "============================================================",
            paste("Beta path:                ", betaPath), paste(
                "M-values path:            ",
                mPath
            ), paste("CN path:                  ", cnPath),
            paste("Beta dimensions:          ", paste(dim(beta),
                collapse = " x "
            )), paste(
                "M dimensions:             ",
                paste(dim(m), collapse = " x ")
            ), paste(
                "CN dimensions:            ",
                paste(dim(cn), collapse = " x ")
            ), unlist(lapply(
                validations,
                function(validation) {
                    formatMethylationBoundariesLogDnaEpico(validation$boundaries)
                }
            ), use.names = FALSE), "Preview of beta values:",
            previewLinesMinfiEwasWater(beta[preview_rows, preview_cols,
                drop = FALSE
            ]), "============================================================"
        ),
        verbose = verbose, log_path = log_path
    )

    structure(list(beta = beta, m = m, cn = cn, validation = validations),
        class = "dnaEPICO_preprocessingPheno_metrics"
    )
}

#' Split phenotype and methylation data by timepoint
#'
#' @param pheno Data frame containing phenotype information.
#' @param metricsData Object returned by `loadMetricsPreprocessingPheno()`.
#' @param SampleID Character. Name of the sample identifier column in `pheno`.
#' @param timeVar Character. Name of the timepoint column in `pheno`.
#' @param timepoints Character vector or comma-separated string of timepoints to
#'   retain.
#' @param methylationScale Character. Methylation metric to use for the merged
#'   modeling table. One of `'Beta'`, `'M'`, or `'CN'`, case-insensitively.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return A list with class `'dnaEPICO_preprocessingPheno_timepoints'`
#'   containing the parsed timepoints and aligned per-timepoint subsets.
#'
#' @examples
#' ex <- dnaEPICO:::examplePreprocessingPhenoStateDnaEpico()
#' timepoint_data <- splitTimepointsPreprocessingPheno(
#'     pheno = ex$pheno,
#'     metricsData = ex$metricsData,
#'     SampleID = "Sample_Name",
#'     timeVar = "Timepoint",
#'     timepoints = "1,2",
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#' timepoint_data$timepoints
#'
#' @description
#' Align phenotype rows and metric matrices for each requested timepoint, and
#' precompute the per-timepoint phenotype-plus-beta objects used by downstream
#' modeling functions.
#'
#' @export
splitTimepointsPreprocessingPheno <- function(
    pheno, metricsData,
    SampleID = "Sample_Name", timeVar = "Timepoint", timepoints = "1,2",
    methylationScale = "beta", verbose = FALSE, logs = FALSE,
    log_dir = NULL, log_file = "log_splitTimepointsPreprocessingPheno.txt"
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )
    methylation_scale <- normalizeMethylationScaleDnaEpico(methylationScale)
    methylation_prefix <-
        methylationScaleObjectPrefixDnaEpico(methylation_scale)

    if (!(SampleID %in% colnames(pheno))) {
        stop("SampleID column not found in phenotype data: ",
            SampleID,
            call. = FALSE
        )
    }

    if (!(timeVar %in% colnames(pheno))) {
        stop("timeVar column not found in phenotype data: ",
            timeVar,
            call. = FALSE
        )
    }

    required_metrics <- c("beta", "m", "cn")
    if (!all(required_metrics %in% names(metricsData))) {
        stop("metricsData must contain beta, m, and cn matrices.",
            call. = FALSE
        )
    }

    requested_timepoints <- parseTimepointsPreprocessingPheno(
        values = timepoints,
        label = "timepoints"
    )
    available_timepoints <- table(pheno[[timeVar]], useNA = "ifany")
    subsets <- list()
    summary_lines <- c(paste("Requested timepoints:      ",
        paste(requested_timepoints,
        collapse = ", "
    )), paste(
        "Available values in", timeVar,
        "column:"
    ))
    summary_lines <- c(summary_lines,
        previewLinesMinfiEwasWater(available_timepoints))

    time_values <- as.character(pheno[[timeVar]])
    for (tp in requested_timepoints) {
        selected_timepoint <- !is.na(time_values) & time_values ==
            tp
        pheno_subset <- pheno[selected_timepoint, , drop = FALSE]

        if (nrow(pheno_subset) == 0L) {
            stop("No phenotype rows were found for timepoint: ",
                tp,
                call. = FALSE
            )
        }

        sample_ids <- validateSampleIdentifiersDnaEpico(
            pheno_subset[[SampleID]],
            paste0(
                "Phenotype sample identifiers at timepoint ",
                tp
            )
        )
        missing_metric_samples <- setdiff(sample_ids,
            colnames(metricsData$beta))
        if (length(missing_metric_samples) > 0L) {
            stop("Samples from timepoint ", tp,
                " are missing from the metric matrices: ",
                paste(utils::head(missing_metric_samples, 10L),
                    collapse = ", "
                ),
                call. = FALSE
            )
        }
        common_ids <- sample_ids

        pheno_subset <- pheno_subset[match(common_ids,
            as.character(pheno_subset[[SampleID]])), ,
            drop = FALSE
        ]

        beta_subset <- metricsData$beta[, common_ids, drop = FALSE]
        m_subset <- metricsData$m[, common_ids, drop = FALSE]
        cn_subset <- metricsData$cn[, common_ids, drop = FALSE]
        selected_subset <- metricsData[[methylation_scale]][,
            common_ids,
            drop = FALSE
        ]
        pheno_methylation <- mergePhenoMethylationPreprocessingPheno(
            phenoFrame = pheno_subset,
            methylationMatrix = selected_subset, id = SampleID,
            methylationScale = methylation_scale
        )

        subset_data <- list(
            pheno = pheno_subset, beta = beta_subset,
            m = m_subset, cn = cn_subset, phenoMethylation = pheno_methylation,
            methylationScale = methylation_scale,
                methylationObjectPrefix = methylation_prefix
        )
        subset_data[[methylation_prefix]] <- pheno_methylation
        subsets[[tp]] <- subset_data

        summary_lines <- c(
            summary_lines, paste("Timepoint ",
                tp, " rows:          ", nrow(pheno_subset),
                sep = ""
            ),
            paste("Timepoint ", tp, " matched samples:", length(common_ids),
                sep = " "
            ), paste("Timepoint ", tp, " merged object:  ",
                methylation_prefix, "T", tp,
                sep = ""
            )
        )
    }

    emitLogMinfiEwasWater(c(summary_lines,
        "============================================================"),
        verbose = verbose, log_path = log_path
    )

    structure(
        list(
            timepoints = requested_timepoints, data = subsets,
            methylationScale = methylation_scale,
                methylationObjectPrefix = methylation_prefix
        ),
        class = "dnaEPICO_preprocessingPheno_timepoints"
    )
}

#' Combine selected timepoints for downstream longitudinal modeling
#'
#' @param timepointData Object returned by
#' `splitTimepointsPreprocessingPheno()`.
#' @param combineTimepoints Character vector or comma-separated string of
#'   timepoints to combine.
#' @param methylationScale Character. Methylation metric used in the merged
#'   modeling table. One of `'Beta'`, `'M'`, or `'CN'`, case-insensitively.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return A list with class `'dnaEPICO_preprocessingPheno_combined'`
#'   containing the combined phenotype table, merged phenotype-plus-beta table,
#'   selected timepoints, and output suffix.
#'
#' @examples
#' ex <- dnaEPICO:::examplePreprocessingPhenoStateDnaEpico()
#' combined_data <- combineTimepointsPreprocessingPheno(
#'     timepointData = ex$timepointData,
#'     combineTimepoints = "1,2",
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#' combined_data$suffix
#'
#' @description
#' Combine selected timepoints that were already aligned by
#' `splitTimepointsPreprocessingPheno()` into the wide phenotype-plus-beta
#' objects used by downstream longitudinal models.
#'
#' @export
combineTimepointsPreprocessingPheno <- function(
    timepointData,
    combineTimepoints = "1,2", methylationScale = "beta", verbose = FALSE,
    logs = FALSE, log_dir = NULL,
        log_file = "log_combineTimepointsPreprocessingPheno.txt"
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )
    methylation_scale <- if (!is.null(timepointData$methylationScale)) {
        normalizeMethylationScaleDnaEpico(timepointData$methylationScale)
    } else {
        normalizeMethylationScaleDnaEpico(methylationScale)
    }
    methylation_prefix <-
        methylationScaleObjectPrefixDnaEpico(methylation_scale)

    requested_timepoints <- parseTimepointsPreprocessingPheno(
        values = combineTimepoints,
        label = "combineTimepoints"
    )

    available_timepoints <- names(timepointData$data)
    missing_timepoints <- setdiff(requested_timepoints, available_timepoints)

    if (length(missing_timepoints) > 0L) {
        stop("Requested combined timepoints are missing from timepointData: ",
            paste(missing_timepoints, collapse = ", "),
            call. = FALSE
        )
    }

    combined_pheno <- do.call(rbind, lapply(
        requested_timepoints,
        function(tp) timepointData$data[[tp]]$pheno
    ))
    combined_pheno_beta <- do.call(rbind, lapply(
        requested_timepoints,
        function(tp) {
            tp_data <- timepointData$data[[tp]]
            if (!is.null(tp_data$phenoMethylation)) {
                return(tp_data$phenoMethylation)
            }
            tp_data[[methylation_prefix]]
        }
    ))
    combine_suffix <- paste0("T", paste(requested_timepoints,
        collapse = "T"
    ))

    emitLogMinfiEwasWater(
        c(paste(
            "Combining timepoints:      ",
            paste(requested_timepoints, collapse = ", ")
        ), paste(
            "Combined phenotype rows:   ",
            nrow(combined_pheno)
        ), paste("Combined ", methylation_prefix,
            " rows:   ", nrow(combined_pheno_beta),
            sep = ""
        ), paste(
            "Combined suffix:           ",
            combine_suffix
        ), "============================================================"),
        verbose = verbose, log_path = log_path
    )

    combined_data <- list(
        timepoints = requested_timepoints,
        suffix = combine_suffix, pheno = combined_pheno,
            phenoMethylation = combined_pheno_beta,
        methylationScale = methylation_scale,
            methylationObjectPrefix = methylation_prefix
    )
    combined_data[[methylation_prefix]] <- combined_pheno_beta

    structure(combined_data, class = "dnaEPICO_preprocessingPheno_combined")
}

#' Build Clock Foundation input tables from preprocessingPheno data
#'
#' @param beta Numeric matrix of beta values with probes in rows and samples in
#'   columns.
#' @param pheno Phenotype data frame aligned with the beta matrix columns.
#' @param SampleID Character. Name of the phenotype sample identifier column.
#' @param sexColumn Character. Name of the phenotype sex column.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return A list with class `'dnaEPICO_preprocessingPheno_clock'`
#'   containing `betaCSV` and `phenoCF`.
#'
#' @examples
#' ex <- dnaEPICO:::examplePreprocessingPhenoStateDnaEpico()
#' clock_inputs <- buildClockFoundationInputsPreprocessingPheno(
#'     beta = ex$timepointData$data[["1"]]$beta,
#'     pheno = ex$timepointData$data[["1"]]$pheno,
#'     SampleID = "Sample_Name",
#'     sexColumn = "Sex",
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#' names(clock_inputs)
#'
#' @description
#' Prepare the beta and phenotype tables commonly exported for Clock Foundation
#' style downstream workflows, without writing them to disk.
#'
#' @export
buildClockFoundationInputsPreprocessingPheno <- function(
    beta,
    pheno, SampleID = "Sample_Name", sexColumn = "Sex", verbose = FALSE,
    logs = FALSE, log_dir = NULL,
        log_file = "log_buildClockFoundationInputsPreprocessingPheno.txt"
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )

    if (!(SampleID %in% colnames(pheno))) {
        stop("SampleID column not found in phenotype data: ",
            SampleID,
            call. = FALSE
        )
    }

    if (!(sexColumn %in% colnames(pheno))) {
        stop("sexColumn not found in phenotype data: ", sexColumn,
            call. = FALSE
        )
    }

    if ("id" %in% colnames(pheno) && !identical(SampleID, "id")) {
        stop("The phenotype data already contains a column named 'id'. ",
            "Please rename either that column or SampleID before building Clock Foundation inputs.",
            call. = FALSE
        )
    }

    if (!is.matrix(beta) || !is.numeric(beta) || is.null(rownames(beta)) ||
        is.null(colnames(beta))) {
        stop("beta must be a numeric matrix with probe and sample names.",
            call. = FALSE
        )
    }
    validateMethylationProbeIdentifiersDnaEpico(
        rownames(beta),
        "Clock Foundation beta row names"
    )
    beta_validation <- inspectMethylationMatrixDnaEpico(
        beta,
        "beta"
    )
    beta <- beta_validation$values
    if (beta_validation$boundaries$NaN.Converted > 0L) {
        warning(beta_validation$boundaries$NaN.Converted,
            " NaN beta value(s) were converted to NA for Clock Foundation input.",
            call. = FALSE
        )
    }
    if (nrow(beta_validation$invalidCpGs) > 0L) {
        warning(nrow(beta_validation$invalidCpGs),
            " invalid CpG(s) were excluded from Clock Foundation input.",
            call. = FALSE
        )
        beta <- beta[!(rownames(beta) %in% beta_validation$invalidCpGs$CpG), ,
            drop = FALSE
        ]
    }
    if (nrow(beta) == 0L) {
        stop("No valid beta CpGs remain for Clock Foundation input.",
            call. = FALSE
        )
    }
    beta_sample_ids <- validateSampleIdentifiersDnaEpico(
        colnames(beta),
        "Beta-matrix sample identifiers"
    )
    pheno_sample_ids <- validateSampleIdentifiersDnaEpico(
        pheno[[SampleID]],
        paste0("Phenotype column '", SampleID, "'")
    )
    pheno_match <- matchSampleIdentifiersDnaEpico(
        query = beta_sample_ids,
        reference = pheno_sample_ids,
            queryLabel = "Beta-matrix sample identifiers",
        referenceLabel = paste0(
            "phenotype column '", SampleID,
            "'"
        ), requireSameSet = TRUE
    )
    pheno <- pheno[pheno_match, , drop = FALSE]
    rownames(pheno) <- NULL

    beta_csv <- as.data.frame(beta, stringsAsFactors = FALSE)
    beta_csv <- cbind(
        ProbeID = rownames(beta_csv), beta_csv,
        row.names = NULL, stringsAsFactors = FALSE
    )

    pheno_cf <- pheno
    colnames(pheno_cf)[colnames(pheno_cf) == SampleID] <- "id"

    sex_info <- canonicalizeSexDnaEpico(pheno_cf[[sexColumn]])
    if (anyNA(sex_info$code)) {
        stop("The Clock Foundation sex column contains missing or unsupported values",
            if (length(sex_info$unknown) > 0L) {
                paste0(": ", paste(sex_info$unknown, collapse = ", "))
            } else {
                "."
            },
            call. = FALSE
        )
    }
    pheno_cf[[sexColumn]] <- ifelse(sex_info$code == 0L, "Female",
        "Male"
    )
    sex_line <- "Sex values were standardized to Female and Male."

    preview_rows <- seq_len(min(nrow(beta_csv), 5L))
    preview_cols <- seq_len(min(ncol(beta_csv), 5L))

    emitLogMinfiEwasWater(
        c(
            paste(
                "Clock Foundation beta rows:",
                nrow(beta_csv)
            ), paste(
                "Clock Foundation beta cols:",
                ncol(beta_csv)
            ), paste(
                "Clock Foundation pheno rows:",
                nrow(pheno_cf)
            ), formatMethylationBoundariesLogDnaEpico(beta_validation$boundaries),
            sex_line, "Preview of Clock Foundation beta table:",
            previewLinesMinfiEwasWater(beta_csv[preview_rows, preview_cols,
                drop = FALSE
            ]), "============================================================"
        ),
        verbose = verbose, log_path = log_path
    )

    structure(
        list(
            betaCSV = beta_csv, phenoCF = pheno_cf,
                methylationBoundaries = beta_validation$boundaries,
            methylationIssues = beta_validation$issues,
                invalidCpGs = beta_validation$invalidCpGs
        ),
        class = "dnaEPICO_preprocessingPheno_clock"
    )
}

#' Write preprocessingPheno outputs to disk
#'
#' @param preprocessingData Object returned by `preprocessingPheno()` or a list
#'   with the same components.
#' @param outputPheno Character. Directory used for saved phenotype CSV files.
#' @param outputRData Character. Directory used for saved metric `.RData` files.
#' @param outputRDataMerge Character. Directory used for saved merged
#'   phenotype-plus-beta `.RData` files.
#' @param outputDir Character. Directory used for the Clock Foundation export
#'   files.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return A list with class `'dnaEPICO_preprocessingPheno_paths'`
#'   containing the paths written to disk.
#'
#' @examples
#' ex <- dnaEPICO:::examplePreprocessingPhenoStateDnaEpico()
#' output_paths <- writePreprocessingPhenoOutputs(
#'     preprocessingData = ex$preprocessingData,
#'     outputPheno = file.path(ex$tempDir, "pheno"),
#'     outputRData = file.path(ex$tempDir, "metrics"),
#'     outputRDataMerge = file.path(ex$tempDir, "merge"),
#'     outputDir = file.path(ex$tempDir, "clock"),
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#' names(output_paths)
#'
#' @description
#' Write the CSV, ZIP, and `.RData` outputs produced by
#' `preprocessingPheno()`. This helper keeps file writing separate from the
#' in-memory preprocessing steps.
#'
#' @export
writePreprocessingPhenoOutputs <- function(
    preprocessingData,
    outputPheno = "data/preprocessingPheno",
        outputRData = "rData/preprocessingPheno/metrics",
    outputRDataMerge = "rData/preprocessingPheno/mergeData",
    outputDir = "data/preprocessingPheno", verbose = FALSE, logs = FALSE,
    log_dir = NULL, log_file = "log_writePreprocessingPhenoOutputs.txt"
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )

    dir.create(outputPheno, recursive = TRUE, showWarnings = FALSE)
    dir.create(outputRData, recursive = TRUE, showWarnings = FALSE)
    dir.create(outputRDataMerge, recursive = TRUE, showWarnings = FALSE)
    dir.create(outputDir, recursive = TRUE, showWarnings = FALSE)

    methylation_scale <- if (!is.null(preprocessingData$methylationScale)) {
        normalizeMethylationScaleDnaEpico(preprocessingData$methylationScale)
    } else if (!is.null(preprocessingData$combinedData$methylationScale)) {
        normalizeMethylationScaleDnaEpico(preprocessingData$combinedData$methylationScale)
    } else {
        normalizeMethylationScaleDnaEpico("beta")
    }
    methylation_prefix <-
        methylationScaleObjectPrefixDnaEpico(methylation_scale)

    timepoint_paths <- list()

    for (tp in preprocessingData$timepointData$timepoints) {
        tp_data <- preprocessingData$timepointData$data[[tp]]

        pheno_path <- file.path(outputPheno, paste0(
            "phenoT",
            tp, ".csv"
        ))
        beta_path <- file.path(outputRData, paste0(
            "betaT", tp,
            ".RData"
        ))
        m_path <- file.path(outputRData, paste0("mT", tp, ".RData"))
        cn_path <- file.path(outputRData, paste0("cnT", tp, ".RData"))
        merged_path <- file.path(outputRDataMerge, paste0(
            methylation_prefix,
            "T", tp, ".RData"
        ))
        merged_object_name <- paste0(
            methylation_prefix, "T",
            tp
        )
        merged_data <- if (!is.null(tp_data$phenoMethylation)) {
            tp_data$phenoMethylation
        } else {
            tp_data[[methylation_prefix]]
        }

        utils::write.csv(tp_data$pheno, file = pheno_path, row.names = FALSE)
        saveNamedObjectMinfiEwasWater(tp_data$beta, paste0(
            "betaT",
            tp
        ), beta_path)
        saveNamedObjectMinfiEwasWater(tp_data$m, paste0(
            "mT",
            tp
        ), m_path)
        saveNamedObjectMinfiEwasWater(tp_data$cn, paste0(
            "cnT",
            tp
        ), cn_path)
        saveNamedObjectMinfiEwasWater(
            merged_data, merged_object_name,
            merged_path
        )

        tp_paths <- list(
            pheno = pheno_path, beta = beta_path,
            m = m_path, cn = cn_path, phenoMethylation = merged_path
        )
        tp_paths[[methylation_prefix]] <- merged_path
        timepoint_paths[[tp]] <- tp_paths
    }

    combined_pheno_path <- file.path(outputPheno, paste0(
        "pheno",
        preprocessingData$combinedData$suffix, ".csv"
    ))
    combined_merge_path <- file.path(outputRDataMerge, paste0(
        methylation_prefix,
        preprocessingData$combinedData$suffix, ".RData"
    ))
    combined_merge_object_name <- paste0(
        methylation_prefix,
        preprocessingData$combinedData$suffix
    )
    combined_merge_data <-
        if (!is.null(preprocessingData$combinedData$phenoMethylation)) {
        preprocessingData$combinedData$phenoMethylation
    } else {
        preprocessingData$combinedData[[methylation_prefix]]
    }
    beta_csv_path <- file.path(outputDir, "beta.csv")
    zip_path <- file.path(outputDir, "beta.zip")
    pheno_cf_path <- file.path(outputDir, "phenoCF.csv")

    utils::write.csv(preprocessingData$combinedData$pheno,
        file = combined_pheno_path,
        row.names = FALSE
    )
    saveNamedObjectMinfiEwasWater(
        combined_merge_data, combined_merge_object_name,
        combined_merge_path
    )
    utils::write.csv(preprocessingData$clockFoundation$betaCSV,
        file = beta_csv_path, row.names = FALSE
    )
    utils::write.csv(preprocessingData$clockFoundation$phenoCF,
        file = pheno_cf_path, row.names = FALSE
    )
    zip_status <- utils::zip(
        zipfile = zip_path, files = beta_csv_path,
        flags = "-j"
    )
    beta_zip_path <- if (identical(zip_status, 0L) && file.exists(zip_path)) {
        zip_path
    } else {
        NULL
    }

    emitLogMinfiEwasWater(
        c(
            paste(
                "Output phenotype dir:     ",
                outputPheno
            ), paste("RData metrics dir:        ", outputRData),
            paste("RData merge dir:          ", outputRDataMerge),
            paste("Clock Foundation dir:     ", outputDir), paste(
                "Saved combined phenotype: ",
                combined_pheno_path
            ), paste("Saved combined ", methylation_prefix,
                ": ", combined_merge_path,
                sep = ""
            ), paste(
                "Saved beta CSV:           ",
                beta_csv_path
            ), if (is.null(beta_zip_path)) {
                "Beta ZIP file was not created."
            } else {
                paste("Saved beta ZIP:           ", beta_zip_path)
            }, paste("Saved phenoCF:            ", pheno_cf_path),
            paste("ZIP status code:          ", zip_status),
                "============================================================"
        ),
        verbose = verbose, log_path = log_path
    )

    output_paths <- list(
        timepoints = timepoint_paths, combinedPheno = combined_pheno_path,
        combinedPhenoMethylation = combined_merge_path,
            methylationScale = methylation_scale,
        methylationObjectPrefix = methylation_prefix, betaCSV = beta_csv_path,
        betaZIP = beta_zip_path, phenoCF = pheno_cf_path
    )
    combined_key <- switch(methylation_scale,
        beta = "combinedPhenoB",
        m = "combinedPhenoM",
        cn = "combinedPhenoCN"
    )
    output_paths[[combined_key]] <- combined_merge_path

    structure(output_paths, class = "dnaEPICO_preprocessingPheno_paths")
}

create_preprocessing_pheno_example <- function(path) {
    pheno <- data.frame(
        Sample_Name = c("S1", "S2", "S3"),
        Timepoint = c("1", "1", "2"),
        Sex = c(0, 1, 0),
        stringsAsFactors = FALSE
    )

    beta <- matrix(
        c(0.10, 0.20, 0.30, 0.40, 0.50, 0.60),
        nrow = 2,
        dimnames = list(c("cg1", "cg2"), pheno$Sample_Name)
    )
    m <- beta * 10
    cn <- beta * 100

    pheno_file <- file.path(path, "pheno.csv")
    beta_path <- file.path(path, "beta.RData")
    m_path <- file.path(path, "m.RData")
    cn_path <- file.path(path, "cn.RData")

    utils::write.csv(pheno, pheno_file, row.names = FALSE)
    save(beta, file = beta_path)
    save(m, file = m_path)
    save(cn, file = cn_path)

    list(
        pheno = pheno,
        beta = beta,
        m = m,
        cn = cn,
        phenoFile = pheno_file,
        betaPath = beta_path,
        mPath = m_path,
        cnPath = cn_path
    )
}

test_that("preprocessingPheno returns in-memory objects quietly by default", {
    tmp <- withr::local_tempdir()
    example_data <- create_preprocessing_pheno_example(tmp)

    expect_message(
        result <- preprocessingPheno(
            phenoFile = example_data$phenoFile,
            betaPath = example_data$betaPath,
            mPath = example_data$mPath,
            cnPath = example_data$cnPath,
            SampleID = "Sample_Name",
            timeVar = "Timepoint",
            timepoints = "1,2",
            combineTimepoints = "1,2",
            outputPheno = file.path(tmp, "data", "preprocessingPheno"),
            outputRData = file.path(tmp, "rData", "preprocessingPheno", "metrics"),
            outputRDataMerge = file.path(tmp, "rData", "preprocessingPheno", "mergeData"),
            sexColumn = "Sex",
            outputLogs = file.path(tmp, "logs"),
            outputDir = file.path(tmp, "clockFoundation"),
            saveOutputs = FALSE
        ),
        NA
    )

    expect_s3_class(result, "dnaEPICO_preprocessingPheno")
    expect_null(result$savedFiles)
    expect_identical(result$timepointData$timepoints, c("1", "2"))
    expect_equal(nrow(result$combinedData$pheno), 3)
    expect_equal(nrow(result$combinedData$phenoB), 3)
    expect_true(all(c("ProbeID", "S1", "S2", "S3") %in% colnames(result$clockFoundation$betaCSV)))
    expect_true("id" %in% colnames(result$clockFoundation$phenoCF))
    expect_false(dir.exists(file.path(tmp, "data", "preprocessingPheno")))
})

test_that("preprocessingPheno can write legacy outputs and logs on request", {
    tmp <- withr::local_tempdir()
    example_data <- create_preprocessing_pheno_example(tmp)

    expect_message(
        result <- preprocessingPheno(
            phenoFile = example_data$phenoFile,
            betaPath = example_data$betaPath,
            mPath = example_data$mPath,
            cnPath = example_data$cnPath,
            SampleID = "Sample_Name",
            timeVar = "Timepoint",
            timepoints = "1,2",
            combineTimepoints = "1,2",
            outputPheno = file.path(tmp, "data", "preprocessingPheno"),
            outputRData = file.path(tmp, "rData", "preprocessingPheno", "metrics"),
            outputRDataMerge = file.path(tmp, "rData", "preprocessingPheno", "mergeData"),
            sexColumn = "Sex",
            outputLogs = file.path(tmp, "logs"),
            outputDir = file.path(tmp, "clockFoundation"),
            verbose = TRUE,
            logs = TRUE,
            saveOutputs = TRUE
        ),
        "Starting preprocessingPheno"
    )

    expect_s3_class(result$savedFiles, "dnaEPICO_preprocessingPheno_paths")
    expect_true(file.exists(file.path(tmp, "logs", "log_preprocessingPheno.txt")))
    expect_true(file.exists(file.path(tmp, "data", "preprocessingPheno", "phenoT1.csv")))
    expect_true(file.exists(file.path(tmp, "rData", "preprocessingPheno", "metrics", "betaT1.RData")))
    pheno_b_t1 <- file.path(
        tmp,
        "rData",
        "preprocessingPheno",
        "mergeData",
        "phenoBT1.RData"
    )
    expect_true(file.exists(pheno_b_t1))
    expect_equal(
        result$savedFiles$combinedPhenoB,
        file.path(
            tmp,
            "rData",
            "preprocessingPheno",
            "mergeData",
            "phenoBT1T2.RData"
        )
    )
    expect_true(file.exists(file.path(tmp, "clockFoundation", "beta.csv")))
    expect_true(file.exists(file.path(tmp, "clockFoundation", "phenoCF.csv")))
})

test_that("preprocessingPheno selects M and CN modeling scales without changing Clock Foundation beta", {
    tmp <- withr::local_tempdir()
    example_data <- create_preprocessing_pheno_example(tmp)

    result_m <- preprocessingPheno(
        phenoFile = example_data$phenoFile,
        betaPath = example_data$betaPath,
        mPath = example_data$mPath,
        cnPath = example_data$cnPath,
        SampleID = "Sample_Name",
        timeVar = "Timepoint",
        timepoints = "1,2",
        combineTimepoints = "1,2",
        methylationScale = "m",
        outputPheno = file.path(tmp, "data", "preprocessingPhenoM"),
        outputRData = file.path(tmp, "rData", "preprocessingPhenoM", "metrics"),
        outputRDataMerge = file.path(tmp, "rData", "preprocessingPhenoM", "mergeData"),
        sexColumn = "Sex",
        outputLogs = file.path(tmp, "logs"),
        outputDir = file.path(tmp, "clockFoundationM"),
        saveOutputs = FALSE
    )

    expect_equal(result_m$methylationScale, "m")
    expect_true("phenoM" %in% names(result_m$combinedData))
    expect_equal(result_m$combinedData$phenoM$cg1, c(1, 3, 5))
    expect_equal(
        result_m$clockFoundation$betaCSV$S1,
        c(0.10, 0.20)
    )

    result_cn <- preprocessingPheno(
        phenoFile = example_data$phenoFile,
        betaPath = example_data$betaPath,
        mPath = example_data$mPath,
        cnPath = example_data$cnPath,
        SampleID = "Sample_Name",
        timeVar = "Timepoint",
        timepoints = "1,2",
        combineTimepoints = "1,2",
        methylationScale = "cn",
        outputPheno = file.path(tmp, "data", "preprocessingPhenoCN"),
        outputRData = file.path(tmp, "rData", "preprocessingPhenoCN", "metrics"),
        outputRDataMerge = file.path(tmp, "rData", "preprocessingPhenoCN", "mergeData"),
        sexColumn = "Sex",
        outputLogs = file.path(tmp, "logs"),
        outputDir = file.path(tmp, "clockFoundationCN"),
        saveOutputs = TRUE
    )

    pheno_cn_t1 <- file.path(tmp, "rData", "preprocessingPhenoCN", "mergeData", "phenoCNT1.RData")
    pheno_cn_t1t2 <- file.path(tmp, "rData", "preprocessingPhenoCN", "mergeData", "phenoCNT1T2.RData")
    expect_true(file.exists(pheno_cn_t1))
    expect_true(file.exists(pheno_cn_t1t2))
    merge_files <- basename(list.files(file.path(tmp, "rData", "preprocessingPhenoCN", "mergeData")))
    expect_false("phenoCnT1.RData" %in% merge_files)
    expect_equal(result_cn$savedFiles$combinedPhenoCN, pheno_cn_t1t2)

    loaded <- new.env(parent = emptyenv())
    load(pheno_cn_t1, envir = loaded)
    expect_true("phenoCNT1" %in% ls(loaded))
    expect_equal(loaded$phenoCNT1$cg1, c(10, 30))
})

test_that("Clock Foundation input reports and excludes invalid beta CpGs", {
    beta <- rbind(
        cg_boundary = c(0, 1),
        cg_nan = c(NaN, 0.4),
        cg_outside = c(-0.01, 0.5),
        cg_missing = c(NA_real_, NA_real_)
    )
    colnames(beta) <- c("S1", "S2")
    pheno <- data.frame(
        Sample_Name = c("S2", "S1"),
        Sex = c("M", "F"),
        stringsAsFactors = FALSE
    )

    result <- suppressWarnings(buildClockFoundationInputsPreprocessingPheno(
        beta = beta,
        pheno = pheno,
        verbose = FALSE,
        logs = FALSE
    ))

    expect_identical(result$betaCSV$ProbeID, c("cg_boundary", "cg_nan"))
    expect_true(is.na(result$betaCSV$S1[result$betaCSV$ProbeID == "cg_nan"]))
    expect_identical(result$phenoCF$id, c("S1", "S2"))
    expect_identical(result$invalidCpGs$CpG, c("cg_outside", "cg_missing"))
    expect_equal(result$methylationBoundaries$At.Lower.Boundary, 1L)
    expect_equal(result$methylationBoundaries$At.Upper.Boundary, 1L)
    expect_equal(result$methylationBoundaries$NaN.Converted, 1L)
})

test_that("Clock Foundation input rejects ambiguous CpG identifiers", {
    beta <- matrix(
        c(0.1, 0.2, 0.3, 0.4),
        nrow = 2,
        dimnames = list(c("cg1", "cg1"), c("S1", "S2"))
    )
    pheno <- data.frame(
        Sample_Name = c("S1", "S2"),
        Sex = c("F", "M"),
        stringsAsFactors = FALSE
    )

    expect_error(
        buildClockFoundationInputsPreprocessingPheno(
            beta = beta,
            pheno = pheno,
            logs = FALSE
        ),
        "duplicate CpG identifiers"
    )
})

test_that("timepoint parsing ignores missing rows and removes duplicate requests", {
    beta <- matrix(
        c(0.1, 0.2, 0.3),
        nrow = 1,
        dimnames = list("cg1", c("S1", "S2", "S3"))
    )
    metrics <- list(beta = beta, m = beta, cn = beta)
    pheno <- data.frame(
        Sample_Name = c("S1", "S2", "S3"),
        Timepoint = c("1", NA_character_, "2"),
        stringsAsFactors = FALSE
    )

    result <- splitTimepointsPreprocessingPheno(
        pheno = pheno,
        metricsData = metrics,
        timepoints = "1,1,2",
        logs = FALSE
    )

    expect_identical(result$timepoints, c("1", "2"))
    expect_identical(result$data[["1"]]$pheno$Sample_Name, "S1")
    expect_identical(result$data[["2"]]$pheno$Sample_Name, "S3")
})

#' Normalize a report path for cross-platform rendering
#'
#' @param path Character. Path to normalize.
#'
#' @return Character scalar containing a normalized path.
#'
#' @keywords internal
#' @noRd
normalizePathDnamReport <- function(path) {
    gsub("\\\\", "/", normalizePath(path, winslash = "/", mustWork = FALSE))
}

#' Format a report print path
#'
#' @param path Character path to format.
#' @param missing Character value used when the path is unavailable.
#'
#' @return Character scalar for console printing.
#'
#' @keywords internal
#' @noRd
formatPrintPathDnamReport <- function(path, missing = "not written") {
    if (is.null(path) || !length(path) || is.na(path[[1]]) ||
        !nzchar(path[[1]])) {
        return(missing)
    }

    normalizePathDnamReport(path[[1]])
}

#' Collect report figure inventory from a directory
#'
#' @param directory Character. Directory containing report figures.
#' @param patterns Character vector of regular expressions used to find matching
#'   files.
#' @param label Character. Short label used in report logs.
#' @param recursive Logical. If `TRUE`, search subdirectories as well.
#'
#' @return A list containing the normalized directory, whether it exists, the
#'   matching files, the number of matches, and the label.
#'
#' @keywords internal
#' @noRd
collectFigureInventoryDnamReport <- function(
    directory, patterns,
    label, recursive = TRUE
) {
    normalized_dir <- normalizePathDnamReport(directory)
    files <- character(0)
    directory_exists <- dir.exists(directory)

    if (directory_exists) {
        file_sets <- lapply(patterns, function(pattern) {
            list.files(directory,
                pattern = pattern, full.names = TRUE,
                recursive = recursive, ignore.case = TRUE
            )
        })
        files <- sort(unique(unlist(file_sets, use.names = FALSE)))
        files <- vapply(files, normalizePathDnamReport, character(1))
    }

    list(
        label = label, directory = normalized_dir, exists = directory_exists,
        files = files, count = length(files)
    )
}

#' Prepare inputs for a DNA methylation report
#'
#' @param outputDir Character. Directory where the report project is written.
#' @param qcDir Character. Directory containing ENmix quality-control figures.
#' @param preprocessingDir Character. Directory containing preprocessing
#'   quality-control figures.
#' @param postprocessingDir Character. Directory containing postprocessing
#' metric
#'   figures.
#' @param svaDir Character. Directory containing SVA or batch-effect figures.
#' @param glmDir Character. Directory containing GLM figures.
#' @param lmeDir Character. Directory containing LME figures.
#' @param figDir Character. Directory used for generated report figure assets.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write progress messages to
#'   `file.path(logDir, 'log_dnamReport.txt')`.
#' @param logDir Character. Directory for optional log files.
#'
#' @return A list with class `'dnaEPICO_dnamReport_prepared'`.
#'
#' @examples
#' report_root <- file.path(tempdir(), "dnaepico-report-inputs")
#' prepared <- prepareDnamReportInputs(
#'     outputDir = file.path(report_root, "reports"),
#'     qcDir = file.path(
#'         report_root,
#'         "figures",
#'         "preprocessingMinfiEwasWater",
#'         "enmix"
#'     ),
#'     preprocessingDir = file.path(
#'         report_root,
#'         "figures",
#'         "preprocessingMinfiEwasWater",
#'         "qc"
#'     ),
#'     postprocessingDir = file.path(
#'         report_root,
#'         "figures",
#'         "preprocessingMinfiEwasWater",
#'         "metrics"
#'     ),
#'     svaDir = file.path(report_root, "figures", "svaEnmix")
#' )
#' inherits(prepared, "dnaEPICO_dnamReport_prepared")
#'
#' @export
prepareDnamReportInputs <- function(
    outputDir = "reports", qcDir = file.path(
        "figures",
        "preprocessingMinfiEwasWater", "enmix"
    ), preprocessingDir = file.path(
        "figures",
        "preprocessingMinfiEwasWater", "qc"
    ), postprocessingDir = file.path(
        "figures",
        "preprocessingMinfiEwasWater", "metrics"
    ), svaDir = file.path(
        "figures",
        "svaEnmix"
    ), glmDir = file.path("figures", "methylationGLM"),
    lmeDir = file.path("figures", "methylationLME"), figDir = file.path(
        outputDir,
        "assets", "figures"
    ), verbose = FALSE, logs = FALSE,
    logDir = outputDir
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = logDir,
        log_file = "log_dnamReport.txt"
    )

    figure_inventory <- list(
        qc = collectFigureInventoryDnamReport(
            directory = qcDir,
            patterns = c(
                "\\.jpg$", "\\.jpeg$", "\\.png$", "\\.svg$",
                "\\.tif$", "\\.tiff$"
            ), label = "ENmix QC"
        ), preprocessing = collectFigureInventoryDnamReport(
            directory = preprocessingDir,
            patterns = c(
                "\\.jpg$", "\\.jpeg$", "\\.png$", "\\.svg$",
                "\\.tif$", "\\.tiff$"
            ), label = "Quality control"
        ),
        postprocessing = collectFigureInventoryDnamReport(
            directory = postprocessingDir,
            patterns = c(
                "\\.jpg$", "\\.jpeg$", "\\.png$", "\\.svg$",
                "\\.tif$", "\\.tiff$"
            ), label = "Postprocessing"
        ),
        sva = collectFigureInventoryDnamReport(
            directory = svaDir,
            patterns = c(
                "\\.jpg$", "\\.jpeg$", "\\.png$", "\\.svg$",
                "\\.tif$", "\\.tiff$"
            ), label = "SVA"
        ), glm = collectFigureInventoryDnamReport(
            directory = glmDir,
            patterns = c(
                "\\.jpg$", "\\.jpeg$", "\\.png$", "\\.svg$",
                "\\.tif$", "\\.tiff$"
            ), label = "GLM"
        ), lme = collectFigureInventoryDnamReport(
            directory = lmeDir,
            patterns = c(
                "\\.jpg$", "\\.jpeg$", "\\.png$", "\\.svg$",
                "\\.tif$", "\\.tiff$"
            ), label = "LME"
        )
    )

    missing_directories <- vapply(
        figure_inventory, function(section) !isTRUE(section$exists),
        logical(1)
    )
    missing_directories <- names(missing_directories)[missing_directories]

    emitLogMinfiEwasWater(
        c(
            "============================================================",
            "Prepared DNA methylation dashboard report inputs", paste(
                "Output directory:",
                normalizePathDnamReport(outputDir)
            ), paste(
                "Figure directory:",
                normalizePathDnamReport(figDir)
            ), "============================================================"
        ),
        verbose = verbose, log_path = log_path
    )

    output_file <- normalizePathDnamReport(file.path(
        outputDir,
        "docs", "index.html"
    ))

    structure(list(
        output = basename(output_file),
            outputDir = normalizePathDnamReport(outputDir),
        outputFile = output_file, figDir = normalizePathDnamReport(figDir),
        figureInventory = figure_inventory,
            missingFigureDirectories = missing_directories,
        logFile = log_path
    ), class = "dnaEPICO_dnamReport_prepared")
}

#' Render a prepared DNA methylation report
#'
#' @param preparedReport Object returned by `prepareDnamReportInputs()`.
#' @param verbose Logical. If `TRUE`, emit progress messages.
#' @param logs Logical. If `TRUE`, write progress messages to a log file.
#' @param logDir Character or `NULL`. Directory for optional log files.
#' @param clean Logical. Retained for backwards compatibility.
#'
#' @return A list with class `'dnaEPICO_dnamReport_render'`.
#'
#' @examples
#' report_root <- file.path(tempdir(), "dnaepico-render-example")
#' prepared <- prepareDnamReportInputs(
#'     outputDir = file.path(report_root, "reports")
#' )
#' rendered <- renderDnamReport(prepared)
#' rendered$status
#'
#' @export
renderDnamReport <- function(
    preparedReport, verbose = FALSE,
    logs = FALSE, logDir = NULL, clean = TRUE
) {
    if (!inherits(preparedReport, "dnaEPICO_dnamReport_prepared")) {
        stop("preparedReport must be an object returned by prepareDnamReportInputs().",
            call. = FALSE
        )
    }

    log_path <- resolveLogPathMinfiEwasWater(logs = logs,
        log_dir = if (is.null(logDir)) {
        dirname(preparedReport$outputFile)
    } else {
        logDir
    }, log_file = "log_dnamReport.txt")

    emitLogMinfiEwasWater(
        c(
            "============================================================",
            "renderDnamReport() is retained for compatibility.",
            "Use dnamReport() to generate the Quarto dashboard report.",
            "============================================================"
        ),
        verbose = verbose, log_path = log_path
    )

    structure(list(
        preparedReport = preparedReport, status = "skipped",
        renderedFile = NULL,
            errorMessage = "Use dnamReport() to generate the Quarto dashboard report.",
        logFile = log_path
    ), class = "dnaEPICO_dnamReport_render")
}

#' Generate a DNA methylation dashboard report
#'
#' @param outputDir Character. Directory where the Quarto project is written.
#' @param phenoTab Character or `NULL`. CSV file shown in the Data tab.
#'   When `NULL`, the path is inferred from the Makefile output layout.
#' @param enmixTab Character. Directory containing ENmix quality-control
#' figures.
#' @param qcTab Character. Directory containing Quality Control figures.
#' @param svaTab Character. Directory containing Batch Effect or SVA figures.
#' @param metricTab Character. Directory containing Metrics figures.
#' @param glmTab Character or `NULL`. XLSX workbook shown in the GLM Analysis
#' tab.
#'   When `NULL`, the path is inferred from the Makefile output layout. Report
#'   sidecars produced by `methylationGLM()` are read from this report project's
#'   `assets/results/glm_results` directory when available.
#' @param lmeTab Character or `NULL`. XLSX workbook shown in the LME Analysis
#' tab.
#'   When `NULL`, the path is inferred from the Makefile output layout. Report
#'   sidecars produced by `methylationLME()` are read from this report project's
#'   `assets/results/lme_results` directory when available.
#' @param logTab Character. Directory containing workflow logs shown in the Logs
#'   tab.
#' @param verbose Logical. If `TRUE`, emit progress messages.
#' @param logs Logical. If `TRUE`, write a report log.
#' @param projectName Character. Name used for the generated Quarto project.
#' @param detPPath Character or `NULL`. RData file containing the detection
#'   P-value matrix object `detP`, used to build the quality-control tables.
#'   When `NULL`, the path is inferred from the Makefile output layout.
#' @param detPThreshold Numeric. Detection P-value threshold used when
#'   summarising the `detP` matrix.
#' @param cpgDetectionPath Character or `NULL`. Optional fallback CpG detection
#'   summary CSV.
#' @param sampleDetectionPath Character or `NULL`. Optional fallback sample
#'   detection summary CSV.
#' @param logoPath Character. Path to the navbar logo. Defaults to the packaged
#'   `inst/extdata/dnaEPICO.svg` asset.
#' @param imagePattern Character. Regular expression used to identify image
#'   files inside the section directories.
#' @param recursive Logical. If `TRUE`, search section directories recursively.
#'
#' @return A list with class `'dnaEPICO_dnamReport'`.
#'
#' @examples
#' report_root <- file.path(tempdir(), "dnaepico-dnam-report")
#' pheno_file <- file.path(
#'     report_root,
#'     "data",
#'     "model1",
#'     "preprocessingMinfiEwasWater",
#'     "phenoLC.csv"
#' )
#' dir.create(dirname(pheno_file), recursive = TRUE, showWarnings = FALSE)
#' utils::write.csv(
#'     data.frame(
#'         UID = c("sample1", "sample2"),
#'         Timepoint = c(1, 2),
#'         Sex = c("F", "M")
#'     ),
#'     pheno_file,
#'     row.names = FALSE
#' )
#'
#' result <- dnamReport(
#'     outputDir = file.path(report_root, "reports", "model1"),
#'     phenoTab = pheno_file,
#'     enmixTab = file.path(
#'         report_root,
#'         "figures",
#'         "model1",
#'         "preprocessingMinfiEwasWater",
#'         "enmix"
#'     ),
#'     qcTab = file.path(
#'         report_root,
#'         "figures",
#'         "model1",
#'         "preprocessingMinfiEwasWater",
#'         "qc"
#'     ),
#'     svaTab = file.path(report_root, "figures", "model1", "svaEnmix"),
#'     metricTab = file.path(
#'         report_root,
#'         "figures",
#'         "model1",
#'         "preprocessingMinfiEwasWater",
#'         "metrics"
#'     ),
#'     logTab = file.path(report_root, "logs", "model1")
#' )
#' result$status
#'
#' @export
dnamReport <- function(
    outputDir = "reports", phenoTab = NULL,
    enmixTab = file.path(
        "figures", "preprocessingMinfiEwasWater",
        "enmix"
    ), qcTab = file.path(
        "figures", "preprocessingMinfiEwasWater",
        "qc"
    ), svaTab = file.path("figures", "svaEnmix"), metricTab = file.path(
        "figures",
        "preprocessingMinfiEwasWater", "metrics"
    ), glmTab = NULL,
    lmeTab = NULL, logTab = outputDir, verbose = FALSE, logs = FALSE,
    projectName = "dnaEPICO", detPPath = NULL, detPThreshold = 0.01,
    cpgDetectionPath = NULL, sampleDetectionPath = NULL,
        logoPath = system.file("extdata",
        "dnaEPICO.svg",
        package = "dnaEPICO"
    ), imagePattern = "\\.(png|jpg|jpeg|gif|webp|svg|tif|tiff)$",
    recursive = TRUE
) {
    old_options <- options(stringsAsFactors = FALSE)
    on.exit(options(old_options), add = TRUE)
    detPThreshold <- validateProbabilityDnaEpico(
        detPThreshold,
        "detPThreshold"
    )

    is_absolute_path <- function(path) {
        grepl("^[A-Za-z]:[/\\\\]|^/", path)
    }

    infer_workflow_root <- function(output_dir) {
        normalized_output <- if (is_absolute_path(output_dir)) {
            normalizePath(output_dir, winslash = "/", mustWork = FALSE)
        } else {
            normalizePath(file.path(getwd(), output_dir),
                winslash = "/",
                mustWork = FALSE
            )
        }

        if (identical(basename(dirname(normalized_output)), "reports")) {
            return(dirname(dirname(normalized_output)))
        }
        if (identical(basename(normalized_output), "reports")) {
            return(dirname(normalized_output))
        }

        dirname(normalized_output)
    }

    root_dir <- infer_workflow_root(outputDir)
    resolve_report_path <- function(path, base = root_dir) {
        if (is.null(path) || !nzchar(path)) {
            return("")
        }
        if (is_absolute_path(path)) {
            return(normalizePath(path, winslash = "/", mustWork = FALSE))
        }
        normalizePath(file.path(base, path),
            winslash = "/",
            mustWork = FALSE
        )
    }
    project_name <- projectName
    qc_dir <- resolve_report_path(enmixTab)
    preprocessing_dir <- resolve_report_path(qcTab)
    postprocessing_dir <- resolve_report_path(metricTab)
    sva_dir <- resolve_report_path(svaTab)

    infer_model_from_path <- function(path,
        step = "preprocessingMinfiEwasWater") {
        pieces <- strsplit(gsub("\\\\", "/", normalizePath(path,
            winslash = "/", mustWork = FALSE
        )), "/", fixed = TRUE)[[1]]
        step_idx <- which(pieces == step)
        figures_idx <- which(pieces == "figures")
        if (length(step_idx) && length(figures_idx)) {
            usable <- figures_idx[figures_idx < step_idx[[1]]]
            if (length(usable) && step_idx[[1]] - usable[[length(usable)]] >=
                2L) {
                return(pieces[[step_idx[[1]] - 1L]])
            }
        }
        basename(normalizePath(outputDir, winslash = "/", mustWork = FALSE))
    }

    model_name <- infer_model_from_path(qc_dir)
    logs_dir <- resolve_report_path(logTab)

    pheno_file <- if (is.null(phenoTab) || !nzchar(phenoTab)) {
        resolve_report_path(file.path(
            "data", model_name, "preprocessingMinfiEwasWater",
            "phenoLC.csv"
        ))
    } else {
        resolve_report_path(phenoTab)
    }
    detp_path <- if (is.null(detPPath) || !nzchar(detPPath)) {
        resolve_report_path(file.path(
            "rData", model_name, "preprocessingMinfiEwasWater",
            "qc", "detP_RGSet.RData"
        ))
    } else {
        resolve_report_path(detPPath)
    }
    cpg_detection_path <- if (is.null(cpgDetectionPath) ||
        !nzchar(cpgDetectionPath)) {
        resolve_report_path(file.path(
            "data", model_name, "preprocessingMinfiEwasWater",
            "cpgD.csv"
        ))
    } else {
        resolve_report_path(cpgDetectionPath)
    }
    sample_detection_path <- if (is.null(sampleDetectionPath) ||
        !nzchar(sampleDetectionPath)) {
        resolve_report_path(file.path(
            "data", model_name, "preprocessingMinfiEwasWater",
            "sampleD.csv"
        ))
    } else {
        resolve_report_path(sampleDetectionPath)
    }
    glm_table_path <- if (is.null(glmTab) || !nzchar(glmTab)) {
        resolve_report_path(file.path(
            "data", model_name, "methylationGLM",
            "annotatedGLM.xlsx"
        ))
    } else {
        resolve_report_path(glmTab)
    }
    lme_table_path <- if (is.null(lmeTab) || !nzchar(lmeTab)) {
        resolve_report_path(file.path(
            "data", model_name, "methylationLME",
            "annotatedLME.xlsx"
        ))
    } else {
        resolve_report_path(lmeTab)
    }
    glm_dir <- resolve_report_path(file.path(
        "figures", model_name,
        "methylationGLM"
    ))
    lme_dir <- resolve_report_path(file.path(
        "figures", model_name,
        "methylationLME"
    ))
    logo_candidates <- c(logoPath, system.file("extdata", "dnaEPICO.svg",
        package = "dnaEPICO"
    ), file.path(
        root_dir, "inst", "extdata",
        "dnaEPICO.svg"
    ), file.path(root_dir, "dnaEPICO.svg"))
    logo_candidates <- logo_candidates[nzchar(logo_candidates)]
    logo_source_path <- logo_candidates[file.exists(logo_candidates)][1]
    if (is.na(logo_source_path)) {
        logo_source_path <- ""
    }
    viewer_js_candidates <- c(system.file("extdata", "cpg-viewer.js",
        package = "dnaEPICO"
    ), file.path(
        getwd(), "inst", "extdata",
        "cpg-viewer.js"
    ), file.path(
        root_dir, "inst", "extdata",
        "cpg-viewer.js"
    ))
    viewer_js_candidates <- viewer_js_candidates[nzchar(viewer_js_candidates)]
    viewer_js_source_path <-
        viewer_js_candidates[file.exists(viewer_js_candidates)][1]
    if (is.na(viewer_js_source_path)) {
        viewer_js_source_path <- ""
    }

    project_dir <- resolve_report_path(outputDir)
    assets_dir <- file.path(project_dir, "assets")
    assets_figures_dir <- file.path(assets_dir, "figures")
    assets_logs_dir <- file.path(assets_dir, "logs")
    fig_dir <- assets_figures_dir

    image_pattern <- imagePattern
    magick_available <- requireNamespace("magick", quietly = TRUE)
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = logs_dir,
        log_file = "log_dnamReport.txt"
    )

    emitLogMinfiEwasWater(
        c(
            "============================================================",
            paste("Starting DNA Methylation Report Step:", project_name),
            paste("Project root:", root_dir), paste(
                "Output project:",
                project_dir
            ), "============================================================"
        ),
        verbose = FALSE, log_path = log_path
    )

    slash <- function(path) {
        gsub("\\\\", "/", path)
    }

    write_utf8 <- function(path, lines) {
        dir.create(dirname(path), recursive = TRUE, showWarnings = FALSE)
        con <- file(path, open = "w", encoding = "UTF-8")
        on.exit(close(con), add = TRUE)
        writeLines(lines, con = con, useBytes = TRUE)
    }

    find_quarto <- function() {
        env_candidates <- c(Sys.getenv("QUARTO_BIN"), Sys.getenv("QUARTO"))
        env_candidates <- env_candidates[nzchar(env_candidates)]
        env_candidates <- normalizePath(env_candidates,
            winslash = "/",
            mustWork = FALSE
        )
        env_candidates <- env_candidates[file.exists(env_candidates)]
        if (length(env_candidates)) {
            return(env_candidates[[1]])
        }

        quarto_bin <- Sys.which("quarto")
        if (nzchar(quarto_bin)) {
            return(unname(quarto_bin))
        }

        candidates <- c(
            "C:/Program Files/RStudio/resources/app/bin/quarto/bin/quarto.exe",
            "C:/Program Files/Quarto/bin/quarto.exe", file.path(
                Sys.getenv("LOCALAPPDATA"),
                "Programs", "Quarto", "bin", "quarto.exe"
            ), file.path(
                Sys.getenv("LOCALAPPDATA"),
                "quarto", "bin", "quarto.exe"
            )
        )
        candidates <- normalizePath(candidates,
            winslash = "/",
            mustWork = FALSE
        )
        candidates <- candidates[file.exists(candidates)]
        if (length(candidates)) {
            return(candidates[[1]])
        }

        ""
    }

    html_escape <- function(text) {
        text <- gsub("&", "&amp;", text, fixed = TRUE)
        text <- gsub("<", "&lt;", text, fixed = TRUE)
        text <- gsub(">", "&gt;", text, fixed = TRUE)
        text <- gsub("\"", "&quot;", text, fixed = TRUE)
        text
    }

    r_string <- function(text) {
        text <- gsub("\\\\", "\\\\\\\\", text)
        text <- gsub("\"", "\\\"", text, fixed = TRUE)
        paste0("\"", text, "\"")
    }

    slugify <- function(text) {
        text <- tools::file_path_sans_ext(basename(text))
        text <- tolower(text)
        text <- gsub("[^a-z0-9]+", "-", text)
        text <- gsub("(^-+|-+$)", "", text)
        if (!nzchar(text)) {
            text <- "figure"
        }
        text
    }

    pretty_label <- function(path) {
        label <- tools::file_path_sans_ext(basename(path))
        label <- gsub("([a-z0-9])([A-Z])", "\\1 \\2", label,
            perl = TRUE
        )
        label <- gsub("[_\\-]+", " ", label)
        label <- gsub("\\s+", " ", label)
        label <- trimws(label)
        words <- strsplit(tolower(label), "\\s+", perl = TRUE)[[1]]
        label <- paste0(toupper(substr(words, 1L, 1L)), substr(
            words,
            2L, nchar(words)
        ))
        label <- paste(label, collapse = " ")
        label <- gsub("\\bQc\\b", "QC", label)
        label <- gsub("\\bPca\\b", "PCA", label)
        label <- gsub("\\bSva\\b", "SVA", label)
        label <- gsub("\\bEnmix\\b", "ENmix", label)
        label
    }

    callout_lines <- function(text, type = "note") {
        c(sprintf("::: {.callout-%s}", type), text, ":::", "")
    }

    copy_log_asset <- function(src_path, output_name) {
        dir.create(assets_logs_dir, recursive = TRUE, showWarnings = FALSE)
        dest_path <- file.path(assets_logs_dir, output_name)

        if (file.exists(src_path)) {
            ok <- file.copy(src_path, dest_path, overwrite = TRUE)
            if (!ok) {
                stop("Failed to copy log file: ", src_path)
            }
        }

        list(
            exists = file.exists(src_path), source_path = slash(src_path),
            asset_path = slash(file.path("assets", "logs", output_name))
        )
    }

    copy_first_existing_log_asset <- function(src_paths, output_name) {
        existing <- src_paths[file.exists(src_paths)]
        src_path <- if (length(existing) > 0L) {
            existing[[1L]]
        } else {
            src_paths[[1L]]
        }
        copy_log_asset(src_path, output_name)
    }

    copy_figure_assets <- function(src_dir, asset_subdir) {
        dir.create(file.path(assets_figures_dir, asset_subdir),
            recursive = TRUE, showWarnings = FALSE
        )

        if (!dir.exists(src_dir)) {
            return(data.frame(
                title = character(), original_name = character(),
                asset_path = character(), browser_ready = logical(),
                converted = logical(), stringsAsFactors = FALSE
            ))
        }

        files <- list.files(src_dir,
            pattern = image_pattern,
            full.names = TRUE, recursive = recursive, ignore.case = TRUE
        )
        files <- sort(files)

        if (!length(files)) {
            return(data.frame(
                title = character(), original_name = character(),
                asset_path = character(), browser_ready = logical(),
                converted = logical(), stringsAsFactors = FALSE
            ))
        }

        rows <- vector("list", length(files))

        for (idx in seq_along(files)) {
            src_path <- files[[idx]]
            src_ext <- tolower(tools::file_ext(src_path))
            base_slug <- sprintf("%02d-%s", idx, slugify(src_path))
            converted <- FALSE
            browser_ready <- TRUE

            if (src_ext %in% c("tif", "tiff") && magick_available) {
                dest_name <- paste0(base_slug, ".png")
                dest_path <- file.path(
                    assets_figures_dir, asset_subdir,
                    dest_name
                )
                converted <- tryCatch(
                    {
                        img <- magick::image_read(src_path)
                        magick::image_write(img,
                            path = dest_path,
                            format = "png"
                        )
                        TRUE
                    },
                    error = function(e) FALSE
                )

                if (!isTRUE(converted)) {
                    dest_name <- paste0(base_slug, ".", src_ext)
                    dest_path <- file.path(
                        assets_figures_dir,
                        asset_subdir, dest_name
                    )
                    ok <- file.copy(src_path, dest_path, overwrite = TRUE)
                    if (!ok) {
                        stop("Failed to copy figure: ", src_path)
                    }
                    browser_ready <- FALSE
                }
            } else {
                dest_name <- paste0(base_slug, ".", src_ext)
                dest_path <- file.path(
                    assets_figures_dir, asset_subdir,
                    dest_name
                )
                ok <- file.copy(src_path, dest_path, overwrite = TRUE)
                if (!ok) {
                    stop("Failed to copy figure: ", src_path)
                }
                browser_ready <- !(src_ext %in% c("tif", "tiff"))
            }

            rows[[idx]] <- data.frame(
                title = pretty_label(src_path),
                original_name = basename(src_path),
                    asset_path = slash(file.path(
                    "assets",
                    "figures", asset_subdir, dest_name
                )), browser_ready = browser_ready,
                converted = converted, stringsAsFactors = FALSE
            )
        }

        do.call(rbind, rows)
    }

    build_figure_cards <- function(items, empty_message,
        title_prefix = "Figure",
                                    show_filename = FALSE,
                                        figure_titles = NULL) {
        if (!nrow(items)) {
            return(callout_lines(empty_message, type = "warning"))
        }

        lines <- character()

        if (any(!items$browser_ready)) {
            lines <- c(lines, callout_lines(
                paste(
                    "Some TIFF figures were copied without PNG conversion.",
                    "Install the `magick` package and rerun `dnamReport()` if you want those figures to render directly in the browser."
                ),
                type = "warning"
            ))
        }

        for (idx in seq_len(nrow(items))) {
            row <- items[idx, , drop = FALSE]
            caption_title <- if (!is.null(figure_titles) &&
                length(figure_titles) >=
                idx && nzchar(figure_titles[[idx]])) {
                figure_titles[[idx]]
            } else if ("title" %in% names(row) && nzchar(row$title[[1]])) {
                row$title[[1]]
            } else {
                sprintf("%s %s", title_prefix, idx)
            }
            lines <- c(lines, sprintf(
                "::: {.card title=\"%s\"}",
                html_escape(caption_title)
            ), "")

            if (isTRUE(row$browser_ready)) {
                lines <- c(
                    lines, "<figure class=\"qpasst-figure-card\">",
                    sprintf(
                        "  <img src=\"%s\" alt=\"%s\" loading=\"lazy\" />",
                        row$asset_path, html_escape(row$title)
                    )
                )
                if (isTRUE(show_filename)) {
                    lines <- c(lines, sprintf(
                        "  <figcaption><code>%s</code></figcaption>",
                        html_escape(row$original_name)
                    ))
                }
                lines <- c(lines, "</figure>", "")
            } else {
                download_label <- if (isTRUE(show_filename)) {
                    row$original_name
                } else {
                    caption_title
                }
                lines <- c(
                    lines, sprintf(
                        "<p class=\"qpasst-download\">Browser preview is unavailable for this TIFF file. <a href=\"%s\">Download %s</a>.</p>",
                        row$asset_path, html_escape(download_label)
                    ),
                    ""
                )
            }
            lines <- c(lines, ":::", "")
        }

        lines
    }

    js_quote_result_values <- function(values) {
        missing <- is.na(values)
        values <- enc2utf8(as.character(values))
        values <- gsub("\\", "\\\\", values, fixed = TRUE)
        values <- gsub("\"", "\\\"", values, fixed = TRUE)
        values <- gsub("\r", "\\r", values, fixed = TRUE)
        values <- gsub("\n", "\\n", values, fixed = TRUE)
        values <- gsub("\t", "\\t", values, fixed = TRUE)
        values <- gsub("<", "\\u003c", values, fixed = TRUE)
        values <- gsub(">", "\\u003e", values, fixed = TRUE)
        values <- gsub("&", "\\u0026", values, fixed = TRUE)
        values <- paste0("\"", values, "\"")
        values[missing] <- "null"
        values
    }

    js_result_array <- function(values) {
        paste0(
            "[", paste(js_quote_result_values(values), collapse = ","),
            "]"
        )
    }

    read_optional_workbook_sheet <- function(path, sheets, sheet) {
        if (!(sheet %in% sheets)) {
            return(data.frame())
        }
        tryCatch(openxlsx::read.xlsx(path, sheet = sheet, check.names = FALSE),
            error = function(e) data.frame()
        )
    }

    metadata_frame_to_list <- function(metadata) {
        if (!is.data.frame(metadata) || !all(c("Key", "Value") %in%
            names(metadata))) {
            return(list())
        }
        values <- as.character(metadata$Value)
        names(values) <- as.character(metadata$Key)
        as.list(values[nzchar(names(values))])
    }

    split_report_metadata_values <- function(value) {
        if (is.null(value) || !length(value) || is.na(value[[1L]]) ||
            !nzchar(value[[1L]])) {
            return(character())
        }
        values <- trimws(unlist(strsplit(as.character(value[[1L]]),
            ",",
            fixed = TRUE
        ), use.names = FALSE))
        unique(values[nzchar(values)])
    }

    extract_report_formula_phenotype <- function(formula) {
        formula <- as.character(formula[[1L]])
        rhs <- trimws(sub("^[^~]*~", "", formula))
        quoted_term <- regmatches(rhs, regexpr("`[^`]+`", rhs,
            perl = TRUE
        ))
        if (length(quoted_term) && nzchar(quoted_term) && !identical(
            quoted_term,
            character(0)
        )) {
            return(substring(quoted_term, 2L, nchar(quoted_term) -
                1L))
        }
        trimws(sub("\\s*(?:\\+|\\*|:|\\|).*", "", rhs, perl = TRUE))
    }

    resolve_report_formula_records <- function(dictionary, metadata = list()) {
        empty_records <- data.frame(
            phenotype = character(),
            result_column = character(), formula = character(),
            stringsAsFactors = FALSE, check.names = FALSE
        )
        if (!is.data.frame(dictionary) || !("Formula" %in% names(dictionary))) {
            return(empty_records)
        }

        keep <- !is.na(dictionary$Formula) & nzchar(trimws(as.character(dictionary$Formula)))
        if (!any(keep)) {
            return(empty_records)
        }
        formula_rows <- dictionary[keep, , drop = FALSE]
        formulas <- as.character(formula_rows$Formula)
        result_columns <- if ("Column" %in% names(formula_rows)) {
            as.character(formula_rows$Column)
        } else {
            rep("", length(formulas))
        }
        explicit_phenotypes <- if ("Phenotype" %in% names(formula_rows)) {
            as.character(formula_rows$Phenotype)
        } else {
            rep("", length(formulas))
        }
        known_phenotypes <- split_report_metadata_values(metadata$phenotypes)

        phenotype_labels <- vapply(seq_along(formulas), function(index) {
            if (!is.na(explicit_phenotypes[[index]]) &&
                nzchar(trimws(explicit_phenotypes[[index]]))) {
                return(trimws(explicit_phenotypes[[index]]))
            }
            if (length(known_phenotypes)) {
                matches <- known_phenotypes[vapply(
                    known_phenotypes,
                    function(phenotype) {
                        grepl(paste0("`", phenotype, "`"), formulas[[index]],
                            fixed = TRUE
                        ) || startsWith(
                            result_columns[[index]],
                            phenotype
                        )
                    }, logical(1)
                )]
                if (length(matches)) {
                    return(matches[[1L]])
                }
            }
            extracted <- extract_report_formula_phenotype(formulas[[index]])
            if (nzchar(extracted)) {
                extracted
            } else {
                result_columns[[index]]
            }
        }, character(1))

        records <- data.frame(
            phenotype = phenotype_labels, result_column = result_columns,
            formula = formulas, stringsAsFactors = FALSE, check.names = FALSE
        )
        duplicate_key <- paste(records$phenotype, records$formula,
            sep = "\r"
        )
        records <- records[!duplicated(duplicate_key), , drop = FALSE]
        rownames(records) <- NULL
        records
    }

    last_log_field <- function(path, label) {
        if (!file.exists(path)) {
            return("")
        }
        lines <- readLines(path, warn = FALSE, encoding = "UTF-8")
        matches <- grep(paste0("^", label, ":"), trimws(lines),
            value = TRUE
        )
        if (!length(matches)) {
            return("")
        }
        trimws(sub("^[^:]+:", "", matches[[length(matches)]]))
    }

    resolve_lme_report_metadata <- function(metadata, dictionary,
                                            log_path) {
        values <- metadata_frame_to_list(metadata)
        fallback <- list(libraries = last_log_field(
            log_path,
            "LME libraries"
        ), correlation_structure = last_log_field(
            log_path,
            "Correlation structure"
        ), correlation_variable = last_log_field(
            log_path,
            "Correlation variable"
        ), interaction_term = last_log_field(
            log_path,
            "Interaction term"
        ), annotation_columns = last_log_field(
            log_path,
            "Annotation columns used"
        ), missing_annotation_columns = last_log_field(
            log_path,
            "Missing annotation columns"
        ))
        for (key in names(fallback)) {
            if (is.null(values[[key]]) || !nzchar(values[[key]])) {
                values[[key]] <- fallback[[key]]
            }
        }

        if (is.null(values$backend) || !nzchar(values$backend)) {
            values$backend <- if (grepl("nlme", values$libraries,
                ignore.case = TRUE
            )) {
                "nlme"
            } else if (grepl("lme4|lmerTest", values$libraries,
                ignore.case = TRUE
            )) {
                "lme4"
            } else {
                ""
            }
        }
        if (is.null(values$fitting_function) ||
            !nzchar(values$fitting_function)) {
            values$fitting_function <- if (identical(
                values$backend,
                "nlme"
            )) {
                "nlme::lme"
            } else if (identical(values$backend, "lme4")) {
                "lmerTest::lmer"
            } else {
                ""
            }
        }
        if (is.null(values$correlation_structure) ||
            !nzchar(values$correlation_structure)) {
            values$correlation_structure <- "none"
        }
        if (is.null(values$correlation_variable) ||
            !nzchar(values$correlation_variable)) {
            values$correlation_variable <- "None"
        }
        if (is.null(values$interaction_term) ||
            !nzchar(values$interaction_term)) {
            values$interaction_term <- "None"
        }
        values
    }

    same_report_path <- function(source, target) {
        source <- normalizePath(source, winslash = "/", mustWork = FALSE)
        target <- normalizePath(target, winslash = "/", mustWork = FALSE)
        if (identical(.Platform$OS.type, "windows")) {
            identical(tolower(source), tolower(target))
        } else {
            identical(source, target)
        }
    }

    prepare_table_viewer_directory <- function(var_prefix) {
        result_dir <- file.path(assets_dir, "results", var_prefix)
        dir.create(result_dir, recursive = TRUE, showWarnings = FALSE)
        stale_chunks <- list.files(result_dir,
            pattern = "^chunk-[0-9]+\\.js$",
            full.names = TRUE
        )
        if (length(stale_chunks)) {
            unlink(stale_chunks, force = TRUE)
        }
        result_dir
    }

    write_table_viewer_chunk <- function(chunk, chunk_number,
                                            result_dir, var_prefix, id_column) {
        encoded_columns <- lapply(chunk, js_quote_result_values)
        encoded_rows <- if (nrow(chunk)) {
            do.call(paste, c(encoded_columns, sep = ","))
        } else {
            character()
        }
        chunk_key <- paste0(var_prefix, ":", chunk_number)
        write_utf8(file.path(result_dir, sprintf(
            "chunk-%04d.js",
            chunk_number
        )), paste0(
            "window.dnaEPICOResultChunks=window.dnaEPICOResultChunks||{};",
            "window.dnaEPICOResultChunks[", js_quote_result_values(chunk_key),
            "]={rows:[", paste0("[", encoded_rows, "]", collapse = ","),
            "]};"
        ))
        ids <- if (nzchar(id_column)) {
            as.character(chunk[[id_column]])
        } else {
            rep("", nrow(chunk))
        }
        ids <- ids[!is.na(ids)]
        data.frame(number = chunk_number, first_id = if (length(ids)) {
            ids[[1L]]
        } else {
            ""
        }, last_id = if (length(ids)) {
            ids[[length(ids)]]
        } else {
            ""
        }, stringsAsFactors = FALSE)
    }

    write_table_viewer_manifest <- function(columns, n_rows,
                                            chunk_size, chunk_index, var_prefix,
                                                id_column, downloads,
                                            item_singular, item_plural) {
        chunk_objects <- if (nrow(chunk_index)) {
            vapply(seq_len(nrow(chunk_index)), function(i) {
                paste0(
                    "{\"number\":", chunk_index$number[[i]],
                    ",\"firstId\":",
                        js_quote_result_values(chunk_index$first_id[[i]]),
                    ",\"lastId\":",
                        js_quote_result_values(chunk_index$last_id[[i]]),
                    "}"
                )
            }, character(1))
        } else {
            character()
        }
        download_objects <- if (length(downloads)) {
            vapply(downloads, function(download) {
                paste0(
                    "{\"href\":", js_quote_result_values(download$href),
                    ",\"label\":", js_quote_result_values(download$label),
                    "}"
                )
            }, character(1))
        } else {
            character()
        }
        manifest <- paste0(
            "window.dnaEPICOResultManifests=window.dnaEPICOResultManifests||{};",
            "window.dnaEPICOResultManifests[",
                js_quote_result_values(var_prefix),
            "]={", "\"key\":", js_quote_result_values(var_prefix),
            ",", "\"columns\":", js_result_array(columns), ",",
            "\"totalRows\":", n_rows, ",", "\"chunkSize\":",
            chunk_size, ",", "\"maxCachedChunks\":4,", "\"idColumn\":",
            js_quote_result_values(id_column), ",", "\"basePath\":",
            js_quote_result_values(paste0(
                "assets/results/",
                var_prefix
            )), ",", "\"itemSingular\":", js_quote_result_values(item_singular),
            ",", "\"itemPlural\":", js_quote_result_values(item_plural),
            ",", "\"downloads\":[", paste(download_objects, collapse = ","),
            "],", "\"chunks\":[", paste(chunk_objects, collapse = ","),
            "]};"
        )
        result_dir <- file.path(assets_dir, "results", var_prefix)
        write_utf8(file.path(result_dir, "manifest.js"), manifest)
    }

    table_viewer_asset_result <- function(columns, n_rows, var_prefix,
                                            id_column, item_singular,
                                                item_plural, search_label,
                                            search_placeholder) {
        list(
            key = var_prefix, columns = columns, n_rows = n_rows,
            n_cols = length(columns), id_column = id_column,
            item_singular = item_singular, item_plural = item_plural,
            search_label = search_label,
                search_placeholder = search_placeholder,
            manifest_path = paste0(
                "assets/results/", var_prefix,
                "/manifest.js"
            )
        )
    }

    write_table_viewer_assets <- function(table_data, var_prefix,
                                            id_column, downloads,
                                                chunk_size = 5000L,
                                                item_singular = "CpG",
                                            item_plural = "CpGs",
                                                search_label = "Find CpG",
                                                search_placeholder = "e.g. cg00000029") {
        table_data <- as.data.frame(table_data,
            stringsAsFactors = FALSE,
            check.names = FALSE
        )
        chunk_size <- max(100L, as.integer(chunk_size))
        result_dir <- prepare_table_viewer_directory(var_prefix)

        n_rows <- nrow(table_data)
        n_chunks <- if (n_rows) {
            ceiling(n_rows / chunk_size)
        } else {
            0L
        }
        chunk_rows <- vector("list", n_chunks)
        if (n_chunks) {
            for (chunk_idx in seq_len(n_chunks)) {
                start <- (chunk_idx - 1L) * chunk_size + 1L
                end <- min(chunk_idx * chunk_size, n_rows)
                chunk <- table_data[start:end, , drop = FALSE]
                chunk_rows[[chunk_idx]] <- write_table_viewer_chunk(
                    chunk = chunk,
                    chunk_number = chunk_idx, result_dir = result_dir,
                    var_prefix = var_prefix, id_column = id_column
                )
            }
        }
        chunk_index <- if (length(chunk_rows)) {
            do.call(rbind, chunk_rows)
        } else {
            data.frame()
        }
        write_table_viewer_manifest(
            columns = names(table_data),
            n_rows = n_rows, chunk_size = chunk_size, chunk_index = chunk_index,
            var_prefix = var_prefix, id_column = id_column,
                downloads = downloads,
            item_singular = item_singular, item_plural = item_plural
        )

        table_viewer_asset_result(
            columns = names(table_data),
            n_rows = n_rows, var_prefix = var_prefix, id_column = id_column,
            item_singular = item_singular, item_plural = item_plural,
            search_label = search_label, search_placeholder = search_placeholder
        )
    }

    write_delimited_table_viewer_assets <- function(table_path,
                                                    var_prefix, id_column,
                                                        downloads,
                                                        expected_rows,
                                                        expected_columns,
                                                    chunk_size = 5000L,
                                                        item_singular = "CpG",
                                                        item_plural = "CpGs",
                                                    search_label = "Find CpG",
                                                        search_placeholder = "e.g. cg00000029") {
        chunk_size <- max(100L, as.integer(chunk_size))
        result_dir <- prepare_table_viewer_directory(var_prefix)
        chunk_state <- new.env(parent = emptyenv())
        chunk_state$rows <- list()
        stream_result <- streamReportTableDnaEpico(
            tableFile = table_path,
            chunkSize = chunk_size, expectedRows = expected_rows,
            expectedColumns = expected_columns, chunkHandler = function(chunk,
                                                                        chunk_number) {
                chunk_state$rows[[chunk_number]] <- write_table_viewer_chunk(
                    chunk = chunk,
                    chunk_number = chunk_number, result_dir = result_dir,
                    var_prefix = var_prefix, id_column = id_column
                )
            }
        )
        chunk_rows <- chunk_state$rows
        columns <- stream_result$columns
        n_rows <- stream_result$nRows
        if (!(id_column %in% columns)) {
            stop("The report table identifier column was not found.",
                call. = FALSE
            )
        }
        chunk_index <- if (length(chunk_rows)) {
            do.call(rbind, chunk_rows)
        } else {
            data.frame()
        }
        write_table_viewer_manifest(
            columns = columns, n_rows = n_rows,
            chunk_size = chunk_size, chunk_index = chunk_index,
            var_prefix = var_prefix, id_column = id_column,
                downloads = downloads,
            item_singular = item_singular, item_plural = item_plural
        )

        table_viewer_asset_result(
            columns = columns, n_rows = n_rows,
            var_prefix = var_prefix, id_column = id_column,
                item_singular = item_singular,
            item_plural = item_plural, search_label = search_label,
            search_placeholder = search_placeholder
        )
    }

    prepare_xlsx_table_assets <- function(data_path, sheet, var_prefix,
                                            analysis = c("glm", "lme"),
                                                chunk_size = 5000L,
                                                model_log_path = "") {
        analysis <- match.arg(analysis)
        source_data_path <- if (grepl(
            "^[A-Za-z]:[/\\\\]|^/",
            data_path
        )) {
            data_path
        } else {
            file.path(root_dir, data_path)
        }
        if (!file.exists(source_data_path)) {
            return(list(
                ok = FALSE, error = paste0(
                    "Data file not found at `",
                    slash(source_data_path), "`."
                ), source_path = source_data_path,
                analysis = analysis, metadata = list()
            ))
        }

        result_dir <- file.path(assets_dir, "results", var_prefix)
        dir.create(result_dir, recursive = TRUE, showWarnings = FALSE)
        xlsx_name <- basename(source_data_path)
        legacy_xlsx_target <- file.path(result_dir, xlsx_name)
        if (!same_report_path(source_data_path, legacy_xlsx_target) &&
            file.exists(legacy_xlsx_target)) {
            unlink(legacy_xlsx_target, force = TRUE)
        }

        sidecar <- resolveReportTableSidecarDnaEpico(
            workbookFile = source_data_path,
            sidecarDir = result_dir,
            sheet = sheet
        )
        if (isTRUE(sidecar$ok)) {
            dictionary <- tryCatch(utils::read.delim(sidecar$dictionary,
                check.names = FALSE, stringsAsFactors = FALSE,
                quote = "\"", comment.char = ""
            ), error = function(error) error)
            workbook_metadata <- if (file.exists(sidecar$workbookMetadata)) {
                tryCatch(utils::read.delim(sidecar$workbookMetadata,
                    check.names = FALSE, stringsAsFactors = FALSE,
                    quote = "\"", comment.char = ""
                ), error = function(error) error)
            } else {
                data.frame()
            }
            if (inherits(dictionary, "error") || inherits(
                workbook_metadata,
                "error"
            )) {
                sidecar$ok <- FALSE
            }
        }
        if (!isTRUE(sidecar$ok)) {
            sheets <- tryCatch(openxlsx::getSheetNames(source_data_path),
                error = function(e) character()
            )
            dictionary <- read_optional_workbook_sheet(
                source_data_path,
                sheets, "dictionary"
            )
            workbook_metadata <- read_optional_workbook_sheet(
                source_data_path,
                sheets, "metadata"
            )
        }
        metadata <- if (identical(analysis, "lme")) {
            resolve_lme_report_metadata(
                workbook_metadata, dictionary,
                model_log_path
            )
        } else {
            metadata_frame_to_list(workbook_metadata)
        }
        metadata$formula_records <- resolve_report_formula_records(
            dictionary,
            metadata
        )
        metadata$formulas <- unique(metadata$formula_records$formula)

        text_name <- paste0(sheet, ".tsv.gz")
        text_target <- file.path(result_dir, text_name)
        downloads <- list(list(href = paste0(
                "assets/results/", var_prefix,
                "/", text_name
            ), label = "Download all results (TSV.gz)")
        )

        viewer_assets <- NULL
        source_mode <- "xlsx_fallback"
        if (isTRUE(sidecar$ok)) {
            sidecar_staged <- same_report_path(
                sidecar$table,
                text_target
            ) || isTRUE(file.copy(sidecar$table,
                text_target,
                overwrite = TRUE
            ))
            if (sidecar_staged && file.exists(text_target)) {
                viewer_assets <- tryCatch(
                    write_delimited_table_viewer_assets(
                        table_path = text_target,
                        var_prefix = var_prefix, id_column = sidecar$idColumn,
                        downloads = downloads, expected_rows = sidecar$rows,
                        expected_columns = sidecar$columns,
                            chunk_size = chunk_size
                    ),
                    error = function(error) NULL
                )
                if (!is.null(viewer_assets)) {
                    source_mode <- "streamed_sidecar"
                }
            }
        }

        if (is.null(viewer_assets)) {
            table_data <- tryCatch(openxlsx::read.xlsx(source_data_path,
                sheet = sheet, check.names = FALSE
            ), error = function(e) e)
            if (inherits(table_data, "error")) {
                return(list(
                    ok = FALSE, error = conditionMessage(table_data),
                    source_path = source_data_path, analysis = analysis,
                    metadata = list()
                ))
            }
            report_table <- sortReportTableDnaEpico(table_data)
            table_data <- report_table$data
            data.table::fwrite(table_data,
                file = text_target,
                sep = "\t", quote = TRUE, na = "", compress = "gzip",
                showProgress = FALSE
            )
            viewer_assets <- write_table_viewer_assets(
                table_data = table_data,
                var_prefix = var_prefix, id_column = report_table$idColumn,
                downloads = downloads, chunk_size = chunk_size
            )
        }

        c(list(
            ok = TRUE, error = NULL, source_path = source_data_path,
            analysis = analysis, dictionary = dictionary, metadata = metadata,
            source_mode = source_mode, text_path = paste0(
                "assets/results/",
                var_prefix, "/", text_name
            )
        ), viewer_assets)
    }

    prepare_csv_table_assets <- function(data_path, var_prefix,
                                            front_columns = character(),
                                                chunk_size = 5000L) {
        source_data_path <- if (grepl(
            "^[A-Za-z]:[/\\\\]|^/",
            data_path
        )) {
            data_path
        } else {
            file.path(root_dir, data_path)
        }
        if (!file.exists(source_data_path)) {
            return(list(
                ok = FALSE, error = paste0(
                    "Data file not found at `",
                    slash(source_data_path), "`."
                ), source_path = source_data_path,
                metadata = list()
            ))
        }

        table_data <- tryCatch(utils::read.csv(source_data_path,
            check.names = FALSE
        ), error = function(e) e)
        if (inherits(table_data, "error")) {
            return(list(
                ok = FALSE, error = conditionMessage(table_data),
                source_path = source_data_path, metadata = list()
            ))
        }
        table_data <- as.data.frame(table_data,
            stringsAsFactors = FALSE,
            check.names = FALSE
        )
        front_columns <- front_columns[nzchar(front_columns) &
            front_columns %in% names(table_data)]
        table_data <- table_data[, c(front_columns, setdiff(
            names(table_data),
            front_columns
        )), drop = FALSE]
        id_column <- if (length(front_columns)) {
            front_columns[[1L]]
        } else if (ncol(table_data)) {
            names(table_data)[[1L]]
        } else {
            ""
        }
        sort_columns <- unique(c(id_column, front_columns[-1L]))
        sort_columns <- sort_columns[nzchar(sort_columns) & sort_columns %in%
            names(table_data)]
        if (length(sort_columns) && nrow(table_data) > 1L) {
            sort_values <- lapply(table_data[sort_columns], as.character)
            table_data <- table_data[do.call(order, c(
                sort_values,
                list(na.last = TRUE)
            )), , drop = FALSE]
            rownames(table_data) <- NULL
        }

        result_dir <- file.path(assets_dir, "results", var_prefix)
        dir.create(result_dir, recursive = TRUE, showWarnings = FALSE)
        csv_name <- basename(source_data_path)
        csv_target <- file.path(result_dir, csv_name)
        staged <- same_report_path(source_data_path, csv_target) ||
            isTRUE(file.copy(source_data_path, csv_target, overwrite = TRUE))
        if (!staged || !file.exists(csv_target)) {
            return(list(
                ok = FALSE, error = paste0(
                    "Could not copy the data file into the report assets from `",
                    slash(source_data_path), "`."
                ), source_path = source_data_path,
                metadata = list()
            ))
        }

        search_label <- if (nzchar(id_column)) {
            paste("Find", id_column)
        } else {
            "Find row"
        }
        first_id <- if (nzchar(id_column) && nrow(table_data)) {
            as.character(table_data[[id_column]][[1L]])
        } else {
            ""
        }
        viewer_assets <- write_table_viewer_assets(
            table_data = table_data,
            var_prefix = var_prefix, id_column = id_column,
                downloads = list(list(href = paste0(
                "assets/results/",
                var_prefix, "/", csv_name
            ), label = "Download data (CSV)")),
            chunk_size = chunk_size, item_singular = "record",
            item_plural = "records", search_label = search_label,
            search_placeholder = if (nzchar(first_id)) {
                paste("e.g.", first_id)
            } else {
                "Enter an identifier"
            }
        )

        c(list(
            ok = TRUE, error = NULL, source_path = source_data_path,
            metadata = list(), csv_path = paste0(
                "assets/results/",
                var_prefix, "/", csv_name
            )
        ), viewer_assets)
    }

    build_result_table_section <- function(title, table_assets) {
        if (!isTRUE(table_assets$ok)) {
            return(c(
                sprintf("::: {.card title=\"%s\"}", html_escape(title)),
                "", callout_lines(table_assets$error, type = "warning"),
                ":::", ""
            ))
        }

        search_label <- if (!is.null(table_assets$search_label) &&
            nzchar(table_assets$search_label)) {
            table_assets$search_label
        } else {
            "Find identifier"
        }
        search_placeholder <- if (!is.null(table_assets$search_placeholder) &&
            nzchar(table_assets$search_placeholder)) {
            table_assets$search_placeholder
        } else {
            "Enter an identifier"
        }
        pagination_lines <- function(position) {
            c(
                sprintf(
                    "  <div class=\"dnaepico-viewer-pagination dnaepico-viewer-pagination-%s\">",
                    position
                ), "    <button type=\"button\" data-role=\"first\" title=\"First page\" aria-label=\"First page\">&laquo;</button>",
                "    <button type=\"button\" data-role=\"previous\" title=\"Previous page\" aria-label=\"Previous page\">&lsaquo;</button>",
                "    <label>Page <input data-role=\"page-number\" type=\"number\" min=\"1\" value=\"1\" /></label>",
                "    <span data-role=\"page-count\"></span>",
                "    <button type=\"button\" data-role=\"next\" title=\"Next page\" aria-label=\"Next page\">&rsaquo;</button>",
                "    <button type=\"button\" data-role=\"last\" title=\"Last page\" aria-label=\"Last page\">&raquo;</button>",
                "  </div>"
            )
        }

        c(
            sprintf("::: {.card title=\"%s\"}", html_escape(title)),
            "", sprintf(
                "<div class=\"dnaepico-cpg-viewer\" data-result-key=\"%s\">",
                html_escape(table_assets$key)
            ), "", "```{=html}",
            "  <div class=\"dnaepico-viewer-toolbar\">",
                "    <label>Rows per page <select data-role=\"page-size\"><option>25</option><option>50</option><option>100</option></select></label>",
            sprintf(
                "    <label>%s <input data-role=\"cpg-search\" type=\"search\" placeholder=\"%s\" /></label>",
                html_escape(search_label), html_escape(search_placeholder)
            ),
            "    <button type=\"button\" data-role=\"find-cpg\">Find</button>",
            "  </div>", "  <div class=\"dnaepico-viewer-filter\">",
            "    <label>Filter column <select data-role=\"filter-column\"><option value=\"\">Choose a column</option></select></label>",
            "    <label>Condition <select data-role=\"filter-operator\"><option value=\"contains\">contains</option><option value=\"equals\">equals</option><option value=\"lt\">&lt;</option><option value=\"lte\">&le;</option><option value=\"gt\">&gt;</option><option value=\"gte\">&ge;</option></select></label>",
            "    <label>Value <input data-role=\"filter-value\" type=\"text\" placeholder=\"e.g. 0.05\" /></label>",
            "    <button type=\"button\" data-role=\"apply-filter\">Apply filter</button>",
            "    <button type=\"button\" data-role=\"clear-filter\" disabled>Clear</button>",
            "    <span data-role=\"filter-summary\" aria-live=\"polite\"></span>",
            "  </div>",
                "  <div class=\"dnaepico-viewer-downloads\" data-role=\"downloads\"></div>",
            "  <div class=\"dnaepico-viewer-status\" data-role=\"status\" aria-live=\"polite\">Loading results&hellip;</div>",
            pagination_lines("top"),
                "  <div class=\"dnaepico-viewer-table-wrap\">",
            "    <table class=\"table table-striped table-sm dnaepico-viewer-table\">",
            "      <thead data-role=\"head\"></thead>",
                "      <tbody data-role=\"body\"></tbody>",
            "    </table>", "  </div>", pagination_lines("bottom"),
            "  <noscript>JavaScript is required for paged browsing. Use a complete-file download instead.</noscript>",
            "```", "", "</div>", sprintf(
                "<script src=\"%s\"></script>",
                table_assets$manifest_path
            ), "<script src=\"assets/cpg-viewer.js\"></script>",
            ":::", ""
        )
    }

    build_data_frame_table_section <- function(title, data, empty_message,
                                                preview_rows = 25L) {
        if (!is.data.frame(data) || !nrow(data)) {
            return(c(
                sprintf("::: {.card title=\"%s\"}", html_escape(title)),
                "", callout_lines(empty_message, type = "warning"),
                ":::", ""
            ))
        }

        if (is.finite(preview_rows) && nrow(data) > as.integer(preview_rows)) {
            data <- utils::head(data, as.integer(preview_rows))
        }

        header_cells <- paste0(
            "    <th>", html_escape(names(data)),
            "</th>"
        )
        body_rows <- unlist(lapply(seq_len(nrow(data)), function(i) {
            row_values <- vapply(data[i, , drop = FALSE], function(value) {
                if (is.na(value)) {
                    ""
                } else {
                    as.character(value)
                }
            }, character(1))
            c("  <tr>", paste0(
                "    <td>", html_escape(row_values),
                "</td>"
            ), "  </tr>")
        }), use.names = FALSE)

        c(
            sprintf("::: {.card title=\"%s\"}", html_escape(title)),
            "", "<div class=\"table-responsive\">",
                "  <table class=\"table table-striped table-sm\">",
            "  <thead>", "  <tr>", header_cells, "  </tr>", "  </thead>",
            "  <tbody>", body_rows, "  </tbody>", "  </table>",
            "</div>", ":::", ""
        )
    }

    format_count <- function(value) {
        if (!length(value) || is.na(value)) {
            return("not available")
        }
        format(as.integer(round(value)),
            big.mark = ",", scientific = FALSE,
            trim = TRUE
        )
    }

    format_model_formula <- function(value) {
        value <- paste(as.character(value), collapse = "")
        value <- gsub("`", "", value, fixed = TRUE)
        value <- gsub("\\s+", " ", value, perl = TRUE)
        value <- gsub("\\s+\\)", ")", value, perl = TRUE)
        trimws(value)
    }

    build_model_formula_notes <- function(formula_records, model_label) {
        if (!is.data.frame(formula_records) || !nrow(formula_records)) {
            return(character())
        }

        display_formulas <- vapply(
            formula_records$formula, format_model_formula,
            character(1)
        )
        if (identical(tolower(model_label), "nlme")) {
            display_formulas <- sub("^LME:", "nlme:", display_formulas,
                ignore.case = TRUE
            )
        }
        if (nrow(formula_records) == 1L) {
            return(paste0(
                "The recorded model formula is <code class=\"dnaepico-model-formula\">",
                html_escape(display_formulas[[1L]]), "</code>."
            ))
        }

        formula_items <- vapply(
            seq_len(nrow(formula_records)),
            function(index) {
                phenotype <- formula_records$phenotype[[index]]
                if (is.na(phenotype) || !nzchar(trimws(phenotype))) {
                    phenotype <- paste("Model", index)
                }
                paste0(
                    "<div class=\"dnaepico-model-formula-item\">",
                    "<strong>", html_escape(phenotype), "</strong>",
                    "<code class=\"dnaepico-model-formula\">",
                    html_escape(display_formulas[[index]]), "</code>",
                    "</div>"
                )
            }, character(1)
        )

        c(sprintf(
            "%d phenotype-specific models were fitted.",
            nrow(formula_records)
        ), paste0(
            "<details class=\"dnaepico-model-formulas\">",
            "<summary>View recorded model formulas (", nrow(formula_records),
            ")</summary>", "<div class=\"dnaepico-model-formula-list\">",
            paste(formula_items, collapse = ""), "</div>", "</details>"
        ))
    }

    format_decimal <- function(value, digits = 2L) {
        if (!length(value) || is.na(value)) {
            return("not available")
        }
        format(round(as.numeric(value), digits),
            nsmall = digits,
            scientific = FALSE, trim = TRUE
        )
    }

    plural <- function(n, singular, plural_form = paste0(
                            singular,
                            "s"
                        )) {
        if (length(n) && !is.na(n) && as.integer(n) == 1L) {
            singular
        } else {
            plural_form
        }
    }

    is_numeric_like <- function(values) {
        values <- as.character(values)
        grepl("^\\s*[+-]?(?:\\d+\\.?\\d*|\\.\\d+)(?:[eE][+-]?\\d+)?\\s*$",
            values,
            perl = TRUE
        )
    }

    sort_values <- function(values) {
        values <- unique(as.character(values[!is.na(values)]))
        values <- values[nzchar(trimws(values))]
        if (!length(values)) {
            return(character())
        }
        if (all(is_numeric_like(values))) {
            numeric_values <- as.numeric(values)
            values[order(numeric_values)]
        } else {
            sort(values)
        }
    }

    collapse_values <- function(values) {
        values <- sort_values(values)
        if (!length(values)) {
            return("not available")
        }
        if (length(values) == 1L) {
            return(values[[1]])
        }
        if (length(values) == 2L) {
            return(paste(values, collapse = " and "))
        }
        paste0(
            paste(values[-length(values)], collapse = ", "),
            ", and ", values[[length(values)]]
        )
    }

    is_blank <- function(values) {
        is.na(values) | !nzchar(trimws(as.character(values)))
    }

    safe_read_table_file <- function(path, sheet = 1) {
        if (!file.exists(path)) {
            return(NULL)
        }
        extension <- tolower(sub("^.*\\.([^.]+)$", "\\1", basename(path)))
        tryCatch(if (extension %in% c("xlsx", "xlsm")) {
            openxlsx::read.xlsx(path, sheet = sheet, check.names = FALSE)
        } else {
            utils::read.csv(path, check.names = FALSE)
        }, error = function(e) NULL)
    }

    read_detection_tables <- function(detp_path, threshold, cpg_path = "",
                                        sample_path = "") {
        result <- list(
            source = "detP", path = slash(detp_path),
            exists = FALSE, threshold = threshold, cpg = NULL,
            sample = NULL, error = NULL
        )

        if (file.exists(detp_path)) {
            loaded <- tryCatch(
                {
                    detp_env <- new.env(parent = emptyenv())
                    object_names <- load(detp_path, envir = detp_env)
                    detp_object <- if ("detP" %in% object_names) {
                        detp_env$detP
                    } else {
                        matrix_objects <- object_names[vapply(
                            object_names,
                            function(name) {
                                obj <- get(name, envir = detp_env)
                                is.matrix(obj) || is.data.frame(obj)
                            }, logical(1)
                        )]
                        if (length(matrix_objects)) {
                            get(matrix_objects[[1]], envir = detp_env)
                        } else {
                            NULL
                        }
                    }

                    if (is.null(detp_object)) {
                        stop("No `detP` matrix-like object was found.",
                            call. = FALSE
                        )
                    }

                    detp_matrix <- as.matrix(detp_object)
                    storage.mode(detp_matrix) <- "numeric"
                    if (is.null(colnames(detp_matrix))) {
                        colnames(detp_matrix) <- sprintf(
                            "Sample_%s",
                            seq_len(ncol(detp_matrix))
                        )
                    }

                    assessed <- is.finite(detp_matrix) & detp_matrix >=
                        0 & detp_matrix <= 1
                    detected <- assessed & detp_matrix < threshold
                    assessed_by_cpg <- rowSums(assessed)
                    detected_by_cpg <- rowSums(detected)
                    cpg <- data.frame(
                        metric = c(
                            "Total CpGs assessed",
                            "CpGs detected in at least one sample",
                                "CpGs detected in all samples",
                            "CpGs never detected",
                                "CpGs with no valid detection P values"
                        ),
                        nCpGs = c(nrow(detp_matrix), sum(detected_by_cpg >=
                            1L), sum(assessed_by_cpg == ncol(detp_matrix) &
                            detected_by_cpg == ncol(detp_matrix)),
                                sum(assessed_by_cpg >
                            0L & detected_by_cpg == 0L), sum(assessed_by_cpg ==
                            0L)), stringsAsFactors = FALSE
                    )
                    assessed_by_sample <- colSums(assessed)
                    detected_by_sample <- colSums(detected)
                    percent_detected <- rep(NA_real_,
                        length(assessed_by_sample))
                    has_assessed_values <- assessed_by_sample > 0L
                    percent_detected[has_assessed_values] <- 100 *
                        detected_by_sample[has_assessed_values] / assessed_by_sample[has_assessed_values]
                    sample <- data.frame(
                        UID = colnames(detp_matrix),
                        nAssessed = assessed_by_sample,
                            nDetected = detected_by_sample,
                        pDetected = round(percent_detected, 2),
                            stringsAsFactors = FALSE
                    )
                    list(cpg = cpg, sample = sample)
                },
                error = function(e) e
            )

            if (inherits(loaded, "error")) {
                result$error <- conditionMessage(loaded)
            } else {
                result$exists <- TRUE
                result$cpg <- loaded$cpg
                result$sample <- loaded$sample
                return(result)
            }
        }

        cpg <- safe_read_table_file(cpg_path)
        sample <- safe_read_table_file(sample_path)
        if (!is.null(cpg) || !is.null(sample)) {
            result$source <- "csv"
            result$exists <- TRUE
            result$cpg <- cpg
            result$sample <- sample
        }

        result
    }

    detection_table_warning <- function(detection_tables) {
        if (!is.null(detection_tables$error)) {
            return(paste0(
                "Unable to build detection tables from `",
                detection_tables$path, "`: ", detection_tables$error
            ))
        }
        paste0(
            "Detection P-value file not found at `", detection_tables$path,
            "`."
        )
    }

    pick_column <- function(data, candidates) {
        if (is.null(data) || !ncol(data)) {
            return(NULL)
        }
        if (any(candidates %in% names(data))) {
            return(candidates[candidates %in% names(data)][[1]])
        }
        lower_names <- tolower(names(data))
        for (candidate in candidates) {
            idx <- match(tolower(candidate), lower_names)
            if (!is.na(idx)) {
                return(names(data)[[idx]])
            }
        }
        NULL
    }

    pick_participant_column <- function(data) {
        exact_col <- pick_column(data, c(
            "Person_ID", "PersonID",
            "person_id", "Person", "person", "Participant_ID",
            "ParticipantID", "participant_id", "Participant",
            "Subject_ID", "SubjectID", "Individual_ID", "IndividualID",
            "UID", "ID"
        ))
        if (!is.null(exact_col)) {
            return(exact_col)
        }

        id_cols <- grep("id", names(data),
            ignore.case = TRUE,
            value = TRUE
        )
        technical_id <-
            grepl("sentrix|sample|array|slide|position|time|visit|plate|well|pool|probe",
            id_cols,
            ignore.case = TRUE
        )
        id_cols <- id_cols[!technical_id]
        if (length(id_cols)) {
            return(id_cols[[1]])
        }

        NULL
    }

    pick_timepoint_column <- function(data) {
        pick_column(data, c(
            "Timepoint", "Tiempoint", "Time_Point",
            "Timepoint_ID", "TimepointID", "Visit", "VisitID",
            "Visit_ID"
        ))
    }

    summarize_dataset <- function(data_path) {
        data <- safe_read_table_file(data_path)
        summary <- list(
            path = slash(data_path), exists = !is.null(data),
            n_rows = NA_integer_, n_cols = NA_integer_, participant_col = NULL,
            n_participants = NA_integer_, timepoint_col = NULL,
            timepoints = character(), n_timepoints = NA_integer_,
            complete_timepoint_participants = NA_integer_,
                repeated_participant_timepoint_rows = NA_integer_,
            survey_columns = character(),
                participants_missing_survey = NA_integer_
        )

        if (is.null(data)) {
            return(summary)
        }

        summary$n_rows <- nrow(data)
        summary$n_cols <- ncol(data)
        participant_col <- pick_participant_column(data)
        timepoint_col <- pick_timepoint_column(data)
        summary$participant_col <- participant_col
        summary$timepoint_col <- timepoint_col

        if (!is.null(participant_col)) {
            participants <- as.character(data[[participant_col]])
            participants <- participants[!is_blank(participants)]
            summary$n_participants <- length(unique(participants))
        }

        if (!is.null(timepoint_col)) {
            timepoints <- as.character(data[[timepoint_col]])
            summary$timepoints <- sort_values(timepoints)
            summary$n_timepoints <- length(summary$timepoints)
        }

        if (!is.null(participant_col) && !is.null(timepoint_col)) {
            participants <- as.character(data[[participant_col]])
            timepoints <- as.character(data[[timepoint_col]])
            complete_rows <- !is_blank(participants) & !is_blank(timepoints)
            keys <- paste(participants[complete_rows],
                timepoints[complete_rows],
                sep = "\r"
            )
            summary$repeated_participant_timepoint_rows <- length(keys) -
                length(unique(keys))

            split_timepoints <- split(
                timepoints[complete_rows],
                participants[complete_rows]
            )
            summary$complete_timepoint_participants <- sum(vapply(
                split_timepoints,
                function(values) {
                    all(summary$timepoints %in%
                        unique(as.character(values)))
                }, logical(1)
            ))
        }

        survey_columns <- grep("^(MHC_|BDSST|BRS_|SS_|WHO_)",
            names(data),
            value = TRUE
        )
        summary$survey_columns <- survey_columns
        if (!is.null(participant_col) && length(survey_columns)) {
            participants <- as.character(data[[participant_col]])
            unique_participants <- sort_values(participants)
            missing_survey <- vapply(unique_participants,
                function(participant) {
                rows <- !is_blank(participants) & participants ==
                    participant
                values <- unlist(data[rows, survey_columns, drop = FALSE],
                    use.names = FALSE
                )
                all(is_blank(values))
            }, logical(1))
            summary$participants_missing_survey <- sum(missing_survey)
        }

        summary
    }

    summarize_cpg_detection <- function(data) {
        summary <- list(exists = is.data.frame(data))
        if (is.null(data) || !all(c("metric", "nCpGs") %in% names(data))) {
            summary$total_assessed <- NA_integer_
            summary$detected_any <- NA_integer_
            summary$detected_all <- NA_integer_
            summary$never_detected <- NA_integer_
            return(summary)
        }

        get_metric <- function(pattern) {
            idx <- grep(pattern, data$metric, ignore.case = TRUE)
            if (length(idx)) {
                as.numeric(data$nCpGs[[idx[[1]]]])
            } else {
                NA_real_
            }
        }

        summary$total_assessed <- get_metric("^Total CpGs assessed$")
        summary$detected_any <- get_metric("at least one sample")
        summary$detected_all <- get_metric("all samples")
        summary$never_detected <- get_metric("never detected")
        summary
    }

    summarize_sample_detection <- function(data) {
        summary <- list(exists = is.data.frame(data))
        if (is.null(data) || !"pDetected" %in% names(data)) {
            summary$n_samples <- NA_integer_
            summary$min_p_detected <- NA_real_
            summary$median_p_detected <- NA_real_
            summary$max_p_detected <- NA_real_
            return(summary)
        }

        p_values <- as.character(data$pDetected)
        numeric_idx <- is_numeric_like(p_values)
        p_detected <- rep(NA_real_, length(p_values))
        p_detected[numeric_idx] <- as.numeric(p_values[numeric_idx])
        summary$n_samples <- nrow(data)
        if (any(!is.na(p_detected))) {
            summary$min_p_detected <- min(p_detected, na.rm = TRUE)
            summary$median_p_detected <- stats::median(p_detected,
                na.rm = TRUE
            )
            summary$max_p_detected <- max(p_detected, na.rm = TRUE)
        } else {
            summary$min_p_detected <- NA_real_
            summary$median_p_detected <- NA_real_
            summary$max_p_detected <- NA_real_
        }
        summary
    }

    extract_sva_log_summary <- function(log_path) {
        summary <- list(k = NA_integer_, percent_variation = NA_real_)
        if (!file.exists(log_path)) {
            return(summary)
        }

        text <- paste(readLines(log_path, warn = FALSE, encoding = "UTF-8"),
            collapse = " "
        )
        match <- regexec("([0-9]+)\\s+surrogate variables explain\\s+([0-9.]+)\\s*%",
            text,
            perl = TRUE
        )
        values <- regmatches(text, match)[[1]]
        if (length(values) == 3L) {
            summary$k <- as.integer(values[[2]])
            summary$percent_variation <- as.numeric(values[[3]])
        }
        summary
    }

    summarize_sva <- function(pheno_path, log_path) {
        pheno_data <- safe_read_table_file(pheno_path)
        log_summary <- extract_sva_log_summary(log_path)
        summary <- list(
            path = slash(pheno_path), exists = !is.null(pheno_data),
            n_samples = NA_integer_, n_surrogate_variables = NA_integer_,
            surrogate_variables = character(), log_k = log_summary$k,
            percent_variation = log_summary$percent_variation,
            sentrix_id_levels = NA_integer_,
                sentrix_position_levels = NA_integer_
        )

        if (!is.na(log_summary$k)) {
            summary$n_surrogate_variables <- log_summary$k
            summary$surrogate_variables <- paste0("SV", seq_len(log_summary$k))
        }

        if (!is.null(pheno_data)) {
            summary$n_samples <- nrow(pheno_data)
            sentrix_id_col <- pick_column(pheno_data, c(
                "Sentrix_ID",
                "SentrixID", "Sentrix Id"
            ))
            sentrix_position_col <- pick_column(pheno_data, c(
                "Sentrix_Position",
                "SentrixPosition", "Sentrix Position"
            ))
            if (!is.null(sentrix_id_col)) {
                summary$sentrix_id_levels <-
                    length(sort_values(pheno_data[[sentrix_id_col]]))
            }
            if (!is.null(sentrix_position_col)) {
                summary$sentrix_position_levels <-
                    length(sort_values(pheno_data[[sentrix_position_col]]))
            }
        }

        summary
    }

    summarize_logs <- function(log_assets) {
        labels <- c(
            methylation = "Methylation Analysis", data = "Data Preparation",
            batch = "Batch Effect", glm = "GLM", lme = "LME"
        )
        rows <- lapply(names(log_assets), function(name) {
            asset <- log_assets[[name]]
            line_count <- if (isTRUE(asset$exists)) {
                length(readLines(asset$source_path,
                    warn = FALSE,
                    encoding = "UTF-8"
                ))
            } else {
                NA_integer_
            }
            data.frame(
                key = name, label = labels[[name]],
                    exists = isTRUE(asset$exists),
                source_path = asset$source_path, line_count = line_count,
                stringsAsFactors = FALSE
            )
        })
        rows <- do.call(rbind, rows)
        list(
            rows = rows, total = nrow(rows), found = sum(rows$exists),
            total_lines = sum(rows$line_count, na.rm = TRUE)
        )
    }

    make_data_notes <- function(summary) {
        if (!isTRUE(summary$exists)) {
            return("The source CSV was not found.")
        }

        notes <- c(sprintf(
            "The table has %s rows and %s columns.",
            format_count(summary$n_rows), format_count(summary$n_cols)
        ))

        if (!is.null(summary$participant_col) &&
            !is.null(summary$timepoint_col)) {
            notes <- c(notes, sprintf(
                "Using `%s` as the participant identifier and `%s` as the timepoint column, the file contains %s unique %s across timepoint values %s.",
                summary$participant_col, summary$timepoint_col,
                format_count(summary$n_participants), plural(
                    summary$n_participants,
                    "participant"
                ), collapse_values(summary$timepoints)
            ))
        } else if (!is.null(summary$participant_col)) {
            notes <- c(notes, sprintf(
                "Using `%s` as the participant identifier, the file contains %s unique %s. No Timepoint column was detected.",
                summary$participant_col, format_count(summary$n_participants),
                plural(summary$n_participants, "participant")
            ))
        } else {
            notes <- c(notes,
                "No biological participant identifier column was detected. Technical identifiers such as sample, Sentrix, array, and slide IDs are not treated as participant IDs.")
        }

        notes
    }

    describe_figure_title <- function(title) {
        title <- sub("^Figure\\s+[0-9]+:\\s*", "", title)
        title <- sub("\\.$", "", title)
        first <- substr(title, 1L, 1L)
        rest <- substr(title, 2L, nchar(title))
        paste0(tolower(first), rest)
    }

    highlight_report_label <- function(title) {
        sub("^((?:Figure|Table)\\s+[0-9]+)([:.])", "`\\1`\\2",
            title,
            perl = TRUE
        )
    }

    make_figure_notes <- function(items, figure_titles, section_label,
                                    figure_descriptions = NULL) {
        n_items <- nrow(items)
        if (!n_items) {
            return(sprintf(
                "No supported figures were found for the %s tab.",
                section_label
            ))
        }

        vapply(seq_len(n_items), function(idx) {
            title <- if (length(figure_titles) >= idx &&
                nzchar(figure_titles[[idx]])) {
                figure_titles[[idx]]
            } else if ("title" %in% names(items) &&
                nzchar(items$title[[idx]])) {
                items$title[[idx]]
            } else {
                sprintf("Figure %s", idx)
            }
            highlighted_title <- highlight_report_label(title)
            description <- if (!is.null(figure_descriptions) &&
                length(figure_descriptions) >= idx &&
                    nzchar(figure_descriptions[[idx]])) {
                figure_descriptions[[idx]]
            } else {
                describe_figure_title(title)
            }
            description <- sub("\\.$", "", description)
            if (grepl("^(presents|shows|summarises|compares)\\b",
                description,
                ignore.case = TRUE
            )) {
                sprintf("%s %s.", highlighted_title, description)
            } else {
                sprintf(
                    "%s presents %s.", highlighted_title,
                    description
                )
            }
        }, character(1))
    }

    make_metrics_notes <- function(items, figure_titles, figure_descriptions,
                                    data_summary) {
        make_figure_notes(items, figure_titles, "Metrics", figure_descriptions)
    }

    make_quality_control_notes <- function(items, figure_titles,
                                            figure_descriptions, cpg_summary,
                                                sample_summary) {
        notes <- make_figure_notes(
            items, figure_titles, "Quality Control",
            figure_descriptions
        )

        if (isTRUE(cpg_summary$exists)) {
            notes <- c(notes, sprintf(
                "`Table 1` reports %s assessed CpGs, %s detected in all samples, and %s never detected.",
                format_count(cpg_summary$total_assessed),
                    format_count(cpg_summary$detected_all),
                format_count(cpg_summary$never_detected)
            ))
        }

        if (isTRUE(sample_summary$exists)) {
            notes <- c(notes, sprintf(
                "`Table 2` shows sample detection percentages ranging from %s%% to %s%% across %s %s.",
                format_decimal(sample_summary$min_p_detected),
                format_decimal(sample_summary$max_p_detected),
                format_count(sample_summary$n_samples), plural(
                    sample_summary$n_samples,
                    "sample"
                )
            ))
        }

        notes
    }

    make_batch_effect_notes <- function(items, figure_titles,
                                        figure_descriptions, sva_summary) {
        notes <- make_figure_notes(
            items, figure_titles, "Batch Effect",
            figure_descriptions
        )

        if (!is.na(sva_summary$log_k) &&
            !is.na(sva_summary$percent_variation)) {
            notes <- c(notes, sprintf(
                "The SVA log reports %s surrogate %s explaining %s%% of data variation.",
                format_count(sva_summary$log_k), plural(
                    sva_summary$log_k,
                    "variable"
                ), format_decimal(sva_summary$percent_variation)
            ))
        }

        notes
    }

    make_logs_notes <- function(log_summary, lme_label = "LME Analysis") {
        descriptions <- c(
            methylation = "displays methylation preprocessing, including IDAT loading, normalisation, filtering, and cell composition estimation",
            data = "displays phenotype preparation, timepoint splitting, and methylation matrix export steps",
            batch = "displays the hidden-effect and surrogate-variable analysis workflow",
            glm = "displays the generalised linear model workflow and CpG annotation steps",
            lme = "displays the linear mixed-effects model workflow and CpG annotation steps"
        )
        labels <- c(
            methylation = "Methylation Analysis", data = "Data Preparation",
            batch = "Batch Effect", glm = "GLM Analysis", lme = lme_label
        )

        notes <- c(
            "The Logs tab displays workflow log files generated by the pipeline, with one section per analysis stage.",
            vapply(log_summary$rows$key, function(key) {
                status <-
                    if (isTRUE(log_summary$rows$exists[log_summary$rows$key ==
                    key])) {
                    "is available"
                } else {
                    "was not found"
                }
                sprintf(
                    "`%s` %s and %s.", labels[[key]], status,
                    descriptions[[key]]
                )
            }, character(1))
        )

        notes
    }

    html_paragraph <- function(text) {
        sprintf("<p>%s</p>", html_escape(text))
    }

    html_section <- function(title, paragraphs) {
        c(sprintf("<h3>%s</h3>", html_escape(title)),
            html_paragraph(paragraphs))
    }

    strip_inline_markdown <- function(text) {
        text <- gsub("`([^`]+)`", "\\1", text, perl = TRUE)
        text <- gsub("<[^>]+>", "", text, perl = TRUE)
        text <- gsub("\\s+", " ", text)
        trimws(text)
    }

    sentence_case <- function(text) {
        if (!nzchar(text)) {
            return(text)
        }
        paste0(toupper(substr(text, 1L, 1L)), substr(
            text, 2L,
            nchar(text)
        ))
    }

    tab_report_paragraph <- function(notes, fallback) {
        notes <- notes[!grepl("dnaepico-model-formulas", notes,
            fixed = TRUE
        )]
        notes <- strip_inline_markdown(notes)
        notes <- notes[nzchar(notes)]
        if (!length(notes)) {
            return(sprintf("%s.", sentence_case(fallback)))
        }

        paste(notes, collapse = " ")
    }

    build_report_page <- function(project_name, project_dir,
                                    data_notes, enmix_notes,
                                        quality_control_notes,
                                        batch_effect_notes,
                                    metrics_notes, glm_notes, lme_notes,
                                        logs_notes,
                                        lme_label = "LME Analysis") {
        overview <-
            sprintf("This report is generated automatically from the available datasets, figures, tables, and workflow logs for the project.")

        data_paragraph <- tab_report_paragraph(data_notes,
            "displays the phenotype data preview and detected participant/timepoint fields")
        enmix_paragraph <- tab_report_paragraph(
            enmix_notes,
            "displays ENmix control plots and interpretation notes"
        )
        quality_control_paragraph <- tab_report_paragraph(
            quality_control_notes,
            "displays quality-control figures and detection summary tables"
        )
        batch_effect_paragraph <- tab_report_paragraph(
            batch_effect_notes,
            "displays SVA and batch-effect figures"
        )
        metrics_paragraph <- tab_report_paragraph(
            metrics_notes,
            "displays post-filtering methylation metric figures"
        )
        glm_paragraph <- tab_report_paragraph(glm_notes,
            "displays the annotated generalised linear model results table")
        lme_paragraph <- tab_report_paragraph(lme_notes,
            "displays the annotated linear mixed-effects model results table")
        logs_paragraph <- tab_report_paragraph(logs_notes,
            "displays workflow log files for each analysis stage")

        c(
            "---", "title: \"Report\"", "format:", "  html:", "    css:",
            "      - assets/qpasst.css", "page-layout: full",
            "body-classes: qpasst-report-page", "---", "",
                "<div class=\"qpasst-report-viewport\">",
            "<div class=\"qpasst-report-card\">",
                "<div class=\"qpasst-report-sheet\">",
            sprintf("<h2>%s Report</h2>", html_escape(project_name)),
            html_paragraph(overview), html_section("Data", data_paragraph),
            html_section("ENmix QC", enmix_paragraph), html_section(
                "Quality Control",
                quality_control_paragraph
            ), html_section(
                "Batch Effect",
                batch_effect_paragraph
            ), html_section(
                "Metrics",
                metrics_paragraph
            ), html_section("GLM", glm_paragraph),
            html_section(sub(" Analysis$", "", lme_label), lme_paragraph),
            html_section("Logs", logs_paragraph), "</div>", "</div>",
            "</div>"
        )
    }

    sidebar_lines <- function(notes) {
        bullet_lines <- if (length(notes)) {
            paste0("- ", notes)
        } else {
            "- Add notes here."
        }

        c(
            "## {.sidebar width=\"320px\"}", "### Notes", "", bullet_lines,
            ""
        )
    }

    compose_page <- function(title, notes, body_lines, body_classes = NULL) {
        front_matter <- c(
            "---", sprintf("title: \"%s\"", title),
            "format: dashboard"
        )

        if (!is.null(body_classes) && nzchar(body_classes)) {
            front_matter <- c(front_matter, sprintf(
                "body-classes: %s",
                body_classes
            ))
        }

        c(
            front_matter, "---", "", sidebar_lines(notes), "## Column",
            "", body_lines
        )
    }

    prepared_report <- prepareDnamReportInputs(
        outputDir = project_dir,
        qcDir = qc_dir, preprocessingDir = preprocessing_dir,
        postprocessingDir = postprocessing_dir, svaDir = sva_dir,
        glmDir = glm_dir, lmeDir = lme_dir, figDir = fig_dir,
        verbose = FALSE, logs = FALSE, logDir = logs_dir
    )

    enmix_items <- copy_figure_assets(qc_dir, asset_subdir = "enmix-qc")
    metrics_items <- copy_figure_assets(postprocessing_dir,
        asset_subdir = "metrics")
    qc_items <- copy_figure_assets(preprocessing_dir,
        asset_subdir = "quality-control")
    batch_items <- copy_figure_assets(sva_dir, asset_subdir = "batch-effects")

    log_assets <- list(
        methylation = copy_log_asset(file.path(
            logs_dir,
            "log_preprocessingMinfiEwasWater.txt"
        ), "preprocessingMinfiEwasWater.txt"),
        data = copy_log_asset(
            file.path(logs_dir, "log_preprocessingPheno.txt"),
            "preprocessingPheno.txt"
        ), batch = copy_log_asset(file.path(
            logs_dir,
            "log_svaEnmix.txt"
        ), "svaEnmix.txt"), glm = copy_first_existing_log_asset(file.path(
            logs_dir,
            "log_methylationGLM.txt"
        ), "methylationGLM.txt"),
        lme = copy_first_existing_log_asset(file.path(
            logs_dir,
            "log_methylationLME.txt"
        ), "methylationLME.txt")
    )

    data_summary <- summarize_dataset(pheno_file)
    detection_tables <- read_detection_tables(
        detp_path = detp_path,
        threshold = detPThreshold, cpg_path = cpg_detection_path,
        sample_path = sample_detection_path
    )
    cpg_detection_summary <- summarize_cpg_detection(detection_tables$cpg)
    sample_detection_summary <-
        summarize_sample_detection(detection_tables$sample)
    sva_summary <- summarize_sva(pheno_path = pheno_file, log_path = file.path(
        logs_dir,
        "log_svaEnmix.txt"
    ))
    log_summary <- summarize_logs(log_assets)

    enmix_figure_titles <- enmix_items$title
    enmix_notes <- c(paste(
        "`ENmix` produces control plots similar to those generated by Illumina's GenomeStudio software.",
        "Infinium controls are not evaluated using absolute intensity values.",
        "Instead, we interpret them based on their expected signal range across the dataset.",
        "This approach is intended to provide greater flexibility and reliability when accounting for natural biological variation."
    ))
    metrics_figure_titles <- c(
        "Figure 1: The Density Distributions of Beta and M-values Across Samples Post-Filtering Methylation Plot by Timepoint",
        "Figure 2: Principal Component Analysis Post-Filtering Methylation Plot by Timepoint"
    )
    metrics_figure_descriptions <- c(
        "beta-value and M-value density distributions across samples after filtering, grouped by timepoint",
        "a principal component analysis plot after filtering, grouped by timepoint"
    )
    quality_control_metadata <- data.frame(
        pattern = c(
            "densityBeta",
            "detection_pvalues", "quality_control", "sexClinical",
            "sexComparison_RawNorm", "sexPrediction"
        ), title = c(
            "Figure 1: Beta Value Density Distribution (MSET)",
            "Figure 2: Detection P-values (RGSET)",
                "Figure 3: Quality Control Plot (MSET)",
            "Figure 4: Clinical Sex Distribution (GSET)",
                "Figure 5: Sex Comparison Before and After Normalisation (MSETF)",
            "Figure 6: Sex Prediction Plot (GSET)"
        ), description = c(
            "presents the distribution of methylation beta values across samples, allowing visual assessment of overall methylation patterns and potential sample outliers",
            "shows the distribution of detection p-values across samples, providing a quality check for probe signal reliability and identifying samples with poor detection performance",
            "presents sample-level quality metrics based on methylated and unmethylated signal intensities, helping to identify low-quality samples or technical outliers",
            "summarises the sex information recorded in the clinical or phenotype data, supporting checks of sample annotation consistency",
            "compares sex-related methylation patterns in raw and normalised data, helping to confirm whether normalisation preserves expected biological differences",
            "presents methylation-based sex prediction across samples, allowing comparison with recorded clinical sex and detection of possible sample swaps or annotation errors"
        ),
        stringsAsFactors = FALSE
    )
    quality_control_figure_titles <- qc_items$title
    quality_control_figure_descriptions <- rep("", nrow(qc_items))
    for (metadata_idx in seq_len(nrow(quality_control_metadata))) {
        item_idx <- grep(quality_control_metadata$pattern[[metadata_idx]],
            qc_items$original_name,
            ignore.case = TRUE
        )
        if (length(item_idx)) {
            quality_control_figure_titles[item_idx] <-
                quality_control_metadata$title[[metadata_idx]]
            quality_control_figure_descriptions[item_idx] <-
                quality_control_metadata$description[[metadata_idx]]
        }
    }
    batch_effect_figure_titles <- c(
        "Figure 1: Distribution of Samples Across the First Two Surrogate Variables by SentrixID",
        "Figure 2: Pairwise Distribution of the First Three Surrogate Variables by SentrixID and SentrixPosition",
        "Figure 3: Distribution of Samples Across the First Two Surrogate Variables by SentrixPosition"
    )
    batch_effect_figure_descriptions <- c(
        "sample distribution across the first two surrogate variables, coloured by Sentrix ID",
        "pairwise distributions of the first three surrogate variables by Sentrix ID and Sentrix position",
        "sample distribution across the first two surrogate variables by Sentrix position"
    )
    data_table_assets <- prepare_csv_table_assets(
        data_path = pheno_file,
        var_prefix = "phenotype_data", front_columns = c(
            if (is.null(data_summary$participant_col)) "" else data_summary$participant_col,
            if (is.null(data_summary$timepoint_col)) "" else data_summary$timepoint_col
        )
    )
    glm_table_assets <- prepare_xlsx_table_assets(
        data_path = glm_table_path,
        sheet = "annotatedGLM", var_prefix = "glm_results", analysis = "glm",
        model_log_path = file.path(logs_dir, "log_methylationGLM.txt")
    )
    lme_table_assets <- prepare_xlsx_table_assets(
        data_path = lme_table_path,
        sheet = "annotatedLME", var_prefix = "lme_results", analysis = "lme",
        model_log_path = file.path(logs_dir, "log_methylationLME.txt")
    )
    emitLogMinfiEwasWater(
        c(
            paste(
                "GLM report table source:",
                if (is.null(glm_table_assets$source_mode)) "unavailable" else glm_table_assets$source_mode
            ),
            paste("LME report table source:",
                if (is.null(lme_table_assets$source_mode)) "unavailable" else lme_table_assets$source_mode)
        ),
        verbose = FALSE, log_path = log_path
    )

    lme_backend <- lme_table_assets$metadata$backend
    if (is.null(lme_backend) || !nzchar(lme_backend)) {
        lme_backend <- "lme4"
    }
    lme_backend <- tolower(lme_backend)
    lme_analysis_label <- if (identical(lme_backend, "nlme")) {
        "nlme Analysis"
    } else {
        "LME Analysis"
    }
    lme_interaction_term <- lme_table_assets$metadata$interaction_term
    has_lme_interaction <- !is.null(lme_interaction_term) &&
        nzchar(lme_interaction_term) && !tolower(lme_interaction_term) %in%
        c("none", "null", "na")

    glm_table_title <-
        "Table 1. Generalised Linear Model Results and Genomic Annotation of CpG Sites by Phenotype(s)"
    lme_table_title <- if (has_lme_interaction) {
        sprintf(
            "Table 1. Linear Mixed-Effects Model Results and Genomic Annotation of CpG Sites by Phenotype(s) and %s",
            lme_interaction_term
        )
    } else {
        "Table 1. Linear Mixed-Effects Model Results and Genomic Annotation of CpG Sites by Phenotype(s)"
    }

    describe_available_annotations <- function(columns) {
        notes <- character()
        if (all(c("IlmnID", "Name") %in% columns)) {
            notes <- c(notes,
                "The `IlmnID` and `Name` columns identify each Illumina CpG probe.")
        } else if ("IlmnID" %in% columns) {
            notes <- c(notes,
                "The `IlmnID` column identifies each Illumina CpG probe.")
        }
        if (all(c("chr", "pos") %in% columns)) {
            notes <- c(notes,
                "The `chr` and `pos` columns give the probe's genomic position.")
        }
        if (any(c("UCSC_RefGene_Group", "UCSC_RefGene_Name") %in%
            columns)) {
            notes <- c(notes,
                "Available `UCSC_RefGene` columns provide gene-related annotation.")
        }
        if ("Relation_to_Island" %in% columns) {
            notes <- c(notes,
                "`Relation_to_Island` reports the probe's CpG-island context.")
        }
        if ("GencodeV41_Group" %in% columns) {
            notes <- c(notes,
                "`GencodeV41_Group` provides GENCODE version 41 gene-region annotation.")
        }
        notes
    }

    glm_table_description <- if (isTRUE(glm_table_assets$ok)) {
        glm_p_columns <- grep("P\\.Value$|P\\.value$", glm_table_assets$columns,
            value = TRUE
        )
        c(
            "CpGs were analysed using a generalised linear model fitted with `glm2`.",
            build_model_formula_notes(
                glm_table_assets$metadata$formula_records,
                "GLM"
            ), if (length(glm_p_columns)) {
                paste0(
                    "The result p-value column(s) are `",
                    paste(glm_p_columns, collapse = "`, `"), "`."
                )
            }, describe_available_annotations(glm_table_assets$columns)
        )
    } else {
        "The annotated GLM result workbook was not available when this report was generated."
    }

    lme_table_description <- if (isTRUE(lme_table_assets$ok)) {
        lme_p_columns <- grep("P\\.Value$|P\\.value$", lme_table_assets$columns,
            value = TRUE
        )
        fitting_description <- if (identical(lme_backend, "nlme")) {
            "CpGs were analysed using `nlme::lme()`."
        } else {
            "CpGs were analysed using `lmerTest::lmer()` with the `lme4` mixed-effects framework."
        }
        correlation_structure <- lme_table_assets$metadata$correlation_structure
        correlation_variable <- lme_table_assets$metadata$correlation_variable
        correlation_description <- if (identical(
            lme_backend,
            "nlme"
        ) && !is.null(correlation_structure) && nzchar(correlation_structure) &&
            !tolower(correlation_structure) %in% c(
                "none", "null",
                "na"
            )) {
            sprintf(
                "The `%s` residual correlation structure orders repeated observations using `%s`.",
                correlation_structure, if (is.null(correlation_variable) ||
                    !nzchar(correlation_variable)) {
                    "the configured correlation variable"
                } else {
                    correlation_variable
                }
            )
        } else {
            character()
        }
        interaction_description <- if (has_lme_interaction) {
            sprintf(
                "The reported phenotype terms test their interaction with `%s`.",
                lme_interaction_term
            )
        } else {
            "No phenotype-by-time interaction was fitted; the reported p-value column(s) correspond to phenotype coefficient(s)."
        }
        formula_description <- build_model_formula_notes(
            lme_table_assets$metadata$formula_records,
            if (identical(lme_backend, "nlme")) {
                "nlme"
            } else {
                "LME"
            }
        )
        c(
            fitting_description, correlation_description,
                interaction_description,
            formula_description, if (length(lme_p_columns)) {
                paste0(
                    "The result p-value column(s) are `",
                    paste(lme_p_columns, collapse = "`, `"), "`."
                )
            }, describe_available_annotations(lme_table_assets$columns)
        )
    } else {
        "The annotated LME result workbook was not available when this report was generated."
    }

    data_notes <- make_data_notes(data_summary)
    metrics_notes <- make_metrics_notes(
        metrics_items, metrics_figure_titles,
        metrics_figure_descriptions, data_summary
    )
    quality_control_notes <- make_quality_control_notes(
        qc_items,
        quality_control_figure_titles, quality_control_figure_descriptions,
        cpg_detection_summary, sample_detection_summary
    )
    batch_effect_notes <- make_batch_effect_notes(
        batch_items,
        batch_effect_figure_titles, batch_effect_figure_descriptions,
        sva_summary
    )
    glm_notes <- glm_table_description
    lme_notes <- lme_table_description
    logs_notes <- make_logs_notes(log_summary, lme_label = lme_analysis_label)

    if (file.exists(logo_source_path)) {
        dir.create(assets_dir, recursive = TRUE, showWarnings = FALSE)
        ok <- file.copy(logo_source_path, file.path(
            assets_dir,
            "dnaEPICO.svg"
        ), overwrite = TRUE)
        if (!ok) {
            warning("Failed to copy navbar logo: ", logo_source_path)
        }
    }
    if (file.exists(viewer_js_source_path)) {
        dir.create(assets_dir, recursive = TRUE, showWarnings = FALSE)
        ok <- file.copy(viewer_js_source_path, file.path(
            assets_dir,
            "cpg-viewer.js"
        ), overwrite = TRUE)
        if (!ok) {
            warning(
                "Failed to copy CpG viewer JavaScript: ",
                viewer_js_source_path
            )
        }
    } else {
        warning("CpG viewer JavaScript was not found in the installed package.")
    }

    quarto_yml <- c(
        "project:", "  type: website", "  output-dir: docs",
        "  resources:", "    - assets/", "", "website:", sprintf(
            "  title: \"%s\"",
            project_name
        ), "  navbar:", "    logo: assets/dnaEPICO.svg",
        "    left:", "      - href: index.qmd", "        text: \"Data\"",
        "      - href: enmix-qc.qmd", "        text: \"ENmix QC\"",
        "      - href: quality-control.qmd",
            "        text: \"Quality Control\"",
        "      - href: batch-effect.qmd", "        text: \"Batch Effect\"",
        "      - href: metrics.qmd", "        text: \"Metrics\"",
        "      - href: glm.qmd", "        text: \"GLM Analysis\"",
        "      - href: lme.qmd", sprintf(
            "        text: \"%s\"",
            lme_analysis_label
        ), "      - href: report.qmd",
        "        text: \"Report\"", "      - href: logs.qmd",
        "        text: \"Logs\"", "    search: true",
            "  page-navigation: false",
        "  bread-crumbs: false", "  reader-mode: false", "",
        "format:", "  dashboard:", "    theme: cosmo", "    css:",
        "      - assets/qpasst.css", "    toc: false", "", "execute:",
        "  warning: false", "  message: false"
    )

    site_css <- c(
        ":root {", "  --qpasst-border: #e5e7eb;", "  --qpasst-bg: #d5d9df;",
        "  --qpasst-navbar-bg: #f4f6f8;", "  --qpasst-card-bg: #ffffff;",
        "  --qpasst-accent: #0d6efd;", "}", "", "body {",
            "  background: #ffffff;",
        "}", "", ".navbar,", ".quarto-navbar,", "#quarto-header .navbar {",
        "  background: var(--qpasst-navbar-bg) !important;",
        "  border-bottom: 1px solid var(--qpasst-border) !important;",
        "  box-shadow: none !important;", "}", "", ".navbar-brand,",
        ".navbar-title {", "  color: #3f444a !important;",
            "  font-weight: 600;",
        "}", "", "#quarto-header .navbar-brand-container {",
        "  display: flex;", "  align-items: center;", "}", "",
        "#quarto-header .navbar-brand.navbar-brand-logo {",
            "  display: flex !important;",
        "  align-items: center;", "  padding-top: 0;", "  padding-bottom: 0;",
        "  margin-right: 0.35rem;", "}", "",
            "#quarto-header .navbar-brand.navbar-brand-logo img,",
        "#quarto-header .navbar-brand.navbar-brand-logo .navbar-logo,",
        "#quarto-header .navbar-brand.navbar-brand-logo .navbar-logo-image {",
        "  display: block;", "  height: 48px !important;", "  width: auto;",
        "  max-height: 48px !important;", "  max-width: 48px !important;",
        "  object-fit: contain;", "}", "",
            "#quarto-header .navbar-brand:not(.navbar-brand-logo) {",
        "  display: none !important;", "}", "", ".navbar-nav .nav-link {",
        "  color: #495057 !important;", "  font-weight: 700;",
        "}", "", ".navbar-nav .nav-link.active,",
            ".navbar-nav .show > .nav-link,",
        ".navbar-nav .nav-link:hover {",
            "  color: var(--qpasst-accent) !important;",
        "}", "", "#quarto-dashboard-header {", "  display: none !important;",
        "}", "", ".quarto-dashboard .bslib-sidebar-layout > .main,",
        ".quarto-dashboard .bslib-sidebar-layout .sidebar-content {",
        "  overflow-y: auto !important;", "}", "",
            ".quarto-dashboard .sidebar-content > .bslib-grid,",
        ".quarto-dashboard .sidebar-content > .bslib-grid > .bslib-grid,",
        ".quarto-dashboard .sidebar-content > .bslib-grid > .bslib-grid > .bslib-grid {",
        "  grid-auto-rows: max-content !important;", "}", "",
        ".quarto-dashboard .bslib-card {", "  min-height: 0 !important;",
        "  margin-bottom: 1.5rem;", "}", "",
            ".quarto-dashboard .bslib-card .card-body {",
        "  height: auto !important;", "  overflow: visible !important;",
        "  padding: 1.5rem !important;", "}", "",
            "body.qpasst-logs-page .sidebar-content > .bslib-grid > .bslib-grid {",
        "  grid-template-rows: minmax(3em, max-content) minmax(3em, max-content) minmax(3em, max-content) !important;",
        "}", "", "body.qpasst-logs-page .panel-tabset.bslib-grid {",
        "  grid-template-columns: minmax(3em, 1fr) !important;",
        "  grid-auto-rows: max-content !important;", "}", "",
        "body.qpasst-logs-page .bslib-card {",
            "  min-height: calc(100vh - 12rem) !important;",
        "}", "", "body.qpasst-logs-page .bslib-card .card-body {",
        "  overflow: auto !important;", "}", "",
            "body.qpasst-logs-page pre.text {",
        "  margin: 0;", "  max-height: calc(100vh - 16rem);",
        "  overflow: auto;", "  white-space: pre;", "}", "",
        "body.qpasst-report-page #title-block-header {",
            "  display: none !important;",
        "}", "", "body.qpasst-report-page #quarto-content,",
        "body.qpasst-report-page main.content,",
            "body.qpasst-report-page #quarto-document-content,",
        "body.qpasst-report-page .page-columns,",
            "body.qpasst-report-page .page-full {",
        "  width: 100% !important;", "  max-width: none !important;",
        "}", "", "body.qpasst-report-page main.content {",
            "  margin: 0 !important;",
        "  padding: 0.75rem !important;", "}", "",
            "body.qpasst-report-page .qpasst-report-viewport {",
        "  width: 100%;", "  min-height: calc(100vh - 7rem);",
        "}", "", "body.qpasst-report-page .qpasst-report-card {",
        "  width: 100%;", "  min-height: calc(100vh - 7rem);",
        "  border: 1px solid var(--qpasst-border);", "  border-radius: 12px;",
        "  background: var(--qpasst-card-bg);",
            "  box-shadow: 0 4px 14px rgba(15, 23, 42, 0.05);",
        "  overflow-y: auto;", "  padding: 1.25rem 1.5rem;",
        "}", "", "body.qpasst-report-page .qpasst-report-sheet {",
        "  width: 100%;", "  min-height: 297mm;", "  margin: 0;",
        "  padding: 0.5rem 0 2rem;", "}", "",
            "body.qpasst-report-page .qpasst-report-sheet h2,",
        "body.qpasst-report-page .qpasst-report-sheet h3 {",
        "  color: #2f3540;", "  margin-top: 0;", "}", "",
            "body.qpasst-report-page .qpasst-report-sheet p {",
        "  font-size: 1.02rem;", "  line-height: 1.75;",
            "  margin-bottom: 1rem;",
        "}", "", ".dashboard .sidebar,", ".quarto-dashboard .sidebar,",
        ".sidebar {", "  background: var(--qpasst-bg) !important;",
        "  border-right: 1px solid var(--qpasst-border) !important;",
        "  overflow-x: hidden !important;", "}", "", ".sidebar h3,",
        ".sidebar .h3 {", "  color: #2f3540 !important;", "}",
        "", ".sidebar p,", ".sidebar li,", ".sidebar code {",
        "  font-size: 1.06rem;", "  line-height: 1.7;",
            "  overflow-wrap: anywhere;",
        "  word-break: break-word;", "}", "",
            "body.qpasst-data-page .sidebar li {",
        "  font-size: 1.08rem;", "  line-height: 1.65;", "}",
        "", ".sidebar li::marker {", "  color: var(--qpasst-accent);",
        "}", "", ".qpasst-gallery {", "  display: grid;",
            "  grid-template-columns: 1fr;",
        "  gap: 1.25rem;", "  margin-top: 0.5rem;", "}", "",
        ".qpasst-gallery-item {", "  margin: 0;",
            "  border: 1px solid var(--qpasst-border);",
        "  border-radius: 12px;", "  background: var(--qpasst-card-bg);",
        "  padding: 1rem;", "  box-shadow: 0 4px 14px rgba(15, 23, 42, 0.05);",
        "}", "", ".qpasst-gallery-item figcaption {",
            "  margin-bottom: 0.75rem;",
        "  line-height: 1.5;", "}", "", ".qpasst-gallery-item img {",
        "  display: block;", "  width: 50%;", "  height: auto;",
        "  border-radius: 8px;", "  background: #fff;", "}",
        "", ".qpasst-gallery-item code {", "  white-space: normal;",
        "  word-break: break-word;", "}", "", ".qpasst-figure-card {",
        "  margin: 0;", "}", "", ".qpasst-figure-card img {",
        "  display: block;", "  width: 100%;", "  height: auto;",
        "  border-radius: 8px;", "  background: #fff;", "}",
        "", ".qpasst-figure-card figcaption {", "  margin-top: 0.75rem;",
        "}", "", ".qpasst-download {", "  margin: 0;",
            "  padding: 0.75rem 0.9rem;",
        "  border-radius: 8px;", "  background: #f8fafc;",
            "  border: 1px dashed var(--qpasst-border);",
        "}", "", ".table {", "  font-size: 0.92rem;", "}", "",
        ".dnaepico-cpg-viewer {", "  display: flex !important;",
        "  flex-direction: column !important;", "  gap: 0.45rem;",
        "  width: 100%;", "  min-height: 0 !important;", "}",
        "", ".dnaepico-viewer-toolbar,", ".dnaepico-viewer-filter,",
        ".dnaepico-viewer-pagination,", ".dnaepico-viewer-downloads {",
        "  display: flex !important;", "  flex: 0 0 auto !important;",
        "  flex-direction: row !important;",
            "  align-items: center !important;",
        "  flex-wrap: wrap !important;", "  gap: 0.45rem;",
            "  min-height: 0 !important;",
        "}", "", ".dnaepico-viewer-toolbar label,",
            ".dnaepico-viewer-filter label,",
        ".dnaepico-viewer-pagination label {",
            "  display: inline-flex !important;",
        "  flex-direction: row !important;",
            "  align-items: center !important;",
        "  gap: 0.4rem;", "  margin: 0;", "}", "", ".dnaepico-viewer-filter {",
        "  padding: 0.4rem 0.5rem;", "  border: 1px solid #e2e8f0;",
        "  border-radius: 0.4rem;", "  background: #f8fafc;",
        "}", "", ".dnaepico-viewer-toolbar input {", "  min-width: 14rem;",
        "}", "", ".dnaepico-viewer-filter input {", "  width: 8rem;",
        "}", "", ".dnaepico-viewer-pagination input {", "  width: 4.5rem;",
        "}", "", ".dnaepico-viewer-toolbar input,",
            ".dnaepico-viewer-toolbar select,",
        ".dnaepico-viewer-filter input,", ".dnaepico-viewer-filter select,",
        ".dnaepico-viewer-pagination input {", "  border: 1px solid #ced4da;",
        "  border-radius: 0.375rem;", "  padding: 0.25rem 0.45rem;",
        "  background: #fff;", "}", "", ".dnaepico-viewer-toolbar button,",
        ".dnaepico-viewer-filter button,",
            ".dnaepico-viewer-pagination button {",
        "  border: 1px solid #adb5bd;", "  border-radius: 0.375rem;",
        "  padding: 0.25rem 0.55rem;", "  background: #fff;",
        "  color: #212529;", "}", "", ".dnaepico-viewer-toolbar button:hover,",
        ".dnaepico-viewer-filter button:hover:not(:disabled),",
        ".dnaepico-viewer-pagination button:hover:not(:disabled) {",
        "  background: #eef3f8;", "}", "",
            ".dnaepico-viewer-filter button:disabled,",
        ".dnaepico-viewer-pagination button:disabled {", "  opacity: 0.5;",
        "}", "", ".dnaepico-viewer-pagination {",
            "  justify-content: center !important;",
        "  font-size: 0.88rem;", "}", "",
            ".dnaepico-viewer-pagination button {",
        "  width: 2rem;", "  height: 1.85rem;", "  padding: 0;",
        "  font-size: 1rem;", "  line-height: 1;", "}", "",
            ".dnaepico-viewer-table-wrap {",
        "  display: block !important;", "  flex: 1 1 auto !important;",
        "  overflow: auto !important;",
            "  border: 1px solid var(--qpasst-border);",
        "  border-radius: 0.4rem;", "  height: clamp(28rem, 62vh, 52rem);",
        "  min-height: 28rem !important;", "  max-height: none;",
        "}", "", ".dnaepico-viewer-table {", "  margin-bottom: 0;",
        "  white-space: nowrap;", "}", "", ".dnaepico-viewer-table thead th {",
        "  position: sticky;", "  top: 0;", "  z-index: 1;",
        "  background: #f4f6f8;", "}", "", ".dnaepico-viewer-status {",
        "  display: block !important;", "  flex: 0 0 auto !important;",
        "  min-height: 0 !important;", "  color: #495057;", "}",
        "", ".dnaepico-viewer-filter [data-role='filter-summary'] {",
        "  color: #495057;", "  font-size: 0.9rem;", "}", "",
        ".sidebar .dnaepico-model-formula {", "  display: inline;",
        "  white-space: normal !important;", "  overflow-wrap: anywhere;",
        "  word-break: normal;", "}", "", ".sidebar .dnaepico-model-formulas {",
        "  margin: 0.35rem 0;", "}", "",
            ".sidebar .dnaepico-model-formulas summary {",
        "  cursor: pointer;", "  font-weight: 600;", "  color: #2f3540;",
        "}", "", ".sidebar .dnaepico-model-formula-list {", "  display: grid;",
        "  gap: 0.65rem;", "  margin-top: 0.65rem;", "}", "",
        ".sidebar .dnaepico-model-formula-item {",
            "  padding: 0.55rem 0.65rem;",
        "  border: 1px solid #e2e8f0;", "  border-radius: 0.4rem;",
        "  background: #f8fafc;", "}", "",
            ".sidebar .dnaepico-model-formula-item strong,",
        ".sidebar .dnaepico-model-formula-item code {", "  display: block;",
        "}", "", ".sidebar .dnaepico-model-formula-item code {",
        "  margin-top: 0.3rem;", "}", "", ".dnaepico-viewer-error {",
        "  color: #b02a37;", "}", "", "pre code {", "  white-space: pre-wrap;",
        "  word-break: break-word;", "}", "", "@media (max-width: 768px) {",
        "  .qpasst-gallery {", "    grid-template-columns: 1fr;",
        "  }", "  .dnaepico-viewer-toolbar input {", "    min-width: 10rem;",
        "  }", "}"
    )

    data_page <- compose_page(
        title = "Data", notes = data_notes,
        body_classes = "qpasst-data-page",
            body_lines = c(build_result_table_section(
            title = "Data Preview",
            table_assets = data_table_assets
        ))
    )

    enmix_page <- compose_page(
        title = "ENmix QC", notes = enmix_notes,
        body_lines = c(build_figure_cards(enmix_items,
            "No supported image files were found for the ENmix QC tab.",
            title_prefix = "Figure", show_filename = FALSE,
                figure_titles = enmix_figure_titles
        ))
    )

    metrics_page <- compose_page(
        title = "Metrics", notes = metrics_notes,
        body_lines = c(build_figure_cards(metrics_items,
            "No supported image files were found for the Metrics tab.",
            title_prefix = "Figure", show_filename = FALSE,
                figure_titles = metrics_figure_titles
        ))
    )

    quality_control_page <- compose_page(
        title = "Quality Control",
        notes = quality_control_notes, body_lines = c(
            build_figure_cards(qc_items,
                "No supported image files were found for the Quality Control tab.",
                title_prefix = "Figure", show_filename = FALSE,
                    figure_titles = quality_control_figure_titles
            ),
            build_data_frame_table_section(
                title = "Table 1: CpG Detection Summary",
                data = detection_tables$cpg,
                    empty_message = detection_table_warning(detection_tables),
                preview_rows = 10L
            ), build_data_frame_table_section(
                title = "Table 2: Sample Detection Summary",
                data = detection_tables$sample,
                    empty_message = detection_table_warning(detection_tables),
                preview_rows = 100L
            )
        )
    )

    batch_effect_page <- compose_page(
        title = "Batch Effect",
        notes = batch_effect_notes,
            body_lines = c(build_figure_cards(batch_items,
            "No supported image files were found for the Batch Effect tab.",
            title_prefix = "Figure", show_filename = FALSE,
                figure_titles = batch_effect_figure_titles
        ))
    )

    glm_page <- compose_page(
        title = "GLM Analysis", notes = glm_notes,
        body_lines = c(build_result_table_section(
            title = glm_table_title,
            table_assets = glm_table_assets
        ))
    )

    lme_page <- compose_page(
        title = lme_analysis_label, notes = lme_notes,
        body_lines = c(build_result_table_section(
            title = lme_table_title,
            table_assets = lme_table_assets
        ))
    )

    logs_page <- compose_page(
        title = "Logs", notes = logs_notes,
        body_classes = "qpasst-logs-page", body_lines = c(
            "```{r}",
            "#| echo: false", "#| include: false",
                "render_log_block <- function(path, label) {",
            "  if (!file.exists(path)) {",
                "    cat('::: {.callout-warning}\\n')",
            "    cat(sprintf('Log file not found for `%s`.\\n', label))",
            "    cat(':::\\n')", "    return(invisible(NULL))",
            "  }", "  lines <- readLines(path, warn = FALSE, encoding = 'UTF-8')",
            "  if (!length(lines)) {", "    cat('::: {.callout-note}\\n')",
            "    cat(sprintf('`%s` is empty.\\n', label))", "    cat(':::\\n')",
            "    return(invisible(NULL))", "  }", "  cat('```text\\n')",
            "  cat(paste(lines, collapse = '\\n'))", "  cat('\\n```\\n')",
            "}", "```", "", "::: {.card title=\"Methylation Analysis\"}",
            "", "Displays the methylation preprocessing log, including IDAT loading, normalisation, filtering, and cell composition estimation steps.",
            "", "```{r}", "#| echo: false", "#| results: asis",
            sprintf(
                "render_log_block(%s, 'Methylation Analysis')",
                r_string(log_assets$methylation$asset_path)
            ),
            "```", ":::", "", "::: {.card title=\"Data Preparation\"}",
            "", "Displays the phenotype preparation log, including timepoint splitting and methylation matrix export steps.",
            "", "```{r}", "#| echo: false", "#| results: asis",
            sprintf(
                "render_log_block(%s, 'Data Preparation')",
                r_string(log_assets$data$asset_path)
            ), "```",
            ":::", "", "::: {.card title=\"Batch Effect\"}",
            "", "Displays the hidden-effect and surrogate-variable analysis log used for batch-effect assessment.",
            "", "```{r}", "#| echo: false", "#| results: asis",
            sprintf("render_log_block(%s, 'Batch Effect')",
                r_string(log_assets$batch$asset_path)),
            "```", ":::", "", "::: {.card title=\"GLM Analysis\"}",
            "", "Displays the generalised linear model log, including phenotype association testing and CpG annotation steps.",
            "", "```{r}", "#| echo: false", "#| results: asis",
            sprintf("render_log_block(%s, 'GLM Analysis')",
                r_string(log_assets$glm$asset_path)),
            "```", ":::", "", sprintf(
                "::: {.card title=\"%s\"}",
                html_escape(lme_analysis_label)
            ), "", if (has_lme_interaction) {
                sprintf(
                    "Displays the linear mixed-effects model log, including the phenotype interaction with %s and CpG annotation steps.",
                    lme_interaction_term
                )
            } else {
                "Displays the linear mixed-effects model log, including phenotype coefficient testing and CpG annotation steps."
            }, "", "```{r}", "#| echo: false", "#| results: asis",
            sprintf(
                "render_log_block(%s, %s)", r_string(log_assets$lme$asset_path),
                r_string(lme_analysis_label)
            ), "```", ":::"
        )
    )

    report_page <- build_report_page(
        project_name = project_name,
        project_dir = project_dir, data_notes = data_notes,
            enmix_notes = enmix_notes,
        quality_control_notes = quality_control_notes,
            batch_effect_notes = batch_effect_notes,
        metrics_notes = metrics_notes, glm_notes = glm_notes,
        lme_notes = lme_notes, logs_notes = logs_notes,
            lme_label = lme_analysis_label
    )

    unlink(c(
        file.path(project_dir, c("data.qmd", "DNAm.html")),
        file.path(project_dir, "docs", "data.html")
    ), force = TRUE)

    write_utf8(file.path(project_dir, "_quarto.yml"), quarto_yml)
    write_utf8(file.path(assets_dir, "qpasst.css"), site_css)
    write_utf8(file.path(project_dir, "index.qmd"), data_page)
    write_utf8(file.path(project_dir, "enmix-qc.qmd"), enmix_page)
    write_utf8(file.path(project_dir, "metrics.qmd"), metrics_page)
    write_utf8(
        file.path(project_dir, "quality-control.qmd"),
        quality_control_page
    )
    write_utf8(file.path(project_dir, "batch-effect.qmd"), batch_effect_page)
    write_utf8(file.path(project_dir, "glm.qmd"), glm_page)
    write_utf8(file.path(project_dir, "lme.qmd"), lme_page)
    write_utf8(file.path(project_dir, "report.qmd"), report_page)
    write_utf8(file.path(project_dir, "logs.qmd"), logs_page)

    quarto_bin <- find_quarto()

    source_files <- file.path(project_dir, c(
        "_quarto.yml", "index.qmd",
        "enmix-qc.qmd", "metrics.qmd", "quality-control.qmd",
        "batch-effect.qmd", "glm.qmd", "lme.qmd", "report.qmd",
        "logs.qmd"
    ))

    emitLogMinfiEwasWater(
        c(paste(
            "Generated", length(source_files),
            "report source files in:", project_dir
        )),
        verbose = FALSE,
        log_path = log_path
    )

    render_status <- "skipped"
    rendered_file <- file.path(project_dir, "docs", "index.html")
    error_message <- NULL

    if (nzchar(quarto_bin)) {
        emitLogMinfiEwasWater(c("Rendering report site..."),
            verbose = FALSE, log_path = log_path
        )
        old_wd <- getwd()
        on.exit(setwd(old_wd), add = TRUE)
        setwd(project_dir)
        render_output <- system2(quarto_bin, "render",
            stdout = TRUE,
            stderr = TRUE
        )
        status <- attr(render_output, "status")
        if (is.null(status)) {
            status <- 0L
        }
        if (length(render_output)) {
            emitLogMinfiEwasWater(c("Quarto output:", render_output),
                verbose = FALSE, log_path = log_path
            )
        }
        if (identical(status, 0L)) {
            render_status <- "rendered"
            emitLogMinfiEwasWater(c("Render complete."),
                verbose = FALSE,
                log_path = log_path
            )
        } else {
            render_status <- "failed"
            error_message <- paste(
                "Quarto render failed with status",
                status
            )
            emitLogMinfiEwasWater(error_message,
                verbose = verbose,
                log_path = log_path
            )
        }
    } else {
        render_status <- "failed"
        error_message <- paste(
            "Quarto CLI not found. The project files were created but not rendered.",
            "Install Quarto or set QUARTO_BIN to the full path of the quarto executable."
        )
        emitLogMinfiEwasWater(c(error_message, project_dir),
            verbose = verbose, log_path = log_path
        )
    }

    if (!magick_available && any(grepl("\\.tiff?$", c(
        enmix_items$original_name,
        metrics_items$original_name, qc_items$original_name,
        batch_items$original_name
    ), ignore.case = TRUE))) {
        emitLogMinfiEwasWater(
            c(
                "Note: TIFF figures were copied as-is because the `magick` package is not installed.",
                "Install `magick` and rerun this function if you want automatic PNG conversion for browser previews."
            ),
            verbose = FALSE, log_path = log_path
        )
    }

    emitLogMinfiEwasWater(paste("Report path:",
        normalizePathDnamReport(file.path(
        project_dir,
        "docs", "index.html"
    ))), verbose = verbose, log_path = log_path)

    render_result <- structure(list(
        preparedReport = prepared_report,
        status = render_status, renderedFile = if (identical(
            render_status,
            "rendered"
        )) {
            rendered_file
        } else {
            NULL
        }, errorMessage = error_message,
        logFile = log_path
    ), class = "dnaEPICO_dnamReport_render")

    structure(list(
        preparedReport = prepared_report, renderResult = render_result,
        status = render_status, outputFile = if (identical(
            render_status,
            "rendered"
        )) {
            rendered_file
        } else {
            normalizePathDnamReport(file.path(
                project_dir, "docs",
                "index.html"
            ))
        }, projectDir = project_dir, sourceFiles = source_files,
        resultTableSources = c(
            GLM = if (is.null(glm_table_assets$source_mode)) "unavailable" else glm_table_assets$source_mode,
            LME = if (is.null(lme_table_assets$source_mode)) "unavailable" else lme_table_assets$source_mode
        ),
        docsDir = file.path(project_dir, "docs"), logoPath = logo_source_path,
        errorMessage = error_message, logFile = log_path
    ), class = "dnaEPICO_dnamReport")
}

#' Print a DNA methylation report result
#'
#' @param x Object returned by [dnamReport()].
#' @param ... Additional arguments ignored.
#'
#' @return Invisibly returns `x`.
#'
#' @export
print.dnaEPICO_dnamReport <- function(x, ...) {
    report_path <- x$outputFile
    if (is.null(report_path) || !length(report_path) ||
        is.na(report_path[[1]]) ||
        !nzchar(report_path[[1]])) {
        report_path <- x$renderResult$renderedFile
    }

    cat("Class type: ", class(x)[[1]], "\n", sep = "")
    cat("Log output path: ", formatPrintPathDnamReport(x$logFile),
        "\n",
        sep = ""
    )
    cat("Report output path: ", formatPrintPathDnamReport(report_path),
        "\n",
        sep = ""
    )

    invisible(x)
}

#' Read IDAT files into an annotated RGChannelSet
#'
#' Read methylation-array IDAT files with `minfi::read.metharray.exp()`, set
#' sample names from the phenotype table, apply the requested annotation, and
#' return the resulting `RGChannelSet`.
#'
#' @param idatFolder Character. Directory containing the IDAT files.
#' @param targets Data frame returned by `readPhenotypeTargets()` or an
#'   equivalent phenotype table.
#' @param SampleID Character. Name of the phenotype column containing sample
#'   identifiers used to label the `RGChannelSet`.
#' @param arrayType Character. Array name passed to
#'   `Biobase::annotation(RGSet)`, for example
#'   `'IlluminaHumanMethylationEPICv2'`.
#' @param annotationVersion Character. Annotation build passed to
#'   `Biobase::annotation(RGSet)`, for example `'20a1.hg38'` for EPIC v2 hg38
#'   annotations or `'ilmn12.hg19'` for 450K hg19 annotations.
#' @param force Logical. Passed to `minfi::read.metharray.exp()`. Use `TRUE`
#'   only after confirming that the selected IDAT files should be read together.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return An annotated `RGChannelSet`.
#'
#' @examples
#' if (requireNamespace("minfiData", quietly = TRUE) &&
#'     requireNamespace("IlluminaHumanMethylation450kmanifest", quietly = TRUE) &&
#'     requireNamespace("IlluminaHumanMethylation450kanno.ilmn12.hg19", quietly = TRUE)) {
#'     ex <- dnaEPICO:::exampleMinfiIdatInputsDnaEpico(n = 4)
#'     rgset <- readRGSetMinfiEwasWater(
#'         idatFolder = ex$idatFolder,
#'         targets = ex$targets,
#'         SampleID = "Sample_Name",
#'         arrayType = ex$arrayType,
#'         annotationVersion = ex$annotationVersion
#'     )
#'     class(rgset)
#' }
#'
#' @export
readRGSetMinfiEwasWater <- function(
    idatFolder, targets, SampleID = "Sample_Name",
    arrayType = "IlluminaHumanMethylationEPICv2",
        annotationVersion = "20a1.hg38",
    force = FALSE, verbose = FALSE, logs = FALSE, log_dir = NULL,
    log_file = "log_readRGSetMinfiEwasWater.txt"
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )

    if (!dir.exists(idatFolder)) {
        stop("idatFolder does not exist: ", idatFolder, call. = FALSE)
    }

    if (!(SampleID %in% colnames(targets))) {
        stop("SampleID column not found in targets: ", SampleID,
            call. = FALSE
        )
    }
    force <- validateLogicalScalarDnaEpico(force, "force")

    emitLogMinfiEwasWater(
        c(
            "============================================================",
            paste("IDAT folder:              ", idatFolder), paste(
                "Array type:               ",
                arrayType
            ), paste("Annotation version:       ", annotationVersion),
            paste("Force IDAT read:          ", force)
        ),
        verbose = verbose,
        log_path = log_path
    )

    RGSet <- minfi::read.metharray.exp(
        base = idatFolder, targets = targets,
        extended = FALSE, recursive = FALSE, verbose = FALSE,
        force = force
    )

    sample_ids <- validateSampleIdentifiersDnaEpico(
        targets[[SampleID]],
        paste0("Phenotype column '", SampleID, "'")
    )
    if (ncol(RGSet) != length(sample_ids)) {
        stop("The IDAT reader returned ", ncol(RGSet), " samples for ",
            length(sample_ids), " phenotype rows.",
            call. = FALSE
        )
    }

    colnames(RGSet) <- sample_ids
    Biobase::annotation(RGSet) <- c(array = arrayType,
        annotation = annotationVersion)

    manifest_lines <-
        utils::capture.output(methods::show(minfi::getManifest(RGSet)))

    emitLogMinfiEwasWater(
        c(
            paste(
                "RGSet loaded with", ncol(RGSet),
                "samples."
            ), paste("Applied annotation:       ",
                paste(Biobase::annotation(RGSet),
                collapse = ", "
            )), "Manifest used:", manifest_lines,
            "============================================================"
        ),
        verbose = verbose, log_path = log_path
    )

    RGSet
}

#' Build raw minfi preprocessing objects
#'
#' Create a raw `MethylSet`, `RatioSet`, and genome-mapped object from an
#' `RGChannelSet`, and return them together in a single structured object.
#'
#' @param RGSet An `RGChannelSet`.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return A list with class `'dnaEPICO_minfiEwasWater_raw'` containing `MSet`,
#'   `RatioSet`, and `GSet`.
#'
#' @examplesIf requireNamespace("minfiData", quietly = TRUE)
#' ex <- dnaEPICO:::exampleMinfiBaseDataDnaEpico()
#' raw_data <- buildRawMinfiEwasWater(
#'     RGSet = ex$RGSet,
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#' names(raw_data)
#'
#' @export
buildRawMinfiEwasWater <- function(
    RGSet, verbose = FALSE, logs = FALSE,
    log_dir = NULL, log_file = "log_buildRawMinfiEwasWater.txt"
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )

    emitLogMinfiEwasWater("Running preprocessRaw(), ratioConvert(), and mapToGenome().",
        verbose = verbose, log_path = log_path
    )

    MSet <- minfi::preprocessRaw(RGSet)
    RatioSet <- minfi::ratioConvert(MSet, what = "both", keepCN = TRUE)
    GSet <- minfi::mapToGenome(RatioSet)

    meth_cols <- seq_len(min(ncol(MSet), 3L))
    gset_cols <- seq_len(min(ncol(GSet), 5L))

    emitLogMinfiEwasWater(
        c(
            paste(
                "MSet dimensions:          ",
                paste(dim(MSet), collapse = " x ")
            ), paste(
                "RatioSet dimensions:      ",
                paste(dim(RatioSet), collapse = " x ")
            ), paste(
                "GSet dimensions:          ",
                paste(dim(GSet), collapse = " x ")
            ), "Preview of methylated intensities:",
            previewLinesMinfiEwasWater(minfi::getMeth(MSet)[, meth_cols,
                drop = FALSE
            ]), "Preview of unmethylated intensities:",
            previewLinesMinfiEwasWater(minfi::getUnmeth(MSet)[, meth_cols,
                drop = FALSE
            ]), "Preview of beta values:",
                previewLinesMinfiEwasWater(minfi::getBeta(GSet)[,
                gset_cols,
                drop = FALSE
            ]), "Preview of M-values:",
            previewLinesMinfiEwasWater(minfi::getM(GSet)[, gset_cols,
                drop = FALSE
            ]), "Preview of copy-number values:",
            previewLinesMinfiEwasWater(minfi::getCN(GSet)[, gset_cols,
                drop = FALSE
            ]), "============================================================"
        ),
        verbose = verbose, log_path = log_path
    )

    structure(list(MSet = MSet, RatioSet = RatioSet, GSet = GSet),
        class = "dnaEPICO_minfiEwasWater_raw"
    )
}

#' Assess sample quality before sample filtering
#'
#' Compute minfi QC metrics and detection P values, identify failed samples
#' using the requested threshold, and return the assessment as a single object.
#'
#' @param rawData Object returned by `buildRawMinfiEwasWater()`.
#' @param RGSet An `RGChannelSet` aligned with `rawData`.
#' @param qcCutoff Numeric. Cutoff passed to `minfi::plotQC()` when the QC plot
#'   is drawn.
#' @param detPtype Character. Detection P-value mode passed to
#'   `minfi::detectionP()`. Common values in minfi workflows are `'m+u'` and
#'   `'negative'`. The default used here is `'m+u'`.
#' @param detPThreshold Numeric. Samples with mean detection P value above this
#'   threshold are marked as failed.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return A list with class `'dnaEPICO_minfiEwasWater_assessment'` containing
#'   the QC object, detection P matrix, mean detection P values, and failed
#'   sample identifiers.
#'
#' @examplesIf requireNamespace("minfiData", quietly = TRUE)
#' ex <- dnaEPICO:::exampleMinfiBaseDataDnaEpico()
#' raw_data <- buildRawMinfiEwasWater(ex$RGSet, verbose = FALSE, logs = FALSE)
#' assessment <- assessSamplesMinfiEwasWater(
#'     rawData = raw_data,
#'     RGSet = ex$RGSet,
#'     detPThreshold = 1,
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#' names(assessment)
#'
#' @export
assessSamplesMinfiEwasWater <- function(
    rawData, RGSet, qcCutoff = 10.5,
    detPtype = "m+u", detPThreshold = 0.05, verbose = FALSE,
    logs = FALSE, log_dir = NULL,
        log_file = "log_assessSamplesMinfiEwasWater.txt"
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )

    qc <- minfi::getQC(rawData$MSet)
    detP <- minfi::detectionP(RGSet, type = detPtype)
    detPThreshold <- validateProbabilityDnaEpico(
        detPThreshold,
        "detPThreshold"
    )
    nanDetPConverted <- sum(is.nan(detP))
    if (nanDetPConverted > 0L) {
        detP[is.nan(detP)] <- NA_real_
    }

    missingDetP <- colSums(!is.finite(detP))
    observedDetP <- colSums(is.finite(detP))
    meanDetP <- vapply(seq_len(ncol(detP)), function(index) {
        meanFiniteOrNADnaEpico(detP[
            ,
            index
        ])
    }, numeric(1))
    names(meanDetP) <- colnames(detP)
    failedSamples <- names(meanDetP[!is.finite(meanDetP) | meanDetP >
        detPThreshold])

    preview_cols <- seq_len(min(ncol(detP), 5L))

    emitLogMinfiEwasWater(
        c(
            paste(
                "QC cutoff (median):       ",
                qcCutoff
            ), paste("Detection P type:         ", detPtype),
            paste("Detection P threshold:    ", detPThreshold), paste(
                "Detection P NaN to NA:     ",
                nanDetPConverted
            ), paste(
                "Missing detection values: ",
                sum(missingDetP)
            ), "Preview of detection P values:",
            previewLinesMinfiEwasWater(detP[, preview_cols, drop = FALSE]),
            paste("Number of failed samples: ", length(failedSamples)),
            if (length(failedSamples) > 0L) {
                paste("Failed sample IDs:        ", paste(failedSamples,
                    collapse = ", "
                ))
            } else {
                "Failed sample IDs:         none"
            }, "============================================================"
        ),
        verbose = verbose, log_path = log_path
    )

    structure(
        list(
            qc = qc, qcCutoff = qcCutoff, detP = detP,
            detPtype = detPtype, detPThreshold = detPThreshold,
                nanDetPConverted = nanDetPConverted,
            observedDetP = observedDetP, missingDetP = missingDetP,
            meanDetP = meanDetP, failedSamples = failedSamples
        ),
        class = "dnaEPICO_minfiEwasWater_assessment"
    )
}

#' Plot quality-assessment outputs for preprocessingMinfiEwasWater
#'
#' Draw either the minfi QC plot or the detection P-value plot from an
#' assessment object returned by `assessSamplesMinfiEwasWater()`.
#'
#' @param assessment Object returned by `assessSamplesMinfiEwasWater()`.
#' @param plot Character. Plot type: `'qc'` or `'detection'`.
#' @param display Logical. If `TRUE`, draw the plot on the active graphics
#'   device.
#' @param file Character or `NULL`. TIFF file written when supplied.
#' @param width Integer. TIFF width in pixels when `file` is supplied.
#' @param height Integer. TIFF height in pixels when `file` is supplied.
#' @param res Integer. TIFF resolution in DPI when `file` is supplied.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return Invisibly returns the saved TIFF path when `file` is supplied,
#'   otherwise `NULL`.
#'
#' @examples
#' assessment <- list(
#'     meanDetP = c(S1 = 0.01, S2 = 0.02, S3 = 0.04),
#'     detPThreshold = 0.05
#' )
#' plotAssessmentMinfiEwasWater(
#'     assessment = assessment,
#'     plot = "detection",
#'     display = FALSE,
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#'
#' @export
plotAssessmentMinfiEwasWater <- function(
    assessment, plot = c(
        "qc",
        "detection"
    ), display = FALSE, file = NULL, width = 2000L,
    height = 1000L, res = 150L, verbose = FALSE, logs = FALSE,
    log_dir = NULL, log_file = "log_plotAssessmentMinfiEwasWater.txt"
) {
    plot <- match.arg(plot)

    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )

    draw_fun <- switch(plot,
        qc = function() {
            minfi::plotQC(assessment$qc, badSampleCutoff = assessment$qcCutoff)
        },
        detection = function() {
            graphics::barplot(assessment$meanDetP,
                las = 3, cex.names = 0.8,
                ylab = "Mean detection p-values"
            )
            graphics::abline(
                h = assessment$detPThreshold, col = "red",
                lwd = 2, lty = 2
            )
        }
    )

    runPlotMinfiEwasWater(
        draw_fun = draw_fun, display = display,
        file = file, width = width, height = height, res = res
    )

    emitLogMinfiEwasWater(
        c(paste(
            "Assessment plot created:   ",
            plot
        ), if (is.null(file)) {
            "Assessment plot file:      none"
        } else {
            paste("Assessment plot file:      ", file)
        }, "============================================================"),
        verbose = verbose, log_path = log_path
    )

    invisible(file)
}

#' Filter failed samples from an RGSet and phenotype table
#'
#' Remove failed samples identified during sample assessment and return the
#' filtered `RGChannelSet` together with the aligned phenotype table.
#'
#' @param RGSet An `RGChannelSet`.
#' @param targets Data frame containing phenotype information.
#' @param failedSamples Character vector of sample identifiers to remove.
#' @param SampleID Character. Name of the sample identifier column in `targets`.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return A list with class `'dnaEPICO_minfiEwasWater_samples'` containing the
#'   filtered `RGSet`, aligned phenotype table, and failed sample identifiers.
#'
#' @examplesIf requireNamespace("minfiData", quietly = TRUE)
#' ex <- dnaEPICO:::exampleMinfiBaseDataDnaEpico()
#' filtered_samples <- filterSamplesMinfiEwasWater(
#'     RGSet = ex$RGSet,
#'     targets = ex$targets,
#'     failedSamples = ex$targets$Sample_Name[1],
#'     SampleID = "Sample_Name",
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#' nrow(filtered_samples$targets)
#'
#' @export
filterSamplesMinfiEwasWater <- function(
    RGSet, targets, failedSamples = character(0),
    SampleID = "Sample_Name", verbose = FALSE, logs = FALSE,
    log_dir = NULL, log_file = "log_filterSamplesMinfiEwasWater.txt"
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )

    if (!(SampleID %in% colnames(targets))) {
        stop("SampleID column not found in targets: ", SampleID,
            call. = FALSE
        )
    }
    sample_names <- validateSampleIdentifiersDnaEpico(
        colnames(RGSet),
        "RGSet sample identifiers"
    )
    target_sample_ids <- validateSampleIdentifiersDnaEpico(
        targets[[SampleID]],
        paste0("phenotype column '", SampleID, "'")
    )
    matchSampleIdentifiersDnaEpico(
        query = sample_names, reference = target_sample_ids,
        queryLabel = "RGSet sample identifiers", referenceLabel = paste0(
            "phenotype column '",
            SampleID, "'"
        ), requireSameSet = TRUE
    )
    if (length(failedSamples) > 0L) {
        failedSamples <- validateSampleIdentifiersDnaEpico(
            failedSamples,
            "failedSamples"
        )
        unknown_failed_samples <- setdiff(failedSamples, sample_names)
        if (length(unknown_failed_samples) > 0L) {
            stop("failedSamples are not present in RGSet: ",
                paste(unknown_failed_samples, collapse = ", "),
                call. = FALSE
            )
        }
    }

    keep_samples <- !(sample_names %in% failedSamples)
    if (!any(keep_samples)) {
        stop("Sample filtering would remove every sample.", call. = FALSE)
    }
    RGSet_filtered <- RGSet[, keep_samples]

    sample_names <- colnames(RGSet_filtered)
    matched <- matchSampleIdentifiersDnaEpico(
        query = sample_names,
        reference = targets[[SampleID]],
            queryLabel = "RGSet sample identifiers",
        referenceLabel = paste0(
            "phenotype column '", SampleID,
            "'"
        )
    )
    targets_filtered <- targets[matched, , drop = FALSE]
    rownames(targets_filtered) <- NULL

    emitLogMinfiEwasWater(
        c(
            paste(
                "Samples before filtering: ",
                ncol(RGSet)
            ), paste("Samples after filtering:  ", ncol(RGSet_filtered)),
            "============================================================"
        ),
        verbose = verbose, log_path = log_path
    )

    structure(list(
        RGSet = RGSet_filtered, targets = targets_filtered,
        failedSamples = failedSamples
    ), class = "dnaEPICO_minfiEwasWater_samples")
}

#' Predict biological sex from a filtered raw-data object
#'
#' Predict sample sex from a genome-mapped methylation object, align the
#' predictions with phenotype data, and return a structured object that can be
#' plotted or merged into downstream phenotype tables.
#'
#' @param rawData Object returned by `buildRawMinfiEwasWater()`.
#' @param targets Filtered phenotype data frame aligned with `rawData`.
#' @param SampleID Character. Name of the sample identifier column in `targets`.
#' @param sexColumn Character. Name of the phenotype column containing reported
#'   sex.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return A list with class `'dnaEPICO_minfiEwasWater_sex'` containing the sex
#'   prediction result, aligned phenotype data, plotting data, and mismatch
#'   table.
#'
#' @examplesIf requireNamespace("minfiData", quietly = TRUE)
#' ex <- dnaEPICO:::exampleMinfiWorkflowStateDnaEpico()
#' sex_data <- predictSexMinfiEwasWater(
#'     rawData = ex$rawFiltered,
#'     targets = ex$sampleData$targets,
#'     SampleID = "Sample_Name",
#'     sexColumn = "Sex",
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#' names(sex_data)
#'
#' @export
predictSexMinfiEwasWater <- function(
    rawData, targets, SampleID = "Sample_Name",
    sexColumn = "Sex", verbose = FALSE, logs = FALSE, log_dir = NULL,
    log_file = "log_predictSexMinfiEwasWater.txt"
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )

    if (!(SampleID %in% colnames(targets))) {
        stop("SampleID column not found in targets: ", SampleID,
            call. = FALSE
        )
    }

    if (!(sexColumn %in% colnames(targets))) {
        stop("sexColumn not found in targets: ", sexColumn, call. = FALSE)
    }

    pSex <- as.data.frame(minfi::getSex(rawData$GSet), stringsAsFactors = FALSE)
    numeric_sex_columns <- names(pSex)[vapply(
        pSex, is.numeric,
        logical(1)
    )]
    nanSexConverted <- sum(vapply(
        pSex[numeric_sex_columns],
        function(values) sum(is.nan(values)), integer(1)
    ))
    for (column in numeric_sex_columns) {
        pSex[[column]][is.nan(pSex[[column]])] <- NA_real_
    }
    sex_plot_data <- as.data.frame(pSex, stringsAsFactors = FALSE)
    sex_plot_data$SampleID <- rownames(sex_plot_data)

    matched <- matchSampleIdentifiersDnaEpico(
        query = sex_plot_data$SampleID,
        reference = targets[[SampleID]],
            queryLabel = "Predicted-sex sample identifiers",
        referenceLabel = paste0(
            "phenotype column '", SampleID,
            "'"
        ), requireSameSet = TRUE
    )
    targets_aligned <- targets[matched, , drop = FALSE]
    rownames(targets_aligned) <- NULL

    reported_sex <- canonicalizeSexDnaEpico(targets_aligned[[sexColumn]])
    predicted_sex <- canonicalizeSexDnaEpico(pSex$predictedSex)
    targets_aligned$PredSex <- predicted_sex$code
    targets_aligned$SexMismatch <- !is.na(reported_sex$code) &
        !is.na(predicted_sex$code) & reported_sex$code != predicted_sex$code

    sex_plot_data[[sexColumn]] <- reported_sex$code
    sex_plot_data$PredSex <- predicted_sex$code
    sex_plot_data$SexMismatch <- targets_aligned$SexMismatch
    mismatches <- targets_aligned[targets_aligned$SexMismatch, ,
        drop = FALSE
    ]

    sex_values <- targets[[sexColumn]]
    nSexNA <- sum(is.na(sex_values))
    unknown_sex <- reported_sex$unknown

    emitLogMinfiEwasWater(
        c(
            "Sex column integrity check:", paste(
                "  Sex column used:        ",
                sexColumn
            ), paste("  NA values:              ", nSexNA),
            paste("  NaN values converted:   ", nanSexConverted),
            paste("  Unknown labels:         ", if (length(unknown_sex) ==
                0L) {
                "none"
            } else {
                paste(unknown_sex, collapse = ", ")
            }), paste("Mismatches found:         ", nrow(mismatches)),
            if (nrow(mismatches) > 0L) {
                previewLinesMinfiEwasWater(mismatches[, seq_len(min(
                    3L,
                    ncol(mismatches)
                )), drop = FALSE])
            } else {
                character(0)
            }, "============================================================"
        ),
        verbose = verbose, log_path = log_path
    )

    structure(
        list(
            pSex = pSex, targets = targets_aligned, sexPlotData = sex_plot_data,
            mismatches = mismatches,
                reportedSexOriginal = reported_sex$original,
            reportedSexCode = reported_sex$code,
                reportedSexLabel = reported_sex$label,
            predictedSexCode = predicted_sex$code,
                predictedSexLabel = predicted_sex$label,
            unknownReportedSex = reported_sex$unknown,
                unknownPredictedSex = predicted_sex$unknown,
            nanConverted = nanSexConverted, removedSampleIDs = character(0),
            removeSexMismatch = FALSE, SampleID = SampleID,
                sexColumn = sexColumn
        ),
        class = "dnaEPICO_minfiEwasWater_sex"
    )
}

#' Plot predicted or clinical sex from `predictSexMinfiEwasWater()`
#'
#' @param sexData Object returned by `predictSexMinfiEwasWater()`.
#' @param type Character. Plot type: `'predicted'` for methylation-predicted
#'   sex or `'clinical'` for reported sex.
#' @param display Logical. If `TRUE`, draw the plot on the active graphics
#'   device.
#' @param file Character or `NULL`. TIFF file written when supplied.
#' @param width Integer. TIFF width in pixels when `file` is supplied.
#' @param height Integer. TIFF height in pixels when `file` is supplied.
#' @param res Integer. TIFF resolution in DPI when `file` is supplied.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return Invisibly returns the saved TIFF path when `file` is supplied,
#'   otherwise `NULL`.
#'
#' @examples
#' ex <- dnaEPICO:::exampleSexPlotStateDnaEpico()
#' plotSexMinfiEwasWater(
#'     sexData = ex,
#'     type = "predicted",
#'     display = FALSE,
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#'
#' @export
plotSexMinfiEwasWater <- function(
    sexData, type = c(
        "predicted",
        "clinical"
    ), display = FALSE, file = NULL, width = 2000L,
    height = 1000L, res = 70L, verbose = FALSE, logs = FALSE,
    log_dir = NULL, log_file = "log_plotSexMinfiEwasWater.txt"
) {
    type <- match.arg(type)

    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )

    draw_fun <- switch(type,
        predicted = function() {
            graphics::plot(
                x = sexData$pSex$xMed, y = sexData$pSex$yMed,
                type = "n", xlab = "X chr, median total intensity (log2)",
                ylab = "Y chr, median total intensity (log2)"
            )
            graphics::text(
                x = sexData$pSex$xMed, y = sexData$pSex$yMed,
                labels = sexData$targets[[sexData$SampleID]],
                    col = ifelse(sexData$pSex$predictedSex ==
                    "M", "deepskyblue", "deeppink3")
            )
            graphics::legend("bottomleft", c("M", "F"), col = c(
                "deepskyblue",
                "deeppink3"
            ), pch = 16)
        },
        clinical = function() {
            graphics::plot(
                x = sexData$sexPlotData$xMed, y = sexData$sexPlotData$yMed,
                type = "n", xlab = "X chr, median total intensity (log2)",
                ylab = "Y chr, median total intensity (log2)"
            )
            graphics::text(
                x = sexData$sexPlotData$xMed, y = sexData$sexPlotData$yMed,
                labels = sexData$sexPlotData$SampleID,
                    col = ifelse(sexData$sexPlotData[[sexData$sexColumn]] ==
                    1L, "deepskyblue", "deeppink3")
            )
            graphics::legend("bottomleft", c("M", "F"), col = c(
                "deepskyblue",
                "deeppink3"
            ), pch = 16)
        }
    )

    runPlotMinfiEwasWater(
        draw_fun = draw_fun, display = display,
        file = file, width = width, height = height, res = res
    )

    emitLogMinfiEwasWater(
        c(paste(
            "Sex plot created:         ",
            type
        ), if (is.null(file)) {
            "Sex plot file:            none"
        } else {
            paste("Sex plot file:            ", file)
        }, "============================================================"),
        verbose = verbose, log_path = log_path
    )

    invisible(file)
}

#' Normalize filtered samples with minfi and wateRmelon methods
#'
#' Apply one or more supported normalization methods to a filtered `RGSet` and
#' return all normalized objects together in a single result object.
#'
#' @param sampleData Object returned by `filterSamplesMinfiEwasWater()`.
#' @param sexColumn Character. Name of the phenotype column used as the optional
#'   sex covariate for normalization methods that support it.
#' @param normMethods Character vector or semicolon-separated string of
#'   normalization methods. Supported values are `'adjustedfunnorm'`,
#'   `'funnorm'`, `'illumina'`, `'quantile'`, and `'swan'`.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return A list with class `'dnaEPICO_minfiEwasWater_norm'` containing the
#'   requested normalized objects and the first method as `primary`.
#'
#' @examplesIf requireNamespace("minfiData", quietly = TRUE)
#' ex <- dnaEPICO:::exampleMinfiBaseDataDnaEpico()
#' sample_data <- filterSamplesMinfiEwasWater(
#'     RGSet = ex$RGSet,
#'     targets = ex$targets,
#'     failedSamples = character(0),
#'     SampleID = "Sample_Name",
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#' norm_data <- normalizeMinfiEwasWater(
#'     sampleData = sample_data,
#'     sexColumn = "Sex",
#'     normMethods = "quantile",
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#' names(norm_data$normalized)
#'
#' @export
normalizeMinfiEwasWater <- function(
    sampleData, sexColumn = "Sex",
    normMethods = "adjustedfunnorm", verbose = FALSE, logs = FALSE,
    log_dir = NULL, log_file = "log_normalizeMinfiEwasWater.txt"
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )

    method_list <- tolower(splitOptionMinfiEwasWater(normMethods,
        sep = ";"
    ))
    supported_methods <- c(
        "adjustedfunnorm", "funnorm", "illumina",
        "quantile", "swan"
    )
    if (length(method_list) == 0L) {
        stop("At least one normalization method must be supplied.",
            call. = FALSE
        )
    }
    unknown_methods <- setdiff(method_list, supported_methods)
    if (length(unknown_methods) > 0L) {
        stop("Unknown normalization method(s): ", paste(unknown_methods,
            collapse = ", "
        ), call. = FALSE)
    }
    normalized <- vector("list", length(method_list))
    names(normalized) <- method_list

    col_data <- SummarizedExperiment::colData(sampleData$RGSet)
    sex_vec <- resolveNormalizationSexDnaEpico(col_data, sexColumn)

    emitLogMinfiEwasWater(c(paste(
        "Normalization methods:    ",
        paste(method_list, collapse = ", ")
    ), paste(
        "Sex column:               ",
        sexColumn
    )), verbose = verbose, log_path = log_path)

    for (i in seq_along(method_list)) {
        method <- method_list[[i]]
        emitLogMinfiEwasWater(paste(
            "Applying normalization:   ",
            method
        ), verbose = verbose, log_path = log_path)

        normalized[[i]] <- switch(method,
            adjustedfunnorm = wateRmelon::adjustedFunnorm(sampleData$RGSet,
                sex = sex_vec
            ),
            funnorm = minfi::preprocessFunnorm(sampleData$RGSet,
                sex = sex_vec
            ),
            illumina = minfi::preprocessIllumina(sampleData$RGSet),
            quantile = minfi::preprocessQuantile(sampleData$RGSet,
                sex = sex_vec
            ),
            swan = minfi::preprocessSWAN(sampleData$RGSet),
            stop("Unknown normalization method: ", method, call. = FALSE)
        )
    }

    emitLogMinfiEwasWater("============================================================",
        verbose = verbose, log_path = log_path
    )

    structure(list(
        primary = normalized[[1L]], normalized = normalized,
        methods = method_list, sexColumn = sexColumn
    ), class = "dnaEPICO_minfiEwasWater_norm")
}

#' Plot raw and normalized methylation distributions
#'
#' Draw the density comparison plot used to inspect raw versus normalized data.
#'
#' @param RGSet An `RGChannelSet` aligned with `targets`.
#' @param normData Object returned by `normalizeMinfiEwasWater()`.
#' @param targets Filtered phenotype data aligned with `RGSet`.
#' @param sexColumn Character. Name of the phenotype column used to colour the
#'   density curves.
#' @param display Logical. If `TRUE`, draw the plot on the active graphics
#'   device.
#' @param file Character or `NULL`. TIFF file written when supplied.
#' @param width Integer. TIFF width in pixels when `file` is supplied.
#' @param height Integer. TIFF height in pixels when `file` is supplied.
#' @param res Integer. TIFF resolution in DPI when `file` is supplied.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return Invisibly returns the saved TIFF path when `file` is supplied,
#'   otherwise `NULL`.
#'
#' @examples
#' ex <- dnaEPICO:::exampleMinfiMetricsStateDnaEpico()
#' plotNormalizationMinfiEwasWater(
#'     RGSet = ex$beta,
#'     normData = ex$normData,
#'     targets = ex$targets,
#'     sexColumn = "Sex",
#'     display = FALSE,
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#'
#' @export
plotNormalizationMinfiEwasWater <- function(
    RGSet, normData,
    targets, sexColumn = "Sex", display = FALSE, file = NULL,
    width = 2000L, height = 1000L, res = 150L, verbose = FALSE,
    logs = FALSE, log_dir = NULL,
        log_file = "log_plotNormalizationMinfiEwasWater.txt"
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )

    draw_fun <- function() {
        graphics::par(mfrow = c(1, 2))
        minfi::densityPlot(RGSet,
            sampGroups = targets[[sexColumn]],
            main = "Raw", legend = FALSE
        )
        graphics::legend("top",
            legend = levels(factor(targets[[sexColumn]])),
            text.col = RColorBrewer::brewer.pal(8, "Dark2")
        )

        minfi::densityPlot(minfi::getBeta(normData$primary),
            sampGroups = targets[[sexColumn]], main = "Normalized",
            legend = FALSE
        )
        graphics::legend("top",
            legend = levels(factor(targets[[sexColumn]])),
            text.col = RColorBrewer::brewer.pal(8, "Dark2")
        )
    }

    runPlotMinfiEwasWater(
        draw_fun = draw_fun, display = display,
        file = file, width = width, height = height, res = res
    )

    emitLogMinfiEwasWater(
        c(if (is.null(file)) {
            "Raw-vs-normalized plot:   not written to file"
        } else {
            paste("Raw-vs-normalized plot:   ", file)
        }, "============================================================"),
        verbose = verbose, log_path = log_path
    )

    invisible(file)
}

#' Plot raw beta-value density from a raw preprocessing object
#'
#' Draw the pre-normalization beta density plot from a raw minfi object and a
#' grouping variable in the phenotype table.
#'
#' @param rawData Object returned by `buildRawMinfiEwasWater()`.
#' @param targets Filtered phenotype data aligned with `rawData`.
#' @param plotGroupVar Character. Phenotype column used to group samples in the
#'   density plot.
#' @param display Logical. If `TRUE`, draw the plot on the active graphics
#'   device.
#' @param file Character or `NULL`. TIFF file written when supplied.
#' @param width Integer. TIFF width in pixels when `file` is supplied.
#' @param height Integer. TIFF height in pixels when `file` is supplied.
#' @param res Integer. TIFF resolution in DPI when `file` is supplied.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return Invisibly returns the saved TIFF path when `file` is supplied,
#'   otherwise `NULL`.
#'
#' @examples
#' ex <- dnaEPICO:::exampleMinfiMetricsStateDnaEpico()
#' plotRawDensityMinfiEwasWater(
#'     rawData = ex$rawData,
#'     targets = ex$targets,
#'     plotGroupVar = "Sex",
#'     display = FALSE,
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#'
#' @export
plotRawDensityMinfiEwasWater <- function(
    rawData, targets, plotGroupVar = "Sex",
    display = FALSE, file = NULL, width = 2000L, height = 1000L,
    res = 150L, verbose = FALSE, logs = FALSE, log_dir = NULL,
    log_file = "log_plotRawDensityMinfiEwasWater.txt"
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )

    draw_fun <- function() {
        minfi::densityPlot(minfi::getBeta(rawData$MSet),
            sampGroups = targets[[plotGroupVar]],
            pal = RColorBrewer::brewer.pal(8, "Dark2"), main = paste(
                "Density Plot of Beta Values by",
                plotGroupVar
            ), legend = TRUE
        )
    }

    runPlotMinfiEwasWater(
        draw_fun = draw_fun, display = display,
        file = file, width = width, height = height, res = res
    )

    emitLogMinfiEwasWater(
        c(if (is.null(file)) {
            "Raw beta density plot:    not written to file"
        } else {
            paste("Raw beta density plot:    ", file)
        }, "============================================================"),
        verbose = verbose, log_path = log_path
    )

    invisible(file)
}

#' Read probe-exclusion identifiers from one CSV file
#'
#' @param probeExclusionPath Character. Path to a CSV file containing probe
#'   identifiers to exclude.
#' @param probeExclusionIdColumn Character or `NULL`. Column containing probe
#'   identifiers. When `NULL` or `''`, common names are auto-detected before
#'   falling back to an unlabeled or probe-like first column.
#' @param featureNames Character vector or `NULL`. Optional array feature names
#'   used to count overlap.
#'
#' @return A list with class `'dnaEPICO_probeExclusion_file_ids'`.
#'
#' @keywords internal
#' @noRd
readProbeExclusionFileIdsMinfiEwasWater <- function(
    probeExclusionPath,
    probeExclusionIdColumn = NULL, featureNames = NULL
) {
    if (!file.exists(probeExclusionPath)) {
        stop("probeExclusionPath does not exist: ", probeExclusionPath,
            call. = FALSE
        )
    }

    probe_exclusion <- utils::read.csv(probeExclusionPath,
        stringsAsFactors = FALSE,
        check.names = FALSE
    )

    if (ncol(probe_exclusion) == 0L) {
        stop("probeExclusionPath contains no columns: ", probeExclusionPath,
            call. = FALSE
        )
    }

    id_column <- probeExclusionIdColumn
    if (length(id_column) == 0L || is.na(id_column[[1L]])) {
        id_column <- NULL
    } else {
        id_column <- trimws(as.character(id_column[[1L]]))
        if (!nzchar(id_column) || identical(
            toupper(id_column),
            "NULL"
        )) {
            id_column <- NULL
        }
    }

    available_columns <- colnames(probe_exclusion)
    standard_columns <- c("ProbeID", "TargetID", "IlmnID", "Name")

    if (!is.null(id_column)) {
        if (!(id_column %in% available_columns)) {
            stop("probeExclusionIdColumn not found in probe-exclusion file: ",
                id_column, ". Available columns: ",
                    paste(ifelse(nzchar(available_columns),
                    available_columns, "<blank>"
                ), collapse = ", "),
                call. = FALSE
            )
        }
        selected_column <- id_column
        selected_index <- match(id_column, available_columns)
        source <- "configured"
    } else {
        standard_match <- standard_columns[standard_columns %in%
            available_columns]
        if (length(standard_match) > 0L) {
            selected_column <- standard_match[[1L]]
            selected_index <- match(selected_column, available_columns)
            source <- "standard"
        } else {
            first_column <- available_columns[[1L]]
            first_values <- trimws(as.character(probe_exclusion[[1L]]))
            first_values <- first_values[!is.na(first_values) &
                nzchar(first_values)]
            first_column_is_unlabeled <- !nzchar(first_column)
            first_column_is_probe_like <- length(first_values) >
                0L && all(grepl("^(cg|ch\\.)", first_values))

            if (isTRUE(first_column_is_unlabeled) ||
                isTRUE(first_column_is_probe_like)) {
                selected_column <- first_column
                selected_index <- 1L
                source <- if (isTRUE(first_column_is_unlabeled)) {
                    "unlabeled first column"
                } else {
                    "probe-like first column"
                }
            } else {
                stop("Could not detect a probe-exclusion ID column in ",
                    probeExclusionPath,
                        ". Set probeExclusionIdColumn to one of: ",
                    paste(ifelse(nzchar(available_columns), available_columns,
                        "<blank>"
                    ), collapse = ", "),
                    call. = FALSE
                )
            }
        }
    }

    ids <- trimws(as.character(probe_exclusion[[selected_index]]))
    ids <- unique(ids[!is.na(ids) & nzchar(ids)])

    if (length(ids) == 0L) {
        stop("Probe-exclusion ID column contains no usable IDs: ",
            ifelse(nzchar(selected_column), selected_column,
                "<blank>"
            ),
            call. = FALSE
        )
    }

    overlap <- NA_integer_
    if (!is.null(featureNames)) {
        overlap <- sum(ids %in% featureNames)
    }

    structure(list(
        ids = ids, path = probeExclusionPath, column = selected_column,
        source = source, nRows = nrow(probe_exclusion), nIds = length(ids),
        overlap = overlap
    ), class = "dnaEPICO_probeExclusion_file_ids")
}

normalizeProbeExclusionPathsMinfiEwasWater <- function(probeExclusionPath) {
    if (length(probeExclusionPath) == 0L || all(is.na(probeExclusionPath))) {
        return(character(0))
    }

    paths <- splitOptionMinfiEwasWater(probeExclusionPath, sep = ";")
    paths[toupper(paths) != "NULL"]
}

normalizeProbeExclusionColumnsMinfiEwasWater <- function(
    probeExclusionIdColumn,
    nFiles
) {
    if (length(probeExclusionIdColumn) == 0L ||
        all(is.na(probeExclusionIdColumn))) {
        return(rep(list(NULL), nFiles))
    }

    columns <- splitOptionMinfiEwasWater(probeExclusionIdColumn,
        sep = ";"
    )
    if (length(columns) == 0L || identical(
        toupper(columns[[1L]]),
        "NULL"
    )) {
        return(rep(list(NULL), nFiles))
    }

    if (length(columns) == 1L) {
        return(rep(as.list(columns), nFiles))
    }

    if (length(columns) != nFiles) {
        stop("probeExclusionIdColumn must be NULL, a single column name, or one ",
            "semicolon-separated column name per probe-exclusion file.",
            call. = FALSE
        )
    }

    as.list(columns)
}

#' Read probe-exclusion identifiers from one or more CSV files
#'
#' @param probeExclusionPath Character vector or semicolon-separated string of
#'   CSV files containing probe identifiers to exclude.
#' @param probeExclusionIdColumn Character, `NULL`, or semicolon-separated
#'   string of column names.
#' @param featureNames Character vector or `NULL`. Optional array feature names
#'   used to count overlap.
#'
#' @return A list with class `'dnaEPICO_probeExclusion_ids'`.
#'
#' @keywords internal
#' @noRd
readProbeExclusionIdsMinfiEwasWater <- function(
    probeExclusionPath,
    probeExclusionIdColumn = NULL, featureNames = NULL
) {
    paths <- normalizeProbeExclusionPathsMinfiEwasWater(probeExclusionPath)
    columns <- normalizeProbeExclusionColumnsMinfiEwasWater(
        probeExclusionIdColumn = probeExclusionIdColumn,
        nFiles = length(paths)
    )

    if (length(paths) == 0L) {
        return(structure(
            list(
                ids = character(0), files = data.frame(
                    path = character(0),
                    column = character(0), source = character(0),
                        nRows = integer(0),
                    nIds = integer(0), overlap = integer(0),
                        stringsAsFactors = FALSE
                ),
                nIds = 0L,
                    overlap = if (is.null(featureNames)) NA_integer_ else 0L
            ),
            class = "dnaEPICO_probeExclusion_ids"
        ))
    }

    file_results <- Map(f = function(path, column) {
        readProbeExclusionFileIdsMinfiEwasWater(
            probeExclusionPath = path,
            probeExclusionIdColumn = column, featureNames = featureNames
        )
    }, path = paths, column = columns)

    ids <- unique(unlist(lapply(file_results, `[[`, "ids"), use.names = FALSE))
    overlap <- NA_integer_
    if (!is.null(featureNames)) {
        overlap <- sum(ids %in% featureNames)
    }

    files <- do.call(rbind, lapply(file_results, function(result) {
        data.frame(
            path = result$path, column = ifelse(nzchar(result$column),
                result$column, "<blank>"
            ), source = result$source,
            nRows = result$nRows, nIds = result$nIds, overlap = result$overlap,
            stringsAsFactors = FALSE
        )
    }))

    structure(list(
        ids = ids, files = files, nIds = length(ids),
        overlap = overlap
    ), class = "dnaEPICO_probeExclusion_ids")
}

normalizeEpicV2ManifestFlagsMinfiEwasWater <- function(epicV2ManifestFlags) {
    default_flags <- c(
        CH_WGBS_evidence = TRUE, CH_BLAT = TRUE,
        MissingPos = TRUE, MismatchPos = FALSE
    )

    if (length(epicV2ManifestFlags) == 0L || all(is.na(epicV2ManifestFlags))) {
        return(default_flags)
    }

    if (is.logical(epicV2ManifestFlags)) {
        if (is.null(names(epicV2ManifestFlags))) {
            if (length(epicV2ManifestFlags) != length(default_flags)) {
                stop("Unnamed epicV2ManifestFlags must have ",
                    length(default_flags), " values.",
                    call. = FALSE
                )
            }
            names(epicV2ManifestFlags) <- names(default_flags)
        }

        unknown_flags <- setdiff(
            names(epicV2ManifestFlags),
            names(default_flags)
        )
        if (length(unknown_flags) > 0L) {
            stop("Unknown EPICv2 manifest flag(s): ", paste(unknown_flags,
                collapse = ", "
            ), call. = FALSE)
        }

        if (anyNA(epicV2ManifestFlags)) {
            stop("epicV2ManifestFlags must contain only TRUE or FALSE values.",
                call. = FALSE
            )
        }

        default_flags[names(epicV2ManifestFlags)] <- epicV2ManifestFlags
        return(default_flags)
    }

    flag_values <- splitOptionMinfiEwasWater(epicV2ManifestFlags,
        sep = ";"
    )
    if (length(flag_values) == 0L) {
        return(default_flags)
    }

    flag_names <- sub("=.*$", "", flag_values)
    flag_enabled <- rep(TRUE, length(flag_values))
    has_equals <- grepl("=", flag_values, fixed = TRUE)
    flag_enabled[has_equals] <- as.logical(toupper(sub(
        "^.*=",
        "", flag_values[has_equals]
    )))
    if (anyNA(flag_enabled)) {
        stop("epicV2ManifestFlags must contain only TRUE or FALSE values.",
            call. = FALSE
        )
    }

    names(flag_enabled) <- flag_names
    normalizeEpicV2ManifestFlagsMinfiEwasWater(flag_enabled)
}

extractEpicV2ManifestExclusionIdsMinfiEwasWater <- function(
    manifest,
    epicV2ManifestFlags = c(
        CH_WGBS_evidence = TRUE, CH_BLAT = TRUE,
        MissingPos = TRUE, MismatchPos = FALSE
    ), featureNames = NULL
) {
    flags <- normalizeEpicV2ManifestFlagsMinfiEwasWater(epicV2ManifestFlags)
    enabled_flags <- names(flags)[!is.na(flags) & flags]

    if (length(enabled_flags) == 0L) {
        return(structure(
            list(
                ids = character(0), flags = flags,
                flagCounts = stats::setNames(integer(0), character(0)),
                nIds = 0L,
                    overlap = if (is.null(featureNames)) NA_integer_ else 0L
            ),
            class = "dnaEPICO_epicV2Manifest_ids"
        ))
    }

    manifest <- as.data.frame(manifest, stringsAsFactors = FALSE)
    missing_flags <- setdiff(enabled_flags, colnames(manifest))
    if (length(missing_flags) > 0L) {
        stop("EPICv2 manifest is missing requested flag column(s): ",
            paste(missing_flags, collapse = ", "),
            call. = FALSE
        )
    }

    ids <- rownames(manifest)
    default_row_names <- identical(ids, as.character(seq_len(nrow(manifest))))
    if (is.null(ids) || all(!nzchar(ids)) || isTRUE(default_row_names)) {
        if (!("IlmnID" %in% colnames(manifest))) {
            stop("EPICv2 manifest must have IlmnID row names or an IlmnID column.",
                call. = FALSE
            )
        }
        ids <- as.character(manifest$IlmnID)
    }

    selected <- rep(FALSE, nrow(manifest))
    flag_counts <- stats::setNames(
        integer(length(enabled_flags)),
        enabled_flags
    )
    for (flag in enabled_flags) {
        flag_values <- toupper(trimws(as.character(manifest[[flag]])))
        flag_selected <- !is.na(flag_values) & flag_values %in%
            c("Y", "TRUE", "1")
        flag_counts[[flag]] <- sum(flag_selected)
        selected <- selected | flag_selected
    }

    selected_ids <- unique(ids[selected & !is.na(ids) & nzchar(ids)])
    overlap <- NA_integer_
    if (!is.null(featureNames)) {
        overlap <- sum(selected_ids %in% featureNames)
    }

    structure(list(
        ids = selected_ids, flags = flags, flagCounts = flag_counts,
        nIds = length(selected_ids), overlap = overlap
    ), class = "dnaEPICO_epicV2Manifest_ids")
}

readEpicV2ManifestExclusionIdsMinfiEwasWater <- function(
    useEpicV2Manifest = FALSE,
    epicV2ManifestFlags = c(
        CH_WGBS_evidence = TRUE, CH_BLAT = TRUE,
        MissingPos = TRUE, MismatchPos = FALSE
    ), featureNames = NULL
) {
    if (!isTRUE(useEpicV2Manifest)) {
        return(extractEpicV2ManifestExclusionIdsMinfiEwasWater(
            manifest = data.frame(row.names = character(0)),
            epicV2ManifestFlags = stats::setNames(rep(
                FALSE,
                4L
            ), c(
                "CH_WGBS_evidence", "CH_BLAT", "MissingPos",
                "MismatchPos"
            )), featureNames = featureNames
        ))
    }

    if (!requireNamespace("AnnotationHub", quietly = TRUE)) {
        stop("AnnotationHub is required when useEpicV2Manifest = TRUE. ",
            "Install AnnotationHub or set useEpicV2Manifest = FALSE.",
            call. = FALSE
        )
    }

    hub <- AnnotationHub::AnnotationHub()
    manifest <- hub[["AH116484"]]

    extractEpicV2ManifestExclusionIdsMinfiEwasWater(
        manifest = manifest,
        epicV2ManifestFlags = epicV2ManifestFlags, featureNames = featureNames
    )
}

#' Filter probes from a normalized methylation object
#'
#' Apply detection P-value, chromosome, SNP, and probe-exclusion filters to the
#' primary normalized object and return the filtered result.
#'
#' @param normData Object returned by `normalizeMinfiEwasWater()`.
#' @param RGSet Filtered `RGChannelSet` aligned with `normData`.
#' @param pvalThreshold Numeric. Probes must have detection P values below this
#'   threshold in all samples to be retained.
#' @param chrToRemove Character vector or comma-separated string of chromosome
#'   names to remove, for example `'chrX,chrY'`.
#' @param snpsToRemove Character vector or comma-separated string of SNP probe
#'   types to remove, for example `'SBE,CpG'`.
#' @param mafThreshold Numeric. Minor allele frequency threshold passed to
#'   `minfi::dropLociWithSnps()`.
#' @param probeExclusionPath Character vector or semicolon-separated string of
#'   CSV files containing probe IDs to remove.
#' @param probeExclusionIdColumn Character or `NULL`. Column containing probe
#'   IDs. When `NULL` or `''`, each file is auto-detected using `ProbeID`,
#'   `TargetID`, `IlmnID`, or `Name`, then falling back to an unlabeled or
#'   probe-like first column.
#' @param useEpicV2Manifest Logical. If `TRUE`, also remove EPICv2 probes
#'   flagged in the Peters et al. expanded manifest from AnnotationHub resource
#'   `AH116484`.
#' @param epicV2ManifestFlags Named logical vector controlling which EPICv2
#'   manifest flags are removed. Defaults remove `CH_WGBS_evidence`, `CH_BLAT`,
#'   and `MissingPos`, but not `MismatchPos`.
#' @param crossReactivePath Deprecated alias for `probeExclusionPath`.
#' @param crossReactiveIdColumn Deprecated alias for `probeExclusionIdColumn`.
#' @param detPtype Character. Detection P-value mode passed to
#'   `minfi::detectionP()` for the probe filter. Common values in minfi
#'   workflows are `'m+u'` and `'negative'`.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return A list with class `'dnaEPICO_minfiEwasWater_filter'` containing the
#'   filtered object and counts for each filtering stage.
#'
#' @examplesIf requireNamespace("minfiData", quietly = TRUE)
#' ex <- dnaEPICO:::exampleMinfiWorkflowStateDnaEpico()
#' filtered_data <- filterProbesMinfiEwasWater(
#'     normData = ex$normData,
#'     RGSet = ex$sampleData$RGSet,
#'     pvalThreshold = 1,
#'     chrToRemove = "chrY",
#'     snpsToRemove = "SBE",
#'     mafThreshold = 1,
#'     probeExclusionPath = ex$probeExclusionPath,
#'     detPtype = "m+u",
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#' filtered_data$counts[["probeExclusion"]]
#'
#' @export
filterProbesMinfiEwasWater <- function(
    normData, RGSet, pvalThreshold = 0.01,
    chrToRemove = "chrX,chrY", snpsToRemove = "SBE,CpG", mafThreshold = 0.1,
    probeExclusionPath, probeExclusionIdColumn = NULL,
        useEpicV2Manifest = FALSE,
    epicV2ManifestFlags = c(
        CH_WGBS_evidence = TRUE, CH_BLAT = TRUE,
        MissingPos = TRUE, MismatchPos = FALSE
    ), detPtype = "m+u",
    verbose = FALSE, logs = FALSE, log_dir = NULL,
        log_file = "log_filterProbesMinfiEwasWater.txt",
    crossReactivePath = NULL, crossReactiveIdColumn = NULL
) {
    if (!is.null(crossReactivePath)) {
        probeExclusionPath <- crossReactivePath
    }
    if (!is.null(crossReactiveIdColumn)) {
        probeExclusionIdColumn <- crossReactiveIdColumn
    }

    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )

    chr_list <- splitOptionMinfiEwasWater(chrToRemove, sep = ",")
    snp_list <- splitOptionMinfiEwasWater(snpsToRemove, sep = ",")

    pvalThreshold <- validateProbabilityDnaEpico(
        pvalThreshold,
        "pvalThreshold"
    )
    mafThreshold <- validateProbabilityDnaEpico(
        mafThreshold,
        "mafThreshold"
    )
    useEpicV2Manifest <- validateLogicalScalarDnaEpico(
        useEpicV2Manifest,
        "useEpicV2Manifest"
    )

    detP <- minfi::detectionP(RGSet, type = detPtype)
    nan_detp_converted <- sum(is.nan(detP))
    if (nan_detp_converted > 0L) {
        detP[is.nan(detP)] <- NA_real_
    }

    normalized_samples <- colnames(normData$primary)
    rgset_samples <- colnames(RGSet)
    matchSampleIdentifiersDnaEpico(
        query = normalized_samples,
        reference = rgset_samples, queryLabel = "Normalized sample identifiers",
        referenceLabel = "RGSet sample identifiers", requireSameSet = TRUE
    )

    feature_match <- match(
        Biobase::featureNames(normData$primary),
        rownames(detP)
    )
    if (anyNA(feature_match)) {
        stop("Detection P values are missing for ", sum(is.na(feature_match)),
            " normalized probes.",
            call. = FALSE
        )
    }
    detP <- detP[feature_match, normalized_samples, drop = FALSE]

    keep_detp <- rowSums(!is.na(detP) & detP < pvalThreshold) ==
        ncol(detP)
    filtered_detp <- normData$primary[keep_detp, ]

    ann <- minfi::getAnnotation(RGSet)
    remove_probes <- ann$Name[ann$chr %in% chr_list]
    keep_chr <- !(Biobase::featureNames(filtered_detp) %in% remove_probes)
    filtered_chr <- filtered_detp[keep_chr, ]

    filtered_snp <- minfi::dropLociWithSnps(filtered_chr,
        snps = snp_list,
        maf = mafThreshold
    )

    probe_file_ids <- readProbeExclusionIdsMinfiEwasWater(
        probeExclusionPath = probeExclusionPath,
        probeExclusionIdColumn = probeExclusionIdColumn,
            featureNames = Biobase::featureNames(filtered_snp)
    )
    epicv2_manifest_ids <- readEpicV2ManifestExclusionIdsMinfiEwasWater(
        useEpicV2Manifest = useEpicV2Manifest,
        epicV2ManifestFlags = epicV2ManifestFlags,
            featureNames = Biobase::featureNames(filtered_snp)
    )

    probe_exclusion_ids <- unique(c(probe_file_ids$ids,
        epicv2_manifest_ids$ids))
    probe_exclusion_overlap <- NA_integer_
    if (length(probe_exclusion_ids) > 0L) {
        probe_exclusion_overlap <- sum(probe_exclusion_ids %in%
            Biobase::featureNames(filtered_snp))
    }

    if (length(probe_exclusion_ids) > 0L &&
        isTRUE(!is.na(probe_exclusion_overlap)) &&
        identical(probe_exclusion_overlap, 0L)) {
        warning("No probe-exclusion IDs overlap the filtered array feature names. ",
            "Check that probeExclusionPath, probeExclusionIdColumn, and manifest ",
            "settings match the array platform.",
            call. = FALSE
        )
    }

    keep_probe_exclusion <- !(Biobase::featureNames(filtered_snp) %in%
        probe_exclusion_ids)
    filtered_final <- filtered_snp[keep_probe_exclusion, ]

    counts <- c(
        start = nrow(normData$primary), detP = nrow(filtered_detp),
        chromosome = nrow(filtered_chr), snp = nrow(filtered_snp),
        probeExclusion = nrow(filtered_final)
    )
    if (nrow(filtered_final) == 0L) {
        stop("Probe filtering removed every probe. Review the detection P-value, chromosome, SNP, and probe-exclusion settings.",
            call. = FALSE
        )
    }

    probe_file_lines <- if (nrow(probe_file_ids$files) > 0L) {
        apply(probe_file_ids$files, 1L, function(row) {
            paste0(
                "  - ", row[["path"]], " [", row[["column"]],
                "; ", row[["source"]], "; IDs=", row[["nIds"]],
                "; overlap=", row[["overlap"]], "]"
            )
        })
    } else {
        "  - none"
    }
    epicv2_flag_lines <- if (isTRUE(useEpicV2Manifest)) {
        paste0(
            "  - ", names(epicv2_manifest_ids$flagCounts),
            ": ", epicv2_manifest_ids$flagCounts
        )
    } else {
        "  - disabled"
    }

    emitLogMinfiEwasWater(
        c(
            paste(
                "Probe filter threshold:    ",
                pvalThreshold
            ), paste(
                "Detection P NaN to NA:     ",
                nan_detp_converted
            ), paste(
                "Chromosomes removed:       ",
                paste(chr_list, collapse = ", ")
            ), paste(
                "SNP filters removed:       ",
                paste(snp_list, collapse = ", ")
            ), paste(
                "MAF threshold:             ",
                mafThreshold
            ), "Probe-exclusion files:", probe_file_lines,
            paste("Probe-exclusion file IDs: ", probe_file_ids$nIds),
            paste("EPICv2 manifest enabled:  ", useEpicV2Manifest),
            "EPICv2 manifest flag counts:", epicv2_flag_lines, paste(
                "EPICv2 manifest IDs:      ",
                epicv2_manifest_ids$nIds
            ), paste(
                "Probe-exclusion IDs total:",
                length(probe_exclusion_ids)
            ), paste(
                "Probe-exclusion overlap:  ",
                probe_exclusion_overlap
            ), paste(
                "Probes after detP filter:  ",
                counts[["detP"]]
            ), paste(
                "Probes after chr filter:   ",
                counts[["chromosome"]]
            ), paste(
                "Probes after SNP filter:   ",
                counts[["snp"]]
            ), paste(
                "Probes after exclusion:    ",
                counts[["probeExclusion"]]
            ), "============================================================"
        ),
        verbose = verbose, log_path = log_path
    )

    structure(list(
        filtered = filtered_final, detPFiltered = filtered_detp,
        chrFiltered = filtered_chr, snpFiltered = filtered_snp,
        detP = detP, counts = counts, pvalThreshold = pvalThreshold,
        nanDetPConverted = nan_detp_converted, chrToRemove = chr_list,
        snpsToRemove = snp_list, mafThreshold = mafThreshold,
        probeExclusionPath = probeExclusionPath,
            probeExclusionIdColumn = probeExclusionIdColumn,
        probeExclusionIds = probe_exclusion_ids,
            probeExclusionFileIds = probe_file_ids,
        epicV2ManifestIds = epicv2_manifest_ids,
            useEpicV2Manifest = useEpicV2Manifest,
        epicV2ManifestFlags = normalizeEpicV2ManifestFlagsMinfiEwasWater(epicV2ManifestFlags),
        crossReactivePath = probeExclusionPath,
            crossReactiveIdColumn = probeExclusionIdColumn,
        crossReactiveIds = probe_file_ids
    ), class = "dnaEPICO_minfiEwasWater_filter")
}

#' Extract beta, M, and copy-number matrices from a filtered object
#'
#' @param filteredData Object returned by `filterProbesMinfiEwasWater()`.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return A list with class `'dnaEPICO_minfiEwasWater_metrics'` containing
#'   `beta`, `m`, and `cn`.
#'
#' @examples
#' ex <- dnaEPICO:::exampleMinfiMetricsStateDnaEpico()
#' metrics_data <- extractMetricsMinfiEwasWater(
#'     filteredData = ex$filteredData,
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#' names(metrics_data)
#'
#' @export
extractMetricsMinfiEwasWater <- function(
    filteredData, verbose = FALSE,
    logs = FALSE, log_dir = NULL,
        log_file = "log_extractMetricsMinfiEwasWater.txt"
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )

    beta <- minfi::getBeta(filteredData$filtered)
    m <- minfi::getM(filteredData$filtered)
    cn <- minfi::getCN(filteredData$filtered)

    if (is.null(rownames(beta)) || is.null(colnames(beta))) {
        stop("The extracted beta matrix must have probe and sample identifiers.",
            call. = FALSE
        )
    }
    validateMethylationProbeIdentifiersDnaEpico(
        rownames(beta),
        "Extracted methylation probe identifiers"
    )
    validateSampleIdentifiersDnaEpico(colnames(beta),
        "Extracted methylation sample identifiers")
    metrics_match <- vapply(list(m = m, cn = cn), function(metric) {
        identical(dim(metric), dim(beta)) && identical(
            rownames(metric),
            rownames(beta)
        ) && identical(colnames(metric), colnames(beta))
    }, logical(1))
    if (!all(metrics_match)) {
        stop("Extracted Beta, M, and CN matrices must have identical probe and sample order.",
            call. = FALSE
        )
    }

    validations <- list(
        beta = inspectMethylationMatrixDnaEpico(
            beta,
            "beta"
        ), m = inspectMethylationMatrixDnaEpico(m, "m"),
        cn = inspectMethylationMatrixDnaEpico(cn, "cn")
    )
    beta <- validations$beta$values
    m <- validations$m$values
    cn <- validations$cn$values

    converted_nan <- sum(vapply(
        validations, function(validation) validation$boundaries$NaN.Converted,
        numeric(1)
    ))
    invalid_cpgs <- sum(vapply(
        validations, function(validation) validation$boundaries$Invalid.CpGs,
        numeric(1)
    ))
    if (converted_nan > 0L) {
        warning(converted_nan,
            " NaN values returned by methylation accessors were converted to NA.",
            call. = FALSE
        )
    }
    if (invalid_cpgs > 0L) {
        warning(invalid_cpgs,
            " invalid scale-specific CpG records were retained for reporting.",
            call. = FALSE
        )
    }

    preview_cols <- seq_len(min(ncol(beta), 5L))

    emitLogMinfiEwasWater(
        c(
            "Extracting final DNAm matrices (M, Beta, CN).",
            unlist(lapply(validations, function(validation) {
                formatMethylationBoundariesLogDnaEpico(validation$boundaries)
            }), use.names = FALSE), "Preview of beta values:",
                previewLinesMinfiEwasWater(beta[,
                preview_cols,
                drop = FALSE
            ]), "Preview of M-values:",
            previewLinesMinfiEwasWater(m[, preview_cols, drop = FALSE]),
            "Preview of copy-number values:", previewLinesMinfiEwasWater(cn[,
                preview_cols,
                drop = FALSE
            ]), "============================================================"
        ),
        verbose = verbose, log_path = log_path
    )

    structure(list(beta = beta, m = m, cn = cn, validation = validations),
        class = "dnaEPICO_minfiEwasWater_metrics"
    )
}

#' Plot multidimensional scaling or density summaries from final metrics
#'
#' @param metricsData Object returned by `extractMetricsMinfiEwasWater()`.
#' @param targets Filtered phenotype data aligned with `metricsData`.
#' @param plot Character. Plot type: `'mds'` or `'density'`.
#' @param plotGroupVar Character. Phenotype column used for the main grouping.
#' @param sexColumn Character. Phenotype column used for the sex grouping in the
#'   MDS plot.
#' @param display Logical. If `TRUE`, draw the plot on the active graphics
#'   device.
#' @param file Character or `NULL`. TIFF file written when supplied.
#' @param width Integer. TIFF width in pixels when `file` is supplied.
#' @param height Integer. TIFF height in pixels when `file` is supplied.
#' @param res Integer. TIFF resolution in DPI when `file` is supplied.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return Invisibly returns the saved TIFF path when `file` is supplied,
#'   otherwise `NULL`.
#'
#' @examples
#' ex <- dnaEPICO:::exampleMinfiMetricsStateDnaEpico()
#' plotMetricsMinfiEwasWater(
#'     metricsData = ex$metricsData,
#'     targets = ex$targets,
#'     plot = "density",
#'     plotGroupVar = "Sex",
#'     sexColumn = "Sex",
#'     display = FALSE,
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#'
#' @export
plotMetricsMinfiEwasWater <- function(
    metricsData, targets, plot = c(
        "mds",
        "density"
    ), plotGroupVar = "Sex", sexColumn = "Sex", display = FALSE,
    file = NULL, width = 2000L, height = 1000L, res = 150L, verbose = FALSE,
    logs = FALSE, log_dir = NULL, log_file = "log_plotMetricsMinfiEwasWater.txt"
) {
    plot <- match.arg(plot)

    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )

    draw_fun <- switch(plot,
        mds = function() {
            group_factor <- factor(targets[[plotGroupVar]])
            group_sex <- factor(targets[[sexColumn]])
            pal <- RColorBrewer::brewer.pal(8, "Dark2")

            graphics::par(mfrow = c(1, 2))
            limma::plotMDS(metricsData$m,
                main = plotGroupVar, top = 1000,
                gene.selection = "common", col = pal[group_factor],
                dim = c(1, 2)
            )
            graphics::legend("right",
                legend = levels(group_factor),
                text.col = pal, cex = 0.7, bg = "white"
            )

            limma::plotMDS(metricsData$m,
                main = sexColumn, top = 1000,
                gene.selection = "common", col = pal[group_sex],
                dim = c(2, 3)
            )
            graphics::legend("topright",
                legend = levels(group_sex),
                text.col = pal, cex = 0.7, bg = "white"
            )
        },
        density = function() {
            group_factor <- factor(targets[[plotGroupVar]])
            graphics::par(mfrow = c(1, 2))
            minfi::densityPlot(metricsData$beta,
                sampGroups = group_factor,
                main = "Beta values", legend = FALSE, xlab = "Beta values"
            )
            graphics::legend("top",
                legend = levels(group_factor),
                text.col = RColorBrewer::brewer.pal(8, "Dark2")
            )

            minfi::densityPlot(metricsData$m,
                sampGroups = group_factor,
                main = "M-values", legend = FALSE, xlab = "M values"
            )
            graphics::legend("topleft",
                legend = levels(group_factor),
                text.col = RColorBrewer::brewer.pal(8, "Dark2")
            )
        }
    )

    runPlotMinfiEwasWater(
        draw_fun = draw_fun, display = display,
        file = file, width = width, height = height, res = res
    )

    emitLogMinfiEwasWater(
        c(paste(
            "Metrics plot created:     ",
            plot
        ), if (is.null(file)) {
            "Metrics plot file:        none"
        } else {
            paste("Metrics plot file:        ", file)
        }, "============================================================"),
        verbose = verbose, log_path = log_path
    )

    invisible(file)
}

#' Plot ENmix control images from an RGSet
#'
#' Call `ENmix::plotCtrl()` for a supplied `RGSet`. This function only writes
#' files when `output_dir` is provided because `ENmix::plotCtrl()` produces JPG
#' files on disk rather than returning a plot object.
#'
#' @param RGSet An `RGChannelSet`.
#' @param output_dir Character or `NULL`. Directory where ENmix control JPG
#'   files should be written. If `NULL`, the function returns without writing
#'   files.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return Invisibly returns `output_dir`.
#'
#' @examplesIf requireNamespace("minfiData", quietly = TRUE)
#' ex <- dnaEPICO:::exampleMinfiBaseDataDnaEpico()
#' output_dir <- file.path(tempdir(), "enmix-control-plots")
#' plotCtrlMinfiEwasWater(
#'     RGSet = ex$RGSet,
#'     output_dir = output_dir,
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#' dir.exists(output_dir)
#'
#' @export
plotCtrlMinfiEwasWater <- function(
    RGSet, output_dir = NULL,
    verbose = FALSE, logs = FALSE, log_dir = NULL,
        log_file = "log_plotCtrlMinfiEwasWater.txt"
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )

    if (is.null(output_dir)) {
        emitLogMinfiEwasWater(
            c(
                "ENmix control plots skipped because output_dir is NULL.",
                "============================================================"
            ),
            verbose = verbose, log_path = log_path
        )

        return(invisible(NULL))
    }

    dir.create(output_dir, recursive = TRUE, showWarnings = FALSE)

    old_wd <- getwd()
    on.exit(setwd(old_wd), add = TRUE)
    setwd(output_dir)

    old_options <- options(bitmapType = "cairo")
    on.exit(options(old_options), add = TRUE)

    ENmix::plotCtrl(RGSet)

    emitLogMinfiEwasWater(
        c(paste(
            "Generated ENmix control JPGs in:",
            output_dir
        ), "============================================================"),
        verbose = verbose, log_path = log_path
    )

    invisible(output_dir)
}

test_that("probe-exclusion IDs can be read from an unlabeled first column", {
    path <- tempfile(fileext = ".csv")
    writeLines(
        c(
            '"","match47","match48","match49","match50","Total"',
            '"cg00001510",2,0,1,0,3',
            '"cg00002590",5,0,0,0,5',
            '"cg00003969",0,0,2,0,2'
        ),
        con = path
    )

    result <- dnaEPICO:::readProbeExclusionIdsMinfiEwasWater(
        probeExclusionPath = path,
        featureNames = c("cg00001510", "cg99999999")
    )

    expect_identical(result$files$source, "unlabeled first column")
    expect_identical(result$files$column, "<blank>")
    expect_identical(result$ids, c("cg00001510", "cg00002590", "cg00003969"))
    expect_equal(result$overlap, 1)
})

test_that("probe-exclusion IDs can be read from TargetID files", {
    path <- tempfile(fileext = ".csv")
    writeLines(
        c(
            "TargetID,47,48,49,50",
            "cg00001510,2,0,1,0",
            "cg00003969,0,0,2,0",
            "cg00004121,0,4,1,0"
        ),
        con = path
    )

    auto_result <- dnaEPICO:::readProbeExclusionIdsMinfiEwasWater(
        probeExclusionPath = path,
        featureNames = c("cg00003969", "cg00004121")
    )
    configured_result <- dnaEPICO:::readProbeExclusionIdsMinfiEwasWater(
        probeExclusionPath = path,
        probeExclusionIdColumn = "TargetID",
        featureNames = c("cg00003969", "cg00004121")
    )
    null_string_result <- dnaEPICO:::readProbeExclusionIdsMinfiEwasWater(
        probeExclusionPath = path,
        probeExclusionIdColumn = "NULL",
        featureNames = c("cg00003969", "cg00004121")
    )

    expect_identical(auto_result$files$source, "standard")
    expect_identical(auto_result$files$column, "TargetID")
    expect_identical(configured_result$files$column, "TargetID")
    expect_identical(null_string_result$files$source, "standard")
    expect_identical(null_string_result$files$column, "TargetID")
    expect_equal(auto_result$overlap, 2)
    expect_equal(configured_result$overlap, 2)
    expect_equal(null_string_result$overlap, 2)
})

test_that("multiple probe-exclusion files are combined with per-file columns", {
    path1 <- tempfile(fileext = ".csv")
    path2 <- tempfile(fileext = ".csv")
    writeLines(
        c(
            "TargetID,score",
            "cg00001510,1",
            "cg00003969,1"
        ),
        con = path1
    )
    writeLines(
        c(
            "IlmnID,score",
            "cg00003969,1",
            "cg00004121,1"
        ),
        con = path2
    )

    result <- dnaEPICO:::readProbeExclusionIdsMinfiEwasWater(
        probeExclusionPath = paste(path1, path2, sep = ";"),
        probeExclusionIdColumn = "TargetID;IlmnID",
        featureNames = c("cg00001510", "cg00004121", "cg99999999")
    )

    expect_identical(result$ids, c("cg00001510", "cg00003969", "cg00004121"))
    expect_identical(result$files$column, c("TargetID", "IlmnID"))
    expect_equal(result$nIds, 3)
    expect_equal(result$overlap, 2)
})

test_that("probe-exclusion ID column errors are clear", {
    path <- tempfile(fileext = ".csv")
    writeLines(
        c(
            "TargetID,47,48,49,50",
            "cg00001510,2,0,1,0"
        ),
        con = path
    )

    expect_error(
        dnaEPICO:::readProbeExclusionIdsMinfiEwasWater(
            probeExclusionPath = path,
            probeExclusionIdColumn = "ProbeID"
        ),
        "probeExclusionIdColumn not found"
    )
})

test_that("EPICv2 manifest flags select expected exclusion IDs", {
    manifest <- data.frame(
        CH_WGBS_evidence = c("Y", "N", "N", "N"),
        CH_BLAT = c("N", "Y", "N", "N"),
        MissingPos = c("N", "N", "Y", "N"),
        MismatchPos = c("N", "N", "N", "Y"),
        row.names = c("cgWGBS", "cgBLAT", "cgMissing", "cgMismatch"),
        stringsAsFactors = FALSE
    )

    default_result <- dnaEPICO:::extractEpicV2ManifestExclusionIdsMinfiEwasWater(
        manifest = manifest,
        featureNames = c("cgWGBS", "cgMismatch")
    )
    mismatch_result <- dnaEPICO:::extractEpicV2ManifestExclusionIdsMinfiEwasWater(
        manifest = manifest,
        epicV2ManifestFlags = c(
            CH_WGBS_evidence = FALSE,
            CH_BLAT = FALSE,
            MissingPos = FALSE,
            MismatchPos = TRUE
        ),
        featureNames = c("cgWGBS", "cgMismatch")
    )

    expect_identical(default_result$ids, c("cgWGBS", "cgBLAT", "cgMissing"))
    expect_equal(default_result$overlap, 1)
    expect_identical(mismatch_result$ids, "cgMismatch")
    expect_equal(mismatch_result$overlap, 1)
})

test_that("EPICv2 manifest extraction can use an IlmnID column", {
    manifest <- data.frame(
        IlmnID = c("cgWGBS", "cgBLAT"),
        CH_WGBS_evidence = c("Y", "N"),
        CH_BLAT = c("N", "Y"),
        MissingPos = c("N", "N"),
        MismatchPos = c("N", "N"),
        stringsAsFactors = FALSE
    )

    result <- dnaEPICO:::extractEpicV2ManifestExclusionIdsMinfiEwasWater(
        manifest = manifest
    )

    expect_identical(result$ids, c("cgWGBS", "cgBLAT"))
})

#' Estimate surrogate variables from ENmix control probes
#'
#' @param RGSet An `RGChannelSet`.
#' @param ctrlSvaPercVar Numeric. Proportion of variance explained by control
#'   probes, passed to `ENmix::ctrlsva()`.
#' @param ctrlSvaFlag Integer. Control-probe flag passed to
#'   `ENmix::ctrlsva()`.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return A list with class `'dnaEPICO_svaEnmix_sva'` containing the surrogate
#'   variable matrix and the parameters used to estimate it.
#'
#' @description
#' Run `ENmix::ctrlsva()` on an `RGChannelSet` and return the surrogate variable
#' matrix as an in-memory object.
#'
#' @examplesIf requireNamespace("minfiData", quietly = TRUE)
#' ex <- dnaEPICO:::exampleMinfiBaseDataDnaEpico()
#' sva_data <- estimateSvaEnmixControls(
#'     RGSet = ex$RGSet,
#'     ctrlSvaPercVar = 0.5,
#'     ctrlSvaFlag = 1,
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#' sva_data$K
#'
#' @export
estimateSvaEnmixControls <- function(
    RGSet, ctrlSvaPercVar = 0.9,
    ctrlSvaFlag = 1, verbose = FALSE, logs = FALSE, log_dir = NULL,
    log_file = "log_estimateSvaEnmixControls.txt"
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )

    ctrlSvaPercVar <- validateProbabilityDnaEpico(
        ctrlSvaPercVar,
        "ctrlSvaPercVar"
    )
    if (length(ctrlSvaFlag) != 1L || !is.numeric(ctrlSvaFlag) ||
        is.na(ctrlSvaFlag) || !is.finite(ctrlSvaFlag) || !(ctrlSvaFlag %in%
        c(1, 2))) {
        stop("ctrlSvaFlag must be 1 (percvar) or 2 (one requested component).",
            call. = FALSE
        )
    }

    sva <- ENmix::ctrlsva(
        rgSet = RGSet, percvar = ctrlSvaPercVar,
        flag = ctrlSvaFlag
    )
    sva <- as.matrix(sva)
    nan_sva_converted <- sum(is.nan(sva))
    if (nan_sva_converted > 0L) {
        sva[is.nan(sva)] <- NA_real_
    }
    sample_names <- validateSampleIdentifiersDnaEpico(
        svaEnmixGetRGSetSampleNames(RGSet),
        "RGSet sample identifiers"
    )
    if (!is.numeric(sva) || nrow(sva) != length(sample_names) ||
        ncol(sva) == 0L || anyNA(sva) || any(!is.finite(sva)) ||
        is.null(rownames(sva))) {
        if (nan_sva_converted > 0L) {
            stop("ENmix::ctrlsva() must return a finite numeric matrix with one named row per RGSet sample; ",
                nan_sva_converted,
                    " upstream NaN value(s) were converted to NA.",
                call. = FALSE
            )
        }
        stop("ENmix::ctrlsva() must return a finite numeric matrix with one named row per RGSet sample.",
            call. = FALSE
        )
    }
    sample_match <- matchSampleIdentifiersDnaEpico(
        query = sample_names,
        reference = rownames(sva), queryLabel = "RGSet sample identifiers",
        referenceLabel = "ENmix SVA row names", requireSameSet = TRUE
    )
    sva <- sva[sample_match, , drop = FALSE]
    colnames(sva) <- paste0("PC", seq_len(ncol(sva)))

    emitLogMinfiEwasWater(
        c(
            paste(
                "ctrlSva percvar:          ",
                ctrlSvaPercVar
            ), paste(
                "ctrlSva flag:             ",
                ctrlSvaFlag
            ), paste("ctrlSva NaN to NA:        ", nan_sva_converted),
            paste("Number of surrogate variables:", ncol(sva)),
                "Surrogate variables matrix (first few rows):",
            previewLinesMinfiEwasWater(sva),
                "============================================================"
        ),
        verbose = verbose, log_path = log_path
    )

    structure(
        list(
            sva = sva, K = ncol(sva), ctrlSvaPercVar = ctrlSvaPercVar,
            ctrlSvaFlag = ctrlSvaFlag, nanConverted = nan_sva_converted
        ),
        class = "dnaEPICO_svaEnmix_sva"
    )
}

#' Merge surrogate variables into the phenotype table
#'
#' @param targets Phenotype data frame aligned with the samples in `sva`.
#' @param sva Numeric matrix of surrogate variables with samples in rows.
#' @param SampleID Character. Name of the phenotype sample identifier column.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return A phenotype data frame with the surrogate variables appended.
#'
#' @description
#' Merge the surrogate variable matrix back into the phenotype table while
#' preserving the original row order of `targets`.
#'
#' @examples
#' ex <- dnaEPICO:::exampleSvaAnalysisStateDnaEpico()
#' merged_pheno <- mergeSvaTargetsEnmix(
#'     targets = ex$targets,
#'     sva = ex$sva,
#'     SampleID = "Sample_Name",
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#' colnames(merged_pheno)[seq_len(4)]
#'
#' @export
mergeSvaTargetsEnmix <- function(
    targets, sva, SampleID = "Sample_Name",
    verbose = FALSE, logs = FALSE, log_dir = NULL,
        log_file = "log_mergeSvaTargetsEnmix.txt"
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )

    if (!(SampleID %in% colnames(targets))) {
        stop("SampleID column not found in targets: ", SampleID,
            call. = FALSE
        )
    }

    if (!is.matrix(sva) || !is.numeric(sva) || ncol(sva) == 0L) {
        stop("sva must be a numeric matrix with at least one column.",
            call. = FALSE
        )
    }
    if (is.null(rownames(sva))) {
        stop("sva must have row names that match the phenotype SampleID column.",
            call. = FALSE
        )
    }
    if (anyNA(sva) || any(!is.finite(sva))) {
        stop("sva contains missing or non-finite values.", call. = FALSE)
    }

    sva_data <- data.frame(
        sample_id = rownames(sva), sva, row.names = NULL,
        check.names = FALSE, stringsAsFactors = FALSE
    )
    colnames(sva_data)[1] <- SampleID
    surrogate_columns <- setdiff(colnames(sva_data), SampleID)
    duplicate_columns <- intersect(colnames(targets), surrogate_columns)
    if (length(duplicate_columns) > 0L) {
        stop("The phenotype table already contains output columns that would be duplicated: ",
            paste(duplicate_columns, collapse = ", "),
                ". Start from the pre-SVA phenotype file or remove the existing columns explicitly.",
            call. = FALSE
        )
    }

    match_idx <- matchSampleIdentifiersDnaEpico(
        query = targets[[SampleID]],
        reference = sva_data[[SampleID]], queryLabel = paste0(
            "Phenotype column '",
            SampleID, "'"
        ), referenceLabel = "SVA row names",
        requireSameSet = TRUE
    )
    merged_pheno <- cbind(targets, sva_data[match_idx, surrogate_columns,
        drop = FALSE
    ])

    emitLogMinfiEwasWater(
        c(paste(
            "Merged phenotype rows:     ",
            nrow(merged_pheno)
        ), paste(
            "Merged phenotype columns:  ",
            ncol(merged_pheno)
        ), "============================================================"),
        verbose = verbose, log_path = log_path
    )

    merged_pheno
}

#' Analyze surrogate variables against Sentrix chip and position factors
#'
#' @param sva Numeric matrix of surrogate variables with samples in rows.
#' @param RGSet An `RGChannelSet` aligned with `sva`.
#' @param SentrixIDColumn Character. Name of the chip identifier column in
#'   `SummarizedExperiment::colData(RGSet)`.
#' @param SentrixPositionColumn Character. Name of the chip position column in
#'   `SummarizedExperiment::colData(RGSet)`.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return A list with class `'dnaEPICO_svaEnmix_analysis'` containing the
#'   aligned Sentrix factors, full and reduced linear models, and ANOVA tables.
#'
#' @description
#' Fit linear models for each surrogate variable against informative Sentrix
#' chip and position factors, perform backward elimination with
#' `stats::drop1()`, and return the in-memory analysis objects. A constant
#' technical factor is omitted; if both are constant, an intercept-only model
#' is retained.
#'
#' @examples
#' ex <- dnaEPICO:::exampleSvaAnalysisStateDnaEpico()
#' analysis_data <- analyzeSvaEnmix(
#'     sva = ex$sva,
#'     RGSet = ex$RGSet,
#'     SentrixIDColumn = "Sentrix_ID",
#'     SentrixPositionColumn = "Sentrix_Position",
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#' analysis_data$K
#'
#' @export
analyzeSvaEnmix <- function(
    sva, RGSet, SentrixIDColumn = "Sentrix_ID",
    SentrixPositionColumn = "Sentrix_Position", verbose = FALSE,
    logs = FALSE, log_dir = NULL, log_file = "log_analyzeSvaEnmix.txt"
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )

    sample_names <- svaEnmixGetRGSetSampleNames(RGSet)

    if (length(sample_names) == 0L) {
        stop("Could not determine sample names from the loaded RGSet.",
            call. = FALSE
        )
    }

    if (!is.matrix(sva) || !is.numeric(sva) || ncol(sva) == 0L) {
        stop("sva must be a numeric matrix with at least one surrogate variable.",
            call. = FALSE
        )
    }
    if (anyNA(sva) || any(!is.finite(sva))) {
        stop("sva contains missing or non-finite values.", call. = FALSE)
    }
    if (is.null(rownames(sva))) {
        stop("sva must have sample identifiers in its row names.",
            call. = FALSE
        )
    }

    match_idx <- matchSampleIdentifiersDnaEpico(
        query = sample_names,
        reference = rownames(sva), queryLabel = "RGSet sample identifiers",
        referenceLabel = "SVA row names", requireSameSet = TRUE
    )

    sva <- sva[match_idx, , drop = FALSE]
    col_data <- svaEnmixGetRGSetColData(RGSet)
    missing_sentrix <- setdiff(
        c(SentrixIDColumn, SentrixPositionColumn),
        colnames(col_data)
    )
    if (length(missing_sentrix) > 0L) {
        stop("RGSet column data is missing: ", paste(missing_sentrix,
            collapse = ", "
        ), call. = FALSE)
    }
    sentrix_id <- as.factor(col_data[[SentrixIDColumn]])
    sentrix_position <- as.factor(col_data[[SentrixPositionColumn]])
    if (anyNA(sentrix_id) || anyNA(sentrix_position)) {
        stop("Sentrix chip and position columns cannot contain missing values.",
            call. = FALSE
        )
    }
    K <- ncol(sva)

    technical_data <- data.frame(SentrixID = sentrix_id,
        SentrixPosition = sentrix_position)
    technical_terms <- colnames(technical_data)[vapply(
        technical_data,
        function(values) nlevels(droplevels(values)) > 1L, logical(1)
    )]
    technical_formula <- if (length(technical_terms) == 0L) {
        stats::as.formula("sva_value ~ 1")
    } else {
        stats::reformulate(termlabels = technical_terms, response = "sva_value")
    }

    full_models <- lapply(seq_len(K), function(i) {
        model_data <- technical_data
        model_data$sva_value <- sva[, i]
        stats::lm(technical_formula, data = model_data)
    })

    reduced_models <- vector("list", K)
    dropterm_steps <- vector("list", K)
    warning_state <- new.env(parent = emptyenv())
    warning_state$messages <- character(0)
    capture_analysis_warnings <- function(expression) {
        withCallingHandlers(expression, warning = function(condition) {
            warning_state$messages <- unique(c(
                warning_state$messages,
                conditionMessage(condition)
            ))
            invokeRestart("muffleWarning")
        })
    }
    for (i in seq_len(K)) {
        model_tmp <- full_models[[i]]
        model_steps <- list()

        repeat {
            drop_table <- capture_analysis_warnings(stats::drop1(model_tmp,
                test = "F"
            ))
            drop_p <- drop_table$`Pr(F)`
            valid_idx <- which(!is.na(drop_p))

            if (length(valid_idx) == 0L) {
                break
            }

            max_idx <- valid_idx[which.max(drop_p[valid_idx])]
            max_p <- drop_p[[max_idx]]
            term_to_drop <- rownames(drop_table)[[max_idx]]

            if (!is.finite(max_p) || max_p <= 0.05 || identical(
                term_to_drop,
                "<none>"
            )) {
                break
            }

            model_tmp <- stats::update(model_tmp, paste(
                ". ~ . -",
                term_to_drop
            ))
            model_steps[[length(model_steps) + 1L]] <- list(
                dropterm = drop_table,
                summary = summary(model_tmp)
            )
        }

        reduced_models[[i]] <- model_tmp
        dropterm_steps[[i]] <- model_steps
    }

    anova_full <- lapply(full_models,
        function(model) capture_analysis_warnings(stats::anova(model)))
    anova_reduced <- lapply(reduced_models,
        function(model) capture_analysis_warnings(stats::anova(model)))
    analysis_warnings <- warning_state$messages

    emitLogMinfiEwasWater(
        c(
            paste(
                "Number of surrogate variables (K):",
                K
            ), paste("SentrixID class:            ", class(sentrix_id)[1]),
            paste("SentrixID unique levels:    ", length(unique(sentrix_id))),
            previewLinesMinfiEwasWater(table(sentrix_id)), paste(
                "SentrixPosition class:      ",
                class(sentrix_position)[1]
            ), paste(
                "SentrixPosition unique levels:",
                length(unique(sentrix_position))
            ), previewLinesMinfiEwasWater(table(sentrix_position)),
            paste("Technical terms modelled:     ",
                if (length(technical_terms)) {
                paste(technical_terms,
                    collapse = ", "
                )
            } else {
                "intercept only"
            }), "First row of SVA matrix:",
            previewLinesMinfiEwasWater(sva[1, , drop = FALSE]), paste(
                "Sample names in SVA matrix:",
                paste(rownames(sva)[seq_len(min(5L, nrow(sva)))],
                    collapse = ", "
                )
            ), paste(
                "Sample names in colData(RGSet):",
                paste(sample_names[seq_len(min(5L, length(sample_names)))],
                    collapse = ", "
                )
            ), paste(
                "Model-selection warnings:    ",
                length(analysis_warnings)
            ), if (length(analysis_warnings) >
                0L) {
                analysis_warnings
            } else {
                character(0)
            }, "============================================================"
        ),
        verbose = verbose, log_path = log_path
    )

    structure(
        list(
            sva = sva, K = K, sentrixID = sentrix_id,
            sentrixPosition = sentrix_position,
                technicalTerms = technical_terms,
            fullModels = full_models, reducedModels = reduced_models,
            droptermSteps = dropterm_steps, anovaFull = anova_full,
            anovaReduced = anova_reduced, warnings = analysis_warnings
        ),
        class = "dnaEPICO_svaEnmix_analysis"
    )
}

#' @keywords internal
#' @noRd
svaEnmixPositiveNumber <- function(value, fallback) {
    if (!is.numeric(value) || length(value) == 0L) {
        return(fallback)
    }

    value <- value[1]
    if (!is.finite(value) || value <= 0) {
        return(fallback)
    }

    value
}

#' @keywords internal
#' @noRd
svaEnmixSplitIndices <- function(K, page_limit = 6L) {
    K <- as.integer(svaEnmixPositiveNumber(K, 1))
    page_limit <- as.integer(svaEnmixPositiveNumber(
        page_limit,
        6
    ))
    indices <- seq_len(K)
    split(indices, ceiling(indices / page_limit))
}

#' @keywords internal
#' @noRd
svaEnmixMatrixPages <- function(K, page_limit = 6L) {
    row_blocks <- svaEnmixSplitIndices(K, page_limit)
    col_blocks <- svaEnmixSplitIndices(K, page_limit)
    pages <- vector("list", length(row_blocks) * length(col_blocks))
    page <- 1L

    for (row_block in row_blocks) {
        for (col_block in col_blocks) {
            pages[[page]] <- list(
                page = page, rowIndices = row_block,
                colIndices = col_block
            )
            page <- page + 1L
        }
    }

    pages
}

#' @keywords internal
#' @noRd
svaEnmixMatrixPlotDimensions <- function(
    width, height, res,
    panel_count
) {
    fallback_width <- 2000
    fallback_height <- 1000
    width <- svaEnmixPositiveNumber(width, fallback_width)
    height <- svaEnmixPositiveNumber(height, fallback_height)
    res <- svaEnmixPositiveNumber(res, NA_real_)
    panel_count <- svaEnmixPositiveNumber(panel_count, NA_real_)

    if (!is.finite(res) || res <= 0 || !is.finite(panel_count) ||
        panel_count <= 0L) {
        return(c(width = as.integer(ceiling(width)),
            height = as.integer(ceiling(height))))
    }

    minimum_pixels <- ceiling(round(((panel_count * 1.6) + 1) *
        res, digits = 6L))
    c(width = as.integer(max(width, minimum_pixels)), height = as.integer(max(
        height,
        minimum_pixels
    )))
}

#' @keywords internal
#' @noRd
svaEnmixMatrixScale <- function(K, multiplier, minimum, maximum) {
    K <- svaEnmixPositiveNumber(K, 1)

    max(minimum, min(maximum, multiplier / K))
}

#' @keywords internal
#' @noRd
svaEnmixMatrixPanelMargin <- function(K) {
    if (K <= 5L) {
        return(c(2.4, 2.4, 0.5, 0.2))
    }
    if (K <= 8L) {
        return(c(1.9, 1.9, 0.3, 0.1))
    }

    c(1.5, 1.5, 0.2, 0.1)
}

#' @keywords internal
#' @noRd
svaEnmixMatrixPlotState <- function(
    analysisData, width, height,
    res
) {
    sentrix_id_levels <- levels(analysisData$sentrixID)
    sentrix_id_legend_limit <- 20L
    page_panel_limit <- 6L
    pages <- svaEnmixMatrixPages(analysisData$K, page_panel_limit)
    page_panel_count <- min(analysisData$K, page_panel_limit)
    dimensions <- svaEnmixMatrixPlotDimensions(
        width = width,
        height = height, res = res, panel_count = page_panel_count
    )

    list(
        width = dimensions[["width"]], height = dimensions[["height"]],
        pages = pages, pageCount = length(pages),
            pagePanelLimit = page_panel_limit,
        sentrixIDLevels = sentrix_id_levels,
            sentrixIDLegendLimit = sentrix_id_legend_limit,
        showSentrixIDLegend = length(sentrix_id_levels) <=
            sentrix_id_legend_limit
    )
}

#' @keywords internal
#' @noRd
svaEnmixLogMatrixPlotState <- function(state, res, verbose, log_path) {
    log_lines <- c(paste(
        "SVA matrix plot dimensions:", state$width,
        "x", state$height, "@", res
    ))

    if (state$pageCount > 1L) {
        log_lines <- c(log_lines, paste(
            "SVA matrix plot pages:",
            state$pageCount, "pages with up to", state$pagePanelLimit,
            "x", state$pagePanelLimit, "panels per page"
        ))
    }

    if (!isTRUE(state$showSentrixIDLegend)) {
        log_lines <- c(log_lines, paste(
            "SentrixID legend suppressed for matrix plot:",
            length(state$sentrixIDLevels), "levels exceed limit of",
            state$sentrixIDLegendLimit
        ))
    }

    emitLogMinfiEwasWater(log_lines, verbose = verbose, log_path = log_path)
}

#' @keywords internal
#' @noRd
svaEnmixMatrixPageFile <- function(file, page, page_count) {
    if (is.null(file) || page_count <= 1L) {
        return(file)
    }

    extension <- tools::file_ext(file)
    stem <- tools::file_path_sans_ext(file)
    if (!nzchar(extension)) {
        return(sprintf("%s_page%02d", stem, page))
    }

    sprintf("%s_page%02d.%s", stem, page, extension)
}

#' @keywords internal
#' @noRd
svaEnmixPlotSentrixID <- function(analysisData) {
    color_map <- grDevices::rainbow(length(levels(analysisData$sentrixID)))
    sva <- analysisData$sva
    graphics::plot(sva[, 1], sva[, 2],
        col = color_map[analysisData$sentrixID],
        pch = 16, xlab = "Surrogate Variable 1 (PC1)",
            ylab = "Surrogate Variable 2 (PC2)",
        main = "Surrogate Variables Colored by Chip (SentrixID)"
    )
    graphics::legend("topright",
        legend = levels(analysisData$sentrixID),
        col = color_map, pch = 16, title = "SentrixID", cex = 0.6
    )
}

#' @keywords internal
#' @noRd
svaEnmixPlotSentrixPosition <- function(analysisData) {
    color_map <-
        grDevices::rainbow(length(levels(analysisData$sentrixPosition)))
    sva <- analysisData$sva
    graphics::plot(sva[, 1], sva[, 2],
        col = color_map[analysisData$sentrixPosition],
        pch = 16, xlab = "Surrogate Variable 1 (PC1)",
            ylab = "Surrogate Variable 2 (PC2)",
        main = "Surrogate Variables Colored by Sentrix Position"
    )
    graphics::legend("topright",
        legend = levels(analysisData$sentrixPosition),
        col = color_map, pch = 16, title = "SentrixPosition",
        cex = 0.6
    )
}

#' @keywords internal
#' @noRd
svaEnmixMatrixPlotScales <- function(K) {
    list(axis = svaEnmixMatrixScale(K,
        multiplier = 3.4, minimum = 0.35,
        maximum = 0.7
    ), label = svaEnmixMatrixScale(K,
        multiplier = 3.8,
        minimum = 0.45, maximum = 0.8
    ), point = svaEnmixMatrixScale(K,
        multiplier = 4, minimum = 0.45, maximum = 0.8
    ), legend = svaEnmixMatrixScale(K,
        multiplier = 3.2, minimum = 0.35, maximum = 0.6
    ))
}

#' @keywords internal
#' @noRd
svaEnmixPlotMatrixLegends <- function(
    state, sentrix_position_levels,
    color_map, pch_map, legend_cex
) {
    if (isTRUE(state$showSentrixIDLegend)) {
        graphics::legend("topright",
            legend = state$sentrixIDLevels,
            col = color_map, pch = 15, title = "SentrixID", cex = legend_cex,
            bty = "n"
        )
    }

    graphics::legend(
        if (isTRUE(state$showSentrixIDLegend)) {
            "bottomright"
        } else {
            "topright"
        },
        legend = sentrix_position_levels, pch = pch_map,
        title = "SentrixPosition", cex = legend_cex, bty = "n"
    )
}

#' @keywords internal
#' @noRd
svaEnmixPlotMatrixPageTitle <- function(page, page_count) {
    title <- "Effects of Sentrix ID (color) and Sentrix Position (shape)"
    if (page_count <= 1L) {
        return(title)
    }

    paste(title, sprintf(
        "(page %d of %d: SV%d-SV%d by SV%d-SV%d)",
        page$page, page_count, min(page$rowIndices), max(page$rowIndices),
        min(page$colIndices), max(page$colIndices)
    ))
}

#' @keywords internal
#' @noRd
svaEnmixPlotMatrixPage <- function(analysisData, state, page) {
    sva <- analysisData$sva
    row_indices <- page$rowIndices
    col_indices <- page$colIndices
    page_panel_count <- max(length(row_indices), length(col_indices))
    sentrix_position_levels <- levels(analysisData$sentrixPosition)
    color_map <- grDevices::rainbow(length(state$sentrixIDLevels))
    pch_map <- seq_along(sentrix_position_levels)
    scales <- svaEnmixMatrixPlotScales(page_panel_count)

    old_par <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(old_par), add = TRUE)
    graphics::par(
        mfrow = c(length(row_indices), length(col_indices)),
        family = "Times", las = 1,
            mar = svaEnmixMatrixPanelMargin(page_panel_count),
        oma = c(3.2, 3.2, 2, 0.5), mgp = c(1.1, 0.35, 0), tcl = -0.18,
        cex.axis = scales$axis, cex.lab = scales$label
    )

    for (row_pos in seq_along(row_indices)) {
        for (col_pos in seq_along(col_indices)) {
            row_index <- row_indices[[row_pos]]
            col_index <- col_indices[[col_pos]]
            graphics::plot(sva[, col_index], sva[, row_index],
                col = color_map[analysisData$sentrixID],
                    pch = pch_map[analysisData$sentrixPosition],
                cex = scales$point, xlab = if (row_pos == length(row_indices)) {
                    paste("SV", col_index)
                } else {
                    ""
                }, ylab = if (col_pos == 1L) {
                    paste("SV", row_index)
                } else {
                    ""
                }, main = "", xaxt = if (row_pos == length(row_indices)) {
                    "s"
                } else {
                    "n"
                }, yaxt = if (col_pos == 1L) {
                    "s"
                } else {
                    "n"
                }
            )
            if (row_pos == 1L && col_pos == 1L) {
                svaEnmixPlotMatrixLegends(
                    state, sentrix_position_levels,
                    color_map, pch_map, scales$legend
                )
            }
        }
    }

    graphics::mtext(svaEnmixPlotMatrixPageTitle(page, state$pageCount),
        side = 3, outer = TRUE, line = 0.5, cex = 0.9
    )
}

#' @keywords internal
#' @noRd
svaEnmixSaveMatrixPage <- function(
    file, draw_fun, width, height,
    res
) {
    dir.create(dirname(file), recursive = TRUE, showWarnings = FALSE)
    grDevices::tiff(
        filename = file, width = width, height = height,
        res = res, type = "cairo"
    )
    tryCatch(draw_fun(), finally = grDevices::dev.off())

    invisible(file)
}

#' @keywords internal
#' @noRd
svaEnmixRunMatrixPlot <- function(
    analysisData, state, display,
    file, res
) {
    if (is.null(file) && !isTRUE(display)) {
        return(invisible(NULL))
    }

    output_files <- NULL
    if (!is.null(file)) {
        output_files <- vapply(state$pages, function(page) {
            svaEnmixMatrixPageFile(
                file,
                page$page, state$pageCount
            )
        }, character(1))
        output_files <- unname(output_files)
    }

    if (!is.null(file)) {
        for (page in state$pages) {
            page_file <- svaEnmixMatrixPageFile(
                file, page$page,
                state$pageCount
            )
            svaEnmixSaveMatrixPage(
                file = page_file, draw_fun = function() {
                    svaEnmixPlotMatrixPage(
                        analysisData,
                        state, page
                    )
                }, width = state$width, height = state$height,
                res = res
            )
        }
    }

    if (isTRUE(display)) {
        for (page in state$pages) {
            svaEnmixPlotMatrixPage(analysisData, state, page)
        }
    }

    invisible(if (is.null(file)) NULL else output_files)
}

#' @keywords internal
#' @noRd
svaEnmixSavedPlotLogLines <- function(output) {
    if (is.null(output)) {
        return("Saved plot path:           not saved")
    }
    if (length(output) <= 1L) {
        return(paste("Saved plot path:          ", output))
    }

    c(
        paste("Saved plot paths:         ", length(output), "files"),
        paste("  ", output)
    )
}

#' Plot surrogate variables for svaEnmix
#'
#' @param analysisData Object returned by `analyzeSvaEnmix()`.
#' @param plot Character. Plot type: `'sentrix_id'`, `'sentrix_position'`, or
#'   `'matrix'`.
#' @param display Logical. If `TRUE`, draw the plot on the active graphics
#'   device.
#' @param file Character or `NULL`. TIFF file path used for saved output.
#' @param width Integer. Plot width in pixels when `file` is supplied.
#' @param height Integer. Plot height in pixels when `file` is supplied.
#' @param res Integer. TIFF resolution in DPI when `file` is supplied.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return Invisibly returns saved TIFF path(s), otherwise `NULL`.
#'
#' @description
#' Draw one of the standard surrogate-variable plots used by `svaEnmix()`.
#'
#' @examples
#' ex <- dnaEPICO:::exampleSvaAnalysisStateDnaEpico()
#' plotSvaEnmix(
#'     analysisData = ex$analysisData,
#'     plot = "sentrix_id",
#'     display = FALSE,
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#'
#' @export
plotSvaEnmix <- function(
    analysisData, plot = c(
        "sentrix_id",
        "sentrix_position", "matrix"
    ), display = FALSE, file = NULL,
    width = 2000L, height = 1000L, res = 150L, verbose = FALSE,
    logs = FALSE, log_dir = NULL, log_file = "log_plotSvaEnmix.txt"
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )
    plot <- match.arg(plot)
    sva <- analysisData$sva
    matrix_state <- list(width = width, height = height)
    will_draw <- !is.null(file) || isTRUE(display)

    if (plot %in% c("sentrix_id", "sentrix_position") && ncol(sva) <
        2L) {
        stop("At least two surrogate variables are required for this plot.",
            call. = FALSE
        )
    }

    if (identical(plot, "matrix")) {
        matrix_state <- svaEnmixMatrixPlotState(
            analysisData,
            width, height, res
        )
        if (will_draw) {
            svaEnmixLogMatrixPlotState(
                matrix_state, res, verbose,
                log_path
            )
        }
    }

    draw_fun <- switch(plot,
        sentrix_id = function() svaEnmixPlotSentrixID(analysisData),
        sentrix_position = function() svaEnmixPlotSentrixPosition(analysisData),
        matrix = NULL
    )

    if (identical(plot, "matrix")) {
        output <- svaEnmixRunMatrixPlot(
            analysisData = analysisData,
            state = matrix_state, display = display, file = file,
            res = res
        )
    } else {
        output <- runPlotMinfiEwasWater(
            draw_fun = draw_fun,
            display = display, file = file, width = matrix_state$width,
            height = matrix_state$height, res = res
        )
    }

    emitLogMinfiEwasWater(
        c(
            paste(
                "SVA plot type:            ",
                plot
            ), paste("Display plot:             ", display),
            svaEnmixSavedPlotLogLines(output),
                "============================================================"
        ),
        verbose = verbose, log_path = log_path
    )

    invisible(output)
}

#' @keywords internal
#' @noRd
readPhenotypeFileSvaEnmix <- function(file, sepType = NULL) {
    separator <- resolveSeparatorMinfiEwasWater(sepType)
    if (is.null(separator)) {
        return(utils::read.csv(file,
            stringsAsFactors = FALSE,
            check.names = FALSE
        ))
    }
    utils::read.table(file,
        header = TRUE, sep = separator, stringsAsFactors = FALSE,
        check.names = FALSE, quote = "\"", comment.char = ""
    )
}

#' @keywords internal
#' @noRd
validatePhenotypeReplacementSvaEnmix <- function(
    original, updated,
    SampleID, pcColumns
) {
    if (!is.data.frame(original) || !is.data.frame(updated)) {
        stop("The original and updated phenotype objects must be data frames.",
            call. = FALSE
        )
    }
    if (!(SampleID %in% colnames(original)) || !(SampleID %in%
        colnames(updated))) {
        stop("SampleID column is missing from the phenotype replacement: ",
            SampleID,
            call. = FALSE
        )
    }
    if (nrow(original) != nrow(updated)) {
        stop("The updated phenotype has ", nrow(updated), " rows; expected ",
            nrow(original), ". The original file was not changed.",
            call. = FALSE
        )
    }
    original_ids <- validateSampleIdentifiersDnaEpico(
        original[[SampleID]],
        paste0("Original phenotype column '", SampleID, "'")
    )
    updated_ids <- validateSampleIdentifiersDnaEpico(
        updated[[SampleID]],
        paste0("Updated phenotype column '", SampleID, "'")
    )
    if (!identical(original_ids, updated_ids)) {
        stop("Sample identifiers or their order changed; the original file was not changed.",
            call. = FALSE
        )
    }
    if (anyDuplicated(colnames(updated))) {
        stop("The updated phenotype contains duplicated column names.",
            call. = FALSE
        )
    }
    missing_original <- setdiff(colnames(original), colnames(updated))
    if (length(missing_original) > 0L) {
        stop("The updated phenotype is missing original column(s): ",
            paste(missing_original, collapse = ", "),
            call. = FALSE
        )
    }
    for (column in colnames(original)) {
        if (!isTRUE(all.equal(original[[column]], updated[[column]],
            check.attributes = FALSE
        ))) {
            stop("Original phenotype column changed: ", column,
                call. = FALSE
            )
        }
    }
    if (length(pcColumns) == 0L || anyNA(pcColumns) ||
        any(!nzchar(pcColumns))) {
        stop("The SVA output does not contain named PC columns.",
            call. = FALSE
        )
    }
    if (anyDuplicated(pcColumns) || !all(grepl(
        "^PC[1-9][0-9]*$",
        pcColumns
    ))) {
        stop("SVA columns must have unique names PC1, PC2, and so on.",
            call. = FALSE
        )
    }
    missing_pc <- setdiff(pcColumns, colnames(updated))
    if (length(missing_pc) > 0L) {
        stop("The updated phenotype is missing PC column(s): ",
            paste(missing_pc, collapse = ", "),
            call. = FALSE
        )
    }
    if (any(pcColumns %in% colnames(original))) {
        stop("PC column names already exist in the original phenotype file.",
            call. = FALSE
        )
    }
    pc_numeric <- vapply(updated[pcColumns], is.numeric, logical(1))
    if (!all(pc_numeric)) {
        stop("All appended PC columns must be numeric.", call. = FALSE)
    }
    if (!all(is.finite(as.matrix(updated[pcColumns])))) {
        stop("Appended PC columns must contain only finite values.",
            call. = FALSE
        )
    }
    invisible(TRUE)
}

#' @keywords internal
#' @noRd
writePhenotypeFileSvaEnmix <- function(data, file, sepType = NULL) {
    separator <- resolveSeparatorMinfiEwasWater(sepType)
    if (is.null(separator)) {
        utils::write.csv(data, file = file, row.names = FALSE)
    } else {
        utils::write.table(data,
            file = file, sep = separator,
            row.names = FALSE, col.names = TRUE, quote = TRUE,
            na = "NA"
        )
    }
    invisible(file)
}

#' @keywords internal
#' @noRd
replacePhenotypeFileSvaEnmix <- function(
    mergedPheno, phenoFile,
    SampleID, pcColumns, sepType = NULL
) {
    if (!file.exists(phenoFile)) {
        stop("Phenotype file does not exist: ", phenoFile, call. = FALSE)
    }
    original <- readPhenotypeFileSvaEnmix(phenoFile, sepType = sepType)
    validatePhenotypeReplacementSvaEnmix(
        original = original,
        updated = mergedPheno, SampleID = SampleID, pcColumns = pcColumns
    )

    output_dir <- dirname(phenoFile)
    temp_file <- tempfile(pattern = paste0(
        ".", basename(phenoFile),
        ".sva-"
    ), tmpdir = output_dir, fileext = ".tmp")
    backup_file <- tempfile(pattern = paste0(
        ".", basename(phenoFile),
        ".backup-"
    ), tmpdir = output_dir, fileext = ".tmp")
    backup_active <- FALSE
    replacement_complete <- FALSE
    on.exit(
        {
            if (file.exists(temp_file)) unlink(temp_file)
            if (isTRUE(backup_active) && file.exists(backup_file)) {
                if (file.exists(phenoFile)) unlink(phenoFile)
                file.rename(backup_file, phenoFile)
            }
            if (isTRUE(replacement_complete) &&
                file.exists(backup_file)) unlink(backup_file)
        },
        add = TRUE
    )

    writePhenotypeFileSvaEnmix(mergedPheno, temp_file, sepType = sepType)
    written <- readPhenotypeFileSvaEnmix(temp_file, sepType = sepType)
    validatePhenotypeReplacementSvaEnmix(
        original = original,
        updated = written, SampleID = SampleID, pcColumns = pcColumns
    )

    if (!file.rename(phenoFile, backup_file)) {
        stop("Could not create a rollback copy of the phenotype file.",
            call. = FALSE
        )
    }
    backup_active <- TRUE
    if (!file.rename(temp_file, phenoFile)) {
        restored <- file.rename(backup_file, phenoFile)
        backup_active <- FALSE
        if (!isTRUE(restored)) {
            stop("Phenotype replacement failed and the original could not be restored from: ",
                backup_file,
                call. = FALSE
            )
        }
        stop("Phenotype replacement failed; the original file was restored.",
            call. = FALSE
        )
    }
    final_validation <- tryCatch(
        {
            final <- readPhenotypeFileSvaEnmix(phenoFile, sepType = sepType)
            validatePhenotypeReplacementSvaEnmix(
                original = original,
                updated = final, SampleID = SampleID, pcColumns = pcColumns
            )
            NULL
        },
        error = identity
    )
    if (inherits(final_validation, "error")) {
        unlink(phenoFile)
        restored <- file.rename(backup_file, phenoFile)
        backup_active <- FALSE
        if (!isTRUE(restored)) {
            stop("Final phenotype validation failed and the original could not be restored from: ",
                backup_file,
                call. = FALSE
            )
        }
        stop("Final phenotype validation failed; the original file was restored: ",
            conditionMessage(final_validation),
            call. = FALSE
        )
    }
    if (file.exists(backup_file) && unlink(backup_file) != 0L) {
        stop("The phenotype file was replaced, but its temporary backup could not be removed.",
            call. = FALSE
        )
    }
    backup_active <- FALSE
    replacement_complete <- TRUE
    invisible(phenoFile)
}

#' Write svaEnmix outputs to disk
#'
#' @param svaData Object returned by `estimateSvaEnmixControls()`.
#' @param mergedPheno Phenotype data frame returned by `mergeSvaTargetsEnmix()`.
#' @param analysisData Optional object returned by `analyzeSvaEnmix()`.
#' @param phenoFile Character or `NULL`. Existing phenotype file replaced with
#'   `mergedPheno` after validation.
#' @param SampleID Character. Phenotype column containing sample identifiers.
#' @param sepType Character or `NULL`. Field separator used by `phenoFile`.
#' @param dataBaseDir Character. Base directory used for saved data outputs.
#' @param rBaseDir Character. Base directory used for saved `.RData` outputs.
#' @param scriptLabel Character. Label used to create the output subdirectory.
#' @param verbose Logical. If `TRUE`, emit progress messages with `message()`.
#' @param logs Logical. If `TRUE`, write the same messages to a log file.
#' @param log_dir Character or `NULL`. Directory used for the log file when
#'   `logs = TRUE`.
#' @param log_file Character. File name used when `logs = TRUE`.
#'
#' @return A list with class `'dnaEPICO_svaEnmix_paths'` containing the paths
#'   written to disk.
#'
#' @description
#' Write the SVA matrix, phenotype, and model-summary outputs.
#'
#' @examples
#' ex <- dnaEPICO:::exampleSvaAnalysisStateDnaEpico()
#' temp_dir <- tempdir()
#' pheno_file <- file.path(temp_dir, "phenoLC.csv")
#' utils::write.csv(ex$targets, pheno_file, row.names = FALSE)
#' output_paths <- writeSvaEnmixOutputs(
#'     svaData = list(sva = ex$sva),
#'     mergedPheno = ex$mergedPheno,
#'     analysisData = ex$analysisData,
#'     phenoFile = pheno_file,
#'     SampleID = "Sample_Name",
#'     dataBaseDir = file.path(temp_dir, "data"),
#'     rBaseDir = file.path(temp_dir, "rData"),
#'     scriptLabel = "svaEnmixExample",
#'     verbose = FALSE,
#'     logs = FALSE
#' )
#' names(output_paths)
#'
#' @export
writeSvaEnmixOutputs <- function(
    svaData, mergedPheno, analysisData = NULL,
    phenoFile = NULL, SampleID = "Sample_Name", sepType = NULL,
    dataBaseDir = "data", rBaseDir = "rData", scriptLabel = "svaEnmix",
    verbose = FALSE, logs = FALSE, log_dir = NULL,
        log_file = "log_writeSvaEnmixOutputs.txt"
) {
    log_path <- resolveLogPathMinfiEwasWater(
        logs = logs, log_dir = log_dir,
        log_file = log_file
    )

    data_dir <- file.path(dataBaseDir, scriptLabel)
    r_dir <- file.path(rBaseDir, scriptLabel)
    dir.create(data_dir, recursive = TRUE, showWarnings = FALSE)
    dir.create(r_dir, recursive = TRUE, showWarnings = FALSE)

    sva_rdata_path <- file.path(r_dir, "svaMatrix.RData")
    sva_csv_path <- file.path(data_dir, "svaMatrix.csv")
    pheno_output_path <- phenoFile

    saveNamedObjectMinfiEwasWater(svaData$sva, "sva", sva_rdata_path)
    utils::write.csv(svaData$sva, sva_csv_path, row.names = TRUE)

    if (!is.null(analysisData)) {
        for (i in seq_len(analysisData$K)) {
            utils::capture.output(summary(analysisData$fullModels[[i]]),
                file = file.path(data_dir, paste0(
                    "summary_full_sva",
                    i, ".txt"
                ))
            )

            if (length(analysisData$droptermSteps[[i]]) > 0L) {
                for (step in analysisData$droptermSteps[[i]]) {
                    utils::capture.output(step$dropterm,
                        file = file.path(
                            data_dir,
                            paste0("dropterm_step_sva", i, ".txt")
                        ),
                        append = TRUE
                    )
                    utils::capture.output(step$summary,
                        file = file.path(
                            data_dir,
                            paste0("dropterm_model_sva", i, ".txt")
                        ),
                        append = TRUE
                    )
                }
            }

            utils::capture.output(analysisData$anovaFull[[i]],
                file = file.path(data_dir, paste0(
                    "anova_full_sva",
                    i, ".txt"
                ))
            )
            utils::capture.output(analysisData$anovaReduced[[i]],
                file = file.path(data_dir, paste0(
                    "anova_reduced_sva",
                    i, ".txt"
                ))
            )
        }
    }

    if (!is.null(pheno_output_path)) {
        replacePhenotypeFileSvaEnmix(
            mergedPheno = mergedPheno,
            phenoFile = pheno_output_path, SampleID = SampleID,
            pcColumns = colnames(svaData$sva), sepType = sepType
        )
    }

    emitLogMinfiEwasWater(
        c(
            paste(
                "SVA Matrix RData saved to: ",
                sva_rdata_path
            ), paste(
                "SVA Matrix CSV saved to:   ",
                sva_csv_path
            ), paste("Updated phenotype + SVA:   ",
                if (is.null(pheno_output_path)) "not saved" else pheno_output_path),
            "============================================================"
        ),
        verbose = verbose, log_path = log_path
    )

    structure(list(
        svaRData = sva_rdata_path, svaCSV = sva_csv_path,
        phenoWithSva = pheno_output_path, dataDir = data_dir,
        rDir = r_dir
    ), class = "dnaEPICO_svaEnmix_paths")
}

create_toy_sva_analysis <- function() {
    sva <- matrix(
        c(-1.0, 1.0, 0.0, 0.5, -0.5, 0.0),
        nrow = 3,
        dimnames = list(c("S1", "S2", "S3"), c("PC1", "PC2"))
    )
    sentrix_id <- factor(c("Chip1", "Chip1", "Chip2"))
    sentrix_position <- factor(c("R02C02", "R04C01", "R04C02"))

    full_models <- lapply(
        seq_len(ncol(sva)),
        function(i) {
            stats::lm(
                sva[, i] ~ sentrix_id + sentrix_position,
                data = data.frame(
                    sentrix_id = sentrix_id,
                    sentrix_position = sentrix_position
                )
            )
        }
    )

    structure(
        list(
            sva = sva,
            K = ncol(sva),
            sentrixID = sentrix_id,
            sentrixPosition = sentrix_position,
            fullModels = full_models,
            reducedModels = full_models,
            droptermSteps = list(list(), list()),
            anovaFull = suppressWarnings(lapply(full_models, stats::anova)),
            anovaReduced = suppressWarnings(lapply(full_models, stats::anova))
        ),
        class = "dnaEPICO_svaEnmix_analysis"
    )
}

create_large_sva_analysis <- function(
    sample_count = 48L,
    K = 6L,
    sentrix_id_count = 44L
) {
    sample_ids <- paste0("S", seq_len(sample_count))
    sentrix_id_levels <- paste0("Chip", seq_len(sentrix_id_count))
    sentrix_position_levels <- paste0("R", sprintf("%02d", seq_len(8L)), "C01")
    sva <- matrix(
        seq_len(sample_count * K) / 100,
        nrow = sample_count,
        dimnames = list(sample_ids, paste0("PC", seq_len(K)))
    )

    structure(
        list(
            sva = sva,
            K = K,
            sentrixID = factor(
                sentrix_id_levels[
                    ((seq_len(sample_count) - 1L) %% sentrix_id_count) + 1L
                ],
                levels = sentrix_id_levels
            ),
            sentrixPosition = factor(
                sentrix_position_levels[
                    ((seq_len(sample_count) - 1L) %%
                        length(sentrix_position_levels)) + 1L
                ],
                levels = sentrix_position_levels
            )
        ),
        class = "dnaEPICO_svaEnmix_analysis"
    )
}

test_that("mergeSvaTargetsEnmix preserves target order", {
    targets <- data.frame(
        Sample_Name = c("S2", "S1", "S3"),
        Group = c("A", "A", "B"),
        stringsAsFactors = FALSE
    )
    sva <- create_toy_sva_analysis()$sva

    merged <- mergeSvaTargetsEnmix(
        targets = targets,
        sva = sva,
        SampleID = "Sample_Name"
    )

    expect_identical(merged$Sample_Name, targets$Sample_Name)
    expect_true(all(c("PC1", "PC2") %in% colnames(merged)))
})

test_that("mergeSvaTargetsEnmix prevents repeated surrogate-variable columns", {
    targets <- data.frame(
        Sample_Name = c("S1", "S2", "S3"),
        PC1 = c(0, 0, 0),
        stringsAsFactors = FALSE
    )
    expect_error(
        mergeSvaTargetsEnmix(
            targets = targets,
            sva = create_toy_sva_analysis()$sva,
            SampleID = "Sample_Name"
        ),
        "would be duplicated"
    )
})

test_that("ENmix control-SVA options follow the documented flag contract", {
    expect_error(
        estimateSvaEnmixControls(NULL, ctrlSvaPercVar = -0.01),
        "between 0 and 1"
    )
    expect_error(
        estimateSvaEnmixControls(NULL, ctrlSvaFlag = 0),
        "must be 1"
    )
    expect_error(
        estimateSvaEnmixControls(NULL, ctrlSvaFlag = 3),
        "must be 1"
    )
})

test_that("Sentrix analysis omits technical factors without variation", {
    testthat::skip_if_not_installed("SummarizedExperiment")
    testthat::skip_if_not_installed("S4Vectors")

    sample_ids <- paste0("S", seq_len(6))
    sva <- matrix(
        seq_len(12) / 10,
        nrow = 6,
        dimnames = list(sample_ids, c("PC1", "PC2"))
    )
    rgset <- SummarizedExperiment::SummarizedExperiment(
        assays = list(signal = matrix(1, nrow = 2, ncol = 6)),
        colData = S4Vectors::DataFrame(
            Sentrix_ID = rep("Chip1", 6),
            Sentrix_Position = rep("R01C01", 6),
            row.names = sample_ids
        )
    )

    result <- analyzeSvaEnmix(
        sva = sva,
        RGSet = rgset,
        logs = FALSE
    )

    expect_length(result$technicalTerms, 0L)
    expect_true(all(vapply(result$fullModels, function(model) {
        identical(attr(stats::terms(model), "term.labels"), character(0))
    }, logical(1))))
})

test_that("plotSvaEnmix saves a sentrix-id plot", {
    tmp <- withr::local_tempdir()
    analysis_data <- create_toy_sva_analysis()
    plot_file <- file.path(tmp, "sva_SentrixID.tiff")

    output <- plotSvaEnmix(
        analysisData = analysis_data,
        plot = "sentrix_id",
        file = plot_file
    )

    expect_identical(output, plot_file)
    expect_true(file.exists(plot_file))
})

test_that("plotSvaEnmix adapts matrix plots and SentrixID legends", {
    tmp <- withr::local_tempdir()
    log_dir <- file.path(tmp, "logs")
    analysis_data <- create_large_sva_analysis()
    plot_file <- file.path(tmp, "sva_SentrixIDPosition.tiff")

    output <- plotSvaEnmix(
        analysisData = analysis_data,
        plot = "matrix",
        file = plot_file,
        width = 600,
        height = 400,
        res = 150,
        logs = TRUE,
        log_dir = log_dir
    )

    expect_identical(output, plot_file)
    expect_true(file.exists(plot_file))

    log_lines <- readLines(file.path(log_dir, "log_plotSvaEnmix.txt"))
    expect_true(any(grepl(
        "SVA matrix plot dimensions:",
        log_lines,
        fixed = TRUE
    )))
    expect_true(any(grepl(
        "SentrixID legend suppressed for matrix plot",
        log_lines,
        fixed = TRUE
    )))
})

test_that("plotSvaEnmix paginates large matrix plots", {
    tmp <- withr::local_tempdir()
    log_dir <- file.path(tmp, "logs")
    analysis_data <- create_large_sva_analysis(sample_count = 56L, K = 7L)
    plot_file <- file.path(tmp, "sva_SentrixIDPosition.tiff")

    output <- plotSvaEnmix(
        analysisData = analysis_data,
        plot = "matrix",
        file = plot_file,
        width = 600,
        height = 400,
        res = 150,
        logs = TRUE,
        log_dir = log_dir
    )

    expect_length(output, 4L)
    expect_true(all(file.exists(output)))
    expect_false(file.exists(plot_file))
    expect_true(all(grepl("_page0[1-4]\\.tiff$", output)))

    log_lines <- readLines(file.path(log_dir, "log_plotSvaEnmix.txt"))
    expect_true(any(grepl(
        "SVA matrix plot pages: 4 pages",
        log_lines,
        fixed = TRUE
    )))
    expect_true(any(grepl("Saved plot paths:", log_lines, fixed = TRUE)))
})

test_that("writeSvaEnmixOutputs writes matrix and summary files", {
    tmp <- withr::local_tempdir()
    analysis_data <- create_toy_sva_analysis()
    sva_data <- structure(
        list(
            sva = analysis_data$sva,
            K = analysis_data$K,
            ctrlSvaPercVar = 0.9,
            ctrlSvaFlag = 1
        ),
        class = "dnaEPICO_svaEnmix_sva"
    )
    merged_pheno <- data.frame(
        Sample_Name = c("S1", "S2", "S3"),
        CD8T = c(0.20, 0.25, 0.30),
        PC1 = analysis_data$sva[, 1],
        PC2 = analysis_data$sva[, 2],
        stringsAsFactors = FALSE
    )
    pheno_file <- file.path(tmp, "pheno.csv")
    utils::write.csv(
        merged_pheno[c("Sample_Name", "CD8T")],
        pheno_file,
        row.names = FALSE
    )

    paths <- writeSvaEnmixOutputs(
        svaData = sva_data,
        mergedPheno = merged_pheno,
        analysisData = analysis_data,
        phenoFile = pheno_file,
        dataBaseDir = file.path(tmp, "data"),
        rBaseDir = file.path(tmp, "rData"),
        scriptLabel = "svaEnmix"
    )

    expect_s3_class(paths, "dnaEPICO_svaEnmix_paths")
    expect_true(file.exists(paths$svaRData))
    expect_true(file.exists(paths$svaCSV))
    expect_true(file.exists(paths$phenoWithSva))
    expect_true(file.exists(file.path(paths$dataDir, "summary_full_sva1.txt")))
    expect_true(file.exists(file.path(paths$dataDir, "anova_full_sva1.txt")))
    expect_identical(
        names(utils::read.csv(pheno_file, check.names = FALSE)),
        c("Sample_Name", "CD8T", "PC1", "PC2")
    )
})

test_that("writeSvaEnmixOutputs transactionally updates one phenotype file", {
    tmp <- withr::local_tempdir()
    analysis_data <- create_toy_sva_analysis()
    pheno_file <- file.path(tmp, "phenoLC.csv")
    input_data <- data.frame(
        Sample_Name = rownames(analysis_data$sva),
        CD8T = c(0.20, 0.25, 0.30),
        status = c("case", "control", "case"),
        stringsAsFactors = FALSE
    )
    utils::write.csv(input_data, pheno_file, row.names = FALSE)
    merged <- cbind(input_data, as.data.frame(analysis_data$sva))

    paths <- writeSvaEnmixOutputs(
        svaData = list(sva = analysis_data$sva),
        mergedPheno = merged,
        phenoFile = pheno_file,
        SampleID = "Sample_Name",
        dataBaseDir = file.path(tmp, "data"),
        rBaseDir = file.path(tmp, "rData"),
        scriptLabel = "svaEnmix"
    )

    updated <- utils::read.csv(pheno_file, check.names = FALSE)
    expect_identical(paths$phenoWithSva, pheno_file)
    expect_identical(updated$Sample_Name, input_data$Sample_Name)
    expect_equal(updated[c("CD8T", "status")], input_data[c("CD8T", "status")])
    expect_equal(
        unname(as.matrix(updated[c("PC1", "PC2")])),
        unname(as.matrix(merged[c("PC1", "PC2")]))
    )
    expect_identical(
        list.files(tmp, pattern = "^\\.phenoLC\\.csv\\.(sva|backup)-", all.files = TRUE),
        character(0)
    )
    expect_identical(
        list.files(tmp, pattern = "^phenoLC", all.files = TRUE),
        "phenoLC.csv"
    )
})

test_that("failed phenotype validation leaves the original file unchanged", {
    tmp <- withr::local_tempdir()
    pheno_file <- file.path(tmp, "phenoLC.csv")
    original <- data.frame(
        Sample_Name = c("S1", "S2", "S3"),
        CD8T = c(0.20, 0.25, 0.30),
        stringsAsFactors = FALSE
    )
    utils::write.csv(original, pheno_file, row.names = FALSE)
    checksum <- tools::md5sum(pheno_file)
    invalid <- cbind(original, PC1 = c(0.1, Inf, 0.3))

    expect_error(
        replacePhenotypeFileSvaEnmix(
            mergedPheno = invalid,
            phenoFile = pheno_file,
            SampleID = "Sample_Name",
            pcColumns = "PC1"
        ),
        "only finite values"
    )

    expect_identical(unname(tools::md5sum(pheno_file)), unname(checksum))
    expect_equal(utils::read.csv(pheno_file), original)
    expect_identical(
        list.files(tmp, pattern = "^\\.phenoLC\\.csv\\.(sva|backup)-", all.files = TRUE),
        character(0)
    )
})

test_that("svaEnmix exposes one phenotype-file interface", {
    expect_false("outputPhenoFile" %in% names(formals(svaEnmix)))
})

test_that("svaEnmix logs fatal errors before stopping", {
    tmp <- withr::local_tempdir()
    pheno_file <- file.path(tmp, "pheno.csv")
    log_dir <- file.path(tmp, "logs")
    pheno <- data.frame(
        UID = "S1",
        Sentrix_ID = "Chip1",
        Sentrix_Position = "R01C01",
        stringsAsFactors = FALSE
    )
    utils::write.csv(pheno, pheno_file, row.names = FALSE)

    expect_error(
        svaEnmix(
            phenoFile = pheno_file,
            rgsetData = file.path(tmp, "missing_RGSet.RData"),
            SampleID = "UID",
            SentrixIDColumn = "Sentrix_ID",
            SentrixPositionColumn = "Sentrix_Position",
            outputLogs = log_dir,
            logs = TRUE,
            verbose = FALSE,
            saveOutputs = FALSE
        ),
        "Input file does not exist"
    )

    log_file <- file.path(log_dir, "log_svaEnmix.txt")
    expect_true(file.exists(log_file))
    log_lines <- readLines(log_file, warn = FALSE)
    expect_true(any(grepl("ERROR in svaEnmix", log_lines, fixed = TRUE)))
    expect_true(any(grepl("Input file does not exist", log_lines, fixed = TRUE)))
    expect_true(any(grepl("Failing call:", log_lines, fixed = TRUE)))
    expect_true(any(grepl("Call stack:", log_lines, fixed = TRUE)))
    expect_true(any(grepl("svaEnmix", log_lines, fixed = TRUE)))
})

test_that("svaEnmix accepts saved RGSet inputs", {
    testthat::skip_if_not_installed("minfiData")

    tmp <- withr::local_tempdir()
    ex <- dnaEPICO:::exampleMinfiBaseDataDnaEpico()
    pheno_file <- file.path(tmp, "pheno.csv")
    rgset_path <- file.path(tmp, "RGSet.RData")
    RGSet <- ex$RGSet
    utils::write.csv(ex$targets, pheno_file, row.names = FALSE)
    save(RGSet, file = rgset_path)

    result <- svaEnmix(
        phenoFile = pheno_file,
        rgsetData = rgset_path,
        SampleID = "Sample_Name",
        arrayType = "IlluminaHumanMethylation450k",
        annotationVersion = "ilmn12.hg19",
        SentrixIDColumn = "Sentrix_ID",
        SentrixPositionColumn = "Sentrix_Position",
        outputLogs = file.path(tmp, "logs"),
        figureBaseDir = file.path(tmp, "figures"),
        dataBaseDir = file.path(tmp, "data"),
        rBaseDir = file.path(tmp, "rData"),
        verbose = FALSE,
        logs = FALSE,
        saveOutputs = FALSE
    )

    expect_s3_class(result, "dnaEPICO_svaEnmix")
    expect_equal(nrow(result$targets), ncol(result$RGSet))
    expect_equal(colnames(result$svaData$sva), paste0("PC", seq_len(result$svaData$K)))
    expect_false(anyDuplicated(colnames(result$mergedPheno)) > 0L)
})

test_that("svaEnmix accepts wrapper files that contain an RGSet element", {
    testthat::skip_if_not_installed("minfiData")

    tmp <- withr::local_tempdir()
    ex <- dnaEPICO:::exampleMinfiBaseDataDnaEpico()
    pheno_file <- file.path(tmp, "pheno.csv")
    rgset_path <- file.path(tmp, "RGSet_wrapper.RData")
    rgset_state <- list(
        RGSet = ex$RGSet,
        note = "wrapper object"
    )
    utils::write.csv(ex$targets, pheno_file, row.names = FALSE)
    save(rgset_state, file = rgset_path)

    result <- svaEnmix(
        phenoFile = pheno_file,
        rgsetData = rgset_path,
        SampleID = "Sample_Name",
        arrayType = "IlluminaHumanMethylation450k",
        annotationVersion = "ilmn12.hg19",
        SentrixIDColumn = "Sentrix_ID",
        SentrixPositionColumn = "Sentrix_Position",
        outputLogs = file.path(tmp, "logs"),
        figureBaseDir = file.path(tmp, "figures"),
        dataBaseDir = file.path(tmp, "data"),
        rBaseDir = file.path(tmp, "rData"),
        verbose = FALSE,
        logs = FALSE,
        saveOutputs = FALSE
    )

    expect_s3_class(result, "dnaEPICO_svaEnmix")
    expect_equal(nrow(result$targets), ncol(result$RGSet))
})

test_that("svaEnmix reports invalid saved RGSet objects clearly", {
    tmp <- withr::local_tempdir()
    pheno_file <- file.path(tmp, "pheno.csv")
    bad_path <- file.path(tmp, "bad_RGSet.RData")
    bad_object <- list(not_rgset = 1)
    pheno <- data.frame(
        UID = "S1",
        Sentrix_ID = "Chip1",
        Sentrix_Position = "R01C01",
        stringsAsFactors = FALSE
    )
    utils::write.csv(pheno, pheno_file, row.names = FALSE)
    save(bad_object, file = bad_path)

    expect_error(
        svaEnmix(
            phenoFile = pheno_file,
            rgsetData = bad_path,
            SampleID = "UID",
            SentrixIDColumn = "Sentrix_ID",
            SentrixPositionColumn = "Sentrix_Position",
            outputLogs = file.path(tmp, "logs"),
            logs = TRUE,
            verbose = FALSE,
            saveOutputs = FALSE
        ),
        "does not expose a usable sample count"
    )
})

#' Logging helpers for preprocessingMinfiEwasWater
#'
#' @param lines Character vector of log lines to emit.
#' @param verbose Logical. If `TRUE`, send log lines to the console with
#'   `message()`.
#' @param log_path Character or `NULL`. Path to the log file. If `NULL`, log
#'   lines are not written to disk.
#'
#' @return Invisibly returns `NULL`.
#' @description
#' Internal helper that writes preprocessing log lines to the console and/or a
#' log file while keeping user-facing functions quiet by default.
#' @keywords internal
#' @noRd
emitLogMinfiEwasWater <- function(lines, verbose = FALSE, log_path = NULL) {
    if (length(lines) == 0L) {
        return(invisible(NULL))
    }

    lines <- as.character(lines)

    if (!is.null(log_path)) {
        con <- file(log_path, open = "at")
        on.exit(close(con), add = TRUE)
        writeLines(lines, con = con, sep = "\n", useBytes = TRUE)
    }

    if (isTRUE(verbose)) {
        for (line in lines) {
            message(line)
        }
    }

    invisible(NULL)
}

#' Format call-stack lines for dnaEPICO workflow error logs
#'
#' @param calls List of calls, typically from `sys.calls()`.
#'
#' @return Character vector of formatted stack-trace lines.
#' @description
#' Internal helper that converts R calls into readable log lines while dropping
#' the noisier `tryCatch()` bookkeeping frames.
#' @keywords internal
#' @noRd
formatCallStackMinfiEwasWater <- function(calls) {
    if (length(calls) == 0L) {
        return("Call stack: unavailable")
    }

    call_text <- vapply(calls, function(call) {
        paste(deparse(call,
            width.cutoff = 500L
        ), collapse = " ")
    }, character(1))

    keep <- !grepl(
        "^(tryCatch|tryCatchList|tryCatchOne|doTryCatch|withCallingHandlers)\\b",
        call_text
    )
    keep <- keep & !grepl(
        "^withLoggedErrorsMinfiEwasWater\\b",
        call_text
    )

    filtered_calls <- call_text[keep]
    if (length(filtered_calls) == 0L) {
        filtered_calls <- call_text
    }

    c("Call stack:", paste0(
        "  ", seq_along(filtered_calls),
        ": ", filtered_calls
    ))
}

#' Wrap a workflow block and log any error before rethrowing
#'
#' @param expr Expression to evaluate.
#' @param log_path Character or `NULL`. Path to the log file.
#' @param verbose Logical. If `TRUE`, mirror the error message with `message()`.
#' @param context Character. Short label describing where the error happened.
#'
#' @return Returns the value of `expr` when successful.
#' @description
#' Internal helper that ensures top-level workflow wrappers record fatal errors
#' in
#' their log files before the error is rethrown to the caller.
#' @keywords internal
#' @noRd
withLoggedErrorsMinfiEwasWater <- function(
    expr, log_path = NULL,
    verbose = FALSE, context = "dnaEPICO workflow"
) {
    tryCatch(expr, error = function(e) {
        failing_call <- conditionCall(e)
        call_lines <- formatCallStackMinfiEwasWater(sys.calls())
        emitLogMinfiEwasWater(
            c(
                "============================================================",
                paste("ERROR in", context, ":"), conditionMessage(e),
                if (is.null(failing_call)) {
                    "Failing call: unavailable"
                } else {
                    paste("Failing call:", paste(deparse(failing_call,
                        width.cutoff = 500L
                    ), collapse = " "))
                }, call_lines,
                    "============================================================"
            ),
            verbose = verbose, log_path = log_path
        )
        stop(e)
    })
}

#' Resolve the log file path for preprocessingMinfiEwasWater helpers
#'
#' @param logs Logical. If `TRUE`, create a log file path.
#' @param log_dir Character or `NULL`. Directory used for log files.
#' @param log_file Character. File name used for the log file.
#'
#' @return Character scalar with the resolved log path, or `NULL` when
#'   `logs = FALSE`.
#' @description
#' Internal helper that creates the target log directory when needed and returns
#' the path used for file-based logs.
#' @keywords internal
#' @noRd
resolveLogPathMinfiEwasWater <- function(
    logs = FALSE, log_dir = NULL,
    log_file = "log_preprocessingMinfiEwasWater.txt"
) {
    if (!isTRUE(logs)) {
        return(NULL)
    }

    if (is.null(log_dir)) {
        log_dir <- getwd()
    }

    dir.create(log_dir, recursive = TRUE, showWarnings = FALSE)
    log_dir <- normalizePath(log_dir, winslash = "/", mustWork = FALSE)

    file.path(log_dir, log_file)
}

#' Resolve the phenotype separator used in preprocessingMinfiEwasWater helpers
#'
#' @param sepType Character or `NULL`. Separator description supplied by the
#'   user.
#'
#' @return Character scalar with the resolved separator, or `NULL` for the
#'   default comma separator used by `utils::read.csv()`.
#' @description
#' Internal helper that standardizes separator handling across modular
#' preprocessing functions.
#' @keywords internal
#' @noRd
resolveSeparatorMinfiEwasWater <- function(sepType = NULL) {
    if (length(sepType) == 0L || is.na(sepType[[1L]])) {
        return(NULL)
    }

    sepType <- as.character(sepType[[1L]])
    sepTypeLabel <- trimws(sepType)

    if (!nzchar(sepType) || identical(
        toupper(sepTypeLabel),
        "NULL"
    )) {
        return(NULL)
    }

    if (identical(sepType, "\\t")) {
        return("\t")
    }

    sepType
}

#' Split character options used in preprocessingMinfiEwasWater helpers
#'
#' @param x Character vector or scalar string to split.
#' @param sep Character scalar used to split `x` when `x` has length 1.
#'
#' @return Character vector with empty values removed and whitespace trimmed.
#' @description
#' Internal helper that standardizes comma- and semicolon-separated option
#' parsing across preprocessing helpers.
#' @keywords internal
#' @noRd
splitOptionMinfiEwasWater <- function(x, sep = ",") {
    if (length(x) == 0L || all(is.na(x))) {
        return(character(0))
    }

    if (length(x) == 1L) {
        pieces <- strsplit(as.character(x), split = sep, fixed = TRUE)[[1]]
    } else {
        pieces <- as.character(x)
    }

    pieces <- trimws(pieces)
    pieces[!is.na(pieces) & nzchar(pieces)]
}

#' Validate a scalar logical option
#'
#' @keywords internal
#' @noRd
validateLogicalScalarDnaEpico <- function(value, name) {
    if (!is.logical(value) || length(value) != 1L || is.na(value)) {
        stop(name, " must be either TRUE or FALSE.", call. = FALSE)
    }

    value
}

#' Calculate a finite mean without creating NaN
#'
#' @keywords internal
#' @noRd
meanFiniteOrNADnaEpico <- function(values) {
    finite_values <- values[is.finite(values)]
    if (length(finite_values) == 0L) {
        return(NA_real_)
    }

    mean(finite_values)
}

#' Validate sample identifiers used to align package objects
#'
#' @keywords internal
#' @noRd
validateSampleIdentifiersDnaEpico <- function(ids, label) {
    ids <- as.character(ids)

    if (length(ids) == 0L) {
        stop(label, " must contain at least one sample identifier.",
            call. = FALSE
        )
    }

    missing_ids <- is.na(ids) | !nzchar(trimws(ids))
    if (any(missing_ids)) {
        stop(label, " contains missing or blank sample identifiers.",
            call. = FALSE
        )
    }

    duplicated_ids <- unique(ids[duplicated(ids)])
    if (length(duplicated_ids) > 0L) {
        stop(label, " contains duplicate sample identifiers: ",
            paste(duplicated_ids, collapse = ", "),
            call. = FALSE
        )
    }

    ids
}

#' Match sample identifiers without allowing silent missing rows
#'
#' @keywords internal
#' @noRd
matchSampleIdentifiersDnaEpico <- function(
    query, reference,
    queryLabel = "query sample identifiers",
        referenceLabel = "reference sample identifiers",
    requireSameSet = FALSE
) {
    query <- validateSampleIdentifiersDnaEpico(query, queryLabel)
    reference <- validateSampleIdentifiersDnaEpico(
        reference,
        referenceLabel
    )

    matched <- match(query, reference)
    if (anyNA(matched)) {
        stop(queryLabel, " not found in ", referenceLabel, ": ",
            paste(query[is.na(matched)], collapse = ", "),
            call. = FALSE
        )
    }

    if (isTRUE(requireSameSet) && length(query) != length(reference)) {
        extra_reference <- setdiff(reference, query)
        stop(queryLabel, " and ", referenceLabel,
            " do not contain the same samples",
            if (length(extra_reference) > 0L) {
                paste0(": extra reference samples are ", paste(extra_reference,
                    collapse = ", "
                ))
            } else {
                "."
            },
            call. = FALSE
        )
    }

    matched
}

#' Convert reported or predicted sex values to a common representation
#'
#' @keywords internal
#' @noRd
canonicalizeSexDnaEpico <- function(values) {
    original <- values
    text <- trimws(as.character(values))
    lower <- tolower(text)
    missing <- is.na(values) | !nzchar(text)

    code <- rep(NA_integer_, length(values))
    code[!missing & lower %in% c("f", "female", "0")] <- 0L
    code[!missing & lower %in% c("m", "male", "1")] <- 1L

    label <- rep(NA_character_, length(values))
    label[code == 0L & !is.na(code)] <- "F"
    label[code == 1L & !is.na(code)] <- "M"

    unknown <- unique(text[!missing & is.na(code)])

    list(original = original, code = code, label = label, unknown = unknown)
}

#' Resolve sex labels supplied to methylation normalization
#'
#' @keywords internal
#' @noRd
resolveNormalizationSexDnaEpico <- function(colData, sexColumn) {
    if (is.null(sexColumn) || !(sexColumn %in% colnames(colData))) {
        return(NULL)
    }

    reported_sex <- canonicalizeSexDnaEpico(colData[[sexColumn]])
    sex_labels <- reported_sex$label
    if (anyNA(sex_labels) && "PredSex" %in% colnames(colData)) {
        predicted_sex <- canonicalizeSexDnaEpico(colData$PredSex)$label
        replace_with_predicted <- is.na(sex_labels) & !is.na(predicted_sex)
        sex_labels[replace_with_predicted] <-
            predicted_sex[replace_with_predicted]
    }

    if (anyNA(sex_labels)) {
        stop("The normalization sex information contains missing or unsupported values",
            if (length(reported_sex$unknown) > 0L) {
                paste0(": ", paste(reported_sex$unknown, collapse = ", "))
            } else {
                "."
            },
            call. = FALSE
        )
    }

    sex_labels
}

#' Capture preview lines for logging
#'
#' @param x Object to preview.
#' @param n Integer. Number of rows to show when `x` is a matrix or data frame.
#'
#' @return Character vector containing formatted preview lines.
#' @description
#' Internal helper that standardizes preview output for logs.
#' @keywords internal
#' @noRd
previewLinesMinfiEwasWater <- function(x, n = 6L) {
    if (is.matrix(x) || is.data.frame(x)) {
        return(utils::capture.output(utils::head(x, n = n)))
    }

    utils::capture.output(methods::show(x))
}

#' Draw a stored plot object for preprocessing helpers
#'
#' @param plot_object Plot object to draw.
#'
#' @return Invisibly returns `NULL`.
#' @description
#' Internal helper that draws stored ggplot or grid objects without relying on
#' `print()` in package code.
#' @keywords internal
#' @noRd
drawPlotObjectMinfiEwasWater <- function(plot_object) {
    if (inherits(plot_object, "ggplot")) {
        grid::grid.newpage()
        grid::grid.draw(ggplot2::ggplotGrob(plot_object))
        return(invisible(NULL))
    }

    if (grid::is.grob(plot_object)) {
        grid::grid.newpage()
        grid::grid.draw(plot_object)
        return(invisible(NULL))
    }

    stop("plot_object must inherit from 'ggplot' or be a grid grob.",
        call. = FALSE
    )
}

#' Draw and optionally save a plot for preprocessingMinfiEwasWater helpers
#'
#' @param draw_fun Function with no arguments that draws the plot.
#' @param display Logical. If `TRUE`, draw on the active graphics device.
#' @param file Character or `NULL`. TIFF file path used for saved output.
#' @param width Integer. Plot width in pixels when `file` is supplied.
#' @param height Integer. Plot height in pixels when `file` is supplied.
#' @param res Integer. TIFF resolution in DPI when `file` is supplied.
#'
#' @return Invisibly returns `file` when a TIFF is written, otherwise `NULL`.
#' @description
#' Internal helper that centralizes plot saving for preprocessing helpers.
#' @keywords internal
#' @noRd
runPlotMinfiEwasWater <- function(
    draw_fun, display = FALSE,
    file = NULL, width = 2000L, height = 1000L, res = 150L
) {
    if (is.null(file) && !isTRUE(display)) {
        return(invisible(NULL))
    }

    if (!is.null(file)) {
        dir.create(dirname(file), recursive = TRUE, showWarnings = FALSE)
        save_plot <- function() {
            grDevices::tiff(
                filename = file, width = width, height = height,
                res = res, type = "cairo"
            )
            on.exit(grDevices::dev.off(), add = TRUE)
            draw_fun()
        }
        save_plot()
    }

    if (isTRUE(display)) {
        draw_fun()
    }

    invisible(file)
}

#' Save a named preprocessingMinfiEwasWater object
#'
#' @param object Object to save.
#' @param object_name Character. Object name to preserve in the `.RData` file.
#' @param file Character. Target file path.
#'
#' @return Invisibly returns `file`.
#' @description
#' Internal helper used by the convenience wrapper to preserve the documented
#' `.RData` object names consumed by other package functions.
#' @keywords internal
#' @noRd
saveNamedObjectMinfiEwasWater <- function(
    object, object_name,
    file
) {
    dir.create(dirname(file), recursive = TRUE, showWarnings = FALSE)

    save_env <- list2env(stats::setNames(list(object), object_name),
        parent = emptyenv()
    )
    save(list = object_name, file = file, envir = save_env)

    invisible(file)
}

---
title: "Local Use of dnaEPICO"
author:
  - name: Paul Ruiz
    affiliation:
      - Queensland University of Technology
    email: ruizpint@qut.edu.au
output:
  BiocStyle::html_document:
    self_contained: yes
    toc: true
    toc_float: true
    toc_depth: 2
date: "`r Sys.Date()`"
package: dnaEPICO
bibliography: dnaEPICO.bib
vignette: >
  %\VignetteIndexEntry{2. Local Use of dnaEPICO}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
---

```{r, include = FALSE}
startTime <- Sys.time()
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>",
    crop = NULL
)

library(dnaEPICO)

can_run_preprocessing <- requireNamespace("minfiData", quietly = TRUE) &&
    requireNamespace("IlluminaHumanMethylation450kmanifest", quietly = TRUE) &&
    requireNamespace("IlluminaHumanMethylation450kanno.ilmn12.hg19", quietly = TRUE)

can_run_sva <- requireNamespace("minfiData", quietly = TRUE)
```

# Introduction

This vignette demonstrates the local, interactive use of `dnaEPICO`. In this
mode, the main functions return structured in-memory result objects that can be
inspected directly in the R session. This is the recommended route when you are
checking inputs, reviewing quality-control outputs, or learning the structure of
the package results before moving to larger analyses.

The examples below focus on the first three workflow stages:

1. `preprocessingMinfiEwasWater()`
2. `svaEnmix()`
3. `preprocessingPheno()`

All examples use `saveOutputs = FALSE`, which keeps the workflow in memory and
avoids writing file outputs.

# Citation

The `r Biocpkg("dnaEPICO")` package uses methods from several external
packages. Because no single manuscript describes all components, the guidance
below explains how to cite `dnaEPICO` depending on the functions you use.

This citation guidance is adapted from the vignettes and user guides of
[minfi](https://github.com/hansenlab/minfi/blob/devel/vignettes/minfi.Rmd),
[ENmix](https://github.com/xuz1/ENmix/blob/master/vignettes/ENmix.Rmd), and
[wateRmelon](https://github.com/schalkwyk/wateRmelon/blob/master/vignettes/references.bib).

- If you use `preprocessingMinfiEwasWater()`, please cite @minfi; @minfiEPIC;
    @ENmix; @wateRmelon; @ewastools; @SWAN; @funnorm; @noob; @quantile.
- If you use `svaEnmix()`, please cite @minfi; @ENmix.


# Required knowledge

`r Biocpkg("dnaEPICO")` is built on core Bioconductor infrastructure for
high-dimensional genomic data, with a focus on Illumina DNA methylation arrays.
To read this vignette, we assume that you are already familiar with the general
DNA methylation workflow. If not, we recommend first reading
[this tutorial](https://paulyrp.github.io/2025-cpgpneurogenomics-workshop/tutorial.html),
which provides a practical introduction to the main concepts and analysis steps.

Preprocessing and quality control are performed using established Bioconductor
tools, including `r Biocpkg("minfi")`, `r Biocpkg("ENmix")`, and
`r Biocpkg("wateRmelon")`. Users are expected to have basic familiarity with R,
Bioconductor workflows, and Illumina IDAT file structures.

If you are asking yourself the question "Where do I start using
Bioconductor?", you might be interested in
[this post](https://www.bioconductor.org/install/).

## Probe-exclusion reference files

The lists of excluded probes depend on the Illumina methylation-array platform.

- For **Illumina HumanMethylationEPIC v2.0**, use the cross-reactive probe-exclusion file from [Peters et al. (2024)](https://springernature.figshare.com/articles/dataset/Additional_file_8_of_Characterisation_and_reproducibility_of_the_HumanMethylationEPIC_v2_0_BeadChip_for_DNA_methylation_profiling/26689706?file=48531323).

- For **Illumina MethylationEPIC**, also known as the **850k array**, use the probe-exclusion resources from [Pidsley et al. (2016)](https://springernature.figshare.com/articles/dataset/Additional_file_1_Table_S1_of_Critical_evaluation_of_the_Illumina_MethylationEPIC_BeadChip_microarray_for_whole-genome_DNA_methylation_profiling/4405946?file=7132334). The supplementary files commonly used together are [13059_2016_1066_MOESM1_ESM.csv](https://static-content.springer.com/esm/art%3A10.1186%2Fs13059-016-1066-1/MediaObjects/13059_2016_1066_MOESM1_ESM.csv), [13059_2016_1066_MOESM4_ESM.csv](https://static-content.springer.com/esm/art%3A10.1186%2Fs13059-016-1066-1/MediaObjects/13059_2016_1066_MOESM4_ESM.csv), [13059_2016_1066_MOESM5_ESM.csv](https://static-content.springer.com/esm/art%3A10.1186%2Fs13059-016-1066-1/MediaObjects/13059_2016_1066_MOESM5_ESM.csv), and [13059_2016_1066_MOESM6_ESM.csv](https://static-content.springer.com/esm/art%3A10.1186%2Fs13059-016-1066-1/MediaObjects/13059_2016_1066_MOESM6_ESM.csv).

- For **Illumina HumanMethylation450k**, use the cross-reactive and polymorphic probe resources from [Chen et al. (2013)](https://figshare.com/articles/dataset/methylAnalysisDataV3_tar_gz/4800970?file=7896205).

- Multiple probe-exclusion files can be supplied as a semicolon-separated value in `probeExclusionPath`. `dnaEPICO` reads probe IDs from each file, uses `probeExclusionIdColumn` when supplied, or otherwise auto-detects common probe-ID columns such as `ProbeID`, `TargetID`, `IlmnID`, and `Name`. The unique union of all probe IDs is then used to filter the normalised object.

- For **EPICv2**, setting `useEpicV2Manifest = TRUE` also retrieves the expanded Peters et al. manifest from AnnotationHub resource [AH116484](https://bioconductor.org/packages/release/data/annotation/vignettes/EPICv2manifest/inst/doc/EPICv2manifest.html). Probes flagged in selected manifest columns are added to the same exclusion set. By default, probes flagged by `CH_WGBS_evidence`, `CH_BLAT`, or `MissingPos` are removed, while `MismatchPos` is retained unless explicitly enabled.

# Installation

The `r Biocpkg("dnaEPICO")` package depends on `r Biocpkg("minfi")`,
`r Biocpkg("ENmix")`, and `r Biocpkg("wateRmelon")`, among others. Install
dependencies with BiocManager:

```{r, eval = FALSE}
if (!requireNamespace("BiocManager", quietly = TRUE)) {
    install.packages("BiocManager")
}

BiocManager::install("dnaEPICO")

BiocManager::valid()
```

```{r, eval = FALSE, message = FALSE}
library("dnaEPICO")
```

# Step 1: preprocessingMinfiEwasWater

The preprocessing function reads a phenotype table and IDAT files, performs
quality control, normalisation, filtering, and cell-composition estimation, and
returns a list-based result object. The next chunk runs the function on a small
example input and prints the class, element names, and a preview of the
filtered phenotype table.

### Create example input files

The next chunk shows how to recreate the temporary files used in this example.
It writes a small phenotype table to a temporary directory, copies a small set
of IDAT files, and records the probe-exclusion reference used during probe
filtering.

```{r preprocessing-local-inputs, eval = can_run_preprocessing, message = FALSE, warning = FALSE}
preprocessing_inputs <- dnaEPICO:::exampleMinfiIdatInputsDnaEpico(n = 6L)

names(preprocessing_inputs)
preprocessing_inputs$tempDir
basename(preprocessing_inputs$phenoFile)
head(basename(list.files(preprocessing_inputs$idatFolder, full.names = TRUE)), 4)
basename(preprocessing_inputs$probeExclusionPath)
```

`preprocessing_inputs` is a small list returned by the internal example helper.
`names(preprocessing_inputs)` shows its main elements: `tempDir` is the
temporary working directory, `idatFolder` contains the copied example IDAT
files, `phenoFile` is the phenotype table used by the function, `targets` is the
same phenotype information already loaded into memory, `arrayType` and
`annotationVersion` describe the array platform, and `probeExclusionPath`
points to the probe-exclusion reference used later during filtering.

```{r preprocessing-local, eval = can_run_preprocessing, message = FALSE, warning = FALSE}
preprocessResult <- preprocessingMinfiEwasWater(
    phenoFile = preprocessing_inputs$phenoFile,
    idatFolder = preprocessing_inputs$idatFolder,
    outputLogs = file.path(preprocessing_inputs$tempDir, "logs"),
    nSamples = 6,
    SampleID = "Sample_Name",
    arrayType = preprocessing_inputs$arrayType,
    annotationVersion = preprocessing_inputs$annotationVersion,
    scriptLabel = "preprocessingMinfiEwasWater",
    baseDataFolder = file.path(preprocessing_inputs$tempDir, "rData"),
    figureBaseDir = file.path(preprocessing_inputs$tempDir, "figures"),
    detPThreshold = 1,
    normMethods = "quantile",
    sexColumn = "Sex",
    removeSexMismatch = FALSE,
    pvalThreshold = 1,
    chrToRemove = "",
    snpsToRemove = "SBE",
    mafThreshold = 1,
    probeExclusionPath = preprocessing_inputs$probeExclusionPath,
    plotGroupVar = "Sex",
    lcRef = "saliva",
    phenoOrder = "Sample_Name;Sex;Basename;Sentrix_ID;Sentrix_Position",
    lcPhenoDir = preprocessing_inputs$tempDir,
    display = FALSE,
    verbose = FALSE,
    logs = FALSE,
    saveOutputs = FALSE
)

class(preprocessResult)
names(preprocessResult)
head(preprocessResult$targets[, c(
    "Sample_Name",
    "Sex",
    "Sentrix_ID",
    "Sentrix_Position"
)])
```

The returned object has class `dnaEPICO_preprocessingMinfiEwasWater`.
`names(preprocessResult)` lists the main elements returned by the function. The
`targets` element is the filtered phenotype table aligned to the retained
samples. The remaining elements are nested result objects for the major
preprocessing stages, including the `RGSet`, the methylation metrics, and the
cell-composition results.

Arguments:

- `phenoFile` and `idatFolder` define the phenotype table and the directory of
    raw IDAT pairs.
- `nSamples = 6` keeps the example small enough for interactive use.
- `SampleID = "Sample_Name"` identifies the column used to align phenotype data
    and methylation objects.
- `arrayType` and `annotationVersion` must match the array platform being
    processed.
- `detPThreshold`, `normMethods`, `pvalThreshold`, `chrToRemove`,
    `snpsToRemove`, and `mafThreshold` control quality filtering and probe
    filtering.
- `probeExclusionPath` supplies the probe-exclusion reference used during
    filtering.
- `sexColumn` identifies reported sex for comparison with methylation-predicted
    sex. `removeSexMismatch = FALSE` retains and reports confirmed mismatches;
    `TRUE` removes samples only when both values are known and disagree. Missing
    or unknown sex values are retained.
- `plotGroupVar`, `lcRef`, and `phenoOrder` control quality-control grouping,
    cell-type estimation, and the structure of the returned phenotype table.
- `display = FALSE`, `verbose = FALSE`, `logs = FALSE`, and
    `saveOutputs = FALSE` keep the example quiet, in memory, and free of file
    outputs.

The next chunk inspects three important elements: the filtered `RGSet`, the
metric matrices, and the phenotype table augmented with cell-composition
estimates.

```{r preprocessing-local-inspect, eval = can_run_preprocessing, message = FALSE, warning = FALSE}
class(preprocessResult$RGSet)
names(preprocessResult$metricsData)
head(preprocessResult$lcData$phenoLC[, 1:6])
```

In local use, these are the most useful objects to inspect first.
`preprocessResult$RGSet` is the core methylation object. `metricsData`
contains the beta-value, M-value, and copy-number matrices used in downstream
steps. `lcData$phenoLC` is the phenotype table augmented with estimated cell
composition and is the object passed forward to later workflow stages.

# Step 2: svaEnmix

The SVA function loads a phenotype table and a saved `RGSet`, estimates
surrogate variables from ENmix control probes, and returns both the surrogate
variables and the merged phenotype table.

### Create example input files

The next chunk shows how to recreate the temporary files consumed by
`svaEnmix()`: a phenotype CSV file and a saved `RGSet` object in `.RData`
format.

```{r sva-local-inputs, eval = can_run_sva, message = FALSE, warning = FALSE}
sva_inputs <- dnaEPICO:::exampleMinfiBaseDataDnaEpico()
sva_temp_dir <- tempdir()
sva_targets_file <- file.path(sva_temp_dir, "sva_targets.csv")
sva_rgset_file <- file.path(sva_temp_dir, "sva_RGSet.RData")

names(sva_inputs)
class(sva_inputs$RGSet)
utils::write.csv(sva_inputs$targets, sva_targets_file, row.names = FALSE)
RGSet <- sva_inputs$RGSet
save(RGSet, file = sva_rgset_file)

basename(c(sva_targets_file, sva_rgset_file))
file.exists(c(sva_targets_file, sva_rgset_file))
```

`sva_inputs` is another helper list used to make the vignette self-contained.
`names(sva_inputs)` shows that it provides the in-memory `RGSet`, the matching
`targets` table, and a `probeExclusionPath` reference. In this step, the function
needs file inputs rather than the in-memory objects, so the chunk writes
`targets` to `sva_targets.csv` and saves the `RGSet` object to
`sva_RGSet.RData`.

```{r sva-local, eval = can_run_sva, message = FALSE, warning = FALSE}
svaResult <- svaEnmix(
    phenoFile = sva_targets_file,
    rgsetData = sva_rgset_file,
    outputLogs = file.path(tempdir(), "logs"),
    SampleID = "Sample_Name",
    arrayType = "IlluminaHumanMethylation450k",
    annotationVersion = "ilmn12.hg19",
    SentrixIDColumn = "Sentrix_ID",
    SentrixPositionColumn = "Sentrix_Position",
    ctrlSvaPercVar = 0.90,
    ctrlSvaFlag = 1,
    scriptLabel = "svaEnmix",
    display = FALSE,
    verbose = FALSE,
    logs = FALSE,
    saveOutputs = FALSE
)

class(svaResult)
names(svaResult)
dim(svaResult$svaData$sva)
head(svaResult$mergedPheno[, 1:min(6, ncol(svaResult$mergedPheno))])
```

The returned object has class `dnaEPICO_svaEnmix`, and `names(svaResult)`
shows the main SVA components returned by the function. The `svaData$sva`
matrix contains one row per sample and one column per surrogate variable. The
`mergedPheno` table is often the most convenient object to inspect because it
combines the original phenotype data and the new SVA columns in one place.

Arguments:

- `phenoFile` points to the phenotype table that already includes the cell
    composition estimates from preprocessing.
- `rgsetData` points to the saved `RGSet` used to derive ENmix control-probe
    surrogate variables.
- `SampleID`, `SentrixIDColumn`, and `SentrixPositionColumn` identify the
    sample and array-position columns used in the SVA analysis.
- `ctrlSvaPercVar = 0.90` asks the procedure to keep enough control-derived
    surrogate variables to explain 90% of the control-probe variance.
- `ctrlSvaFlag = 1` enables the control-based SVA workflow.
- `display = FALSE`, `verbose = FALSE`, `logs = FALSE`, and
    `saveOutputs = FALSE` keep the example interactive and in-memory.

The association analysis between the surrogate variables and array-position
metadata is stored separately in `analysisData`.

```{r sva-local-inspect, eval = can_run_sva, message = FALSE, warning = FALSE}
class(svaResult$analysisData)
names(svaResult$analysisData)
```

This object contains the fitted models and ANOVA summaries used to assess how
the surrogate variables relate to Sentrix chip and position effects.

# Step 3: preprocessingPheno

The phenotype-preparation function aligns the phenotype table with the beta,
M-value, and copy-number matrices; splits the data by timepoint; constructs a
combined longitudinal object; and prepares Clock Foundation export tables.

### Create example input files

The next chunk shows how to recreate the temporary phenotype and matrix files
used in this example. The helper writes a phenotype table plus aligned beta,
M-value, and copy-number objects to a temporary directory.

```{r preprocessingPheno-local-inputs, message = FALSE, warning = FALSE}
pheno_inputs <- dnaEPICO:::examplePreprocessingPhenoStateDnaEpico()

names(pheno_inputs)
pheno_inputs$tempDir
basename(c(
    pheno_inputs$phenoPath,
    pheno_inputs$betaPath,
    pheno_inputs$mPath,
    pheno_inputs$cnPath
))
```

`pheno_inputs` is a list that bundles both the temporary files and the
corresponding in-memory objects used in this stage. `names(pheno_inputs)` shows
that it contains the temporary directory, the phenotype table, the three saved
matrix files (`phenoPath`, `betaPath`, `mPath`, and `cnPath`), and precomputed
helper objects such as `metricsData`, `timepointData`, `combinedData`, and
`clockFoundation`.

```{r preprocessingPheno-local, message = FALSE, warning = FALSE}
phenoResult <- preprocessingPheno(
    phenoFile = pheno_inputs$phenoPath,
    betaPath = pheno_inputs$betaPath,
    mPath = pheno_inputs$mPath,
    cnPath = pheno_inputs$cnPath,
    SampleID = "Sample_Name",
    timeVar = "Timepoint",
    timepoints = "1,2",
    combineTimepoints = "1,2",
    outputPheno = file.path(pheno_inputs$tempDir, "data", "preprocessingPheno"),
    outputRData = file.path(
        pheno_inputs$tempDir,
        "rData",
        "preprocessingPheno",
        "metrics"
    ),
    outputRDataMerge = file.path(
        pheno_inputs$tempDir,
        "rData",
        "preprocessingPheno",
        "mergeData"
    ),
    sexColumn = "Sex",
    outputLogs = file.path(pheno_inputs$tempDir, "logs"),
    outputDir = file.path(pheno_inputs$tempDir, "clockFoundation"),
    verbose = FALSE,
    logs = FALSE,
    saveOutputs = FALSE
)

class(phenoResult)
names(phenoResult)
names(phenoResult$combinedData)
head(phenoResult$combinedData$phenoMethylation[, 1:6])
```

The returned object has class `dnaEPICO_preprocessingPheno`, and
`names(phenoResult)` shows the main returned components. The
`combinedData` element groups the combined longitudinal objects, and
`phenoResult$combinedData$phenoMethylation` is the key downstream table for
modeling. By default it combines phenotype columns and CpG beta values in a
single data frame that can be passed to the GLM or LME functions. Set
`methylationScale = "m"` for M-values or `"cn"` for copy-number values.

Arguments:

- `phenoFile`, `betaPath`, `mPath`, and `cnPath` define the phenotype table and
    the three methylation matrices that are aligned in this step.
- `SampleID = "Sample_Name"` is the key used to match samples across the loaded
    objects.
- `timeVar = "Timepoint"`, `timepoints = "1,2"`, and
    `combineTimepoints = "1,2"` define the separate exports and the combined
    longitudinal object.
- `outputPheno`, `outputRData`, `outputRDataMerge`, and `outputDir` still need
    to be specified even when `saveOutputs = FALSE`, because they define the
    canonical output structure for the function.
- `sexColumn = "Sex"` identifies the sex variable used by the helper functions.
- `verbose = FALSE`, `logs = FALSE`, and `saveOutputs = FALSE` keep the example
    focused on the returned object rather than file writing.

The Clock Foundation export objects are returned in the same result, which
makes it possible to inspect the export-ready tables before deciding whether
they should be used in a reporting or export workflow.

```{r preprocessingPheno-local-inspect, message = FALSE, warning = FALSE}
names(phenoResult$clockFoundation)
head(phenoResult$clockFoundation$phenoCF)
```

The `clockFoundation` element contains the export-ready objects, and
`phenoResult$clockFoundation$phenoCF` is the phenotype table prepared for
[Clock Foundation](https://dnamage.clockfoundation.org/user) use.

## Variables used in GLM and LME models

The phenotype-methylation tables returned by `preprocessingPheno()` can be used
by `methylationGLM()` or `methylationLME()`. `factorVars` identifies categorical
fixed-effect variables, while `scaleVars` identifies numeric fixed-effect
variables to centre and divide by their sample standard deviations before model
fitting. For example, `factorVars = "Sex,TreatmentGroup"` treats those variables
as categorical and `scaleVars = "Age,PC1"` standardises `Age` and `PC1`.

For GLM and LME analyses, the selected variables must be numeric, have at least
two finite observations, and have non-zero sample variance. A variable cannot be
listed in both `factorVars` and `scaleVars`. The applied centres and sample
standard deviations are stored in the annotated workbook metadata.

# Summary

In interactive local use, the most important habit is to inspect the returned
objects after each step. In practice, the following checks are enough to
confirm that the workflow is behaving as expected:

- `class(result)` to confirm which function produced the object
- `names(result)` to see the main return elements
- `head()` on the phenotype-like tables
- `class()` or `dim()` on the methylation objects and matrices

The next vignette shows how to use `saveOutputs = TRUE` and the Makefile
workflow. The file-based vignette also describes the regression and
mixed-effects model structures used downstream, including how phenotype-specific
PRS terms are added through `prsMap`.

# Basics

Date the vignette was generated.

```{r, echo = FALSE}
Sys.time()
```

Wallclock time spent generating the vignette.

```{r, echo = FALSE}
totalTime <- diff(c(startTime, Sys.time()))
round(totalTime, digits = 3)
```

`R` session information.

```{r, echo = FALSE}
sessionInfo()
```

## Asking for help

As package developers, we try to explain clearly how to use our packages and in
which order to use the functions. But `R` and `Bioconductor` have a steep
learning curve, so it is critical to learn where to ask for help. We would like
to highlight the [Bioconductor support site](https://support.bioconductor.org/)
as the main resource for getting help. Please remember to use the `dnaEPICO`
tag and check the
[older posts](https://support.bioconductor.org/tag/dnaEPICO/). If you want to
receive help, please provide a small reproducible example and your session
information so the source of the problem can be tracked efficiently.

# References

---
title: "Pipeline Use of dnaEPICO"
author:
  - name: Paul Ruiz
    affiliation:
      - Queensland University of Technology
    email: ruizpint@qut.edu.au
output:
  BiocStyle::html_document:
    self_contained: yes
    toc: true
    toc_float: true
    toc_depth: 2
date: "`r Sys.Date()`"
package: dnaEPICO
bibliography: dnaEPICO.bib
vignette: >
  %\VignetteIndexEntry{3. Pipeline Use of dnaEPICO}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
---

```{r, include = FALSE}
startTime <- Sys.time()
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>",
    crop = NULL
)

library(dnaEPICO)

can_run_preprocessing <- requireNamespace("minfiData", quietly = TRUE) &&
    requireNamespace("IlluminaHumanMethylation450kmanifest", quietly = TRUE) &&
    requireNamespace("IlluminaHumanMethylation450kanno.ilmn12.hg19", quietly = TRUE)

can_run_sva <- can_run_preprocessing
can_run_glm <- requireNamespace(
    "IlluminaHumanMethylation450kanno.ilmn12.hg19",
    quietly = TRUE
)
can_run_lme <- can_run_glm && requireNamespace("lmerTest", quietly = TRUE)
```

# Introduction

This vignette describes the file-based route in `dnaEPICO`. In this mode, the
same main functions still return structured result objects, but they also
write files to disk when `saveOutputs = TRUE`. This route is appropriate when
you want a reproducible folder structure, Makefile-driven execution, or a
pipeline that can be resumed on a local workstation or an HPC system.

# Citation

The `r Biocpkg("dnaEPICO")` package uses methods from several external
packages. Because no single manuscript describes all components, the guidance
below explains how to cite `dnaEPICO` depending on the functions you use.

This citation guidance is adapted from the vignettes and user guides of
[minfi](https://github.com/hansenlab/minfi/blob/devel/vignettes/minfi.Rmd),
[ENmix](https://github.com/xuz1/ENmix/blob/master/vignettes/ENmix.Rmd), and
[wateRmelon](https://github.com/schalkwyk/wateRmelon/blob/master/vignettes/references.bib).

- If you use `make f3 MODEL=model1`, please cite @minfi; @minfiEPIC; @ENmix;
    @wateRmelon; @SWAN; @funnorm; @noob; @quantile; @ewastools.
- If you use `make f4 MODEL=model1`, please cite @minfi; @minfiEPIC; @ENmix;
    @wateRmelon; @SWAN; @funnorm; @noob; @quantile; @ewastools; @glm2.
- If you use `make f3lme MODEL=model1`, please cite @minfi; @minfiEPIC;
    @ENmix; @wateRmelon; @SWAN; @funnorm; @noob; @quantile; @ewastools; @lme4.
- If you use `make all MODEL=model1`, please cite @minfi; @minfiEPIC; @ENmix;
    @wateRmelon; @SWAN; @funnorm; @noob; @quantile; @ewastools; @glm2; @lme4.

# Required knowledge

`r Biocpkg("dnaEPICO")` is built on core Bioconductor infrastructure for
high-dimensional genomic data, with a focus on Illumina DNA methylation arrays.
To read this vignette, we assume that you are already familiar with the general
DNA methylation workflow. If not, we recommend first reading
[this tutorial](https://paulyrp.github.io/2025-cpgpneurogenomics-workshop/tutorial.html),
which provides a practical introduction to the main concepts and analysis steps.

Preprocessing and quality control are performed using established Bioconductor
tools, including `r Biocpkg("minfi")`, `r Biocpkg("ENmix")`, and
`r Biocpkg("wateRmelon")`. Downstream statistical modelling uses CpG-wise
generalised linear models and longitudinal mixed-effects models fitted with
`lmerTest`/`lme4` or `nlme`. The `nlme` engine can include AR1 or CAR1
within-participant residual correlation. Users are expected to have basic
familiarity with R, command-line execution, and Illumina IDAT file structures.

If you are asking yourself the question "Where do I start using
Bioconductor?", you might be interested in
[this blog post](https://www.bioconductor.org/install/).

## Probe-exclusion reference files

The lists of excluded probes depend on the Illumina methylation-array platform.

- For **Illumina HumanMethylationEPIC v2.0**, use the cross-reactive probe-exclusion file from [Peters et al. (2024)](https://springernature.figshare.com/articles/dataset/Additional_file_8_of_Characterisation_and_reproducibility_of_the_HumanMethylationEPIC_v2_0_BeadChip_for_DNA_methylation_profiling/26689706?file=48531323).

- For **Illumina MethylationEPIC**, also known as the **850k array**, use the probe-exclusion resources from [Pidsley et al. (2016)](https://springernature.figshare.com/articles/dataset/Additional_file_1_Table_S1_of_Critical_evaluation_of_the_Illumina_MethylationEPIC_BeadChip_microarray_for_whole-genome_DNA_methylation_profiling/4405946?file=7132334). The supplementary files commonly used together are [13059_2016_1066_MOESM1_ESM.csv](https://static-content.springer.com/esm/art%3A10.1186%2Fs13059-016-1066-1/MediaObjects/13059_2016_1066_MOESM1_ESM.csv), [13059_2016_1066_MOESM4_ESM.csv](https://static-content.springer.com/esm/art%3A10.1186%2Fs13059-016-1066-1/MediaObjects/13059_2016_1066_MOESM4_ESM.csv), [13059_2016_1066_MOESM5_ESM.csv](https://static-content.springer.com/esm/art%3A10.1186%2Fs13059-016-1066-1/MediaObjects/13059_2016_1066_MOESM5_ESM.csv), and [13059_2016_1066_MOESM6_ESM.csv](https://static-content.springer.com/esm/art%3A10.1186%2Fs13059-016-1066-1/MediaObjects/13059_2016_1066_MOESM6_ESM.csv).

- For **Illumina HumanMethylation450k**, use the cross-reactive and polymorphic probe resources from [Chen et al. (2013)](https://figshare.com/articles/dataset/methylAnalysisDataV3_tar_gz/4800970?file=7896205).

- Multiple probe-exclusion files can be supplied as a semicolon-separated value in `probeExclusionPath`. `dnaEPICO` reads probe IDs from each file, uses `probeExclusionIdColumn` when supplied, or otherwise auto-detects common probe-ID columns such as `ProbeID`, `TargetID`, `IlmnID`, and `Name`. The unique union of all probe IDs is then used to filter the normalised object.

- For **EPICv2**, setting `useEpicV2Manifest = TRUE` also retrieves the expanded Peters et al. manifest from AnnotationHub resource [AH116484](https://bioconductor.org/packages/release/data/annotation/vignettes/EPICv2manifest/inst/doc/EPICv2manifest.html). Probes flagged in selected manifest columns are added to the same exclusion set. By default, probes flagged by `CH_WGBS_evidence`, `CH_BLAT`, or `MissingPos` are removed, while `MismatchPos` is retained unless explicitly enabled.

# Installation

The `r Biocpkg("dnaEPICO")` package depends on `r Biocpkg("minfi")`,
`r Biocpkg("ENmix")`, and `r Biocpkg("wateRmelon")`, among others. Install
dependencies with BiocManager:

```{r, eval = FALSE}
if (!requireNamespace("BiocManager", quietly = TRUE)) {
    install.packages("BiocManager")
}

BiocManager::install("dnaEPICO")

BiocManager::valid()
```

```{r, eval = FALSE, message = FALSE}
library("dnaEPICO")
```

# Extract the example Makefile

The package includes a template Makefile that can be copied into a project
directory and adapted to local paths and model definitions.

```{r makefile-example}
makefilePath <- extractMake(destDir = tempdir(), overwrite = TRUE)
basename(makefilePath)
```

After extraction, the file should be edited so that the project-specific paths,
model labels, and cluster settings match the target analysis environment.

Arguments:

- `destDir = tempdir()` writes the template to a temporary directory so the
    example does not modify the working tree.
- `overwrite = TRUE` allows the template to be replaced if it already exists in
    that temporary location.

# Write preprocessingMinfiEwasWater outputs to disk

The file-based route usually starts with `preprocessingMinfiEwasWater()`. This
step writes the filtered `RGSet`, normalized metric matrices, quality-control
figures, and the `phenoLC.csv` file that will be consumed by `svaEnmix()` and
`preprocessingPheno()`.

### Create example input files

The next chunk shows how to recreate the temporary files used in this example.
It writes a small phenotype table to a temporary directory, copies a small set
of IDAT files, and records the probe-exclusion reference used during probe
filtering.

```{r preprocessing-pipeline-inputs, eval = can_run_preprocessing, message = FALSE, warning = FALSE}
preprocessing_inputs <- dnaEPICO:::exampleMinfiIdatInputsDnaEpico(n = 6L)

names(preprocessing_inputs)
preprocessing_inputs$tempDir
basename(preprocessing_inputs$phenoFile)
head(basename(list.files(preprocessing_inputs$idatFolder, full.names = TRUE)), 4)
basename(preprocessing_inputs$probeExclusionPath)
```

`preprocessing_inputs` is a small list returned by the internal example helper.
`names(preprocessing_inputs)` shows its main elements: `tempDir` is the
temporary working directory, `idatFolder` contains the copied example IDAT
files, `phenoFile` is the phenotype table used by the function, `targets` is the
same phenotype information already loaded into memory, `arrayType` and
`annotationVersion` describe the array platform, and `probeExclusionPath`
points to the probe-exclusion reference used during probe filtering.

```{r preprocessing-pipeline, eval = can_run_preprocessing, message = FALSE, warning = FALSE}
preprocessPipelineResult <- preprocessingMinfiEwasWater(
    phenoFile = preprocessing_inputs$phenoFile,
    idatFolder = preprocessing_inputs$idatFolder,
    outputLogs = file.path(preprocessing_inputs$tempDir, "logs"),
    nSamples = 6,
    SampleID = "Sample_Name",
    arrayType = preprocessing_inputs$arrayType,
    annotationVersion = preprocessing_inputs$annotationVersion,
    scriptLabel = "preprocessingMinfiEwasWater",
    baseDataFolder = file.path(preprocessing_inputs$tempDir, "rData"),
    figureBaseDir = file.path(preprocessing_inputs$tempDir, "figures"),
    detPThreshold = 1,
    normMethods = "quantile",
    sexColumn = "Sex",
    removeSexMismatch = FALSE,
    pvalThreshold = 1,
    chrToRemove = "",
    snpsToRemove = "SBE",
    mafThreshold = 1,
    probeExclusionPath = preprocessing_inputs$probeExclusionPath,
    plotGroupVar = "Sex",
    lcRef = "saliva",
    phenoOrder = "Sample_Name;Sex;Basename;Sentrix_ID;Sentrix_Position",
    lcPhenoDir = file.path(
        preprocessing_inputs$tempDir,
        "data",
        "preprocessingMinfiEwasWater"
    ),
    display = FALSE,
    verbose = FALSE,
    logs = TRUE,
    saveOutputs = TRUE
)

preprocess_paths <- c(
    phenoLC = file.path(
        preprocessing_inputs$tempDir,
        "data",
        "preprocessingMinfiEwasWater",
        "phenoLC.csv"
    ),
    rgset = file.path(
        preprocessing_inputs$tempDir,
        "rData",
        "preprocessingMinfiEwasWater",
        "objects",
        "RGSet.RData"
    ),
    beta = file.path(
        preprocessing_inputs$tempDir,
        "rData",
        "preprocessingMinfiEwasWater",
        "metrics",
        "beta_NomFilt_MSetF_Flt_Rxy_Ds_Rc.RData"
    ),
    qc = file.path(
        preprocessing_inputs$tempDir,
        "figures",
        "preprocessingMinfiEwasWater",
        "qc",
        "quality_control(MSet).tiff"
    )
)

class(preprocessPipelineResult)
basename(preprocess_paths)
file.exists(preprocess_paths)
```

The returned object still has class `dnaEPICO_preprocessingMinfiEwasWater`, but
the printed vectors in this chunk focus on the written outputs. In
`preprocess_paths`, `phenoLC` is the phenotype table with estimated cell
proportions, `rgset` is the filtered `RGSet`, `beta` is the saved beta matrix
used later by phenotype preparation, and `qc` is an example quality-control
figure. `basename(preprocess_paths)` shows the file names, while
`file.exists(preprocess_paths)` confirms that each one was written successfully.

Arguments:

- `phenoFile` and `idatFolder` define the phenotype table and the IDAT folder
    used to start the pipeline.
- `SampleID = "Sample_Name"` is the key used to align phenotype rows with the
    methylation objects.
- `arrayType` and `annotationVersion` define the array manifest and annotation
    used by `minfi`.
- `baseDataFolder`, `figureBaseDir`, and `lcPhenoDir` define where serialized
    objects, figures, and `phenoLC.csv` are written.
- `detPThreshold`, `normMethods`, `pvalThreshold`, `chrToRemove`,
    `snpsToRemove`, `mafThreshold`, and `probeExclusionPath` define the main
    preprocessing and filtering choices.
- `sexColumn` identifies reported sex for comparison with methylation-predicted
    sex. `removeSexMismatch = FALSE` retains and reports confirmed mismatches;
    `TRUE` removes samples only when both values are known and disagree. Missing
    or unknown sex values are retained.
- `plotGroupVar`, `lcRef`, and `phenoOrder` control QC grouping, cell-type
    estimation, and the structure of the exported phenotype table.
- `saveOutputs = TRUE` is what turns this function into the first file-producing
    step of the pipeline.

# Write SVA outputs to disk

The next example shows `svaEnmix()` in file-writing mode. The function still
returns a structured result object, but the `savedFiles` element now records the
main output files written to disk.

### Reuse the temporary files written in Step 1

In the file-based route, `svaEnmix()` consumes the `phenoLC.csv` and
`RGSet.RData` files written by `preprocessingMinfiEwasWater()`. The next chunk
shows how those temporary file paths are reconstructed locally.

```{r sva-pipeline-inputs, eval = can_run_sva, message = FALSE, warning = FALSE}
sva_targets_file <- file.path(
    preprocessing_inputs$tempDir,
    "data",
    "preprocessingMinfiEwasWater",
    "phenoLC.csv"
)
sva_rgset_file <- file.path(
    preprocessing_inputs$tempDir,
    "rData",
    "preprocessingMinfiEwasWater",
    "objects",
    "RGSet.RData"
)

basename(c(sva_targets_file, sva_rgset_file))
file.exists(c(sva_targets_file, sva_rgset_file))
```

`sva_targets_file` and `sva_rgset_file` point to the Step 1 outputs:
`phenoLC.csv`, containing the phenotype table and cell-composition estimates,
and `RGSet.RData`, containing the filtered methylation object.

```{r sva-pipeline, eval = can_run_sva, message = FALSE, warning = FALSE}
svaPipelineResult <- svaEnmix(
    phenoFile = sva_targets_file,
    rgsetData = sva_rgset_file,
    outputLogs = file.path(preprocessing_inputs$tempDir, "logs"),
    SampleID = "Sample_Name",
    arrayType = "IlluminaHumanMethylation450k",
    annotationVersion = "ilmn12.hg19",
    SentrixIDColumn = "Sentrix_ID",
    SentrixPositionColumn = "Sentrix_Position",
    ctrlSvaPercVar = 0.90,
    ctrlSvaFlag = 1,
    scriptLabel = "svaEnmix",
    dataBaseDir = file.path(preprocessing_inputs$tempDir, "data"),
    rBaseDir = file.path(preprocessing_inputs$tempDir, "rData"),
    figureBaseDir = file.path(preprocessing_inputs$tempDir, "figures"),
    display = FALSE,
    verbose = FALSE,
    logs = TRUE,
    saveOutputs = TRUE
)

class(svaPipelineResult$savedFiles)
names(svaPipelineResult$savedFiles)
basename(unlist(svaPipelineResult$savedFiles, use.names = FALSE))
```

The returned `savedFiles` object records the main paths written by this step.
`names(svaPipelineResult$savedFiles)` identifies the output groups: `svaRData`
is the serialized surrogate-variable matrix, `svaCSV` is the CSV version of
that matrix, `phenoWithSva` is the same `phenoLC.csv` after the SVA columns are
appended, and `dataDir` / `rDir` are the parent directories used for the saved
outputs. The `basename(...)` call shows the file names generated from those
paths. The phenotype is validated, written to a temporary file, read back, and
then replaced with rollback protection. Sample identifiers, row order,
cell-composition columns, and all other existing columns must remain unchanged.
Existing PC names are rejected before replacement.

Arguments:

- `phenoFile` and `rgsetData` identify the phenotype table and saved `RGSet`
    produced by the earlier preprocessing step.
- `SampleID`, `SentrixIDColumn`, and `SentrixPositionColumn` specify how
    samples and array positions are identified in the phenotype table.
- `ctrlSvaPercVar = 0.90` keeps enough control-derived surrogate variables to
    explain 90% of the control-probe variance.
- `ctrlSvaFlag = 1` enables the ENmix control-based SVA workflow.
- `dataBaseDir` and `rBaseDir` define where file outputs are written when
    `saveOutputs = TRUE`.
- `display = FALSE` and `verbose = FALSE` keep the console output quiet, while
    `logs = TRUE` writes the log files used by the report.

# Write preprocessingPheno outputs to disk

`preprocessingPheno()` is the step that usually prepares the largest number of
pipeline-ready files because it writes timepoint-specific tables, combined
longitudinal tables, and the Clock Foundation export inputs.

### Create example input files

The next chunk shows how to recreate the temporary phenotype and matrix files
used in this example. The helper writes a phenotype table plus aligned beta,
M-value, and copy-number objects to a temporary directory.

```{r preprocessingPheno-pipeline-inputs, message = FALSE, warning = FALSE}
pheno_inputs <- dnaEPICO:::examplePreprocessingPhenoStateDnaEpico()

names(pheno_inputs)
pheno_inputs$tempDir
basename(c(
    pheno_inputs$phenoPath,
    pheno_inputs$betaPath,
    pheno_inputs$mPath,
    pheno_inputs$cnPath
))
```

`pheno_inputs` is a list that bundles the temporary files and the aligned
in-memory objects used in this stage. `names(pheno_inputs)` shows that it
contains the temporary directory, the phenotype table, the saved beta, M-value,
and copy-number files, and precomputed helper objects such as `metricsData`,
`timepointData`, `combinedData`, and `clockFoundation`.

```{r preprocessingPheno-pipeline, message = FALSE, warning = FALSE}
phenoPipelineResult <- preprocessingPheno(
    phenoFile = pheno_inputs$phenoPath,
    betaPath = pheno_inputs$betaPath,
    mPath = pheno_inputs$mPath,
    cnPath = pheno_inputs$cnPath,
    SampleID = "Sample_Name",
    timeVar = "Timepoint",
    timepoints = "1,2",
    combineTimepoints = "1,2",
    outputPheno = file.path(pheno_inputs$tempDir, "data", "preprocessingPheno"),
    outputRData = file.path(
        pheno_inputs$tempDir,
        "rData",
        "preprocessingPheno",
        "metrics"
    ),
    outputRDataMerge = file.path(
        pheno_inputs$tempDir,
        "rData",
        "preprocessingPheno",
        "mergeData"
    ),
    sexColumn = "Sex",
    outputLogs = file.path(pheno_inputs$tempDir, "logs"),
    outputDir = file.path(pheno_inputs$tempDir, "clockFoundation"),
    verbose = FALSE,
    logs = TRUE,
    saveOutputs = TRUE
)

names(phenoPipelineResult$savedFiles)
names(phenoPipelineResult$savedFiles$timepoints)
basename(unlist(phenoPipelineResult$savedFiles$timepoints[["1"]]))
basename(unlist(phenoPipelineResult$savedFiles[c(
    "combinedPheno",
    "combinedPhenoMethylation",
    "betaCSV",
    "phenoCF"
)]))
```

These outputs are the files most commonly consumed by downstream modeling
functions. `names(phenoPipelineResult$savedFiles)` shows the high-level output
groups returned by the function. `names(phenoPipelineResult$savedFiles$timepoints)`
lists the available timepoint-specific subsets. The `basename(...)` calls then
show examples of the concrete files written for one timepoint and for the
combined outputs. In this structure, `combinedPhenoMethylation` is the merged
phenotype-plus-methylation object that becomes the direct input to the GLM and
LME functions. It defaults to beta values, but can be written as M-values or
copy-number values when `methylationScale` is changed. The `betaCSV`, `betaZIP`,
and `phenoCF` outputs remain beta-based Clock Foundation exports.

Arguments:

- `phenoFile`, `betaPath`, `mPath`, and `cnPath` point to the phenotype table
    and the aligned methylation matrices from the previous workflow stages.
- `SampleID = "Sample_Name"` defines the sample key used during alignment.
- `timeVar = "Timepoint"`, `timepoints = "1,2"`, and
    `combineTimepoints = "1,2"` determine the timepoint-specific outputs and the
    combined longitudinal object.
- `methylationScale = "beta"` selects the merged modeling scale. Use `"m"` for
    M-values or `"cn"` for copy-number values. Clock Foundation exports continue
    to use beta values.
- `outputPheno`, `outputRData`, `outputRDataMerge`, and `outputDir` define the
    directories used for phenotype exports, metric objects, merged objects, and
    Clock Foundation inputs.
- `saveOutputs = TRUE` is what makes this step useful in a file-based pipeline,
    because later modeling steps consume these written files.

# Write GLM model outputs to disk

The modeling function also separate in-memory results from file outputs. The
next example runs a small GLM analysis and prints the names of the files written
to disk.

### Create example input files

The next chunk shows how to recreate the temporary merged phenotype-plus-beta
input file used by the GLM example.

```{r glm-pipeline-inputs, eval = can_run_glm, message = FALSE, warning = FALSE}
glm_inputs <- dnaEPICO:::exampleMethylationGLMStateDnaEpico()

names(glm_inputs)
glm_inputs$tempDir
basename(glm_inputs$inputPath)
file.exists(glm_inputs$inputPath)
```

`glm_inputs` is a list returned by the GLM example helper. `names(glm_inputs)`
shows that it contains the temporary directory, the saved merged
phenotype-plus-beta input file (`inputPath`), and the corresponding in-memory
objects produced during helper construction: `preparedData`, `modelResults`, and
`modelSummaries`.

```{r glm-pipeline, eval = can_run_glm, message = FALSE, warning = FALSE}
glmPipelineResult <- methylationGLM(
    inputPheno = glm_inputs$inputPath,
    phenotypes = "status",
    covariates = "sex,age",
    factorVars = "status,sex",
    scaleVars = "age",
    cpgLimit = 2,
    nCores = 1,
    outputLogs = file.path(glm_inputs$tempDir, "logs"),
    outputRData = file.path(glm_inputs$tempDir, "rData", "models"),
    outputPlots = file.path(glm_inputs$tempDir, "figures"),
    significantCpGDir = file.path(glm_inputs$tempDir, "significant"),
    summaryTxtDir = file.path(glm_inputs$tempDir, "summaries"),
    summaryPval = 1,
    significantCpGPval = 1,
    annotationPackage = "IlluminaHumanMethylation450kanno.ilmn12.hg19",
    annotationCols = "Name,chr,pos",
    annotatedGLMOut = file.path(glm_inputs$tempDir, "annotated"),
    reportAssetsDir = file.path(
        preprocessing_inputs$tempDir,
        "reports", "example", "assets", "results", "glm_results"
    ),
    display = FALSE,
    verbose = FALSE,
    logs = TRUE,
    saveOutputs = TRUE
)

class(glmPipelineResult$savedFiles)
names(glmPipelineResult$savedFiles)
```

The returned `savedFiles` object records the groups of outputs written by the
GLM step. `names(glmPipelineResult$savedFiles)` typically includes the model
files, summary files, diagnostic plot files, significant-CpG exports, summary
text files, and the annotated results workbook. This is the object to inspect when
you want to confirm which modeling outputs were written and how they are
grouped.

Arguments:

- `inputPheno` points to the merged phenotype-plus-beta object generated by
    `preprocessingPheno()`.
- `phenotypes` lists the outcome variables that will be modeled one at a time.
- `covariates` lists the adjustment variables included in every model.
- `factorVars` identifies the categorical variables that should be treated as
    factors before model fitting.
- `scaleVars` identifies numeric fixed-effect variables to centre and divide by
    their sample standard deviations before fitting.
- `cpgLimit = 2` keeps the example fast by fitting only two CpG models.
- `nCores = 1` keeps the example deterministic and lightweight; production HPC
    runs typically use higher values.
- `annotationPackage` and `annotationCols` control which array annotation is
    merged into the final summary tables.
- `outputRData`, `outputPlots`, `significantCpGDir`, `summaryTxtDir`, and
    `annotatedGLMOut` define where model objects, plots, summaries, and the
    annotated GLM workbook are written.
- `reportAssetsDir` writes the compressed report table directly into the report
    project, leaving the annotated workbook directory free of report sidecars.
- `saveOutputs = TRUE` enables the file-based pipeline behaviour expected by
    Makefile and HPC use.

# Write LME model outputs to disk

The longitudinal modeling step follows the same pattern as the GLM function, but
the saved outputs now correspond to mixed-effects models, longitudinal summary
tables, and interaction-specific result files.

### Create example input files

The next chunk shows how to recreate the temporary combined longitudinal
phenotype-plus-beta input file used by the LME example.

```{r lme-pipeline-inputs, eval = can_run_lme, message = FALSE, warning = FALSE}
lme_inputs <- dnaEPICO:::exampleMethylationLMEStateDnaEpico()

names(lme_inputs)
lme_inputs$tempDir
basename(lme_inputs$inputPath)
file.exists(lme_inputs$inputPath)
```

`lme_inputs` plays the same role for the longitudinal example. The helper
returns a list with the temporary directory, the saved combined longitudinal
input file (`inputPath`), and the in-memory objects used to build that example:
`preparedData`, `modelResults`, and `modelSummaries`.

```{r lme-pipeline, eval = can_run_lme, message = FALSE, warning = FALSE}
lmePipelineResult <- methylationLME(
    inputPheno = lme_inputs$inputPath,
    outputLogs = file.path(lme_inputs$tempDir, "logs"),
    outputRData = file.path(lme_inputs$tempDir, "rData", "models"),
    outputPlots = file.path(lme_inputs$tempDir, "figures"),
    personVar = "person",
    timeVar = "Timepoint",
    phenotypes = "score",
    covariates = "sex",
    factorVars = "sex",
    scaleVars = "score",
    cpgLimit = 2,
    nCores = 1,
    summaryPval = 1,
    saveSignificantInteractions = TRUE,
    significantInteractionDir = file.path(
        lme_inputs$tempDir,
        "results",
        "cpgs",
        "methylationLME"
    ),
    significantInteractionPval = 1,
    saveTxtSummaries = TRUE,
    summaryTxtDir = file.path(
        lme_inputs$tempDir,
        "results",
        "summary",
        "methylationLME"
    ),
    annotationPackage = "IlluminaHumanMethylation450kanno.ilmn12.hg19",
    annotationCols = "Name,chr,pos",
    annotatedLMEOut = file.path(lme_inputs$tempDir, "annotated"),
    reportAssetsDir = file.path(
        preprocessing_inputs$tempDir,
        "reports", "example", "assets", "results", "lme_results"
    ),
    display = FALSE,
    verbose = FALSE,
    logs = TRUE,
    saveOutputs = TRUE
)

class(lmePipelineResult$savedFiles)
names(lmePipelineResult$savedFiles)
```

The returned `savedFiles` object records the groups of outputs written by the
LME step. `names(lmePipelineResult$savedFiles)` typically includes the model
files, summary files, diagnostic plot files, summary text files, significant
interaction exports, and the annotated longitudinal results table. This is the
object to inspect when you want to confirm which longitudinal modeling outputs
were written and how they are grouped.

Arguments:

- `inputPheno` points to the combined longitudinal phenotype-plus-beta object.
- `personVar` defines the participant identifier used for the random effect.
- `timeVar` defines the repeated-measures time variable used for longitudinal
    summaries and preprocessing checks.
- `phenotypes`, `covariates`, and optional `interactionTerm` define the
    fixed-effect portion of the mixed model.
- `cpgLimit = 2` keeps the example short by fitting only two CpG models.
- `nCores = 1` keeps the example lightweight; production HPC runs usually use
    higher parallel settings.
- `saveSignificantInteractions`, `significantInteractionDir`, and
    `significantInteractionPval` control the export of interaction-specific
    results.
- `summaryTxtDir` defines where plain-text model summaries are written.
- `annotationPackage`, `annotationCols`, and `annotatedLMEOut` define the
    annotation resources and output location for the final longitudinal summary
    workbook.
- `reportAssetsDir` writes the compressed longitudinal report table directly
    into the report project rather than beside the workbook.
- `saveOutputs = TRUE` enables the file-based longitudinal workflow expected by
    Makefile and HPC use.

## Generate the report from the example outputs

After the file-based examples above have been run, `dnamReport()` can assemble
the phenotype table, ENmix control figures, quality-control figures, batch
effect figures, metrics figures, model annotation tables, and logs into one
website report. The example below uses the concrete output paths created by the
previous chunks.

The preprocessing and SVA examples share `preprocessing_inputs$tempDir`, while
the GLM and LME examples write model outputs into their own temporary
directories. Because `dnamReport()` accepts one path per tab, those outputs can
be passed directly.

When running this section from an R session, `dnamReport()` still needs the
Quarto command line interface because the function renders the `.qmd` files into
the final website. You can check whether R can find Quarto with:

```{r quarto-check-r, eval = FALSE}
Sys.which("quarto")
system2("quarto", "--version")
```

If `Sys.which("quarto")` returns an empty string, install the Quarto command
line interface from the
[Quarto get-started page](https://quarto.org/docs/get-started/) before
rendering the report. The R package `quarto` provides R helper functions for an
existing Quarto installation; it does not replace the Quarto command line
interface used by `dnamReport()`.

If Quarto is installed but not on `PATH`, set the executable path before calling
`dnamReport()`:

```{r quarto-path-r, eval = FALSE}
Sys.setenv(QUARTO_BIN = "/full/path/to/quarto")
```

```{r dnamReport-pipeline-example, eval = FALSE}
report_log_dir <- file.path(preprocessing_inputs$tempDir, "logs", "report")
dir.create(report_log_dir, recursive = TRUE, showWarnings = FALSE)

report_logs <- c(
    file.path(
        preprocessing_inputs$tempDir,
        "logs",
        "log_preprocessingMinfiEwasWater.txt"
    ),
    file.path(pheno_inputs$tempDir, "logs", "log_preprocessingPheno.txt"),
    file.path(preprocessing_inputs$tempDir, "logs", "log_svaEnmix.txt"),
    file.path(glm_inputs$tempDir, "logs", "log_methylationGLM.txt"),
    file.path(lme_inputs$tempDir, "logs", "log_methylationLME.txt")
)

existing_logs <- report_logs[file.exists(report_logs)]
if (length(existing_logs)) {
    file.copy(existing_logs, report_log_dir, overwrite = TRUE)
}

reportResult <- dnamReport(
    outputDir = file.path(preprocessing_inputs$tempDir, "reports", "example"),
    phenoTab = file.path(
        preprocessing_inputs$tempDir,
        "data",
        "preprocessingMinfiEwasWater",
        "phenoLC.csv"
    ),
    enmixTab = file.path(
        preprocessing_inputs$tempDir,
        "figures",
        "preprocessingMinfiEwasWater",
        "enmix"
    ),
    qcTab = file.path(
        preprocessing_inputs$tempDir,
        "figures",
        "preprocessingMinfiEwasWater",
        "qc"
    ),
    svaTab = file.path(preprocessing_inputs$tempDir, "figures", "svaEnmix"),
    metricTab = file.path(
        preprocessing_inputs$tempDir,
        "figures",
        "preprocessingMinfiEwasWater",
        "metrics"
    ),
    glmTab = glmPipelineResult$savedFiles$annotatedGLM,
    lmeTab = lmePipelineResult$savedFiles$annotatedLME,
    logTab = report_log_dir,
    detPPath = file.path(
        preprocessing_inputs$tempDir,
        "rData",
        "preprocessingMinfiEwasWater",
        "qc",
        "detP_RGSet.RData"
    ),
    detPThreshold = 0.01,
    verbose = FALSE,
    logs = TRUE
)

reportResult$outputFile
```

## GLM model structure and PRS terms

For each CpG, `methylationGLM()` fits a regression model of the form:

$$
\text{Methylation}_{CpG_i} =
\beta_0 +
\beta_1 \cdot \text{Phenotype} +
\sum_{k = 1}^{K} \beta_{k + 1} \cdot \text{Covariate}_k +
\varepsilon
$$

where $\text{Methylation}_{CpG_i}$ is the methylation value at CpG $i$,
$\beta_0$ is the intercept, the phenotype term is the variable of interest, and
the remaining fixed effects represent the listed covariates.

If no polygenic risk score is included, a model can be read schematically as:

$$
\begin{aligned}
\text{Methylation}_{CpG_i} =\;&
\beta_0 +
\beta_1 \cdot \text{Phenotype} +
\beta_2 \cdot \text{Age} +
\beta_3 \cdot \text{Sex} +
\beta_4 \cdot \text{Ethnicity} \\
&+
\varepsilon
\end{aligned}
$$

When `prsMap` is supplied, a phenotype-specific PRS term is appended only for
the matching phenotype. For example, the mapping
`"Pheno1:PRS_1,Pheno2:PRS_2"` means that:

- models for `Pheno1` include `PRS_1`
- models for `Pheno2` include `PRS_2`
- phenotypes without a mapping are fit without a PRS term

For a phenotype mapped to a PRS, the model becomes:

$$
\text{Methylation}_{CpG_i} =
\beta_0 +
\beta_1 \cdot \text{Phenotype} +
\sum_{k = 1}^{K} \beta_{k + 1} \cdot \text{Covariate}_k +
\beta_{\text{PRS}} \cdot \text{PRS} +
\varepsilon
$$

If `interactionTerm` is supplied, the same framework is extended by adding the
requested interaction to the fixed-effect portion of the model.

## Longitudinal mixed-effects structure

The longitudinal function `methylationLME()` follows the same file-based
pattern, but uses a mixed-effects model with a participant-level random
intercept. In schematic form:

$$
\begin{aligned}
\text{Methylation}_{CpG_i} =\;&
\beta_0 +
\beta_1 \cdot \text{Phenotype} +
\beta_2 \cdot (\text{Phenotype} \times \text{Interaction}) +
\sum_{k = 1}^{K} \beta_{k + 2} \cdot \text{Covariate}_k +
\beta_{\text{PRS}} \cdot \text{PRS} +
b_{\text{person}} +
\varepsilon
\end{aligned}
$$

where $b_{\text{person}}$ is the subject-specific random intercept. As in the
GLM workflow, the PRS term is included only when the current phenotype is
matched in `prsMap`.

Common arguments in this longitudinal stage are:

- `inputPheno` for the combined longitudinal phenotype-plus-beta object
- `personVar` for the participant identifier used in the random effect
- `timeVar` for repeated-measures summaries and preprocessing checks
- `phenotypes`, `covariates`, and optional `interactionTerm` for the fixed-effect
    structure
- `prsMap` for phenotype-specific PRS terms
- `cpgLimit` and `nCores` for computational control
- `annotationPackage` and `annotationCols` for annotated summaries

# Makefile use

In a project directory, the Makefile route is simpler than calling each
function by hand. A minimal project layout looks like this:

```text
my_project/
    Makefile
    metadata/
    pheno_model1.csv
    data/
    preprocessingMinfiEwasWater/
        idats/
        <IDAT files>
```

The exported template should then be edited so that at least the model name,
phenotype path, and directory variables match the project. A minimal example is:

```make
MODEL = model1
PHENO_FILE = metadata/pheno_model1.csv
LOGS_DIR := $(abspath logs)
DATA_DIR = data
RDATA_DIR = rData
RESULTS_DIR = results
FIGURES_DIR = figures
```

With that structure in place, a typical sequence of commands looks like this:

```bash
make step1 MODEL=model1
make step2 MODEL=model1
make step3 MODEL=model1
make f3 MODEL=model1
make f4 MODEL=model1
make f3lme MODEL=model1
make all MODEL=model1
make status MODEL=model1
make clean MODEL=model1
```

These targets correspond to the grouped outputs defined in the exported rules:

- `make step1 MODEL=model1` runs preprocessing only.
- `make step2 MODEL=model1` runs the SVA step using the files written by Step 1.
- `make step3 MODEL=model1` runs phenotype preparation after Step 2 has safely
    appended PC columns to `phenoLC.csv`.
- `make f3 MODEL=model1` runs preprocessing, SVA, phenotype preparation, and the report target defined by the current rules.
- `make f4 MODEL=model1` runs the same route, adds the GLM step, and refreshes the report target.
- `make f3lme MODEL=model1` runs the same route, adds the longitudinal LME step, and refreshes the report target.
- `make all MODEL=model1` runs the full pipeline, including both modeling steps and the report.
- `make status MODEL=model1` checks which expected output files are already present.
- `make clean MODEL=model1` removes the files generated for the selected model while keeping the shared raw input area.

If several models are declared in `MODELS`, the grouped multi-model targets can
be used instead:

```bash
make f3_models
make f4_models
make f3lme_models
make models
```

These grouped targets loop over every model listed in `MODELS`:

- `make f3_models` runs the `f3` route for each declared model.
- `make f4_models` runs the `f4` route for each declared model.
- `make f3lme_models` runs the `f3lme` route for each declared model.
- `make models` runs the full `all` pipeline for each declared model.

Two maintenance targets are also useful during routine work:

```bash
make status MODEL=model1
make clean MODEL=model1
make clean_models
```

- `make status MODEL=model1` reports which expected outputs are already present for one model.
- `make clean MODEL=model1` removes the generated outputs for one model.
- `make clean_models` removes the generated outputs for every model listed in
    `MODELS`.

The exact target names and variables depend on the Makefile settings extracted
for the project. In practice, `saveOutputs = TRUE` is what makes the individual
steps compatible with this route because later steps consume the files written
by earlier ones. Step 2 creates
`data/<model>/svaEnmix/.sva_complete` only after the updated `phenoLC.csv` has
been validated and installed. Downstream targets depend on both files, so Make
reruns SVA when Step 1 replaces `phenoLC.csv` and skips SVA when the completed
inputs are unchanged.

## Report generation

The exported Makefile contains a report rule whose output is
`$(STEP6)/$(MODEL)/docs/index.html`. With the default directory settings, this
becomes `reports/<model>/docs/index.html`. Users usually do not need to call
that file target directly: the grouped targets above depend on it, so Make
refreshes the report automatically after the required pipeline outputs are
available.

For example, the `f3` route builds the report from the Data, ENmix QC, Quality
Control, Batch Effect, Metrics, Report, and Logs inputs. The `f4`, `f3lme`, and
`all` routes additionally make the GLM and LME tables available when the
corresponding model outputs exist.

The report target reads the same types of outputs generated in the examples
above, organised under the Makefile project layout: the phenotype table in
`data/`, quality-control and metric figures in `figures/`, the detection
P-value object in `rData/`, model annotation tables in `data/`, and workflow
logs in `logs/`.

The generated site is opened from the report target path:

```text
$(STEP6)/$(MODEL)/docs/index.html
```

Internally, the Makefile calls `dnamReport()` with one argument per tab, using
the project paths defined by the Makefile variables. The report target requires
the Quarto command line interface in the same environment where `make` is run.
If Quarto is not already available, install the Quarto command line interface
from the [Quarto get-started page](https://quarto.org/docs/get-started/), or
use the installation method provided by the local computing environment. Before
launching the report target, check:

```bash
quarto --version
```

If Quarto is installed outside `PATH`, export its executable path before running
`make`:

```bash
export QUARTO_BIN=/full/path/to/quarto
```

# Full exported Makefile

The full Makefile exported by `extractMake()` is shown below. This is the
template that should be adapted for a project before running the pipeline in the
target analysis environment.

```{r makefile-template-full, echo = FALSE, results = "asis"}
makefile_lines <- readLines(makefilePath, warn = FALSE)
cat("```make\n", paste(makefile_lines, collapse = "\n"), "\n```\n", sep = "")
```

# Makefile section guide

The exported Makefile is organised into sections so that project-specific
choices can be edited without changing the pipeline rules themselves.

## User configuration

- `MAKEFLAGS += --output-sync` keeps parallel Make output grouped by recipe,
    which makes logs easier to read.

## Models selection and parallel run

- `MODEL` defines a single pipeline configuration to run.
- `MODELS` defines a set of configurations that can be launched in parallel by
    the grouped Make targets.

In single-model use, commands such as `make f4 MODEL=model1` run one workflow.
In multi-model use, commands such as `make f4_models` evaluate the same template
for each name listed in `MODELS`.

## Per-model overrides

- `PHENO_FILE` is reassigned according to the selected `MODEL`.

This block is useful when different cohorts, case definitions, or phenotype
encodings should reuse the same analysis pipeline without duplicating the full
Makefile.

## Directories

- `LOGS_DIR`, `DATA_DIR`, `RDATA_DIR`, `RESULTS_DIR`, and `FIGURES_DIR` define
    the top-level folder structure.
- `STEP1` to `STEP6` define the canonical names of the major pipeline stages.
- `METRICS_DIR`, `IDAT_DIR`, `OBJ_DIR`, `MERGE_DIR`, `MODEL_DIR`, `CPG_DIR`,
    `SUMMARY_DIR`, `ENMIX_DIR`, `SVA_DIR`, and `QC_DIR` define the subdirectory
    names used by the rules file.
- `R_DIR` optionally points to a custom R library path on shared systems. When
    set to `NULL`, the default library path is used.

These variables control where files are written and how paths are composed
across all stages of the pipeline.

## Global parameters

These parameters are shared across multiple workflow steps.

### Step 1 to Step 5 shared parameters

- `PHENO_FILE`: input phenotype table used to start the workflow.
- `SEP_TYPE`: optional separator used when reading text files; `NULL` uses the
    default comma separator.
- `SAMPLE_ID`: phenotype column used to align samples.
- `N_SAMPLES`: optional sample subset for testing; `NA` uses all samples.
- `ARRAY_TYPE` and `ANNOTATION_VERSION`: array platform and annotation build.
- `IDAT_FORCE`: whether `minfi::read.metharray.exp()` should force reading
    selected IDAT files with different internal array sizes; the default is
    `FALSE`.
- `TIFF_WIDTH`, `TIFF_HEIGHT`, and `TIFF_RES`: default plot dimensions and
    resolution.
- `SEX_COLUMN`: phenotype column used for sex-based checks and prediction.
- `SENTRIX_ID_COLUMN` and `SENTRIX_POSITION_COLUMN`: chip-position metadata
    used by the SVA stage.
- `BASENAME_COLUMN`: column holding the array basename.
- `TIME_VAR`: phenotype column used for repeated-measures or longitudinal
    workflows.
- `METHYLATION_SCALE`: methylation metric used in merged modeling tables;
    choose `beta`, `m`, or `cn`. The default is `beta`; Clock Foundation exports
    continue to use beta values.
- `PHENO_ORDER`: desired order of key phenotype columns in exported tables.

### Step 4 and Step 5 shared modeling parameters

- `COVARIATES`: comma-separated adjustment variables included in each model.
- `FACTOR_VARS`: variables that should be treated as categorical before model
    fitting.
- `SCALE_VARS`: numeric fixed-effect variables to centre and divide by their
    sample standard deviations; `NULL` leaves selected model predictors on their
    original scales.
- `SCALE_VARS_GLM` and `SCALE_VARS_LME`: optional engine-specific selections;
    each inherits `SCALE_VARS` unless it is overridden.
- `CPG_PREFIX`: prefix used to identify CpG columns.
- `CPG_LIMIT`: optional cap on the number of CpGs processed; `NA` uses all.
- `PRS_MAP`: optional mapping from phenotype names to PRS variables; `NULL`
    omits PRS terms from the models.
- `SUMMARY_PVAL`: optional p-value cutoff for summary tables; `NA` keeps all
    summary rows.
- `N_CORES`: number of cores used within each R job.
- `SAVE_TXT_SUMMARIES`: whether plain-text summaries should be written.
- `CHUNK_SIZE`: number of CpGs processed per chunk; `NULL` lets the R helpers
    choose the chunk size automatically.
- `FDR_THRESHOLD`: false-discovery-rate threshold used for filtering.
- `PADJ_METHOD`: multiple-testing correction method.
- `ANNOTATION_PACKAGE`: array annotation package used for annotated summaries.
- `ANNOTATION_COLS`: annotation fields to merge into output tables.

## Step 1 parameters

- `QC_CUTOFF`: sample-quality threshold applied during preprocessing.
- `DET_PTYPE`: detection p-value method.
- `DET_PTHRESHOLD`: detection p-value cutoff.
- `PVAL_THRESHOLD`: probe-level p-value filter applied after preprocessing.
- `CHR_TO_REMOVE`: chromosomes to exclude from downstream matrices.
- `SNPS_TO_REMOVE`: SNP-related probe categories to remove.
- `PROBE_EXCLUSION_FILE`: semicolon-separated probe-exclusion reference files.
- `PROBE_EXCLUSION_ID_COLUMN`: optional column or semicolon-separated columns
    containing probe IDs; `NULL` auto-detects each reference file.
- `USE_EPICV2_MANIFEST`: whether to also remove EPICv2 probes flagged in the
    Peters et al. expanded manifest from AnnotationHub resource `AH116484`.
- `EPICV2_MANIFEST_CH_WGBS_EVIDENCE`, `EPICV2_MANIFEST_CH_BLAT`,
    `EPICV2_MANIFEST_MISSING_POS`, and `EPICV2_MANIFEST_MISMATCH_POS`: manifest
    flags included in the EPICv2 probe-exclusion set.
- `MAF_THRESHOLD`: minor-allele-frequency threshold used during SNP filtering.
- `PLOT_GROUP_VAR`: phenotype variable used for grouped QC plots.
- `LC_REF`: reference panel used for cell-type estimation.
- `REMOVE_SEX_MISMATCH`: `FALSE` retains and reports confirmed sex mismatches;
    `TRUE` removes samples only when reported and methylation-predicted sex are
    both known and disagree. Samples with missing or unknown sex are retained.

These values control the main preprocessing and filtering choices in
`preprocessingMinfiEwasWater()`.

## Step 2 parameters

- `CTRL_SVA_PERC_VAR`: target proportion of control-probe variance explained by
    the retained surrogate variables.
- `CTRL_SVA_FLAG`: enables the control-based ENmix SVA workflow.

These values determine how `svaEnmix()` estimates and retains surrogate
variables. The completed step updates the Step 1 `phenoLC.csv` in place and
records successful completion with `.sva_complete`; it does not create a second
phenotype file.

## Step 3 parameters

- `TIMEPOINTS`: comma-separated timepoints exported separately.
- `COMBINE_TIMEPOINTS`: comma-separated timepoints combined into the
    longitudinal object.
- `METHYLATION_SCALE`: methylation metric used for the merged GLM/LME input.
    `Beta` writes `phenoB*`, `M` writes `phenoM*`, and `CN` writes `phenoCN*`;
    scale input is case-insensitive.

These values determine how `preprocessingPheno()` splits and recombines the
phenotype-methylation data.

## Step 4 parameters

- `PHENOTYPES_GLM`: comma-separated phenotype variables that will be modelled
    one at a time by `methylationGLM()`.
- `GLM_LIBS`: GLM implementation used by `methylationGLM()`.
- `INTERACTION_GLM`: optional interaction term added to the fixed-effect part
    of the GLM; `NULL` omits the interaction.
- `SUMMARY_RESIDUAL_SD`: whether residual standard deviations are reported in
    summaries.
- `SAVE_SIGNIFICANT_CPGS`: whether tables of significant CpGs are exported.
- `SIGNIFICANT_CPG_PVAL`: p-value threshold used for those significant-CpG
    exports.

## Step 5 parameters

- `PHENOTYPES_LME`: comma-separated phenotype variables that will be modelled
    one at a time by `methylationLME()`.
- `PERSON_VAR`: participant identifier used as the random effect grouping
    factor.
- `LME_LIBS`: mixed-effects libraries used by `methylationLME()`.
- `LME_CORRELATION_STRUCTURE`: residual correlation structure for the `nlme`
    LME backend; use `none`, `AR1`, or `CAR1`.
- `LME_CORRELATION_VAR`: variable used to order repeated observations within
    participant for `AR1` or `CAR1`; required for those structures.
- `INTERACTION_LME`: optional interaction term used in the longitudinal model;
    `NULL` omits the interaction.
- `SAVE_SIGNIFICANT_INTERACTIONS`: whether significant interaction tables are
    exported.
- `SIGNIFICANT_INTERACTION_PVAL`: p-value threshold for those exported
    interaction results.

## Argument normalisation

- `COVARIATES_ARG`, `FACTOR_VARS_ARG`, `SCALE_VARS_GLM_ARG`,
    `SCALE_VARS_LME_ARG`, `PRS_MAP_ARG`, `SEP_TYPE_ARG`, `R_DIR_ARG`,
    `PROBE_EXCLUSION_ID_COLUMN_ARG`, `INTERACTION_GLM_ARG`, and
    `INTERACTION_LME_ARG` convert Makefile string values into arguments that can
    be passed safely to R.

This section is especially important for values such as `NULL`, because it
prevents the literal string `"NULL"` from being interpreted as a real model
value inside the main functions.

The template convention is therefore: use `NA` for optional numeric limits or
filters where the meaning is "use all" or "keep all", and use `NULL` when an
optional separator, variable, path, mapping, or model term should be absent.
`CHUNK_SIZE` uses `NULL` for automatic chunking because it disables a
user-specified tuning value rather than applying a filter.

## Include pipeline from dnaEPICO

- `DNAPIPE_MK` resolves the packaged `Makefile.rules.pipeline` file.
- `include $(DNAPIPE_MK)` imports the actual pipeline rules after the user
    configuration has been defined.

This separation is what allows the template Makefile to stay compact while the
package keeps the rule implementation in one maintained location.

# Summary

The file-based route is the correct choice when you need:

- a reproducible folder structure,
- files that can be reused by later workflow stages,
- Makefile execution, or
- repeated analyses across one or more models.

The interactive route and the file-based route are therefore complementary:
local use is best for understanding the returned objects, while pipeline use is
best for reproducible multi-step execution.

# Basics

Date the vignette was generated.

```{r, echo = FALSE}
Sys.time()
```

Wallclock time spent generating the vignette.

```{r, echo = FALSE}
totalTime <- diff(c(startTime, Sys.time()))
round(totalTime, digits = 3)
```

`R` session information.

```{r, echo = FALSE}
sessionInfo()
```

## Asking for help

As package developers, we try to explain clearly how to use our packages and in
which order to use the functions. But `R` and `Bioconductor` have a steep
learning curve, so it is critical to learn where to ask for help. We would like
to highlight the [Bioconductor support site](https://support.bioconductor.org/)
as the main resource for getting help. Please remember to use the `dnaEPICO`
tag and check the
[older posts](https://support.bioconductor.org/tag/dnaEPICO/). If you want to
receive help, please provide a small reproducible example and your session
information so the source of the problem can be tracked efficiently.

# References

---
title: "dnaEPICO Overview"
author:
  - name: Paul Ruiz
    affiliation:
      - Queensland University of Technology
    email: ruizpint@qut.edu.au
output:
  BiocStyle::html_document:
    self_contained: yes
    toc: true
    toc_float: true
    toc_depth: 2
date: "`r Sys.Date()`"
package: dnaEPICO
resource_files:
  - preprocessingMinfiEwasWater_pipeline_overview.svg
  - svaEnmix_pipeline_overview.svg
  - preprocessingPheno_pipeline_overview.svg
  - methylationGLM_pipeline_overview.svg
  - methylationLME_pipeline_overview.svg
  - dnamReport_pipeline_overview.svg
vignette: >
    %\VignetteIndexEntry{1. dnaEPICO Overview}
    %\VignetteEncoding{UTF-8}
    %\VignetteEngine{knitr::rmarkdown}
---

```{r, include = FALSE}
startTime <- Sys.time()
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>",
    crop = NULL
)

library(dnaEPICO)
```

# Introduction

This vignette provides a visual overview of the main `r Biocpkg("dnaEPICO")`
functions roles. It shows how the major functions organise their inputs, internal
processing stages, and outputs.

The diagrams are intended as orientation maps. They are useful when deciding
which function to run and checking how outputs move between stages. For detailed arguments and executable
examples, use the local-use and pipeline-use vignettes.

Read each overview as follows:

- The arrows show the expected flow from inputs to derived objects and files.
- The grouped regions show the main analysis tasks performed inside a function.
- The final nodes show the outputs that are returned to R or written to disk.
- The same function can be used interactively or as part of a file-based
    pipeline, depending on the value of `saveOutputs`.

# Required knowledge

`r Biocpkg("dnaEPICO")` is built on core Bioconductor infrastructure for
high-dimensional genomic data, with a focus on Illumina DNA methylation arrays.
To read this vignette, we assume that you are already familiar with the general
DNA methylation pipeline. If not, we recommend first reading
[this tutorial](https://paulyrp.github.io/2025-cpgpneurogenomics-workshop/tutorial.html),
which provides a practical introduction to the main concepts and analysis
steps.

Preprocessing and quality control are performed using established
Bioconductor tools, including `r Biocpkg("minfi")`, `r Biocpkg("ENmix")`, and
`r Biocpkg("wateRmelon")`. Downstream statistical modelling relies on base R
and CRAN frameworks, including generalised linear models and linear
mixed-effects models. Users are expected to have basic familiarity with R,
Bioconductor pipelines, command-line execution, and Illumina IDAT file
structures.

If you are asking yourself the question "Where do I start using
Bioconductor?", you might be interested in
[this blog post](https://www.bioconductor.org/install/).

## Probe-exclusion reference files

The lists of excluded probes depend on the Illumina methylation-array platform.

- For **Illumina HumanMethylationEPIC v2.0**, use the cross-reactive probe-exclusion file from [Peters et al. (2024)](https://springernature.figshare.com/articles/dataset/Additional_file_8_of_Characterisation_and_reproducibility_of_the_HumanMethylationEPIC_v2_0_BeadChip_for_DNA_methylation_profiling/26689706?file=48531323).

- For **Illumina MethylationEPIC**, also known as the **850k array**, use the probe-exclusion resources from [Pidsley et al. (2016)](https://springernature.figshare.com/articles/dataset/Additional_file_1_Table_S1_of_Critical_evaluation_of_the_Illumina_MethylationEPIC_BeadChip_microarray_for_whole-genome_DNA_methylation_profiling/4405946?file=7132334). The supplementary files commonly used together are [13059_2016_1066_MOESM1_ESM.csv](https://static-content.springer.com/esm/art%3A10.1186%2Fs13059-016-1066-1/MediaObjects/13059_2016_1066_MOESM1_ESM.csv), [13059_2016_1066_MOESM4_ESM.csv](https://static-content.springer.com/esm/art%3A10.1186%2Fs13059-016-1066-1/MediaObjects/13059_2016_1066_MOESM4_ESM.csv), [13059_2016_1066_MOESM5_ESM.csv](https://static-content.springer.com/esm/art%3A10.1186%2Fs13059-016-1066-1/MediaObjects/13059_2016_1066_MOESM5_ESM.csv), and [13059_2016_1066_MOESM6_ESM.csv](https://static-content.springer.com/esm/art%3A10.1186%2Fs13059-016-1066-1/MediaObjects/13059_2016_1066_MOESM6_ESM.csv).

- For **Illumina HumanMethylation450k**, use the cross-reactive and polymorphic probe resources from [Chen et al. (2013)](https://figshare.com/articles/dataset/methylAnalysisDataV3_tar_gz/4800970?file=7896205).

- Multiple probe-exclusion files can be supplied as a semicolon-separated value in `probeExclusionPath`. `dnaEPICO` reads probe IDs from each file, uses `probeExclusionIdColumn` when supplied, or otherwise auto-detects common probe-ID columns such as `ProbeID`, `TargetID`, `IlmnID`, and `Name`. The unique union of all probe IDs is then used to filter the normalised object.

- For **EPICv2**, setting `useEpicV2Manifest = TRUE` also retrieves the expanded Peters et al. manifest from AnnotationHub resource [AH116484](https://bioconductor.org/packages/release/data/annotation/vignettes/EPICv2manifest/inst/doc/EPICv2manifest.html). Probes flagged in selected manifest columns are added to the same exclusion set. By default, probes flagged by `CH_WGBS_evidence`, `CH_BLAT`, or `MissingPos` are removed, while `MismatchPos` is retained unless explicitly enabled.

```{r overview-helpers, echo = FALSE, results = "asis"}
cat("
<style>
.overview-figure {
    margin: 1.5rem 0 2rem;
}

.overview-controls {
    display: flex;
    flex-wrap: wrap;
    gap: 8px;
    margin: 0 0 0.5rem;
}

.overview-zoom-button {
    background: #ffffff;
    border: 1px solid #2c7fb8;
    border-radius: 6px;
    color: #2c7fb8;
    cursor: pointer;
    font-size: 0.875rem;
    line-height: 1.2;
    padding: 5px 9px;
}

.overview-zoom-button:hover,
.overview-zoom-button:focus {
    background: #edf6fb;
}

.overview-zoom-value {
    align-items: center;
    color: #4b5563;
    display: inline-flex;
    font-size: 0.875rem;
    min-width: 44px;
}

.overview-svg-frame {
    background: #ffffff;
    border: 1px solid #d9e2ec;
    border-radius: 6px;
    max-height: 680px;
    overflow: auto;
    padding: 10px;
}

.overview-svg-frame img {
    display: block;
    height: auto;
    max-width: none;
    width: calc(1600px * var(--overview-scale, 1));
}

.overview-figure:fullscreen {
    background: #ffffff;
    margin: 0;
    overflow: hidden;
    padding: 16px;
}

.overview-figure:fullscreen .overview-svg-frame {
    height: calc(100vh - 88px);
    max-height: none;
}

@media (max-width: 768px) {
    .overview-svg-frame {
    max-height: 520px;
    }

    .overview-svg-frame img {
    width: calc(1200px * var(--overview-scale, 1));
    }
}
</style>
<script>
function setOverviewZoom(figureId, scale) {
    var figure = document.getElementById(figureId);
    if (!figure) {
    return;
    }

    var next = Math.max(0.5, Math.min(3, scale));
    figure.dataset.scale = next.toFixed(2);
    figure.style.setProperty('--overview-scale', next.toString());

    var label = figure.querySelector('.overview-zoom-value');
    if (label) {
    label.textContent = Math.round(next * 100) + '%';
    }
}

function changeOverviewZoom(figureId, delta) {
    var figure = document.getElementById(figureId);
    var current = figure ? parseFloat(figure.dataset.scale || '1') : 1;
    setOverviewZoom(figureId, current + delta);
}

function openOverviewFullscreen(figureId) {
    var figure = document.getElementById(figureId);
    if (!figure || !figure.requestFullscreen) {
    return;
    }

    figure.requestFullscreen();
}
</script>
")

overview_svgs <- c(
    preprocessingMinfiEwasWater =
        "preprocessingMinfiEwasWater_pipeline_overview.svg",
    svaEnmix =
        "svaEnmix_pipeline_overview.svg",
    preprocessingPheno =
        "preprocessingPheno_pipeline_overview.svg",
    methylationGLM =
        "methylationGLM_pipeline_overview.svg",
    methylationLME =
        "methylationLME_pipeline_overview.svg",
    dnamReport =
        "dnamReport_pipeline_overview.svg"
)

html_escape <- function(x) {
    x <- gsub("&", "&amp;", x, fixed = TRUE)
    x <- gsub("<", "&lt;", x, fixed = TRUE)
    x <- gsub(">", "&gt;", x, fixed = TRUE)
    x <- gsub('"', "&quot;", x, fixed = TRUE)
    x
}

svg_to_html <- function(svg_path, title = NULL) {
    if (!file.exists(svg_path)) {
        stop("Missing overview SVG: ", svg_path)
    }

    if (is.null(title)) {
        title <- sub(
            "_pipeline_overview$",
            "() overview",
            tools::file_path_sans_ext(basename(svg_path))
        )
    }

    figure_id <- gsub("[^A-Za-z0-9_-]", "-", tools::file_path_sans_ext(
        basename(svg_path)
    ))

    cat(sprintf(
        '<div class="overview-figure" id="%s" data-scale="1">\n',
        figure_id
    ))
    cat('<div class="overview-controls">\n')
    cat(sprintf(
        '<button class="overview-zoom-button" type="button" onclick="changeOverviewZoom(\'%s\', 0.25)">Zoom +</button>\n',
        figure_id
    ))
    cat(sprintf(
        '<button class="overview-zoom-button" type="button" onclick="changeOverviewZoom(\'%s\', -0.25)">Zoom -</button>\n',
        figure_id
    ))
    cat(sprintf(
        '<button class="overview-zoom-button" type="button" onclick="setOverviewZoom(\'%s\', 1)">Reset</button>\n',
        figure_id
    ))
    cat(sprintf(
        '<button class="overview-zoom-button" type="button" onclick="openOverviewFullscreen(\'%s\')">Full screen</button>\n',
        figure_id
    ))
    cat('<span class="overview-zoom-value">100%</span>\n')
    cat('</div>\n')
    cat('<div class="overview-svg-frame">\n')
    cat(sprintf(
        '<img src="%s" alt="%s">\n',
        html_escape(svg_path),
        html_escape(title)
    ))
    cat('\n</div>\n</div>\n')
}
```

# Functions

The sections below describe the role of each main function in the package `r Biocpkg("dnaEPICO")`.
They are ordered according to the analysis path: preprocessing,
surrogate-variable estimation, phenotype preparation, cross-sectional modelling, longitudinal
modelling, and report generation.

## preprocessingMinfiEwasWater()

`preprocessingMinfiEwasWater()` reads the phenotype table and IDAT files, builds the methylation
objects, performs quality control and normalisation, filters probes, and
estimates cell composition.

Its role in the package is to create analysis-ready methylation data:

- Inputs: phenotype metadata, IDAT files, array annotation, and filtering
    settings.
- Processing tasks: import, quality control, normalisation, probe
    filtering, metric extraction, and cell-composition estimation.
- Outputs: filtered phenotype data, `RGSet`, beta values, M-values,
    copy-number values, quality-control figures, and `phenoLC`.

```{r preprocessingMinfiEwasWater-overview, echo = FALSE, results = "asis"}
svg_to_html(
    overview_svgs[["preprocessingMinfiEwasWater"]]
)
```

## svaEnmix()

`svaEnmix()` estimates surrogate variables from control-probe information and
adds them to the phenotype table. This step helps represent technical
variation that may otherwise influence downstream association models.

Its role is to prepare covariates for batch and technical adjustment:

- Inputs: the phenotype table from preprocessing and the saved `RGSet`.
- Processing tasks: control-probe extraction, surrogate-variable
    estimation, association checks with array-position metadata, and phenotype
    merging.
- Outputs: surrogate-variable matrices, an updated phenotype table, and
    diagnostic summaries.

```{r svaEnmix-overview, echo = FALSE, results = "asis"}
svg_to_html(
    overview_svgs[["svaEnmix"]])
```

## preprocessingPheno()

`preprocessingPheno()` aligns phenotype information with methylation metrics.
It prepares timepoint-specific data, combines longitudinal records, and creates export-ready files for external methylation-age tools.

Its role is to organise samples and methylation matrices for modelling:

- Inputs: phenotype data, beta values, M-values, copy-number values, and
    timepoint settings.
- Processing tasks: sample alignment, timepoint splitting, longitudinal
    merging, and Clock Foundation export preparation.
- Outputs: timepoint-specific phenotype-methylation tables, combined
    longitudinal tables, and Clock Foundation input files.

```{r preprocessingPheno-overview, echo = FALSE, results = "asis"}
svg_to_html(
    overview_svgs[["preprocessingPheno"]])
```

## methylationGLM()

`methylationGLM()` fits cross-sectional methylation association models. It
is designed for analyses where one phenotype is tested against CpG-level
methylation while adjusting for selected covariates.

Its role is to run single-timepoint association testing:

- Inputs: a phenotype-methylation table, phenotype variables,
    covariates, optional PRS mappings, and annotation settings.
- Processing tasks: model preparation, CpG-GLM model fitting, diagnostic
    plotting, result summarisation, multiple-testing adjustment, and annotation.
- Outputs: model objects, summary tables, diagnostic figures,
    significant-CpG exports, and annotated workbooks.

```{r methylationGLM-overview, echo = FALSE, results = "asis"}
svg_to_html(
    overview_svgs[["methylationGLM"]])
```

## methylationLME()

`methylationLME()` fits longitudinal mixed-effects models. It supports
repeated-measures designs by including participant-level random effects and
timepoint-related terms.

Its role is to model methylation change across repeated observations:

- Inputs: combined longitudinal phenotype-methylation data, participant
    identifiers, timepoint variables, phenotypes, covariates, and annotation
    settings.
- Processing tasks: longitudinal model preparation, CpG-LME model
    fitting, interaction testing, diagnostic plotting, summarisation, and
    annotation.
- Outputs: mixed-model objects, interaction summaries, diagnostic figures,
    significant-interaction exports, and annotated longitudinal workbooks.

```{r methylationLME-overview, echo = FALSE, results = "asis"}
svg_to_html(
    overview_svgs[["methylationLME"]])
```

## dnamReport()

`dnamReport()` assembles the main tables, figures, logs, and model summaries
into a report website. It can be run after preprocessing and modelling
outputs have been written to disk.

Its role is to make the package outputs easier to inspect and share:

- Inputs: phenotype tables, QC figures, ENmix figures, SVA figures,
    metric figures, model annotation tables, detection P-value data, and logs.
- Processing tasks: input validation, report file preparation, Quarto
    rendering, and post-processing.
- Outputs: a rendered report site with organised tabs for data,
    preprocessing, modelling, logs, and supporting diagnostics.

```{r dnamReport-overview, echo = FALSE, results = "asis"}
svg_to_html(
    overview_svgs[["dnamReport"]])
```

# Summary

The preprocessing function create quality-controlled methylation data and
analysis-ready phenotype tables. The modelling functions fit cross-sectional and
longitudinal association models. The report function gathers the resulting
tables, figures, and logs into a browsable output.

In practice, the main functions are:

- run `preprocessingMinfiEwasWater()` to prepare methylation objects and QC
    outputs,
- run `svaEnmix()` when control-probe surrogate variables are needed,
- run `preprocessingPheno()` to prepare modelling tables,
- run `methylationGLM()` or `methylationLME()` for association
    testing, and
- run `dnamReport()` to review the completed outputs.

# Basics

Date the vignette was generated.

```{r, echo = FALSE}
Sys.time()
```

Wallclock time spent generating the vignette.

```{r, echo = FALSE}
totalTime <- diff(c(startTime, Sys.time()))
round(totalTime, digits = 3)
```

`R` session information.

```{r, echo = FALSE}
sessionInfo()
```

## Asking for help

As package developers, we try to explain clearly how to use our packages and
in which order to use the functions. But `R` and `Bioconductor` have a steep
learning curve, so it is critical to learn where to ask for help. We would
like to highlight the [Bioconductor support site](https://support.bioconductor.org/)
as the main resource for getting help. Please remember to use the `dnaEPICO`
tag and check the
[older posts](https://support.bioconductor.org/tag/dnaEPICO/). If you want to
receive help, please provide a small reproducible example and your session
information so the source of the problem can be tracked efficiently.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/result_classes.R
\docType{class}
\name{dnaEPICO_methylationLME-class}
\alias{dnaEPICO_methylationLME-class}
\alias{dnaEPICO_methylationLME}
\title{Result class returned by methylationLME}
\description{
Objects of class \code{'dnaEPICO_methylationLME'} are list-based results
returned by \code{\link[=methylationLME]{methylationLME()}}. They collect the prepared longitudinal
analysis table, fitted mixed models, summaries, diagnostics, annotations, and
optional saved files.
}
\section{Structure}{

\describe{
\item{preparedData}{Object returned by \code{\link[=prepareMethylationLMEData]{prepareMethylationLMEData()}}.}
\item{modelFits}{Object returned by \code{\link[=fitMethylationLMEModels]{fitMethylationLMEModels()}}.}
\item{modelSummaries}{Object returned by
\code{\link[=summarizeMethylationLMEModels]{summarizeMethylationLMEModels()}}.}
\item{significantInteractions}{Object returned by
\code{\link[=collectSignificantInteractionsMethylationLME]{collectSignificantInteractionsMethylationLME()}}.}
\item{diagnosticPlots}{Object returned by
\code{\link[=plotMethylationLMEDiagnostics]{plotMethylationLMEDiagnostics()}}.}
\item{annotation}{Object returned by
\code{\link[=annotateMethylationLMESummaries]{annotateMethylationLMESummaries()}}.}
\item{savedFiles}{Object returned by
\code{\link[=writeMethylationLMEOutputs]{writeMethylationLMEOutputs()}} when \code{saveOutputs = TRUE}, otherwise
\code{NULL}.}
}
}

\seealso{
\code{\link[=methylationLME]{methylationLME()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessingMinfiEwasWater_steps.R
\name{filterSamplesMinfiEwasWater}
\alias{filterSamplesMinfiEwasWater}
\title{Filter failed samples from an RGSet and phenotype table}
\usage{
filterSamplesMinfiEwasWater(
  RGSet,
  targets,
  failedSamples = character(0),
  SampleID = "Sample_Name",
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_filterSamplesMinfiEwasWater.txt"
)
}
\arguments{
\item{RGSet}{An \code{RGChannelSet}.}

\item{targets}{Data frame containing phenotype information.}

\item{failedSamples}{Character vector of sample identifiers to remove.}

\item{SampleID}{Character. Name of the sample identifier column in \code{targets}.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
A list with class \code{'dnaEPICO_minfiEwasWater_samples'} containing the
filtered \code{RGSet}, aligned phenotype table, and failed sample identifiers.
}
\description{
Remove failed samples identified during sample assessment and return the
filtered \code{RGChannelSet} together with the aligned phenotype table.
}
\examples{
\dontshow{if (requireNamespace("minfiData", quietly = TRUE)) withAutoprint(\{ # examplesIf}
ex <- dnaEPICO:::exampleMinfiBaseDataDnaEpico()
filtered_samples <- filterSamplesMinfiEwasWater(
    RGSet = ex$RGSet,
    targets = ex$targets,
    failedSamples = ex$targets$Sample_Name[1],
    SampleID = "Sample_Name",
    verbose = FALSE,
    logs = FALSE
)
nrow(filtered_samples$targets)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dnamReport.R
\name{renderDnamReport}
\alias{renderDnamReport}
\title{Render a prepared DNA methylation report}
\usage{
renderDnamReport(
  preparedReport,
  verbose = FALSE,
  logs = FALSE,
  logDir = NULL,
  clean = TRUE
)
}
\arguments{
\item{preparedReport}{Object returned by \code{prepareDnamReportInputs()}.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages.}

\item{logs}{Logical. If \code{TRUE}, write progress messages to a log file.}

\item{logDir}{Character or \code{NULL}. Directory for optional log files.}

\item{clean}{Logical. Retained for backwards compatibility.}
}
\value{
A list with class \code{'dnaEPICO_dnamReport_render'}.
}
\description{
Render a prepared DNA methylation report
}
\examples{
report_root <- file.path(tempdir(), "dnaepico-render-example")
prepared <- prepareDnamReportInputs(
    outputDir = file.path(report_root, "reports")
)
rendered <- renderDnamReport(prepared)
rendered$status

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methylationGLM_steps.R
\name{collectSignificantCpGsMethylationGLM}
\alias{collectSignificantCpGsMethylationGLM}
\title{Collect significant CpG coefficient tables from fitted one-timepoint GLMs}
\usage{
collectSignificantCpGsMethylationGLM(
  modelResults,
  pvalThreshold = 0.05,
  interactionTerm = NULL,
  excludeNonConverged = FALSE,
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_methylationGLM.txt"
)
}
\arguments{
\item{modelResults}{Object returned by \code{fitMethylationGLMModels()}.}

\item{pvalThreshold}{Numeric. Threshold applied to phenotype main-effect or
interaction p-values.}

\item{interactionTerm}{Character or \code{NULL}. Optional interaction term.}

\item{excludeNonConverged}{Logical. If \code{TRUE}, do not collect CpGs whose GLM
did not converge; their rows and reasons remain in diagnostic and annotated
output.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
A list with class \code{'dnaEPICO_methylationGLM_significant_cpgs'}.
}
\description{
Collect the raw coefficient tables for CpGs whose phenotype main effect or
interaction p-value passes the requested threshold.
}
\examples{
ex <- dnaEPICO:::exampleMethylationGLMStateDnaEpico()
significant_cpgs <- collectSignificantCpGsMethylationGLM(
    modelResults = ex$modelResults,
    pvalThreshold = 1,
    verbose = FALSE,
    logs = FALSE
)
names(significant_cpgs)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readPhenotypeTargets.R
\name{readPhenotypeTargets}
\alias{readPhenotypeTargets}
\title{Read phenotype targets for shared dnaEPICO workflows}
\usage{
readPhenotypeTargets(
  phenoFile,
  sepType = NULL,
  nSamples = NA,
  SampleID = "Sample_Name",
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_readPhenotypeTargets.txt"
)
}
\arguments{
\item{phenoFile}{Character. Path to the phenotype table on disk.}

\item{sepType}{Character or \code{NULL}. Field separator used in \code{phenoFile}. Use
\code{NULL} for a standard comma-separated file, \code{'\\\\t'} for a tab-delimited
file, or another single-character separator accepted by \code{utils::read.csv()}.}

\item{nSamples}{Integer or \code{NA}. Number of rows to keep from the start of the
phenotype table. The default \code{NA} reads and returns all rows.}

\item{SampleID}{Character. Name of the column containing sample identifiers
that will later be used to name methylation-array samples.}

\item{verbose}{Logical. If \code{TRUE}, emit progress and preview messages with
\code{message()}. The default is \code{FALSE}, so the function is quiet unless the
user explicitly requests messages.}

\item{logs}{Logical. If \code{TRUE}, write the same progress messages to a log
file. The default is \code{FALSE}.}

\item{log_dir}{Character or \code{NULL}. Directory where the log file should be
written when \code{logs = TRUE}. If \code{NULL}, the current working directory is
used.}

\item{log_file}{Character. File name used when \code{logs = TRUE}. The default is
\code{'log_readPhenotypeTargets.txt'}.}
}
\value{
A \code{data.frame} containing the phenotype targets.
}
\description{
Read the phenotype table used by shared \code{dnaEPICO} workflows, validate the
sample identifier column, optionally subset the first \code{nSamples}, and return
the targets as a base \code{data.frame}.
}
\examples{
tmp <- tempdir()
pheno <- data.frame(
    Sample_Name = c("S1", "S2"),
    Sex = c("F", "M"),
    stringsAsFactors = FALSE
)
pheno_file <- file.path(tmp, "pheno.csv")
utils::write.csv(pheno, pheno_file, row.names = FALSE)
targets <- readPhenotypeTargets(
    phenoFile = pheno_file,
    SampleID = "Sample_Name"
)
stopifnot(is.data.frame(targets))
stopifnot(nrow(targets) == 2L)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessingMinfiEwasWater_steps.R
\name{readRGSetMinfiEwasWater}
\alias{readRGSetMinfiEwasWater}
\title{Read IDAT files into an annotated RGChannelSet}
\usage{
readRGSetMinfiEwasWater(
  idatFolder,
  targets,
  SampleID = "Sample_Name",
  arrayType = "IlluminaHumanMethylationEPICv2",
  annotationVersion = "20a1.hg38",
  force = FALSE,
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_readRGSetMinfiEwasWater.txt"
)
}
\arguments{
\item{idatFolder}{Character. Directory containing the IDAT files.}

\item{targets}{Data frame returned by \code{readPhenotypeTargets()} or an
equivalent phenotype table.}

\item{SampleID}{Character. Name of the phenotype column containing sample
identifiers used to label the \code{RGChannelSet}.}

\item{arrayType}{Character. Array name passed to
\code{Biobase::annotation(RGSet)}, for example
\code{'IlluminaHumanMethylationEPICv2'}.}

\item{annotationVersion}{Character. Annotation build passed to
\code{Biobase::annotation(RGSet)}, for example \code{'20a1.hg38'} for EPIC v2 hg38
annotations or \code{'ilmn12.hg19'} for 450K hg19 annotations.}

\item{force}{Logical. Passed to \code{minfi::read.metharray.exp()}. Use \code{TRUE}
only after confirming that the selected IDAT files should be read together.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
An annotated \code{RGChannelSet}.
}
\description{
Read methylation-array IDAT files with \code{minfi::read.metharray.exp()}, set
sample names from the phenotype table, apply the requested annotation, and
return the resulting \code{RGChannelSet}.
}
\examples{
if (requireNamespace("minfiData", quietly = TRUE) &&
    requireNamespace("IlluminaHumanMethylation450kmanifest", quietly = TRUE) &&
    requireNamespace("IlluminaHumanMethylation450kanno.ilmn12.hg19", quietly = TRUE)) {
    ex <- dnaEPICO:::exampleMinfiIdatInputsDnaEpico(n = 4)
    rgset <- readRGSetMinfiEwasWater(
        idatFolder = ex$idatFolder,
        targets = ex$targets,
        SampleID = "Sample_Name",
        arrayType = ex$arrayType,
        annotationVersion = ex$annotationVersion
    )
    class(rgset)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dnamReport.R
\name{prepareDnamReportInputs}
\alias{prepareDnamReportInputs}
\title{Prepare inputs for a DNA methylation report}
\usage{
prepareDnamReportInputs(
  outputDir = "reports",
  qcDir = file.path("figures", "preprocessingMinfiEwasWater", "enmix"),
  preprocessingDir = file.path("figures", "preprocessingMinfiEwasWater", "qc"),
  postprocessingDir = file.path("figures", "preprocessingMinfiEwasWater", "metrics"),
  svaDir = file.path("figures", "svaEnmix"),
  glmDir = file.path("figures", "methylationGLM"),
  lmeDir = file.path("figures", "methylationLME"),
  figDir = file.path(outputDir, "assets", "figures"),
  verbose = FALSE,
  logs = FALSE,
  logDir = outputDir
)
}
\arguments{
\item{outputDir}{Character. Directory where the report project is written.}

\item{qcDir}{Character. Directory containing ENmix quality-control figures.}

\item{preprocessingDir}{Character. Directory containing preprocessing
quality-control figures.}

\item{postprocessingDir}{Character. Directory containing postprocessing
metric
figures.}

\item{svaDir}{Character. Directory containing SVA or batch-effect figures.}

\item{glmDir}{Character. Directory containing GLM figures.}

\item{lmeDir}{Character. Directory containing LME figures.}

\item{figDir}{Character. Directory used for generated report figure assets.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write progress messages to
\code{file.path(logDir, 'log_dnamReport.txt')}.}

\item{logDir}{Character. Directory for optional log files.}
}
\value{
A list with class \code{'dnaEPICO_dnamReport_prepared'}.
}
\description{
Prepare inputs for a DNA methylation report
}
\examples{
report_root <- file.path(tempdir(), "dnaepico-report-inputs")
prepared <- prepareDnamReportInputs(
    outputDir = file.path(report_root, "reports"),
    qcDir = file.path(
        report_root,
        "figures",
        "preprocessingMinfiEwasWater",
        "enmix"
    ),
    preprocessingDir = file.path(
        report_root,
        "figures",
        "preprocessingMinfiEwasWater",
        "qc"
    ),
    postprocessingDir = file.path(
        report_root,
        "figures",
        "preprocessingMinfiEwasWater",
        "metrics"
    ),
    svaDir = file.path(report_root, "figures", "svaEnmix")
)
inherits(prepared, "dnaEPICO_dnamReport_prepared")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessingMinfiEwasWater_steps.R
\name{extractMetricsMinfiEwasWater}
\alias{extractMetricsMinfiEwasWater}
\title{Extract beta, M, and copy-number matrices from a filtered object}
\usage{
extractMetricsMinfiEwasWater(
  filteredData,
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_extractMetricsMinfiEwasWater.txt"
)
}
\arguments{
\item{filteredData}{Object returned by \code{filterProbesMinfiEwasWater()}.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
A list with class \code{'dnaEPICO_minfiEwasWater_metrics'} containing
\code{beta}, \code{m}, and \code{cn}.
}
\description{
Extract beta, M, and copy-number matrices from a filtered object
}
\examples{
ex <- dnaEPICO:::exampleMinfiMetricsStateDnaEpico()
metrics_data <- extractMetricsMinfiEwasWater(
    filteredData = ex$filteredData,
    verbose = FALSE,
    logs = FALSE
)
names(metrics_data)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/svaEnmix_steps.R
\name{analyzeSvaEnmix}
\alias{analyzeSvaEnmix}
\title{Analyze surrogate variables against Sentrix chip and position factors}
\usage{
analyzeSvaEnmix(
  sva,
  RGSet,
  SentrixIDColumn = "Sentrix_ID",
  SentrixPositionColumn = "Sentrix_Position",
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_analyzeSvaEnmix.txt"
)
}
\arguments{
\item{sva}{Numeric matrix of surrogate variables with samples in rows.}

\item{RGSet}{An \code{RGChannelSet} aligned with \code{sva}.}

\item{SentrixIDColumn}{Character. Name of the chip identifier column in
\code{SummarizedExperiment::colData(RGSet)}.}

\item{SentrixPositionColumn}{Character. Name of the chip position column in
\code{SummarizedExperiment::colData(RGSet)}.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
A list with class \code{'dnaEPICO_svaEnmix_analysis'} containing the
aligned Sentrix factors, full and reduced linear models, and ANOVA tables.
}
\description{
Fit linear models for each surrogate variable against informative Sentrix
chip and position factors, perform backward elimination with
\code{stats::drop1()}, and return the in-memory analysis objects. A constant
technical factor is omitted; if both are constant, an intercept-only model
is retained.
}
\examples{
ex <- dnaEPICO:::exampleSvaAnalysisStateDnaEpico()
analysis_data <- analyzeSvaEnmix(
    sva = ex$sva,
    RGSet = ex$RGSet,
    SentrixIDColumn = "Sentrix_ID",
    SentrixPositionColumn = "Sentrix_Position",
    verbose = FALSE,
    logs = FALSE
)
analysis_data$K

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessingMinfiEwasWater_steps.R
\name{plotRawDensityMinfiEwasWater}
\alias{plotRawDensityMinfiEwasWater}
\title{Plot raw beta-value density from a raw preprocessing object}
\usage{
plotRawDensityMinfiEwasWater(
  rawData,
  targets,
  plotGroupVar = "Sex",
  display = FALSE,
  file = NULL,
  width = 2000L,
  height = 1000L,
  res = 150L,
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_plotRawDensityMinfiEwasWater.txt"
)
}
\arguments{
\item{rawData}{Object returned by \code{buildRawMinfiEwasWater()}.}

\item{targets}{Filtered phenotype data aligned with \code{rawData}.}

\item{plotGroupVar}{Character. Phenotype column used to group samples in the
density plot.}

\item{display}{Logical. If \code{TRUE}, draw the plot on the active graphics
device.}

\item{file}{Character or \code{NULL}. TIFF file written when supplied.}

\item{width}{Integer. TIFF width in pixels when \code{file} is supplied.}

\item{height}{Integer. TIFF height in pixels when \code{file} is supplied.}

\item{res}{Integer. TIFF resolution in DPI when \code{file} is supplied.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
Invisibly returns the saved TIFF path when \code{file} is supplied,
otherwise \code{NULL}.
}
\description{
Draw the pre-normalization beta density plot from a raw minfi object and a
grouping variable in the phenotype table.
}
\examples{
ex <- dnaEPICO:::exampleMinfiMetricsStateDnaEpico()
plotRawDensityMinfiEwasWater(
    rawData = ex$rawData,
    targets = ex$targets,
    plotGroupVar = "Sex",
    display = FALSE,
    verbose = FALSE,
    logs = FALSE
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessingPheno_steps.R
\name{loadMetricsPreprocessingPheno}
\alias{loadMetricsPreprocessingPheno}
\title{Load methylation metric matrices for preprocessingPheno}
\usage{
loadMetricsPreprocessingPheno(
  betaPath,
  mPath,
  cnPath,
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_loadMetricsPreprocessingPheno.txt"
)
}
\arguments{
\item{betaPath}{Character. Path to the saved beta-value object.}

\item{mPath}{Character. Path to the saved M-value object.}

\item{cnPath}{Character. Path to the saved copy-number object.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
A list with class \code{'dnaEPICO_preprocessingPheno_metrics'}
containing \code{beta}, \code{m}, and \code{cn}.
}
\description{
Load the metric matrices generated by \code{preprocessingMinfiEwasWater()} and
return them as a single in-memory object for downstream phenotype alignment.
}
\examples{
ex <- dnaEPICO:::examplePreprocessingPhenoStateDnaEpico()
metrics_data <- loadMetricsPreprocessingPheno(
    betaPath = ex$betaPath,
    mPath = ex$mPath,
    cnPath = ex$cnPath,
    verbose = FALSE,
    logs = FALSE
)
names(metrics_data)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessingPheno_steps.R
\name{writePreprocessingPhenoOutputs}
\alias{writePreprocessingPhenoOutputs}
\title{Write preprocessingPheno outputs to disk}
\usage{
writePreprocessingPhenoOutputs(
  preprocessingData,
  outputPheno = "data/preprocessingPheno",
  outputRData = "rData/preprocessingPheno/metrics",
  outputRDataMerge = "rData/preprocessingPheno/mergeData",
  outputDir = "data/preprocessingPheno",
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_writePreprocessingPhenoOutputs.txt"
)
}
\arguments{
\item{preprocessingData}{Object returned by \code{preprocessingPheno()} or a list
with the same components.}

\item{outputPheno}{Character. Directory used for saved phenotype CSV files.}

\item{outputRData}{Character. Directory used for saved metric \code{.RData} files.}

\item{outputRDataMerge}{Character. Directory used for saved merged
phenotype-plus-beta \code{.RData} files.}

\item{outputDir}{Character. Directory used for the Clock Foundation export
files.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
A list with class \code{'dnaEPICO_preprocessingPheno_paths'}
containing the paths written to disk.
}
\description{
Write the CSV, ZIP, and \code{.RData} outputs produced by
\code{preprocessingPheno()}. This helper keeps file writing separate from the
in-memory preprocessing steps.
}
\examples{
ex <- dnaEPICO:::examplePreprocessingPhenoStateDnaEpico()
output_paths <- writePreprocessingPhenoOutputs(
    preprocessingData = ex$preprocessingData,
    outputPheno = file.path(ex$tempDir, "pheno"),
    outputRData = file.path(ex$tempDir, "metrics"),
    outputRDataMerge = file.path(ex$tempDir, "merge"),
    outputDir = file.path(ex$tempDir, "clock"),
    verbose = FALSE,
    logs = FALSE
)
names(output_paths)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estimateLC.R
\name{estimateLC}
\alias{estimateLC}
\title{Estimate saliva cell proportions from DNA methylation beta values}
\usage{
estimateLC(meth, ref, constrained = FALSE)
}
\arguments{
\item{meth}{Numeric matrix of beta values with CpGs in rows and samples in
columns. Row names must contain probe identifiers compatible with the
selected reference.}

\item{ref}{Character. Reference panel name. Supported values are \code{'saliva'}
and \code{'salivaEPIC'}.}

\item{constrained}{Logical. If \code{TRUE}, estimated cell proportions are
constrained to sum to one.}
}
\value{
A \code{data.table} with one row per sample and one column per estimated
cell type.
}
\description{
Estimate cell-type proportions with the saliva reference panels bundled in
\code{dnaEPICO}. This function keeps the original \code{estimateLC()} interface used by
the package while using the internal reference files distributed in
\code{inst/extdata}.
}
\examples{
ref_file <- system.file("extdata", "saliva.txt", package = "dnaEPICO")
ref_panel <- as.matrix(utils::read.table(ref_file))
meth <- ref_panel[1:20, , drop = FALSE]
colnames(meth) <- c("sample1", "sample2")
estimateLC(
    meth = meth,
    ref = "saliva",
    constrained = FALSE
)

}
\references{
Murat K, et al. Ewastools: Infinium Human Methylation BeadChip pipeline for
population epigenetics integrated into Galaxy. \emph{GigaScience}.
2020;9(5):giaa049.
Houseman EA, Accomando WP, Koestler DC, et al. DNA methylation arrays as
surrogate measures of cell mixture distribution. \emph{BMC Bioinformatics}.
2012;13:86.
Reinius LE, Acevedo N, Joerink M, et al. Differential DNA methylation in
purified human blood cells: implications for cell lineage and studies on
disease susceptibility. \emph{PLoS One}. 2012;7(7):e41361.
Bakulski KM, Feinberg JI, Andrews SV, et al. DNA methylation of cord blood
cell types: applications for mixed cell birth studies. \emph{Epigenetics}.
2016;11(5):354-362.
de Goede OM, Razzaghian HR, Price EM, et al. Nucleated red blood cells
impact DNA methylation and expression analyses of cord blood hematopoietic
cells. \emph{Clinical Epigenetics}. 2015;7:95.
Gervin K, Salas LA, Bakulski KM, et al. Cell type specific DNA methylation
in cord blood: a 450K reference data set and cell count-based validation of
estimated cell type composition. \emph{Epigenetics}. 2016;11(9):690-698.
Gervin K, Salas LA, Bakulski KM, et al. Systematic evaluation and validation
of reference and library selection methods for deconvolution of cord blood
DNA methylation data. \emph{bioRxiv}. 2019. doi:10.1101/570457.
Salas LA, Koestler DC, Butler RA, et al. An optimized library for
reference-based deconvolution of whole-blood biospecimens assayed using the
Illumina HumanMethylationEPIC BeadArray. \emph{Genome Biology}. 2018;19:64.
Heiss JA, Just AC, Brenner H. Training a model for estimating leukocyte
composition using whole-blood DNA methylation and cell counts as reference.
\emph{Epigenomics}. 2017;9(1):13-20.
Middleton LYM, Dou J, Mill J, et al. Saliva cell type DNA methylation
reference panel for epidemiology studies in children. 2020.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methylationLME_steps.R
\name{collectSignificantInteractionsMethylationLME}
\alias{collectSignificantInteractionsMethylationLME}
\title{Collect significant longitudinal model terms from fitted mixed-effects models}
\usage{
collectSignificantInteractionsMethylationLME(
  modelResults,
  pvalThreshold = 0.05,
  interactionTerm = NULL,
  excludeSingular = FALSE,
  excludeNonConverged = FALSE,
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_methylationLME.txt"
)
}
\arguments{
\item{modelResults}{Object returned by \code{fitMethylationLMEModels()}.}

\item{pvalThreshold}{Numeric. Threshold applied to the extracted phenotype or
interaction p-values.}

\item{interactionTerm}{Character or \code{NULL}. Optional interaction term. When
\code{NULL}, phenotype main effects are used.}

\item{excludeSingular}{Logical. If \code{TRUE}, do not collect significant CpGs
with singular lme4 fits.}

\item{excludeNonConverged}{Logical. If \code{TRUE}, do not collect significant
CpGs whose mixed model did not converge. Excluded CpG rows and reasons
remain available in diagnostic and annotated output.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
A list with class \code{'dnaEPICO_methylationLME_significant'}
containing the retained coefficient tables for each phenotype.
}
\description{
Collect raw coefficient tables for CpGs whose phenotype main effect or
requested interaction p-value passes the chosen threshold.
}
\examples{
ex <- dnaEPICO:::exampleMethylationLMEStateDnaEpico()
significant_hits <- collectSignificantInteractionsMethylationLME(
    modelResults = ex$modelResults,
    pvalThreshold = 1,
    verbose = FALSE,
    logs = FALSE
)
names(significant_hits)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/svaEnmix_steps.R
\name{writeSvaEnmixOutputs}
\alias{writeSvaEnmixOutputs}
\title{Write svaEnmix outputs to disk}
\usage{
writeSvaEnmixOutputs(
  svaData,
  mergedPheno,
  analysisData = NULL,
  phenoFile = NULL,
  SampleID = "Sample_Name",
  sepType = NULL,
  dataBaseDir = "data",
  rBaseDir = "rData",
  scriptLabel = "svaEnmix",
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_writeSvaEnmixOutputs.txt"
)
}
\arguments{
\item{svaData}{Object returned by \code{estimateSvaEnmixControls()}.}

\item{mergedPheno}{Phenotype data frame returned by \code{mergeSvaTargetsEnmix()}.}

\item{analysisData}{Optional object returned by \code{analyzeSvaEnmix()}.}

\item{phenoFile}{Character or \code{NULL}. Existing phenotype file replaced with
\code{mergedPheno} after validation.}

\item{SampleID}{Character. Phenotype column containing sample identifiers.}

\item{sepType}{Character or \code{NULL}. Field separator used by \code{phenoFile}.}

\item{dataBaseDir}{Character. Base directory used for saved data outputs.}

\item{rBaseDir}{Character. Base directory used for saved \code{.RData} outputs.}

\item{scriptLabel}{Character. Label used to create the output subdirectory.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
A list with class \code{'dnaEPICO_svaEnmix_paths'} containing the paths
written to disk.
}
\description{
Write the SVA matrix, phenotype, and model-summary outputs.
}
\examples{
ex <- dnaEPICO:::exampleSvaAnalysisStateDnaEpico()
temp_dir <- tempdir()
pheno_file <- file.path(temp_dir, "phenoLC.csv")
utils::write.csv(ex$targets, pheno_file, row.names = FALSE)
output_paths <- writeSvaEnmixOutputs(
    svaData = list(sva = ex$sva),
    mergedPheno = ex$mergedPheno,
    analysisData = ex$analysisData,
    phenoFile = pheno_file,
    SampleID = "Sample_Name",
    dataBaseDir = file.path(temp_dir, "data"),
    rBaseDir = file.path(temp_dir, "rData"),
    scriptLabel = "svaEnmixExample",
    verbose = FALSE,
    logs = FALSE
)
names(output_paths)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methylationGLM_steps.R
\name{summarizeMethylationGLMModels}
\alias{summarizeMethylationGLMModels}
\title{Summarize CpG-wise Gaussian GLM fits for one-timepoint analyses}
\usage{
summarizeMethylationGLMModels(
  modelResults,
  preparedData,
  summaryResidualSD = TRUE,
  summaryPval = NA,
  excludeNonConverged = FALSE,
  nCores = 1L,
  libPath = NULL,
  glmLibs = "glm2",
  chunkSize = NULL,
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_methylationGLM.txt"
)
}
\arguments{
\item{modelResults}{Object returned by \code{fitMethylationGLMModels()}.}

\item{preparedData}{Object returned by \code{prepareMethylationGLMData()}.}

\item{summaryResidualSD}{Logical. If \code{TRUE}, add residual standard deviations
to each CpG summary row.}

\item{summaryPval}{Numeric or \code{NA}. Optional p-value filter applied to the
returned summary tables. \code{NA} keeps all rows.}

\item{excludeNonConverged}{Logical. If \code{TRUE}, retain non-converged CpG rows
and their reasons but set their inferential p-values to \code{NA}.}

\item{nCores}{Integer. Number of worker processes to use while extracting
summary rows.}

\item{libPath}{Character vector or \code{NULL}. Optional library paths forwarded
to worker processes.}

\item{glmLibs}{Character vector or comma-separated string of package names to
check on worker processes. The default is \code{'glm2'}.}

\item{chunkSize}{Integer or \code{NULL}. Number of CpGs to process per parallel
chunk. \code{NULL} chooses a value automatically.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
A list with class \code{'dnaEPICO_methylationGLM_summaries'}
containing the optionally filtered summary tables in \code{summaries} and the
complete CpG-level tables in \code{diagnosticSummaries}. Diagnostics,
annotation, and report output use the complete tables so \code{summaryPval}
does not remove CpGs from those outputs.
}
\description{
Extract phenotype-specific CpG coefficient tables from the fitted model
object returned by \code{fitMethylationGLMModels()}.
}
\examples{
ex <- dnaEPICO:::exampleMethylationGLMStateDnaEpico()
summary_results <- summarizeMethylationGLMModels(
    modelResults = ex$modelResults,
    preparedData = ex$preparedData,
    summaryResidualSD = TRUE,
    summaryPval = NA,
    nCores = 1,
    verbose = FALSE,
    logs = FALSE
)
names(summary_results$summaries)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methylationLME_steps.R
\name{prepareMethylationLMEData}
\alias{prepareMethylationLMEData}
\title{Prepare longitudinal phenotype-plus-methylation data for mixed-effects
analyses}
\usage{
prepareMethylationLMEData(
  inputPheno,
  personVar = "person",
  timeVar = "Timepoint",
  phenotypes,
  covariates,
  factorVars,
  scaleVars = NULL,
  prsMap = NULL,
  cpgPrefix = "cg",
  cpgLimit = NA,
  methylationScale = "beta",
  interactionTerm = NULL,
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_methylationLME.txt"
)
}
\arguments{
\item{inputPheno}{Character. Path to the merged longitudinal phenotype-plus-
methylation object created by \code{preprocessingPheno()}.}

\item{personVar}{Character. Name of the subject identifier column.}

\item{timeVar}{Character. Name of the time variable.}

\item{phenotypes}{Character vector or comma-separated string of phenotype
variables to model.}

\item{covariates}{Character vector or comma-separated string of covariate
variables to adjust for.}

\item{factorVars}{Character vector or comma-separated string of variables
that
should be converted to factors before modeling.}

\item{scaleVars}{Character vector, comma-separated variable names, or \code{NULL}.
Numeric fixed-effect variables to standardize before model fitting.}

\item{prsMap}{Character vector or comma-separated string of phenotype-to-PRS
mappings in the form \code{'Phenotype:PRS'}.}

\item{cpgPrefix}{Character. Prefix used to identify methylation columns.}

\item{cpgLimit}{Integer or \code{NA}. Maximum number of CpGs to retain. \code{NA}
keeps all matching CpGs.}

\item{methylationScale}{Character. Methylation metric represented by the CpG
columns. One of \code{'Beta'}, \code{'M'}, or \code{'CN'}, case-insensitively.}

\item{interactionTerm}{Character or \code{NULL}. Optional interaction term.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
A list with class \code{'dnaEPICO_methylationLME_data'} containing
the prepared analysis data, parsed variable selections, CpG columns,
timepoint summaries, and subject-ID diagnostics.
}
\description{
Load the merged longitudinal phenotype-plus-methylation object, ensure that
a subject
identifier column is available, validate the requested modeling variables,
convert selected variables to factors, and return a single in-memory object
for downstream mixed-effects modeling helpers.
}
\examples{
ex <- dnaEPICO:::exampleMethylationLMEStateDnaEpico()
prepared_data <- prepareMethylationLMEData(
    inputPheno = ex$inputPath,
    personVar = "person",
    timeVar = "Timepoint",
    phenotypes = "score",
    covariates = "sex",
    factorVars = "sex",
    cpgLimit = 2,
    verbose = FALSE,
    logs = FALSE
)
names(prepared_data)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/result_classes.R
\docType{class}
\name{dnaEPICO_preprocessingPheno-class}
\alias{dnaEPICO_preprocessingPheno-class}
\alias{dnaEPICO_preprocessingPheno}
\title{Result class returned by preprocessingPheno}
\description{
Objects of class \code{'dnaEPICO_preprocessingPheno'} are list-based results
returned by \code{\link[=preprocessingPheno]{preprocessingPheno()}}. They describe the phenotype data,
methylation matrices, timepoint splits, longitudinal merges, and optional
exported files.
}
\section{Structure}{

\describe{
\item{pheno}{Phenotype table read from \code{phenoFile}.}
\item{metricsData}{Object returned by \code{\link[=loadMetricsPreprocessingPheno]{loadMetricsPreprocessingPheno()}}.}
\item{timepointData}{Object returned by
\code{\link[=splitTimepointsPreprocessingPheno]{splitTimepointsPreprocessingPheno()}}.}
\item{combinedData}{Object returned by
\code{\link[=combineTimepointsPreprocessingPheno]{combineTimepointsPreprocessingPheno()}}.}
\item{clockFoundation}{Object returned by
\code{\link[=buildClockFoundationInputsPreprocessingPheno]{buildClockFoundationInputsPreprocessingPheno()}}.}
\item{savedFiles}{Object returned by \code{\link[=writePreprocessingPhenoOutputs]{writePreprocessingPhenoOutputs()}}
when \code{saveOutputs = TRUE}, otherwise \code{NULL}.}
\item{logFile}{Resolved path to the optional log file, or \code{NULL} when
logging was disabled.}
}
}

\seealso{
\code{\link[=preprocessingPheno]{preprocessingPheno()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methylationLME_steps.R
\name{summarizeTimepointsMethylationLME}
\alias{summarizeTimepointsMethylationLME}
\title{Summarize phenotype values by timepoint for longitudinal methylation analyses}
\usage{
summarizeTimepointsMethylationLME(
  data,
  timeVar = "Timepoint",
  phenotypes,
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_methylationLME.txt"
)
}
\arguments{
\item{data}{Data frame containing the longitudinal phenotype-plus-beta data.}

\item{timeVar}{Character. Name of the time variable.}

\item{phenotypes}{Character vector or comma-separated string of phenotype
variables to summarize.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
A data frame with one row per timepoint and summary columns for each
requested phenotype.
}
\description{
Summarize the requested phenotype variables by timepoint. Numeric phenotypes
are reported with mean, standard deviation, and non-missing counts;
non-numeric
phenotypes are reported with non-missing counts and the observed levels.
}
\examples{
ex <- dnaEPICO:::exampleMethylationLMEStateDnaEpico()
timepoint_summary <- summarizeTimepointsMethylationLME(
    data = ex$preparedData$data,
    timeVar = "Timepoint",
    phenotypes = "score",
    verbose = FALSE,
    logs = FALSE
)
nrow(timepoint_summary)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessingPheno_steps.R
\name{buildClockFoundationInputsPreprocessingPheno}
\alias{buildClockFoundationInputsPreprocessingPheno}
\title{Build Clock Foundation input tables from preprocessingPheno data}
\usage{
buildClockFoundationInputsPreprocessingPheno(
  beta,
  pheno,
  SampleID = "Sample_Name",
  sexColumn = "Sex",
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_buildClockFoundationInputsPreprocessingPheno.txt"
)
}
\arguments{
\item{beta}{Numeric matrix of beta values with probes in rows and samples in
columns.}

\item{pheno}{Phenotype data frame aligned with the beta matrix columns.}

\item{SampleID}{Character. Name of the phenotype sample identifier column.}

\item{sexColumn}{Character. Name of the phenotype sex column.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
A list with class \code{'dnaEPICO_preprocessingPheno_clock'}
containing \code{betaCSV} and \code{phenoCF}.
}
\description{
Prepare the beta and phenotype tables commonly exported for Clock Foundation
style downstream workflows, without writing them to disk.
}
\examples{
ex <- dnaEPICO:::examplePreprocessingPhenoStateDnaEpico()
clock_inputs <- buildClockFoundationInputsPreprocessingPheno(
    beta = ex$timepointData$data[["1"]]$beta,
    pheno = ex$timepointData$data[["1"]]$pheno,
    SampleID = "Sample_Name",
    sexColumn = "Sex",
    verbose = FALSE,
    logs = FALSE
)
names(clock_inputs)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessingMinfiEwasWater_steps.R
\name{plotSexMinfiEwasWater}
\alias{plotSexMinfiEwasWater}
\title{Plot predicted or clinical sex from \code{predictSexMinfiEwasWater()}}
\usage{
plotSexMinfiEwasWater(
  sexData,
  type = c("predicted", "clinical"),
  display = FALSE,
  file = NULL,
  width = 2000L,
  height = 1000L,
  res = 70L,
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_plotSexMinfiEwasWater.txt"
)
}
\arguments{
\item{sexData}{Object returned by \code{predictSexMinfiEwasWater()}.}

\item{type}{Character. Plot type: \code{'predicted'} for methylation-predicted
sex or \code{'clinical'} for reported sex.}

\item{display}{Logical. If \code{TRUE}, draw the plot on the active graphics
device.}

\item{file}{Character or \code{NULL}. TIFF file written when supplied.}

\item{width}{Integer. TIFF width in pixels when \code{file} is supplied.}

\item{height}{Integer. TIFF height in pixels when \code{file} is supplied.}

\item{res}{Integer. TIFF resolution in DPI when \code{file} is supplied.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
Invisibly returns the saved TIFF path when \code{file} is supplied,
otherwise \code{NULL}.
}
\description{
Plot predicted or clinical sex from \code{predictSexMinfiEwasWater()}
}
\examples{
ex <- dnaEPICO:::exampleSexPlotStateDnaEpico()
plotSexMinfiEwasWater(
    sexData = ex,
    type = "predicted",
    display = FALSE,
    verbose = FALSE,
    logs = FALSE
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methylationLME_steps.R
\name{writeMethylationLMEOutputs}
\alias{writeMethylationLMEOutputs}
\title{Write optional disk outputs for longitudinal mixed-effects analyses}
\usage{
writeMethylationLMEOutputs(
  modelResults,
  modelSummaries,
  annotatedResults,
  significantInteractions = NULL,
  outputRData,
  summaryTxtDir,
  significantInteractionDir,
  annotatedLMEOut,
  reportAssetsDir = NULL,
  saveTxtSummaries = TRUE,
  saveSignificantInteractions = FALSE,
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_methylationLME.txt"
)
}
\arguments{
\item{modelResults}{Object returned by \code{fitMethylationLMEModels()}.}

\item{modelSummaries}{Object returned by \code{summarizeMethylationLMEModels()}.}

\item{annotatedResults}{Object returned by
\code{annotateMethylationLMESummaries()} or a compatible data frame.}

\item{significantInteractions}{Object returned by
\code{collectSignificantInteractionsMethylationLME()} or \code{NULL}.}

\item{outputRData}{Character. Directory used for serialized model and summary
outputs.}

\item{summaryTxtDir}{Character. Directory used for tab-delimited summary
tables.}

\item{significantInteractionDir}{Character. Directory used for significant
interaction coefficient tables.}

\item{annotatedLMEOut}{Character. Directory used for the annotated summary
XLSX workbook.}

\item{reportAssetsDir}{Character or \code{NULL}. Report results directory used for
the compressed TSV table and compact metadata sidecars.}

\item{saveTxtSummaries}{Logical. If \code{TRUE}, write tab-delimited summary
tables.}

\item{saveSignificantInteractions}{Logical. If \code{TRUE}, write significant
interaction coefficient tables.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
A list with class \code{'dnaEPICO_methylationLME_paths'}
containing the paths of the files written to disk, including the annotated
workbook and any requested report-table sidecars.
}
\description{
Write optional serialized outputs, summary tables, significant interaction
tables, and annotated results from the longitudinal mixed-effects workflow.
}
\examples{
ex <- dnaEPICO:::exampleMethylationLMEStateDnaEpico()
annotation_data <- annotateMethylationLMESummaries(
    modelSummaries = ex$modelSummaries,
    annotationObject = ex$annotationData,
    annotationCols = "Name,chr,pos",
    verbose = FALSE,
    logs = FALSE
)
significant_hits <- collectSignificantInteractionsMethylationLME(
    modelResults = ex$modelResults,
    pvalThreshold = 1,
    verbose = FALSE,
    logs = FALSE
)
output_paths <- writeMethylationLMEOutputs(
    modelResults = ex$modelResults,
    modelSummaries = ex$modelSummaries,
    annotatedResults = annotation_data,
    significantInteractions = significant_hits,
    outputRData = file.path(ex$tempDir, "models"),
    summaryTxtDir = file.path(ex$tempDir, "summary"),
    significantInteractionDir = file.path(ex$tempDir, "significant"),
    annotatedLMEOut = file.path(ex$tempDir, "annotated"),
    saveTxtSummaries = TRUE,
    saveSignificantInteractions = TRUE,
    verbose = FALSE,
    logs = FALSE
)
names(output_paths)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessingMinfiEwasWater_steps.R
\name{predictSexMinfiEwasWater}
\alias{predictSexMinfiEwasWater}
\title{Predict biological sex from a filtered raw-data object}
\usage{
predictSexMinfiEwasWater(
  rawData,
  targets,
  SampleID = "Sample_Name",
  sexColumn = "Sex",
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_predictSexMinfiEwasWater.txt"
)
}
\arguments{
\item{rawData}{Object returned by \code{buildRawMinfiEwasWater()}.}

\item{targets}{Filtered phenotype data frame aligned with \code{rawData}.}

\item{SampleID}{Character. Name of the sample identifier column in \code{targets}.}

\item{sexColumn}{Character. Name of the phenotype column containing reported
sex.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
A list with class \code{'dnaEPICO_minfiEwasWater_sex'} containing the sex
prediction result, aligned phenotype data, plotting data, and mismatch
table.
}
\description{
Predict sample sex from a genome-mapped methylation object, align the
predictions with phenotype data, and return a structured object that can be
plotted or merged into downstream phenotype tables.
}
\examples{
\dontshow{if (requireNamespace("minfiData", quietly = TRUE)) withAutoprint(\{ # examplesIf}
ex <- dnaEPICO:::exampleMinfiWorkflowStateDnaEpico()
sex_data <- predictSexMinfiEwasWater(
    rawData = ex$rawFiltered,
    targets = ex$sampleData$targets,
    SampleID = "Sample_Name",
    sexColumn = "Sex",
    verbose = FALSE,
    logs = FALSE
)
names(sex_data)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dnamReport.R
\name{print.dnaEPICO_dnamReport}
\alias{print.dnaEPICO_dnamReport}
\title{Print a DNA methylation report result}
\usage{
\method{print}{dnaEPICO_dnamReport}(x, ...)
}
\arguments{
\item{x}{Object returned by \code{\link[=dnamReport]{dnamReport()}}.}

\item{...}{Additional arguments ignored.}
}
\value{
Invisibly returns \code{x}.
}
\description{
Print a DNA methylation report result
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methylationLME_steps.R
\name{fitMethylationLMEModels}
\alias{fitMethylationLMEModels}
\title{Fit CpG-wise mixed-effects models for longitudinal methylation analyses}
\usage{
fitMethylationLMEModels(
  preparedData,
  nCores = 1L,
  libPath = NULL,
  lmeLibs = "lme4,lmerTest",
  correlationStructure = "none",
  correlationVar = NULL,
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_methylationLME.txt"
)
}
\arguments{
\item{preparedData}{Object returned by \code{prepareMethylationLMEData()}.}

\item{nCores}{Integer. Maximum number of worker processes to use. Automatic
fitting uses the empirical lme4 or nlme crossover and caps workers by
available CpGs and detected physical cores.}

\item{libPath}{Character vector or \code{NULL}. Optional library paths forwarded
to worker processes.}

\item{lmeLibs}{Character vector or comma-separated string of package names to
check on worker processes. The default is \code{'lme4,lmerTest'}.}

\item{correlationStructure}{Character. Residual correlation structure used
when \code{lmeLibs} selects \code{'nlme'}. One of \code{'none'}, \code{'AR1'}, or \code{'CAR1'}.}

\item{correlationVar}{Character or \code{NULL}. Variable used to order repeated
observations within \code{personVar} for \code{AR1} or \code{CAR1} residual correlation
structures. Must be supplied explicitly for \code{AR1} or \code{CAR1}.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
A list with class \code{'dnaEPICO_methylationLME_models'}
containing fitted model lists, model formulas, and counts of failed CpG
fits.
}
\description{
Fit one linear mixed-effects model per CpG for each phenotype requested in
the
object returned by \code{prepareMethylationLMEData()}.
}
\examples{
ex <- dnaEPICO:::exampleMethylationLMEStateDnaEpico()
model_results <- fitMethylationLMEModels(
    preparedData = ex$preparedData,
    nCores = 1,
    verbose = FALSE,
    logs = FALSE
)
names(model_results$fits)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessingMinfiEwasWater_steps.R
\name{assessSamplesMinfiEwasWater}
\alias{assessSamplesMinfiEwasWater}
\title{Assess sample quality before sample filtering}
\usage{
assessSamplesMinfiEwasWater(
  rawData,
  RGSet,
  qcCutoff = 10.5,
  detPtype = "m+u",
  detPThreshold = 0.05,
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_assessSamplesMinfiEwasWater.txt"
)
}
\arguments{
\item{rawData}{Object returned by \code{buildRawMinfiEwasWater()}.}

\item{RGSet}{An \code{RGChannelSet} aligned with \code{rawData}.}

\item{qcCutoff}{Numeric. Cutoff passed to \code{minfi::plotQC()} when the QC plot
is drawn.}

\item{detPtype}{Character. Detection P-value mode passed to
\code{minfi::detectionP()}. Common values in minfi workflows are \code{'m+u'} and
\code{'negative'}. The default used here is \code{'m+u'}.}

\item{detPThreshold}{Numeric. Samples with mean detection P value above this
threshold are marked as failed.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
A list with class \code{'dnaEPICO_minfiEwasWater_assessment'} containing
the QC object, detection P matrix, mean detection P values, and failed
sample identifiers.
}
\description{
Compute minfi QC metrics and detection P values, identify failed samples
using the requested threshold, and return the assessment as a single object.
}
\examples{
\dontshow{if (requireNamespace("minfiData", quietly = TRUE)) withAutoprint(\{ # examplesIf}
ex <- dnaEPICO:::exampleMinfiBaseDataDnaEpico()
raw_data <- buildRawMinfiEwasWater(ex$RGSet, verbose = FALSE, logs = FALSE)
assessment <- assessSamplesMinfiEwasWater(
    rawData = raw_data,
    RGSet = ex$RGSet,
    detPThreshold = 1,
    verbose = FALSE,
    logs = FALSE
)
names(assessment)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/result_classes.R
\docType{class}
\name{dnaEPICO_dnamReport_prepared-class}
\alias{dnaEPICO_dnamReport_prepared-class}
\alias{dnaEPICO_dnamReport_prepared}
\title{Result class returned by prepareDnamReportInputs}
\description{
Objects of class \code{'dnaEPICO_dnamReport_prepared'} are list-based results
returned by \code{\link[=prepareDnamReportInputs]{prepareDnamReportInputs()}}. They capture normalized report
paths,
available figures, and logging metadata before rendering.
}
\section{Structure}{

\describe{
\item{output}{Requested output file name.}
\item{outputDir}{Normalized report output directory.}
\item{outputFile}{Normalized full path to the intended report output file.}
\item{figDir}{Normalized directory used by the report template for copied
figures.}
\item{figureInventory}{Named list describing the available figures for each
report section.}
\item{missingFigureDirectories}{Character vector of expected figure
directories that were not present at preparation time.}
\item{logFile}{Resolved path to the optional log file, or \code{NULL} when
logging was disabled.}
}
}

\seealso{
\code{\link[=prepareDnamReportInputs]{prepareDnamReportInputs()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methylationLME.R
\name{methylationLME}
\alias{methylationLME}
\title{Fit CpG-wise linear mixed-effects models for longitudinal methylation
analyses}
\usage{
methylationLME(
  inputPheno = "rData/preprocessingPheno/mergeData/phenoBT1T2.RData",
  outputLogs = "logs",
  outputRData = "rData/methylationLME/models",
  outputPlots = "figures/methylationLME",
  personVar = "person",
  timeVar = "Timepoint",
  phenotypes = c("DASS_Depression", "DASS_Anxiety", "DASS_Stress", "PCL5_TotalScore",
    "MHCSF_TotalScore", "BRS_TotalScore"),
  covariates = "Sex,Age,Ethnicity,TraumaDefinition,Leukocytes,Epithelial.cells",
  factorVars = "Sex,Ethnicity,TraumaDefinition",
  scaleVars = NULL,
  lmeLibs = "lme4,lmerTest",
  correlationStructure = "none",
  correlationVar = NULL,
  prsMap = NULL,
  libPath = NULL,
  cpgPrefix = "cg",
  cpgLimit = NA,
  methylationScale = "beta",
  nCores = 32,
  summaryPval = NA,
  excludeSingular = FALSE,
  excludeNonConverged = FALSE,
  plotWidth = 2000,
  plotHeight = 1000,
  plotDPI = 150,
  interactionTerm = NULL,
  saveSignificantInteractions = TRUE,
  significantInteractionDir = "preliminaryResults/cpgs/methylationLME",
  significantInteractionPval = 0.05,
  saveTxtSummaries = TRUE,
  chunkSize = NULL,
  summaryTxtDir = "preliminaryResults/summary/methylationLME",
  fdrThreshold = 0.05,
  padjmethod = "fdr",
  annotationPackage = "IlluminaHumanMethylationEPICv2anno.20a1.hg38",
  annotationCols = c("Name", "chr", "pos", "UCSC_RefGene_Group", "UCSC_RefGene_Name",
    "Relation_to_Island", "GencodeV41_Group"),
  annotatedLMEOut = "data/methylationLME",
  reportAssetsDir = NULL,
  display = FALSE,
  verbose = FALSE,
  logs = FALSE,
  saveOutputs = FALSE
)
}
\arguments{
\item{inputPheno}{Character. Path to the merged longitudinal
phenotype-plus-methylation \code{.RData} or \code{.rds} object created by
\code{preprocessingPheno()}. The default points to the combined timepoint object
produced by the package workflow.}

\item{outputLogs}{Character. Directory used for optional log files.}

\item{outputRData}{Character. Directory used for optional serialized
mixed-model
and summary outputs.}

\item{outputPlots}{Character. Directory used for optional TIFF diagnostic
plots.}

\item{personVar}{Character. Subject identifier variable used for the random
intercept. When this column is missing, it is derived from \code{SID} using the
package's existing sample naming convention.}

\item{timeVar}{Character. Name of the longitudinal time variable used for
timepoint summaries and preprocessing checks.}

\item{phenotypes}{Character vector or comma-separated phenotype variables to
model.}

\item{covariates}{Character. Comma-separated fixed-effect covariates included
in every mixed model.}

\item{factorVars}{Character. Comma-separated variables that should be coerced
to factors before modeling. This usually includes categorical phenotypes,
covariates, or interaction variables.}

\item{scaleVars}{Character vector, comma-separated variable names, or \code{NULL}.
Numeric fixed-effect variables to centre and divide by their sample standard
deviations before fitting.}

\item{lmeLibs}{Character. Comma-separated package names to validate on worker
processes and select the LME backend. Use \code{'lme4,lmerTest'} or \code{'lme4'}
for the \code{lmerTest}/\code{lme4} path, or \code{'nlme'} for the \code{nlme::lme()} path.}

\item{correlationStructure}{Character. Residual correlation structure used
when \code{lmeLibs = 'nlme'}. One of \code{'none'}, \code{'AR1'}, or \code{'CAR1'}. The
default is \code{'none'}.}

\item{correlationVar}{Character or \code{NULL}. Variable used to order repeated
observations within \code{personVar} for \code{AR1} or \code{CAR1} residual correlation
structures. Must be supplied explicitly for \code{AR1} or \code{CAR1}.}

\item{prsMap}{Character or \code{NULL}. Optional phenotype-to-PRS mapping in the
form \code{'Phenotype1:PRS_1,Phenotype2:PRS_2'}.}

\item{libPath}{Character vector or \code{NULL}. Optional library paths forwarded
to
worker processes. By default, the current \code{.libPaths()} are used.}

\item{cpgPrefix}{Character. Prefix used to identify methylation columns in
the
merged phenotype-plus-methylation input object. The default is \code{'cg'}.}

\item{cpgLimit}{Integer or \code{NA}. Maximum number of CpGs to analyse. Use \code{NA}
to keep all CpGs matching \code{cpgPrefix}.}

\item{methylationScale}{Character. Methylation metric represented by the CpG
columns. One of \code{'Beta'}, \code{'M'}, or \code{'CN'}, in any combination of
upper- and lower-case letters. The default is \code{'beta'}.}

\item{nCores}{Integer. Number of worker processes to use while fitting models
and extracting summaries. Automatic fitting uses engine-specific empirical
crossovers and caps workers by the CpG workload and detected physical cores.}

\item{summaryPval}{Numeric or \code{NA}. Optional p-value threshold applied to the
returned longitudinal CpG summary tables. Use \code{NA} to keep all summary rows.}

\item{excludeSingular}{Logical. If \code{TRUE}, retain singular lme4 CpG rows and
reasons but exclude their p-values from inference.}

\item{excludeNonConverged}{Logical. If \code{TRUE}, retain non-converged CpG rows
and reasons but exclude their p-values from inference.}

\item{plotWidth}{Integer. TIFF width in pixels when plots are written to
disk.}

\item{plotHeight}{Integer. TIFF height in pixels when plots are written to
disk.}

\item{plotDPI}{Integer. TIFF resolution in DPI when plots are written to
disk.}

\item{interactionTerm}{Character or \code{NULL}. Optional interaction term. When
supplied and present in the input data, the phenotype is modeled together
with its interaction against this variable.}

\item{saveSignificantInteractions}{Logical. If \code{TRUE}, collect coefficient
tables for CpGs passing \code{significantInteractionPval} in the returned object
and optionally write them to disk when \code{saveOutputs = TRUE}.}

\item{significantInteractionDir}{Character. Directory used for optional
significant-interaction coefficient tables.}

\item{significantInteractionPval}{Numeric. P-value threshold used to collect
or
write significant interaction coefficient tables.}

\item{saveTxtSummaries}{Logical. If \code{TRUE} and \code{saveOutputs = TRUE}, write
tab-delimited summary tables to \code{summaryTxtDir}.}

\item{chunkSize}{Integer or \code{NULL}. Number of CpGs processed per summary
extraction chunk. \code{NULL} chooses a value automatically.}

\item{summaryTxtDir}{Character. Directory used for optional tab-delimited LME
summary tables.}

\item{fdrThreshold}{Numeric. False-discovery-rate threshold used to highlight
CpGs in the residual-significance diagnostic plots.}

\item{padjmethod}{Character. P-value adjustment method passed to
\code{stats::p.adjust()}. The default is \code{'fdr'}.}

\item{annotationPackage}{Character. Annotation package or object name passed
to
\code{minfi::getAnnotation()}, for example
\code{'IlluminaHumanMethylationEPICv2anno.20a1.hg38'}.}

\item{annotationCols}{Character vector or comma-separated annotation columns
to append to the combined LME summary table. Available columns depend on
the selected annotation package.}

\item{annotatedLMEOut}{Character. Directory used for the optional annotated
LME
summary XLSX workbook.}

\item{reportAssetsDir}{Character or \code{NULL}. Report results directory used for
the compressed TSV table and its compact metadata sidecars. \code{NULL} writes
only the model outputs and annotated workbook.}

\item{display}{Logical. If \code{TRUE}, draw diagnostic plots on the active
graphics device.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.
The
default is \code{FALSE}, so the function is quiet unless requested.}

\item{logs}{Logical. If \code{TRUE}, write the same progress messages to
\code{file.path(outputLogs, 'log_methylationLME.txt')}.}

\item{saveOutputs}{Logical. If \code{TRUE}, write optional serialized model files,
summary tables, significant-interaction tables, annotated results, and TIFF
plots to the requested output directories. The default is \code{FALSE}, so the
function returns in-memory results without writing files.}
}
\value{
A list with class \code{'dnaEPICO_methylationLME'}.
\describe{
\item{preparedData}{Object returned by \code{\link[=prepareMethylationLMEData]{prepareMethylationLMEData()}}
containing the merged longitudinal phenotype-plus-methylation analysis table
and
modeling metadata.}
\item{modelFits}{Object returned by \code{\link[=fitMethylationLMEModels]{fitMethylationLMEModels()}}
containing the per-phenotype CpG mixed-effects model fits.}
\item{modelSummaries}{Object returned by
\code{\link[=summarizeMethylationLMEModels]{summarizeMethylationLMEModels()}} containing the combined CpG summary
tables used for reporting and annotation.}
\item{significantInteractions}{Object returned by
\code{\link[=collectSignificantInteractionsMethylationLME]{collectSignificantInteractionsMethylationLME()}} containing optional
phenotype-specific significant-interaction tables.}
\item{diagnosticPlots}{Object returned by
\code{\link[=plotMethylationLMEDiagnostics]{plotMethylationLMEDiagnostics()}} describing the diagnostic plot
objects and any written TIFF files.}
\item{annotation}{Object returned by
\code{\link[=annotateMethylationLMESummaries]{annotateMethylationLMESummaries()}} containing the annotated combined
summary table.}
\item{savedFiles}{Object returned by
\code{\link[=writeMethylationLMEOutputs]{writeMethylationLMEOutputs()}} when \code{saveOutputs = TRUE}, otherwise
\code{NULL}.}
\item{runSettings}{High-level run metadata including the generic analysis
label, methylation scale, display label, selected merged-object prefix,
internal response-column name, and longitudinal time variable.}
}
See \linkS4class{dnaEPICO_methylationLME} for a class-level overview.
}
\description{
\code{methylationLME()} is the high-level coordinator for the longitudinal
linear mixed-effects stage of the \code{dnaEPICO} workflow. It prepares the merged
phenotype-plus-methylation input, fits one mixed-effects model per CpG for
each
requested phenotype, extracts phenotype-specific coefficient summaries,
optionally collects significant interaction tables, generates diagnostic
plots,
annotates the combined summary table, and optionally writes model outputs to
disk. By default, the function runs quietly and returns its results in
memory.
}
\examples{
if (
    requireNamespace("IlluminaHumanMethylation450kanno.ilmn12.hg19", quietly = TRUE) &&
        requireNamespace("lmerTest", quietly = TRUE)
) {
    tmp <- tempdir()
    toy_path <- file.path(tmp, "phenoBT1T2.RData")
    phenoBT1T2 <- data.frame(
        SID = c("P1A", "P1B", "P2A", "P2B", "P3A", "P3B", "P4A", "P4B"),
        person = c(1, 1, 2, 2, 3, 3, 4, 4),
        Timepoint = factor(c("1", "2", "1", "2", "1", "2", "1", "2")),
        score = c(10, 12, 9, 11, 13, 14, 8, 9),
        sex = factor(c("F", "F", "M", "M", "F", "F", "M", "M")),
        cg00000029 = c(0.25, 0.27, 0.20, 0.22, 0.30, 0.31, 0.18, 0.20),
        cg00000108 = c(0.50, 0.53, 0.55, 0.57, 0.48, 0.49, 0.60, 0.61),
        check.names = FALSE
    )
    save(phenoBT1T2, file = toy_path)

    result <- methylationLME(
        inputPheno = toy_path,
        phenotypes = "score",
        covariates = "sex",
        factorVars = "sex",
        cpgLimit = 2,
        nCores = 1,
        summaryPval = 1,
        annotationPackage = "IlluminaHumanMethylation450kanno.ilmn12.hg19",
        annotationCols = "Name,chr,pos",
        display = FALSE,
        verbose = FALSE,
        logs = FALSE,
        saveOutputs = FALSE
    )

    class(result)
}

}
\seealso{
\linkS4class{dnaEPICO_methylationLME}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extractMake.R
\name{extractMake}
\alias{extractMake}
\title{Copy dnaEPICO Makefile to a user directory}
\usage{
extractMake(destDir, overwrite = FALSE)
}
\arguments{
\item{destDir}{Character. Destination directory where the Makefile
will be copied.}

\item{overwrite}{Logical. Whether to overwrite an existing \code{Makefile} in
\code{destDir}. The default is \code{FALSE}.}
}
\value{
Character scalar containing the path to the copied \code{Makefile}.
}
\description{
Copies the example Makefile pipeline shipped with dnaEPICO to a
user-specified directory for local execution or modification.
}
\examples{
tmp <- file.path(tempdir(), "dnaEPICO-make-example")
dir.create(tmp, recursive = TRUE, showWarnings = FALSE)
makefile_path <- extractMake(
    destDir = tmp,
    overwrite = TRUE
)
stopifnot(file.exists(makefile_path))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessingMinfiEwasWater_steps.R
\name{plotNormalizationMinfiEwasWater}
\alias{plotNormalizationMinfiEwasWater}
\title{Plot raw and normalized methylation distributions}
\usage{
plotNormalizationMinfiEwasWater(
  RGSet,
  normData,
  targets,
  sexColumn = "Sex",
  display = FALSE,
  file = NULL,
  width = 2000L,
  height = 1000L,
  res = 150L,
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_plotNormalizationMinfiEwasWater.txt"
)
}
\arguments{
\item{RGSet}{An \code{RGChannelSet} aligned with \code{targets}.}

\item{normData}{Object returned by \code{normalizeMinfiEwasWater()}.}

\item{targets}{Filtered phenotype data aligned with \code{RGSet}.}

\item{sexColumn}{Character. Name of the phenotype column used to colour the
density curves.}

\item{display}{Logical. If \code{TRUE}, draw the plot on the active graphics
device.}

\item{file}{Character or \code{NULL}. TIFF file written when supplied.}

\item{width}{Integer. TIFF width in pixels when \code{file} is supplied.}

\item{height}{Integer. TIFF height in pixels when \code{file} is supplied.}

\item{res}{Integer. TIFF resolution in DPI when \code{file} is supplied.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
Invisibly returns the saved TIFF path when \code{file} is supplied,
otherwise \code{NULL}.
}
\description{
Draw the density comparison plot used to inspect raw versus normalized data.
}
\examples{
ex <- dnaEPICO:::exampleMinfiMetricsStateDnaEpico()
plotNormalizationMinfiEwasWater(
    RGSet = ex$beta,
    normData = ex$normData,
    targets = ex$targets,
    sexColumn = "Sex",
    display = FALSE,
    verbose = FALSE,
    logs = FALSE
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/result_classes.R
\docType{class}
\name{dnaEPICO_dnamReport-class}
\alias{dnaEPICO_dnamReport-class}
\alias{dnaEPICO_dnamReport}
\title{Result class returned by dnamReport}
\description{
Objects of class \code{'dnaEPICO_dnamReport'} are list-based results returned by
\code{\link[=dnamReport]{dnamReport()}}. They combine the prepared report inputs, render result, and
final status metadata into one convenience object.
}
\section{Structure}{

\describe{
\item{preparedReport}{Object returned by \code{\link[=prepareDnamReportInputs]{prepareDnamReportInputs()}}.}
\item{renderResult}{Structured render metadata created by \code{\link[=dnamReport]{dnamReport()}}.}
\item{status}{Final status string such as \code{'rendered'}, \code{'skipped'}, or
\code{'failed'}.}
\item{outputFile}{Path to \code{docs/index.html}.}
\item{resultTableSources}{Named character vector identifying whether each
model table used a streamed sidecar or the compatible XLSX fallback.}
\item{errorMessage}{Final render error or skip message when available,
otherwise \code{NULL}.}
\item{logFile}{Resolved path to the optional log file, or \code{NULL} when
logging was disabled.}
}
}

\seealso{
\code{\link[=dnamReport]{dnamReport()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessingMinfiEwasWater_steps.R
\name{filterProbesMinfiEwasWater}
\alias{filterProbesMinfiEwasWater}
\title{Filter probes from a normalized methylation object}
\usage{
filterProbesMinfiEwasWater(
  normData,
  RGSet,
  pvalThreshold = 0.01,
  chrToRemove = "chrX,chrY",
  snpsToRemove = "SBE,CpG",
  mafThreshold = 0.1,
  probeExclusionPath,
  probeExclusionIdColumn = NULL,
  useEpicV2Manifest = FALSE,
  epicV2ManifestFlags = c(CH_WGBS_evidence = TRUE, CH_BLAT = TRUE, MissingPos = TRUE,
    MismatchPos = FALSE),
  detPtype = "m+u",
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_filterProbesMinfiEwasWater.txt",
  crossReactivePath = NULL,
  crossReactiveIdColumn = NULL
)
}
\arguments{
\item{normData}{Object returned by \code{normalizeMinfiEwasWater()}.}

\item{RGSet}{Filtered \code{RGChannelSet} aligned with \code{normData}.}

\item{pvalThreshold}{Numeric. Probes must have detection P values below this
threshold in all samples to be retained.}

\item{chrToRemove}{Character vector or comma-separated string of chromosome
names to remove, for example \code{'chrX,chrY'}.}

\item{snpsToRemove}{Character vector or comma-separated string of SNP probe
types to remove, for example \code{'SBE,CpG'}.}

\item{mafThreshold}{Numeric. Minor allele frequency threshold passed to
\code{minfi::dropLociWithSnps()}.}

\item{probeExclusionPath}{Character vector or semicolon-separated string of
CSV files containing probe IDs to remove.}

\item{probeExclusionIdColumn}{Character or \code{NULL}. Column containing probe
IDs. When \code{NULL} or \code{''}, each file is auto-detected using \code{ProbeID},
\code{TargetID}, \code{IlmnID}, or \code{Name}, then falling back to an unlabeled or
probe-like first column.}

\item{useEpicV2Manifest}{Logical. If \code{TRUE}, also remove EPICv2 probes
flagged in the Peters et al. expanded manifest from AnnotationHub resource
\code{AH116484}.}

\item{epicV2ManifestFlags}{Named logical vector controlling which EPICv2
manifest flags are removed. Defaults remove \code{CH_WGBS_evidence}, \code{CH_BLAT},
and \code{MissingPos}, but not \code{MismatchPos}.}

\item{detPtype}{Character. Detection P-value mode passed to
\code{minfi::detectionP()} for the probe filter. Common values in minfi
workflows are \code{'m+u'} and \code{'negative'}.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}

\item{crossReactivePath}{Deprecated alias for \code{probeExclusionPath}.}

\item{crossReactiveIdColumn}{Deprecated alias for \code{probeExclusionIdColumn}.}
}
\value{
A list with class \code{'dnaEPICO_minfiEwasWater_filter'} containing the
filtered object and counts for each filtering stage.
}
\description{
Apply detection P-value, chromosome, SNP, and probe-exclusion filters to the
primary normalized object and return the filtered result.
}
\examples{
\dontshow{if (requireNamespace("minfiData", quietly = TRUE)) withAutoprint(\{ # examplesIf}
ex <- dnaEPICO:::exampleMinfiWorkflowStateDnaEpico()
filtered_data <- filterProbesMinfiEwasWater(
    normData = ex$normData,
    RGSet = ex$sampleData$RGSet,
    pvalThreshold = 1,
    chrToRemove = "chrY",
    snpsToRemove = "SBE",
    mafThreshold = 1,
    probeExclusionPath = ex$probeExclusionPath,
    detPtype = "m+u",
    verbose = FALSE,
    logs = FALSE
)
filtered_data$counts[["probeExclusion"]]
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessingPheno_steps.R
\name{combineTimepointsPreprocessingPheno}
\alias{combineTimepointsPreprocessingPheno}
\title{Combine selected timepoints for downstream longitudinal modeling}
\usage{
combineTimepointsPreprocessingPheno(
  timepointData,
  combineTimepoints = "1,2",
  methylationScale = "beta",
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_combineTimepointsPreprocessingPheno.txt"
)
}
\arguments{
\item{timepointData}{Object returned by
\code{splitTimepointsPreprocessingPheno()}.}

\item{combineTimepoints}{Character vector or comma-separated string of
timepoints to combine.}

\item{methylationScale}{Character. Methylation metric used in the merged
modeling table. One of \code{'Beta'}, \code{'M'}, or \code{'CN'}, case-insensitively.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
A list with class \code{'dnaEPICO_preprocessingPheno_combined'}
containing the combined phenotype table, merged phenotype-plus-beta table,
selected timepoints, and output suffix.
}
\description{
Combine selected timepoints that were already aligned by
\code{splitTimepointsPreprocessingPheno()} into the wide phenotype-plus-beta
objects used by downstream longitudinal models.
}
\examples{
ex <- dnaEPICO:::examplePreprocessingPhenoStateDnaEpico()
combined_data <- combineTimepointsPreprocessingPheno(
    timepointData = ex$timepointData,
    combineTimepoints = "1,2",
    verbose = FALSE,
    logs = FALSE
)
combined_data$suffix

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessingMinfiEwasWater_steps.R
\name{normalizeMinfiEwasWater}
\alias{normalizeMinfiEwasWater}
\title{Normalize filtered samples with minfi and wateRmelon methods}
\usage{
normalizeMinfiEwasWater(
  sampleData,
  sexColumn = "Sex",
  normMethods = "adjustedfunnorm",
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_normalizeMinfiEwasWater.txt"
)
}
\arguments{
\item{sampleData}{Object returned by \code{filterSamplesMinfiEwasWater()}.}

\item{sexColumn}{Character. Name of the phenotype column used as the optional
sex covariate for normalization methods that support it.}

\item{normMethods}{Character vector or semicolon-separated string of
normalization methods. Supported values are \code{'adjustedfunnorm'},
\code{'funnorm'}, \code{'illumina'}, \code{'quantile'}, and \code{'swan'}.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
A list with class \code{'dnaEPICO_minfiEwasWater_norm'} containing the
requested normalized objects and the first method as \code{primary}.
}
\description{
Apply one or more supported normalization methods to a filtered \code{RGSet} and
return all normalized objects together in a single result object.
}
\examples{
\dontshow{if (requireNamespace("minfiData", quietly = TRUE)) withAutoprint(\{ # examplesIf}
ex <- dnaEPICO:::exampleMinfiBaseDataDnaEpico()
sample_data <- filterSamplesMinfiEwasWater(
    RGSet = ex$RGSet,
    targets = ex$targets,
    failedSamples = character(0),
    SampleID = "Sample_Name",
    verbose = FALSE,
    logs = FALSE
)
norm_data <- normalizeMinfiEwasWater(
    sampleData = sample_data,
    sexColumn = "Sex",
    normMethods = "quantile",
    verbose = FALSE,
    logs = FALSE
)
names(norm_data$normalized)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methylationGLM.R
\name{methylationGLM}
\alias{methylationGLM}
\title{Fit CpG-wise GLMs for one-timepoint methylation analyses}
\usage{
methylationGLM(
  inputPheno = "rData/preprocessingPheno/mergeData/phenoBT1.RData",
  outputLogs = "logs",
  outputRData = "rData/methylationGLM/models",
  outputPlots = "figures/methylationGLM",
  phenotypes = c("DASS_Depression", "DASS_Anxiety", "DASS_Stress", "PCL5_TotalScore",
    "MHCSF_TotalScore", "BRS_TotalScore"),
  covariates = "Sex,Age,Ethnicity,TraumaDefinition,Leukocytes,Epithelial.cells",
  factorVars = "Sex,Ethnicity,TraumaDefinition",
  scaleVars = NULL,
  cpgPrefix = "cg",
  cpgLimit = NA,
  methylationScale = "beta",
  nCores = 32,
  plotWidth = 2000,
  plotHeight = 1000,
  plotDPI = 150,
  interactionTerm = NULL,
  libPath = NULL,
  glmLibs = "glm2",
  prsMap = NULL,
  summaryPval = NA,
  summaryResidualSD = TRUE,
  excludeNonConverged = FALSE,
  saveSignificantCpGs = FALSE,
  significantCpGDir = "preliminaryResults/cpgs/methylationGLM",
  significantCpGPval = 0.05,
  saveTxtSummaries = TRUE,
  chunkSize = NULL,
  summaryTxtDir = "preliminaryResults/summary/methylationGLM",
  fdrThreshold = 0.05,
  padjmethod = "fdr",
  annotationPackage = "IlluminaHumanMethylationEPICv2anno.20a1.hg38",
  annotationCols = c("Name", "chr", "pos", "UCSC_RefGene_Group", "UCSC_RefGene_Name",
    "Relation_to_Island", "GencodeV41_Group"),
  annotatedGLMOut = "data/methylationGLM",
  reportAssetsDir = NULL,
  display = FALSE,
  verbose = FALSE,
  logs = FALSE,
  saveOutputs = FALSE
)
}
\arguments{
\item{inputPheno}{Character. Path to the merged phenotype-plus-methylation
\code{.RData}
or \code{.rds} object created by \code{preprocessingPheno()}. The default points to
the timepoint-1 object produced by the package workflow.}

\item{outputLogs}{Character. Directory used for optional log files.}

\item{outputRData}{Character. Directory used for optional serialized model
and
summary outputs.}

\item{outputPlots}{Character. Directory used for optional TIFF plots.}

\item{phenotypes}{Character vector or comma-separated phenotype variables to
model.}

\item{covariates}{Character. Comma-separated covariate variables included in
each GLM.}

\item{factorVars}{Character. Comma-separated variables that should be treated
as factors before modeling.}

\item{scaleVars}{Character vector, comma-separated variable names, or \code{NULL}.
Numeric fixed-effect variables to centre and divide by their sample standard
deviations before model fitting.}

\item{cpgPrefix}{Character. Prefix used to identify methylation columns in
the
merged phenotype-plus-methylation input object. The default is \code{'cg'}.}

\item{cpgLimit}{Integer or \code{NA}. Maximum number of CpGs to analyse. Use \code{NA}
to keep all CpGs matching \code{cpgPrefix}.}

\item{methylationScale}{Character. Methylation metric represented by the CpG
columns. One of \code{'Beta'}, \code{'M'}, or \code{'CN'}, in any combination of
upper- and lower-case letters. The default is \code{'beta'}.}

\item{nCores}{Integer. Number of worker processes to use while fitting models
and extracting summaries. Automatic fitting remains serial below the
empirical glm2 crossover and caps workers by the CpG workload and detected
physical cores.}

\item{plotWidth}{Integer. TIFF width in pixels when plots are written to
disk.}

\item{plotHeight}{Integer. TIFF height in pixels when plots are written to
disk.}

\item{plotDPI}{Integer. TIFF resolution in DPI when plots are written to
disk.}

\item{interactionTerm}{Character or \code{NULL}. Optional interaction term. When
supplied and present in the input data, the phenotype is modeled together
with its interaction against this variable.}

\item{libPath}{Character vector or \code{NULL}. Optional library paths forwarded
to
worker processes. By default, the current \code{.libPaths()} are used.}

\item{glmLibs}{Character. Comma-separated package names to validate on worker
processes. The default is \code{'glm2'}.}

\item{prsMap}{Character or \code{NULL}. Optional phenotype-to-PRS mapping in the
form \code{'Phenotype1:PRS_1,Phenotype2:PRS_2'}.}

\item{summaryPval}{Numeric or \code{NA}. Optional p-value threshold applied to the
returned CpG summary tables. Use \code{NA} to keep all summary rows.}

\item{summaryResidualSD}{Logical. If \code{TRUE}, append residual standard
deviations to the CpG summary tables and residual diagnostic plots.}

\item{excludeNonConverged}{Logical. If \code{TRUE}, retain non-converged CpG rows
and reasons but exclude their p-values from inference.}

\item{saveSignificantCpGs}{Logical. If \code{TRUE}, collect significant CpG
coefficient tables in the returned object and optionally write them to disk
when \code{saveOutputs = TRUE}.}

\item{significantCpGDir}{Character. Directory used for optional significant
CpG coefficient tables.}

\item{significantCpGPval}{Numeric. P-value threshold used to collect or write
significant CpG coefficient tables.}

\item{saveTxtSummaries}{Logical. If \code{TRUE} and \code{saveOutputs = TRUE}, write
tab-delimited summary tables to \code{summaryTxtDir}.}

\item{chunkSize}{Integer or \code{NULL}. Number of CpGs processed per summary
extraction chunk. \code{NULL} chooses a value automatically.}

\item{summaryTxtDir}{Character. Directory used for optional tab-delimited GLM
summary tables.}

\item{fdrThreshold}{Numeric. False-discovery-rate threshold used to highlight
CpGs in the residual-significance diagnostic plots.}

\item{padjmethod}{Character. P-value adjustment method passed to
\code{stats::p.adjust()}. The default is \code{'fdr'}.}

\item{annotationPackage}{Character. Annotation package or object name passed
to \code{minfi::getAnnotation()}, for example
\code{'IlluminaHumanMethylationEPICv2anno.20a1.hg38'}.}

\item{annotationCols}{Character vector or comma-separated annotation columns
to append to the combined GLM summary table. Available columns depend on
the selected annotation package.}

\item{annotatedGLMOut}{Character. Directory used for the optional annotated
GLM summary XLSX workbook.}

\item{reportAssetsDir}{Character or \code{NULL}. Report results directory used for
the compressed TSV table and its compact metadata sidecars. \code{NULL} writes
only the model outputs and annotated workbook.}

\item{display}{Logical. If \code{TRUE}, draw exploratory and diagnostic plots on
the active graphics device.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.
The default is \code{FALSE}, so the function is quiet unless requested.}

\item{logs}{Logical. If \code{TRUE}, write the same progress messages to
\code{file.path(outputLogs, 'log_methylationGLM.txt')}.}

\item{saveOutputs}{Logical. If \code{TRUE}, write optional serialized model files,
summary tables, significant-CpG tables, annotated results, and TIFF plots to
the requested output directories. The default is \code{FALSE}, so the function
returns in-memory results without writing files.}
}
\value{
A list with class \code{'dnaEPICO_methylationGLM'}.
\describe{
\item{preparedData}{Object returned by \code{\link[=prepareMethylationGLMData]{prepareMethylationGLMData()}}
containing the merged phenotype-plus-methylation analysis table and modeling
metadata.}
\item{distributionPlots}{Object returned by
\code{\link[=plotMethylationGLMDistributions]{plotMethylationGLMDistributions()}} describing any exploratory plots that
were generated or written.}
\item{modelFits}{Object returned by \code{\link[=fitMethylationGLMModels]{fitMethylationGLMModels()}}
containing the per-phenotype CpG model fits.}
\item{modelSummaries}{Object returned by
\code{\link[=summarizeMethylationGLMModels]{summarizeMethylationGLMModels()}} containing the combined CpG summary
tables used for reporting and annotation.}
\item{significantCpGs}{Object returned by
\code{\link[=collectSignificantCpGsMethylationGLM]{collectSignificantCpGsMethylationGLM()}} containing optional
phenotype-specific significant-CpG tables.}
\item{diagnosticPlots}{Object returned by
\code{\link[=plotMethylationGLMDiagnostics]{plotMethylationGLMDiagnostics()}} describing the diagnostic plot objects
and any written TIFF files.}
\item{annotation}{Object returned by
\code{\link[=annotateMethylationGLMSummaries]{annotateMethylationGLMSummaries()}} containing the annotated combined
summary table.}
\item{savedFiles}{Object returned by \code{\link[=writeMethylationGLMOutputs]{writeMethylationGLMOutputs()}} when
\code{saveOutputs = TRUE}, otherwise \code{NULL}.}
\item{runSettings}{High-level run metadata including the generic analysis
label, methylation scale, display label, selected merged-object prefix, and
internal response-column name.}
}
See \linkS4class{dnaEPICO_methylationGLM} for a class-level overview.
}
\description{
\code{methylationGLM()} is the high-level coordinator for the methylation GLM
stage of the \code{dnaEPICO} workflow. It prepares the merged
phenotype-plus-methylation
input, optionally creates exploratory plots, fits one Gaussian GLM per CpG
for
each requested phenotype, extracts CpG-level summaries, optionally collects
significant CpG coefficient tables, generates diagnostic plots, annotates the
combined summary table, and optionally writes model outputs to disk. By
default, the function runs quietly and returns its results in memory.
}
\examples{
if (requireNamespace("IlluminaHumanMethylation450kanno.ilmn12.hg19", quietly = TRUE)) {
    tmp <- tempdir()
    toy_path <- file.path(tmp, "phenoBT1.RData")
    phenoBT1 <- data.frame(
        Sample_Name = c("S1", "S2", "S3", "S4"),
        status = factor(c("Case", "Case", "Control", "Control")),
        sex = factor(c("F", "M", "F", "M")),
        cg00000029 = c(0.20, 0.25, 0.22, 0.27),
        cg00000108 = c(0.60, 0.55, 0.52, 0.58),
        check.names = FALSE
    )
    save(phenoBT1, file = toy_path)

    result <- methylationGLM(
        inputPheno = toy_path,
        phenotypes = "status",
        covariates = "sex",
        factorVars = "status,sex",
        cpgLimit = 2,
        nCores = 1,
        summaryPval = 1,
        annotationPackage = "IlluminaHumanMethylation450kanno.ilmn12.hg19",
        annotationCols = "Name,chr,pos",
        display = FALSE,
        verbose = FALSE,
        logs = FALSE,
        saveOutputs = FALSE
    )

    class(result)
}

}
\seealso{
\linkS4class{dnaEPICO_methylationGLM}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/result_classes.R
\docType{class}
\name{dnaEPICO_preprocessingMinfiEwasWater-class}
\alias{dnaEPICO_preprocessingMinfiEwasWater-class}
\alias{dnaEPICO_preprocessingMinfiEwasWater}
\title{Result class returned by preprocessingMinfiEwasWater}
\description{
Objects of class \code{'dnaEPICO_preprocessingMinfiEwasWater'} are list-based
results returned by \code{\link[=preprocessingMinfiEwasWater]{preprocessingMinfiEwasWater()}}. They are lightweight
S3-style containers rather than formal S4 classes.
}
\section{Structure}{

\describe{
\item{targets}{Filtered phenotype table aligned to the retained samples.}
\item{RGSet}{Filtered \code{RGChannelSet} used in downstream preprocessing.}
\item{rawData}{Object returned by \code{\link[=buildRawMinfiEwasWater]{buildRawMinfiEwasWater()}}.}
\item{assessment}{Object returned by \code{\link[=assessSamplesMinfiEwasWater]{assessSamplesMinfiEwasWater()}}.}
\item{sexData}{Object returned by \code{\link[=predictSexMinfiEwasWater]{predictSexMinfiEwasWater()}}.}
\item{normData}{Object returned by \code{\link[=normalizeMinfiEwasWater]{normalizeMinfiEwasWater()}}.}
\item{filterData}{Object returned by \code{\link[=filterProbesMinfiEwasWater]{filterProbesMinfiEwasWater()}}.}
\item{metricsData}{Object returned by \code{\link[=extractMetricsMinfiEwasWater]{extractMetricsMinfiEwasWater()}}.}
\item{lcData}{Object returned by \code{\link[=estimateLCMinfiEwasWater]{estimateLCMinfiEwasWater()}}.}
\item{logFile}{Resolved path to the optional log file, or \code{NULL} when
logging was disabled.}
}
}

\seealso{
\code{\link[=preprocessingMinfiEwasWater]{preprocessingMinfiEwasWater()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methylationLME_steps.R
\name{summarizeMethylationLMEModels}
\alias{summarizeMethylationLMEModels}
\title{Summarize CpG-wise mixed-effects model fits for longitudinal analyses}
\usage{
summarizeMethylationLMEModels(
  modelResults,
  preparedData,
  summaryPval = NA,
  excludeSingular = FALSE,
  excludeNonConverged = FALSE,
  nCores = 1L,
  chunkSize = NULL,
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_methylationLME.txt"
)
}
\arguments{
\item{modelResults}{Object returned by \code{fitMethylationLMEModels()}.}

\item{preparedData}{Object returned by \code{prepareMethylationLMEData()}.}

\item{summaryPval}{Numeric or \code{NA}. Optional p-value filter applied to the
returned summary tables. \code{NA} keeps all rows.}

\item{excludeSingular}{Logical. If \code{TRUE}, retain singular CpG rows and
reasons but set their inferential p-values to \code{NA}.}

\item{excludeNonConverged}{Logical. If \code{TRUE}, retain non-converged CpG rows
and reasons but set their inferential p-values to \code{NA}.}

\item{nCores}{Integer. Number of worker processes to use while extracting
summary rows.}

\item{chunkSize}{Integer or \code{NULL}. Number of CpGs processed per parallel
chunk. \code{NULL} chooses a value automatically.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
A list with class \code{'dnaEPICO_methylationLME_summaries'}
containing the optionally filtered summary tables in \code{summaries} and the
complete CpG-level tables in \code{diagnosticSummaries}. Diagnostics,
annotation, and report output use the complete tables so \code{summaryPval}
does not remove CpGs from those outputs. \code{fitDiagnostics} retains the
phenotype-specific fit status, lme4 singular-fit flag, and warning text for
every attempted CpG.
}
\description{
Extract phenotype-specific fixed-effect tables from the fitted mixed-effects
model object returned by \code{fitMethylationLMEModels()}.
}
\examples{
ex <- dnaEPICO:::exampleMethylationLMEStateDnaEpico()
summary_results <- summarizeMethylationLMEModels(
    modelResults = ex$modelResults,
    preparedData = ex$preparedData,
    summaryPval = NA,
    nCores = 1,
    verbose = FALSE,
    logs = FALSE
)
names(summary_results$summaries)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methylationLME_steps.R
\name{plotMethylationLMEDiagnostics}
\alias{plotMethylationLMEDiagnostics}
\title{Plot longitudinal mixed-effects model diagnostics}
\usage{
plotMethylationLMEDiagnostics(
  modelSummaries,
  preparedData,
  fdrThreshold = 0.05,
  padjmethod = "fdr",
  outputDir = NULL,
  plotWidth = 2000L,
  plotHeight = 1000L,
  plotDPI = 150L,
  display = FALSE,
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_methylationLME.txt"
)
}
\arguments{
\item{modelSummaries}{Object returned by \code{summarizeMethylationLMEModels()}.}

\item{preparedData}{Object returned by \code{prepareMethylationLMEData()}.}

\item{fdrThreshold}{Numeric. False-discovery-rate threshold used to highlight
CpGs in the diagnostic plots.}

\item{padjmethod}{Character. P-value adjustment method passed to
\code{stats::p.adjust()}.}

\item{outputDir}{Character or \code{NULL}. Directory used for TIFF files. When
\code{NULL}, plots are returned in memory only.}

\item{plotWidth}{Integer. TIFF width in pixels when plots are written to
disk.}

\item{plotHeight}{Integer. TIFF height in pixels when plots are written to
disk.}

\item{plotDPI}{Integer. TIFF resolution in DPI when plots are written to
disk.}

\item{display}{Logical. If \code{TRUE}, draw plots on the active graphics device.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
A list with class \code{'dnaEPICO_methylationLME_diagnostic_plots'}
containing the generated \code{ggplot2} objects, genomic inflation factors, and
any saved TIFF file paths.
}
\description{
Create Q-Q and standard-error diagnostic plots from the mixed-effects summary
tables returned by \code{summarizeMethylationLMEModels()}.
}
\examples{
ex <- dnaEPICO:::exampleMethylationLMEStateDnaEpico()
diagnostic_plots <- plotMethylationLMEDiagnostics(
    modelSummaries = ex$modelSummaries,
    preparedData = ex$preparedData,
    display = FALSE,
    verbose = FALSE,
    logs = FALSE
)
names(diagnostic_plots$plots)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/svaEnmix.R
\name{svaEnmix}
\alias{svaEnmix}
\title{Estimate surrogate variables from ENmix control probes}
\usage{
svaEnmix(
  phenoFile = "data/preprocessingMinfiEwasWater/phenoLC.csv",
  rgsetData = "rData/preprocessingMinfiEwasWater/objects/RGSet.RData",
  sepType = NULL,
  outputLogs = "logs",
  nSamples = NA,
  SampleID = "Sample_Name",
  arrayType = "IlluminaHumanMethylationEPICv2",
  annotationVersion = "20a1.hg38",
  SentrixIDColumn = "Sentrix_ID",
  SentrixPositionColumn = "Sentrix_Position",
  ctrlSvaPercVar = 0.9,
  ctrlSvaFlag = 1,
  scriptLabel = "svaEnmix",
  tiffWidth = 2000,
  tiffHeight = 1000,
  tiffRes = 150,
  figureBaseDir = "figures",
  dataBaseDir = "data",
  rBaseDir = "rData",
  display = FALSE,
  verbose = FALSE,
  logs = FALSE,
  saveOutputs = FALSE
)
}
\arguments{
\item{phenoFile}{Character. Path to the phenotype file with cell-composition
data. When \code{saveOutputs = TRUE}, the validated PC columns are appended to
this file using a rollback-safe replacement.}

\item{rgsetData}{Character. Path to a saved \code{RGChannelSet} object. Both
\code{.RData} and \code{.rds} files are supported.}

\item{sepType}{Character or \code{NULL}. Field separator used in \code{phenoFile}. Use
\code{NULL} for a comma-separated file, \code{'\\\\t'} for a tab-delimited file, or
another separator accepted by \code{utils::read.csv()}.}

\item{outputLogs}{Character. Directory used for log files when \code{logs = TRUE}.}

\item{nSamples}{Integer or \code{NA}. Number of rows to keep from the phenotype
table. Use \code{NA} to keep all samples.}

\item{SampleID}{Character. Name of the phenotype column containing sample
identifiers.}

\item{arrayType}{Character. Illumina array identifier assigned to
\code{Biobase::annotation(RGSet)}.}

\item{annotationVersion}{Character. Annotation build assigned to
\code{Biobase::annotation(RGSet)}.}

\item{SentrixIDColumn}{Character. Name of the chip identifier column in the
phenotype data.}

\item{SentrixPositionColumn}{Character. Name of the chip position column in
the phenotype data.}

\item{ctrlSvaPercVar}{Numeric. Proportion of control-probe variance explained
when running \code{ENmix::ctrlsva()}.}

\item{ctrlSvaFlag}{Integer. Control-probe flag passed to \code{ENmix::ctrlsva()}.}

\item{scriptLabel}{Character. Label used to name output folders when
\code{saveOutputs = TRUE}.}

\item{tiffWidth}{Integer. Width of saved TIFF plots in pixels.}

\item{tiffHeight}{Integer. Height of saved TIFF plots in pixels.}

\item{tiffRes}{Integer. Resolution in DPI for saved TIFF plots.}

\item{figureBaseDir}{Character. Base directory used for saved figure outputs
when \code{saveOutputs = TRUE}.}

\item{dataBaseDir}{Character. Base directory used for saved CSV and text
outputs when \code{saveOutputs = TRUE}.}

\item{rBaseDir}{Character. Base directory used for saved \code{.RData} outputs
when \code{saveOutputs = TRUE}.}

\item{display}{Logical. If \code{TRUE}, draw plots on the active graphics device.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.
The default is \code{FALSE}.}

\item{logs}{Logical. If \code{TRUE}, write the same progress messages to
\code{outputLogs}. The default is \code{FALSE}.}

\item{saveOutputs}{Logical. If \code{TRUE}, write the CSV, \code{.RData}, text,
and TIFF outputs to disk. The default is \code{FALSE}.}
}
\value{
A list with class \code{'dnaEPICO_svaEnmix'}.
\describe{
\item{targets}{Phenotype table read from \code{phenoFile} after any optional row
subsetting.}
\item{RGSet}{Loaded \code{RGChannelSet} with sample names realigned to
\code{targets[[SampleID]]}.}
\item{svaData}{Object returned by \code{\link[=estimateSvaEnmixControls]{estimateSvaEnmixControls()}} containing
the surrogate-variable matrix and the control-probe settings used to
estimate it.}
\item{mergedPheno}{Phenotype table returned by \code{\link[=mergeSvaTargetsEnmix]{mergeSvaTargetsEnmix()}}
after the surrogate variables were appended as additional columns.}
\item{analysisData}{Object returned by \code{\link[=analyzeSvaEnmix]{analyzeSvaEnmix()}} containing the
surrogate-variable association models, ANOVA tables, and Sentrix metadata.}
\item{plotFiles}{Named list describing the plot file paths requested for the
SVA figures. When \code{saveOutputs = FALSE}, the entries are typically \code{NULL}.}
\item{savedFiles}{Object returned by \code{\link[=writeSvaEnmixOutputs]{writeSvaEnmixOutputs()}} when
\code{saveOutputs = TRUE}, otherwise \code{NULL}.}
\item{logFile}{Resolved path to the optional log file, or \code{NULL} when
logging was disabled.}
}
See \linkS4class{dnaEPICO_svaEnmix} for a class-level overview.
}
\description{
Read the phenotype table and a saved \code{RGChannelSet}, estimate surrogate
variables from ENmix control probes, analyze their association with Sentrix
chip and position factors, and return a structured in-memory result. CSV,
\code{.RData}, text-summary, and figure outputs are written only when
\code{saveOutputs = TRUE}.
}
\examples{
tmp <- tempdir()
stopifnot(dir.exists(tmp))

if (requireNamespace("minfiData", quietly = TRUE)) {
    ex <- dnaEPICO:::exampleMinfiBaseDataDnaEpico()
    pheno_file <- file.path(tmp, "pheno.csv")
    rgset_path <- file.path(tmp, "RGSet.RData")
    RGSet <- ex$RGSet
    utils::write.csv(ex$targets, pheno_file, row.names = FALSE)
    save(RGSet, file = rgset_path)
    sva_result <- svaEnmix(
        phenoFile = pheno_file,
        rgsetData = rgset_path,
        SampleID = "Sample_Name",
        arrayType = "IlluminaHumanMethylation450k",
        annotationVersion = "ilmn12.hg19",
        SentrixIDColumn = "Sentrix_ID",
        SentrixPositionColumn = "Sentrix_Position",
        outputLogs = file.path(tmp, "logs"),
        figureBaseDir = file.path(tmp, "figures"),
        dataBaseDir = file.path(tmp, "data"),
        rBaseDir = file.path(tmp, "rData"),
        saveOutputs = FALSE
    )
    stopifnot(inherits(sva_result, "dnaEPICO_svaEnmix"))
}

}
\seealso{
\linkS4class{dnaEPICO_svaEnmix}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessingPheno.R
\name{preprocessingPheno}
\alias{preprocessingPheno}
\title{Prepare phenotype and methylation matrices for downstream modeling}
\usage{
preprocessingPheno(
  phenoFile = "data/preprocessingMinfiEwasWater/phenoLC.csv",
  sepType = NULL,
  betaPath =
    "rData/preprocessingMinfiEwasWater/metrics/beta_NomFilt_MSetF_Flt_Rxy_Ds_Rc.RData",
  mPath = "rData/preprocessingMinfiEwasWater/metrics/m_NomFilt_MSetF_Flt_Rxy_Ds_Rc.RData",
  cnPath =
    "rData/preprocessingMinfiEwasWater/metrics/cn_NomFilt_MSetF_Flt_Rxy_Ds_Rc.RData",
  SampleID = "Sample_Name",
  timeVar = "Timepoint",
  timepoints = "1,2",
  combineTimepoints = "1,2",
  methylationScale = "beta",
  outputPheno = "data/preprocessingPheno",
  outputRData = "rData/preprocessingPheno/metrics",
  outputRDataMerge = "rData/preprocessingPheno/mergeData",
  sexColumn = "Sex",
  outputLogs = "logs",
  outputDir = "data/preprocessingPheno",
  verbose = FALSE,
  logs = FALSE,
  saveOutputs = FALSE
)
}
\arguments{
\item{phenoFile}{Character. Path to the phenotype CSV file.}

\item{sepType}{Character or \code{NULL}. Field separator used in \code{phenoFile}. Use
\code{NULL} for a comma-separated file, \code{'\\\\t'} for a tab-delimited file, or
another separator accepted by \code{utils::read.csv()}.}

\item{betaPath}{Character. Path to the saved beta-value object. Both \code{.RData}
and \code{.rds} files are supported.}

\item{mPath}{Character. Path to the saved M-value object. Both \code{.RData} and
\code{.rds} files are supported.}

\item{cnPath}{Character. Path to the saved copy-number object. Both \code{.RData}
and \code{.rds} files are supported.}

\item{SampleID}{Character. Name of the phenotype column containing sample
identifiers used to align phenotype and methylation data.}

\item{timeVar}{Character. Name of the phenotype column containing timepoint
labels.}

\item{timepoints}{Character vector or comma-separated string of timepoints to
retain and split into separate in-memory subsets.}

\item{combineTimepoints}{Character vector or comma-separated string of
timepoints to combine into the longitudinal phenotype-plus-methylation
object.}

\item{methylationScale}{Character. Methylation metric to use in merged
modeling tables. One of \code{'Beta'}, \code{'M'}, or \code{'CN'}, in any combination
of upper- and lower-case letters. The default is \code{'beta'}. Beta values
are always used for Clock Foundation exports.}

\item{outputPheno}{Character. Directory used for saved phenotype CSV files
when \code{saveOutputs = TRUE}.}

\item{outputRData}{Character. Directory used for saved metric \code{.RData} files
when \code{saveOutputs = TRUE}.}

\item{outputRDataMerge}{Character. Directory used for saved merged
phenotype-plus-methylation \code{.RData} files when \code{saveOutputs = TRUE}.}

\item{sexColumn}{Character. Name of the phenotype sex column used when
building Clock Foundation exports.}

\item{outputLogs}{Character. Directory used for log files when \code{logs = TRUE}.}

\item{outputDir}{Character. Directory used for Clock Foundation export files
when \code{saveOutputs = TRUE}.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.
The default is \code{FALSE}.}

\item{logs}{Logical. If \code{TRUE}, write the same progress messages to
\code{outputLogs}. The default is \code{FALSE}.}

\item{saveOutputs}{Logical. If \code{TRUE}, write the CSV, ZIP, and \code{.RData}
outputs to disk. The default is \code{FALSE}.}
}
\value{
A list with class \code{'dnaEPICO_preprocessingPheno'}.
\describe{
\item{pheno}{Phenotype table read from \code{phenoFile}.}
\item{metricsData}{Object returned by \code{\link[=loadMetricsPreprocessingPheno]{loadMetricsPreprocessingPheno()}}
containing the beta-value, M-value, and copy-number matrices loaded from
\code{betaPath}, \code{mPath}, and \code{cnPath}.}
\item{timepointData}{Object returned by
\code{\link[=splitTimepointsPreprocessingPheno]{splitTimepointsPreprocessingPheno()}}
containing per-timepoint phenotype tables and methylation matrices.}
\item{combinedData}{Object returned by
\code{\link[=combineTimepointsPreprocessingPheno]{combineTimepointsPreprocessingPheno()}} containing the merged longitudinal
phenotype-plus-methylation object and the timepoint combination metadata.}
\item{clockFoundation}{Object returned by
\code{\link[=buildClockFoundationInputsPreprocessingPheno]{buildClockFoundationInputsPreprocessingPheno()}} containing the beta table
and phenotype table prepared for Clock Foundation export.}
\item{savedFiles}{Object returned by \code{\link[=writePreprocessingPhenoOutputs]{writePreprocessingPhenoOutputs()}} when
\code{saveOutputs = TRUE}, otherwise \code{NULL}.}
\item{logFile}{Resolved path to the optional log file, or \code{NULL} when
logging was disabled.}
}
See \linkS4class{dnaEPICO_preprocessingPheno} for a class-level overview.
}
\description{
Read the phenotype table and the preprocessed beta, M-value, and copy-number
matrices; align them by sample identifier; split them by timepoint; prepare
combined longitudinal objects for a selected modeling scale; and build Clock
Foundation export tables. The
function returns a structured in-memory result, while files are
written only when \code{saveOutputs = TRUE}.
}
\examples{
tmp <- tempdir()
pheno <- data.frame(
    Sample_Name = c("S1", "S2", "S3"),
    Timepoint = c("1", "1", "2"),
    Sex = c(0, 1, 0),
    stringsAsFactors = FALSE
)
beta <- matrix(
    c(0.10, 0.20, 0.30, 0.40, 0.50, 0.60),
    nrow = 2,
    dimnames = list(c("cg1", "cg2"), pheno$Sample_Name)
)
m <- beta * 10
cn <- beta * 100
pheno_file <- file.path(tmp, "pheno.csv")
beta_path <- file.path(tmp, "beta.RData")
m_path <- file.path(tmp, "m.RData")
cn_path <- file.path(tmp, "cn.RData")
utils::write.csv(pheno, pheno_file, row.names = FALSE)
save(beta, file = beta_path)
save(m, file = m_path)
save(cn, file = cn_path)
result <- preprocessingPheno(
    phenoFile = pheno_file,
    betaPath = beta_path,
    mPath = m_path,
    cnPath = cn_path,
    SampleID = "Sample_Name",
    timeVar = "Timepoint",
    timepoints = "1,2",
    combineTimepoints = "1,2",
    outputPheno = file.path(tmp, "data", "preprocessingPheno"),
    outputRData = file.path(tmp, "rData", "preprocessingPheno", "metrics"),
    outputRDataMerge = file.path(tmp, "rData", "preprocessingPheno", "mergeData"),
    sexColumn = "Sex",
    outputLogs = file.path(tmp, "logs"),
    outputDir = file.path(tmp, "clockFoundation"),
    saveOutputs = FALSE
)
stopifnot(inherits(result, "dnaEPICO_preprocessingPheno"))

}
\seealso{
\linkS4class{dnaEPICO_preprocessingPheno}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/svaEnmix_steps.R
\name{estimateSvaEnmixControls}
\alias{estimateSvaEnmixControls}
\title{Estimate surrogate variables from ENmix control probes}
\usage{
estimateSvaEnmixControls(
  RGSet,
  ctrlSvaPercVar = 0.9,
  ctrlSvaFlag = 1,
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_estimateSvaEnmixControls.txt"
)
}
\arguments{
\item{RGSet}{An \code{RGChannelSet}.}

\item{ctrlSvaPercVar}{Numeric. Proportion of variance explained by control
probes, passed to \code{ENmix::ctrlsva()}.}

\item{ctrlSvaFlag}{Integer. Control-probe flag passed to
\code{ENmix::ctrlsva()}.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
A list with class \code{'dnaEPICO_svaEnmix_sva'} containing the surrogate
variable matrix and the parameters used to estimate it.
}
\description{
Run \code{ENmix::ctrlsva()} on an \code{RGChannelSet} and return the surrogate variable
matrix as an in-memory object.
}
\examples{
\dontshow{if (requireNamespace("minfiData", quietly = TRUE)) withAutoprint(\{ # examplesIf}
ex <- dnaEPICO:::exampleMinfiBaseDataDnaEpico()
sva_data <- estimateSvaEnmixControls(
    RGSet = ex$RGSet,
    ctrlSvaPercVar = 0.5,
    ctrlSvaFlag = 1,
    verbose = FALSE,
    logs = FALSE
)
sva_data$K
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dnamReport.R
\name{dnamReport}
\alias{dnamReport}
\title{Generate a DNA methylation dashboard report}
\usage{
dnamReport(
  outputDir = "reports",
  phenoTab = NULL,
  enmixTab = file.path("figures", "preprocessingMinfiEwasWater", "enmix"),
  qcTab = file.path("figures", "preprocessingMinfiEwasWater", "qc"),
  svaTab = file.path("figures", "svaEnmix"),
  metricTab = file.path("figures", "preprocessingMinfiEwasWater", "metrics"),
  glmTab = NULL,
  lmeTab = NULL,
  logTab = outputDir,
  verbose = FALSE,
  logs = FALSE,
  projectName = "dnaEPICO",
  detPPath = NULL,
  detPThreshold = 0.01,
  cpgDetectionPath = NULL,
  sampleDetectionPath = NULL,
  logoPath = system.file("extdata", "dnaEPICO.svg", package = "dnaEPICO"),
  imagePattern = "\\\\.(png|jpg|jpeg|gif|webp|svg|tif|tiff)$",
  recursive = TRUE
)
}
\arguments{
\item{outputDir}{Character. Directory where the Quarto project is written.}

\item{phenoTab}{Character or \code{NULL}. CSV file shown in the Data tab.
When \code{NULL}, the path is inferred from the Makefile output layout.}

\item{enmixTab}{Character. Directory containing ENmix quality-control
figures.}

\item{qcTab}{Character. Directory containing Quality Control figures.}

\item{svaTab}{Character. Directory containing Batch Effect or SVA figures.}

\item{metricTab}{Character. Directory containing Metrics figures.}

\item{glmTab}{Character or \code{NULL}. XLSX workbook shown in the GLM Analysis
tab.
When \code{NULL}, the path is inferred from the Makefile output layout. Report
sidecars produced by \code{methylationGLM()} are read from this report project's
\code{assets/results/glm_results} directory when available.}

\item{lmeTab}{Character or \code{NULL}. XLSX workbook shown in the LME Analysis
tab.
When \code{NULL}, the path is inferred from the Makefile output layout. Report
sidecars produced by \code{methylationLME()} are read from this report project's
\code{assets/results/lme_results} directory when available.}

\item{logTab}{Character. Directory containing workflow logs shown in the Logs
tab.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages.}

\item{logs}{Logical. If \code{TRUE}, write a report log.}

\item{projectName}{Character. Name used for the generated Quarto project.}

\item{detPPath}{Character or \code{NULL}. RData file containing the detection
P-value matrix object \code{detP}, used to build the quality-control tables.
When \code{NULL}, the path is inferred from the Makefile output layout.}

\item{detPThreshold}{Numeric. Detection P-value threshold used when
summarising the \code{detP} matrix.}

\item{cpgDetectionPath}{Character or \code{NULL}. Optional fallback CpG detection
summary CSV.}

\item{sampleDetectionPath}{Character or \code{NULL}. Optional fallback sample
detection summary CSV.}

\item{logoPath}{Character. Path to the navbar logo. Defaults to the packaged
\code{inst/extdata/dnaEPICO.svg} asset.}

\item{imagePattern}{Character. Regular expression used to identify image
files inside the section directories.}

\item{recursive}{Logical. If \code{TRUE}, search section directories recursively.}
}
\value{
A list with class \code{'dnaEPICO_dnamReport'}.
}
\description{
Generate a DNA methylation dashboard report
}
\examples{
report_root <- file.path(tempdir(), "dnaepico-dnam-report")
pheno_file <- file.path(
    report_root,
    "data",
    "model1",
    "preprocessingMinfiEwasWater",
    "phenoLC.csv"
)
dir.create(dirname(pheno_file), recursive = TRUE, showWarnings = FALSE)
utils::write.csv(
    data.frame(
        UID = c("sample1", "sample2"),
        Timepoint = c(1, 2),
        Sex = c("F", "M")
    ),
    pheno_file,
    row.names = FALSE
)

result <- dnamReport(
    outputDir = file.path(report_root, "reports", "model1"),
    phenoTab = pheno_file,
    enmixTab = file.path(
        report_root,
        "figures",
        "model1",
        "preprocessingMinfiEwasWater",
        "enmix"
    ),
    qcTab = file.path(
        report_root,
        "figures",
        "model1",
        "preprocessingMinfiEwasWater",
        "qc"
    ),
    svaTab = file.path(report_root, "figures", "model1", "svaEnmix"),
    metricTab = file.path(
        report_root,
        "figures",
        "model1",
        "preprocessingMinfiEwasWater",
        "metrics"
    ),
    logTab = file.path(report_root, "logs", "model1")
)
result$status

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/svaEnmix_steps.R
\name{plotSvaEnmix}
\alias{plotSvaEnmix}
\title{Plot surrogate variables for svaEnmix}
\usage{
plotSvaEnmix(
  analysisData,
  plot = c("sentrix_id", "sentrix_position", "matrix"),
  display = FALSE,
  file = NULL,
  width = 2000L,
  height = 1000L,
  res = 150L,
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_plotSvaEnmix.txt"
)
}
\arguments{
\item{analysisData}{Object returned by \code{analyzeSvaEnmix()}.}

\item{plot}{Character. Plot type: \code{'sentrix_id'}, \code{'sentrix_position'}, or
\code{'matrix'}.}

\item{display}{Logical. If \code{TRUE}, draw the plot on the active graphics
device.}

\item{file}{Character or \code{NULL}. TIFF file path used for saved output.}

\item{width}{Integer. Plot width in pixels when \code{file} is supplied.}

\item{height}{Integer. Plot height in pixels when \code{file} is supplied.}

\item{res}{Integer. TIFF resolution in DPI when \code{file} is supplied.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
Invisibly returns saved TIFF path(s), otherwise \code{NULL}.
}
\description{
Draw one of the standard surrogate-variable plots used by \code{svaEnmix()}.
}
\examples{
ex <- dnaEPICO:::exampleSvaAnalysisStateDnaEpico()
plotSvaEnmix(
    analysisData = ex$analysisData,
    plot = "sentrix_id",
    display = FALSE,
    verbose = FALSE,
    logs = FALSE
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methylationGLM_steps.R
\name{prepareMethylationGLMData}
\alias{prepareMethylationGLMData}
\title{Prepare phenotype-plus-methylation data for one-timepoint GLM analyses}
\usage{
prepareMethylationGLMData(
  inputPheno,
  phenotypes,
  covariates,
  factorVars,
  scaleVars = NULL,
  cpgPrefix = "cg",
  cpgLimit = NA,
  methylationScale = "beta",
  interactionTerm = NULL,
  prsMap = NULL,
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_methylationGLM.txt"
)
}
\arguments{
\item{inputPheno}{Character. Path to the merged phenotype-plus-methylation
object
created by \code{preprocessingPheno()}.}

\item{phenotypes}{Character vector or comma-separated string of phenotype
variables to model.}

\item{covariates}{Character vector or comma-separated string of covariate
variables to adjust for.}

\item{factorVars}{Character vector or comma-separated string of variables
that
should be converted to factors before modeling.}

\item{scaleVars}{Character vector, comma-separated variable names, or \code{NULL}.
Numeric fixed-effect variables to standardize before model fitting.}

\item{cpgPrefix}{Character. Prefix used to identify methylation columns.}

\item{cpgLimit}{Integer or \code{NA}. Maximum number of CpGs to retain. \code{NA}
keeps all matching CpGs.}

\item{methylationScale}{Character. Methylation metric represented by the CpG
columns. One of \code{'Beta'}, \code{'M'}, or \code{'CN'}, case-insensitively.}

\item{interactionTerm}{Character or \code{NULL}. Optional interaction term.}

\item{prsMap}{Character vector or comma-separated string of phenotype-to-PRS
mappings in the form \code{'Phenotype:PRS'}.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
A list with class \code{'dnaEPICO_methylationGLM_data'} containing the
prepared analysis data, parsed variable selections, CpG columns, and
exploratory summaries.
}
\description{
Load the merged phenotype-plus-methylation input object, validate the
requested
modeling variables, convert selected variables to factors, and return a
single in-memory object for downstream helpers.
}
\examples{
ex <- dnaEPICO:::exampleMethylationGLMStateDnaEpico()
prepared_data <- prepareMethylationGLMData(
    inputPheno = ex$inputPath,
    phenotypes = "status",
    covariates = "sex,age",
    factorVars = "status,sex",
    cpgLimit = 2,
    verbose = FALSE,
    logs = FALSE
)
names(prepared_data)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/result_classes.R
\docType{class}
\name{dnaEPICO_dnamReport_render-class}
\alias{dnaEPICO_dnamReport_render-class}
\alias{dnaEPICO_dnamReport_render}
\title{Result class returned by renderDnamReport}
\description{
Objects of class \code{'dnaEPICO_dnamReport_render'} are list-based results
returned by \code{\link[=renderDnamReport]{renderDnamReport()}}. They describe whether a prepared report was
rendered, skipped, or failed.
}
\section{Structure}{

\describe{
\item{preparedReport}{The input object supplied to \code{\link[=renderDnamReport]{renderDnamReport()}}.}
\item{status}{Render status string such as \code{'rendered'}, \code{'skipped'}, or
\code{'failed'}.}
\item{renderedFile}{Normalized path to the rendered HTML file when rendering
succeeded, otherwise \code{NULL}.}
\item{errorMessage}{Render error or skip message when available, otherwise
\code{NULL}.}
\item{logFile}{Resolved path to the optional log file, or \code{NULL} when
logging was disabled.}
}
}

\seealso{
\code{\link[=renderDnamReport]{renderDnamReport()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dnaEPICO-package.R
\docType{package}
\name{dnaEPICO}
\alias{dnaEPICO-package}
\alias{dnaEPICO}
\title{dnaEPICO: DNA methylation preprocessing and modeling workflows}
\description{
The \code{dnaEPICO} package provides a structured workflow for preprocessing and
analyzing Illumina DNA methylation array data, including quality control,
normalization, cell-type estimation, surrogate-variable analysis, phenotype
preparation, CpG-wise generalised linear models, longitudinal mixed-effects
models using \code{lmerTest}/\code{lme4} or \code{nlme} with optional residual correlation
structures, and interactive web reporting.
}
\details{
The package supports two complementary usage styles:
\itemize{
\item interactive use, where functions return structured in-memory result
objects for inspection and composition; and
\item file-based pipeline use, where the same functions can write logs,
plots, tables, and serialized objects when \code{saveOutputs = TRUE}.
}

The main high-level entry points are:
\itemize{
\item \code{\link[=preprocessingMinfiEwasWater]{preprocessingMinfiEwasWater()}}
\item \code{\link[=svaEnmix]{svaEnmix()}}
\item \code{\link[=preprocessingPheno]{preprocessingPheno()}}
\item \code{\link[=methylationGLM]{methylationGLM()}}
\item \code{\link[=methylationLME]{methylationLME()}}
\item \code{\link[=dnamReport]{dnamReport()}}
}
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/paulYRP/dnaEPICO}
  \item Report bugs at \url{https://github.com/paulYRP/dnaEPICO/issues}
}

}
\author{
\strong{Maintainer}: Paul Ruiz \email{ruizpint@qut.edu.au} (\href{https://orcid.org/0009-0007-6714-3566}{ORCID})

}
\keyword{package}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methylationGLM_steps.R
\name{writeMethylationGLMOutputs}
\alias{writeMethylationGLMOutputs}
\title{Write optional disk outputs for one-timepoint GLM analyses}
\usage{
writeMethylationGLMOutputs(
  modelResults,
  modelSummaries,
  annotatedResults,
  significantCpGs = NULL,
  outputRData,
  summaryTxtDir,
  significantCpGDir,
  annotatedGLMOut,
  reportAssetsDir = NULL,
  saveTxtSummaries = TRUE,
  saveSignificantCpGs = FALSE,
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_methylationGLM.txt"
)
}
\arguments{
\item{modelResults}{Object returned by \code{fitMethylationGLMModels()}.}

\item{modelSummaries}{Object returned by \code{summarizeMethylationGLMModels()}.}

\item{annotatedResults}{Object returned by
\code{annotateMethylationGLMSummaries()} or a compatible data frame.}

\item{significantCpGs}{Object returned by
\code{collectSignificantCpGsMethylationGLM()} or \code{NULL}.}

\item{outputRData}{Character. Directory used for serialized model and summary
outputs.}

\item{summaryTxtDir}{Character. Directory used for tab-delimited summary
tables.}

\item{significantCpGDir}{Character. Directory used for significant-CpG
coefficient tables.}

\item{annotatedGLMOut}{Character. Directory used for the annotated summary
XLSX workbook.}

\item{reportAssetsDir}{Character or \code{NULL}. Report results directory used for
the compressed TSV table and compact metadata sidecars.}

\item{saveTxtSummaries}{Logical. If \code{TRUE}, write tab-delimited summary
tables.}

\item{saveSignificantCpGs}{Logical. If \code{TRUE}, write significant-CpG
coefficient tables.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
A list with class \code{'dnaEPICO_methylationGLM_paths'} containing
the paths of the files written to disk, including the annotated workbook
and any requested report-table sidecars.
}
\description{
Write optional serialized outputs, summary tables, significant-CpG tables,
and annotated results from the one-timepoint GLM workflow.
}
\examples{
ex <- dnaEPICO:::exampleMethylationGLMStateDnaEpico()
annotation_data <- annotateMethylationGLMSummaries(
    modelSummaries = ex$modelSummaries,
    annotationObject = ex$annotationData,
    annotationCols = "Name,chr,pos",
    verbose = FALSE,
    logs = FALSE
)
significant_cpgs <- collectSignificantCpGsMethylationGLM(
    modelResults = ex$modelResults,
    pvalThreshold = 1,
    verbose = FALSE,
    logs = FALSE
)
output_paths <- writeMethylationGLMOutputs(
    modelResults = ex$modelResults,
    modelSummaries = ex$modelSummaries,
    annotatedResults = annotation_data,
    significantCpGs = significant_cpgs,
    outputRData = file.path(ex$tempDir, "models"),
    summaryTxtDir = file.path(ex$tempDir, "summary"),
    significantCpGDir = file.path(ex$tempDir, "significant"),
    annotatedGLMOut = file.path(ex$tempDir, "annotated"),
    saveTxtSummaries = TRUE,
    saveSignificantCpGs = TRUE,
    verbose = FALSE,
    logs = FALSE
)
names(output_paths)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methylationGLM_steps.R
\name{plotMethylationGLMDistributions}
\alias{plotMethylationGLMDistributions}
\title{Plot phenotype and covariate distributions for one-timepoint GLM analyses}
\usage{
plotMethylationGLMDistributions(
  preparedData,
  plotWidth = 2000L,
  plotHeight = 1000L,
  plotDPI = 150L,
  outputDir = NULL,
  display = FALSE,
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_methylationGLM.txt"
)
}
\arguments{
\item{preparedData}{Object returned by \code{prepareMethylationGLMData()}.}

\item{plotWidth}{Integer. TIFF width in pixels when plots are written to
disk.}

\item{plotHeight}{Integer. TIFF height in pixels when plots are written to
disk.}

\item{plotDPI}{Integer. TIFF resolution in DPI when plots are written to
disk.}

\item{outputDir}{Character or \code{NULL}. Directory used for TIFF files. When
\code{NULL}, plots are returned in memory only.}

\item{display}{Logical. If \code{TRUE}, draw plots on the active graphics device.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
A list with class \code{'dnaEPICO_methylationGLM_distribution_plots'}
containing the generated \code{ggplot2} objects and any saved TIFF file paths.
}
\description{
Create phenotype, factor-variable, and numeric-covariate distribution plots
from the object returned by \code{prepareMethylationGLMData()}.
}
\examples{
ex <- dnaEPICO:::exampleMethylationGLMStateDnaEpico()
distribution_plots <- plotMethylationGLMDistributions(
    preparedData = ex$preparedData,
    display = FALSE,
    verbose = FALSE,
    logs = FALSE
)
names(distribution_plots)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estimateLCMinfiEwasWater.R
\name{writePhenoLCMinfiEwasWater}
\alias{writePhenoLCMinfiEwasWater}
\title{Write the merged phenotype plus cell-composition table}
\usage{
writePhenoLCMinfiEwasWater(
  lcData,
  file,
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_writePhenoLCMinfiEwasWater.txt"
)
}
\arguments{
\item{lcData}{Object returned by \code{estimateLCMinfiEwasWater()}.}

\item{file}{Character. Path to the CSV file to write.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
Invisibly returns \code{file}.
}
\description{
Write the merged phenotype plus cell-composition table
}
\examples{
ref_file <- system.file("extdata", "saliva.txt", package = "dnaEPICO")
beta <- as.matrix(utils::read.table(ref_file))[1:20, , drop = FALSE]
colnames(beta) <- c("sample1", "sample2")
targets <- data.frame(
    Sample_Name = colnames(beta),
    Timepoint = c("T1", "T2"),
    stringsAsFactors = FALSE
)
lc_data <- estimateLCMinfiEwasWater(
    beta = beta,
    targets = targets,
    lcRef = "saliva",
    phenoOrder = "Sample_Name;Timepoint"
)
output_file <- file.path(tempdir(), "phenoLC.csv")
writePhenoLCMinfiEwasWater(lcData = lc_data, file = output_file)
file.exists(output_file)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessingMinfiEwasWater_steps.R
\name{plotAssessmentMinfiEwasWater}
\alias{plotAssessmentMinfiEwasWater}
\title{Plot quality-assessment outputs for preprocessingMinfiEwasWater}
\usage{
plotAssessmentMinfiEwasWater(
  assessment,
  plot = c("qc", "detection"),
  display = FALSE,
  file = NULL,
  width = 2000L,
  height = 1000L,
  res = 150L,
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_plotAssessmentMinfiEwasWater.txt"
)
}
\arguments{
\item{assessment}{Object returned by \code{assessSamplesMinfiEwasWater()}.}

\item{plot}{Character. Plot type: \code{'qc'} or \code{'detection'}.}

\item{display}{Logical. If \code{TRUE}, draw the plot on the active graphics
device.}

\item{file}{Character or \code{NULL}. TIFF file written when supplied.}

\item{width}{Integer. TIFF width in pixels when \code{file} is supplied.}

\item{height}{Integer. TIFF height in pixels when \code{file} is supplied.}

\item{res}{Integer. TIFF resolution in DPI when \code{file} is supplied.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
Invisibly returns the saved TIFF path when \code{file} is supplied,
otherwise \code{NULL}.
}
\description{
Draw either the minfi QC plot or the detection P-value plot from an
assessment object returned by \code{assessSamplesMinfiEwasWater()}.
}
\examples{
assessment <- list(
    meanDetP = c(S1 = 0.01, S2 = 0.02, S3 = 0.04),
    detPThreshold = 0.05
)
plotAssessmentMinfiEwasWater(
    assessment = assessment,
    plot = "detection",
    display = FALSE,
    verbose = FALSE,
    logs = FALSE
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessingMinfiEwasWater_steps.R
\name{plotCtrlMinfiEwasWater}
\alias{plotCtrlMinfiEwasWater}
\title{Plot ENmix control images from an RGSet}
\usage{
plotCtrlMinfiEwasWater(
  RGSet,
  output_dir = NULL,
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_plotCtrlMinfiEwasWater.txt"
)
}
\arguments{
\item{RGSet}{An \code{RGChannelSet}.}

\item{output_dir}{Character or \code{NULL}. Directory where ENmix control JPG
files should be written. If \code{NULL}, the function returns without writing
files.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
Invisibly returns \code{output_dir}.
}
\description{
Call \code{ENmix::plotCtrl()} for a supplied \code{RGSet}. This function only writes
files when \code{output_dir} is provided because \code{ENmix::plotCtrl()} produces JPG
files on disk rather than returning a plot object.
}
\examples{
\dontshow{if (requireNamespace("minfiData", quietly = TRUE)) withAutoprint(\{ # examplesIf}
ex <- dnaEPICO:::exampleMinfiBaseDataDnaEpico()
output_dir <- file.path(tempdir(), "enmix-control-plots")
plotCtrlMinfiEwasWater(
    RGSet = ex$RGSet,
    output_dir = output_dir,
    verbose = FALSE,
    logs = FALSE
)
dir.exists(output_dir)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/result_classes.R
\docType{class}
\name{dnaEPICO_methylationGLM-class}
\alias{dnaEPICO_methylationGLM-class}
\alias{dnaEPICO_methylationGLM}
\title{Result class returned by methylationGLM}
\description{
Objects of class \code{'dnaEPICO_methylationGLM'} are list-based results
returned by \code{\link[=methylationGLM]{methylationGLM()}}. They collect the prepared analysis table,
fitted models, summaries, diagnostics, annotations, and optional saved files.
}
\section{Structure}{

\describe{
\item{preparedData}{Object returned by \code{\link[=prepareMethylationGLMData]{prepareMethylationGLMData()}}.}
\item{distributionPlots}{Object returned by
\code{\link[=plotMethylationGLMDistributions]{plotMethylationGLMDistributions()}}.}
\item{modelFits}{Object returned by \code{\link[=fitMethylationGLMModels]{fitMethylationGLMModels()}}.}
\item{modelSummaries}{Object returned by
\code{\link[=summarizeMethylationGLMModels]{summarizeMethylationGLMModels()}}.}
\item{significantCpGs}{Object returned by
\code{\link[=collectSignificantCpGsMethylationGLM]{collectSignificantCpGsMethylationGLM()}}.}
\item{diagnosticPlots}{Object returned by
\code{\link[=plotMethylationGLMDiagnostics]{plotMethylationGLMDiagnostics()}}.}
\item{annotation}{Object returned by
\code{\link[=annotateMethylationGLMSummaries]{annotateMethylationGLMSummaries()}}.}
\item{savedFiles}{Object returned by \code{\link[=writeMethylationGLMOutputs]{writeMethylationGLMOutputs()}} when
\code{saveOutputs = TRUE}, otherwise \code{NULL}.}
}
}

\seealso{
\code{\link[=methylationGLM]{methylationGLM()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessingMinfiEwasWater_steps.R
\name{buildRawMinfiEwasWater}
\alias{buildRawMinfiEwasWater}
\title{Build raw minfi preprocessing objects}
\usage{
buildRawMinfiEwasWater(
  RGSet,
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_buildRawMinfiEwasWater.txt"
)
}
\arguments{
\item{RGSet}{An \code{RGChannelSet}.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
A list with class \code{'dnaEPICO_minfiEwasWater_raw'} containing \code{MSet},
\code{RatioSet}, and \code{GSet}.
}
\description{
Create a raw \code{MethylSet}, \code{RatioSet}, and genome-mapped object from an
\code{RGChannelSet}, and return them together in a single structured object.
}
\examples{
\dontshow{if (requireNamespace("minfiData", quietly = TRUE)) withAutoprint(\{ # examplesIf}
ex <- dnaEPICO:::exampleMinfiBaseDataDnaEpico()
raw_data <- buildRawMinfiEwasWater(
    RGSet = ex$RGSet,
    verbose = FALSE,
    logs = FALSE
)
names(raw_data)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estimateLCMinfiEwasWater.R
\name{estimateLCMinfiEwasWater}
\alias{estimateLCMinfiEwasWater}
\title{Estimate cell composition for preprocessingMinfiEwasWater}
\usage{
estimateLCMinfiEwasWater(
  beta,
  targets,
  SampleID = "Sample_Name",
  lcRef = "salivaEPIC",
  phenoOrder = "Sample_Name;Timepoint;Sex;PredSex;Basename;Sentrix_ID;Sentrix_Position",
  constrained = FALSE,
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_estimateLCMinfiEwasWater.txt"
)
}
\arguments{
\item{beta}{Numeric matrix of beta values with probes in rows and samples in
columns.}

\item{targets}{Phenotype data frame aligned with the columns of \code{beta}.}

\item{SampleID}{Character. Phenotype column containing sample identifiers.}

\item{lcRef}{Character. Cell-composition reference. Internal saliva-based
references supported through \code{estimateLC()} are \code{'saliva'} and
\code{'salivaEPIC'}. Other references are passed to
\code{ENmix::estimateCellProp()}.}

\item{phenoOrder}{Character vector or semicolon-separated string describing
the phenotype columns that should appear first in the merged \code{phenoLC}
output.}

\item{constrained}{Logical. Passed to \code{estimateLC()} when an internal saliva
reference is used. If \code{TRUE}, estimated proportions are constrained to sum
to one.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
A list with class \code{'dnaEPICO_minfiEwasWater_lc'} containing the cell
proportion matrix, merged phenotype table, reference name, and method used.
}
\description{
Estimate cell proportions from beta values using \code{estimateLC()} for saliva
reference panels or \code{ENmix::estimateCellProp()} for other supported
references, then merge the estimates into the phenotype table.
}
\examples{
ref_file <- system.file("extdata", "saliva.txt", package = "dnaEPICO")
beta <- as.matrix(utils::read.table(ref_file))[1:20, , drop = FALSE]
colnames(beta) <- c("sample1", "sample2")
targets <- data.frame(
    Sample_Name = colnames(beta),
    Timepoint = c("T1", "T2"),
    stringsAsFactors = FALSE
)
lc_data <- estimateLCMinfiEwasWater(
    beta = beta,
    targets = targets,
    lcRef = "saliva",
    phenoOrder = "Sample_Name;Timepoint"
)
stopifnot(is.data.frame(lc_data$phenoLC))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methylationGLM_steps.R
\name{fitMethylationGLMModels}
\alias{fitMethylationGLMModels}
\title{Fit CpG-wise Gaussian GLMs for one-timepoint methylation analyses}
\usage{
fitMethylationGLMModels(
  preparedData,
  nCores = 1L,
  libPath = NULL,
  glmLibs = "glm2",
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_methylationGLM.txt"
)
}
\arguments{
\item{preparedData}{Object returned by \code{prepareMethylationGLMData()}.}

\item{nCores}{Integer. Maximum number of worker processes to use. Automatic
fitting remains serial below the empirical glm2 crossover and caps workers
by available CpGs and detected physical cores.}

\item{libPath}{Character vector or \code{NULL}. Optional library paths forwarded
to worker processes.}

\item{glmLibs}{Character vector or comma-separated string of package names to
check on worker processes. The default is \code{'glm2'}.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
A list with class \code{'dnaEPICO_methylationGLM_models'} containing
fitted model lists, model formulas, and counts of failed CpG fits.
}
\description{
Fit one Gaussian GLM per CpG for each phenotype requested in the object
returned by \code{prepareMethylationGLMData()}.
}
\examples{
ex <- dnaEPICO:::exampleMethylationGLMStateDnaEpico()
model_results <- fitMethylationGLMModels(
    preparedData = ex$preparedData,
    nCores = 1,
    verbose = FALSE,
    logs = FALSE
)
names(model_results$fits)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessingMinfiEwasWater.R
\name{preprocessingMinfiEwasWater}
\alias{preprocessingMinfiEwasWater}
\title{Convenience preprocessing pipeline for Illumina methylation arrays}
\usage{
preprocessingMinfiEwasWater(
  phenoFile = "data/preprocessingMinfiEwasWater/pheno.csv",
  idatFolder = "data/preprocessingMinfiEwasWater/idats",
  outputLogs = "logs",
  nSamples = NA,
  SampleID = "Sample_Name",
  arrayType = "IlluminaHumanMethylationEPICv2",
  annotationVersion = "20a1.hg38",
  idatForce = FALSE,
  scriptLabel = "preprocessingMinfiEwasWater",
  baseDataFolder = "rData",
  figureBaseDir = "figures",
  sepType = NULL,
  tiffWidth = 2000,
  tiffHeight = 1000,
  tiffRes = 150,
  qcCutoff = 10.5,
  detPtype = "m+u",
  detPThreshold = 0.05,
  normMethods = "adjustedfunnorm",
  sexColumn = "Sex",
  removeSexMismatch = FALSE,
  pvalThreshold = 0.01,
  chrToRemove = "chrX,chrY",
  snpsToRemove = "SBE,CpG",
  mafThreshold = 0.1,
  probeExclusionPath = "data/preprocessingMinfiEwasWater/12864_2024_10027_MOESM8_ESM.csv",
  probeExclusionIdColumn = NULL,
  useEpicV2Manifest = FALSE,
  epicV2ManifestFlags = c(CH_WGBS_evidence = TRUE, CH_BLAT = TRUE, MissingPos = TRUE,
    MismatchPos = FALSE),
  plotGroupVar = "Sex",
  lcRef = "salivaEPIC",
  phenoOrder = "Sample_Name;Timepoint;Sex;PredSex;Basename;Sentrix_ID;Sentrix_Position",
  lcPhenoDir = "data/preprocessingMinfiEwasWater",
  display = FALSE,
  verbose = FALSE,
  logs = FALSE,
  saveOutputs = FALSE,
  crossReactivePath = NULL,
  crossReactiveIdColumn = NULL
)
}
\arguments{
\item{phenoFile}{Character. Path to the phenotype CSV file.}

\item{idatFolder}{Character. Directory containing the IDAT files.}

\item{outputLogs}{Character. Directory used for log files when \code{logs = TRUE}.}

\item{nSamples}{Integer or \code{NA}. Number of rows to keep from the phenotype
table. Use \code{NA} to keep all samples.}

\item{SampleID}{Character. Name of the phenotype column containing sample
identifiers.}

\item{arrayType}{Character. Illumina array identifier passed to
\code{Biobase::annotation()}, for example \code{'IlluminaHumanMethylationEPICv2'}.}

\item{annotationVersion}{Character. Annotation build passed to
\code{Biobase::annotation()}, for example \code{'20a1.hg38'} or \code{'ilmn12.hg19'}.}

\item{idatForce}{Logical. Passed to \code{minfi::read.metharray.exp()} when
reading IDAT files. Use \code{TRUE} only after confirming that the selected IDAT
files should be read together.}

\item{scriptLabel}{Character. Label used to name output folders when
\code{saveOutputs = TRUE}.}

\item{baseDataFolder}{Character. Base directory used for saved \code{.RData}
outputs when \code{saveOutputs = TRUE}.}

\item{figureBaseDir}{Character. Base directory used for saved figure outputs
when \code{saveOutputs = TRUE}.}

\item{sepType}{Character or \code{NULL}. Field separator used in \code{phenoFile}. Use
\code{NULL} for a comma-separated file, \code{'\\\\t'} for a tab-delimited file, or
another separator accepted by \code{utils::read.csv()}.}

\item{tiffWidth}{Integer. Width of saved TIFF plots in pixels.}

\item{tiffHeight}{Integer. Height of saved TIFF plots in pixels.}

\item{tiffRes}{Integer. Resolution in DPI for saved TIFF plots.}

\item{qcCutoff}{Numeric. QC cutoff passed to \code{minfi::plotQC()}.}

\item{detPtype}{Character. Detection P-value mode passed to
\code{minfi::detectionP()}. Common values in minfi workflows are \code{'m+u'} and
\code{'negative'}. The default here is \code{'m+u'}.}

\item{detPThreshold}{Numeric. Samples with mean detection P value above this
threshold are removed.}

\item{normMethods}{Character vector or semicolon-separated string of
normalization methods. Supported values are \code{'adjustedfunnorm'},
\code{'funnorm'}, \code{'illumina'}, \code{'quantile'}, and \code{'swan'}.}

\item{sexColumn}{Character. Name of the phenotype column containing reported
sex.}

\item{removeSexMismatch}{Logical. If \code{TRUE}, remove samples whose reported
sex and methylation-predicted sex are both known and disagree. Missing or
unknown sex values are retained. The default is \code{FALSE}.}

\item{pvalThreshold}{Numeric. Probe-level detection P-value threshold used in
the probe filter.}

\item{chrToRemove}{Character vector or comma-separated string of chromosome
names to remove, for example \code{'chrX,chrY'}.}

\item{snpsToRemove}{Character vector or comma-separated string of SNP probe
types to remove, for example \code{'SBE,CpG'}.}

\item{mafThreshold}{Numeric. Minor allele frequency threshold passed to
\code{minfi::dropLociWithSnps()}.}

\item{probeExclusionPath}{Character vector or semicolon-separated string of
CSV files containing probe IDs to remove.}

\item{probeExclusionIdColumn}{Character or \code{NULL}. Column containing probe
IDs. When \code{NULL} or \code{''}, each file is auto-detected using \code{ProbeID},
\code{TargetID}, \code{IlmnID}, or \code{Name}, then falling back to an unlabeled or
probe-like first column.}

\item{useEpicV2Manifest}{Logical. If \code{TRUE}, also remove EPICv2 probes
flagged in the Peters et al. expanded manifest from AnnotationHub resource
\code{AH116484}.}

\item{epicV2ManifestFlags}{Named logical vector controlling which EPICv2
manifest flags are removed. Defaults remove \code{CH_WGBS_evidence}, \code{CH_BLAT},
and \code{MissingPos}, but not \code{MismatchPos}.}

\item{plotGroupVar}{Character. Phenotype column used for density and MDS
grouping plots.}

\item{lcRef}{Character. Reference panel used for cell composition estimation.
\code{'saliva'} and \code{'salivaEPIC'} use \code{estimateLC()}. Other values are passed
to \code{ENmix::estimateCellProp()}.}

\item{phenoOrder}{Character vector or semicolon-separated string describing
which phenotype columns should appear first in the merged \code{phenoLC} table.}

\item{lcPhenoDir}{Character. Directory used for the saved \code{phenoLC.csv} file
when \code{saveOutputs = TRUE}.}

\item{display}{Logical. If \code{TRUE}, draw plots on the active graphics device.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.
The default is \code{FALSE}.}

\item{logs}{Logical. If \code{TRUE}, write log messages to \code{outputLogs}. The
default is \code{FALSE}.}

\item{saveOutputs}{Logical. If \code{TRUE}, write the \code{.RData}, figure, and
\code{phenoLC.csv} outputs to disk. The default is \code{FALSE}.}

\item{crossReactivePath}{Deprecated alias for \code{probeExclusionPath}.}

\item{crossReactiveIdColumn}{Deprecated alias for \code{probeExclusionIdColumn}.}
}
\value{
A list with class \code{'dnaEPICO_preprocessingMinfiEwasWater'}.
\describe{
\item{targets}{Filtered phenotype table aligned to the retained samples.}
\item{RGSet}{Filtered \code{RGChannelSet} used in downstream preprocessing and
available for direct interactive inspection.}
\item{rawData}{Object returned by \code{\link[=buildRawMinfiEwasWater]{buildRawMinfiEwasWater()}} containing the
raw \code{MSet}, \code{RatioSet}, and genome-mapped object derived from \code{RGSet}.}
\item{assessment}{Object returned by \code{\link[=assessSamplesMinfiEwasWater]{assessSamplesMinfiEwasWater()}}
containing detection P values, QC summaries, and failed-sample tracking.}
\item{sexData}{Object returned by \code{\link[=predictSexMinfiEwasWater]{predictSexMinfiEwasWater()}} containing
predicted sex labels, mismatch summaries, plotting data, and the sample IDs
removed when \code{removeSexMismatch = TRUE}.}
\item{normData}{Object returned by \code{\link[=normalizeMinfiEwasWater]{normalizeMinfiEwasWater()}} containing
the requested normalized objects and metadata on the methods that were run.}
\item{filterData}{Object returned by \code{\link[=filterProbesMinfiEwasWater]{filterProbesMinfiEwasWater()}}
containing the probe-filtered methylation objects at each filtering stage.}
\item{metricsData}{Object returned by \code{\link[=extractMetricsMinfiEwasWater]{extractMetricsMinfiEwasWater()}}
containing the beta-value, M-value, and copy-number matrices used by later
workflow steps.}
\item{lcData}{Object returned by \code{\link[=estimateLCMinfiEwasWater]{estimateLCMinfiEwasWater()}} containing
the estimated cell-type proportions and the phenotype table augmented with
those proportions.}
\item{logFile}{Resolved path to the optional log file, or \code{NULL} when
logging was disabled.}
}
See \linkS4class{dnaEPICO_preprocessingMinfiEwasWater} for a class-level overview.
}
\description{
Run the \code{dnaEPICO} preprocessing workflow as a convenience wrapper around the
smaller minfi/ENmix/wateRmelon helper functions in this package. The wrapper
returns a structured result object containing the in-memory outputs from each
stage. Files are written only when \code{saveOutputs = TRUE}.
}
\examples{
if (requireNamespace("minfiData", quietly = TRUE) &&
    requireNamespace("IlluminaHumanMethylation450kmanifest", quietly = TRUE) &&
    requireNamespace("IlluminaHumanMethylation450kanno.ilmn12.hg19", quietly = TRUE)) {
    ex <- dnaEPICO:::exampleMinfiIdatInputsDnaEpico(n = 4)
    result <- preprocessingMinfiEwasWater(
        phenoFile = ex$phenoFile,
        idatFolder = ex$idatFolder,
        outputLogs = file.path(ex$tempDir, "logs"),
        nSamples = 4,
        SampleID = "Sample_Name",
        arrayType = ex$arrayType,
        annotationVersion = ex$annotationVersion,
        scriptLabel = "preprocessingMinfiEwasWater",
        baseDataFolder = file.path(ex$tempDir, "rData"),
        figureBaseDir = file.path(ex$tempDir, "figures"),
        detPThreshold = 1,
        normMethods = "quantile",
        sexColumn = "Sex",
        pvalThreshold = 1,
        chrToRemove = "",
        snpsToRemove = "SBE",
        mafThreshold = 1,
        probeExclusionPath = ex$probeExclusionPath,
        plotGroupVar = "Sex",
        lcRef = "saliva",
        phenoOrder = "Sample_Name;Sex;Basename;Sentrix_ID;Sentrix_Position",
        lcPhenoDir = ex$tempDir,
        saveOutputs = FALSE,
        verbose = FALSE,
        logs = FALSE
    )
    inherits(result, "dnaEPICO_preprocessingMinfiEwasWater")
}

}
\seealso{
\linkS4class{dnaEPICO_preprocessingMinfiEwasWater}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessingMinfiEwasWater_steps.R
\name{plotMetricsMinfiEwasWater}
\alias{plotMetricsMinfiEwasWater}
\title{Plot multidimensional scaling or density summaries from final metrics}
\usage{
plotMetricsMinfiEwasWater(
  metricsData,
  targets,
  plot = c("mds", "density"),
  plotGroupVar = "Sex",
  sexColumn = "Sex",
  display = FALSE,
  file = NULL,
  width = 2000L,
  height = 1000L,
  res = 150L,
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_plotMetricsMinfiEwasWater.txt"
)
}
\arguments{
\item{metricsData}{Object returned by \code{extractMetricsMinfiEwasWater()}.}

\item{targets}{Filtered phenotype data aligned with \code{metricsData}.}

\item{plot}{Character. Plot type: \code{'mds'} or \code{'density'}.}

\item{plotGroupVar}{Character. Phenotype column used for the main grouping.}

\item{sexColumn}{Character. Phenotype column used for the sex grouping in the
MDS plot.}

\item{display}{Logical. If \code{TRUE}, draw the plot on the active graphics
device.}

\item{file}{Character or \code{NULL}. TIFF file written when supplied.}

\item{width}{Integer. TIFF width in pixels when \code{file} is supplied.}

\item{height}{Integer. TIFF height in pixels when \code{file} is supplied.}

\item{res}{Integer. TIFF resolution in DPI when \code{file} is supplied.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
Invisibly returns the saved TIFF path when \code{file} is supplied,
otherwise \code{NULL}.
}
\description{
Plot multidimensional scaling or density summaries from final metrics
}
\examples{
ex <- dnaEPICO:::exampleMinfiMetricsStateDnaEpico()
plotMetricsMinfiEwasWater(
    metricsData = ex$metricsData,
    targets = ex$targets,
    plot = "density",
    plotGroupVar = "Sex",
    sexColumn = "Sex",
    display = FALSE,
    verbose = FALSE,
    logs = FALSE
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/svaEnmix_steps.R
\name{mergeSvaTargetsEnmix}
\alias{mergeSvaTargetsEnmix}
\title{Merge surrogate variables into the phenotype table}
\usage{
mergeSvaTargetsEnmix(
  targets,
  sva,
  SampleID = "Sample_Name",
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_mergeSvaTargetsEnmix.txt"
)
}
\arguments{
\item{targets}{Phenotype data frame aligned with the samples in \code{sva}.}

\item{sva}{Numeric matrix of surrogate variables with samples in rows.}

\item{SampleID}{Character. Name of the phenotype sample identifier column.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
A phenotype data frame with the surrogate variables appended.
}
\description{
Merge the surrogate variable matrix back into the phenotype table while
preserving the original row order of \code{targets}.
}
\examples{
ex <- dnaEPICO:::exampleSvaAnalysisStateDnaEpico()
merged_pheno <- mergeSvaTargetsEnmix(
    targets = ex$targets,
    sva = ex$sva,
    SampleID = "Sample_Name",
    verbose = FALSE,
    logs = FALSE
)
colnames(merged_pheno)[seq_len(4)]

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessingPheno_steps.R
\name{splitTimepointsPreprocessingPheno}
\alias{splitTimepointsPreprocessingPheno}
\title{Split phenotype and methylation data by timepoint}
\usage{
splitTimepointsPreprocessingPheno(
  pheno,
  metricsData,
  SampleID = "Sample_Name",
  timeVar = "Timepoint",
  timepoints = "1,2",
  methylationScale = "beta",
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_splitTimepointsPreprocessingPheno.txt"
)
}
\arguments{
\item{pheno}{Data frame containing phenotype information.}

\item{metricsData}{Object returned by \code{loadMetricsPreprocessingPheno()}.}

\item{SampleID}{Character. Name of the sample identifier column in \code{pheno}.}

\item{timeVar}{Character. Name of the timepoint column in \code{pheno}.}

\item{timepoints}{Character vector or comma-separated string of timepoints to
retain.}

\item{methylationScale}{Character. Methylation metric to use for the merged
modeling table. One of \code{'Beta'}, \code{'M'}, or \code{'CN'}, case-insensitively.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
A list with class \code{'dnaEPICO_preprocessingPheno_timepoints'}
containing the parsed timepoints and aligned per-timepoint subsets.
}
\description{
Align phenotype rows and metric matrices for each requested timepoint, and
precompute the per-timepoint phenotype-plus-beta objects used by downstream
modeling functions.
}
\examples{
ex <- dnaEPICO:::examplePreprocessingPhenoStateDnaEpico()
timepoint_data <- splitTimepointsPreprocessingPheno(
    pheno = ex$pheno,
    metricsData = ex$metricsData,
    SampleID = "Sample_Name",
    timeVar = "Timepoint",
    timepoints = "1,2",
    verbose = FALSE,
    logs = FALSE
)
timepoint_data$timepoints

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methylationGLM_steps.R
\name{annotateMethylationGLMSummaries}
\alias{annotateMethylationGLMSummaries}
\title{Annotate one-timepoint GLM summary tables with array annotation metadata}
\usage{
annotateMethylationGLMSummaries(
  modelSummaries,
  annotationObject,
  annotationCols = c("Name", "chr", "pos", "UCSC_RefGene_Group", "UCSC_RefGene_Name",
    "Relation_to_Island", "GencodeV41_Group"),
  verbose = FALSE,
  logs = FALSE,
  log_dir = NULL,
  log_file = "log_methylationGLM.txt"
)
}
\arguments{
\item{modelSummaries}{Object returned by \code{summarizeMethylationGLMModels()}
or a named list of CpG summary data frames.}

\item{annotationObject}{Character package/object name, annotation data frame,
or annotation object understood by \code{minfi::getAnnotation()}.}

\item{annotationCols}{Character vector or comma-separated string of
annotation
columns to append.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages with \code{message()}.}

\item{logs}{Logical. If \code{TRUE}, write the same messages to a log file.}

\item{log_dir}{Character or \code{NULL}. Directory used for the log file when
\code{logs = TRUE}.}

\item{log_file}{Character. File name used when \code{logs = TRUE}.}
}
\value{
A list with class \code{'dnaEPICO_methylationGLM_annotation'}
containing the annotated summary table and any requested annotation columns
that were unavailable in the chosen annotation object.
}
\description{
Merge phenotype-specific CpG summary tables with probe annotation metadata
and
return a single annotated result table.
}
\examples{
ex <- dnaEPICO:::exampleMethylationGLMStateDnaEpico()
annotation_data <- annotateMethylationGLMSummaries(
    modelSummaries = ex$modelSummaries,
    annotationObject = ex$annotationData,
    annotationCols = "Name,chr,pos",
    verbose = FALSE,
    logs = FALSE
)
names(annotation_data)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/result_classes.R
\docType{class}
\name{dnaEPICO_svaEnmix-class}
\alias{dnaEPICO_svaEnmix-class}
\alias{dnaEPICO_svaEnmix}
\title{Result class returned by svaEnmix}
\description{
Objects of class \code{'dnaEPICO_svaEnmix'} are list-based results returned by
\code{\link[=svaEnmix]{svaEnmix()}}. They collect the loaded inputs, surrogate-variable results,
association-analysis summaries, and optional file outputs.
}
\section{Structure}{

\describe{
\item{targets}{Phenotype table read from \code{phenoFile} after any optional row
subsetting.}
\item{RGSet}{Loaded \code{RGChannelSet} with sample names reset to match
\code{targets}.}
\item{svaData}{Object returned by \code{\link[=estimateSvaEnmixControls]{estimateSvaEnmixControls()}}.}
\item{mergedPheno}{Phenotype table returned by \code{\link[=mergeSvaTargetsEnmix]{mergeSvaTargetsEnmix()}}
after surrogate variables were appended.}
\item{analysisData}{Object returned by \code{\link[=analyzeSvaEnmix]{analyzeSvaEnmix()}}.}
\item{plotFiles}{Named list of TIFF output paths for the SVA figures when
\code{saveOutputs = TRUE}, otherwise \code{NULL} entries for plots not written.}
\item{savedFiles}{Object returned by \code{\link[=writeSvaEnmixOutputs]{writeSvaEnmixOutputs()}} when
\code{saveOutputs = TRUE}, otherwise \code{NULL}.}
\item{logFile}{Resolved path to the optional log file, or \code{NULL} when
logging was disabled.}
}
}

\seealso{
\code{\link[=svaEnmix]{svaEnmix()}}
}
